/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (https://www.swig.org).
 * Version 4.2.1
 *
 * Do not make changes to this file unless you know what you are doing - modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */

#include "znc/zncconfig.h"



#define SWIG_VERSION 0x040201
#define SWIGPYTHON
#define SWIG_PYTHON_DIRECTOR_NO_VTABLE

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

#if defined(__cplusplus) && __cplusplus >=201103L
# define SWIG_NULLPTR nullptr
#else
# define SWIG_NULLPTR NULL
#endif 

/* -----------------------------------------------------------------------------
 * swigcompat.swg
 *
 * Macros to provide support compatibility with older C and C++ standards.
 * ----------------------------------------------------------------------------- */

/* C99 and C++11 should provide snprintf, but define SWIG_NO_SNPRINTF
 * if you're missing it.
 */
#if ((defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L) || \
     (defined __cplusplus && __cplusplus >= 201103L) || \
     defined SWIG_HAVE_SNPRINTF) && \
    !defined SWIG_NO_SNPRINTF
# define SWIG_snprintf(O,S,F,A) snprintf(O,S,F,A)
# define SWIG_snprintf2(O,S,F,A,B) snprintf(O,S,F,A,B)
#else
/* Fallback versions ignore the buffer size, but most of our uses either have a
 * fixed maximum possible size or dynamically allocate a buffer that's large
 * enough.
 */
# define SWIG_snprintf(O,S,F,A) sprintf(O,F,A)
# define SWIG_snprintf2(O,S,F,A,B) sprintf(O,F,A,B)
#endif


#if defined(__GNUC__) && defined(_WIN32) && !defined(SWIG_PYTHON_NO_HYPOT_WORKAROUND)
/* Workaround for '::hypot' has not been declared', see https://bugs.python.org/issue11566 */
# include <math.h>
#endif

#if !defined(PY_SSIZE_T_CLEAN) && !defined(SWIG_NO_PY_SSIZE_T_CLEAN)
#define PY_SSIZE_T_CLEAN
#endif

#if __GNUC__ >= 7
#pragma GCC diagnostic push
#if defined(__cplusplus) && __cplusplus >=201703L
#pragma GCC diagnostic ignored "-Wregister" /* For python-2.7 headers that use register */
#endif
#endif

#if defined(_DEBUG) && defined(SWIG_PYTHON_INTERPRETER_NO_DEBUG)
/* Use debug wrappers with the Python release dll */

#if defined(_MSC_VER) && _MSC_VER >= 1929
/* Workaround compilation errors when redefining _DEBUG in MSVC 2019 version 16.10 and later
 * See https://github.com/swig/swig/issues/2090 */
# include <corecrt.h>
#endif

# undef _DEBUG
# include <Python.h>
# define _DEBUG 1
#else
# include <Python.h>
#endif

#if __GNUC__ >= 7
#pragma GCC diagnostic pop
#endif

#include <stdio.h>

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "4"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2
#define SWIG_POINTER_NO_NULL       0x4
#define SWIG_POINTER_CLEAR         0x8
#define SWIG_POINTER_RELEASE       (SWIG_POINTER_CLEAR | SWIG_POINTER_DISOWN)

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows returning the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
/* Runtime errors are < 0 */
#define SWIG_ERROR                 (-1)
/* Errors in range -1 to -99 are in swigerrors.swg (errors for all languages including those not using the runtime) */
/* Errors in range -100 to -199 are language specific errors defined in *errors.swg */
/* Errors < -200 are generic runtime specific errors */
#define SWIG_ERROR_RELEASE_NOT_OWNED (-200)

#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporary objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del object mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next cast in linked list */
  struct swig_cast_info  *prev;			/* pointer to the previous cast */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (strcmp(iter->type->name, c) == 0) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Identical to SWIG_TypeCheck, except strcmp is replaced with a pointer comparison
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(const swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (iter->type == from) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  Choose the last
     name. It should be the most specific; a fully resolved name
     but not necessarily with default template parameters expanded. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *cast = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (cast) {
    if (!cast->converter) {
      swig_type_info *tc = cast->type;
      if (!tc->clientdata) {
	SWIG_TypeClientData(tc, clientdata);
      }
    }
    cast = cast->next;
  }
}
SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/* SWIG Errors applicable to all language modules, values are reserved from -1 to -99 */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13


/* Compatibility macros for Python 3 */
#if PY_VERSION_HEX >= 0x03000000

#define PyClass_Check(obj) PyObject_IsInstance(obj, (PyObject *)&PyType_Type)
#define PyInt_Check(x) PyLong_Check(x)
#define PyInt_AsLong(x) PyLong_AsLong(x)
#define PyInt_FromLong(x) PyLong_FromLong(x)
#define PyInt_FromSize_t(x) PyLong_FromSize_t(x)
#define PyString_Check(name) PyBytes_Check(name)
#define PyString_FromString(x) PyUnicode_FromString(x)
#define PyString_Format(fmt, args)  PyUnicode_Format(fmt, args)
#define PyString_AsString(str) PyBytes_AsString(str)
#define PyString_Size(str) PyBytes_Size(str)	
#define PyString_InternFromString(key) PyUnicode_InternFromString(key)
#define Py_TPFLAGS_HAVE_CLASS Py_TPFLAGS_BASETYPE
#define _PyLong_FromSsize_t(x) PyLong_FromSsize_t(x)

#endif

#ifndef Py_TYPE
#  define Py_TYPE(op) ((op)->ob_type)
#endif

/* SWIG APIs for compatibility of both Python 2 & 3 */

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_FromFormat PyUnicode_FromFormat
#else
#  define SWIG_Python_str_FromFormat PyString_FromFormat
#endif


/* Wrapper around PyUnicode_AsUTF8AndSize - call Py_XDECREF on the returned pbytes when finished with the returned string */
SWIGINTERN const char *
SWIG_PyUnicode_AsUTF8AndSize(PyObject *str, Py_ssize_t *psize, PyObject **pbytes)
{
#if PY_VERSION_HEX >= 0x03030000
# if !defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030A0000
  *pbytes = NULL;
  return PyUnicode_AsUTF8AndSize(str, psize);
# else
   *pbytes = PyUnicode_AsUTF8String(str);
   const char *chars = *pbytes ? PyBytes_AsString(*pbytes) : NULL;
   if (chars && psize)
     *psize = PyBytes_Size(*pbytes);
  return chars;
# endif
#else
  char *chars = NULL;
  *pbytes = NULL;
  PyString_AsStringAndSize(str, &chars, psize);
  return chars;
#endif
}

SWIGINTERN PyObject*
SWIG_Python_str_FromChar(const char *c)
{
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_FromString(c); 
#else
  return PyString_FromString(c);
#endif
}

#ifndef PyObject_DEL
# define PyObject_DEL PyObject_Del
#endif

/* SWIGPY_USE_CAPSULE is no longer used within SWIG itself, but some user interface files check for it. */
# define SWIGPY_USE_CAPSULE
#ifdef SWIGPYTHON_BUILTIN
# define SWIGPY_CAPSULE_ATTR_NAME "type_pointer_capsule_builtin" SWIG_TYPE_TABLE_NAME
#else
# define SWIGPY_CAPSULE_ATTR_NAME "type_pointer_capsule" SWIG_TYPE_TABLE_NAME
#endif
# define SWIGPY_CAPSULE_NAME ("swig_runtime_data" SWIG_RUNTIME_VERSION "." SWIGPY_CAPSULE_ATTR_NAME)

#if PY_VERSION_HEX < 0x03020000
#define PyDescr_TYPE(x) (((PyDescrObject *)(x))->d_type)
#define PyDescr_NAME(x) (((PyDescrObject *)(x))->d_name)
#define Py_hash_t long
#endif

#ifdef Py_LIMITED_API
# define PyTuple_GET_ITEM PyTuple_GetItem
/* Note that PyTuple_SetItem() has different semantics from PyTuple_SET_ITEM as it decref's the original tuple item, so in general they cannot be used
  interchangeably. However in SWIG-generated code PyTuple_SET_ITEM is only used with newly initialized tuples without any items and for them this does work. */
# define PyTuple_SET_ITEM PyTuple_SetItem
# define PyTuple_GET_SIZE PyTuple_Size
# define PyCFunction_GET_FLAGS PyCFunction_GetFlags
# define PyCFunction_GET_FUNCTION PyCFunction_GetFunction
# define PyCFunction_GET_SELF PyCFunction_GetSelf
# define PyList_GET_ITEM PyList_GetItem
# define PyList_SET_ITEM PyList_SetItem
# define PySliceObject PyObject
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIME PyObject*
SWIG_Python_ErrorType(int code) {
  PyObject* type = 0;
  switch(code) {
  case SWIG_MemoryError:
    type = PyExc_MemoryError;
    break;
  case SWIG_IOError:
    type = PyExc_IOError;
    break;
  case SWIG_RuntimeError:
    type = PyExc_RuntimeError;
    break;
  case SWIG_IndexError:
    type = PyExc_IndexError;
    break;
  case SWIG_TypeError:
    type = PyExc_TypeError;
    break;
  case SWIG_DivisionByZero:
    type = PyExc_ZeroDivisionError;
    break;
  case SWIG_OverflowError:
    type = PyExc_OverflowError;
    break;
  case SWIG_SyntaxError:
    type = PyExc_SyntaxError;
    break;
  case SWIG_ValueError:
    type = PyExc_ValueError;
    break;
  case SWIG_SystemError:
    type = PyExc_SystemError;
    break;
  case SWIG_AttributeError:
    type = PyExc_AttributeError;
    break;
  default:
    type = PyExc_RuntimeError;
  }
  return type;
}


SWIGRUNTIME void
SWIG_Python_AddErrorMsg(const char* mesg)
{
  PyObject *type = 0;
  PyObject *value = 0;
  PyObject *traceback = 0;

  if (PyErr_Occurred())
    PyErr_Fetch(&type, &value, &traceback);
  if (value) {
    PyObject *old_str = PyObject_Str(value);
    PyObject *bytes = NULL;
    const char *tmp = SWIG_PyUnicode_AsUTF8AndSize(old_str, NULL, &bytes);
    PyErr_Clear();
    Py_XINCREF(type);
    if (tmp)
      PyErr_Format(type, "%s %s", tmp, mesg);
    else
      PyErr_Format(type, "%s", mesg);
    Py_XDECREF(bytes);
    Py_DECREF(old_str);
    Py_DECREF(value);
  } else {
    PyErr_SetString(PyExc_RuntimeError, mesg);
  }
}

SWIGRUNTIME int
SWIG_Python_TypeErrorOccurred(PyObject *obj)
{
  PyObject *error;
  if (obj)
    return 0;
  error = PyErr_Occurred();
  return error && PyErr_GivenExceptionMatches(error, PyExc_TypeError);
}

SWIGRUNTIME void
SWIG_Python_RaiseOrModifyTypeError(const char *message)
{
  if (SWIG_Python_TypeErrorOccurred(NULL)) {
    /* Use existing TypeError to preserve stacktrace and enhance with given message */
    PyObject *newvalue;
    PyObject *type = NULL, *value = NULL, *traceback = NULL;
    PyErr_Fetch(&type, &value, &traceback);
#if PY_VERSION_HEX >= 0x03000000
    newvalue = PyUnicode_FromFormat("%S\nAdditional information:\n%s", value, message);
#else
    newvalue = PyString_FromFormat("%s\nAdditional information:\n%s", PyString_AsString(value), message);
#endif
    if (newvalue) {
      Py_XDECREF(value);
      PyErr_Restore(type, newvalue, traceback);
    } else {
      PyErr_Restore(type, value, traceback);
    }
  } else {
    /* Raise TypeError using given message */
    PyErr_SetString(PyExc_TypeError, message);
  }
}

#if defined(SWIG_PYTHON_NO_THREADS)
#  if defined(SWIG_PYTHON_THREADS)
#    undef SWIG_PYTHON_THREADS
#  endif
#endif
#if defined(SWIG_PYTHON_THREADS) /* Threading support is enabled */
#  if !defined(SWIG_PYTHON_USE_GIL) && !defined(SWIG_PYTHON_NO_USE_GIL)
#    define SWIG_PYTHON_USE_GIL
#  endif
#  if defined(SWIG_PYTHON_USE_GIL) /* Use PyGILState threads calls */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      if PY_VERSION_HEX < 0x03070000
#        define SWIG_PYTHON_INITIALIZE_THREADS PyEval_InitThreads()
#      else
#        define SWIG_PYTHON_INITIALIZE_THREADS
#      endif
#    endif
#    ifdef __cplusplus /* C++ code */
       class SWIG_Python_Thread_Block {
         bool status;
         PyGILState_STATE state;
       public:
         void end() { if (status) { PyGILState_Release(state); status = false;} }
         SWIG_Python_Thread_Block() : status(true), state(PyGILState_Ensure()) {}
         ~SWIG_Python_Thread_Block() { end(); }
       };
       class SWIG_Python_Thread_Allow {
         bool status;
         PyThreadState *save;
       public:
         void end() { if (status) { status = false; PyEval_RestoreThread(save); }}
         SWIG_Python_Thread_Allow() : status(true), save(PyEval_SaveThread()) {}
         ~SWIG_Python_Thread_Allow() { end(); }
       };
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   SWIG_Python_Thread_Block _swig_thread_block
#      define SWIG_PYTHON_THREAD_END_BLOCK     _swig_thread_block.end()
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   SWIG_Python_Thread_Allow _swig_thread_allow
#      define SWIG_PYTHON_THREAD_END_ALLOW     _swig_thread_allow.end()
#    else /* C code */
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   PyGILState_STATE _swig_thread_block = PyGILState_Ensure()
#      define SWIG_PYTHON_THREAD_END_BLOCK     PyGILState_Release(_swig_thread_block)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   PyThreadState *_swig_thread_allow = PyEval_SaveThread()
#      define SWIG_PYTHON_THREAD_END_ALLOW     PyEval_RestoreThread(_swig_thread_allow)
#    endif
#  else /* Old thread way, not implemented, user must provide it */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      define SWIG_PYTHON_INITIALIZE_THREADS
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_BLOCK)
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_BLOCK)
#      define SWIG_PYTHON_THREAD_END_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_ALLOW)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_ALLOW)
#      define SWIG_PYTHON_THREAD_END_ALLOW
#    endif
#  endif
#else /* No thread support */
#  define SWIG_PYTHON_INITIALIZE_THREADS
#  define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#  define SWIG_PYTHON_THREAD_END_BLOCK
#  define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#  define SWIG_PYTHON_THREAD_END_ALLOW
#endif

/* -----------------------------------------------------------------------------
 * Python API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
 * Constant declarations
 * ----------------------------------------------------------------------------- */

/* Constant Types */
#define SWIG_PY_POINTER 4
#define SWIG_PY_BINARY  5

/* Constant information structure */
typedef struct swig_const_info {
  int type;
  const char *name;
  long lvalue;
  double dvalue;
  void   *pvalue;
  swig_type_info **ptype;
} swig_const_info;

#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * pyrun.swg
 *
 * This file contains the runtime support for Python modules
 * and includes code for managing global variables and pointer
 * type checking.
 *
 * ----------------------------------------------------------------------------- */

#if PY_VERSION_HEX < 0x02070000 /* 2.7.0 */
# error "This version of SWIG only supports Python >= 2.7"
#endif

#if PY_VERSION_HEX >= 0x03000000 && PY_VERSION_HEX < 0x03030000
# error "This version of SWIG only supports Python 3 >= 3.3"
#endif

/* Common SWIG API */

/* for raw pointers */
#define SWIG_Python_ConvertPtr(obj, pptr, type, flags)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Python_ConvertPtr(obj, pptr, type, flags)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, own)

#ifdef SWIGPYTHON_BUILTIN
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(self, ptr, type, flags)
#else
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(NULL, ptr, type, flags)
#endif

#define SWIG_InternalNewPointerObj(ptr, type, flags)	SWIG_Python_NewPointerObj(NULL, ptr, type, flags)

#define SWIG_CheckImplicit(ty)                          SWIG_Python_CheckImplicit(ty) 
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Python_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_Python_ConvertFunctionPtr(obj, pptr, type)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_Python_NewPointerObj(NULL, ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Python_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer)             SWIG_Python_SetModule(pointer)
#define SWIG_NewClientData(obj)                         SwigPyClientData_New(obj)

#define SWIG_SetErrorObj                                SWIG_Python_SetErrorObj                            
#define SWIG_SetErrorMsg                        	SWIG_Python_SetErrorMsg				   
#define SWIG_ErrorType(code)                    	SWIG_Python_ErrorType(code)                        
#define SWIG_Error(code, msg)            		SWIG_Python_SetErrorMsg(SWIG_ErrorType(code), msg) 
#define SWIG_fail                        		goto fail					   


/* Runtime API implementation */

/* Error manipulation */

SWIGINTERN void 
SWIG_Python_SetErrorObj(PyObject *errtype, PyObject *obj) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK; 
  PyErr_SetObject(errtype, obj);
  Py_DECREF(obj);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

SWIGINTERN void 
SWIG_Python_SetErrorMsg(PyObject *errtype, const char *msg) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK;
  PyErr_SetString(errtype, msg);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

#define SWIG_Python_Raise(obj, type, desc)  SWIG_Python_SetErrorObj(SWIG_Python_ExceptionType(desc), obj)

/* Set a constant value */

#if defined(SWIGPYTHON_BUILTIN)

SWIGINTERN void
SwigPyBuiltin_AddPublicSymbol(PyObject *seq, const char *key) {
  PyObject *s = PyString_InternFromString(key);
  PyList_Append(seq, s);
  Py_DECREF(s);
}

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, PyObject *public_interface, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  Py_DECREF(obj);
  if (public_interface)
    SwigPyBuiltin_AddPublicSymbol(public_interface, name);
}

#else

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  Py_DECREF(obj);                            
}

#endif

/* Append a value to the result obj */

SWIGINTERN PyObject*
SWIG_Python_AppendOutput(PyObject* result, PyObject* obj) {
  if (!result) {
    result = obj;
  } else if (result == Py_None) {
    Py_DECREF(result);
    result = obj;
  } else {
    if (!PyList_Check(result)) {
      PyObject *o2 = result;
      result = PyList_New(1);
      if (result) {
        PyList_SET_ITEM(result, 0, o2);
      } else {
        Py_DECREF(obj);
        return o2;
      }
    }
    PyList_Append(result,obj);
    Py_DECREF(obj);
  }
  return result;
}

/* Unpack the argument tuple */

SWIGINTERN Py_ssize_t
SWIG_Python_UnpackTuple(PyObject *args, const char *name, Py_ssize_t min, Py_ssize_t max, PyObject **objs)
{
  if (!args) {
    if (!min && !max) {
      return 1;
    } else {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got none", 
		   name, (min == max ? "" : "at least "), (int)min);
      return 0;
    }
  }  
  if (!PyTuple_Check(args)) {
    if (min <= 1 && max >= 1) {
      Py_ssize_t i;
      objs[0] = args;
      for (i = 1; i < max; ++i) {
	objs[i] = 0;
      }
      return 2;
    }
    PyErr_SetString(PyExc_SystemError, "UnpackTuple() argument list is not a tuple");
    return 0;
  } else {
    Py_ssize_t l = PyTuple_GET_SIZE(args);
    if (l < min) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at least "), (int)min, (int)l);
      return 0;
    } else if (l > max) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at most "), (int)max, (int)l);
      return 0;
    } else {
      Py_ssize_t i;
      for (i = 0; i < l; ++i) {
	objs[i] = PyTuple_GET_ITEM(args, i);
      }
      for (; l < max; ++l) {
	objs[l] = 0;
      }
      return i + 1;
    }    
  }
}

SWIGINTERN int
SWIG_Python_CheckNoKeywords(PyObject *kwargs, const char *name) {
  int no_kwargs = 1;
  if (kwargs) {
    assert(PyDict_Check(kwargs));
    if (PyDict_Size(kwargs) > 0) {
      PyErr_Format(PyExc_TypeError, "%s() does not take keyword arguments", name);
      no_kwargs = 0;
    }
  }
  return no_kwargs;
}

/* A functor is a function object with one single object argument */
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunctionObjArgs(functor, obj, NULL);

/*
  Helper for static pointer initialization for both C and C++ code, for example
  static PyObject *SWIG_STATIC_POINTER(MyVar) = NewSomething(...);
*/
#ifdef __cplusplus
#define SWIG_STATIC_POINTER(var)  var
#else
#define SWIG_STATIC_POINTER(var)  var = 0; if (!var) var
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* Python-specific SWIG API */
#define SWIG_newvarlink()                             SWIG_Python_newvarlink()
#define SWIG_addvarlink(p, name, get_attr, set_attr)  SWIG_Python_addvarlink(p, name, get_attr, set_attr)
#define SWIG_InstallConstants(d, constants)           SWIG_Python_InstallConstants(d, constants)
 
/* -----------------------------------------------------------------------------
 * global variable support code.
 * ----------------------------------------------------------------------------- */
 
typedef struct swig_globalvar {   
  char       *name;                  /* Name of global variable */
  PyObject *(*get_attr)(void);       /* Return the current value */
  int       (*set_attr)(PyObject *); /* Set the value */
  struct swig_globalvar *next;
} swig_globalvar;

typedef struct swig_varlinkobject {
  PyObject_HEAD
  swig_globalvar *vars;
} swig_varlinkobject;

SWIGINTERN PyObject *
swig_varlink_repr(PyObject *SWIGUNUSEDPARM(v)) {
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_InternFromString("<Swig global variables>");
#else
  return PyString_FromString("<Swig global variables>");
#endif
}

SWIGINTERN PyObject *
swig_varlink_str(PyObject *o) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
#if PY_VERSION_HEX >= 0x03000000
  PyObject *str = PyUnicode_InternFromString("(");
  PyObject *tail;
  PyObject *joined;
  swig_globalvar *var;
  for (var = v->vars; var; var=var->next) {
    tail = PyUnicode_FromString(var->name);
    joined = PyUnicode_Concat(str, tail);
    Py_DecRef(str);
    Py_DecRef(tail);
    str = joined;
    if (var->next) {
        tail = PyUnicode_InternFromString(", ");
        joined = PyUnicode_Concat(str, tail);
        Py_DecRef(str);
        Py_DecRef(tail);
        str = joined;
    }
  }
  tail = PyUnicode_InternFromString(")");
  joined = PyUnicode_Concat(str, tail);
  Py_DecRef(str);
  Py_DecRef(tail);
  str = joined;
#else
  PyObject *str = PyString_FromString("(");
  swig_globalvar *var;
  for (var = v->vars; var; var=var->next) {
    PyString_ConcatAndDel(&str,PyString_FromString(var->name));
    if (var->next) PyString_ConcatAndDel(&str,PyString_FromString(", "));
  }
  PyString_ConcatAndDel(&str,PyString_FromString(")"));
#endif
  return str;
}

SWIGINTERN void
swig_varlink_dealloc(PyObject *o) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  swig_globalvar *var = v->vars;
  while (var) {
    swig_globalvar *n = var->next;
    free(var->name);
    free(var);
    var = n;
  }
}

SWIGINTERN PyObject *
swig_varlink_getattr(PyObject *o, char *n) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  PyObject *res = NULL;
  swig_globalvar *var = v->vars;
  while (var) {
    if (strcmp(var->name,n) == 0) {
      res = (*var->get_attr)();
      break;
    }
    var = var->next;
  }
  if (res == NULL && !PyErr_Occurred()) {
    PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
  }
  return res;
}

SWIGINTERN int
swig_varlink_setattr(PyObject *o, char *n, PyObject *p) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  int res = 1;
  swig_globalvar *var = v->vars;
  while (var) {
    if (strcmp(var->name,n) == 0) {
      res = (*var->set_attr)(p);
      break;
    }
    var = var->next;
  }
  if (res == 1 && !PyErr_Occurred()) {
    PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
  }
  return res;
}

SWIGINTERN PyTypeObject*
swig_varlink_type(void) {
  static char varlink__doc__[] = "Swig var link object";
#ifndef Py_LIMITED_API
  static PyTypeObject varlink_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                  /* ob_size */
#endif
      "swigvarlink",                      /* tp_name */
      sizeof(swig_varlinkobject),         /* tp_basicsize */
      0,                                  /* tp_itemsize */
      (destructor) swig_varlink_dealloc,  /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                       /*tp_print*/
#else
      (Py_ssize_t)0,                      /*tp_vectorcall_offset*/
#endif
      (getattrfunc) swig_varlink_getattr, /* tp_getattr */
      (setattrfunc) swig_varlink_setattr, /* tp_setattr */
      0,                                  /* tp_compare */
      (reprfunc) swig_varlink_repr,       /* tp_repr */
      0,                                  /* tp_as_number */
      0,                                  /* tp_as_sequence */
      0,                                  /* tp_as_mapping */
      0,                                  /* tp_hash */
      0,                                  /* tp_call */
      (reprfunc) swig_varlink_str,        /* tp_str */
      0,                                  /* tp_getattro */
      0,                                  /* tp_setattro */
      0,                                  /* tp_as_buffer */
      0,                                  /* tp_flags */
      varlink__doc__,                     /* tp_doc */
      0,                                  /* tp_traverse */
      0,                                  /* tp_clear */
      0,                                  /* tp_richcompare */
      0,                                  /* tp_weaklistoffset */
      0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* tp_iter -> tp_weaklist */
      0,                                  /* tp_del */
      0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                  /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                  /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                  /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030C0000
      0,                                  /* tp_watched */
#endif
#ifdef COUNT_ALLOCS
      0,                                  /* tp_allocs */
      0,                                  /* tp_frees */
      0,                                  /* tp_maxalloc */
      0,                                  /* tp_prev */
      0                                   /* tp_next */
#endif
    };
    varlink_type = tmp;
    type_init = 1;
    if (PyType_Ready(&varlink_type) < 0)
      return NULL;
  }
  return &varlink_type;
#else
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)swig_varlink_dealloc },
    { Py_tp_repr, (void *)swig_varlink_repr },
    { Py_tp_getattr, (void *)swig_varlink_getattr },
    { Py_tp_setattr, (void *)swig_varlink_setattr },
    { Py_tp_str, (void *)swig_varlink_str },
    { Py_tp_doc, (void *)varlink__doc__ },
    { 0, NULL }
  };
  PyType_Spec spec = {
    "swigvarlink",
    sizeof(swig_varlinkobject),
    0,
    Py_TPFLAGS_DEFAULT,
    slots
  };
  return (PyTypeObject *)PyType_FromSpec(&spec);
#endif
}

/* Create a variable linking object for use later */
SWIGINTERN PyObject *
SWIG_Python_newvarlink(void) {
  swig_varlinkobject *result = PyObject_New(swig_varlinkobject, swig_varlink_type());
  if (result) {
    result->vars = 0;
  }
  return ((PyObject*) result);
}

SWIGINTERN void 
SWIG_Python_addvarlink(PyObject *p, const char *name, PyObject *(*get_attr)(void), int (*set_attr)(PyObject *p)) {
  swig_varlinkobject *v = (swig_varlinkobject *) p;
  swig_globalvar *gv = (swig_globalvar *) malloc(sizeof(swig_globalvar));
  if (gv) {
    size_t size = strlen(name)+1;
    gv->name = (char *)malloc(size);
    if (gv->name) {
      memcpy(gv->name, name, size);
      gv->get_attr = get_attr;
      gv->set_attr = set_attr;
      gv->next = v->vars;
    }
  }
  v->vars = gv;
}


static PyObject *Swig_Globals_global = NULL;
  
SWIGINTERN PyObject *
SWIG_globals(void) {
  if (Swig_Globals_global == NULL) {
    Swig_Globals_global = SWIG_newvarlink();
  }
  return Swig_Globals_global;
}

#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------
 * Pointer declarations
 * ----------------------------------------------------------------------------- */

/* Flags for new pointer objects */
#define SWIG_POINTER_NOSHADOW       (SWIG_POINTER_OWN      << 1)
#define SWIG_POINTER_NEW            (SWIG_POINTER_NOSHADOW | SWIG_POINTER_OWN)

#define SWIG_POINTER_IMPLICIT_CONV  (SWIG_POINTER_DISOWN   << 1)

#define SWIG_BUILTIN_TP_INIT	    (SWIG_POINTER_OWN << 2)
#define SWIG_BUILTIN_INIT	    (SWIG_BUILTIN_TP_INIT | SWIG_POINTER_OWN)

#ifdef __cplusplus
extern "C" {
#endif

/* The python void return value */

SWIGRUNTIMEINLINE PyObject * 
SWIG_Py_Void(void)
{
  PyObject *none = Py_None;
  Py_INCREF(none);
  return none;
}

/* SwigPyClientData */

typedef struct {
  PyObject *klass;
  PyObject *newraw;
  PyObject *newargs;
  PyObject *destroy;
  int delargs;
  int implicitconv;
  PyTypeObject *pytype;
} SwigPyClientData;

SWIGRUNTIMEINLINE int 
SWIG_Python_CheckImplicit(swig_type_info *ty)
{
  SwigPyClientData *data = (SwigPyClientData *)ty->clientdata;
  int fail = data ? data->implicitconv : 0;
  if (fail)
    PyErr_SetString(PyExc_TypeError, "Implicit conversion is prohibited for explicit constructors.");
  return fail;
}

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_ExceptionType(swig_type_info *desc) {
  SwigPyClientData *data = desc ? (SwigPyClientData *) desc->clientdata : 0;
  PyObject *klass = data ? data->klass : 0;
  return (klass ? klass : PyExc_RuntimeError);
}


SWIGRUNTIME SwigPyClientData * 
SwigPyClientData_New(PyObject* obj)
{
  if (!obj) {
    return 0;
  } else {
    SwigPyClientData *data = (SwigPyClientData *)malloc(sizeof(SwigPyClientData));
    /* the klass element */
    data->klass = obj;
    Py_INCREF(data->klass);
    /* the newraw method and newargs arguments used to create a new raw instance */
    if (PyClass_Check(obj)) {
      data->newraw = 0;
      Py_INCREF(obj);
      data->newargs = obj;
    } else {
      data->newraw = PyObject_GetAttrString(data->klass, "__new__");
      if (data->newraw) {
        data->newargs = PyTuple_New(1);
        if (data->newargs) {
          Py_INCREF(obj);
          PyTuple_SET_ITEM(data->newargs, 0, obj);
        } else {
          Py_DECREF(data->newraw);
          Py_DECREF(data->klass);
          free(data);
          return 0;
        }
      } else {
        Py_INCREF(obj);
        data->newargs = obj;
      }
    }
    /* the destroy method, aka as the C++ delete method */
    data->destroy = PyObject_GetAttrString(data->klass, "__swig_destroy__");
    if (PyErr_Occurred()) {
      PyErr_Clear();
      data->destroy = 0;
    }
    if (data->destroy) {
      data->delargs = !(PyCFunction_GET_FLAGS(data->destroy) & METH_O);
    } else {
      data->delargs = 0;
    }
    data->implicitconv = 0;
    data->pytype = 0;
    return data;
  }
}

SWIGRUNTIME void 
SwigPyClientData_Del(SwigPyClientData *data)
{
  Py_XDECREF(data->klass);
  Py_XDECREF(data->newraw);
  Py_XDECREF(data->newargs);
  Py_XDECREF(data->destroy);
  free(data);
}

/* =============== SwigPyObject =====================*/

typedef struct {
  PyObject_HEAD
  void *ptr;
  swig_type_info *ty;
  int own;
  PyObject *next;
#ifdef SWIGPYTHON_BUILTIN
  PyObject *dict;
#endif
} SwigPyObject;


#ifdef SWIGPYTHON_BUILTIN

SWIGRUNTIME PyObject *
SwigPyObject_get___dict__(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;

  if (!sobj->dict)
    sobj->dict = PyDict_New();

  Py_XINCREF(sobj->dict);
  return sobj->dict;
}

#endif

SWIGRUNTIME PyObject *
SwigPyObject_long(SwigPyObject *v)
{
  return PyLong_FromVoidPtr(v->ptr);
}

SWIGRUNTIME PyObject *
SwigPyObject_format(const char* fmt, SwigPyObject *v)
{
  PyObject *res = NULL;
  PyObject *args = PyTuple_New(1);
  if (args) {
    PyObject *val = SwigPyObject_long(v);
    if (val) {
      PyObject *ofmt;
      PyTuple_SET_ITEM(args, 0, val);
      ofmt = SWIG_Python_str_FromChar(fmt);
      if (ofmt) {
#if PY_VERSION_HEX >= 0x03000000
        res = PyUnicode_Format(ofmt,args);
#else
        res = PyString_Format(ofmt,args);
#endif
        Py_DECREF(ofmt);
      }
    }
    Py_DECREF(args);
  }
  return res;
}

SWIGRUNTIME PyObject *
SwigPyObject_oct(SwigPyObject *v)
{
  return SwigPyObject_format("%o",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_hex(SwigPyObject *v)
{
  return SwigPyObject_format("%x",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_repr(SwigPyObject *v)
{
  const char *name = SWIG_TypePrettyName(v->ty);
  PyObject *repr = SWIG_Python_str_FromFormat("<Swig Object of type '%s' at %p>", (name ? name : "unknown"), (void *)v);
  if (repr && v->next) {
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next);
    if (nrep) {
# if PY_VERSION_HEX >= 0x03000000
      PyObject *joined = PyUnicode_Concat(repr, nrep);
      Py_DecRef(repr);
      Py_DecRef(nrep);
      repr = joined;
# else
      PyString_ConcatAndDel(&repr,nrep);
# endif
    } else {
      Py_DecRef(repr);
      repr = NULL;
    }
  }
  return repr;
}

/* We need a version taking two PyObject* parameters so it's a valid
 * PyCFunction to use in swigobject_methods[]. */
SWIGRUNTIME PyObject *
SwigPyObject_repr2(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  return SwigPyObject_repr((SwigPyObject*)v);
}

SWIGRUNTIME int
SwigPyObject_compare(SwigPyObject *v, SwigPyObject *w)
{
  void *i = v->ptr;
  void *j = w->ptr;
  return (i < j) ? -1 : ((i > j) ? 1 : 0);
}

/* Added for Python 3.x, would it also be useful for Python 2.x? */
SWIGRUNTIME PyObject*
SwigPyObject_richcompare(SwigPyObject *v, SwigPyObject *w, int op)
{
  PyObject* res = NULL;
  if (!PyErr_Occurred()) {
    if (op != Py_EQ && op != Py_NE) {
      Py_INCREF(Py_NotImplemented);
      return Py_NotImplemented;
    }
    res = PyBool_FromLong( (SwigPyObject_compare(v, w)==0) == (op == Py_EQ) ? 1 : 0);
  }
  return res;  
}


SWIGRUNTIME PyTypeObject* SwigPyObject_TypeOnce(void);

#ifdef SWIGPYTHON_BUILTIN
static swig_type_info *SwigPyObject_stype = 0;
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
    SwigPyClientData *cd;
    assert(SwigPyObject_stype);
    cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
    assert(cd);
    assert(cd->pytype);
    return cd->pytype;
}
#else
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyObject_TypeOnce();
  return type;
}
#endif

SWIGRUNTIMEINLINE int
SwigPyObject_Check(PyObject *op) {
  PyTypeObject *target_tp = SwigPyObject_type();
  PyTypeObject *op_type = Py_TYPE(op);
#ifdef SWIGPYTHON_BUILTIN
  if (PyType_IsSubtype(op_type, target_tp))
    return 1;
  return (strcmp(op_type->tp_name, "SwigPyObject") == 0);
#else
  if (op_type == target_tp)
    return 1;
# ifdef Py_LIMITED_API
  int cmp;
  PyObject *tp_name = PyObject_GetAttrString((PyObject *)op_type, "__name__");
  if (!tp_name)
    return 0;
  cmp = PyUnicode_CompareWithASCIIString(tp_name, "SwigPyObject");
  Py_DECREF(tp_name);
  return cmp == 0;
# else
  return (strcmp(op_type->tp_name, "SwigPyObject") == 0);
# endif
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own);

static PyObject* Swig_Capsule_global = NULL;

SWIGRUNTIME void
SwigPyObject_dealloc(PyObject *v)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  PyObject *next = sobj->next;
  if (sobj->own == SWIG_POINTER_OWN) {
    swig_type_info *ty = sobj->ty;
    SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
    PyObject *destroy = data ? data->destroy : 0;
    if (destroy) {
      /* destroy is always a VARARGS method */
      PyObject *res;

      /* PyObject_CallFunction() has the potential to silently drop
         the active exception.  In cases of unnamed temporary
         variable or where we just finished iterating over a generator
         StopIteration will be active right now, and this needs to
         remain true upon return from SwigPyObject_dealloc.  So save
         and restore. */
      
      PyObject *type = NULL, *value = NULL, *traceback = NULL;
      PyErr_Fetch(&type, &value, &traceback);

      if (data->delargs) {
        /* we need to create a temporary object to carry the destroy operation */
        PyObject *tmp = SwigPyObject_New(sobj->ptr, ty, 0);
        if (tmp) {
          res = SWIG_Python_CallFunctor(destroy, tmp);
        } else {
          res = 0;
        }
        Py_XDECREF(tmp);
      } else {
        PyCFunction meth = PyCFunction_GET_FUNCTION(destroy);
        PyObject *mself = PyCFunction_GET_SELF(destroy);
        res = ((*meth)(mself, v));
      }
      if (!res)
        PyErr_WriteUnraisable(destroy);

      PyErr_Restore(type, value, traceback);

      Py_XDECREF(res);
    } 
#if !defined(SWIG_PYTHON_SILENT_MEMLEAK)
    else {
      const char *name = SWIG_TypePrettyName(ty);
      printf("swig/python detected a memory leak of type '%s', no destructor found.\n", (name ? name : "unknown"));
    }
#endif
    Py_XDECREF(Swig_Capsule_global);
  }
  Py_XDECREF(next);
#ifdef SWIGPYTHON_BUILTIN
  Py_XDECREF(sobj->dict);
#endif
  PyObject_DEL(v);
}

SWIGRUNTIME PyObject* 
SwigPyObject_append(PyObject* v, PyObject* next)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (!SwigPyObject_Check(next)) {
    PyErr_SetString(PyExc_TypeError, "Attempt to append a non SwigPyObject");
    return NULL;
  }
  ((SwigPyObject *)next)->next = sobj->next;
  sobj->next = next;
  Py_INCREF(next);
  return SWIG_Py_Void();
}

SWIGRUNTIME PyObject* 
SwigPyObject_next(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (sobj->next) {    
    Py_INCREF(sobj->next);
    return sobj->next;
  } else {
    return SWIG_Py_Void();
  }
}

SWIGINTERN PyObject*
SwigPyObject_disown(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = 0;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_acquire(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = SWIG_POINTER_OWN;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_own(PyObject *v, PyObject *args)
{
  PyObject *val = 0;
  if (!PyArg_UnpackTuple(args, "own", 0, 1, &val)) {
    return NULL;
  } else {
    SwigPyObject *sobj = (SwigPyObject *)v;
    PyObject *obj = PyBool_FromLong(sobj->own);
    if (val) {
      if (PyObject_IsTrue(val)) {
        Py_DECREF(SwigPyObject_acquire(v,args));
      } else {
        Py_DECREF(SwigPyObject_disown(v,args));
      }
    } 
    return obj;
  }
}

static PyMethodDef
swigobject_methods[] = {
  {"disown",  SwigPyObject_disown,  METH_NOARGS,  "releases ownership of the pointer"},
  {"acquire", SwigPyObject_acquire, METH_NOARGS,  "acquires ownership of the pointer"},
  {"own",     SwigPyObject_own,     METH_VARARGS, "returns/sets ownership of the pointer"},
  {"append",  SwigPyObject_append,  METH_O,       "appends another 'this' object"},
  {"next",    SwigPyObject_next,    METH_NOARGS,  "returns the next 'this' object"},
  {"__repr__",SwigPyObject_repr2,   METH_NOARGS,  "returns object representation"},
  {0, 0, 0, 0}  
};

SWIGRUNTIME PyTypeObject*
SwigPyObject_TypeOnce(void) {
  static char swigobject_doc[] = "Swig object carries a C/C++ instance pointer";
#ifndef Py_LIMITED_API
  static PyNumberMethods SwigPyObject_as_number = {
    (binaryfunc)0, /*nb_add*/
    (binaryfunc)0, /*nb_subtract*/
    (binaryfunc)0, /*nb_multiply*/
    /* nb_divide removed in Python 3 */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc)0, /*nb_divide*/
#endif
    (binaryfunc)0, /*nb_remainder*/
    (binaryfunc)0, /*nb_divmod*/
    (ternaryfunc)0,/*nb_power*/
    (unaryfunc)0,  /*nb_negative*/
    (unaryfunc)0,  /*nb_positive*/
    (unaryfunc)0,  /*nb_absolute*/
    (inquiry)0,    /*nb_nonzero*/
    0,		   /*nb_invert*/
    0,		   /*nb_lshift*/
    0,		   /*nb_rshift*/
    0,		   /*nb_and*/
    0,		   /*nb_xor*/
    0,		   /*nb_or*/
#if PY_VERSION_HEX < 0x03000000
    0,   /*nb_coerce*/
#endif
    (unaryfunc)SwigPyObject_long, /*nb_int*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_long, /*nb_long*/
#else
    0, /*nb_reserved*/
#endif
    (unaryfunc)0,                 /*nb_float*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_oct,  /*nb_oct*/
    (unaryfunc)SwigPyObject_hex,  /*nb_hex*/
#endif
#if PY_VERSION_HEX >= 0x03050000 /* 3.5 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_matrix_multiply */
#elif PY_VERSION_HEX >= 0x03000000 /* 3.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index, nb_inplace_divide removed */
#else
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index */
#endif
  };

  static PyTypeObject swigpyobject_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      "SwigPyObject",                       /* tp_name */
      sizeof(SwigPyObject),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyObject_dealloc,     /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                         /*tp_print*/
#else
      (Py_ssize_t)0,                        /*tp_vectorcall_offset*/
#endif
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
      0, /* tp_reserved in 3.0.1, tp_compare in 3.0.0 but not used */
#else
      (cmpfunc)SwigPyObject_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyObject_repr,          /* tp_repr */
      &SwigPyObject_as_number,              /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      0,                                    /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigobject_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      (richcmpfunc)SwigPyObject_richcompare,/* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      swigobject_methods,                   /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030C0000
      0,                                    /* tp_watched */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    swigpyobject_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpyobject_type) != 0)
      return NULL;
  }
  return &swigpyobject_type;
#else
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigPyObject_dealloc },
    { Py_tp_repr, (void *)SwigPyObject_repr },
    { Py_tp_getattro, (void *)PyObject_GenericGetAttr },
    { Py_tp_doc, (void *)swigobject_doc },
    { Py_tp_richcompare, (void *)SwigPyObject_richcompare },
    { Py_tp_methods, (void *)swigobject_methods },
    { Py_nb_int, (void *)SwigPyObject_long },
    { 0, NULL }
  };
  PyType_Spec spec = {
    "SwigPyObject",
    sizeof(SwigPyObject),
    0,
    Py_TPFLAGS_DEFAULT,
    slots
  };
  return (PyTypeObject *)PyType_FromSpec(&spec);
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own)
{
  SwigPyObject *sobj = PyObject_New(SwigPyObject, SwigPyObject_type());
  if (sobj) {
    sobj->ptr  = ptr;
    sobj->ty   = ty;
    sobj->own  = own;
    sobj->next = 0;
#ifdef SWIGPYTHON_BUILTIN
    sobj->dict = 0;
#endif
    if (own == SWIG_POINTER_OWN) {
      /* Obtain a reference to the Python capsule wrapping the module information, so that the
       * module information is correctly destroyed after all SWIG python objects have been freed
       * by the GC (and corresponding destructors invoked) */
      Py_XINCREF(Swig_Capsule_global);
    }
  }
  return (PyObject *)sobj;
}

/* -----------------------------------------------------------------------------
 * Implements a simple Swig Packed type, and use it instead of string
 * ----------------------------------------------------------------------------- */

typedef struct {
  PyObject_HEAD
  void *pack;
  swig_type_info *ty;
  size_t size;
} SwigPyPacked;

SWIGRUNTIME PyObject *
SwigPyPacked_repr(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    return SWIG_Python_str_FromFormat("<Swig Packed at %s%s>", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromFormat("<Swig Packed %s>", v->ty->name);
  }  
}

SWIGRUNTIME PyObject *
SwigPyPacked_str(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))){
    return SWIG_Python_str_FromFormat("%s%s", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromChar(v->ty->name);
  }  
}

SWIGRUNTIME int
SwigPyPacked_compare(SwigPyPacked *v, SwigPyPacked *w)
{
  size_t i = v->size;
  size_t j = w->size;
  int s = (i < j) ? -1 : ((i > j) ? 1 : 0);
  return s ? s : strncmp((const char *)v->pack, (const char *)w->pack, 2*v->size);
}

SWIGRUNTIME PyTypeObject* SwigPyPacked_TypeOnce(void);

SWIGRUNTIME PyTypeObject*
SwigPyPacked_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyPacked_TypeOnce();
  return type;
}

SWIGRUNTIMEINLINE int
SwigPyPacked_Check(PyObject *op) {
  PyTypeObject* op_type = Py_TYPE(op);
  if (op_type == SwigPyPacked_TypeOnce())
    return 1;
#ifdef Py_LIMITED_API
  int cmp;
  PyObject *tp_name = PyObject_GetAttrString((PyObject *)op_type, "__name__");
  if (!tp_name)
    return 0;
  cmp = PyUnicode_CompareWithASCIIString(tp_name, "SwigPyPacked");
  Py_DECREF(tp_name);
  return cmp == 0;
#else
  return (strcmp(op_type->tp_name, "SwigPyPacked") == 0);
#endif
}

SWIGRUNTIME void
SwigPyPacked_dealloc(PyObject *v)
{
  if (SwigPyPacked_Check(v)) {
    SwigPyPacked *sobj = (SwigPyPacked *) v;
    free(sobj->pack);
  }
  PyObject_DEL(v);
}

SWIGRUNTIME PyTypeObject*
SwigPyPacked_TypeOnce(void) {
  static char swigpacked_doc[] = "Swig object carries a C/C++ instance pointer";
#ifndef Py_LIMITED_API
  static PyTypeObject swigpypacked_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX>=0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      "SwigPyPacked",                       /* tp_name */
      sizeof(SwigPyPacked),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyPacked_dealloc,     /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                         /*tp_print*/
#else
      (Py_ssize_t)0,                        /*tp_vectorcall_offset*/
#endif
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX>=0x03000000
      0, /* tp_reserved in 3.0.1 */
#else
      (cmpfunc)SwigPyPacked_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyPacked_repr,          /* tp_repr */
      0,                                    /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      (reprfunc)SwigPyPacked_str,           /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigpacked_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      0,                                    /* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      0,                                    /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030C0000
      0,                                    /* tp_watched */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    swigpypacked_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpypacked_type) != 0)
      return NULL;
  }
  return &swigpypacked_type;
#else
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigPyPacked_dealloc },
    { Py_tp_repr, (void *)SwigPyPacked_repr },
    { Py_tp_str, (void *)SwigPyPacked_str },
    { Py_tp_getattro, (void *)PyObject_GenericGetAttr },
    { Py_tp_doc, (void *)swigpacked_doc },
    { 0, NULL }
  };
  PyType_Spec spec = {
    "SwigPyPacked",
    sizeof(SwigPyPacked),
    0,
    Py_TPFLAGS_DEFAULT,
    slots
  };
  return (PyTypeObject *)PyType_FromSpec(&spec);
#endif
}

SWIGRUNTIME PyObject *
SwigPyPacked_New(void *ptr, size_t size, swig_type_info *ty)
{
  SwigPyPacked *sobj = PyObject_New(SwigPyPacked, SwigPyPacked_type());
  if (sobj) {
    void *pack = malloc(size);
    if (pack) {
      memcpy(pack, ptr, size);
      sobj->pack = pack;
      sobj->ty   = ty;
      sobj->size = size;
    } else {
      PyObject_DEL((PyObject *) sobj);
      sobj = 0;
    }
  }
  return (PyObject *) sobj;
}

SWIGRUNTIME swig_type_info *
SwigPyPacked_UnpackData(PyObject *obj, void *ptr, size_t size)
{
  if (SwigPyPacked_Check(obj)) {
    SwigPyPacked *sobj = (SwigPyPacked *)obj;
    if (sobj->size != size) return 0;
    memcpy(ptr, sobj->pack, size);
    return sobj->ty;
  } else {
    return 0;
  }
}

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

static PyObject *Swig_This_global = NULL;

SWIGRUNTIME PyObject *
SWIG_This(void)
{
  if (Swig_This_global == NULL)
    Swig_This_global = SWIG_Python_str_FromChar("this");
  return Swig_This_global;
}

/* #define SWIG_PYTHON_SLOW_GETSET_THIS */

/* TODO: I don't know how to implement the fast getset in Python 3 right now */
#if PY_VERSION_HEX>=0x03000000
#define SWIG_PYTHON_SLOW_GETSET_THIS 
#endif

SWIGRUNTIME SwigPyObject *
SWIG_Python_GetSwigThis(PyObject *pyobj) 
{
  PyObject *obj;

  if (SwigPyObject_Check(pyobj))
    return (SwigPyObject *) pyobj;

#ifdef SWIGPYTHON_BUILTIN
  (void)obj;
# ifdef PyWeakref_CheckProxy
  if (PyWeakref_CheckProxy(pyobj)) {
    pyobj = PyWeakref_GET_OBJECT(pyobj);
    if (pyobj && SwigPyObject_Check(pyobj))
      return (SwigPyObject*) pyobj;
  }
# endif
  return NULL;
#else

  obj = 0;

#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  if (PyInstance_Check(pyobj)) {
    obj = _PyInstance_Lookup(pyobj, SWIG_This());      
  } else {
    PyObject **dictptr = _PyObject_GetDictPtr(pyobj);
    if (dictptr != NULL) {
      PyObject *dict = *dictptr;
      obj = dict ? PyDict_GetItem(dict, SWIG_This()) : 0;
    } else {
#ifdef PyWeakref_CheckProxy
      if (PyWeakref_CheckProxy(pyobj)) {
	PyObject *wobj = PyWeakref_GET_OBJECT(pyobj);
	return wobj ? SWIG_Python_GetSwigThis(wobj) : 0;
      }
#endif
      obj = PyObject_GetAttr(pyobj,SWIG_This());
      if (obj) {
	Py_DECREF(obj);
      } else {
	if (PyErr_Occurred()) PyErr_Clear();
	return 0;
      }
    }
  }
#else
  obj = PyObject_GetAttr(pyobj,SWIG_This());
  if (obj) {
    Py_DECREF(obj);
  } else {
    if (PyErr_Occurred()) PyErr_Clear();
    return 0;
  }
#endif
  if (obj && !SwigPyObject_Check(obj)) {
    /* a PyObject is called 'this', try to get the 'real this'
       SwigPyObject from it */ 
    return SWIG_Python_GetSwigThis(obj);
  }
  return (SwigPyObject *)obj;
#endif
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Python_AcquirePtr(PyObject *obj, int own) {
  if (own == SWIG_POINTER_OWN) {
    SwigPyObject *sobj = SWIG_Python_GetSwigThis(obj);
    if (sobj) {
      int oldown = sobj->own;
      sobj->own = own;
      return oldown;
    }
  }
  return 0;
}

/* Convert a pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPtrAndOwn(PyObject *obj, void **ptr, swig_type_info *ty, int flags, int *own) {
  int res;
  SwigPyObject *sobj;
  int implicit_conv = (flags & SWIG_POINTER_IMPLICIT_CONV) != 0;

  if (!obj)
    return SWIG_ERROR;
  if (obj == Py_None && !implicit_conv) {
    if (ptr)
      *ptr = 0;
    return (flags & SWIG_POINTER_NO_NULL) ? SWIG_NullReferenceError : SWIG_OK;
  }

  res = SWIG_ERROR;

  sobj = SWIG_Python_GetSwigThis(obj);
  if (own)
    *own = 0;
  while (sobj) {
    void *vptr = sobj->ptr;
    if (ty) {
      swig_type_info *to = sobj->ty;
      if (to == ty) {
        /* no type cast needed */
        if (ptr) *ptr = vptr;
        break;
      } else {
        swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
        if (!tc) {
          sobj = (SwigPyObject *)sobj->next;
        } else {
          if (ptr) {
            int newmemory = 0;
            *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
            if (newmemory == SWIG_CAST_NEW_MEMORY) {
              assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
              if (own)
                *own = *own | SWIG_CAST_NEW_MEMORY;
            }
          }
          break;
        }
      }
    } else {
      if (ptr) *ptr = vptr;
      break;
    }
  }
  if (sobj) {
    if (((flags & SWIG_POINTER_RELEASE) == SWIG_POINTER_RELEASE) && !sobj->own) {
      res = SWIG_ERROR_RELEASE_NOT_OWNED;
    } else {
      if (own)
        *own = *own | sobj->own;
      if (flags & SWIG_POINTER_DISOWN) {
        sobj->own = 0;
      }
      if (flags & SWIG_POINTER_CLEAR) {
        sobj->ptr = 0;
      }
      res = SWIG_OK;
    }
  } else {
    if (implicit_conv) {
      SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
      if (data && !data->implicitconv) {
        PyObject *klass = data->klass;
        if (klass) {
          PyObject *impconv;
          data->implicitconv = 1; /* avoid recursion and call 'explicit' constructors*/
          impconv = SWIG_Python_CallFunctor(klass, obj);
          data->implicitconv = 0;
          if (PyErr_Occurred()) {
            PyErr_Clear();
            impconv = 0;
          }
          if (impconv) {
            SwigPyObject *iobj = SWIG_Python_GetSwigThis(impconv);
            if (iobj) {
              void *vptr;
              res = SWIG_Python_ConvertPtrAndOwn((PyObject*)iobj, &vptr, ty, 0, 0);
              if (SWIG_IsOK(res)) {
                if (ptr) {
                  *ptr = vptr;
                  /* transfer the ownership to 'ptr' */
                  iobj->own = 0;
                  res = SWIG_AddCast(res);
                  res = SWIG_AddNewMask(res);
                } else {
                  res = SWIG_AddCast(res);		    
                }
              }
            }
            Py_DECREF(impconv);
          }
        }
      }
      if (!SWIG_IsOK(res) && obj == Py_None) {
        if (ptr)
          *ptr = 0;
        if (PyErr_Occurred())
          PyErr_Clear();
        res = SWIG_OK;
      }
    }
  }
  return res;
}

/* Convert a function ptr value */

SWIGRUNTIME int
SWIG_Python_ConvertFunctionPtr(PyObject *obj, void **ptr, swig_type_info *ty) {
  if (!PyCFunction_Check(obj)) {
    return SWIG_ConvertPtr(obj, ptr, ty, 0);
  } else {
    void *vptr = 0;
    swig_cast_info *tc;

    /* here we get the method pointer for callbacks */
#ifndef Py_LIMITED_API
    const char *doc = (((PyCFunctionObject *)obj) -> m_ml -> ml_doc);
#else
    PyObject* pystr_doc = PyObject_GetAttrString(obj, "__doc__");
    PyObject *bytes = NULL;
    const char *doc = pystr_doc ? SWIG_PyUnicode_AsUTF8AndSize(pystr_doc, NULL, &bytes) : 0;
#endif
    const char *desc = doc ? strstr(doc, "swig_ptr: ") : 0;
    if (desc)
      desc = ty ? SWIG_UnpackVoidPtr(desc + 10, &vptr, ty->name) : 0;
#ifdef Py_LIMITED_API
    Py_XDECREF(bytes);
    Py_XDECREF(pystr_doc);
#endif
    if (!desc)
      return SWIG_ERROR;
    tc = SWIG_TypeCheck(desc,ty);
    if (tc) {
      int newmemory = 0;
      *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
      assert(!newmemory); /* newmemory handling not yet implemented */
    } else {
      return SWIG_ERROR;
    }
    return SWIG_OK;
  }
}

/* Convert a packed pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPacked(PyObject *obj, void *ptr, size_t sz, swig_type_info *ty) {
  swig_type_info *to = SwigPyPacked_UnpackData(obj, ptr, sz);
  if (!to) return SWIG_ERROR;
  if (ty) {
    if (to != ty) {
      /* check type cast? */
      swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
      if (!tc) return SWIG_ERROR;
    }
  }
  return SWIG_OK;
}  

/* -----------------------------------------------------------------------------
 * Create a new pointer object
 * ----------------------------------------------------------------------------- */

/*
  Create a new instance object, without calling __init__, and set the
  'this' attribute.
*/

SWIGRUNTIME PyObject* 
SWIG_Python_NewShadowInstance(SwigPyClientData *data, PyObject *swig_this)
{
  PyObject *inst = 0;
  PyObject *newraw = data->newraw;
  if (newraw) {
    inst = PyObject_Call(newraw, data->newargs, NULL);
    if (inst) {
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
      PyObject **dictptr = _PyObject_GetDictPtr(inst);
      if (dictptr != NULL) {
        PyObject *dict = *dictptr;
        if (dict == NULL) {
          dict = PyDict_New();
          *dictptr = dict;
        }
        if (dict) {
          PyDict_SetItem(dict, SWIG_This(), swig_this);
        } else{
          Py_DECREF(inst);
          inst = 0;
        }
      }
#else
      if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
        Py_DECREF(inst);
        inst = 0;
      }
#endif
    }
  } else {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *empty_args = PyTuple_New(0);
    if (empty_args) {
      PyObject *empty_kwargs = PyDict_New();
      if (empty_kwargs) {
#ifndef Py_LIMITED_API
        newfunc newfn = ((PyTypeObject *)data->newargs)->tp_new;
#else
        newfunc newfn = (newfunc)PyType_GetSlot((PyTypeObject *)data->newargs, Py_tp_new);
#endif
        inst = newfn((PyTypeObject *)data->newargs, empty_args, empty_kwargs);
        Py_DECREF(empty_kwargs);
        if (inst) {
          if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
            Py_DECREF(inst);
            inst = 0;
          } else {
            PyType_Modified(Py_TYPE(inst));
          }
        }
      }
      Py_DECREF(empty_args);
    }
#else
    PyObject *dict = PyDict_New();
    if (dict) {
      PyDict_SetItem(dict, SWIG_This(), swig_this);
      inst = PyInstance_NewRaw(data->newargs, dict);
      Py_DECREF(dict);
    }
#endif
  }
  return inst;
}

SWIGRUNTIME int
SWIG_Python_SetSwigThis(PyObject *inst, PyObject *swig_this)
{
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  PyObject **dictptr = _PyObject_GetDictPtr(inst);
  if (dictptr != NULL) {
    PyObject *dict = *dictptr;
    if (dict == NULL) {
      dict = PyDict_New();
      *dictptr = dict;
    }
    if (dict) {
      return PyDict_SetItem(dict, SWIG_This(), swig_this);
    } else{
      return -1;
    }
  }
#endif
  return PyObject_SetAttr(inst, SWIG_This(), swig_this);
} 


SWIGINTERN PyObject *
SWIG_Python_InitShadowInstance(PyObject *args) {
  PyObject *obj[2];
  if (!SWIG_Python_UnpackTuple(args, "swiginit", 2, 2, obj)) {
    return NULL;
  } else {
    SwigPyObject *sthis = SWIG_Python_GetSwigThis(obj[0]);
    if (sthis) {
      Py_DECREF(SwigPyObject_append((PyObject*) sthis, obj[1]));
    } else {
      if (SWIG_Python_SetSwigThis(obj[0], obj[1]) != 0)
        return NULL;
    }
    return SWIG_Py_Void();
  }
}

/* Create a new pointer object */

SWIGRUNTIME PyObject *
SWIG_Python_NewPointerObj(PyObject *self, void *ptr, swig_type_info *type, int flags) {
  SwigPyClientData *clientdata;
  PyObject * robj;
  int own;

  if (!ptr)
    return SWIG_Py_Void();

  clientdata = type ? (SwigPyClientData *)(type->clientdata) : 0;
  own = (flags & SWIG_POINTER_OWN) ? SWIG_POINTER_OWN : 0;
  if (clientdata && clientdata->pytype) {
    SwigPyObject *newobj;
    if (flags & SWIG_BUILTIN_TP_INIT) {
      newobj = (SwigPyObject*) self;
      if (newobj->ptr) {
#ifndef Py_LIMITED_API
        allocfunc alloc = clientdata->pytype->tp_alloc;
#else
        allocfunc alloc = (allocfunc)PyType_GetSlot(clientdata->pytype, Py_tp_alloc);
#endif
        PyObject *next_self = alloc(clientdata->pytype, 0);
        while (newobj->next)
	  newobj = (SwigPyObject *) newobj->next;
        newobj->next = next_self;
        newobj = (SwigPyObject *)next_self;
#ifdef SWIGPYTHON_BUILTIN
        newobj->dict = 0;
#endif
      }
    } else {
      newobj = PyObject_New(SwigPyObject, clientdata->pytype);
#ifdef SWIGPYTHON_BUILTIN
      if (newobj) {
        newobj->dict = 0;
      }
#endif
    }
    if (newobj) {
      newobj->ptr = ptr;
      newobj->ty = type;
      newobj->own = own;
      newobj->next = 0;
      return (PyObject*) newobj;
    }
    return SWIG_Py_Void();
  }

  assert(!(flags & SWIG_BUILTIN_TP_INIT));

  robj = SwigPyObject_New(ptr, type, own);
  if (robj && clientdata && !(flags & SWIG_POINTER_NOSHADOW)) {
    PyObject *inst = SWIG_Python_NewShadowInstance(clientdata, robj);
    Py_DECREF(robj);
    robj = inst;
  }
  return robj;
}

/* Create a new packed object */

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_NewPackedObj(void *ptr, size_t sz, swig_type_info *type) {
  return ptr ? SwigPyPacked_New((void *) ptr, sz, type) : SWIG_Py_Void();
}

/* -----------------------------------------------------------------------------*
 *  Get type list 
 * -----------------------------------------------------------------------------*/

#ifdef SWIG_LINK_RUNTIME
void *SWIG_ReturnGlobalTypeList(void *);
#endif

static PyObject *Swig_TypeCache_global = NULL;

/* The python cached type query */
SWIGRUNTIME PyObject *
SWIG_Python_TypeCache(void) {
  if (Swig_TypeCache_global == NULL) {
    Swig_TypeCache_global = PyDict_New();
  }
  return Swig_TypeCache_global;
}

SWIGRUNTIME swig_module_info *
SWIG_Python_GetModule(void *SWIGUNUSEDPARM(clientdata)) {
#ifdef SWIG_LINK_RUNTIME
  static void *type_pointer = (void *)0;
  /* first check if module already created */
  if (!type_pointer) {
    type_pointer = SWIG_ReturnGlobalTypeList((void *)0);
  }
#else
  void *type_pointer = PyCapsule_Import(SWIGPY_CAPSULE_NAME, 0);
  if (PyErr_Occurred()) {
    PyErr_Clear();
    type_pointer = (void *)0;
  }
#endif
  return (swig_module_info *) type_pointer;
}


static int interpreter_counter = 0; /* how many (sub-)interpreters are using swig_module's types */

SWIGRUNTIME void
SWIG_Python_DestroyModule(PyObject *obj)
{
  swig_module_info *swig_module = (swig_module_info *) PyCapsule_GetPointer(obj, SWIGPY_CAPSULE_NAME);
  swig_type_info **types = swig_module->types;
  size_t i;
  if (--interpreter_counter != 0) /* another sub-interpreter may still be using the swig_module's types */
    return;
  for (i =0; i < swig_module->size; ++i) {
    swig_type_info *ty = types[i];
    if (ty->owndata) {
      SwigPyClientData *data = (SwigPyClientData *) ty->clientdata;
      ty->clientdata = 0;
      if (data) SwigPyClientData_Del(data);
    }
  }
  Py_DECREF(SWIG_This());
  Swig_This_global = NULL;
  Py_DECREF(SWIG_globals());
  Swig_Globals_global = NULL;
  Py_DECREF(SWIG_Python_TypeCache());
  Swig_TypeCache_global = NULL;
  Swig_Capsule_global = NULL;
}

SWIGRUNTIME void
SWIG_Python_SetModule(swig_module_info *swig_module) {
#if PY_VERSION_HEX >= 0x03000000
 /* Add a dummy module object into sys.modules */
  PyObject *module = PyImport_AddModule("swig_runtime_data" SWIG_RUNTIME_VERSION);
#else
  static PyMethodDef swig_empty_runtime_method_table[] = { {NULL, NULL, 0, NULL} }; /* Sentinel */
  PyObject *module = Py_InitModule("swig_runtime_data" SWIG_RUNTIME_VERSION, swig_empty_runtime_method_table);
#endif
  PyObject *pointer = PyCapsule_New((void *) swig_module, SWIGPY_CAPSULE_NAME, SWIG_Python_DestroyModule);
  if (pointer && module) {
    if (PyModule_AddObject(module, SWIGPY_CAPSULE_ATTR_NAME, pointer) == 0) {
      ++interpreter_counter;
      Swig_Capsule_global = pointer;
    } else {
      Py_DECREF(pointer);
    }
  } else {
    Py_XDECREF(pointer);
  }
}

SWIGRUNTIME swig_type_info *
SWIG_Python_TypeQuery(const char *type)
{
  PyObject *cache = SWIG_Python_TypeCache();
  PyObject *key = SWIG_Python_str_FromChar(type); 
  PyObject *obj = PyDict_GetItem(cache, key);
  swig_type_info *descriptor;
  if (obj) {
    descriptor = (swig_type_info *) PyCapsule_GetPointer(obj, NULL);
  } else {
    swig_module_info *swig_module = SWIG_GetModule(0);
    descriptor = SWIG_TypeQueryModule(swig_module, swig_module, type);
    if (descriptor) {
      obj = PyCapsule_New((void*) descriptor, NULL, NULL);
      if (obj) {
        PyDict_SetItem(cache, key, obj);
        Py_DECREF(obj);
      }
    }
  }
  Py_DECREF(key);
  return descriptor;
}

/* 
   For backward compatibility only
*/
#define SWIG_POINTER_EXCEPTION  0
#define SWIG_arg_fail(arg)      SWIG_Python_ArgFail(arg)
#define SWIG_MustGetPtr(p, type, argnum, flags)  SWIG_Python_MustGetPtr(p, type, argnum, flags)

SWIGRUNTIME int
SWIG_Python_AddErrMesg(const char* mesg, int infront)
{  
  if (PyErr_Occurred()) {
    PyObject *type = 0;
    PyObject *value = 0;
    PyObject *traceback = 0;
    PyErr_Fetch(&type, &value, &traceback);
    if (value) {
      PyObject *old_str = PyObject_Str(value);
      PyObject *bytes = NULL;
      const char *tmp = SWIG_PyUnicode_AsUTF8AndSize(old_str, NULL, &bytes);
      const char *errmesg = tmp ? tmp : "Invalid error message";
      Py_XINCREF(type);
      PyErr_Clear();
      if (infront) {
	PyErr_Format(type, "%s %s", mesg, errmesg);
      } else {
	PyErr_Format(type, "%s %s", errmesg, mesg);
      }
      Py_XDECREF(bytes);
      Py_DECREF(old_str);
    }
    return 1;
  } else {
    return 0;
  }
}
  
SWIGRUNTIME int
SWIG_Python_ArgFail(int argnum)
{
  if (PyErr_Occurred()) {
    /* add information about failing argument */
    char mesg[256];
    PyOS_snprintf(mesg, sizeof(mesg), "argument number %d:", argnum);
    return SWIG_Python_AddErrMesg(mesg, 1);
  } else {
    return 0;
  }
}

SWIGRUNTIMEINLINE const char *
SwigPyObject_GetDesc(PyObject *self)
{
  SwigPyObject *v = (SwigPyObject *)self;
  swig_type_info *ty = v ? v->ty : 0;
  return ty ? ty->str : "";
}

SWIGRUNTIME void
SWIG_Python_TypeError(const char *type, PyObject *obj)
{
  if (type) {
#if defined(SWIG_COBJECT_TYPES)
    if (obj && SwigPyObject_Check(obj)) {
      const char *otype = (const char *) SwigPyObject_GetDesc(obj);
      if (otype) {
	PyErr_Format(PyExc_TypeError, "a '%s' is expected, 'SwigPyObject(%s)' is received",
		     type, otype);
	return;
      }
    } else 
#endif      
    {
#ifndef Py_LIMITED_API
      /* tp_name is not accessible */
      const char *otype = (obj ? obj->ob_type->tp_name : 0); 
      if (otype) {
	PyObject *str = PyObject_Str(obj);
	PyObject *bytes = NULL;
	const char *cstr = str ? SWIG_PyUnicode_AsUTF8AndSize(str, NULL, &bytes) : 0;
	if (cstr) {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s(%s)' is received",
		       type, otype, cstr);
	} else {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s' is received",
		       type, otype);
	}
	Py_XDECREF(bytes);
	Py_XDECREF(str);
	return;
      }
#endif
    }   
    PyErr_Format(PyExc_TypeError, "a '%s' is expected", type);
  } else {
    PyErr_Format(PyExc_TypeError, "unexpected type is received");
  }
}


/* Convert a pointer value, signal an exception on a type mismatch */
SWIGRUNTIME void *
SWIG_Python_MustGetPtr(PyObject *obj, swig_type_info *ty, int SWIGUNUSEDPARM(argnum), int flags) {
  void *result;
  if (SWIG_Python_ConvertPtr(obj, &result, ty, flags) == -1) {
    PyErr_Clear();
  }
  return result;
}

#ifdef SWIGPYTHON_BUILTIN
SWIGRUNTIME int
SWIG_Python_NonDynamicSetAttr(PyObject *obj, PyObject *name, PyObject *value) {
  PyTypeObject *tp = obj->ob_type;
  PyObject *descr;
  PyObject *encoded_name;
  descrsetfunc f;
  int res = -1;

# ifdef Py_USING_UNICODE
  if (PyString_Check(name)) {
    name = PyUnicode_Decode(PyString_AsString(name), PyString_Size(name), NULL, NULL);
    if (!name)
      return -1;
  } else if (!PyUnicode_Check(name))
# else
  if (!PyString_Check(name))
# endif
  {
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%.200s'", name->ob_type->tp_name);
    return -1;
  } else {
    Py_INCREF(name);
  }

  if (!tp->tp_dict) {
    if (PyType_Ready(tp) != 0)
      goto done;
  }

  descr = _PyType_Lookup(tp, name);
  f = NULL;
  if (descr != NULL)
    f = descr->ob_type->tp_descr_set;
  if (!f) {
    if (PyString_Check(name)) {
      encoded_name = name;
      Py_INCREF(name);
    } else {
      encoded_name = PyUnicode_AsUTF8String(name);
      if (!encoded_name)
        goto done;
    }
    PyErr_Format(PyExc_AttributeError, "'%.100s' object has no attribute '%.200s'", tp->tp_name, PyString_AsString(encoded_name));
    Py_DECREF(encoded_name);
  } else {
    res = f(descr, obj, value);
  }
  
  done:
  Py_DECREF(name);
  return res;
}
#endif


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) do { if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } } while (0) 



  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg); SWIG_fail;; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_m_CModule__f_r_q_const__CString__void swig_types[0]
#define SWIGTYPE_p_CActionMessage swig_types[1]
#define SWIGTYPE_p_CAuthBase swig_types[2]
#define SWIGTYPE_p_CAuthenticateMessage swig_types[3]
#define SWIGTYPE_p_CBufLine swig_types[4]
#define SWIGTYPE_p_CBuffer swig_types[5]
#define SWIGTYPE_p_CCTCPMessage swig_types[6]
#define SWIGTYPE_p_CCapability swig_types[7]
#define SWIGTYPE_p_CChan swig_types[8]
#define SWIGTYPE_p_CChgHostMessage swig_types[9]
#define SWIGTYPE_p_CClient swig_types[10]
#define SWIGTYPE_p_CClientAuth swig_types[11]
#define SWIGTYPE_p_CConfig swig_types[12]
#define SWIGTYPE_p_CConfigEntry swig_types[13]
#define SWIGTYPE_p_CConnectQueueTimer swig_types[14]
#define SWIGTYPE_p_CCoreTranslationMixin swig_types[15]
#define SWIGTYPE_p_CCron swig_types[16]
#define SWIGTYPE_p_CDebug swig_types[17]
#define SWIGTYPE_p_CDebugStream swig_types[18]
#define SWIGTYPE_p_CDelayedTranslation swig_types[19]
#define SWIGTYPE_p_CDir swig_types[20]
#define SWIGTYPE_p_CException swig_types[21]
#define SWIGTYPE_p_CExecSock swig_types[22]
#define SWIGTYPE_p_CFPTimer swig_types[23]
#define SWIGTYPE_p_CFile swig_types[24]
#define SWIGTYPE_p_CGetAddrInfo swig_types[25]
#define SWIGTYPE_p_CHTTPSock swig_types[26]
#define SWIGTYPE_p_CIRCNetwork swig_types[27]
#define SWIGTYPE_p_CIRCSock swig_types[28]
#define SWIGTYPE_p_CIRCSocket swig_types[29]
#define SWIGTYPE_p_CIncomingConnection swig_types[30]
#define SWIGTYPE_p_CInviteMessage swig_types[31]
#define SWIGTYPE_p_CJoinMessage swig_types[32]
#define SWIGTYPE_p_CKickMessage swig_types[33]
#define SWIGTYPE_p_CLanguageScope swig_types[34]
#define SWIGTYPE_p_CListener swig_types[35]
#define SWIGTYPE_p_CMessage swig_types[36]
#define SWIGTYPE_p_CModCommand swig_types[37]
#define SWIGTYPE_p_CModInfo swig_types[38]
#define SWIGTYPE_p_CModPython swig_types[39]
#define SWIGTYPE_p_CModeMessage swig_types[40]
#define SWIGTYPE_p_CModule swig_types[41]
#define SWIGTYPE_p_CModuleEntry swig_types[42]
#define SWIGTYPE_p_CModules swig_types[43]
#define SWIGTYPE_p_CModulesIter swig_types[44]
#define SWIGTYPE_p_CNick swig_types[45]
#define SWIGTYPE_p_CNickMessage swig_types[46]
#define SWIGTYPE_p_CNoticeMessage swig_types[47]
#define SWIGTYPE_p_CNumericMessage swig_types[48]
#define SWIGTYPE_p_COptionalTranslation swig_types[49]
#define SWIGTYPE_p_CPartMessage swig_types[50]
#define SWIGTYPE_p_CPyCapability swig_types[51]
#define SWIGTYPE_p_CPyModCommand swig_types[52]
#define SWIGTYPE_p_CPyModule swig_types[53]
#define SWIGTYPE_p_CPyRetBool swig_types[54]
#define SWIGTYPE_p_CPyRetString swig_types[55]
#define SWIGTYPE_p_CPySocket swig_types[56]
#define SWIGTYPE_p_CPyTimer swig_types[57]
#define SWIGTYPE_p_CQuery swig_types[58]
#define SWIGTYPE_p_CQuitMessage swig_types[59]
#define SWIGTYPE_p_CRealListener swig_types[60]
#define SWIGTYPE_p_CSCharBuffer swig_types[61]
#define SWIGTYPE_p_CSConnection swig_types[62]
#define SWIGTYPE_p_CSListener swig_types[63]
#define SWIGTYPE_p_CSMonitorFD swig_types[64]
#define SWIGTYPE_p_CSSSLConnection swig_types[65]
#define SWIGTYPE_p_CSSockAddr swig_types[66]
#define SWIGTYPE_p_CServer swig_types[67]
#define SWIGTYPE_p_CSockCommon swig_types[68]
#define SWIGTYPE_p_CSockManager swig_types[69]
#define SWIGTYPE_p_CSocket swig_types[70]
#define SWIGTYPE_p_CSocketManager swig_types[71]
#define SWIGTYPE_p_CString__EEscape swig_types[72]
#define SWIGTYPE_p_CString__size_type swig_types[73]
#define SWIGTYPE_p_CTCPListener swig_types[74]
#define SWIGTYPE_p_CTable swig_types[75]
#define SWIGTYPE_p_CTargetMessage swig_types[76]
#define SWIGTYPE_p_CTemplate swig_types[77]
#define SWIGTYPE_p_CTemplateLoopContext swig_types[78]
#define SWIGTYPE_p_CTemplateOptions swig_types[79]
#define SWIGTYPE_p_CTemplateTagHandler swig_types[80]
#define SWIGTYPE_p_CTextMessage swig_types[81]
#define SWIGTYPE_p_CTimer swig_types[82]
#define SWIGTYPE_p_CTopicMessage swig_types[83]
#define SWIGTYPE_p_CTranslation swig_types[84]
#define SWIGTYPE_p_CTranslationDomainRefHolder swig_types[85]
#define SWIGTYPE_p_CTranslationInfo swig_types[86]
#define SWIGTYPE_p_CUnixListener swig_types[87]
#define SWIGTYPE_p_CUser swig_types[88]
#define SWIGTYPE_p_CUtils swig_types[89]
#define SWIGTYPE_p_CWebSession swig_types[90]
#define SWIGTYPE_p_CWebSessionMap swig_types[91]
#define SWIGTYPE_p_CWebSock swig_types[92]
#define SWIGTYPE_p_CWebSubPage swig_types[93]
#define SWIGTYPE_p_CZNC swig_types[94]
#define SWIGTYPE_p_CZNCSock swig_types[95]
#define SWIGTYPE_p_CZNCTagHandler swig_types[96]
#define SWIGTYPE_p_CmdFunc swig_types[97]
#define SWIGTYPE_p_Csock swig_types[98]
#define SWIGTYPE_p_EAcceptType swig_types[99]
#define SWIGTYPE_p_EChanModeArgs swig_types[100]
#define SWIGTYPE_p_EModException swig_types[101]
#define SWIGTYPE_p_EModRet swig_types[102]
#define SWIGTYPE_p_EModes swig_types[103]
#define SWIGTYPE_p_EModuleType swig_types[104]
#define SWIGTYPE_p_EType swig_types[105]
#define SWIGTYPE_p_EUserPerms swig_types[106]
#define SWIGTYPE_p_EntryMap swig_types[107]
#define SWIGTYPE_p_EntryMapIterator swig_types[108]
#define SWIGTYPE_p_MCString swig_types[109]
#define SWIGTYPE_p_MCString_iter swig_types[110]
#define SWIGTYPE_p_ModCmdFunc swig_types[111]
#define SWIGTYPE_p_ModDirList swig_types[112]
#define SWIGTYPE_p_String swig_types[113]
#define SWIGTYPE_p_SubConfig swig_types[114]
#define SWIGTYPE_p_SubConfigEntry swig_types[115]
#define SWIGTYPE_p_SubConfigMap swig_types[116]
#define SWIGTYPE_p_SubConfigMapIterator swig_types[117]
#define SWIGTYPE_p_TSocketManagerT_CZNCSock_t swig_types[118]
#define SWIGTYPE_p_TrafficStatsMap swig_types[119]
#define SWIGTYPE_p_TrafficStatsPair swig_types[120]
#define SWIGTYPE_p_allocator_type swig_types[121]
#define SWIGTYPE_p_bool swig_types[122]
#define SWIGTYPE_p_char swig_types[123]
#define SWIGTYPE_p_difference_type swig_types[124]
#define SWIGTYPE_p_double swig_types[125]
#define SWIGTYPE_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModPython swig_types[126]
#define SWIGTYPE_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModule swig_types[127]
#define SWIGTYPE_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CPyModule swig_types[128]
#define SWIGTYPE_p_f_p_CModule_p_CFPTimer__void swig_types[129]
#define SWIGTYPE_p_f_r_CModInfo__void swig_types[130]
#define SWIGTYPE_p_f_r_std__ostream__r_std__ostream swig_types[131]
#define SWIGTYPE_p_fd_set swig_types[132]
#define SWIGTYPE_p_first_type swig_types[133]
#define SWIGTYPE_p_gid_t swig_types[134]
#define SWIGTYPE_p_in_addr swig_types[135]
#define SWIGTYPE_p_int swig_types[136]
#define SWIGTYPE_p_int32_t swig_types[137]
#define SWIGTYPE_p_int64_t swig_types[138]
#define SWIGTYPE_p_key_type swig_types[139]
#define SWIGTYPE_p_long swig_types[140]
#define SWIGTYPE_p_mapped_type swig_types[141]
#define SWIGTYPE_p_mode_t swig_types[142]
#define SWIGTYPE_p_p_PyObject swig_types[143]
#define SWIGTYPE_p_second_type swig_types[144]
#define SWIGTYPE_p_size_type swig_types[145]
#define SWIGTYPE_p_sockaddr_in swig_types[146]
#define SWIGTYPE_p_sockaddr_storage swig_types[147]
#define SWIGTYPE_p_socklen_t swig_types[148]
#define SWIGTYPE_p_ssize_t swig_types[149]
#define SWIGTYPE_p_stat swig_types[150]
#define SWIGTYPE_p_std__allocatorT_CBufLine_t swig_types[151]
#define SWIGTYPE_p_std__allocatorT_CChan_p_t swig_types[152]
#define SWIGTYPE_p_std__allocatorT_CClient_p_t swig_types[153]
#define SWIGTYPE_p_std__allocatorT_CIRCNetwork_p_t swig_types[154]
#define SWIGTYPE_p_std__allocatorT_CListener_p_t swig_types[155]
#define SWIGTYPE_p_std__allocatorT_CModule_p_t swig_types[156]
#define SWIGTYPE_p_std__allocatorT_CQuery_p_t swig_types[157]
#define SWIGTYPE_p_std__allocatorT_CServer_p_t swig_types[158]
#define SWIGTYPE_p_std__allocatorT_CString_t swig_types[159]
#define SWIGTYPE_p_std__allocatorT_std__pairT_CString_CString_t_t swig_types[160]
#define SWIGTYPE_p_std__allocatorT_std__pairT_CString_const_CNick_t_t swig_types[161]
#define SWIGTYPE_p_std__allocatorT_std__pairT_CString_const_CString_t_t swig_types[162]
#define SWIGTYPE_p_std__allocatorT_std__pairT_CString_const_std__vectorT_CString_t_t_t swig_types[163]
#define SWIGTYPE_p_std__allocatorT_std__shared_ptrT_CWebSubPage_t_t swig_types[164]
#define SWIGTYPE_p_std__allocatorT_std__vectorT_CString_t_t swig_types[165]
#define SWIGTYPE_p_std__dequeT_CBufLine_t swig_types[166]
#define SWIGTYPE_p_std__functionT_void_fCString_const_RF_t swig_types[167]
#define SWIGTYPE_p_std__invalid_argument swig_types[168]
#define SWIGTYPE_p_std__lessT_CModInfo_t swig_types[169]
#define SWIGTYPE_p_std__lessT_CString_t swig_types[170]
#define SWIGTYPE_p_std__listT_CIRCNetwork_p_std__allocatorT_CIRCNetwork_p_t_t swig_types[171]
#define SWIGTYPE_p_std__listT_CString_t swig_types[172]
#define SWIGTYPE_p_std__mapT_CString_CNick_t swig_types[173]
#define SWIGTYPE_p_std__mapT_CString_CString_t swig_types[174]
#define SWIGTYPE_p_std__mapT_CString_CString_t__iterator swig_types[175]
#define SWIGTYPE_p_std__mapT_CString_CTranslationInfo_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CTranslationInfo_t_t_t swig_types[176]
#define SWIGTYPE_p_std__mapT_CString_CUser_p_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CUser_p_t_t_t swig_types[177]
#define SWIGTYPE_p_std__mapT_CString_VCString_t swig_types[178]
#define SWIGTYPE_p_std__mapT_CString_std__pairT_unsigned_long_long_unsigned_long_long_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__pairT_unsigned_long_long_unsigned_long_long_t_t_t_t swig_types[179]
#define SWIGTYPE_p_std__mapT_CString_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_t_t_t__const_iterator swig_types[180]
#define SWIGTYPE_p_std__mapT_char_CIRCSock__EChanModeArgs_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CIRCSock__EChanModeArgs_t_t_t swig_types[181]
#define SWIGTYPE_p_std__mapT_char_CString_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CString_t_t_t swig_types[182]
#define SWIGTYPE_p_std__mapT_char_unsigned_int_std__lessT_char_t_std__allocatorT_std__pairT_char_const_unsigned_int_t_t_t swig_types[183]
#define SWIGTYPE_p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t swig_types[184]
#define SWIGTYPE_p_std__ostream swig_types[185]
#define SWIGTYPE_p_std__pairT_CString_CString_t swig_types[186]
#define SWIGTYPE_p_std__pairT_unsigned_long_long_unsigned_long_long_t swig_types[187]
#define SWIGTYPE_p_std__setT_CChan_p_std__lessT_CChan_p_t_std__allocatorT_CChan_p_t_t swig_types[188]
#define SWIGTYPE_p_std__setT_CModInfo_t swig_types[189]
#define SWIGTYPE_p_std__setT_CSocket_p_std__lessT_CSocket_p_t_std__allocatorT_CSocket_p_t_t__const_iterator swig_types[190]
#define SWIGTYPE_p_std__setT_CString_t swig_types[191]
#define SWIGTYPE_p_std__setT_CTimer_p_std__lessT_CTimer_p_t_std__allocatorT_CTimer_p_t_t__const_iterator swig_types[192]
#define SWIGTYPE_p_std__setT_char_std__lessT_char_t_std__allocatorT_char_t_t swig_types[193]
#define SWIGTYPE_p_std__shared_ptrT_CAuthBase_t swig_types[194]
#define SWIGTYPE_p_std__shared_ptrT_CClientAuth_t swig_types[195]
#define SWIGTYPE_p_std__shared_ptrT_CTemplateOptions_t swig_types[196]
#define SWIGTYPE_p_std__shared_ptrT_CTemplateTagHandler_t swig_types[197]
#define SWIGTYPE_p_std__shared_ptrT_CWebSession_t swig_types[198]
#define SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t swig_types[199]
#define SWIGTYPE_p_std__vectorT_CChan_p_t swig_types[200]
#define SWIGTYPE_p_std__vectorT_CClient_p_t swig_types[201]
#define SWIGTYPE_p_std__vectorT_CCron_p_std__allocatorT_CCron_p_t_t swig_types[202]
#define SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t swig_types[203]
#define SWIGTYPE_p_std__vectorT_CListener_p_t swig_types[204]
#define SWIGTYPE_p_std__vectorT_CModule_p_t swig_types[205]
#define SWIGTYPE_p_std__vectorT_CModule_p_t__const_iterator swig_types[206]
#define SWIGTYPE_p_std__vectorT_CQuery_p_t swig_types[207]
#define SWIGTYPE_p_std__vectorT_CServer_p_t swig_types[208]
#define SWIGTYPE_p_std__vectorT_CString_t swig_types[209]
#define SWIGTYPE_p_std__vectorT_CTemplate_p_std__allocatorT_CTemplate_p_t_t swig_types[210]
#define SWIGTYPE_p_std__vectorT_Csock_p_std__allocatorT_Csock_p_t_t swig_types[211]
#define SWIGTYPE_p_std__vectorT_TWebSubPage_t swig_types[212]
#define SWIGTYPE_p_std__vectorT_VCString_t swig_types[213]
#define SWIGTYPE_p_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t swig_types[214]
#define SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t swig_types[215]
#define SWIGTYPE_p_std__vectorT_std__shared_ptrT_CTemplateTagHandler_t_std__allocatorT_std__shared_ptrT_CTemplateTagHandler_t_t_t swig_types[216]
#define SWIGTYPE_p_swig__SwigPyIterator swig_types[217]
#define SWIGTYPE_p_time_t swig_types[218]
#define SWIGTYPE_p_timeval swig_types[219]
#define SWIGTYPE_p_uid_t swig_types[220]
#define SWIGTYPE_p_uint16_t swig_types[221]
#define SWIGTYPE_p_uint64_t swig_types[222]
#define SWIGTYPE_p_unsigned_int swig_types[223]
#define SWIGTYPE_p_unsigned_short swig_types[224]
#define SWIGTYPE_p_value_type swig_types[225]
#define SWIGTYPE_p_void swig_types[226]
static swig_type_info *swig_types[228];
static swig_module_info swig_module = {swig_types, 227, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#ifdef SWIG_TypeQuery
# undef SWIG_TypeQuery
#endif
#define SWIG_TypeQuery SWIG_Python_TypeQuery

/*-----------------------------------------------
              @(target):= _znc_core.so
  ------------------------------------------------*/
#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_init    PyInit__znc_core

#else
#  define SWIG_init    init_znc_core

#endif
#define SWIG_name    "_znc_core"

#ifdef __cplusplus
#include <utility>
/* SwigValueWrapper is described in swig.swg */
template<typename T> class SwigValueWrapper {
  struct SwigSmartPointer {
    T *ptr;
    SwigSmartPointer(T *p) : ptr(p) { }
    ~SwigSmartPointer() { delete ptr; }
    SwigSmartPointer& operator=(SwigSmartPointer& rhs) { T* oldptr = ptr; ptr = 0; delete oldptr; ptr = rhs.ptr; rhs.ptr = 0; return *this; }
    void reset(T *p) { T* oldptr = ptr; ptr = 0; delete oldptr; ptr = p; }
  } pointer;
  SwigValueWrapper& operator=(const SwigValueWrapper<T>& rhs);
  SwigValueWrapper(const SwigValueWrapper<T>& rhs);
public:
  SwigValueWrapper() : pointer(0) { }
  SwigValueWrapper& operator=(const T& t) { SwigSmartPointer tmp(new T(t)); pointer = tmp; return *this; }
#if __cplusplus >=201103L
  SwigValueWrapper& operator=(T&& t) { SwigSmartPointer tmp(new T(std::move(t))); pointer = tmp; return *this; }
  operator T&&() const { return std::move(*pointer.ptr); }
#else
  operator T&() const { return *pointer.ptr; }
#endif
  T *operator&() const { return pointer.ptr; }
  static void reset(SwigValueWrapper& t, T *p) { t.pointer.reset(p); }
};

/*
 * SwigValueInit() is a generic initialisation solution as the following approach:
 * 
 *       T c_result = T();
 * 
 * doesn't compile for all types for example:
 * 
 *       unsigned int c_result = unsigned int();
 */
template <typename T> T SwigValueInit() {
  return T();
}

#if __cplusplus >=201103L
# define SWIG_STD_MOVE(OBJ) std::move(OBJ)
#else
# define SWIG_STD_MOVE(OBJ) OBJ
#endif

#endif


#define SWIG_as_voidptr(a) const_cast< void * >(static_cast< const void * >(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),reinterpret_cast< void** >(a)) 


#include <stdexcept>


namespace swig {
  class SwigPtr_PyObject {
  protected:
    PyObject *_obj;

  public:
    SwigPtr_PyObject() :_obj(0)
    {
    }

    SwigPtr_PyObject(const SwigPtr_PyObject& item) : _obj(item._obj)
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      Py_XINCREF(_obj);      
      SWIG_PYTHON_THREAD_END_BLOCK;
    }
    
    SwigPtr_PyObject(PyObject *obj, bool initial_ref = true) :_obj(obj)
    {
      if (initial_ref) {
        SWIG_PYTHON_THREAD_BEGIN_BLOCK;
        Py_XINCREF(_obj);
        SWIG_PYTHON_THREAD_END_BLOCK;
      }
    }
    
    SwigPtr_PyObject & operator=(const SwigPtr_PyObject& item) 
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      Py_XINCREF(item._obj);
      Py_XDECREF(_obj);
      _obj = item._obj;
      SWIG_PYTHON_THREAD_END_BLOCK;
      return *this;      
    }
    
    ~SwigPtr_PyObject() 
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      Py_XDECREF(_obj);
      SWIG_PYTHON_THREAD_END_BLOCK;
    }
    
    operator PyObject *() const
    {
      return _obj;
    }

    PyObject *operator->() const
    {
      return _obj;
    }
  };
}


namespace swig {
  struct SwigVar_PyObject : SwigPtr_PyObject {
    SwigVar_PyObject(PyObject* obj = 0) : SwigPtr_PyObject(obj, false) { }
    
    SwigVar_PyObject & operator = (PyObject* obj)
    {
      Py_XDECREF(_obj);
      _obj = obj;
      return *this;      
    }
  };
}


#include <utility>
#include "znc/Utils.h"
#include "znc/Threads.h"
#include "znc/Config.h"
#include "znc/Socket.h"
#include "znc/Modules.h"
#include "znc/Nick.h"
#include "znc/Chan.h"
#include "znc/User.h"
#include "znc/IRCNetwork.h"
#include "znc/Query.h"
#include "znc/Client.h"
#include "znc/IRCSock.h"
#include "znc/Listener.h"
#include "znc/HTTPSock.h"
#include "znc/Template.h"
#include "znc/WebModules.h"
#include "znc/znc.h"
#include "znc/Server.h"
#include "znc/ZNCString.h"
#include "znc/FileUtils.h"
#include "znc/ZNCDebug.h"
#include "znc/ExecSock.h"
#include "znc/Buffer.h"
#include "modpython/module.h"

#include "modpython/ret.h"

#define stat struct stat
using std::allocator;


#include <typeinfo>
#include <stdexcept>


#if defined(__GNUC__)
#  if __GNUC__ == 2 && __GNUC_MINOR <= 96
#     define SWIG_STD_NOMODERN_STL
#  endif
#endif


#include <string>


#include <stddef.h>


#include <iostream>

#if PY_VERSION_HEX >= 0x03020000
# define SWIGPY_SLICEOBJECT PyObject
#else
# define SWIGPY_SLICEOBJECT PySliceObject
#endif


namespace swig {
  struct stop_iteration {
  };

  struct SwigPyIterator {
  private:
    SwigPtr_PyObject _seq;

  protected:
    SwigPyIterator(PyObject *seq) : _seq(seq)
    {
    }
      
  public:
    virtual ~SwigPyIterator() {}

    // Access iterator method, required by Python
    virtual PyObject *value() const = 0;

    // Forward iterator method, required by Python
    virtual SwigPyIterator *incr(size_t n = 1) = 0;
    
    // Backward iterator method, very common in C++, but not required in Python
    virtual SwigPyIterator *decr(size_t /*n*/ = 1)
    {
      throw stop_iteration();
    }

    // Random access iterator methods, but not required in Python
    virtual ptrdiff_t distance(const SwigPyIterator &/*x*/) const
    {
      throw std::invalid_argument("operation not supported");
    }

    virtual bool equal (const SwigPyIterator &/*x*/) const
    {
      throw std::invalid_argument("operation not supported");
    }
    
    // C++ common/needed methods
    virtual SwigPyIterator *copy() const = 0;

    PyObject *next()     
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK; // disable threads       
      PyObject *obj = value();
      incr();       
      SWIG_PYTHON_THREAD_END_BLOCK; // re-enable threads
      return obj;     
    }

    /* Make an alias for Python 3.x */
    PyObject *__next__()
    {
      return next();
    }

    PyObject *previous()
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK; // disable threads       
      decr();
      PyObject *obj = value();
      SWIG_PYTHON_THREAD_END_BLOCK; // re-enable threads       
      return obj;
    }

    SwigPyIterator *advance(ptrdiff_t n)
    {
      return  (n > 0) ?  incr(n) : decr(-n);
    }
      
    bool operator == (const SwigPyIterator& x)  const
    {
      return equal(x);
    }
      
    bool operator != (const SwigPyIterator& x) const
    {
      return ! operator==(x);
    }
      
    SwigPyIterator& operator += (ptrdiff_t n)
    {
      return *advance(n);
    }

    SwigPyIterator& operator -= (ptrdiff_t n)
    {
      return *advance(-n);
    }
      
    SwigPyIterator* operator + (ptrdiff_t n) const
    {
      return copy()->advance(n);
    }

    SwigPyIterator* operator - (ptrdiff_t n) const
    {
      return copy()->advance(-n);
    }
      
    ptrdiff_t operator - (const SwigPyIterator& x) const
    {
      return x.distance(*this);
    }
      
    static swig_type_info* descriptor() {
      static int init = 0;
      static swig_type_info* desc = 0;
      if (!init) {
	desc = SWIG_TypeQuery("swig::SwigPyIterator *");
	init = 1;
      }	
      return desc;
    }    
  };

#if defined(SWIGPYTHON_BUILTIN)
  inline PyObject* make_output_iterator_builtin (PyObject *pyself)
  {
    Py_INCREF(pyself);
    return pyself;
  }
#endif
}


SWIGINTERN int
SWIG_AsVal_double (PyObject *obj, double *val)
{
  int res = SWIG_TypeError;
  if (PyFloat_Check(obj)) {
    if (val) *val = PyFloat_AsDouble(obj);
    return SWIG_OK;
#if PY_VERSION_HEX < 0x03000000
  } else if (PyInt_Check(obj)) {
    if (val) *val = (double) PyInt_AsLong(obj);
    return SWIG_OK;
#endif
  } else if (PyLong_Check(obj)) {
    double v = PyLong_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    double d = PyFloat_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = d;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      long v = PyLong_AsLong(obj);
      if (!PyErr_Occurred()) {
	if (val) *val = v;
	return SWIG_AddCast(SWIG_AddCast(SWIG_OK));
      } else {
	PyErr_Clear();
      }
    }
  }
#endif
  return res;
}


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx, cx, rd;
   errno = 0;
   fx = floor(x);
   cx = ceil(x);
   rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (PyObject *obj, unsigned long *val) 
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    long v = PyInt_AsLong(obj);
    if (v >= 0) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      return SWIG_OverflowError;
    }
  } else
#endif
  if (PyLong_Check(obj)) {
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      // Largest double not larger than ULONG_MAX (not portably calculated easily)
      // Note that double(ULONG_MAX) is stored in a double rounded up by one (for 64-bit unsigned long)
      // 0xfffffffffffff800ULL == (uint64_t)std::nextafter(double(__uint128_t(ULONG_MAX)+1), double(0))
      const double ulong_max = sizeof(unsigned long) == 8 ? 0xfffffffffffff800ULL : ULONG_MAX;
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, ulong_max)) {
	if (val) *val = (unsigned long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


#if defined(LLONG_MAX) && !defined(SWIG_LONG_LONG_AVAILABLE)
#  define SWIG_LONG_LONG_AVAILABLE
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERN int
SWIG_AsVal_unsigned_SS_long_SS_long (PyObject *obj, unsigned long long *val)
{
  int res = SWIG_TypeError;
  if (PyLong_Check(obj)) {
    unsigned long long v = PyLong_AsUnsignedLongLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      res = SWIG_OverflowError;
    }
  } else {
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long (obj,&v);
    if (SWIG_IsOK(res)) {
      if (val) *val = v;
      return res;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    const double mant_max = 1LL << DBL_MANT_DIG;
    double d;
    res = SWIG_AsVal_double (obj,&d);
    if (SWIG_IsOK(res) && !SWIG_CanCastAsInteger(&d, 0, mant_max))
      return SWIG_OverflowError;
    if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, mant_max)) {
      if (val) *val = (unsigned long long)(d);
      return SWIG_AddCast(res);
    }
    res = SWIG_TypeError;
  }
#endif
  return res;
}
#endif


SWIGINTERNINLINE int
SWIG_AsVal_size_t (PyObject * obj, size_t *val)
{
  int res = SWIG_TypeError;
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = static_cast< size_t >(v);
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else if (sizeof(size_t) <= sizeof(unsigned long long)) {
    unsigned long long v;
    res = SWIG_AsVal_unsigned_SS_long_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = static_cast< size_t >(v);
  }
#endif
  return res;
}


  #define SWIG_From_long   PyInt_FromLong 


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE PyObject* 
SWIG_From_long_SS_long  (long long value)
{
  return ((value < LONG_MIN) || (value > LONG_MAX)) ?
    PyLong_FromLongLong(value) : PyInt_FromLong(static_cast< long >(value));
}
#endif


SWIGINTERNINLINE PyObject *
SWIG_From_ptrdiff_t  (ptrdiff_t value)
{    
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(ptrdiff_t) <= sizeof(long)) {
#endif
    return SWIG_From_long  (static_cast< long >(value));
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else {
    /* assume sizeof(ptrdiff_t) <= sizeof(long long) */
    return SWIG_From_long_SS_long  (static_cast< long long >(value));
  }
#endif
}


SWIGINTERNINLINE PyObject*
  SWIG_From_bool  (bool value)
{
  return PyBool_FromLong(value ? 1 : 0);
}


SWIGINTERN int
SWIG_AsVal_long (PyObject *obj, long* val)
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else
#endif
  if (PyLong_Check(obj)) {
    long v = PyLong_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    long v = PyInt_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      // Largest double not larger than LONG_MAX (not portably calculated easily)
      // Note that double(LONG_MAX) is stored in a double rounded up by one (for 64-bit long)
      // 0x7ffffffffffffc00LL == (int64_t)std::nextafter(double(__uint128_t(LONG_MAX)+1), double(0))
      const double long_max = sizeof(long) == 8 ? 0x7ffffffffffffc00LL : LONG_MAX;
      // No equivalent needed for 64-bit double(LONG_MIN) is exactly LONG_MIN
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, long_max)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERN int
SWIG_AsVal_long_SS_long (PyObject *obj, long long *val)
{
  int res = SWIG_TypeError;
  if (PyLong_Check(obj)) {
    long long v = PyLong_AsLongLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      res = SWIG_OverflowError;
    }
  } else {
    long v;
    res = SWIG_AsVal_long (obj,&v);
    if (SWIG_IsOK(res)) {
      if (val) *val = v;
      return res;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    const double mant_max = 1LL << DBL_MANT_DIG;
    const double mant_min = -mant_max;
    double d;
    res = SWIG_AsVal_double (obj,&d);
    if (SWIG_IsOK(res) && !SWIG_CanCastAsInteger(&d, mant_min, mant_max))
      return SWIG_OverflowError;
    if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, mant_min, mant_max)) {
      if (val) *val = (long long)(d);
      return SWIG_AddCast(res);
    }
    res = SWIG_TypeError;
  }
#endif
  return res;
}
#endif


SWIGINTERNINLINE int
SWIG_AsVal_ptrdiff_t (PyObject * obj, ptrdiff_t *val)
{
  int res = SWIG_TypeError;
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(ptrdiff_t) <= sizeof(long)) {
#endif
    long v;
    res = SWIG_AsVal_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = static_cast< ptrdiff_t >(v);
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else if (sizeof(ptrdiff_t) <= sizeof(long long)) {
    long long v;
    res = SWIG_AsVal_long_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = static_cast< ptrdiff_t >(v);
  }
#endif
  return res;
}


#include <algorithm>


#include <vector>


#include <utility>


#include <map>


#include <algorithm>


#include <list>


#include <set>


#include <deque>


SWIGINTERNINLINE PyObject*
  SWIG_From_int  (int value)
{
  return PyInt_FromLong((long) value);
}


namespace swig {
  template <class Type>
  struct noconst_traits {
    typedef Type noconst_type;
  };

  template <class Type>
  struct noconst_traits<const Type> {
    typedef Type noconst_type;
  };

  /*
    type categories
  */
  struct pointer_category { };
  struct value_category { };

  /*
    General traits that provides type_name and type_info
  */
  template <class Type> struct traits { };

  template <class Type>
  inline const char* type_name() {
    return traits<typename noconst_traits<Type >::noconst_type >::type_name();
  }

  template <class Type> struct traits_info {
    static swig_type_info *type_query(std::string name) {
      name += " *";
      return SWIG_TypeQuery(name.c_str());
    }
    static swig_type_info *type_info() {
      static swig_type_info *info = type_query(type_name<Type>());
      return info;
    }
  };

  /*
    Partial specialization for pointers (traits_info)
  */
  template <class Type> struct traits_info<Type *> {
    static swig_type_info *type_query(std::string name) {
      name += " *";
      return SWIG_TypeQuery(name.c_str());
    }
    static swig_type_info *type_info() {
      static swig_type_info *info = type_query(type_name<Type>());
      return info;
    }
  };

  template <class Type>
  inline swig_type_info *type_info() {
    return traits_info<Type>::type_info();
  }

  /*
    Partial specialization for pointers (traits)
  */
  template <class Type> struct traits <Type *> {
    typedef pointer_category category;
    static std::string make_ptr_name(const char* name) {
      std::string ptrname = name;
      ptrname += " *";
      return ptrname;
    }
    static const char* type_name() {
      static std::string name = make_ptr_name(swig::type_name<Type>());
      return name.c_str();
    }
  };

  template <class Type, class Category>
  struct traits_as { };

  template <class Type, class Category>
  struct traits_check { };

}


namespace swig {  
  /*
    Traits that provides the from method
  */
  template <class Type> struct traits_from_ptr {
    static PyObject *from(Type *val, int owner = 0) {
      return SWIG_InternalNewPointerObj(val, type_info<Type>(), owner);
    }
  };

  template <class Type> struct traits_from {
    static PyObject *from(const Type& val) {
      return traits_from_ptr<Type>::from(new Type(val), 1);
    }
  };

  template <class Type> struct traits_from<Type *> {
    static PyObject *from(Type* val) {
      return traits_from_ptr<Type>::from(val, 0);
    }
  };

  template <class Type> struct traits_from<const Type *> {
    static PyObject *from(const Type* val) {
      return traits_from_ptr<Type>::from(const_cast<Type*>(val), 0);
    }
  };


  template <class Type>
  inline PyObject *from(const Type& val) {
    return traits_from<Type>::from(val);
  }

  template <class Type>
  inline PyObject *from_ptr(Type* val, int owner) {
    return traits_from_ptr<Type>::from(val, owner);
  }

  /*
    Traits that provides the asval/as/check method
  */
  template <class Type>
  struct traits_asptr {   
    static int asptr(PyObject *obj, Type **val) {
      int res = SWIG_ERROR;
      swig_type_info *descriptor = type_info<Type>();
      if (val) {
        Type *p = 0;
        int newmem = 0;
        res = descriptor ? SWIG_ConvertPtrAndOwn(obj, (void **)&p, descriptor, 0, &newmem) : SWIG_ERROR;
        if (SWIG_IsOK(res)) {
          if (newmem & SWIG_CAST_NEW_MEMORY) {
            res |= SWIG_NEWOBJMASK;
          }
          *val = p;
        }
      } else {
        res = descriptor ? SWIG_ConvertPtr(obj, 0, descriptor, 0) : SWIG_ERROR;
      }
      return res;
    }
  }; 

  template <class Type>
  inline int asptr(PyObject *obj, Type **vptr) {
    return traits_asptr<Type>::asptr(obj, vptr);
  }

  template <class Type> 
  struct traits_asval {
    static int asval(PyObject *obj, Type *val) {
      if (val) {
	Type *p = 0;
	int res = traits_asptr<Type>::asptr(obj, &p);
	if (!SWIG_IsOK(res)) return res;	
	if (p) {
	  typedef typename noconst_traits<Type>::noconst_type noconst_type;
	  *(const_cast<noconst_type*>(val)) = *p;
	  if (SWIG_IsNewObj(res)){
	    delete p;
	    res = SWIG_DelNewMask(res);
	  }
	  return res;
	} else {
	  return SWIG_ERROR;
	}
      } else {
	return traits_asptr<Type>::asptr(obj, (Type **)(0));
      }
    }
  };

  template <class Type> struct traits_asval<Type*> {
    static int asval(PyObject *obj, Type **val) {
      if (val) {
        typedef typename noconst_traits<Type>::noconst_type noconst_type;
        noconst_type *p = 0;
        int res = traits_asptr<noconst_type>::asptr(obj,  &p);
        if (SWIG_IsOK(res)) {
          *(const_cast<noconst_type**>(val)) = p;
	}
	return res;
      } else {
	return traits_asptr<Type>::asptr(obj, (Type **)(0));
      }
    }
  };
  
  template <class Type>
  inline int asval(PyObject *obj, Type *val) {
    return traits_asval<Type>::asval(obj, val);
  }

  template <class Type> 
  struct traits_as<Type, value_category> {
    static Type as(PyObject *obj) {
      Type v;
      int res = asval(obj, &v);
      if (!obj || !SWIG_IsOK(res)) {
	if (!PyErr_Occurred()) {
	  ::SWIG_Error(SWIG_TypeError,  swig::type_name<Type>());
	}
	throw std::invalid_argument("bad type");
      }
      return v;
    }
  };

  template <class Type> 
  struct traits_as<Type, pointer_category> {
    static Type as(PyObject *obj) {
      Type *v = 0;      
      int res = (obj ? traits_asptr<Type>::asptr(obj, &v) : SWIG_ERROR);
      if (SWIG_IsOK(res) && v) {
	if (SWIG_IsNewObj(res)) {
	  Type r(*v);
	  delete v;
	  return r;
	} else {
	  return *v;
	}
      } else {
	if (!PyErr_Occurred()) {
	  SWIG_Error(SWIG_TypeError,  swig::type_name<Type>());
	}
	throw std::invalid_argument("bad type");
      }
    }
  };

  template <class Type> 
  struct traits_as<Type*, pointer_category> {
    static Type* as(PyObject *obj) {
      Type *v = 0;      
      int res = (obj ? traits_asptr<Type>::asptr(obj, &v) : SWIG_ERROR);
      if (SWIG_IsOK(res)) {
	return v;
      } else {
	if (!PyErr_Occurred()) {
	  SWIG_Error(SWIG_TypeError,  swig::type_name<Type>());
	}
	throw std::invalid_argument("bad type");
      }
    }
  };
    
  template <class Type>
  inline Type as(PyObject *obj) {
    return traits_as<Type, typename traits<Type>::category>::as(obj);
  }

  template <class Type> 
  struct traits_check<Type, value_category> {
    static bool check(PyObject *obj) {
      int res = obj ? asval(obj, (Type *)(0)) : SWIG_ERROR;
      return SWIG_IsOK(res) ? true : false;
    }
  };

  template <class Type> 
  struct traits_check<Type, pointer_category> {
    static bool check(PyObject *obj) {
      int res = obj ? asptr(obj, (Type **)(0)) : SWIG_ERROR;
      return SWIG_IsOK(res) ? true : false;
    }
  };

  template <class Type>
  inline bool check(PyObject *obj) {
    return traits_check<Type, typename traits<Type>::category>::check(obj);
  }
}


#if defined(__SUNPRO_CC) && defined(_RWSTD_VER)
#  if !defined(SWIG_NO_STD_NOITERATOR_TRAITS_STL)
#    define SWIG_STD_NOITERATOR_TRAITS_STL
#  endif
#endif

#if !defined(SWIG_STD_NOITERATOR_TRAITS_STL)
#include <iterator>
#else
namespace std {
  template <class Iterator>
  struct iterator_traits {
    typedef ptrdiff_t difference_type;
    typedef typename Iterator::value_type value_type;
  };

  template <class Iterator, class Category,class T, class Reference, class Pointer, class Distance>
  struct iterator_traits<__reverse_bi_iterator<Iterator,Category,T,Reference,Pointer,Distance> > {
    typedef Distance difference_type;
    typedef T value_type;
  };

  template <class T>
  struct iterator_traits<T*> {
    typedef T value_type;
    typedef ptrdiff_t difference_type;
  };

  template<typename _InputIterator>
  inline typename iterator_traits<_InputIterator>::difference_type
  distance(_InputIterator __first, _InputIterator __last)
  {
    typename iterator_traits<_InputIterator>::difference_type __n = 0;
    while (__first != __last) {
      ++__first; ++__n;
    }
    return __n;
  }
}
#endif


namespace swig {
  template<typename OutIterator>
  class SwigPyIterator_T :  public SwigPyIterator
  {
  public:
    typedef OutIterator out_iterator;
    typedef typename std::iterator_traits<out_iterator>::value_type value_type;    
    typedef SwigPyIterator_T<out_iterator> self_type;

    SwigPyIterator_T(out_iterator curr, PyObject *seq)
      : SwigPyIterator(seq), current(curr)
    {
    }

    const out_iterator& get_current() const
    {
      return current;
    }

    
    bool equal (const SwigPyIterator &iter) const
    {
      const self_type *iters = dynamic_cast<const self_type *>(&iter);
      if (iters) {
	return (current == iters->get_current());
      } else {
	throw std::invalid_argument("bad iterator type");
      }
    }
    
    ptrdiff_t distance(const SwigPyIterator &iter) const
    {
      const self_type *iters = dynamic_cast<const self_type *>(&iter);
      if (iters) {
	return std::distance(current, iters->get_current());
      } else {
	throw std::invalid_argument("bad iterator type");
      }
    }    
    
  protected:
    out_iterator current;
  };
  
  template <class ValueType>
  struct from_oper 
  {
    typedef const ValueType& argument_type;
    typedef PyObject *result_type;
    result_type operator()(argument_type v) const
    {
      return swig::from(v);
    }
  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class SwigPyForwardIteratorOpen_T :  public SwigPyIterator_T<OutIterator>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef SwigPyIterator_T<out_iterator>  base;
    typedef SwigPyForwardIteratorOpen_T<OutIterator, ValueType, FromOper> self_type;
    
    SwigPyForwardIteratorOpen_T(out_iterator curr, PyObject *seq)
      : SwigPyIterator_T<OutIterator>(curr, seq)
    {
    }
    
    PyObject *value() const {
      return from(static_cast<const value_type&>(*(base::current)));
    }
    
    SwigPyIterator *copy() const
    {
      return new self_type(*this);
    }

    SwigPyIterator *incr(size_t n = 1)
    {
      while (n--) {
	++base::current;
      }
      return this;
    }

  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class SwigPyIteratorOpen_T :  public SwigPyForwardIteratorOpen_T<OutIterator, ValueType, FromOper>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef SwigPyIterator_T<out_iterator>  base;
    typedef SwigPyIteratorOpen_T<OutIterator, ValueType, FromOper> self_type;
    
    SwigPyIteratorOpen_T(out_iterator curr, PyObject *seq)
      : SwigPyForwardIteratorOpen_T<OutIterator>(curr, seq)
    {
    }

    SwigPyIterator *decr(size_t n = 1)
    {
      while (n--) {
	--base::current;
      }
      return this;
    }
  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class SwigPyForwardIteratorClosed_T :  public SwigPyIterator_T<OutIterator>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef SwigPyIterator_T<out_iterator>  base;    
    typedef SwigPyForwardIteratorClosed_T<OutIterator, ValueType, FromOper> self_type;
    
    SwigPyForwardIteratorClosed_T(out_iterator curr, out_iterator first, out_iterator last, PyObject *seq)
      : SwigPyIterator_T<OutIterator>(curr, seq), begin(first), end(last)
    {
    }
    
    PyObject *value() const {
      if (base::current == end) {
	throw stop_iteration();
      } else {
	return from(static_cast<const value_type&>(*(base::current)));
      }
    }
    
    SwigPyIterator *copy() const
    {
      return new self_type(*this);
    }

    SwigPyIterator *incr(size_t n = 1)
    {
      while (n--) {
	if (base::current == end) {
	  throw stop_iteration();
	} else {
	  ++base::current;
	}
      }
      return this;
    }

  protected:
    out_iterator begin;
    out_iterator end;
  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class SwigPyIteratorClosed_T :  public SwigPyForwardIteratorClosed_T<OutIterator,ValueType,FromOper>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef SwigPyIterator_T<out_iterator>  base;
    typedef SwigPyForwardIteratorClosed_T<OutIterator, ValueType, FromOper> base0;
    typedef SwigPyIteratorClosed_T<OutIterator, ValueType, FromOper> self_type;
    
    SwigPyIteratorClosed_T(out_iterator curr, out_iterator first, out_iterator last, PyObject *seq)
      : SwigPyForwardIteratorClosed_T<OutIterator,ValueType,FromOper>(curr, first, last, seq)
    {
    }

    SwigPyIterator *decr(size_t n = 1)
    {
      while (n--) {
	if (base::current == base0::begin) {
	  throw stop_iteration();
	} else {
	  --base::current;
	}
      }
      return this;
    }
  };


  template<typename OutIter>
  inline SwigPyIterator*
  make_output_forward_iterator(const OutIter& current, const OutIter& begin,const OutIter& end, PyObject *seq = 0)
  {
    return new SwigPyForwardIteratorClosed_T<OutIter>(current, begin, end, seq);
  }

  template<typename OutIter>
  inline SwigPyIterator*
  make_output_iterator(const OutIter& current, const OutIter& begin,const OutIter& end, PyObject *seq = 0)
  {
    return new SwigPyIteratorClosed_T<OutIter>(current, begin, end, seq);
  }

  template<typename OutIter>
  inline SwigPyIterator*
  make_output_forward_iterator(const OutIter& current, PyObject *seq = 0)
  {
    return new SwigPyForwardIteratorOpen_T<OutIter>(current, seq);
  }

  template<typename OutIter>
  inline SwigPyIterator*
  make_output_iterator(const OutIter& current, PyObject *seq = 0)
  {
    return new SwigPyIteratorOpen_T<OutIter>(current, seq);
  }

}


SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static int init = 0;
  static swig_type_info* info = 0;
  if (!init) {
    info = SWIG_TypeQuery("_p_char");
    init = 1;
  }
  return info;
}


/* Return string from Python obj. NOTE: obj must remain in scope in order
   to use the returned cptr (but only when alloc is set to SWIG_OLDOBJ) */
SWIGINTERN int
SWIG_AsCharPtrAndSize(PyObject *obj, char **cptr, size_t *psize, int *alloc)
{
#if PY_VERSION_HEX>=0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
  if (PyBytes_Check(obj))
#else
  if (PyUnicode_Check(obj))
#endif
#else  
  if (PyString_Check(obj))
#endif
  {
    char *cstr; Py_ssize_t len;
    PyObject *bytes = NULL;
    int ret = SWIG_OK;
    if (alloc)
      *alloc = SWIG_OLDOBJ;
#if PY_VERSION_HEX>=0x03000000 && defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
    if (PyBytes_AsStringAndSize(obj, &cstr, &len) == -1)
      return SWIG_TypeError;
#else
    cstr = (char *)SWIG_PyUnicode_AsUTF8AndSize(obj, &len, &bytes);
    if (!cstr)
      return SWIG_TypeError;
    /* The returned string is only duplicated if the char * returned is not owned and memory managed by obj */
    if (bytes && cptr) {
      if (alloc) {
        cstr = reinterpret_cast< char* >(memcpy(new char[len + 1], cstr, sizeof(char)*(len + 1)));
        *alloc = SWIG_NEWOBJ;
      } else {
        /* alloc must be set in order to clean up allocated memory */
        return SWIG_RuntimeError;
      }
    }
#endif
    if (cptr) *cptr = cstr;
    if (psize) *psize = len + 1;
    Py_XDECREF(bytes);
    return ret;
  } else {
#if defined(SWIG_PYTHON_2_UNICODE)
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
#error "Cannot use both SWIG_PYTHON_2_UNICODE and SWIG_PYTHON_STRICT_BYTE_CHAR at once"
#endif
#if PY_VERSION_HEX<0x03000000
    if (PyUnicode_Check(obj)) {
      char *cstr; Py_ssize_t len;
      if (!alloc && cptr) {
        return SWIG_RuntimeError;
      }
      obj = PyUnicode_AsUTF8String(obj);
      if (!obj)
        return SWIG_TypeError;
      if (PyString_AsStringAndSize(obj, &cstr, &len) != -1) {
        if (cptr) {
          if (alloc) *alloc = SWIG_NEWOBJ;
          *cptr = reinterpret_cast< char* >(memcpy(new char[len + 1], cstr, sizeof(char)*(len + 1)));
        }
        if (psize) *psize = len + 1;

        Py_XDECREF(obj);
        return SWIG_OK;
      } else {
        Py_XDECREF(obj);
      }
    }
#endif
#endif

    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *) vptr;
	if (psize) *psize = vptr ? (strlen((char *)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsPtr_CString (PyObject * obj, CString **val) 
{
  char* buf = 0 ; size_t size = 0; int alloc = SWIG_OLDOBJ;
  if (SWIG_IsOK((SWIG_AsCharPtrAndSize(obj, &buf, &size, &alloc)))) {
    if (buf) {
      if (val) *val = new CString(buf, size - 1);
      if (alloc == SWIG_NEWOBJ) delete[] buf;
      return SWIG_NEWOBJ;
    } else {
      if (val) *val = 0;
      return SWIG_OLDOBJ;
    }
  } else {
    static int init = 0;
    static swig_type_info* descriptor = 0;
    if (!init) {
      descriptor = SWIG_TypeQuery("CString" " *");
      init = 1;
    }
    if (descriptor) {
      CString *vptr;
      int res = SWIG_ConvertPtr(obj, (void**)&vptr, descriptor, 0);
      if (SWIG_IsOK(res) && val) *val = vptr;
      return res;
    }
  }
  return SWIG_ERROR;
}


SWIGINTERN int
SWIG_AsVal_CString (PyObject * obj, CString *val)
{
  CString* v = (CString *) 0;
  int res = SWIG_AsPtr_CString (obj, &v);
  if (!SWIG_IsOK(res)) return res;
  if (v) {
    if (val) *val = *v;
    if (SWIG_IsNewObj(res)) {
      delete v;
      res = SWIG_DelNewMask(res);
    }
    return res;
  }
  return SWIG_ERROR;
}


SWIGINTERNINLINE PyObject *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > INT_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_InternalNewPointerObj(const_cast< char * >(carray), pchar_descriptor, 0) : SWIG_Py_Void();
    } else {
#if PY_VERSION_HEX >= 0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
      return PyBytes_FromStringAndSize(carray, static_cast< Py_ssize_t >(size));
#else
      return PyUnicode_DecodeUTF8(carray, static_cast< Py_ssize_t >(size), "surrogateescape");
#endif
#else
      return PyString_FromStringAndSize(carray, static_cast< Py_ssize_t >(size));
#endif
    }
  } else {
    return SWIG_Py_Void();
  }
}


SWIGINTERNINLINE PyObject *
SWIG_From_CString  (const CString& s)
{
  return SWIG_FromCharPtrAndSize(s.data(), s.size());
}


namespace swig {
  template <> struct traits< CString > {
    typedef value_category category;
    static const char* type_name() { return"CString"; }
  };
  template <>  struct traits_asval< CString > {
    typedef CString value_type;
    static int asval(PyObject *obj, value_type *val) {
      return SWIG_AsVal_CString (obj, val);
    }
  };
  template <>  struct traits_from< CString > {
    typedef CString value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_CString  (val);
    }
  };
}


#include <functional>

namespace std {
  template <>
  struct less <PyObject *>
  {
    bool
    operator()(PyObject * v, PyObject *w) const
    { 
      bool res;
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      res = PyObject_RichCompareBool(v, w, Py_LT) ? true : false;
      /* This may fall into a case of inconsistent
               eg. ObjA > ObjX > ObjB
               but ObjA < ObjB
      */
      if( PyErr_Occurred() && PyErr_ExceptionMatches(PyExc_TypeError) )
      {
        /* Objects can't be compared, this mostly occurred in Python 3.0 */
        /* Compare their ptr directly for a workaround */
        res = (v < w);
        PyErr_Clear();
      }
      SWIG_PYTHON_THREAD_END_BLOCK;
      return res;
    }
  };

  template <>
  struct less <swig::SwigPtr_PyObject>
  {
    bool
    operator()(const swig::SwigPtr_PyObject& v, const swig::SwigPtr_PyObject& w) const
    {
      return std::less<PyObject *>()(v, w);
    }
  };

  template <>
  struct less <swig::SwigVar_PyObject>
  {
    bool
    operator()(const swig::SwigVar_PyObject& v, const swig::SwigVar_PyObject& w) const
    {
      return std::less<PyObject *>()(v, w);
    }
  };

}

namespace swig {
  template <> struct traits<PyObject *> {
    typedef value_category category;
    static const char* type_name() { return "PyObject *"; }
  };  

  template <>  struct traits_asval<PyObject * > {   
    typedef PyObject * value_type;
    static int asval(PyObject *obj, value_type *val) {
      if (val) *val = obj;
      return SWIG_OK;
    }
  };

  template <> 
  struct traits_check<PyObject *, value_category> {
    static bool check(PyObject *) {
      return true;
    }
  };

  template <>  struct traits_from<PyObject *> {
    typedef PyObject * value_type;
    static PyObject *from(const value_type& val) {
      Py_XINCREF(val);
      return val;
    }
  };
  
}

namespace swig {
  template <class Difference>
  inline size_t
  check_index(Difference i, size_t size, bool insert = false) {
    if ( i < 0 ) {
      if ((size_t) (-i) <= size)
	return (size_t) (i + size);
    } else if ( (size_t) i < size ) {
      return (size_t) i;
    } else if (insert && ((size_t) i == size)) {
      return size;
    }
    throw std::out_of_range("index out of range");
  }

  template <class Difference>
  void
  slice_adjust(Difference i, Difference j, Py_ssize_t step, size_t size, Difference &ii, Difference &jj, bool insert = false) {
    if (step == 0) {
      throw std::invalid_argument("slice step cannot be zero");
    } else if (step > 0) {
      // Required range: 0 <= i < size, 0 <= j < size, i <= j
      if (i < 0) {
        ii = 0;
      } else if (i < (Difference)size) {
        ii = i;
      } else if (insert && (i >= (Difference)size)) {
        ii = (Difference)size;
      }
      if (j < 0) {
        jj = 0;
      } else {
        jj = (j < (Difference)size) ? j : (Difference)size;
      }
      if (jj < ii)
        jj = ii;
    } else {
      // Required range: -1 <= i < size-1, -1 <= j < size-1, i >= j
      if (i < -1) {
        ii = -1;
      } else if (i < (Difference) size) {
        ii = i;
      } else if (i >= (Difference)(size-1)) {
        ii = (Difference)(size-1);
      }
      if (j < -1) {
        jj = -1;
      } else {
        jj = (j < (Difference)size ) ? j : (Difference)(size-1);
      }
      if (ii < jj)
        ii = jj;
    }
  }

  template <class Sequence, class Difference>
  inline typename Sequence::iterator
  getpos(Sequence* self, Difference i)  {
    typename Sequence::iterator pos = self->begin();
    std::advance(pos, check_index(i,self->size()));
    return pos;
  }

  template <class Sequence, class Difference>
  inline typename Sequence::const_iterator
  cgetpos(const Sequence* self, Difference i)  {
    typename Sequence::const_iterator pos = self->begin();
    std::advance(pos, check_index(i,self->size()));
    return pos;
  }

  template <class Sequence>
  inline void
  erase(Sequence* seq, const typename Sequence::iterator& position) {
    seq->erase(position);
  }

  template <class Sequence>
  struct traits_reserve {
    static void reserve(Sequence & /*seq*/, typename Sequence::size_type /*n*/) {
      // This should be specialized for types that support reserve
    }
  };

  template <class Sequence, class Difference>
  inline Sequence*
  getslice(const Sequence* self, Difference i, Difference j, Py_ssize_t step) {
    typename Sequence::size_type size = self->size();
    Difference ii = 0;
    Difference jj = 0;
    swig::slice_adjust(i, j, step, size, ii, jj);

    if (step > 0) {
      typename Sequence::const_iterator sb = self->begin();
      typename Sequence::const_iterator se = self->begin();
      std::advance(sb,ii);
      std::advance(se,jj);
      if (step == 1) {
        return new Sequence(sb, se);
      } else {
        Sequence *sequence = new Sequence();
        swig::traits_reserve<Sequence>::reserve(*sequence, (jj - ii + step - 1) / step);
        typename Sequence::const_iterator it = sb;
        while (it!=se) {
          sequence->push_back(*it);
          for (Py_ssize_t c=0; c<step && it!=se; ++c)
            it++;
        }
        return sequence;
      } 
    } else {
      Sequence *sequence = new Sequence();
      swig::traits_reserve<Sequence>::reserve(*sequence, (ii - jj - step - 1) / -step);
      typename Sequence::const_reverse_iterator sb = self->rbegin();
      typename Sequence::const_reverse_iterator se = self->rbegin();
      std::advance(sb,size-ii-1);
      std::advance(se,size-jj-1);
      typename Sequence::const_reverse_iterator it = sb;
      while (it!=se) {
        sequence->push_back(*it);
        for (Py_ssize_t c=0; c<-step && it!=se; ++c)
          it++;
      }
      return sequence;
    }
  }

  template <class Sequence, class Difference, class InputSeq>
  inline void
  setslice(Sequence* self, Difference i, Difference j, Py_ssize_t step, const InputSeq& is = InputSeq()) {
    typename Sequence::size_type size = self->size();
    Difference ii = 0;
    Difference jj = 0;
    swig::slice_adjust(i, j, step, size, ii, jj, true);
    if (step > 0) {
      if (step == 1) {
        size_t ssize = jj - ii;
        if (ssize <= is.size()) {
          // expanding/staying the same size
          swig::traits_reserve<Sequence>::reserve(*self, self->size() - ssize + is.size());
          typename Sequence::iterator sb = self->begin();
          typename InputSeq::const_iterator isit = is.begin();
          std::advance(sb,ii);
          std::advance(isit, jj - ii);
          self->insert(std::copy(is.begin(), isit, sb), isit, is.end());
        } else {
          // shrinking
          typename Sequence::iterator sb = self->begin();
          typename Sequence::iterator se = self->begin();
          std::advance(sb,ii);
          std::advance(se,jj);
          self->erase(sb,se);
          sb = self->begin();
          std::advance(sb,ii);
          self->insert(sb, is.begin(), is.end());
        }
      } else {
        size_t replacecount = (jj - ii + step - 1) / step;
        if (is.size() != replacecount) {
          char msg[1024];
          PyOS_snprintf(msg, sizeof(msg), "attempt to assign sequence of size %lu to extended slice of size %lu", (unsigned long)is.size(), (unsigned long)replacecount);
          throw std::invalid_argument(msg);
        }
        typename Sequence::const_iterator isit = is.begin();
        typename Sequence::iterator it = self->begin();
        std::advance(it,ii);
        for (size_t rc=0; rc<replacecount && it != self->end(); ++rc) {
          *it++ = *isit++;
          for (Py_ssize_t c=0; c<(step-1) && it != self->end(); ++c)
            it++;
        }
      }
    } else {
      size_t replacecount = (ii - jj - step - 1) / -step;
      if (is.size() != replacecount) {
        char msg[1024];
        PyOS_snprintf(msg, sizeof(msg), "attempt to assign sequence of size %lu to extended slice of size %lu", (unsigned long)is.size(), (unsigned long)replacecount);
        throw std::invalid_argument(msg);
      }
      typename Sequence::const_iterator isit = is.begin();
      typename Sequence::reverse_iterator it = self->rbegin();
      std::advance(it,size-ii-1);
      for (size_t rc=0; rc<replacecount && it != self->rend(); ++rc) {
        *it++ = *isit++;
        for (Py_ssize_t c=0; c<(-step-1) && it != self->rend(); ++c)
          it++;
      }
    }
  }

  template <class Sequence, class Difference>
  inline void
  delslice(Sequence* self, Difference i, Difference j, Py_ssize_t step) {
    typename Sequence::size_type size = self->size();
    Difference ii = 0;
    Difference jj = 0;
    swig::slice_adjust(i, j, step, size, ii, jj, true);
    if (step > 0) {
      typename Sequence::iterator sb = self->begin();
      std::advance(sb,ii);
      if (step == 1) {
        typename Sequence::iterator se = self->begin();
        std::advance(se,jj);
        self->erase(sb,se);
      } else {
        typename Sequence::iterator it = sb;
        size_t delcount = (jj - ii + step - 1) / step;
        while (delcount) {
          it = self->erase(it);
          for (Py_ssize_t c=0; c<(step-1) && it != self->end(); ++c)
            it++;
          delcount--;
        }
      }
    } else {
      typename Sequence::reverse_iterator sb = self->rbegin();
      std::advance(sb,size-ii-1);
      typename Sequence::reverse_iterator it = sb;
      size_t delcount = (ii - jj - step - 1) / -step;
      while (delcount) {
        it = typename Sequence::reverse_iterator(self->erase((++it).base()));
        for (Py_ssize_t c=0; c<(-step-1) && it != self->rend(); ++c)
          it++;
        delcount--;
      }
    }
  }
}


namespace swig {
  template <class Seq, class T = typename Seq::value_type >
  struct IteratorProtocol {
    static void assign(PyObject *obj, Seq *seq) {
      SwigVar_PyObject iter = PyObject_GetIter(obj);
      if (iter) {
        SwigVar_PyObject item = PyIter_Next(iter);
        while (item) {
          seq->insert(seq->end(), swig::as<T>(item));
          item = PyIter_Next(iter);
        }
      }
    }

    static bool check(PyObject *obj) {
      bool ret = false;
      SwigVar_PyObject iter = PyObject_GetIter(obj);
      if (iter) {
        SwigVar_PyObject item = PyIter_Next(iter);
        ret = true;
        while (item) {
          ret = swig::check<T>(item);
          item = ret ? PyIter_Next(iter) : 0;
        }
      }
      return ret;
    }
  };

  template <class Seq, class T = typename Seq::value_type >
  struct traits_asptr_stdseq {
    typedef Seq sequence;
    typedef T value_type;

    static bool is_iterable(PyObject *obj) {
      SwigVar_PyObject iter = PyObject_GetIter(obj);
      PyErr_Clear();
      return iter != 0;
    }

    static int asptr(PyObject *obj, sequence **seq) {
      int ret = SWIG_ERROR;
      if (obj == Py_None || SWIG_Python_GetSwigThis(obj)) {
	sequence *p;
	swig_type_info *descriptor = swig::type_info<sequence>();
	if (descriptor && SWIG_IsOK(::SWIG_ConvertPtr(obj, (void **)&p, descriptor, 0))) {
	  if (seq) *seq = p;
	  return SWIG_OLDOBJ;
	}
      } else if (is_iterable(obj)) {
	try {
	  if (seq) {
	    *seq = new sequence();
            IteratorProtocol<Seq, T>::assign(obj, *seq);
            if (!PyErr_Occurred())
              return SWIG_NEWOBJ;
	  } else {
	    return IteratorProtocol<Seq, T>::check(obj) ? SWIG_OK : SWIG_ERROR;
	  }
	} catch (std::exception& e) {
          if (seq && !PyErr_Occurred())
            PyErr_SetString(PyExc_TypeError, e.what());
	}
        if (seq)
          delete *seq;
	return SWIG_ERROR;
      }
      return ret;
    }
  };

  template <class Seq, class T = typename Seq::value_type >
  struct traits_from_stdseq {
    typedef Seq sequence;
    typedef T value_type;
    typedef typename Seq::size_type size_type;
    typedef typename sequence::const_iterator const_iterator;

    static PyObject *from(const sequence& seq) {
#ifdef SWIG_PYTHON_EXTRA_NATIVE_CONTAINERS
      swig_type_info *desc = swig::type_info<sequence>();
      if (desc && desc->clientdata) {
	return SWIG_InternalNewPointerObj(new sequence(seq), desc, SWIG_POINTER_OWN);
      }
#endif
      size_type size = seq.size();
      if (size <= (size_type)INT_MAX) {
	PyObject *obj = PyTuple_New((Py_ssize_t)size);
	Py_ssize_t i = 0;
	for (const_iterator it = seq.begin(); it != seq.end(); ++it, ++i) {
	  PyTuple_SetItem(obj,i,swig::from<value_type>(*it));
	}
	return obj;
      } else {
	PyErr_SetString(PyExc_OverflowError,"sequence size not valid in python");
	return NULL;
      }
    }
  };
}


  namespace swig {
    template <class T >
    struct traits_asptr<std::list<T> >  {
      static int asptr(PyObject *obj, std::list<T> **lis) {
	return traits_asptr_stdseq<std::list<T> >::asptr(obj, lis);
      }
    };

    template <class T>
    struct traits_from<std::list<T> > {
      static PyObject *from(const std::list<T>& vec) {
	return traits_from_stdseq<std::list<T> >::from(vec);
      }
    };
  }


      namespace swig {
	template <>  struct traits<std::list< CString, std::allocator< CString > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::list<" "CString" ", " "std::allocator< CString >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_list_Sl_CString_Sg__iterator(std::list< CString > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_list_Sl_CString_Sg____nonzero__(std::list< CString > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_list_Sl_CString_Sg____bool__(std::list< CString > const *self){
      return !(self->empty());
    }
SWIGINTERN std::list< CString >::size_type std_list_Sl_CString_Sg____len__(std::list< CString > const *self){
      return self->size();
    }

SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLong(value) : PyInt_FromLong(static_cast< long >(value));
}


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long_SS_long  (unsigned long long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLongLong(value) : PyInt_FromLong(static_cast< long >(value));
}
#endif


SWIGINTERNINLINE PyObject *
SWIG_From_size_t  (size_t value)
{    
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    return SWIG_From_unsigned_SS_long  (static_cast< unsigned long >(value));
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else {
    /* assume sizeof(size_t) <= sizeof(unsigned long long) */
    return SWIG_From_unsigned_SS_long_SS_long  (static_cast< unsigned long long >(value));
  }
#endif
}

SWIGINTERN std::list< CString,std::allocator< CString > > *std_list_Sl_CString_Sg____getslice__(std::list< CString > *self,std::list< CString >::difference_type i,std::list< CString >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_list_Sl_CString_Sg____setslice____SWIG_0(std::list< CString > *self,std::list< CString >::difference_type i,std::list< CString >::difference_type j){
      swig::setslice(self, i, j, 1, std::list< CString,std::allocator< CString > >());
    }
SWIGINTERN void std_list_Sl_CString_Sg____setslice____SWIG_1(std::list< CString > *self,std::list< CString >::difference_type i,std::list< CString >::difference_type j,std::list< CString,std::allocator< CString > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_list_Sl_CString_Sg____delslice__(std::list< CString > *self,std::list< CString >::difference_type i,std::list< CString >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_list_Sl_CString_Sg____delitem____SWIG_0(std::list< CString > *self,std::list< CString >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::list< CString,std::allocator< CString > > *std_list_Sl_CString_Sg____getitem____SWIG_0(std::list< CString > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::list< CString,std::allocator< CString > >::difference_type id = i;
      std::list< CString,std::allocator< CString > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_list_Sl_CString_Sg____setitem____SWIG_0(std::list< CString > *self,SWIGPY_SLICEOBJECT *slice,std::list< CString,std::allocator< CString > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::list< CString,std::allocator< CString > >::difference_type id = i;
      std::list< CString,std::allocator< CString > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_list_Sl_CString_Sg____setitem____SWIG_1(std::list< CString > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::list< CString,std::allocator< CString > >::difference_type id = i;
      std::list< CString,std::allocator< CString > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_list_Sl_CString_Sg____delitem____SWIG_1(std::list< CString > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::list< CString,std::allocator< CString > >::difference_type id = i;
      std::list< CString,std::allocator< CString > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::list< CString >::value_type const &std_list_Sl_CString_Sg____getitem____SWIG_1(std::list< CString > const *self,std::list< CString >::difference_type i){
      return *(swig::cgetpos(self, i));
    }

namespace swig {
  static PyObject* container_owner_attribute() {
    static PyObject* attr = SWIG_Python_str_FromChar("__swig_container");
    return attr;
  }

  template <typename T>
  struct container_owner {
    // By default, do not add the back-reference (for value types)
    // Specialization below will check the reference for pointer types.
    static bool back_reference(PyObject* /*child*/, PyObject* /*owner*/) {
      return false;
    }
  };

  template <>
  struct container_owner<swig::pointer_category> {  
    /*
     * Call to add a back-reference to the owning object when returning a 
     * reference from a container.  Will only set the reference if child
     * is a SWIG wrapper object that does not own the pointer.
     *
     * returns whether the reference was set or not
     */
    static bool back_reference(PyObject* child, PyObject* owner) {
      SwigPyObject* swigThis = SWIG_Python_GetSwigThis(child);
      if (swigThis && (swigThis->own & SWIG_POINTER_OWN) != SWIG_POINTER_OWN) {
        return PyObject_SetAttr(child, container_owner_attribute(), owner) != -1;
      }
      return false;
    }
  };
}

SWIGINTERN void std_list_Sl_CString_Sg____setitem____SWIG_2(std::list< CString > *self,std::list< CString >::difference_type i,std::list< CString >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::list< CString >::value_type std_list_Sl_CString_Sg__pop(std::list< CString > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::list< CString,std::allocator< CString > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_list_Sl_CString_Sg__append(std::list< CString > *self,std::list< CString >::value_type const &x){
      self->push_back(x);
    }
SWIGINTERN std::list< CString >::iterator std_list_Sl_CString_Sg__erase__SWIG_0(std::list< CString > *self,std::list< CString >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::list< CString >::iterator std_list_Sl_CString_Sg__erase__SWIG_1(std::list< CString > *self,std::list< CString >::iterator first,std::list< CString >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::list< CString >::iterator std_list_Sl_CString_Sg__insert__SWIG_0(std::list< CString > *self,std::list< CString >::iterator pos,std::list< CString >::value_type const &x){ return self->insert(pos, x); }
SWIGINTERN void std_list_Sl_CString_Sg__insert__SWIG_1(std::list< CString > *self,std::list< CString >::iterator pos,std::list< CString >::size_type n,std::list< CString >::value_type const &x){ self->insert(pos, n, x); }

  namespace swig {
    template <>  struct traits< CIRCNetwork > {
      typedef pointer_category category;
      static const char* type_name() { return"CIRCNetwork"; }
    };
  }


  namespace swig {
    template <class T>
    struct traits_reserve<std::vector<T> > {
      static void reserve(std::vector<T> &seq, typename std::vector<T>::size_type n) {
        seq.reserve(n);
      }
    };

    template <class T>
    struct traits_asptr<std::vector<T> >  {
      static int asptr(PyObject *obj, std::vector<T> **vec) {
	return traits_asptr_stdseq<std::vector<T> >::asptr(obj, vec);
      }
    };
    
    template <class T>
    struct traits_from<std::vector<T> > {
      static PyObject *from(const std::vector<T>& vec) {
	return traits_from_stdseq<std::vector<T> >::from(vec);
      }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< CIRCNetwork*, std::allocator< CIRCNetwork * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "CIRCNetwork" " *," "std::allocator< CIRCNetwork * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_CIRCNetwork_Sm__Sg__iterator(std::vector< CIRCNetwork * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_CIRCNetwork_Sm__Sg____nonzero__(std::vector< CIRCNetwork * > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_CIRCNetwork_Sm__Sg____bool__(std::vector< CIRCNetwork * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< CIRCNetwork * >::size_type std_vector_Sl_CIRCNetwork_Sm__Sg____len__(std::vector< CIRCNetwork * > const *self){
      return self->size();
    }
SWIGINTERN std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > *std_vector_Sl_CIRCNetwork_Sm__Sg____getslice__(std::vector< CIRCNetwork * > *self,std::vector< CIRCNetwork * >::difference_type i,std::vector< CIRCNetwork * >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CIRCNetwork_Sm__Sg____setslice____SWIG_0(std::vector< CIRCNetwork * > *self,std::vector< CIRCNetwork * >::difference_type i,std::vector< CIRCNetwork * >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >());
    }
SWIGINTERN void std_vector_Sl_CIRCNetwork_Sm__Sg____setslice____SWIG_1(std::vector< CIRCNetwork * > *self,std::vector< CIRCNetwork * >::difference_type i,std::vector< CIRCNetwork * >::difference_type j,std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_CIRCNetwork_Sm__Sg____delslice__(std::vector< CIRCNetwork * > *self,std::vector< CIRCNetwork * >::difference_type i,std::vector< CIRCNetwork * >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CIRCNetwork_Sm__Sg____delitem____SWIG_0(std::vector< CIRCNetwork * > *self,std::vector< CIRCNetwork * >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > *std_vector_Sl_CIRCNetwork_Sm__Sg____getitem____SWIG_0(std::vector< CIRCNetwork * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >::difference_type id = i;
      std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CIRCNetwork_Sm__Sg____setitem____SWIG_0(std::vector< CIRCNetwork * > *self,SWIGPY_SLICEOBJECT *slice,std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >::difference_type id = i;
      std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_CIRCNetwork_Sm__Sg____setitem____SWIG_1(std::vector< CIRCNetwork * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >::difference_type id = i;
      std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CIRCNetwork_Sm__Sg____delitem____SWIG_1(std::vector< CIRCNetwork * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >::difference_type id = i;
      std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< CIRCNetwork * >::value_type std_vector_Sl_CIRCNetwork_Sm__Sg____getitem____SWIG_1(std::vector< CIRCNetwork * > *self,std::vector< CIRCNetwork * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_CIRCNetwork_Sm__Sg____setitem____SWIG_2(std::vector< CIRCNetwork * > *self,std::vector< CIRCNetwork * >::difference_type i,std::vector< CIRCNetwork * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< CIRCNetwork * >::value_type std_vector_Sl_CIRCNetwork_Sm__Sg__pop(std::vector< CIRCNetwork * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_CIRCNetwork_Sm__Sg__append(std::vector< CIRCNetwork * > *self,std::vector< CIRCNetwork * >::value_type x){
      self->push_back(x);
    }
SWIGINTERN std::vector< CIRCNetwork * >::iterator std_vector_Sl_CIRCNetwork_Sm__Sg__erase__SWIG_0(std::vector< CIRCNetwork * > *self,std::vector< CIRCNetwork * >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< CIRCNetwork * >::iterator std_vector_Sl_CIRCNetwork_Sm__Sg__erase__SWIG_1(std::vector< CIRCNetwork * > *self,std::vector< CIRCNetwork * >::iterator first,std::vector< CIRCNetwork * >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< CIRCNetwork * >::iterator std_vector_Sl_CIRCNetwork_Sm__Sg__insert__SWIG_0(std::vector< CIRCNetwork * > *self,std::vector< CIRCNetwork * >::iterator pos,std::vector< CIRCNetwork * >::value_type x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_CIRCNetwork_Sm__Sg__insert__SWIG_1(std::vector< CIRCNetwork * > *self,std::vector< CIRCNetwork * >::iterator pos,std::vector< CIRCNetwork * >::size_type n,std::vector< CIRCNetwork * >::value_type x){ self->insert(pos, n, x); }

  namespace swig {
    template <>  struct traits< CChan > {
      typedef pointer_category category;
      static const char* type_name() { return"CChan"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< CChan*, std::allocator< CChan * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "CChan" " *," "std::allocator< CChan * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_CChan_Sm__Sg__iterator(std::vector< CChan * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_CChan_Sm__Sg____nonzero__(std::vector< CChan * > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_CChan_Sm__Sg____bool__(std::vector< CChan * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< CChan * >::size_type std_vector_Sl_CChan_Sm__Sg____len__(std::vector< CChan * > const *self){
      return self->size();
    }
SWIGINTERN std::vector< CChan *,std::allocator< CChan * > > *std_vector_Sl_CChan_Sm__Sg____getslice__(std::vector< CChan * > *self,std::vector< CChan * >::difference_type i,std::vector< CChan * >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CChan_Sm__Sg____setslice____SWIG_0(std::vector< CChan * > *self,std::vector< CChan * >::difference_type i,std::vector< CChan * >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< CChan*,std::allocator< CChan * > >());
    }
SWIGINTERN void std_vector_Sl_CChan_Sm__Sg____setslice____SWIG_1(std::vector< CChan * > *self,std::vector< CChan * >::difference_type i,std::vector< CChan * >::difference_type j,std::vector< CChan *,std::allocator< CChan * > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_CChan_Sm__Sg____delslice__(std::vector< CChan * > *self,std::vector< CChan * >::difference_type i,std::vector< CChan * >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CChan_Sm__Sg____delitem____SWIG_0(std::vector< CChan * > *self,std::vector< CChan * >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< CChan *,std::allocator< CChan * > > *std_vector_Sl_CChan_Sm__Sg____getitem____SWIG_0(std::vector< CChan * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CChan*,std::allocator< CChan * > >::difference_type id = i;
      std::vector< CChan*,std::allocator< CChan * > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CChan_Sm__Sg____setitem____SWIG_0(std::vector< CChan * > *self,SWIGPY_SLICEOBJECT *slice,std::vector< CChan *,std::allocator< CChan * > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CChan*,std::allocator< CChan * > >::difference_type id = i;
      std::vector< CChan*,std::allocator< CChan * > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_CChan_Sm__Sg____setitem____SWIG_1(std::vector< CChan * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CChan*,std::allocator< CChan * > >::difference_type id = i;
      std::vector< CChan*,std::allocator< CChan * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CChan_Sm__Sg____delitem____SWIG_1(std::vector< CChan * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CChan*,std::allocator< CChan * > >::difference_type id = i;
      std::vector< CChan*,std::allocator< CChan * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< CChan * >::value_type std_vector_Sl_CChan_Sm__Sg____getitem____SWIG_1(std::vector< CChan * > *self,std::vector< CChan * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_CChan_Sm__Sg____setitem____SWIG_2(std::vector< CChan * > *self,std::vector< CChan * >::difference_type i,std::vector< CChan * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< CChan * >::value_type std_vector_Sl_CChan_Sm__Sg__pop(std::vector< CChan * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< CChan*,std::allocator< CChan * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_CChan_Sm__Sg__append(std::vector< CChan * > *self,std::vector< CChan * >::value_type x){
      self->push_back(x);
    }
SWIGINTERN std::vector< CChan * >::iterator std_vector_Sl_CChan_Sm__Sg__erase__SWIG_0(std::vector< CChan * > *self,std::vector< CChan * >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< CChan * >::iterator std_vector_Sl_CChan_Sm__Sg__erase__SWIG_1(std::vector< CChan * > *self,std::vector< CChan * >::iterator first,std::vector< CChan * >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< CChan * >::iterator std_vector_Sl_CChan_Sm__Sg__insert__SWIG_0(std::vector< CChan * > *self,std::vector< CChan * >::iterator pos,std::vector< CChan * >::value_type x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_CChan_Sm__Sg__insert__SWIG_1(std::vector< CChan * > *self,std::vector< CChan * >::iterator pos,std::vector< CChan * >::size_type n,std::vector< CChan * >::value_type x){ self->insert(pos, n, x); }

  namespace swig {
    template <>  struct traits< CNick > {
      typedef pointer_category category;
      static const char* type_name() { return"CNick"; }
    };
  }


  namespace swig {


















































    template <class T, class U >
    struct traits_asptr<std::pair<T,U> >  {
      typedef std::pair<T,U> value_type;

      static int get_pair(PyObject* first, PyObject* second,
			  std::pair<T,U> **val) 
      {
	if (val) {
	  value_type *vp = (new std::pair<T,U>());
	  T *pfirst = &(vp->first);
	  int res1 = swig::asval((PyObject*)first, pfirst);
	  if (!SWIG_IsOK(res1)) {
	    delete vp;
	    return res1;
	  }
	  U *psecond = &(vp->second);
	  int res2 = swig::asval((PyObject*)second, psecond);
	  if (!SWIG_IsOK(res2)) {
	    delete vp;
	    return res2;
	  }
	  *val = vp;
	  return SWIG_AddNewMask(res1 > res2 ? res1 : res2);
	} else {
	  T *pfirst = 0;
	  int res1 = swig::asval((PyObject*)first, pfirst);
	  if (!SWIG_IsOK(res1)) return res1;
	  U *psecond = 0;
	  int res2 = swig::asval((PyObject*)second, psecond);
	  if (!SWIG_IsOK(res2)) return res2;
	  return res1 > res2 ? res1 : res2;
	}
      }

      static int asptr(PyObject *obj, std::pair<T,U> **val) {
	int res = SWIG_ERROR;
	if (PyTuple_Check(obj)) {
	  if (PyTuple_GET_SIZE(obj) == 2) {
	    res = get_pair(PyTuple_GET_ITEM(obj,0),PyTuple_GET_ITEM(obj,1), val);
	  }
	} else if (PySequence_Check(obj)) {
	  if (PySequence_Size(obj) == 2) {
	    swig::SwigVar_PyObject first = PySequence_GetItem(obj,0);
	    swig::SwigVar_PyObject second = PySequence_GetItem(obj,1);
	    res = get_pair(first, second, val);
	  }
	} else {
	  value_type *p = 0;
	  swig_type_info *descriptor = swig::type_info<value_type>();
	  res = descriptor ? SWIG_ConvertPtr(obj, (void **)&p, descriptor, 0) : SWIG_ERROR;
	  if (SWIG_IsOK(res) && val)  *val = p;
	}
	return res;
      }
    };


    template <class T, class U >
    struct traits_from<std::pair<T,U> >   {
      static PyObject *from(const std::pair<T,U>& val) {
	PyObject* obj = PyTuple_New(2);
	PyTuple_SetItem(obj,0,swig::from(val.first));
	PyTuple_SetItem(obj,1,swig::from(val.second));
	return obj;
      }
    };
  }





































      namespace swig {
	template <>  struct traits<std::pair< CString, CNick > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::pair<" "CString" "," "CNick" " >";
	  }
	};
      }
    

  namespace swig {
    template <class ValueType>
    struct from_key_oper 
    {
      typedef const ValueType& argument_type;
      typedef  PyObject *result_type;
      result_type operator()(argument_type v) const
      {
	return swig::from(v.first);
      }
    };

    template <class ValueType>
    struct from_value_oper 
    {
      typedef const ValueType& argument_type;
      typedef  PyObject *result_type;
      result_type operator()(argument_type v) const
      {
	return swig::from(v.second);
      }
    };

    template<class OutIterator, class FromOper, class ValueType = typename OutIterator::value_type>
    struct SwigPyMapIterator_T : SwigPyIteratorClosed_T<OutIterator, ValueType, FromOper>
    {
      SwigPyMapIterator_T(OutIterator curr, OutIterator first, OutIterator last, PyObject *seq)
	: SwigPyIteratorClosed_T<OutIterator,ValueType,FromOper>(curr, first, last, seq)
      {
      }
    };


    template<class OutIterator,
	     class FromOper = from_key_oper<typename OutIterator::value_type> >
    struct SwigPyMapKeyIterator_T : SwigPyMapIterator_T<OutIterator, FromOper>
    {
      SwigPyMapKeyIterator_T(OutIterator curr, OutIterator first, OutIterator last, PyObject *seq)
	: SwigPyMapIterator_T<OutIterator, FromOper>(curr, first, last, seq)
      {
      }
    };

    template<typename OutIter>
    inline SwigPyIterator*
    make_output_key_iterator(const OutIter& current, const OutIter& begin, const OutIter& end, PyObject *seq = 0)
    {
      return new SwigPyMapKeyIterator_T<OutIter>(current, begin, end, seq);
    }

    template<class OutIterator,
	     class FromOper = from_value_oper<typename OutIterator::value_type> >
    struct SwigPyMapValueIterator_T : SwigPyMapIterator_T<OutIterator, FromOper>
    {
      SwigPyMapValueIterator_T(OutIterator curr, OutIterator first, OutIterator last, PyObject *seq)
	: SwigPyMapIterator_T<OutIterator, FromOper>(curr, first, last, seq)
      {
      }
    };
    

    template<typename OutIter>
    inline SwigPyIterator*
    make_output_value_iterator(const OutIter& current, const OutIter& begin, const OutIter& end, PyObject *seq = 0)
    {
      return new SwigPyMapValueIterator_T<OutIter>(current, begin, end, seq);
    }
  }


  namespace swig {
    template <class K, class T, class Compare, class Alloc>
    struct traits_asptr<std::map<K,T,Compare,Alloc > >  {
      typedef std::map<K,T,Compare,Alloc > map_type;
      static int asptr(PyObject *obj, map_type **val) {
	int res = SWIG_ERROR;
	SWIG_PYTHON_THREAD_BEGIN_BLOCK;
	if (PyDict_Check(obj)) {
	  SwigVar_PyObject items = PyObject_CallMethod(obj,(char *)"items",NULL);
#if PY_VERSION_HEX >= 0x03000000
          /* In Python 3.x the ".items()" method returns a dict_items object */
          items = PySequence_Fast(items, ".items() didn't return a sequence!");
#endif
	  res = traits_asptr_stdseq<map_type, std::pair<K, T> >::asptr(items, val);
	} else {
	  map_type *p = 0;
	  swig_type_info *descriptor = swig::type_info<map_type>();
	  res = descriptor ? SWIG_ConvertPtr(obj, (void **)&p, descriptor, 0) : SWIG_ERROR;
	  if (SWIG_IsOK(res) && val)  *val = p;
	}
	SWIG_PYTHON_THREAD_END_BLOCK;
	return res;
      }      
    };
      
    template <class K, class T, class Compare, class Alloc >
    struct traits_from<std::map<K,T,Compare,Alloc > >  {
      typedef std::map<K,T,Compare,Alloc > map_type;
      typedef typename map_type::const_iterator const_iterator;
      typedef typename map_type::size_type size_type;

      static PyObject *asdict(const map_type& map) {
	SWIG_PYTHON_THREAD_BEGIN_BLOCK;
	size_type size = map.size();
	Py_ssize_t pysize = (size <= (size_type) INT_MAX) ? (Py_ssize_t) size : -1;
	if (pysize < 0) {
	  PyErr_SetString(PyExc_OverflowError, "map size not valid in python");
	  SWIG_PYTHON_THREAD_END_BLOCK;
	  return NULL;
	}
	PyObject *obj = PyDict_New();
	for (const_iterator i= map.begin(); i!= map.end(); ++i) {
	  swig::SwigVar_PyObject key = swig::from(i->first);
	  swig::SwigVar_PyObject val = swig::from(i->second);
	  PyDict_SetItem(obj, key, val);
	}
	SWIG_PYTHON_THREAD_END_BLOCK;
	return obj;
      }
                
      static PyObject *from(const map_type& map) {
	swig_type_info *desc = swig::type_info<map_type>();
	if (desc && desc->clientdata) {
	  return SWIG_InternalNewPointerObj(new map_type(map), desc, SWIG_POINTER_OWN);
	} else {
	  return asdict(map);
	}
      }
    };
  }


      namespace swig {
	template <>  struct traits<std::map< CString, CNick, std::less< CString >, std::allocator< std::pair< CString const,CNick > > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::map<" "CString" "," "CNick" "," "std::less< CString >" "," "std::allocator< std::pair< CString const,CNick > >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_map_Sl_CString_Sc_CNick_Sg__iterator(std::map< CString,CNick > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_map_Sl_CString_Sc_CNick_Sg____nonzero__(std::map< CString,CNick > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_map_Sl_CString_Sc_CNick_Sg____bool__(std::map< CString,CNick > const *self){
      return !(self->empty());
    }
SWIGINTERN std::map< CString,CNick >::size_type std_map_Sl_CString_Sc_CNick_Sg____len__(std::map< CString,CNick > const *self){
      return self->size();
    }
SWIGINTERN std::map< CString,CNick >::mapped_type const &std_map_Sl_CString_Sc_CNick_Sg____getitem__(std::map< CString,CNick > *self,std::map< CString,CNick >::key_type const &key){
      std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >::const_iterator i = self->find(key);
      if (i != self->end())
	return i->second;
      else
	throw std::out_of_range("key not found");
    }
SWIGINTERN void std_map_Sl_CString_Sc_CNick_Sg____delitem__(std::map< CString,CNick > *self,std::map< CString,CNick >::key_type const &key){
      std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >::iterator i = self->find(key);
      if (i != self->end())
	self->erase(i);
      else
	throw std::out_of_range("key not found");
    }
SWIGINTERN bool std_map_Sl_CString_Sc_CNick_Sg__has_key(std::map< CString,CNick > const *self,std::map< CString,CNick >::key_type const &key){
      std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >::const_iterator i = self->find(key);
      return i != self->end();
    }
SWIGINTERN PyObject *std_map_Sl_CString_Sc_CNick_Sg__keys(std::map< CString,CNick > *self){
      std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >::size_type size = self->size();
      Py_ssize_t pysize = (size <= (std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >::size_type) INT_MAX) ? (Py_ssize_t) size : -1;
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      if (pysize < 0) {
	PyErr_SetString(PyExc_OverflowError, "map size not valid in python");
	SWIG_PYTHON_THREAD_END_BLOCK;
	return NULL;
      }
      PyObject* keyList = PyList_New(pysize);
      std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >::const_iterator i = self->begin();
      for (Py_ssize_t j = 0; j < pysize; ++i, ++j) {
	PyList_SET_ITEM(keyList, j, swig::from(i->first));
      }
      SWIG_PYTHON_THREAD_END_BLOCK;
      return keyList;
    }
SWIGINTERN PyObject *std_map_Sl_CString_Sc_CNick_Sg__values(std::map< CString,CNick > *self){
      std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >::size_type size = self->size();
      Py_ssize_t pysize = (size <= (std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >::size_type) INT_MAX) ? (Py_ssize_t) size : -1;
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      if (pysize < 0) {
	PyErr_SetString(PyExc_OverflowError, "map size not valid in python");
	SWIG_PYTHON_THREAD_END_BLOCK;
	return NULL;
      }
      PyObject* valList = PyList_New(pysize);
      std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >::const_iterator i = self->begin();
      for (Py_ssize_t j = 0; j < pysize; ++i, ++j) {
	PyList_SET_ITEM(valList, j, swig::from(i->second));
      }
      SWIG_PYTHON_THREAD_END_BLOCK;
      return valList;
    }
SWIGINTERN PyObject *std_map_Sl_CString_Sc_CNick_Sg__items(std::map< CString,CNick > *self){
      std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >::size_type size = self->size();
      Py_ssize_t pysize = (size <= (std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >::size_type) INT_MAX) ? (Py_ssize_t) size : -1;
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      if (pysize < 0) {
	PyErr_SetString(PyExc_OverflowError, "map size not valid in python");
	SWIG_PYTHON_THREAD_END_BLOCK;
	return NULL;
      }    
      PyObject* itemList = PyList_New(pysize);
      std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >::const_iterator i = self->begin();
      for (Py_ssize_t j = 0; j < pysize; ++i, ++j) {
	PyList_SET_ITEM(itemList, j, swig::from(*i));
      }
      SWIG_PYTHON_THREAD_END_BLOCK;
      return itemList;
    }
SWIGINTERN bool std_map_Sl_CString_Sc_CNick_Sg____contains__(std::map< CString,CNick > *self,std::map< CString,CNick >::key_type const &key){
      return self->find(key) != self->end();
    }
SWIGINTERN swig::SwigPyIterator *std_map_Sl_CString_Sc_CNick_Sg__key_iterator(std::map< CString,CNick > *self,PyObject **PYTHON_SELF){
      return swig::make_output_key_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN swig::SwigPyIterator *std_map_Sl_CString_Sc_CNick_Sg__value_iterator(std::map< CString,CNick > *self,PyObject **PYTHON_SELF){
      return swig::make_output_value_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN void std_map_Sl_CString_Sc_CNick_Sg____setitem____SWIG_0(std::map< CString,CNick > *self,std::map< CString,CNick >::key_type const &key){
      self->erase(key);
    }
SWIGINTERN void std_map_Sl_CString_Sc_CNick_Sg____setitem____SWIG_1(std::map< CString,CNick > *self,std::map< CString,CNick >::key_type const &key,std::map< CString,CNick >::mapped_type const &x){
#ifdef __cpp_lib_map_try_emplace
      (*self).insert_or_assign(key, x);
#else
      (*self)[key] = x;
#endif
    }
SWIGINTERN PyObject *std_map_Sl_CString_Sc_CNick_Sg__asdict(std::map< CString,CNick > *self){
      return swig::traits_from< std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > > >::asdict(*self);
    }
SWIGINTERN void std_map_Sl_CString_Sc_CNick_Sg__erase__SWIG_1(std::map< CString,CNick > *self,std::map< CString,CNick >::iterator position){ self->erase(position); }
SWIGINTERN void std_map_Sl_CString_Sc_CNick_Sg__erase__SWIG_2(std::map< CString,CNick > *self,std::map< CString,CNick >::iterator first,std::map< CString,CNick >::iterator last){ self->erase(first, last); }

  namespace swig {
    template <>  struct traits< CModInfo > {
      typedef pointer_category category;
      static const char* type_name() { return"CModInfo"; }
    };
  }


  namespace swig {
    template <class T>
    struct traits_asptr<std::set<T> >  {
      static int asptr(PyObject *obj, std::set<T> **s) {
	return traits_asptr_stdseq<std::set<T> >::asptr(obj, s);
      }
    };

    template <class T>
    struct traits_from<std::set<T> > {
      static PyObject *from(const std::set<T>& vec) {
	return traits_from_stdseq<std::set<T> >::from(vec);
      }
    };
  }


      namespace swig {
	template <>  struct traits<std::set< CModInfo, std::less< CModInfo >, std::allocator< CModInfo > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::set<" "CModInfo" "," "std::less< CModInfo >" "," "std::allocator< CModInfo >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_set_Sl_CModInfo_Sg__iterator(std::set< CModInfo > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_set_Sl_CModInfo_Sg____nonzero__(std::set< CModInfo > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_set_Sl_CModInfo_Sg____bool__(std::set< CModInfo > const *self){
      return !(self->empty());
    }
SWIGINTERN std::set< CModInfo >::size_type std_set_Sl_CModInfo_Sg____len__(std::set< CModInfo > const *self){
      return self->size();
    }
SWIGINTERN void std_set_Sl_CModInfo_Sg__append(std::set< CModInfo > *self,std::set< CModInfo >::value_type x){
       self->insert(x);
     }
SWIGINTERN bool std_set_Sl_CModInfo_Sg____contains__(std::set< CModInfo > *self,std::set< CModInfo >::value_type x){
       return self->find(x) != self->end();
     }
SWIGINTERN std::set< CModInfo >::value_type std_set_Sl_CModInfo_Sg____getitem__(std::set< CModInfo > const *self,std::set< CModInfo >::difference_type i){
       return *(swig::cgetpos(self, i));
     }
SWIGINTERN void std_set_Sl_CModInfo_Sg__add(std::set< CModInfo > *self,std::set< CModInfo >::value_type x){
       self->insert(x);
     }
SWIGINTERN void std_set_Sl_CModInfo_Sg__discard(std::set< CModInfo > *self,std::set< CModInfo >::value_type x){
       self->erase(x);
     }
SWIGINTERN void std_set_Sl_CModInfo_Sg__erase__SWIG_1(std::set< CModInfo > *self,std::set< CModInfo >::iterator pos){ self->erase(pos); }
SWIGINTERN void std_set_Sl_CModInfo_Sg__erase__SWIG_2(std::set< CModInfo > *self,std::set< CModInfo >::iterator first,std::set< CModInfo >::iterator last){ self->erase(first, last); }


      namespace swig {
	template <>  struct traits<std::set< CString, std::less< CString >, std::allocator< CString > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::set<" "CString" "," "std::less< CString >" "," "std::allocator< CString >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_set_Sl_CString_Sg__iterator(std::set< CString > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_set_Sl_CString_Sg____nonzero__(std::set< CString > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_set_Sl_CString_Sg____bool__(std::set< CString > const *self){
      return !(self->empty());
    }
SWIGINTERN std::set< CString >::size_type std_set_Sl_CString_Sg____len__(std::set< CString > const *self){
      return self->size();
    }
SWIGINTERN void std_set_Sl_CString_Sg__append(std::set< CString > *self,std::set< CString >::value_type x){
       self->insert(x);
     }
SWIGINTERN bool std_set_Sl_CString_Sg____contains__(std::set< CString > *self,std::set< CString >::value_type x){
       return self->find(x) != self->end();
     }
SWIGINTERN std::set< CString >::value_type std_set_Sl_CString_Sg____getitem__(std::set< CString > const *self,std::set< CString >::difference_type i){
       return *(swig::cgetpos(self, i));
     }
SWIGINTERN void std_set_Sl_CString_Sg__add(std::set< CString > *self,std::set< CString >::value_type x){
       self->insert(x);
     }
SWIGINTERN void std_set_Sl_CString_Sg__discard(std::set< CString > *self,std::set< CString >::value_type x){
       self->erase(x);
     }
SWIGINTERN void std_set_Sl_CString_Sg__erase__SWIG_1(std::set< CString > *self,std::set< CString >::iterator pos){ self->erase(pos); }
SWIGINTERN void std_set_Sl_CString_Sg__erase__SWIG_2(std::set< CString > *self,std::set< CString >::iterator first,std::set< CString >::iterator last){ self->erase(first, last); }

      namespace swig {
	template <>  struct traits<std::vector< CString, std::allocator< CString > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "CString" "," "std::allocator< CString >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_CString_Sg__iterator(std::vector< CString > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_CString_Sg____nonzero__(std::vector< CString > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_CString_Sg____bool__(std::vector< CString > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< CString >::size_type std_vector_Sl_CString_Sg____len__(std::vector< CString > const *self){
      return self->size();
    }
SWIGINTERN std::vector< CString,std::allocator< CString > > *std_vector_Sl_CString_Sg____getslice__(std::vector< CString > *self,std::vector< CString >::difference_type i,std::vector< CString >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CString_Sg____setslice____SWIG_0(std::vector< CString > *self,std::vector< CString >::difference_type i,std::vector< CString >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< CString,std::allocator< CString > >());
    }
SWIGINTERN void std_vector_Sl_CString_Sg____setslice____SWIG_1(std::vector< CString > *self,std::vector< CString >::difference_type i,std::vector< CString >::difference_type j,std::vector< CString,std::allocator< CString > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_CString_Sg____delslice__(std::vector< CString > *self,std::vector< CString >::difference_type i,std::vector< CString >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CString_Sg____delitem____SWIG_0(std::vector< CString > *self,std::vector< CString >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< CString,std::allocator< CString > > *std_vector_Sl_CString_Sg____getitem____SWIG_0(std::vector< CString > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CString,std::allocator< CString > >::difference_type id = i;
      std::vector< CString,std::allocator< CString > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CString_Sg____setitem____SWIG_0(std::vector< CString > *self,SWIGPY_SLICEOBJECT *slice,std::vector< CString,std::allocator< CString > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CString,std::allocator< CString > >::difference_type id = i;
      std::vector< CString,std::allocator< CString > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_CString_Sg____setitem____SWIG_1(std::vector< CString > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CString,std::allocator< CString > >::difference_type id = i;
      std::vector< CString,std::allocator< CString > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CString_Sg____delitem____SWIG_1(std::vector< CString > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CString,std::allocator< CString > >::difference_type id = i;
      std::vector< CString,std::allocator< CString > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< CString >::value_type const &std_vector_Sl_CString_Sg____getitem____SWIG_1(std::vector< CString > const *self,std::vector< CString >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_CString_Sg____setitem____SWIG_2(std::vector< CString > *self,std::vector< CString >::difference_type i,std::vector< CString >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< CString >::value_type std_vector_Sl_CString_Sg__pop(std::vector< CString > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< CString,std::allocator< CString > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_CString_Sg__append(std::vector< CString > *self,std::vector< CString >::value_type const &x){
      self->push_back(x);
    }
SWIGINTERN std::vector< CString >::iterator std_vector_Sl_CString_Sg__erase__SWIG_0(std::vector< CString > *self,std::vector< CString >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< CString >::iterator std_vector_Sl_CString_Sg__erase__SWIG_1(std::vector< CString > *self,std::vector< CString >::iterator first,std::vector< CString >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< CString >::iterator std_vector_Sl_CString_Sg__insert__SWIG_0(std::vector< CString > *self,std::vector< CString >::iterator pos,std::vector< CString >::value_type const &x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_CString_Sg__insert__SWIG_1(std::vector< CString > *self,std::vector< CString >::iterator pos,std::vector< CString >::size_type n,std::vector< CString >::value_type const &x){ self->insert(pos, n, x); }

      namespace swig {
	template <>  struct traits<std::pair< CString, CString > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::pair<" "CString" "," "CString" " >";
	  }
	};
      }
    

      namespace swig {
	template <>  struct traits<std::map< CString, CString, std::less< CString >, std::allocator< std::pair< CString const,CString > > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::map<" "CString" "," "CString" "," "std::less< CString >" "," "std::allocator< std::pair< CString const,CString > >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_map_Sl_CString_Sc_CString_Sg__iterator(std::map< CString,CString > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_map_Sl_CString_Sc_CString_Sg____nonzero__(std::map< CString,CString > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_map_Sl_CString_Sc_CString_Sg____bool__(std::map< CString,CString > const *self){
      return !(self->empty());
    }
SWIGINTERN std::map< CString,CString >::size_type std_map_Sl_CString_Sc_CString_Sg____len__(std::map< CString,CString > const *self){
      return self->size();
    }
SWIGINTERN std::map< CString,CString >::mapped_type const &std_map_Sl_CString_Sc_CString_Sg____getitem__(std::map< CString,CString > *self,std::map< CString,CString >::key_type const &key){
      std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >::const_iterator i = self->find(key);
      if (i != self->end())
	return i->second;
      else
	throw std::out_of_range("key not found");
    }
SWIGINTERN void std_map_Sl_CString_Sc_CString_Sg____delitem__(std::map< CString,CString > *self,std::map< CString,CString >::key_type const &key){
      std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >::iterator i = self->find(key);
      if (i != self->end())
	self->erase(i);
      else
	throw std::out_of_range("key not found");
    }
SWIGINTERN bool std_map_Sl_CString_Sc_CString_Sg__has_key(std::map< CString,CString > const *self,std::map< CString,CString >::key_type const &key){
      std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >::const_iterator i = self->find(key);
      return i != self->end();
    }
SWIGINTERN PyObject *std_map_Sl_CString_Sc_CString_Sg__keys(std::map< CString,CString > *self){
      std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >::size_type size = self->size();
      Py_ssize_t pysize = (size <= (std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >::size_type) INT_MAX) ? (Py_ssize_t) size : -1;
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      if (pysize < 0) {
	PyErr_SetString(PyExc_OverflowError, "map size not valid in python");
	SWIG_PYTHON_THREAD_END_BLOCK;
	return NULL;
      }
      PyObject* keyList = PyList_New(pysize);
      std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >::const_iterator i = self->begin();
      for (Py_ssize_t j = 0; j < pysize; ++i, ++j) {
	PyList_SET_ITEM(keyList, j, swig::from(i->first));
      }
      SWIG_PYTHON_THREAD_END_BLOCK;
      return keyList;
    }
SWIGINTERN PyObject *std_map_Sl_CString_Sc_CString_Sg__values(std::map< CString,CString > *self){
      std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >::size_type size = self->size();
      Py_ssize_t pysize = (size <= (std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >::size_type) INT_MAX) ? (Py_ssize_t) size : -1;
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      if (pysize < 0) {
	PyErr_SetString(PyExc_OverflowError, "map size not valid in python");
	SWIG_PYTHON_THREAD_END_BLOCK;
	return NULL;
      }
      PyObject* valList = PyList_New(pysize);
      std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >::const_iterator i = self->begin();
      for (Py_ssize_t j = 0; j < pysize; ++i, ++j) {
	PyList_SET_ITEM(valList, j, swig::from(i->second));
      }
      SWIG_PYTHON_THREAD_END_BLOCK;
      return valList;
    }
SWIGINTERN PyObject *std_map_Sl_CString_Sc_CString_Sg__items(std::map< CString,CString > *self){
      std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >::size_type size = self->size();
      Py_ssize_t pysize = (size <= (std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >::size_type) INT_MAX) ? (Py_ssize_t) size : -1;
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      if (pysize < 0) {
	PyErr_SetString(PyExc_OverflowError, "map size not valid in python");
	SWIG_PYTHON_THREAD_END_BLOCK;
	return NULL;
      }    
      PyObject* itemList = PyList_New(pysize);
      std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >::const_iterator i = self->begin();
      for (Py_ssize_t j = 0; j < pysize; ++i, ++j) {
	PyList_SET_ITEM(itemList, j, swig::from(*i));
      }
      SWIG_PYTHON_THREAD_END_BLOCK;
      return itemList;
    }
SWIGINTERN bool std_map_Sl_CString_Sc_CString_Sg____contains__(std::map< CString,CString > *self,std::map< CString,CString >::key_type const &key){
      return self->find(key) != self->end();
    }
SWIGINTERN swig::SwigPyIterator *std_map_Sl_CString_Sc_CString_Sg__key_iterator(std::map< CString,CString > *self,PyObject **PYTHON_SELF){
      return swig::make_output_key_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN swig::SwigPyIterator *std_map_Sl_CString_Sc_CString_Sg__value_iterator(std::map< CString,CString > *self,PyObject **PYTHON_SELF){
      return swig::make_output_value_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN void std_map_Sl_CString_Sc_CString_Sg____setitem____SWIG_0(std::map< CString,CString > *self,std::map< CString,CString >::key_type const &key){
      self->erase(key);
    }
SWIGINTERN void std_map_Sl_CString_Sc_CString_Sg____setitem____SWIG_1(std::map< CString,CString > *self,std::map< CString,CString >::key_type const &key,std::map< CString,CString >::mapped_type const &x){
#ifdef __cpp_lib_map_try_emplace
      (*self).insert_or_assign(key, x);
#else
      (*self)[key] = x;
#endif
    }
SWIGINTERN PyObject *std_map_Sl_CString_Sc_CString_Sg__asdict(std::map< CString,CString > *self){
      return swig::traits_from< std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > > >::asdict(*self);
    }
SWIGINTERN void std_map_Sl_CString_Sc_CString_Sg__erase__SWIG_1(std::map< CString,CString > *self,std::map< CString,CString >::iterator position){ self->erase(position); }
SWIGINTERN void std_map_Sl_CString_Sc_CString_Sg__erase__SWIG_2(std::map< CString,CString > *self,std::map< CString,CString >::iterator first,std::map< CString,CString >::iterator last){ self->erase(first, last); }

      namespace swig {
	template <>  struct traits<std::pair< CString, std::vector< CString,std::allocator< CString > > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::pair<" "CString" "," "std::vector< CString,std::allocator< CString > >" " >";
	  }
	};
      }
    

      namespace swig {
	template <>  struct traits<std::map< CString, std::vector< CString,std::allocator< CString > >, std::less< CString >, std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::map<" "CString" "," "std::vector< CString,std::allocator< CString > >" "," "std::less< CString >" "," "std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_map_Sl_CString_Sc_VCString_Sg__iterator(std::map< CString,VCString > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_map_Sl_CString_Sc_VCString_Sg____nonzero__(std::map< CString,VCString > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_map_Sl_CString_Sc_VCString_Sg____bool__(std::map< CString,VCString > const *self){
      return !(self->empty());
    }
SWIGINTERN std::map< CString,std::vector< CString,std::allocator< CString > > >::size_type std_map_Sl_CString_Sc_VCString_Sg____len__(std::map< CString,VCString > const *self){
      return self->size();
    }
SWIGINTERN std::map< CString,std::vector< CString,std::allocator< CString > > >::mapped_type const &std_map_Sl_CString_Sc_VCString_Sg____getitem__(std::map< CString,VCString > *self,std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &key){
      std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >::const_iterator i = self->find(key);
      if (i != self->end())
	return i->second;
      else
	throw std::out_of_range("key not found");
    }
SWIGINTERN void std_map_Sl_CString_Sc_VCString_Sg____delitem__(std::map< CString,VCString > *self,std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &key){
      std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >::iterator i = self->find(key);
      if (i != self->end())
	self->erase(i);
      else
	throw std::out_of_range("key not found");
    }
SWIGINTERN bool std_map_Sl_CString_Sc_VCString_Sg__has_key(std::map< CString,VCString > const *self,std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &key){
      std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >::const_iterator i = self->find(key);
      return i != self->end();
    }
SWIGINTERN PyObject *std_map_Sl_CString_Sc_VCString_Sg__keys(std::map< CString,VCString > *self){
      std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >::size_type size = self->size();
      Py_ssize_t pysize = (size <= (std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >::size_type) INT_MAX) ? (Py_ssize_t) size : -1;
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      if (pysize < 0) {
	PyErr_SetString(PyExc_OverflowError, "map size not valid in python");
	SWIG_PYTHON_THREAD_END_BLOCK;
	return NULL;
      }
      PyObject* keyList = PyList_New(pysize);
      std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >::const_iterator i = self->begin();
      for (Py_ssize_t j = 0; j < pysize; ++i, ++j) {
	PyList_SET_ITEM(keyList, j, swig::from(i->first));
      }
      SWIG_PYTHON_THREAD_END_BLOCK;
      return keyList;
    }
SWIGINTERN PyObject *std_map_Sl_CString_Sc_VCString_Sg__values(std::map< CString,VCString > *self){
      std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >::size_type size = self->size();
      Py_ssize_t pysize = (size <= (std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >::size_type) INT_MAX) ? (Py_ssize_t) size : -1;
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      if (pysize < 0) {
	PyErr_SetString(PyExc_OverflowError, "map size not valid in python");
	SWIG_PYTHON_THREAD_END_BLOCK;
	return NULL;
      }
      PyObject* valList = PyList_New(pysize);
      std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >::const_iterator i = self->begin();
      for (Py_ssize_t j = 0; j < pysize; ++i, ++j) {
	PyList_SET_ITEM(valList, j, swig::from(i->second));
      }
      SWIG_PYTHON_THREAD_END_BLOCK;
      return valList;
    }
SWIGINTERN PyObject *std_map_Sl_CString_Sc_VCString_Sg__items(std::map< CString,VCString > *self){
      std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >::size_type size = self->size();
      Py_ssize_t pysize = (size <= (std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >::size_type) INT_MAX) ? (Py_ssize_t) size : -1;
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      if (pysize < 0) {
	PyErr_SetString(PyExc_OverflowError, "map size not valid in python");
	SWIG_PYTHON_THREAD_END_BLOCK;
	return NULL;
      }    
      PyObject* itemList = PyList_New(pysize);
      std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >::const_iterator i = self->begin();
      for (Py_ssize_t j = 0; j < pysize; ++i, ++j) {
	PyList_SET_ITEM(itemList, j, swig::from(*i));
      }
      SWIG_PYTHON_THREAD_END_BLOCK;
      return itemList;
    }
SWIGINTERN bool std_map_Sl_CString_Sc_VCString_Sg____contains__(std::map< CString,VCString > *self,std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &key){
      return self->find(key) != self->end();
    }
SWIGINTERN swig::SwigPyIterator *std_map_Sl_CString_Sc_VCString_Sg__key_iterator(std::map< CString,VCString > *self,PyObject **PYTHON_SELF){
      return swig::make_output_key_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN swig::SwigPyIterator *std_map_Sl_CString_Sc_VCString_Sg__value_iterator(std::map< CString,VCString > *self,PyObject **PYTHON_SELF){
      return swig::make_output_value_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN void std_map_Sl_CString_Sc_VCString_Sg____setitem____SWIG_0(std::map< CString,VCString > *self,std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &key){
      self->erase(key);
    }
SWIGINTERN void std_map_Sl_CString_Sc_VCString_Sg____setitem____SWIG_1(std::map< CString,VCString > *self,std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &key,std::map< CString,std::vector< CString,std::allocator< CString > > >::mapped_type const &x){
#ifdef __cpp_lib_map_try_emplace
      (*self).insert_or_assign(key, x);
#else
      (*self)[key] = x;
#endif
    }
SWIGINTERN PyObject *std_map_Sl_CString_Sc_VCString_Sg__asdict(std::map< CString,VCString > *self){
      return swig::traits_from< std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > > >::asdict(*self);
    }
SWIGINTERN void std_map_Sl_CString_Sc_VCString_Sg__erase__SWIG_1(std::map< CString,VCString > *self,std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator position){ self->erase(position); }
SWIGINTERN void std_map_Sl_CString_Sc_VCString_Sg__erase__SWIG_2(std::map< CString,VCString > *self,std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator first,std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator last){ self->erase(first, last); }

  namespace swig {
    template <>  struct traits< CModule > {
      typedef pointer_category category;
      static const char* type_name() { return"CModule"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< CModule*, std::allocator< CModule * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "CModule" " *," "std::allocator< CModule * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_CModule_Sm__Sg__iterator(std::vector< CModule * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_CModule_Sm__Sg____nonzero__(std::vector< CModule * > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_CModule_Sm__Sg____bool__(std::vector< CModule * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< CModule * >::size_type std_vector_Sl_CModule_Sm__Sg____len__(std::vector< CModule * > const *self){
      return self->size();
    }
SWIGINTERN std::vector< CModule *,std::allocator< CModule * > > *std_vector_Sl_CModule_Sm__Sg____getslice__(std::vector< CModule * > *self,std::vector< CModule * >::difference_type i,std::vector< CModule * >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CModule_Sm__Sg____setslice____SWIG_0(std::vector< CModule * > *self,std::vector< CModule * >::difference_type i,std::vector< CModule * >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< CModule*,std::allocator< CModule * > >());
    }
SWIGINTERN void std_vector_Sl_CModule_Sm__Sg____setslice____SWIG_1(std::vector< CModule * > *self,std::vector< CModule * >::difference_type i,std::vector< CModule * >::difference_type j,std::vector< CModule *,std::allocator< CModule * > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_CModule_Sm__Sg____delslice__(std::vector< CModule * > *self,std::vector< CModule * >::difference_type i,std::vector< CModule * >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CModule_Sm__Sg____delitem____SWIG_0(std::vector< CModule * > *self,std::vector< CModule * >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< CModule *,std::allocator< CModule * > > *std_vector_Sl_CModule_Sm__Sg____getitem____SWIG_0(std::vector< CModule * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CModule*,std::allocator< CModule * > >::difference_type id = i;
      std::vector< CModule*,std::allocator< CModule * > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CModule_Sm__Sg____setitem____SWIG_0(std::vector< CModule * > *self,SWIGPY_SLICEOBJECT *slice,std::vector< CModule *,std::allocator< CModule * > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CModule*,std::allocator< CModule * > >::difference_type id = i;
      std::vector< CModule*,std::allocator< CModule * > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_CModule_Sm__Sg____setitem____SWIG_1(std::vector< CModule * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CModule*,std::allocator< CModule * > >::difference_type id = i;
      std::vector< CModule*,std::allocator< CModule * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CModule_Sm__Sg____delitem____SWIG_1(std::vector< CModule * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CModule*,std::allocator< CModule * > >::difference_type id = i;
      std::vector< CModule*,std::allocator< CModule * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< CModule * >::value_type std_vector_Sl_CModule_Sm__Sg____getitem____SWIG_1(std::vector< CModule * > *self,std::vector< CModule * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_CModule_Sm__Sg____setitem____SWIG_2(std::vector< CModule * > *self,std::vector< CModule * >::difference_type i,std::vector< CModule * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< CModule * >::value_type std_vector_Sl_CModule_Sm__Sg__pop(std::vector< CModule * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< CModule*,std::allocator< CModule * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_CModule_Sm__Sg__append(std::vector< CModule * > *self,std::vector< CModule * >::value_type x){
      self->push_back(x);
    }
SWIGINTERN std::vector< CModule * >::iterator std_vector_Sl_CModule_Sm__Sg__erase__SWIG_0(std::vector< CModule * > *self,std::vector< CModule * >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< CModule * >::iterator std_vector_Sl_CModule_Sm__Sg__erase__SWIG_1(std::vector< CModule * > *self,std::vector< CModule * >::iterator first,std::vector< CModule * >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< CModule * >::iterator std_vector_Sl_CModule_Sm__Sg__insert__SWIG_0(std::vector< CModule * > *self,std::vector< CModule * >::iterator pos,std::vector< CModule * >::value_type x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_CModule_Sm__Sg__insert__SWIG_1(std::vector< CModule * > *self,std::vector< CModule * >::iterator pos,std::vector< CModule * >::size_type n,std::vector< CModule * >::value_type x){ self->insert(pos, n, x); }

  namespace swig {
    template <>  struct traits< CListener > {
      typedef pointer_category category;
      static const char* type_name() { return"CListener"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< CListener*, std::allocator< CListener * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "CListener" " *," "std::allocator< CListener * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_CListener_Sm__Sg__iterator(std::vector< CListener * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_CListener_Sm__Sg____nonzero__(std::vector< CListener * > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_CListener_Sm__Sg____bool__(std::vector< CListener * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< CListener * >::size_type std_vector_Sl_CListener_Sm__Sg____len__(std::vector< CListener * > const *self){
      return self->size();
    }
SWIGINTERN std::vector< CListener *,std::allocator< CListener * > > *std_vector_Sl_CListener_Sm__Sg____getslice__(std::vector< CListener * > *self,std::vector< CListener * >::difference_type i,std::vector< CListener * >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CListener_Sm__Sg____setslice____SWIG_0(std::vector< CListener * > *self,std::vector< CListener * >::difference_type i,std::vector< CListener * >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< CListener*,std::allocator< CListener * > >());
    }
SWIGINTERN void std_vector_Sl_CListener_Sm__Sg____setslice____SWIG_1(std::vector< CListener * > *self,std::vector< CListener * >::difference_type i,std::vector< CListener * >::difference_type j,std::vector< CListener *,std::allocator< CListener * > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_CListener_Sm__Sg____delslice__(std::vector< CListener * > *self,std::vector< CListener * >::difference_type i,std::vector< CListener * >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CListener_Sm__Sg____delitem____SWIG_0(std::vector< CListener * > *self,std::vector< CListener * >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< CListener *,std::allocator< CListener * > > *std_vector_Sl_CListener_Sm__Sg____getitem____SWIG_0(std::vector< CListener * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CListener*,std::allocator< CListener * > >::difference_type id = i;
      std::vector< CListener*,std::allocator< CListener * > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CListener_Sm__Sg____setitem____SWIG_0(std::vector< CListener * > *self,SWIGPY_SLICEOBJECT *slice,std::vector< CListener *,std::allocator< CListener * > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CListener*,std::allocator< CListener * > >::difference_type id = i;
      std::vector< CListener*,std::allocator< CListener * > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_CListener_Sm__Sg____setitem____SWIG_1(std::vector< CListener * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CListener*,std::allocator< CListener * > >::difference_type id = i;
      std::vector< CListener*,std::allocator< CListener * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CListener_Sm__Sg____delitem____SWIG_1(std::vector< CListener * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CListener*,std::allocator< CListener * > >::difference_type id = i;
      std::vector< CListener*,std::allocator< CListener * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< CListener * >::value_type std_vector_Sl_CListener_Sm__Sg____getitem____SWIG_1(std::vector< CListener * > *self,std::vector< CListener * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_CListener_Sm__Sg____setitem____SWIG_2(std::vector< CListener * > *self,std::vector< CListener * >::difference_type i,std::vector< CListener * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< CListener * >::value_type std_vector_Sl_CListener_Sm__Sg__pop(std::vector< CListener * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< CListener*,std::allocator< CListener * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_CListener_Sm__Sg__append(std::vector< CListener * > *self,std::vector< CListener * >::value_type x){
      self->push_back(x);
    }
SWIGINTERN std::vector< CListener * >::iterator std_vector_Sl_CListener_Sm__Sg__erase__SWIG_0(std::vector< CListener * > *self,std::vector< CListener * >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< CListener * >::iterator std_vector_Sl_CListener_Sm__Sg__erase__SWIG_1(std::vector< CListener * > *self,std::vector< CListener * >::iterator first,std::vector< CListener * >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< CListener * >::iterator std_vector_Sl_CListener_Sm__Sg__insert__SWIG_0(std::vector< CListener * > *self,std::vector< CListener * >::iterator pos,std::vector< CListener * >::value_type x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_CListener_Sm__Sg__insert__SWIG_1(std::vector< CListener * > *self,std::vector< CListener * >::iterator pos,std::vector< CListener * >::size_type n,std::vector< CListener * >::value_type x){ self->insert(pos, n, x); }

  namespace swig {
    template <>  struct traits< CBufLine > {
      typedef pointer_category category;
      static const char* type_name() { return"CBufLine"; }
    };
  }


  namespace swig {
    template <class T>
    struct traits_asptr<std::deque<T> >  {
      static int asptr(PyObject *obj, std::deque<T>  **vec) {
	return traits_asptr_stdseq<std::deque<T> >::asptr(obj, vec);
      }
    };

    template <class T>
    struct traits_from<std::deque<T> > {
      static PyObject *from(const std::deque<T>& vec) {
	return traits_from_stdseq<std::deque<T> >::from(vec);
      }
    };
  }


      namespace swig {
	template <>  struct traits<std::deque< CBufLine, std::allocator< CBufLine > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::deque<" "CBufLine" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_deque_Sl_CBufLine_Sg__iterator(std::deque< CBufLine > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_deque_Sl_CBufLine_Sg____nonzero__(std::deque< CBufLine > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_deque_Sl_CBufLine_Sg____bool__(std::deque< CBufLine > const *self){
      return !(self->empty());
    }
SWIGINTERN std::deque< CBufLine >::size_type std_deque_Sl_CBufLine_Sg____len__(std::deque< CBufLine > const *self){
      return self->size();
    }
SWIGINTERN std::deque< CBufLine,std::allocator< CBufLine > > *std_deque_Sl_CBufLine_Sg____getslice__(std::deque< CBufLine > *self,std::deque< CBufLine >::difference_type i,std::deque< CBufLine >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_deque_Sl_CBufLine_Sg____setslice____SWIG_0(std::deque< CBufLine > *self,std::deque< CBufLine >::difference_type i,std::deque< CBufLine >::difference_type j){
      swig::setslice(self, i, j, 1, std::deque< CBufLine,std::allocator< CBufLine > >());
    }
SWIGINTERN void std_deque_Sl_CBufLine_Sg____setslice____SWIG_1(std::deque< CBufLine > *self,std::deque< CBufLine >::difference_type i,std::deque< CBufLine >::difference_type j,std::deque< CBufLine,std::allocator< CBufLine > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_deque_Sl_CBufLine_Sg____delslice__(std::deque< CBufLine > *self,std::deque< CBufLine >::difference_type i,std::deque< CBufLine >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_deque_Sl_CBufLine_Sg____delitem____SWIG_0(std::deque< CBufLine > *self,std::deque< CBufLine >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::deque< CBufLine,std::allocator< CBufLine > > *std_deque_Sl_CBufLine_Sg____getitem____SWIG_0(std::deque< CBufLine > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::deque< CBufLine,std::allocator< CBufLine > >::difference_type id = i;
      std::deque< CBufLine,std::allocator< CBufLine > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_deque_Sl_CBufLine_Sg____setitem____SWIG_0(std::deque< CBufLine > *self,SWIGPY_SLICEOBJECT *slice,std::deque< CBufLine,std::allocator< CBufLine > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::deque< CBufLine,std::allocator< CBufLine > >::difference_type id = i;
      std::deque< CBufLine,std::allocator< CBufLine > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_deque_Sl_CBufLine_Sg____setitem____SWIG_1(std::deque< CBufLine > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::deque< CBufLine,std::allocator< CBufLine > >::difference_type id = i;
      std::deque< CBufLine,std::allocator< CBufLine > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_deque_Sl_CBufLine_Sg____delitem____SWIG_1(std::deque< CBufLine > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::deque< CBufLine,std::allocator< CBufLine > >::difference_type id = i;
      std::deque< CBufLine,std::allocator< CBufLine > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::deque< CBufLine >::value_type const &std_deque_Sl_CBufLine_Sg____getitem____SWIG_1(std::deque< CBufLine > const *self,std::deque< CBufLine >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_deque_Sl_CBufLine_Sg____setitem____SWIG_2(std::deque< CBufLine > *self,std::deque< CBufLine >::difference_type i,std::deque< CBufLine >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::deque< CBufLine >::value_type std_deque_Sl_CBufLine_Sg__pop(std::deque< CBufLine > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::deque< CBufLine,std::allocator< CBufLine > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_deque_Sl_CBufLine_Sg__append(std::deque< CBufLine > *self,std::deque< CBufLine >::value_type const &x){
      self->push_back(x);
    }
SWIGINTERN std::deque< CBufLine >::iterator std_deque_Sl_CBufLine_Sg__erase__SWIG_0(std::deque< CBufLine > *self,std::deque< CBufLine >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::deque< CBufLine >::iterator std_deque_Sl_CBufLine_Sg__erase__SWIG_1(std::deque< CBufLine > *self,std::deque< CBufLine >::iterator first,std::deque< CBufLine >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::deque< CBufLine >::iterator std_deque_Sl_CBufLine_Sg__insert__SWIG_0(std::deque< CBufLine > *self,std::deque< CBufLine >::iterator pos,std::deque< CBufLine >::value_type const &x){ return self->insert(pos, x); }
SWIGINTERN void std_deque_Sl_CBufLine_Sg__insert__SWIG_1(std::deque< CBufLine > *self,std::deque< CBufLine >::iterator pos,std::deque< CBufLine >::size_type n,std::deque< CBufLine >::value_type const &x){ self->insert(pos, n, x); }

      namespace swig {
	template <>  struct traits<std::vector< std::vector< CString,std::allocator< CString > >, std::allocator< std::vector< CString,std::allocator< CString > > > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "std::vector< CString,std::allocator< CString > >" "," "std::allocator< std::vector< CString,std::allocator< CString > > >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_VCString_Sg__iterator(std::vector< VCString > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_VCString_Sg____nonzero__(std::vector< VCString > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_VCString_Sg____bool__(std::vector< VCString > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< std::vector< CString,std::allocator< CString > > >::size_type std_vector_Sl_VCString_Sg____len__(std::vector< VCString > const *self){
      return self->size();
    }
SWIGINTERN std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *std_vector_Sl_VCString_Sg____getslice__(std::vector< VCString > *self,std::vector< std::vector< CString,std::allocator< CString > > >::difference_type i,std::vector< std::vector< CString,std::allocator< CString > > >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_VCString_Sg____setslice____SWIG_0(std::vector< VCString > *self,std::vector< std::vector< CString,std::allocator< CString > > >::difference_type i,std::vector< std::vector< CString,std::allocator< CString > > >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >());
    }
SWIGINTERN void std_vector_Sl_VCString_Sg____setslice____SWIG_1(std::vector< VCString > *self,std::vector< std::vector< CString,std::allocator< CString > > >::difference_type i,std::vector< std::vector< CString,std::allocator< CString > > >::difference_type j,std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_VCString_Sg____delslice__(std::vector< VCString > *self,std::vector< std::vector< CString,std::allocator< CString > > >::difference_type i,std::vector< std::vector< CString,std::allocator< CString > > >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_VCString_Sg____delitem____SWIG_0(std::vector< VCString > *self,std::vector< std::vector< CString,std::allocator< CString > > >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *std_vector_Sl_VCString_Sg____getitem____SWIG_0(std::vector< VCString > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >::difference_type id = i;
      std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_VCString_Sg____setitem____SWIG_0(std::vector< VCString > *self,SWIGPY_SLICEOBJECT *slice,std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >::difference_type id = i;
      std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_VCString_Sg____setitem____SWIG_1(std::vector< VCString > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >::difference_type id = i;
      std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_VCString_Sg____delitem____SWIG_1(std::vector< VCString > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >::difference_type id = i;
      std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &std_vector_Sl_VCString_Sg____getitem____SWIG_1(std::vector< VCString > const *self,std::vector< std::vector< CString,std::allocator< CString > > >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_VCString_Sg____setitem____SWIG_2(std::vector< VCString > *self,std::vector< std::vector< CString,std::allocator< CString > > >::difference_type i,std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< std::vector< CString,std::allocator< CString > > >::value_type std_vector_Sl_VCString_Sg__pop(std::vector< VCString > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_VCString_Sg__append(std::vector< VCString > *self,std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &x){
      self->push_back(x);
    }
SWIGINTERN std::vector< std::vector< CString,std::allocator< CString > > >::iterator std_vector_Sl_VCString_Sg__erase__SWIG_0(std::vector< VCString > *self,std::vector< std::vector< CString,std::allocator< CString > > >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< std::vector< CString,std::allocator< CString > > >::iterator std_vector_Sl_VCString_Sg__erase__SWIG_1(std::vector< VCString > *self,std::vector< std::vector< CString,std::allocator< CString > > >::iterator first,std::vector< std::vector< CString,std::allocator< CString > > >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< std::vector< CString,std::allocator< CString > > >::iterator std_vector_Sl_VCString_Sg__insert__SWIG_0(std::vector< VCString > *self,std::vector< std::vector< CString,std::allocator< CString > > >::iterator pos,std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_VCString_Sg__insert__SWIG_1(std::vector< VCString > *self,std::vector< std::vector< CString,std::allocator< CString > > >::iterator pos,std::vector< std::vector< CString,std::allocator< CString > > >::size_type n,std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &x){ self->insert(pos, n, x); }

  namespace swig {
    template <>  struct traits< CClient > {
      typedef pointer_category category;
      static const char* type_name() { return"CClient"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< CClient*, std::allocator< CClient * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "CClient" " *," "std::allocator< CClient * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_CClient_Sm__Sg__iterator(std::vector< CClient * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_CClient_Sm__Sg____nonzero__(std::vector< CClient * > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_CClient_Sm__Sg____bool__(std::vector< CClient * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< CClient * >::size_type std_vector_Sl_CClient_Sm__Sg____len__(std::vector< CClient * > const *self){
      return self->size();
    }
SWIGINTERN std::vector< CClient *,std::allocator< CClient * > > *std_vector_Sl_CClient_Sm__Sg____getslice__(std::vector< CClient * > *self,std::vector< CClient * >::difference_type i,std::vector< CClient * >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CClient_Sm__Sg____setslice____SWIG_0(std::vector< CClient * > *self,std::vector< CClient * >::difference_type i,std::vector< CClient * >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< CClient*,std::allocator< CClient * > >());
    }
SWIGINTERN void std_vector_Sl_CClient_Sm__Sg____setslice____SWIG_1(std::vector< CClient * > *self,std::vector< CClient * >::difference_type i,std::vector< CClient * >::difference_type j,std::vector< CClient *,std::allocator< CClient * > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_CClient_Sm__Sg____delslice__(std::vector< CClient * > *self,std::vector< CClient * >::difference_type i,std::vector< CClient * >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CClient_Sm__Sg____delitem____SWIG_0(std::vector< CClient * > *self,std::vector< CClient * >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< CClient *,std::allocator< CClient * > > *std_vector_Sl_CClient_Sm__Sg____getitem____SWIG_0(std::vector< CClient * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CClient*,std::allocator< CClient * > >::difference_type id = i;
      std::vector< CClient*,std::allocator< CClient * > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CClient_Sm__Sg____setitem____SWIG_0(std::vector< CClient * > *self,SWIGPY_SLICEOBJECT *slice,std::vector< CClient *,std::allocator< CClient * > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CClient*,std::allocator< CClient * > >::difference_type id = i;
      std::vector< CClient*,std::allocator< CClient * > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_CClient_Sm__Sg____setitem____SWIG_1(std::vector< CClient * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CClient*,std::allocator< CClient * > >::difference_type id = i;
      std::vector< CClient*,std::allocator< CClient * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CClient_Sm__Sg____delitem____SWIG_1(std::vector< CClient * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CClient*,std::allocator< CClient * > >::difference_type id = i;
      std::vector< CClient*,std::allocator< CClient * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< CClient * >::value_type std_vector_Sl_CClient_Sm__Sg____getitem____SWIG_1(std::vector< CClient * > *self,std::vector< CClient * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_CClient_Sm__Sg____setitem____SWIG_2(std::vector< CClient * > *self,std::vector< CClient * >::difference_type i,std::vector< CClient * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< CClient * >::value_type std_vector_Sl_CClient_Sm__Sg__pop(std::vector< CClient * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< CClient*,std::allocator< CClient * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_CClient_Sm__Sg__append(std::vector< CClient * > *self,std::vector< CClient * >::value_type x){
      self->push_back(x);
    }
SWIGINTERN std::vector< CClient * >::iterator std_vector_Sl_CClient_Sm__Sg__erase__SWIG_0(std::vector< CClient * > *self,std::vector< CClient * >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< CClient * >::iterator std_vector_Sl_CClient_Sm__Sg__erase__SWIG_1(std::vector< CClient * > *self,std::vector< CClient * >::iterator first,std::vector< CClient * >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< CClient * >::iterator std_vector_Sl_CClient_Sm__Sg__insert__SWIG_0(std::vector< CClient * > *self,std::vector< CClient * >::iterator pos,std::vector< CClient * >::value_type x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_CClient_Sm__Sg__insert__SWIG_1(std::vector< CClient * > *self,std::vector< CClient * >::iterator pos,std::vector< CClient * >::size_type n,std::vector< CClient * >::value_type x){ self->insert(pos, n, x); }

  namespace swig {
    template <>  struct traits< CServer > {
      typedef pointer_category category;
      static const char* type_name() { return"CServer"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< CServer*, std::allocator< CServer * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "CServer" " *," "std::allocator< CServer * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_CServer_Sm__Sg__iterator(std::vector< CServer * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_CServer_Sm__Sg____nonzero__(std::vector< CServer * > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_CServer_Sm__Sg____bool__(std::vector< CServer * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< CServer * >::size_type std_vector_Sl_CServer_Sm__Sg____len__(std::vector< CServer * > const *self){
      return self->size();
    }
SWIGINTERN std::vector< CServer *,std::allocator< CServer * > > *std_vector_Sl_CServer_Sm__Sg____getslice__(std::vector< CServer * > *self,std::vector< CServer * >::difference_type i,std::vector< CServer * >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CServer_Sm__Sg____setslice____SWIG_0(std::vector< CServer * > *self,std::vector< CServer * >::difference_type i,std::vector< CServer * >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< CServer*,std::allocator< CServer * > >());
    }
SWIGINTERN void std_vector_Sl_CServer_Sm__Sg____setslice____SWIG_1(std::vector< CServer * > *self,std::vector< CServer * >::difference_type i,std::vector< CServer * >::difference_type j,std::vector< CServer *,std::allocator< CServer * > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_CServer_Sm__Sg____delslice__(std::vector< CServer * > *self,std::vector< CServer * >::difference_type i,std::vector< CServer * >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CServer_Sm__Sg____delitem____SWIG_0(std::vector< CServer * > *self,std::vector< CServer * >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< CServer *,std::allocator< CServer * > > *std_vector_Sl_CServer_Sm__Sg____getitem____SWIG_0(std::vector< CServer * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CServer*,std::allocator< CServer * > >::difference_type id = i;
      std::vector< CServer*,std::allocator< CServer * > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CServer_Sm__Sg____setitem____SWIG_0(std::vector< CServer * > *self,SWIGPY_SLICEOBJECT *slice,std::vector< CServer *,std::allocator< CServer * > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CServer*,std::allocator< CServer * > >::difference_type id = i;
      std::vector< CServer*,std::allocator< CServer * > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_CServer_Sm__Sg____setitem____SWIG_1(std::vector< CServer * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CServer*,std::allocator< CServer * > >::difference_type id = i;
      std::vector< CServer*,std::allocator< CServer * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CServer_Sm__Sg____delitem____SWIG_1(std::vector< CServer * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CServer*,std::allocator< CServer * > >::difference_type id = i;
      std::vector< CServer*,std::allocator< CServer * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< CServer * >::value_type std_vector_Sl_CServer_Sm__Sg____getitem____SWIG_1(std::vector< CServer * > *self,std::vector< CServer * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_CServer_Sm__Sg____setitem____SWIG_2(std::vector< CServer * > *self,std::vector< CServer * >::difference_type i,std::vector< CServer * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< CServer * >::value_type std_vector_Sl_CServer_Sm__Sg__pop(std::vector< CServer * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< CServer*,std::allocator< CServer * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_CServer_Sm__Sg__append(std::vector< CServer * > *self,std::vector< CServer * >::value_type x){
      self->push_back(x);
    }
SWIGINTERN std::vector< CServer * >::iterator std_vector_Sl_CServer_Sm__Sg__erase__SWIG_0(std::vector< CServer * > *self,std::vector< CServer * >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< CServer * >::iterator std_vector_Sl_CServer_Sm__Sg__erase__SWIG_1(std::vector< CServer * > *self,std::vector< CServer * >::iterator first,std::vector< CServer * >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< CServer * >::iterator std_vector_Sl_CServer_Sm__Sg__insert__SWIG_0(std::vector< CServer * > *self,std::vector< CServer * >::iterator pos,std::vector< CServer * >::value_type x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_CServer_Sm__Sg__insert__SWIG_1(std::vector< CServer * > *self,std::vector< CServer * >::iterator pos,std::vector< CServer * >::size_type n,std::vector< CServer * >::value_type x){ self->insert(pos, n, x); }

  namespace swig {
    template <>  struct traits< CQuery > {
      typedef pointer_category category;
      static const char* type_name() { return"CQuery"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< CQuery*, std::allocator< CQuery * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "CQuery" " *," "std::allocator< CQuery * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_CQuery_Sm__Sg__iterator(std::vector< CQuery * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_CQuery_Sm__Sg____nonzero__(std::vector< CQuery * > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_CQuery_Sm__Sg____bool__(std::vector< CQuery * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< CQuery * >::size_type std_vector_Sl_CQuery_Sm__Sg____len__(std::vector< CQuery * > const *self){
      return self->size();
    }
SWIGINTERN std::vector< CQuery *,std::allocator< CQuery * > > *std_vector_Sl_CQuery_Sm__Sg____getslice__(std::vector< CQuery * > *self,std::vector< CQuery * >::difference_type i,std::vector< CQuery * >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CQuery_Sm__Sg____setslice____SWIG_0(std::vector< CQuery * > *self,std::vector< CQuery * >::difference_type i,std::vector< CQuery * >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< CQuery*,std::allocator< CQuery * > >());
    }
SWIGINTERN void std_vector_Sl_CQuery_Sm__Sg____setslice____SWIG_1(std::vector< CQuery * > *self,std::vector< CQuery * >::difference_type i,std::vector< CQuery * >::difference_type j,std::vector< CQuery *,std::allocator< CQuery * > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_CQuery_Sm__Sg____delslice__(std::vector< CQuery * > *self,std::vector< CQuery * >::difference_type i,std::vector< CQuery * >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_CQuery_Sm__Sg____delitem____SWIG_0(std::vector< CQuery * > *self,std::vector< CQuery * >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< CQuery *,std::allocator< CQuery * > > *std_vector_Sl_CQuery_Sm__Sg____getitem____SWIG_0(std::vector< CQuery * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CQuery*,std::allocator< CQuery * > >::difference_type id = i;
      std::vector< CQuery*,std::allocator< CQuery * > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CQuery_Sm__Sg____setitem____SWIG_0(std::vector< CQuery * > *self,SWIGPY_SLICEOBJECT *slice,std::vector< CQuery *,std::allocator< CQuery * > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CQuery*,std::allocator< CQuery * > >::difference_type id = i;
      std::vector< CQuery*,std::allocator< CQuery * > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_CQuery_Sm__Sg____setitem____SWIG_1(std::vector< CQuery * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CQuery*,std::allocator< CQuery * > >::difference_type id = i;
      std::vector< CQuery*,std::allocator< CQuery * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_CQuery_Sm__Sg____delitem____SWIG_1(std::vector< CQuery * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< CQuery*,std::allocator< CQuery * > >::difference_type id = i;
      std::vector< CQuery*,std::allocator< CQuery * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< CQuery * >::value_type std_vector_Sl_CQuery_Sm__Sg____getitem____SWIG_1(std::vector< CQuery * > *self,std::vector< CQuery * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_CQuery_Sm__Sg____setitem____SWIG_2(std::vector< CQuery * > *self,std::vector< CQuery * >::difference_type i,std::vector< CQuery * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< CQuery * >::value_type std_vector_Sl_CQuery_Sm__Sg__pop(std::vector< CQuery * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< CQuery*,std::allocator< CQuery * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_CQuery_Sm__Sg__append(std::vector< CQuery * > *self,std::vector< CQuery * >::value_type x){
      self->push_back(x);
    }
SWIGINTERN std::vector< CQuery * >::iterator std_vector_Sl_CQuery_Sm__Sg__erase__SWIG_0(std::vector< CQuery * > *self,std::vector< CQuery * >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< CQuery * >::iterator std_vector_Sl_CQuery_Sm__Sg__erase__SWIG_1(std::vector< CQuery * > *self,std::vector< CQuery * >::iterator first,std::vector< CQuery * >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< CQuery * >::iterator std_vector_Sl_CQuery_Sm__Sg__insert__SWIG_0(std::vector< CQuery * > *self,std::vector< CQuery * >::iterator pos,std::vector< CQuery * >::value_type x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_CQuery_Sm__Sg__insert__SWIG_1(std::vector< CQuery * > *self,std::vector< CQuery * >::iterator pos,std::vector< CQuery * >::size_type n,std::vector< CQuery * >::value_type x){ self->insert(pos, n, x); }

SWIGINTERN int
SWIG_AsVal_int (PyObject * obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< int >(v);
    }
  }  
  return res;
}





SWIGINTERNINLINE PyObject * 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}


SWIGINTERN int
SWIG_AsVal_bool (PyObject *obj, bool *val)
{
  int r;
  if (!PyBool_Check(obj))
    return SWIG_ERROR;
  r = PyObject_IsTrue(obj);
  if (r == -1)
    return SWIG_ERROR;
  if (val) *val = r ? true : false;
  return SWIG_OK;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_int (PyObject * obj, unsigned int *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UINT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< unsigned int >(v);
    }
  }  
  return res;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_short (PyObject * obj, unsigned short *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > USHRT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< unsigned short >(v);
    }
  }  
  return res;
}


SWIGINTERN int
SWIG_AsVal_short (PyObject * obj, short *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < SHRT_MIN || v > SHRT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< short >(v);
    }
  }  
  return res;
}


/* Getting isfinite working pre C99 across multiple platforms is non-trivial. Users can provide SWIG_isfinite on older platforms. */
#ifndef SWIG_isfinite
/* isfinite() is a macro for C99 */
# if defined(isfinite)
#  define SWIG_isfinite(X) (isfinite(X))
# elif defined(__cplusplus) && __cplusplus >= 201103L
/* Use a template so that this works whether isfinite() is std::isfinite() or
 * in the global namespace.  The reality seems to vary between compiler
 * versions.
 *
 * Make sure namespace std exists to avoid compiler warnings.
 *
 * extern "C++" is required as this fragment can end up inside an extern "C" { } block
 */
namespace std { }
extern "C++" template<typename T>
inline int SWIG_isfinite_func(T x) {
  using namespace std;
  return isfinite(x);
}
#  define SWIG_isfinite(X) (SWIG_isfinite_func(X))
# elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 2))
#  define SWIG_isfinite(X) (__builtin_isfinite(X))
# elif defined(_MSC_VER)
#  define SWIG_isfinite(X) (_finite(X))
# elif defined(__sun) && defined(__SVR4)
#  include <ieeefp.h>
#  define SWIG_isfinite(X) (finite(X))
# endif
#endif


/* Accept infinite as a valid float value unless we are unable to check if a value is finite */
#ifdef SWIG_isfinite
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX) && SWIG_isfinite(X))
#else
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX))
#endif


SWIGINTERN int
SWIG_AsVal_float (PyObject * obj, float *val)
{
  double v;
  int res = SWIG_AsVal_double (obj, &v);
  if (SWIG_IsOK(res)) {
    if (SWIG_Float_Overflow_Check(v)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< float >(v);
    }
  }  
  return res;
}


  #define SWIG_From_double   PyFloat_FromDouble 

SWIGINTERN PyObject *Csock_WriteBytes(Csock *self,PyObject *data){
        if (!PyBytes_Check(data)) {
            PyErr_SetString(PyExc_TypeError, "socket.WriteBytes needs bytes as argument");
            return nullptr;
        }
        char* buffer;
        Py_ssize_t length;
        if (-1 == PyBytes_AsStringAndSize(data, &buffer, &length)) {
            return nullptr;
        }
        if (self->Write(buffer, length)) {
            Py_RETURN_TRUE;
        } else {
            Py_RETURN_FALSE;
        }
    }

SWIGINTERNINLINE PyObject *
SWIG_From_unsigned_SS_short  (unsigned short value)
{    
  return SWIG_From_unsigned_SS_long  (value);
}


SWIGINTERNINLINE PyObject*
  SWIG_From_unsigned_SS_int  (unsigned int value)
{
  return PyInt_FromSize_t((size_t) value);
}


#include <memory>

SWIGINTERN CString CMessage___str__(CMessage *self){
        return self->ToString();
    }
SWIGINTERN CString CMessage___repr__(CMessage *self){
        return self->ToString();
    }

SWIGINTERN int
SWIG_AsCharArray(PyObject * obj, char *val, size_t size)
{ 
  char* cptr = 0; size_t csize = 0; int alloc = SWIG_OLDOBJ;
  int res = SWIG_AsCharPtrAndSize(obj, &cptr, &csize, &alloc);
  if (SWIG_IsOK(res)) {
    /* special case of single char conversion when we don't need space for NUL */
    if (size == 1 && csize == 2 && cptr && !cptr[1]) --csize;
    if (csize <= size) {
      if (val) {
	if (csize) memcpy(val, cptr, csize*sizeof(char));
	if (csize < size) memset(val + csize, 0, (size - csize)*sizeof(char));
      }
      if (alloc == SWIG_NEWOBJ) {
	delete[] cptr;
	res = SWIG_DelNewMask(res);
      }      
      return res;
    }
    if (alloc == SWIG_NEWOBJ) delete[] cptr;
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_char (PyObject * obj, char *val)
{    
  int res = SWIG_AsCharArray(obj, val, 1);
  if (!SWIG_IsOK(res)) {
    long v;
    res = SWIG_AddCast(SWIG_AsVal_long (obj, &v));
    if (SWIG_IsOK(res)) {
      if ((CHAR_MIN <= v) && (v <= CHAR_MAX)) {
	if (val) *val = static_cast< char >(v);
      } else {
	res = SWIG_OverflowError;
      }
    }
  }
  return res;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_char (PyObject * obj, unsigned char *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UCHAR_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< unsigned char >(v);
    }
  }  
  return res;
}

SWIGINTERN CString CModule___str__(CModule *self){
		return self->GetModName();
	}
SWIGINTERN MCString_iter CModule_BeginNV_(CModule *self){
		return MCString_iter(self->BeginNV());
	}
SWIGINTERN bool CModule_ExistsNV(CModule *self,CString const &sName){
		return self->EndNV() != self->FindNV(sName);
	}
SWIGINTERN void CModule_AddServerDependentCapability(CModule *self,CString const &sName,PyObject *serverCb,PyObject *clientCb){
        self->AddServerDependentCapability(sName, std::make_unique<CPyCapability>(serverCb, clientCb));
	}
SWIGINTERN bool CModules_removeModule(CModules *self,CModule *p){
		for (CModules::iterator i = self->begin(); self->end() != i; ++i) {
			if (*i == p) {
				self->erase(i);
				return true;
			}
		}
		return false;
	}

SWIGINTERNINLINE PyObject *
SWIG_From_char  (char c) 
{ 
  return SWIG_FromCharPtrAndSize(&c,1);
}

SWIGINTERN CString CNick___str__(CNick *self){
		return self->GetNick();
	}
SWIGINTERN CString CNick___repr__(CNick *self){
		return "<CNick " + self->GetHostMask() + ">";
	}
SWIGINTERN CString CChan___str__(CChan *self){
		return self->GetName();
	}
SWIGINTERN CString CChan___repr__(CChan *self){
		return "<CChan " + self->GetName() + ">";
	}
SWIGINTERN std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > > CChan_GetNicks_(CChan *self){
		return self->GetNicks();
	}
SWIGINTERN CString CUser___str__(CUser *self){
		return self->GetUsername();
	}
SWIGINTERN CString CUser___repr__(CUser *self){
		return "<CUser " + self->GetUsername() + ">";
	}
SWIGINTERN std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > CUser_GetNetworks_(CUser *self){
		return self->GetNetworks();
	}
SWIGINTERN CString CIRCNetwork___str__(CIRCNetwork *self){
		return self->GetName();
	}
SWIGINTERN CString CIRCNetwork___repr__(CIRCNetwork *self){
		return "<CIRCNetwork " + self->GetName() + ">";
	}
SWIGINTERN std::vector< CChan *,std::allocator< CChan * > > CIRCNetwork_GetChans_(CIRCNetwork *self){
		return self->GetChans();
	}
SWIGINTERN std::vector< CServer *,std::allocator< CServer * > > CIRCNetwork_GetServers_(CIRCNetwork *self){
		return self->GetServers();
	}
SWIGINTERN std::vector< CQuery *,std::allocator< CQuery * > > CIRCNetwork_GetQueries_(CIRCNetwork *self){
		return self->GetQueries();
	}

struct SWIG_null_deleter {
  void operator() (void const *) const {
  }
};
#define SWIG_NO_NULL_DELETER_0 , SWIG_null_deleter()
#define SWIG_NO_NULL_DELETER_1
#define SWIG_NO_NULL_DELETER_SWIG_POINTER_NEW
#define SWIG_NO_NULL_DELETER_SWIG_POINTER_OWN


#define SWIG_NO_NULL_DELETER_SWIG_BUILTIN_INIT

SWIGINTERN void CTemplate_set(CTemplate *self,CString const &key,CString const &value){
		     do {                                  if (CDebug::Debug()) {                CDebugStream sDebug;              sDebug << "WARNING: modpython's CTemplate.set is deprecated and will be removed. Use normal dict's operations like Tmpl['foo'] = 'bar'";                  }                             } while (0);
		(*self)[key] = value;
	}
SWIGINTERN PyObject *CZNC_GetUserMap_(CZNC *self){
        PyObject* result = PyDict_New();
        auto user_type = SWIG_TypeQuery("CUser*");
        for (const auto& p : self->GetUserMap()) {
            PyObject* user = SWIG_NewInstanceObj(p.second, user_type, 0);
            PyDict_SetItemString(result, p.first.c_str(), user);
            Py_CLEAR(user);
        }
        return result;
    }
SWIGINTERN CString String___str__(String *self){
		return self->s;
	}
SWIGINTERN CString CPyRetString___str__(CPyRetString *self){
		return self->s;
	}
SWIGINTERN bool CPyRetBool___bool__(CPyRetBool *self){
		return self->b;
	}

      namespace swig {
	template <>  struct traits<std::vector< std::pair< CString,CString >, std::allocator< std::pair< CString,CString > > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "std::pair< CString,CString >" "," "std::allocator< std::pair< CString,CString > >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__iterator(std::vector< std::pair< CString,CString > > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____nonzero__(std::vector< std::pair< CString,CString > > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____bool__(std::vector< std::pair< CString,CString > > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< std::pair< CString,CString > >::size_type std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____len__(std::vector< std::pair< CString,CString > > const *self){
      return self->size();
    }
SWIGINTERN std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____getslice__(std::vector< std::pair< CString,CString > > *self,std::vector< std::pair< CString,CString > >::difference_type i,std::vector< std::pair< CString,CString > >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____setslice____SWIG_0(std::vector< std::pair< CString,CString > > *self,std::vector< std::pair< CString,CString > >::difference_type i,std::vector< std::pair< CString,CString > >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >());
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____setslice____SWIG_1(std::vector< std::pair< CString,CString > > *self,std::vector< std::pair< CString,CString > >::difference_type i,std::vector< std::pair< CString,CString > >::difference_type j,std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____delslice__(std::vector< std::pair< CString,CString > > *self,std::vector< std::pair< CString,CString > >::difference_type i,std::vector< std::pair< CString,CString > >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____delitem____SWIG_0(std::vector< std::pair< CString,CString > > *self,std::vector< std::pair< CString,CString > >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____getitem____SWIG_0(std::vector< std::pair< CString,CString > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >::difference_type id = i;
      std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____setitem____SWIG_0(std::vector< std::pair< CString,CString > > *self,SWIGPY_SLICEOBJECT *slice,std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >::difference_type id = i;
      std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____setitem____SWIG_1(std::vector< std::pair< CString,CString > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >::difference_type id = i;
      std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____delitem____SWIG_1(std::vector< std::pair< CString,CString > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >::difference_type id = i;
      std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< std::pair< CString,CString > >::value_type const &std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____getitem____SWIG_1(std::vector< std::pair< CString,CString > > const *self,std::vector< std::pair< CString,CString > >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____setitem____SWIG_2(std::vector< std::pair< CString,CString > > *self,std::vector< std::pair< CString,CString > >::difference_type i,std::vector< std::pair< CString,CString > >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< std::pair< CString,CString > >::value_type std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__pop(std::vector< std::pair< CString,CString > > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__append(std::vector< std::pair< CString,CString > > *self,std::vector< std::pair< CString,CString > >::value_type const &x){
      self->push_back(x);
    }
SWIGINTERN std::vector< std::pair< CString,CString > >::iterator std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__erase__SWIG_0(std::vector< std::pair< CString,CString > > *self,std::vector< std::pair< CString,CString > >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< std::pair< CString,CString > >::iterator std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__erase__SWIG_1(std::vector< std::pair< CString,CString > > *self,std::vector< std::pair< CString,CString > >::iterator first,std::vector< std::pair< CString,CString > >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< std::pair< CString,CString > >::iterator std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__insert__SWIG_0(std::vector< std::pair< CString,CString > > *self,std::vector< std::pair< CString,CString > >::iterator pos,std::vector< std::pair< CString,CString > >::value_type const &x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__insert__SWIG_1(std::vector< std::pair< CString,CString > > *self,std::vector< std::pair< CString,CString > >::iterator pos,std::vector< std::pair< CString,CString > >::size_type n,std::vector< std::pair< CString,CString > >::value_type const &x){ self->insert(pos, n, x); }

  namespace swig {
    template <>  struct traits< std::shared_ptr< CWebSubPage > > {
      typedef pointer_category category;
      static const char* type_name() { return"std::shared_ptr< CWebSubPage >"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< std::shared_ptr< CWebSubPage >, std::allocator< std::shared_ptr< CWebSubPage > > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "std::shared_ptr< CWebSubPage >" "," "std::allocator< std::shared_ptr< CWebSubPage > >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_TWebSubPage_Sg__iterator(std::vector< TWebSubPage > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_TWebSubPage_Sg____nonzero__(std::vector< TWebSubPage > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_TWebSubPage_Sg____bool__(std::vector< TWebSubPage > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< std::shared_ptr< CWebSubPage > >::size_type std_vector_Sl_TWebSubPage_Sg____len__(std::vector< TWebSubPage > const *self){
      return self->size();
    }
SWIGINTERN std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *std_vector_Sl_TWebSubPage_Sg____getslice__(std::vector< TWebSubPage > *self,std::vector< std::shared_ptr< CWebSubPage > >::difference_type i,std::vector< std::shared_ptr< CWebSubPage > >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_TWebSubPage_Sg____setslice____SWIG_0(std::vector< TWebSubPage > *self,std::vector< std::shared_ptr< CWebSubPage > >::difference_type i,std::vector< std::shared_ptr< CWebSubPage > >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >());
    }
SWIGINTERN void std_vector_Sl_TWebSubPage_Sg____setslice____SWIG_1(std::vector< TWebSubPage > *self,std::vector< std::shared_ptr< CWebSubPage > >::difference_type i,std::vector< std::shared_ptr< CWebSubPage > >::difference_type j,std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_TWebSubPage_Sg____delslice__(std::vector< TWebSubPage > *self,std::vector< std::shared_ptr< CWebSubPage > >::difference_type i,std::vector< std::shared_ptr< CWebSubPage > >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_TWebSubPage_Sg____delitem____SWIG_0(std::vector< TWebSubPage > *self,std::vector< std::shared_ptr< CWebSubPage > >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *std_vector_Sl_TWebSubPage_Sg____getitem____SWIG_0(std::vector< TWebSubPage > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >::difference_type id = i;
      std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_TWebSubPage_Sg____setitem____SWIG_0(std::vector< TWebSubPage > *self,SWIGPY_SLICEOBJECT *slice,std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >::difference_type id = i;
      std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_TWebSubPage_Sg____setitem____SWIG_1(std::vector< TWebSubPage > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >::difference_type id = i;
      std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_TWebSubPage_Sg____delitem____SWIG_1(std::vector< TWebSubPage > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >::difference_type id = i;
      std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< std::shared_ptr< CWebSubPage > >::value_type const &std_vector_Sl_TWebSubPage_Sg____getitem____SWIG_1(std::vector< TWebSubPage > const *self,std::vector< std::shared_ptr< CWebSubPage > >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_TWebSubPage_Sg____setitem____SWIG_2(std::vector< TWebSubPage > *self,std::vector< std::shared_ptr< CWebSubPage > >::difference_type i,std::vector< std::shared_ptr< CWebSubPage > >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< std::shared_ptr< CWebSubPage > >::value_type std_vector_Sl_TWebSubPage_Sg__pop(std::vector< TWebSubPage > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_TWebSubPage_Sg__append(std::vector< TWebSubPage > *self,std::vector< std::shared_ptr< CWebSubPage > >::value_type const &x){
      self->push_back(x);
    }
SWIGINTERN std::vector< std::shared_ptr< CWebSubPage > >::iterator std_vector_Sl_TWebSubPage_Sg__erase__SWIG_0(std::vector< TWebSubPage > *self,std::vector< std::shared_ptr< CWebSubPage > >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< std::shared_ptr< CWebSubPage > >::iterator std_vector_Sl_TWebSubPage_Sg__erase__SWIG_1(std::vector< TWebSubPage > *self,std::vector< std::shared_ptr< CWebSubPage > >::iterator first,std::vector< std::shared_ptr< CWebSubPage > >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< std::shared_ptr< CWebSubPage > >::iterator std_vector_Sl_TWebSubPage_Sg__insert__SWIG_0(std::vector< TWebSubPage > *self,std::vector< std::shared_ptr< CWebSubPage > >::iterator pos,std::vector< std::shared_ptr< CWebSubPage > >::value_type const &x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_TWebSubPage_Sg__insert__SWIG_1(std::vector< TWebSubPage > *self,std::vector< std::shared_ptr< CWebSubPage > >::iterator pos,std::vector< std::shared_ptr< CWebSubPage > >::size_type n,std::vector< std::shared_ptr< CWebSubPage > >::value_type const &x){ self->insert(pos, n, x); }

	void VPair_Add2Str_(VPair* self, const CString& a, const CString& b) {
		self->push_back(std::make_pair(a, b));
	}


	TWebSubPage CreateWebSubPage_(const CString& sName, const CString& sTitle, const VPair& vParams, unsigned int uFlags) {
		return std::make_shared<CWebSubPage>(sName, sTitle, vParams, uFlags);
	}

#ifdef __cplusplus
extern "C" {
#endif
SWIGINTERN PyObject *_wrap_delete_SwigPyIterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SwigPyIterator" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_value" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  try {
    result = (PyObject *)((swig::SwigPyIterator const *)arg1)->value();
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_incr__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  swig::SwigPyIterator *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_incr" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator_incr" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = static_cast< size_t >(val2);
  try {
    result = (swig::SwigPyIterator *)(arg1)->incr(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_incr__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_incr" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  try {
    result = (swig::SwigPyIterator *)(arg1)->incr();
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_incr(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "SwigPyIterator_incr", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__SwigPyIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_SwigPyIterator_incr__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__SwigPyIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_SwigPyIterator_incr__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'SwigPyIterator_incr'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    swig::SwigPyIterator::incr(size_t)\n"
    "    swig::SwigPyIterator::incr()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_decr__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  swig::SwigPyIterator *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_decr" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator_decr" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = static_cast< size_t >(val2);
  try {
    result = (swig::SwigPyIterator *)(arg1)->decr(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_decr__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_decr" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  try {
    result = (swig::SwigPyIterator *)(arg1)->decr();
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_decr(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "SwigPyIterator_decr", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__SwigPyIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_SwigPyIterator_decr__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__SwigPyIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_SwigPyIterator_decr__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'SwigPyIterator_decr'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    swig::SwigPyIterator::decr(size_t)\n"
    "    swig::SwigPyIterator::decr()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_distance(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  swig::SwigPyIterator *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  ptrdiff_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator_distance", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_distance" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator_distance" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SwigPyIterator_distance" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2);
  try {
    result = ((swig::SwigPyIterator const *)arg1)->distance((swig::SwigPyIterator const &)*arg2);
  } catch(std::invalid_argument &_e) {
    SWIG_Python_Raise(SWIG_NewPointerObj((new std::invalid_argument(static_cast< const std::invalid_argument& >(_e))),SWIGTYPE_p_std__invalid_argument,SWIG_POINTER_OWN), "std::invalid_argument", SWIGTYPE_p_std__invalid_argument); SWIG_fail;
  }
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_equal(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  swig::SwigPyIterator *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator_equal", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_equal" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator_equal" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SwigPyIterator_equal" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2);
  try {
    result = (bool)((swig::SwigPyIterator const *)arg1)->equal((swig::SwigPyIterator const &)*arg2);
  } catch(std::invalid_argument &_e) {
    SWIG_Python_Raise(SWIG_NewPointerObj((new std::invalid_argument(static_cast< const std::invalid_argument& >(_e))),SWIGTYPE_p_std__invalid_argument,SWIG_POINTER_OWN), "std::invalid_argument", SWIGTYPE_p_std__invalid_argument); SWIG_fail;
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_copy(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_copy" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  result = (swig::SwigPyIterator *)((swig::SwigPyIterator const *)arg1)->copy();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_next(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_next" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  try {
    result = (PyObject *)(arg1)->next();
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___next__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___next__" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  try {
    result = (PyObject *)(arg1)->__next__();
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_previous(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_previous" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  try {
    result = (PyObject *)(arg1)->previous();
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_advance(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  ptrdiff_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  swig::SwigPyIterator *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator_advance", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_advance" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator_advance" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    result = (swig::SwigPyIterator *)(arg1)->advance(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___eq__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  swig::SwigPyIterator *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___eq__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator___eq__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SwigPyIterator___eq__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2);
  result = (bool)((swig::SwigPyIterator const *)arg1)->operator ==((swig::SwigPyIterator const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___ne__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  swig::SwigPyIterator *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___ne__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator___ne__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SwigPyIterator___ne__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2);
  result = (bool)((swig::SwigPyIterator const *)arg1)->operator !=((swig::SwigPyIterator const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___iadd__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  ptrdiff_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  swig::SwigPyIterator *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator___iadd__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___iadd__" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator___iadd__" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    result = (swig::SwigPyIterator *) &(arg1)->operator +=(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___isub__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  ptrdiff_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  swig::SwigPyIterator *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator___isub__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___isub__" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator___isub__" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    result = (swig::SwigPyIterator *) &(arg1)->operator -=(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___add__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  ptrdiff_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  swig::SwigPyIterator *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator___add__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___add__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator___add__" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    result = (swig::SwigPyIterator *)((swig::SwigPyIterator const *)arg1)->operator +(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___sub____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  ptrdiff_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  swig::SwigPyIterator *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___sub__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator___sub__" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    result = (swig::SwigPyIterator *)((swig::SwigPyIterator const *)arg1)->operator -(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___sub____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  swig::SwigPyIterator *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  ptrdiff_t result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___sub__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator___sub__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SwigPyIterator___sub__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2);
  result = ((swig::SwigPyIterator const *)arg1)->operator -((swig::SwigPyIterator const &)*arg2);
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___sub__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "SwigPyIterator___sub__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__SwigPyIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_SwigPyIterator___sub____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__SwigPyIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_SwigPyIterator___sub____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *SwigPyIterator_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_swig__SwigPyIterator, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap__stringlist_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_iterator" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  result = (swig::SwigPyIterator *)std_list_Sl_CString_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___nonzero__" "', argument " "1"" of type '" "std::list< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  result = (bool)std_list_Sl_CString_Sg____nonzero__((std::list< CString > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___bool__" "', argument " "1"" of type '" "std::list< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  result = (bool)std_list_Sl_CString_Sg____bool__((std::list< CString > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::list< CString >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___len__" "', argument " "1"" of type '" "std::list< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  result = std_list_Sl_CString_Sg____len__((std::list< CString > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___getslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::difference_type arg2 ;
  std::list< CString >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  std::list< CString,std::allocator< CString > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "_stringlist___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___getslice__" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_stringlist___getslice__" "', argument " "2"" of type '" "std::list< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::list< CString >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "_stringlist___getslice__" "', argument " "3"" of type '" "std::list< CString >::difference_type""'");
  } 
  arg3 = static_cast< std::list< CString >::difference_type >(val3);
  try {
    result = (std::list< CString,std::allocator< CString > > *)std_list_Sl_CString_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__listT_CString_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::difference_type arg2 ;
  std::list< CString >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___setslice__" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_stringlist___setslice__" "', argument " "2"" of type '" "std::list< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::list< CString >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "_stringlist___setslice__" "', argument " "3"" of type '" "std::list< CString >::difference_type""'");
  } 
  arg3 = static_cast< std::list< CString >::difference_type >(val3);
  try {
    std_list_Sl_CString_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::difference_type arg2 ;
  std::list< CString >::difference_type arg3 ;
  std::list< CString,std::allocator< CString > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___setslice__" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_stringlist___setslice__" "', argument " "2"" of type '" "std::list< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::list< CString >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "_stringlist___setslice__" "', argument " "3"" of type '" "std::list< CString >::difference_type""'");
  } 
  arg3 = static_cast< std::list< CString >::difference_type >(val3);
  {
    std::list< CString,std::allocator< CString > > *ptr = (std::list< CString,std::allocator< CString > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "_stringlist___setslice__" "', argument " "4"" of type '" "std::list< CString,std::allocator< CString > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "_stringlist___setslice__" "', argument " "4"" of type '" "std::list< CString,std::allocator< CString > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_list_Sl_CString_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::list< CString,std::allocator< CString > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "_stringlist___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap__stringlist___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::list< CString,std::allocator< CString > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap__stringlist___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_stringlist___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::list< CString >::__setslice__(std::list< CString >::difference_type,std::list< CString >::difference_type)\n"
    "    std::list< CString >::__setslice__(std::list< CString >::difference_type,std::list< CString >::difference_type,std::list< CString,std::allocator< CString > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap__stringlist___delslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::difference_type arg2 ;
  std::list< CString >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "_stringlist___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___delslice__" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_stringlist___delslice__" "', argument " "2"" of type '" "std::list< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::list< CString >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "_stringlist___delslice__" "', argument " "3"" of type '" "std::list< CString >::difference_type""'");
  } 
  arg3 = static_cast< std::list< CString >::difference_type >(val3);
  try {
    std_list_Sl_CString_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___delitem__" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_stringlist___delitem__" "', argument " "2"" of type '" "std::list< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::list< CString >::difference_type >(val2);
  try {
    std_list_Sl_CString_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::list< CString,std::allocator< CString > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___getitem__" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    result = (std::list< CString,std::allocator< CString > > *)std_list_Sl_CString_Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__listT_CString_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  std::list< CString,std::allocator< CString > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___setitem__" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  {
    std::list< CString,std::allocator< CString > > *ptr = (std::list< CString,std::allocator< CString > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "_stringlist___setitem__" "', argument " "3"" of type '" "std::list< CString,std::allocator< CString > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "_stringlist___setitem__" "', argument " "3"" of type '" "std::list< CString,std::allocator< CString > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_list_Sl_CString_Sg____setitem____SWIG_0(arg1,arg2,(std::list< CString,std::allocator< CString > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___setitem__" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_list_Sl_CString_Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___delitem__" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_list_Sl_CString_Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "_stringlist___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap__stringlist___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap__stringlist___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_stringlist___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::list< CString >::__delitem__(std::list< CString >::difference_type)\n"
    "    std::list< CString >::__delitem__(SWIGPY_SLICEOBJECT *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap__stringlist___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::list< CString >::value_type *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___getitem__" "', argument " "1"" of type '" "std::list< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_stringlist___getitem__" "', argument " "2"" of type '" "std::list< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::list< CString >::difference_type >(val2);
  try {
    result = (std::list< CString >::value_type *) &std_list_Sl_CString_Sg____getitem____SWIG_1((std::list< CString > const *)arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  (void)swig::container_owner<swig::traits<std::list< CString >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "_stringlist___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap__stringlist___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap__stringlist___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_stringlist___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::list< CString >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::list< CString >::__getitem__(std::list< CString >::difference_type) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap__stringlist___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::difference_type arg2 ;
  std::list< CString >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist___setitem__" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_stringlist___setitem__" "', argument " "2"" of type '" "std::list< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::list< CString >::difference_type >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "_stringlist___setitem__" "', argument " "3"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "_stringlist___setitem__" "', argument " "3"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_list_Sl_CString_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(CString const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "_stringlist___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap__stringlist___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::list< CString,std::allocator< CString > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap__stringlist___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap__stringlist___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_stringlist___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::list< CString >::__setitem__(SWIGPY_SLICEOBJECT *,std::list< CString,std::allocator< CString > > const &)\n"
    "    std::list< CString >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::list< CString >::__setitem__(std::list< CString >::difference_type,std::list< CString >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap__stringlist_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::list< CString >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_pop" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  try {
    result = std_list_Sl_CString_Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "_stringlist_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_append" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "_stringlist_append" "', argument " "2"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "_stringlist_append" "', argument " "2"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  std_list_Sl_CString_Sg__append(arg1,(CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new__stringlist__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::list< CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::list< CString > *)new std::list< CString >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__listT_CString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new__stringlist__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::list< CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::list< CString,std::allocator< CString > > *ptr = (std::list< CString,std::allocator< CString > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__stringlist" "', argument " "1"" of type '" "std::list< CString > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new__stringlist" "', argument " "1"" of type '" "std::list< CString > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::list< CString > *)new std::list< CString >((std::list< CString > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__listT_CString_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_empty" "', argument " "1"" of type '" "std::list< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  result = (bool)((std::list< CString > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::list< CString >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_size" "', argument " "1"" of type '" "std::list< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  result = ((std::list< CString > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "_stringlist_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_swap" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__listT_CString_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "_stringlist_swap" "', argument " "2"" of type '" "std::list< CString > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "_stringlist_swap" "', argument " "2"" of type '" "std::list< CString > &""'"); 
  }
  arg2 = reinterpret_cast< std::list< CString > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::list< CString >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_begin" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::list< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::list< CString >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_end" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::list< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::list< CString >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_rbegin" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::list< CString >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::list< CString >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_rend" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::list< CString >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_clear" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< CString > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_get_allocator" "', argument " "1"" of type '" "std::list< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  result = ((std::list< CString > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::list< CString >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_CString_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new__stringlist__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::list< CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new__stringlist" "', argument " "1"" of type '" "std::list< CString >::size_type""'");
  } 
  arg1 = static_cast< std::list< CString >::size_type >(val1);
  result = (std::list< CString > *)new std::list< CString >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__listT_CString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_pop_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_pop_back" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_resize" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_stringlist_resize" "', argument " "2"" of type '" "std::list< CString >::size_type""'");
  } 
  arg2 = static_cast< std::list< CString >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::list< CString >::iterator result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_erase" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist_erase" "', argument " "2"" of type '" "std::list< CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::list< CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::list< CString >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist_erase" "', argument " "2"" of type '" "std::list< CString >::iterator""'");
    }
  }
  result = std_list_Sl_CString_Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::list< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::iterator arg2 ;
  std::list< CString >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::list< CString >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_erase" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist_erase" "', argument " "2"" of type '" "std::list< CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::list< CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::list< CString >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist_erase" "', argument " "2"" of type '" "std::list< CString >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist_erase" "', argument " "3"" of type '" "std::list< CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::list< CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::list< CString >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist_erase" "', argument " "3"" of type '" "std::list< CString >::iterator""'");
    }
  }
  result = std_list_Sl_CString_Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::list< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "_stringlist_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::list< CString >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap__stringlist_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::list< CString >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::list< CString >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap__stringlist_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_stringlist_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::list< CString >::erase(std::list< CString >::iterator)\n"
    "    std::list< CString >::erase(std::list< CString >::iterator,std::list< CString >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new__stringlist__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString >::size_type arg1 ;
  std::list< CString >::value_type *arg2 = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  std::list< CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new__stringlist" "', argument " "1"" of type '" "std::list< CString >::size_type""'");
  } 
  arg1 = static_cast< std::list< CString >::size_type >(val1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new__stringlist" "', argument " "2"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new__stringlist" "', argument " "2"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (std::list< CString > *)new std::list< CString >(arg1,(std::list< CString >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__listT_CString_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new__stringlist(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new__stringlist", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new__stringlist__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new__stringlist__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new__stringlist__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new__stringlist__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new__stringlist'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::list< CString >::list()\n"
    "    std::list< CString >::list(std::list< CString > const &)\n"
    "    std::list< CString >::list(std::list< CString >::size_type)\n"
    "    std::list< CString >::list(std::list< CString >::size_type,std::list< CString >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap__stringlist_push_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "_stringlist_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_push_back" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "_stringlist_push_back" "', argument " "2"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "_stringlist_push_back" "', argument " "2"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->push_back((std::list< CString >::value_type const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::list< CString >::value_type *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_front" "', argument " "1"" of type '" "std::list< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  result = (std::list< CString >::value_type *) &((std::list< CString > const *)arg1)->front();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  (void)swig::container_owner<swig::traits<std::list< CString >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::list< CString >::value_type *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_back" "', argument " "1"" of type '" "std::list< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  result = (std::list< CString >::value_type *) &((std::list< CString > const *)arg1)->back();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  (void)swig::container_owner<swig::traits<std::list< CString >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_assign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::size_type arg2 ;
  std::list< CString >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "_stringlist_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_assign" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_stringlist_assign" "', argument " "2"" of type '" "std::list< CString >::size_type""'");
  } 
  arg2 = static_cast< std::list< CString >::size_type >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "_stringlist_assign" "', argument " "3"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "_stringlist_assign" "', argument " "3"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->assign(arg2,(std::list< CString >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::size_type arg2 ;
  std::list< CString >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_resize" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_stringlist_resize" "', argument " "2"" of type '" "std::list< CString >::size_type""'");
  } 
  arg2 = static_cast< std::list< CString >::size_type >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "_stringlist_resize" "', argument " "3"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "_stringlist_resize" "', argument " "3"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->resize(arg2,(std::list< CString >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "_stringlist_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap__stringlist_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap__stringlist_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_stringlist_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::list< CString >::resize(std::list< CString >::size_type)\n"
    "    std::list< CString >::resize(std::list< CString >::size_type,std::list< CString >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap__stringlist_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::iterator arg2 ;
  std::list< CString >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  int res3 = SWIG_OLDOBJ ;
  std::list< CString >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_insert" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist_insert" "', argument " "2"" of type '" "std::list< CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::list< CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::list< CString >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist_insert" "', argument " "2"" of type '" "std::list< CString >::iterator""'");
    }
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "_stringlist_insert" "', argument " "3"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "_stringlist_insert" "', argument " "3"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  result = std_list_Sl_CString_Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),(CString const &)*arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::list< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::iterator arg2 ;
  std::list< CString >::size_type arg3 ;
  std::list< CString >::value_type *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_insert" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist_insert" "', argument " "2"" of type '" "std::list< CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::list< CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::list< CString >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "_stringlist_insert" "', argument " "2"" of type '" "std::list< CString >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "_stringlist_insert" "', argument " "3"" of type '" "std::list< CString >::size_type""'");
  } 
  arg3 = static_cast< std::list< CString >::size_type >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "_stringlist_insert" "', argument " "4"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "_stringlist_insert" "', argument " "4"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    arg4 = ptr;
  }
  std_list_Sl_CString_Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "_stringlist_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::list< CString >::iterator > *>(iter) != 0));
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap__stringlist_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::list< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::list< CString >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap__stringlist_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_stringlist_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::list< CString >::insert(std::list< CString >::iterator,std::list< CString >::value_type const &)\n"
    "    std::list< CString >::insert(std::list< CString >::iterator,std::list< CString >::size_type,std::list< CString >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap__stringlist_pop_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_pop_front" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  (arg1)->pop_front();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_push_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  std::list< CString >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "_stringlist_push_front", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_push_front" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "_stringlist_push_front" "', argument " "2"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "_stringlist_push_front" "', argument " "2"" of type '" "std::list< CString >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->push_front((std::list< CString >::value_type const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap__stringlist_reverse(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_stringlist_reverse" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  (arg1)->reverse();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete__stringlist(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::list< CString > *arg1 = (std::list< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__listT_CString_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__stringlist" "', argument " "1"" of type '" "std::list< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::list< CString > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_stringlist_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__listT_CString_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_stringlist_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_VIRCNetworks_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_iterator" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_CIRCNetwork_Sm__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___nonzero__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = (bool)std_vector_Sl_CIRCNetwork_Sm__Sg____nonzero__((std::vector< CIRCNetwork * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___bool__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = (bool)std_vector_Sl_CIRCNetwork_Sm__Sg____bool__((std::vector< CIRCNetwork * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CIRCNetwork * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___len__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = std_vector_Sl_CIRCNetwork_Sm__Sg____len__((std::vector< CIRCNetwork * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___getslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::difference_type arg2 ;
  std::vector< CIRCNetwork * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VIRCNetworks___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___getslice__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VIRCNetworks___getslice__" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CIRCNetwork * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VIRCNetworks___getslice__" "', argument " "3"" of type '" "std::vector< CIRCNetwork * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CIRCNetwork * >::difference_type >(val3);
  try {
    result = (std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > *)std_vector_Sl_CIRCNetwork_Sm__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::difference_type arg2 ;
  std::vector< CIRCNetwork * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___setslice__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VIRCNetworks___setslice__" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CIRCNetwork * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VIRCNetworks___setslice__" "', argument " "3"" of type '" "std::vector< CIRCNetwork * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CIRCNetwork * >::difference_type >(val3);
  try {
    std_vector_Sl_CIRCNetwork_Sm__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::difference_type arg2 ;
  std::vector< CIRCNetwork * >::difference_type arg3 ;
  std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___setslice__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VIRCNetworks___setslice__" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CIRCNetwork * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VIRCNetworks___setslice__" "', argument " "3"" of type '" "std::vector< CIRCNetwork * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CIRCNetwork * >::difference_type >(val3);
  {
    std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > > *ptr = (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VIRCNetworks___setslice__" "', argument " "4"" of type '" "std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VIRCNetworks___setslice__" "', argument " "4"" of type '" "std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_CIRCNetwork_Sm__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VIRCNetworks___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_VIRCNetworks___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VIRCNetworks___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VIRCNetworks___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CIRCNetwork * >::__setslice__(std::vector< CIRCNetwork * >::difference_type,std::vector< CIRCNetwork * >::difference_type)\n"
    "    std::vector< CIRCNetwork * >::__setslice__(std::vector< CIRCNetwork * >::difference_type,std::vector< CIRCNetwork * >::difference_type,std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___delslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::difference_type arg2 ;
  std::vector< CIRCNetwork * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VIRCNetworks___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___delslice__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VIRCNetworks___delslice__" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CIRCNetwork * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VIRCNetworks___delslice__" "', argument " "3"" of type '" "std::vector< CIRCNetwork * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CIRCNetwork * >::difference_type >(val3);
  try {
    std_vector_Sl_CIRCNetwork_Sm__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___delitem__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VIRCNetworks___delitem__" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CIRCNetwork * >::difference_type >(val2);
  try {
    std_vector_Sl_CIRCNetwork_Sm__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___getitem__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    result = (std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > *)std_vector_Sl_CIRCNetwork_Sm__Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___setitem__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  {
    std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > > *ptr = (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VIRCNetworks___setitem__" "', argument " "3"" of type '" "std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VIRCNetworks___setitem__" "', argument " "3"" of type '" "std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_CIRCNetwork_Sm__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___setitem__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CIRCNetwork_Sm__Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___delitem__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CIRCNetwork_Sm__Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VIRCNetworks___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VIRCNetworks___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VIRCNetworks___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VIRCNetworks___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CIRCNetwork * >::__delitem__(std::vector< CIRCNetwork * >::difference_type)\n"
    "    std::vector< CIRCNetwork * >::__delitem__(SWIGPY_SLICEOBJECT *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< CIRCNetwork * >::value_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___getitem__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VIRCNetworks___getitem__" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CIRCNetwork * >::difference_type >(val2);
  try {
    result = (std::vector< CIRCNetwork * >::value_type)std_vector_Sl_CIRCNetwork_Sm__Sg____getitem____SWIG_1(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VIRCNetworks___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VIRCNetworks___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VIRCNetworks___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VIRCNetworks___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CIRCNetwork * >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CIRCNetwork * >::__getitem__(std::vector< CIRCNetwork * >::difference_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::difference_type arg2 ;
  std::vector< CIRCNetwork * >::value_type arg3 = (std::vector< CIRCNetwork * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks___setitem__" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VIRCNetworks___setitem__" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CIRCNetwork * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VIRCNetworks___setitem__" "', argument " "3"" of type '" "std::vector< CIRCNetwork * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CIRCNetwork * >::value_type >(argp3);
  try {
    std_vector_Sl_CIRCNetwork_Sm__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VIRCNetworks___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VIRCNetworks___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VIRCNetworks___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VIRCNetworks___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VIRCNetworks___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CIRCNetwork * >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > const &)\n"
    "    std::vector< CIRCNetwork * >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CIRCNetwork * >::__setitem__(std::vector< CIRCNetwork * >::difference_type,std::vector< CIRCNetwork * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CIRCNetwork * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_pop" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  try {
    result = (std::vector< CIRCNetwork * >::value_type)std_vector_Sl_CIRCNetwork_Sm__Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::value_type arg2 = (std::vector< CIRCNetwork * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VIRCNetworks_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_append" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VIRCNetworks_append" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CIRCNetwork * >::value_type >(argp2);
  std_vector_Sl_CIRCNetwork_Sm__Sg__append(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VIRCNetworks__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< CIRCNetwork * > *)new std::vector< CIRCNetwork * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VIRCNetworks__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< CIRCNetwork * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > > *ptr = (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_VIRCNetworks" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VIRCNetworks" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< CIRCNetwork * > *)new std::vector< CIRCNetwork * >((std::vector< CIRCNetwork * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_empty" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = (bool)((std::vector< CIRCNetwork * > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CIRCNetwork * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_size" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = ((std::vector< CIRCNetwork * > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VIRCNetworks_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_swap" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VIRCNetworks_swap" "', argument " "2"" of type '" "std::vector< CIRCNetwork * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VIRCNetworks_swap" "', argument " "2"" of type '" "std::vector< CIRCNetwork * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CIRCNetwork * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_begin" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CIRCNetwork * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CIRCNetwork * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_end" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CIRCNetwork * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CIRCNetwork * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_rbegin" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CIRCNetwork * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CIRCNetwork * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_rend" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CIRCNetwork * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_clear" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< CIRCNetwork * > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_get_allocator" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = ((std::vector< CIRCNetwork * > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< CIRCNetwork * >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_CIRCNetwork_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VIRCNetworks__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< CIRCNetwork * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VIRCNetworks" "', argument " "1"" of type '" "std::vector< CIRCNetwork * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CIRCNetwork * >::size_type >(val1);
  result = (std::vector< CIRCNetwork * > *)new std::vector< CIRCNetwork * >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_pop_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_pop_back" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_resize" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VIRCNetworks_resize" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CIRCNetwork * >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< CIRCNetwork * >::iterator result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_erase" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks_erase" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks_erase" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::iterator""'");
    }
  }
  result = std_vector_Sl_CIRCNetwork_Sm__Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CIRCNetwork * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::iterator arg2 ;
  std::vector< CIRCNetwork * >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::vector< CIRCNetwork * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_erase" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks_erase" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks_erase" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks_erase" "', argument " "3"" of type '" "std::vector< CIRCNetwork * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks_erase" "', argument " "3"" of type '" "std::vector< CIRCNetwork * >::iterator""'");
    }
  }
  result = std_vector_Sl_CIRCNetwork_Sm__Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CIRCNetwork * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VIRCNetworks_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_VIRCNetworks_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_VIRCNetworks_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VIRCNetworks_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CIRCNetwork * >::erase(std::vector< CIRCNetwork * >::iterator)\n"
    "    std::vector< CIRCNetwork * >::erase(std::vector< CIRCNetwork * >::iterator,std::vector< CIRCNetwork * >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_VIRCNetworks__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * >::size_type arg1 ;
  std::vector< CIRCNetwork * >::value_type arg2 = (std::vector< CIRCNetwork * >::value_type) 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::vector< CIRCNetwork * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VIRCNetworks" "', argument " "1"" of type '" "std::vector< CIRCNetwork * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CIRCNetwork * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_VIRCNetworks" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CIRCNetwork * >::value_type >(argp2);
  result = (std::vector< CIRCNetwork * > *)new std::vector< CIRCNetwork * >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VIRCNetworks(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_VIRCNetworks", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_VIRCNetworks__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_VIRCNetworks__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_VIRCNetworks__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_VIRCNetworks__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_VIRCNetworks'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CIRCNetwork * >::vector()\n"
    "    std::vector< CIRCNetwork * >::vector(std::vector< CIRCNetwork * > const &)\n"
    "    std::vector< CIRCNetwork * >::vector(std::vector< CIRCNetwork * >::size_type)\n"
    "    std::vector< CIRCNetwork * >::vector(std::vector< CIRCNetwork * >::size_type,std::vector< CIRCNetwork * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_push_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::value_type arg2 = (std::vector< CIRCNetwork * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VIRCNetworks_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_push_back" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VIRCNetworks_push_back" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CIRCNetwork * >::value_type >(argp2);
  (arg1)->push_back(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CIRCNetwork * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_front" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = (std::vector< CIRCNetwork * >::value_type)((std::vector< CIRCNetwork * > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CIRCNetwork * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_back" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = (std::vector< CIRCNetwork * >::value_type)((std::vector< CIRCNetwork * > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_assign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::size_type arg2 ;
  std::vector< CIRCNetwork * >::value_type arg3 = (std::vector< CIRCNetwork * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VIRCNetworks_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_assign" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VIRCNetworks_assign" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CIRCNetwork * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VIRCNetworks_assign" "', argument " "3"" of type '" "std::vector< CIRCNetwork * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CIRCNetwork * >::value_type >(argp3);
  (arg1)->assign(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::size_type arg2 ;
  std::vector< CIRCNetwork * >::value_type arg3 = (std::vector< CIRCNetwork * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_resize" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VIRCNetworks_resize" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CIRCNetwork * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VIRCNetworks_resize" "', argument " "3"" of type '" "std::vector< CIRCNetwork * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CIRCNetwork * >::value_type >(argp3);
  (arg1)->resize(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VIRCNetworks_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VIRCNetworks_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VIRCNetworks_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VIRCNetworks_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CIRCNetwork * >::resize(std::vector< CIRCNetwork * >::size_type)\n"
    "    std::vector< CIRCNetwork * >::resize(std::vector< CIRCNetwork * >::size_type,std::vector< CIRCNetwork * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::iterator arg2 ;
  std::vector< CIRCNetwork * >::value_type arg3 = (std::vector< CIRCNetwork * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  std::vector< CIRCNetwork * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_insert" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks_insert" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks_insert" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VIRCNetworks_insert" "', argument " "3"" of type '" "std::vector< CIRCNetwork * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CIRCNetwork * >::value_type >(argp3);
  result = std_vector_Sl_CIRCNetwork_Sm__Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CIRCNetwork * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::iterator arg2 ;
  std::vector< CIRCNetwork * >::size_type arg3 ;
  std::vector< CIRCNetwork * >::value_type arg4 = (std::vector< CIRCNetwork * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_insert" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks_insert" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VIRCNetworks_insert" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VIRCNetworks_insert" "', argument " "3"" of type '" "std::vector< CIRCNetwork * >::size_type""'");
  } 
  arg3 = static_cast< std::vector< CIRCNetwork * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VIRCNetworks_insert" "', argument " "4"" of type '" "std::vector< CIRCNetwork * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector< CIRCNetwork * >::value_type >(argp4);
  std_vector_Sl_CIRCNetwork_Sm__Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VIRCNetworks_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VIRCNetworks_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CIRCNetwork * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VIRCNetworks_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VIRCNetworks_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CIRCNetwork * >::insert(std::vector< CIRCNetwork * >::iterator,std::vector< CIRCNetwork * >::value_type)\n"
    "    std::vector< CIRCNetwork * >::insert(std::vector< CIRCNetwork * >::iterator,std::vector< CIRCNetwork * >::size_type,std::vector< CIRCNetwork * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_reserve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  std::vector< CIRCNetwork * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VIRCNetworks_reserve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_reserve" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VIRCNetworks_reserve" "', argument " "2"" of type '" "std::vector< CIRCNetwork * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CIRCNetwork * >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VIRCNetworks_capacity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CIRCNetwork * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VIRCNetworks_capacity" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  result = ((std::vector< CIRCNetwork * > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_VIRCNetworks(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CIRCNetwork * > *arg1 = (std::vector< CIRCNetwork * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VIRCNetworks" "', argument " "1"" of type '" "std::vector< CIRCNetwork * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CIRCNetwork * > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *VIRCNetworks_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_CIRCNetwork_p_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *VIRCNetworks_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_VChannels_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_iterator" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_CChan_Sm__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___nonzero__" "', argument " "1"" of type '" "std::vector< CChan * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = (bool)std_vector_Sl_CChan_Sm__Sg____nonzero__((std::vector< CChan * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___bool__" "', argument " "1"" of type '" "std::vector< CChan * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = (bool)std_vector_Sl_CChan_Sm__Sg____bool__((std::vector< CChan * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CChan * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___len__" "', argument " "1"" of type '" "std::vector< CChan * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = std_vector_Sl_CChan_Sm__Sg____len__((std::vector< CChan * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___getslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::difference_type arg2 ;
  std::vector< CChan * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  std::vector< CChan *,std::allocator< CChan * > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VChannels___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___getslice__" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VChannels___getslice__" "', argument " "2"" of type '" "std::vector< CChan * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CChan * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VChannels___getslice__" "', argument " "3"" of type '" "std::vector< CChan * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CChan * >::difference_type >(val3);
  try {
    result = (std::vector< CChan *,std::allocator< CChan * > > *)std_vector_Sl_CChan_Sm__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CChan_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::difference_type arg2 ;
  std::vector< CChan * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___setslice__" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VChannels___setslice__" "', argument " "2"" of type '" "std::vector< CChan * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CChan * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VChannels___setslice__" "', argument " "3"" of type '" "std::vector< CChan * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CChan * >::difference_type >(val3);
  try {
    std_vector_Sl_CChan_Sm__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::difference_type arg2 ;
  std::vector< CChan * >::difference_type arg3 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___setslice__" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VChannels___setslice__" "', argument " "2"" of type '" "std::vector< CChan * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CChan * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VChannels___setslice__" "', argument " "3"" of type '" "std::vector< CChan * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CChan * >::difference_type >(val3);
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VChannels___setslice__" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VChannels___setslice__" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_CChan_Sm__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< CChan *,std::allocator< CChan * > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VChannels___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_VChannels___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VChannels___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VChannels___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CChan * >::__setslice__(std::vector< CChan * >::difference_type,std::vector< CChan * >::difference_type)\n"
    "    std::vector< CChan * >::__setslice__(std::vector< CChan * >::difference_type,std::vector< CChan * >::difference_type,std::vector< CChan *,std::allocator< CChan * > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VChannels___delslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::difference_type arg2 ;
  std::vector< CChan * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VChannels___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___delslice__" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VChannels___delslice__" "', argument " "2"" of type '" "std::vector< CChan * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CChan * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VChannels___delslice__" "', argument " "3"" of type '" "std::vector< CChan * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CChan * >::difference_type >(val3);
  try {
    std_vector_Sl_CChan_Sm__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___delitem__" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VChannels___delitem__" "', argument " "2"" of type '" "std::vector< CChan * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CChan * >::difference_type >(val2);
  try {
    std_vector_Sl_CChan_Sm__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___getitem__" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    result = (std::vector< CChan *,std::allocator< CChan * > > *)std_vector_Sl_CChan_Sm__Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CChan_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___setitem__" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VChannels___setitem__" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VChannels___setitem__" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_CChan_Sm__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< CChan *,std::allocator< CChan * > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___setitem__" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CChan_Sm__Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___delitem__" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CChan_Sm__Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VChannels___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VChannels___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VChannels___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VChannels___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CChan * >::__delitem__(std::vector< CChan * >::difference_type)\n"
    "    std::vector< CChan * >::__delitem__(SWIGPY_SLICEOBJECT *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VChannels___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< CChan * >::value_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___getitem__" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VChannels___getitem__" "', argument " "2"" of type '" "std::vector< CChan * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CChan * >::difference_type >(val2);
  try {
    result = (std::vector< CChan * >::value_type)std_vector_Sl_CChan_Sm__Sg____getitem____SWIG_1(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CChan, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VChannels___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VChannels___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VChannels___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VChannels___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CChan * >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CChan * >::__getitem__(std::vector< CChan * >::difference_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VChannels___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::difference_type arg2 ;
  std::vector< CChan * >::value_type arg3 = (std::vector< CChan * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels___setitem__" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VChannels___setitem__" "', argument " "2"" of type '" "std::vector< CChan * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CChan * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VChannels___setitem__" "', argument " "3"" of type '" "std::vector< CChan * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CChan * >::value_type >(argp3);
  try {
    std_vector_Sl_CChan_Sm__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VChannels___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VChannels___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VChannels___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CChan, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VChannels___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VChannels___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CChan * >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< CChan *,std::allocator< CChan * > > const &)\n"
    "    std::vector< CChan * >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CChan * >::__setitem__(std::vector< CChan * >::difference_type,std::vector< CChan * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VChannels_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CChan * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_pop" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  try {
    result = (std::vector< CChan * >::value_type)std_vector_Sl_CChan_Sm__Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CChan, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::value_type arg2 = (std::vector< CChan * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VChannels_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_append" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VChannels_append" "', argument " "2"" of type '" "std::vector< CChan * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CChan * >::value_type >(argp2);
  std_vector_Sl_CChan_Sm__Sg__append(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VChannels__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< CChan * > *)new std::vector< CChan * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CChan_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VChannels__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< CChan * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_VChannels" "', argument " "1"" of type '" "std::vector< CChan * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VChannels" "', argument " "1"" of type '" "std::vector< CChan * > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< CChan * > *)new std::vector< CChan * >((std::vector< CChan * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CChan_p_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_empty" "', argument " "1"" of type '" "std::vector< CChan * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = (bool)((std::vector< CChan * > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CChan * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_size" "', argument " "1"" of type '" "std::vector< CChan * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = ((std::vector< CChan * > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VChannels_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_swap" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_CChan_p_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VChannels_swap" "', argument " "2"" of type '" "std::vector< CChan * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VChannels_swap" "', argument " "2"" of type '" "std::vector< CChan * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CChan * > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CChan * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_begin" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CChan * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CChan * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_end" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CChan * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CChan * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_rbegin" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CChan * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CChan * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_rend" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CChan * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_clear" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< CChan * > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_get_allocator" "', argument " "1"" of type '" "std::vector< CChan * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = ((std::vector< CChan * > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< CChan * >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_CChan_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VChannels__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< CChan * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VChannels" "', argument " "1"" of type '" "std::vector< CChan * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CChan * >::size_type >(val1);
  result = (std::vector< CChan * > *)new std::vector< CChan * >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CChan_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_pop_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_pop_back" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_resize" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VChannels_resize" "', argument " "2"" of type '" "std::vector< CChan * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CChan * >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< CChan * >::iterator result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_erase" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels_erase" "', argument " "2"" of type '" "std::vector< CChan * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels_erase" "', argument " "2"" of type '" "std::vector< CChan * >::iterator""'");
    }
  }
  result = std_vector_Sl_CChan_Sm__Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CChan * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::iterator arg2 ;
  std::vector< CChan * >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::vector< CChan * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_erase" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels_erase" "', argument " "2"" of type '" "std::vector< CChan * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels_erase" "', argument " "2"" of type '" "std::vector< CChan * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels_erase" "', argument " "3"" of type '" "std::vector< CChan * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels_erase" "', argument " "3"" of type '" "std::vector< CChan * >::iterator""'");
    }
  }
  result = std_vector_Sl_CChan_Sm__Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CChan * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VChannels_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_VChannels_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_VChannels_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VChannels_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CChan * >::erase(std::vector< CChan * >::iterator)\n"
    "    std::vector< CChan * >::erase(std::vector< CChan * >::iterator,std::vector< CChan * >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_VChannels__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * >::size_type arg1 ;
  std::vector< CChan * >::value_type arg2 = (std::vector< CChan * >::value_type) 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::vector< CChan * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VChannels" "', argument " "1"" of type '" "std::vector< CChan * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CChan * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_VChannels" "', argument " "2"" of type '" "std::vector< CChan * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CChan * >::value_type >(argp2);
  result = (std::vector< CChan * > *)new std::vector< CChan * >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CChan_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VChannels(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_VChannels", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_VChannels__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_VChannels__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_VChannels__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CChan, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_VChannels__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_VChannels'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CChan * >::vector()\n"
    "    std::vector< CChan * >::vector(std::vector< CChan * > const &)\n"
    "    std::vector< CChan * >::vector(std::vector< CChan * >::size_type)\n"
    "    std::vector< CChan * >::vector(std::vector< CChan * >::size_type,std::vector< CChan * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VChannels_push_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::value_type arg2 = (std::vector< CChan * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VChannels_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_push_back" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VChannels_push_back" "', argument " "2"" of type '" "std::vector< CChan * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CChan * >::value_type >(argp2);
  (arg1)->push_back(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CChan * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_front" "', argument " "1"" of type '" "std::vector< CChan * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = (std::vector< CChan * >::value_type)((std::vector< CChan * > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CChan, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CChan * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_back" "', argument " "1"" of type '" "std::vector< CChan * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = (std::vector< CChan * >::value_type)((std::vector< CChan * > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CChan, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_assign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::size_type arg2 ;
  std::vector< CChan * >::value_type arg3 = (std::vector< CChan * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VChannels_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_assign" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VChannels_assign" "', argument " "2"" of type '" "std::vector< CChan * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CChan * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VChannels_assign" "', argument " "3"" of type '" "std::vector< CChan * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CChan * >::value_type >(argp3);
  (arg1)->assign(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::size_type arg2 ;
  std::vector< CChan * >::value_type arg3 = (std::vector< CChan * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_resize" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VChannels_resize" "', argument " "2"" of type '" "std::vector< CChan * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CChan * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VChannels_resize" "', argument " "3"" of type '" "std::vector< CChan * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CChan * >::value_type >(argp3);
  (arg1)->resize(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VChannels_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VChannels_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CChan, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VChannels_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VChannels_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CChan * >::resize(std::vector< CChan * >::size_type)\n"
    "    std::vector< CChan * >::resize(std::vector< CChan * >::size_type,std::vector< CChan * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VChannels_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::iterator arg2 ;
  std::vector< CChan * >::value_type arg3 = (std::vector< CChan * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  std::vector< CChan * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_insert" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels_insert" "', argument " "2"" of type '" "std::vector< CChan * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels_insert" "', argument " "2"" of type '" "std::vector< CChan * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VChannels_insert" "', argument " "3"" of type '" "std::vector< CChan * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CChan * >::value_type >(argp3);
  result = std_vector_Sl_CChan_Sm__Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CChan * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::iterator arg2 ;
  std::vector< CChan * >::size_type arg3 ;
  std::vector< CChan * >::value_type arg4 = (std::vector< CChan * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_insert" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels_insert" "', argument " "2"" of type '" "std::vector< CChan * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VChannels_insert" "', argument " "2"" of type '" "std::vector< CChan * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VChannels_insert" "', argument " "3"" of type '" "std::vector< CChan * >::size_type""'");
  } 
  arg3 = static_cast< std::vector< CChan * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VChannels_insert" "', argument " "4"" of type '" "std::vector< CChan * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector< CChan * >::value_type >(argp4);
  std_vector_Sl_CChan_Sm__Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VChannels_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CChan, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VChannels_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CChan*,std::allocator< CChan * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CChan * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CChan, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VChannels_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VChannels_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CChan * >::insert(std::vector< CChan * >::iterator,std::vector< CChan * >::value_type)\n"
    "    std::vector< CChan * >::insert(std::vector< CChan * >::iterator,std::vector< CChan * >::size_type,std::vector< CChan * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VChannels_reserve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  std::vector< CChan * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VChannels_reserve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_reserve" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VChannels_reserve" "', argument " "2"" of type '" "std::vector< CChan * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CChan * >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VChannels_capacity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CChan * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VChannels_capacity" "', argument " "1"" of type '" "std::vector< CChan * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  result = ((std::vector< CChan * > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_VChannels(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CChan * > *arg1 = (std::vector< CChan * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CChan_p_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VChannels" "', argument " "1"" of type '" "std::vector< CChan * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CChan * > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *VChannels_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_CChan_p_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *VChannels_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_MNicks__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::less< CString > *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::map< CString,CNick > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__lessT_CString_t,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_MNicks" "', argument " "1"" of type '" "std::less< CString > const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_MNicks" "', argument " "1"" of type '" "std::less< CString > const &""'"); 
  }
  arg1 = reinterpret_cast< std::less< CString > * >(argp1);
  result = (std::map< CString,CNick > *)new std::map< CString,CNick >((std::less< CString > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_CString_CNick_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_iterator" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (swig::SwigPyIterator *)std_map_Sl_CString_Sc_CNick_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks___nonzero__" "', argument " "1"" of type '" "std::map< CString,CNick > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (bool)std_map_Sl_CString_Sc_CNick_Sg____nonzero__((std::map< CString,CNick > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks___bool__" "', argument " "1"" of type '" "std::map< CString,CNick > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (bool)std_map_Sl_CString_Sc_CNick_Sg____bool__((std::map< CString,CNick > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CNick >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks___len__" "', argument " "1"" of type '" "std::map< CString,CNick > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = std_map_Sl_CString_Sc_CNick_Sg____len__((std::map< CString,CNick > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks___getitem__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::map< CString,CNick >::mapped_type *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "MNicks___getitem__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks___getitem__" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MNicks___getitem__" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MNicks___getitem__" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  try {
    result = (std::map< CString,CNick >::mapped_type *) &std_map_Sl_CString_Sc_CNick_Sg____getitem__(arg1,(CString const &)*arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNick, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks___delitem__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "MNicks___delitem__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks___delitem__" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MNicks___delitem__" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MNicks___delitem__" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  try {
    std_map_Sl_CString_Sc_CNick_Sg____delitem__(arg1,(CString const &)*arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_has_key(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "MNicks_has_key", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_has_key" "', argument " "1"" of type '" "std::map< CString,CNick > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MNicks_has_key" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MNicks_has_key" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)std_map_Sl_CString_Sc_CNick_Sg__has_key((std::map< CString,CNick > const *)arg1,(CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_keys(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_keys" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (PyObject *)std_map_Sl_CString_Sc_CNick_Sg__keys(arg1);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_values(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_values" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (PyObject *)std_map_Sl_CString_Sc_CNick_Sg__values(arg1);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_items(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_items" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (PyObject *)std_map_Sl_CString_Sc_CNick_Sg__items(arg1);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks___contains__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "MNicks___contains__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks___contains__" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MNicks___contains__" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MNicks___contains__" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)std_map_Sl_CString_Sc_CNick_Sg____contains__(arg1,(CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_key_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_key_iterator" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (swig::SwigPyIterator *)std_map_Sl_CString_Sc_CNick_Sg__key_iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_value_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_value_iterator" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (swig::SwigPyIterator *)std_map_Sl_CString_Sc_CNick_Sg__value_iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks___setitem__" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MNicks___setitem__" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MNicks___setitem__" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  std_map_Sl_CString_Sc_CNick_Sg____setitem____SWIG_0(arg1,(CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick >::key_type *arg2 = 0 ;
  std::map< CString,CNick >::mapped_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks___setitem__" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MNicks___setitem__" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MNicks___setitem__" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "MNicks___setitem__" "', argument " "3"" of type '" "std::map< CString,CNick >::mapped_type const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MNicks___setitem__" "', argument " "3"" of type '" "std::map< CString,CNick >::mapped_type const &""'"); 
  }
  arg3 = reinterpret_cast< std::map< CString,CNick >::mapped_type * >(argp3);
  try {
    std_map_Sl_CString_Sc_CNick_Sg____setitem____SWIG_1(arg1,(CString const &)*arg2,(CNick const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "MNicks___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_MNicks___setitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CNick, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_MNicks___setitem____SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'MNicks___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::map< CString,CNick >::__setitem__(std::map< CString,CNick >::key_type const &)\n"
    "    std::map< CString,CNick >::__setitem__(std::map< CString,CNick >::key_type const &,std::map< CString,CNick >::mapped_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_MNicks_asdict(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_asdict" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (PyObject *)std_map_Sl_CString_Sc_CNick_Sg__asdict(arg1);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_MNicks__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::map< CString,CNick > *)new std::map< CString,CNick >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_CString_CNick_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_MNicks__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::map< CString,CNick > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > > *ptr = (std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_MNicks" "', argument " "1"" of type '" "std::map< CString,CNick > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_MNicks" "', argument " "1"" of type '" "std::map< CString,CNick > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::map< CString,CNick > *)new std::map< CString,CNick >((std::map< CString,CNick > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_CString_CNick_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_MNicks(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_MNicks", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_MNicks__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_std__lessT_CString_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_MNicks__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_MNicks__SWIG_2(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_MNicks'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::map< CString,CNick >::map(std::less< CString > const &)\n"
    "    std::map< CString,CNick >::map()\n"
    "    std::map< CString,CNick >::map(std::map< CString,CNick > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_MNicks_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_empty" "', argument " "1"" of type '" "std::map< CString,CNick > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (bool)((std::map< CString,CNick > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CNick >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_size" "', argument " "1"" of type '" "std::map< CString,CNick > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = ((std::map< CString,CNick > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "MNicks_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_swap" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__mapT_CString_CNick_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MNicks_swap" "', argument " "2"" of type '" "std::map< CString,CNick > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MNicks_swap" "', argument " "2"" of type '" "std::map< CString,CNick > &""'"); 
  }
  arg2 = reinterpret_cast< std::map< CString,CNick > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CNick >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_begin" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CNick >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CNick >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_end" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CNick >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CNick >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_rbegin" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CNick >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CNick >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_rend" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CNick >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_clear" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< std::pair< CString const,CNick > > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_get_allocator" "', argument " "1"" of type '" "std::map< CString,CNick > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  result = ((std::map< CString,CNick > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::map< CString,CNick >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_std__pairT_CString_const_CNick_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  std::map< CString,CNick >::size_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_erase" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MNicks_erase" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MNicks_erase" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->erase((std::map< CString,CNick >::key_type const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_count(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::map< CString,CNick >::size_type result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "MNicks_count", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_count" "', argument " "1"" of type '" "std::map< CString,CNick > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MNicks_count" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MNicks_count" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((std::map< CString,CNick > const *)arg1)->count((std::map< CString,CNick >::key_type const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_erase" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "MNicks_erase" "', argument " "2"" of type '" "std::map< CString,CNick >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::map< CString,CNick >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::map< CString,CNick >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "MNicks_erase" "', argument " "2"" of type '" "std::map< CString,CNick >::iterator""'");
    }
  }
  std_map_Sl_CString_Sc_CNick_Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_erase__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick >::iterator arg2 ;
  std::map< CString,CNick >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_erase" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "MNicks_erase" "', argument " "2"" of type '" "std::map< CString,CNick >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::map< CString,CNick >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::map< CString,CNick >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "MNicks_erase" "', argument " "2"" of type '" "std::map< CString,CNick >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "MNicks_erase" "', argument " "3"" of type '" "std::map< CString,CNick >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::map< CString,CNick >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::map< CString,CNick >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "MNicks_erase" "', argument " "3"" of type '" "std::map< CString,CNick >::iterator""'");
    }
  }
  std_map_Sl_CString_Sc_CNick_Sg__erase__SWIG_2(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "MNicks_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::map< CString,CNick >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_MNicks_erase__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_MNicks_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::map< CString,CNick >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::map< CString,CNick >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_MNicks_erase__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'MNicks_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::map< CString,CNick >::erase(std::map< CString,CNick >::key_type const &)\n"
    "    std::map< CString,CNick >::erase(std::map< CString,CNick >::iterator)\n"
    "    std::map< CString,CNick >::erase(std::map< CString,CNick >::iterator,std::map< CString,CNick >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_MNicks_find(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::map< CString,CNick >::iterator result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "MNicks_find", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_find" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MNicks_find" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MNicks_find" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->find((std::map< CString,CNick >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CNick >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_lower_bound(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::map< CString,CNick >::iterator result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "MNicks_lower_bound", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_lower_bound" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MNicks_lower_bound" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MNicks_lower_bound" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->lower_bound((std::map< CString,CNick >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CNick >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_MNicks_upper_bound(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  std::map< CString,CNick >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::map< CString,CNick >::iterator result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "MNicks_upper_bound", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MNicks_upper_bound" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MNicks_upper_bound" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MNicks_upper_bound" "', argument " "2"" of type '" "std::map< CString,CNick >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->upper_bound((std::map< CString,CNick >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CNick >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_MNicks(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CNick > *arg1 = (std::map< CString,CNick > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CNick_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_MNicks" "', argument " "1"" of type '" "std::map< CString,CNick > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CNick > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *MNicks_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__mapT_CString_CNick_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *MNicks_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_SModInfo__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::less< CModInfo > *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::set< CModInfo > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__lessT_CModInfo_t,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_SModInfo" "', argument " "1"" of type '" "std::less< CModInfo > const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_SModInfo" "', argument " "1"" of type '" "std::less< CModInfo > const &""'"); 
  }
  arg1 = reinterpret_cast< std::less< CModInfo > * >(argp1);
  result = (std::set< CModInfo > *)new std::set< CModInfo >((std::less< CModInfo > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__setT_CModInfo_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_iterator" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  result = (swig::SwigPyIterator *)std_set_Sl_CModInfo_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo___nonzero__" "', argument " "1"" of type '" "std::set< CModInfo > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  result = (bool)std_set_Sl_CModInfo_Sg____nonzero__((std::set< CModInfo > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo___bool__" "', argument " "1"" of type '" "std::set< CModInfo > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  result = (bool)std_set_Sl_CModInfo_Sg____bool__((std::set< CModInfo > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< CModInfo >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo___len__" "', argument " "1"" of type '" "std::set< CModInfo > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  result = std_set_Sl_CModInfo_Sg____len__((std::set< CModInfo > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::value_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SModInfo_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_append" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SModInfo_append" "', argument " "2"" of type '" "std::set< CModInfo >::value_type""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SModInfo_append" "', argument " "2"" of type '" "std::set< CModInfo >::value_type""'");
    } else {
      std::set< CModInfo >::value_type * temp = reinterpret_cast< std::set< CModInfo >::value_type * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  std_set_Sl_CModInfo_Sg__append(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo___contains__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::value_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SModInfo___contains__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo___contains__" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SModInfo___contains__" "', argument " "2"" of type '" "std::set< CModInfo >::value_type""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SModInfo___contains__" "', argument " "2"" of type '" "std::set< CModInfo >::value_type""'");
    } else {
      std::set< CModInfo >::value_type * temp = reinterpret_cast< std::set< CModInfo >::value_type * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  result = (bool)std_set_Sl_CModInfo_Sg____contains__(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo___getitem__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  std::set< CModInfo >::value_type result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SModInfo___getitem__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo___getitem__" "', argument " "1"" of type '" "std::set< CModInfo > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SModInfo___getitem__" "', argument " "2"" of type '" "std::set< CModInfo >::difference_type""'");
  } 
  arg2 = static_cast< std::set< CModInfo >::difference_type >(val2);
  try {
    result = std_set_Sl_CModInfo_Sg____getitem__((std::set< CModInfo > const *)arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj((new std::set< CModInfo >::value_type(result)), SWIGTYPE_p_CModInfo, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_add(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::value_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SModInfo_add", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_add" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SModInfo_add" "', argument " "2"" of type '" "std::set< CModInfo >::value_type""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SModInfo_add" "', argument " "2"" of type '" "std::set< CModInfo >::value_type""'");
    } else {
      std::set< CModInfo >::value_type * temp = reinterpret_cast< std::set< CModInfo >::value_type * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  std_set_Sl_CModInfo_Sg__add(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_discard(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::value_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SModInfo_discard", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_discard" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SModInfo_discard" "', argument " "2"" of type '" "std::set< CModInfo >::value_type""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SModInfo_discard" "', argument " "2"" of type '" "std::set< CModInfo >::value_type""'");
    } else {
      std::set< CModInfo >::value_type * temp = reinterpret_cast< std::set< CModInfo >::value_type * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  std_set_Sl_CModInfo_Sg__discard(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_SModInfo__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::set< CModInfo > *)new std::set< CModInfo >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__setT_CModInfo_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_SModInfo__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::set< CModInfo > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > *ptr = (std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_SModInfo" "', argument " "1"" of type '" "std::set< CModInfo > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_SModInfo" "', argument " "1"" of type '" "std::set< CModInfo > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::set< CModInfo > *)new std::set< CModInfo >((std::set< CModInfo > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__setT_CModInfo_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_SModInfo(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_SModInfo", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_SModInfo__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_std__lessT_CModInfo_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_SModInfo__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_SModInfo__SWIG_2(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_SModInfo'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::set< CModInfo >::set(std::less< CModInfo > const &)\n"
    "    std::set< CModInfo >::set()\n"
    "    std::set< CModInfo >::set(std::set< CModInfo > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_SModInfo_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_empty" "', argument " "1"" of type '" "std::set< CModInfo > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  result = (bool)((std::set< CModInfo > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< CModInfo >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_size" "', argument " "1"" of type '" "std::set< CModInfo > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  result = ((std::set< CModInfo > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_clear" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SModInfo_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_swap" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__setT_CModInfo_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SModInfo_swap" "', argument " "2"" of type '" "std::set< CModInfo > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SModInfo_swap" "', argument " "2"" of type '" "std::set< CModInfo > &""'"); 
  }
  arg2 = reinterpret_cast< std::set< CModInfo > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::set< CModInfo >::size_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_erase" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SModInfo_erase" "', argument " "2"" of type '" "std::set< CModInfo >::key_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SModInfo_erase" "', argument " "2"" of type '" "std::set< CModInfo >::key_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::set< CModInfo >::key_type * >(argp2);
  result = (arg1)->erase((std::set< CModInfo >::key_type const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_count(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  std::set< CModInfo >::size_type result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SModInfo_count", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_count" "', argument " "1"" of type '" "std::set< CModInfo > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SModInfo_count" "', argument " "2"" of type '" "std::set< CModInfo >::key_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SModInfo_count" "', argument " "2"" of type '" "std::set< CModInfo >::key_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::set< CModInfo >::key_type * >(argp2);
  result = ((std::set< CModInfo > const *)arg1)->count((std::set< CModInfo >::key_type const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< CModInfo >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_begin" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CModInfo >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< CModInfo >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_end" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CModInfo >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< CModInfo >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_rbegin" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CModInfo >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< CModInfo >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_rend" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CModInfo >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_erase" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "SModInfo_erase" "', argument " "2"" of type '" "std::set< CModInfo >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::set< CModInfo >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::set< CModInfo >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "SModInfo_erase" "', argument " "2"" of type '" "std::set< CModInfo >::iterator""'");
    }
  }
  std_set_Sl_CModInfo_Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_erase__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::iterator arg2 ;
  std::set< CModInfo >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_erase" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "SModInfo_erase" "', argument " "2"" of type '" "std::set< CModInfo >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::set< CModInfo >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::set< CModInfo >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "SModInfo_erase" "', argument " "2"" of type '" "std::set< CModInfo >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "SModInfo_erase" "', argument " "3"" of type '" "std::set< CModInfo >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::set< CModInfo >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::set< CModInfo >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "SModInfo_erase" "', argument " "3"" of type '" "std::set< CModInfo >::iterator""'");
    }
  }
  std_set_Sl_CModInfo_Sg__erase__SWIG_2(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "SModInfo_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CModInfo, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_SModInfo_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::set< CModInfo >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_SModInfo_erase__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::set< CModInfo >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::set< CModInfo >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_SModInfo_erase__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'SModInfo_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::set< CModInfo >::erase(std::set< CModInfo >::key_type const &)\n"
    "    std::set< CModInfo >::erase(std::set< CModInfo >::iterator)\n"
    "    std::set< CModInfo >::erase(std::set< CModInfo >::iterator,std::set< CModInfo >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_SModInfo_find(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  std::set< CModInfo >::iterator result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SModInfo_find", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_find" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SModInfo_find" "', argument " "2"" of type '" "std::set< CModInfo >::key_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SModInfo_find" "', argument " "2"" of type '" "std::set< CModInfo >::key_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::set< CModInfo >::key_type * >(argp2);
  result = (arg1)->find((std::set< CModInfo >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CModInfo >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_lower_bound(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  std::set< CModInfo >::iterator result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SModInfo_lower_bound", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_lower_bound" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SModInfo_lower_bound" "', argument " "2"" of type '" "std::set< CModInfo >::key_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SModInfo_lower_bound" "', argument " "2"" of type '" "std::set< CModInfo >::key_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::set< CModInfo >::key_type * >(argp2);
  result = (arg1)->lower_bound((std::set< CModInfo >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CModInfo >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_upper_bound(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  std::set< CModInfo >::iterator result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SModInfo_upper_bound", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_upper_bound" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SModInfo_upper_bound" "', argument " "2"" of type '" "std::set< CModInfo >::key_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SModInfo_upper_bound" "', argument " "2"" of type '" "std::set< CModInfo >::key_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::set< CModInfo >::key_type * >(argp2);
  result = (arg1)->upper_bound((std::set< CModInfo >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CModInfo >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_equal_range(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  SwigValueWrapper< std::pair< std::set< CModInfo >::iterator,std::set< CModInfo >::iterator > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SModInfo_equal_range", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_equal_range" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SModInfo_equal_range" "', argument " "2"" of type '" "std::set< CModInfo >::key_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SModInfo_equal_range" "', argument " "2"" of type '" "std::set< CModInfo >::key_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::set< CModInfo >::key_type * >(argp2);
  result = (arg1)->equal_range((std::set< CModInfo >::key_type const &)*arg2);
  resultobj = PyTuple_New(2);
  PyTuple_SetItem(resultobj,0,SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::pair< std::set< CModInfo >::iterator,std::set< CModInfo >::iterator > & >(result).first),
      swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN));
  PyTuple_SetItem(resultobj,1,SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::pair< std::set< CModInfo >::iterator,std::set< CModInfo >::iterator > & >(result).second),
      swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SModInfo_insert(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  std::set< CModInfo >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  SwigValueWrapper< std::pair< std::set< CModInfo >::iterator,bool > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SModInfo_insert", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SModInfo_insert" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SModInfo_insert" "', argument " "2"" of type '" "std::set< CModInfo >::value_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SModInfo_insert" "', argument " "2"" of type '" "std::set< CModInfo >::value_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::set< CModInfo >::value_type * >(argp2);
  result = (arg1)->insert((std::set< CModInfo >::value_type const &)*arg2);
  resultobj = PyTuple_New(2);
  PyTuple_SetItem(resultobj,0,SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::pair< std::set< CModInfo >::iterator,bool > & >(result).first),
      swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN));    
  PyTuple_SetItem(resultobj,1,SWIG_From_bool  (static_cast< const std::pair< std::set< CModInfo >::iterator,bool > & >(result).second));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_SModInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CModInfo > *arg1 = (std::set< CModInfo > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CModInfo_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SModInfo" "', argument " "1"" of type '" "std::set< CModInfo > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *SModInfo_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__setT_CModInfo_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *SModInfo_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_SCString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::less< CString > *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::set< CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__lessT_CString_t,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_SCString" "', argument " "1"" of type '" "std::less< CString > const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_SCString" "', argument " "1"" of type '" "std::less< CString > const &""'"); 
  }
  arg1 = reinterpret_cast< std::less< CString > * >(argp1);
  result = (std::set< CString > *)new std::set< CString >((std::less< CString > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__setT_CString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_iterator" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  result = (swig::SwigPyIterator *)std_set_Sl_CString_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString___nonzero__" "', argument " "1"" of type '" "std::set< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  result = (bool)std_set_Sl_CString_Sg____nonzero__((std::set< CString > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString___bool__" "', argument " "1"" of type '" "std::set< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  result = (bool)std_set_Sl_CString_Sg____bool__((std::set< CString > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< CString >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString___len__" "', argument " "1"" of type '" "std::set< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  result = std_set_Sl_CString_Sg____len__((std::set< CString > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::value_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SCString_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_append" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    int res = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "SCString_append" "', argument " "2"" of type '" "std::set< CString >::value_type""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  std_set_Sl_CString_Sg__append(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString___contains__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::value_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SCString___contains__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString___contains__" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    int res = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "SCString___contains__" "', argument " "2"" of type '" "std::set< CString >::value_type""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  result = (bool)std_set_Sl_CString_Sg____contains__(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString___getitem__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  std::set< CString >::value_type result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SCString___getitem__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString___getitem__" "', argument " "1"" of type '" "std::set< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SCString___getitem__" "', argument " "2"" of type '" "std::set< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::set< CString >::difference_type >(val2);
  try {
    result = std_set_Sl_CString_Sg____getitem__((std::set< CString > const *)arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_add(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::value_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SCString_add", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_add" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    int res = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "SCString_add" "', argument " "2"" of type '" "std::set< CString >::value_type""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  std_set_Sl_CString_Sg__add(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_discard(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::value_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SCString_discard", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_discard" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    int res = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "SCString_discard" "', argument " "2"" of type '" "std::set< CString >::value_type""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  std_set_Sl_CString_Sg__discard(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_SCString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::set< CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::set< CString > *)new std::set< CString >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__setT_CString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_SCString__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::set< CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::set< CString,std::less< CString >,std::allocator< CString > > *ptr = (std::set< CString,std::less< CString >,std::allocator< CString > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_SCString" "', argument " "1"" of type '" "std::set< CString > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_SCString" "', argument " "1"" of type '" "std::set< CString > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::set< CString > *)new std::set< CString >((std::set< CString > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__setT_CString_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_SCString(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_SCString", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_SCString__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_std__lessT_CString_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_SCString__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::set< CString,std::less< CString >,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_SCString__SWIG_2(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_SCString'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::set< CString >::set(std::less< CString > const &)\n"
    "    std::set< CString >::set()\n"
    "    std::set< CString >::set(std::set< CString > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_SCString_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_empty" "', argument " "1"" of type '" "std::set< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  result = (bool)((std::set< CString > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< CString >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_size" "', argument " "1"" of type '" "std::set< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  result = ((std::set< CString > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_clear" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SCString_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_swap" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__setT_CString_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SCString_swap" "', argument " "2"" of type '" "std::set< CString > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SCString_swap" "', argument " "2"" of type '" "std::set< CString > &""'"); 
  }
  arg2 = reinterpret_cast< std::set< CString > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  std::set< CString >::size_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_erase" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SCString_erase" "', argument " "2"" of type '" "std::set< CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SCString_erase" "', argument " "2"" of type '" "std::set< CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->erase((std::set< CString >::key_type const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_count(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::set< CString >::size_type result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SCString_count", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_count" "', argument " "1"" of type '" "std::set< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SCString_count" "', argument " "2"" of type '" "std::set< CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SCString_count" "', argument " "2"" of type '" "std::set< CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((std::set< CString > const *)arg1)->count((std::set< CString >::key_type const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< CString >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_begin" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< CString >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_end" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< CString >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_rbegin" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CString >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< CString >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_rend" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CString >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_erase" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "SCString_erase" "', argument " "2"" of type '" "std::set< CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::set< CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::set< CString >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "SCString_erase" "', argument " "2"" of type '" "std::set< CString >::iterator""'");
    }
  }
  std_set_Sl_CString_Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_erase__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::iterator arg2 ;
  std::set< CString >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_erase" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "SCString_erase" "', argument " "2"" of type '" "std::set< CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::set< CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::set< CString >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "SCString_erase" "', argument " "2"" of type '" "std::set< CString >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "SCString_erase" "', argument " "3"" of type '" "std::set< CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::set< CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::set< CString >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "SCString_erase" "', argument " "3"" of type '" "std::set< CString >::iterator""'");
    }
  }
  std_set_Sl_CString_Sg__erase__SWIG_2(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "SCString_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::set< CString,std::less< CString >,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::set< CString >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_SCString_erase__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::set< CString,std::less< CString >,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_SCString_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::set< CString,std::less< CString >,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::set< CString >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::set< CString >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_SCString_erase__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'SCString_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::set< CString >::erase(std::set< CString >::key_type const &)\n"
    "    std::set< CString >::erase(std::set< CString >::iterator)\n"
    "    std::set< CString >::erase(std::set< CString >::iterator,std::set< CString >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_SCString_find(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::set< CString >::iterator result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SCString_find", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_find" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SCString_find" "', argument " "2"" of type '" "std::set< CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SCString_find" "', argument " "2"" of type '" "std::set< CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->find((std::set< CString >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_lower_bound(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::set< CString >::iterator result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SCString_lower_bound", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_lower_bound" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SCString_lower_bound" "', argument " "2"" of type '" "std::set< CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SCString_lower_bound" "', argument " "2"" of type '" "std::set< CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->lower_bound((std::set< CString >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_upper_bound(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::set< CString >::iterator result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SCString_upper_bound", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_upper_bound" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SCString_upper_bound" "', argument " "2"" of type '" "std::set< CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SCString_upper_bound" "', argument " "2"" of type '" "std::set< CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->upper_bound((std::set< CString >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_equal_range(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  SwigValueWrapper< std::pair< std::set< CString >::iterator,std::set< CString >::iterator > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SCString_equal_range", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_equal_range" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SCString_equal_range" "', argument " "2"" of type '" "std::set< CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SCString_equal_range" "', argument " "2"" of type '" "std::set< CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->equal_range((std::set< CString >::key_type const &)*arg2);
  resultobj = PyTuple_New(2);
  PyTuple_SetItem(resultobj,0,SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::pair< std::set< CString >::iterator,std::set< CString >::iterator > & >(result).first),
      swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN));
  PyTuple_SetItem(resultobj,1,SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::pair< std::set< CString >::iterator,std::set< CString >::iterator > & >(result).second),
      swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_SCString_insert(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  std::set< CString >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  SwigValueWrapper< std::pair< std::set< CString >::iterator,bool > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SCString_insert", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SCString_insert" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SCString_insert" "', argument " "2"" of type '" "std::set< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SCString_insert" "', argument " "2"" of type '" "std::set< CString >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->insert((std::set< CString >::value_type const &)*arg2);
  resultobj = PyTuple_New(2);
  PyTuple_SetItem(resultobj,0,SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::pair< std::set< CString >::iterator,bool > & >(result).first),
      swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN));    
  PyTuple_SetItem(resultobj,1,SWIG_From_bool  (static_cast< const std::pair< std::set< CString >::iterator,bool > & >(result).second));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_SCString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::set< CString > *arg1 = (std::set< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__setT_CString_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SCString" "', argument " "1"" of type '" "std::set< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::set< CString > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *SCString_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__setT_CString_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *SCString_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_VCString_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_iterator" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_CString_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___nonzero__" "', argument " "1"" of type '" "std::vector< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = (bool)std_vector_Sl_CString_Sg____nonzero__((std::vector< CString > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___bool__" "', argument " "1"" of type '" "std::vector< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = (bool)std_vector_Sl_CString_Sg____bool__((std::vector< CString > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CString >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___len__" "', argument " "1"" of type '" "std::vector< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = std_vector_Sl_CString_Sg____len__((std::vector< CString > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___getslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::difference_type arg2 ;
  std::vector< CString >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  std::vector< CString,std::allocator< CString > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VCString___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___getslice__" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VCString___getslice__" "', argument " "2"" of type '" "std::vector< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CString >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VCString___getslice__" "', argument " "3"" of type '" "std::vector< CString >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CString >::difference_type >(val3);
  try {
    result = (std::vector< CString,std::allocator< CString > > *)std_vector_Sl_CString_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::difference_type arg2 ;
  std::vector< CString >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___setslice__" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VCString___setslice__" "', argument " "2"" of type '" "std::vector< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CString >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VCString___setslice__" "', argument " "3"" of type '" "std::vector< CString >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CString >::difference_type >(val3);
  try {
    std_vector_Sl_CString_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::difference_type arg2 ;
  std::vector< CString >::difference_type arg3 ;
  std::vector< CString,std::allocator< CString > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___setslice__" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VCString___setslice__" "', argument " "2"" of type '" "std::vector< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CString >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VCString___setslice__" "', argument " "3"" of type '" "std::vector< CString >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CString >::difference_type >(val3);
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VCString___setslice__" "', argument " "4"" of type '" "std::vector< CString,std::allocator< CString > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VCString___setslice__" "', argument " "4"" of type '" "std::vector< CString,std::allocator< CString > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_CString_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< CString,std::allocator< CString > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VCString___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_VCString___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< CString,std::allocator< CString > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VCString___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VCString___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CString >::__setslice__(std::vector< CString >::difference_type,std::vector< CString >::difference_type)\n"
    "    std::vector< CString >::__setslice__(std::vector< CString >::difference_type,std::vector< CString >::difference_type,std::vector< CString,std::allocator< CString > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VCString___delslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::difference_type arg2 ;
  std::vector< CString >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VCString___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___delslice__" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VCString___delslice__" "', argument " "2"" of type '" "std::vector< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CString >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VCString___delslice__" "', argument " "3"" of type '" "std::vector< CString >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CString >::difference_type >(val3);
  try {
    std_vector_Sl_CString_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___delitem__" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VCString___delitem__" "', argument " "2"" of type '" "std::vector< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CString >::difference_type >(val2);
  try {
    std_vector_Sl_CString_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< CString,std::allocator< CString > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___getitem__" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    result = (std::vector< CString,std::allocator< CString > > *)std_vector_Sl_CString_Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  std::vector< CString,std::allocator< CString > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___setitem__" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VCString___setitem__" "', argument " "3"" of type '" "std::vector< CString,std::allocator< CString > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VCString___setitem__" "', argument " "3"" of type '" "std::vector< CString,std::allocator< CString > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_CString_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< CString,std::allocator< CString > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___setitem__" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CString_Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___delitem__" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CString_Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VCString___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VCString___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VCString___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VCString___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CString >::__delitem__(std::vector< CString >::difference_type)\n"
    "    std::vector< CString >::__delitem__(SWIGPY_SLICEOBJECT *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VCString___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< CString >::value_type *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___getitem__" "', argument " "1"" of type '" "std::vector< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VCString___getitem__" "', argument " "2"" of type '" "std::vector< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CString >::difference_type >(val2);
  try {
    result = (std::vector< CString >::value_type *) &std_vector_Sl_CString_Sg____getitem____SWIG_1((std::vector< CString > const *)arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  (void)swig::container_owner<swig::traits<std::vector< CString >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VCString___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VCString___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VCString___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VCString___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CString >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CString >::__getitem__(std::vector< CString >::difference_type) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VCString___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::difference_type arg2 ;
  std::vector< CString >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString___setitem__" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VCString___setitem__" "', argument " "2"" of type '" "std::vector< CString >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CString >::difference_type >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VCString___setitem__" "', argument " "3"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VCString___setitem__" "', argument " "3"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_CString_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(CString const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VCString___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VCString___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CString,std::allocator< CString > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VCString___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VCString___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VCString___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CString >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< CString,std::allocator< CString > > const &)\n"
    "    std::vector< CString >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CString >::__setitem__(std::vector< CString >::difference_type,std::vector< CString >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VCString_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CString >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_pop" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  try {
    result = std_vector_Sl_CString_Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VCString_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_append" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VCString_append" "', argument " "2"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VCString_append" "', argument " "2"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  std_vector_Sl_CString_Sg__append(arg1,(CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VCString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< CString > *)new std::vector< CString >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VCString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_VCString" "', argument " "1"" of type '" "std::vector< CString > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VCString" "', argument " "1"" of type '" "std::vector< CString > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< CString > *)new std::vector< CString >((std::vector< CString > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_empty" "', argument " "1"" of type '" "std::vector< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = (bool)((std::vector< CString > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CString >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_size" "', argument " "1"" of type '" "std::vector< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = ((std::vector< CString > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VCString_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_swap" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_CString_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VCString_swap" "', argument " "2"" of type '" "std::vector< CString > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VCString_swap" "', argument " "2"" of type '" "std::vector< CString > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CString > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CString >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_begin" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CString >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_end" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CString >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_rbegin" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CString >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CString >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_rend" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CString >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_clear" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< CString > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_get_allocator" "', argument " "1"" of type '" "std::vector< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = ((std::vector< CString > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< CString >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_CString_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VCString__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VCString" "', argument " "1"" of type '" "std::vector< CString >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CString >::size_type >(val1);
  result = (std::vector< CString > *)new std::vector< CString >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_pop_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_pop_back" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_resize" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VCString_resize" "', argument " "2"" of type '" "std::vector< CString >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CString >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< CString >::iterator result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_erase" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString_erase" "', argument " "2"" of type '" "std::vector< CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CString >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString_erase" "', argument " "2"" of type '" "std::vector< CString >::iterator""'");
    }
  }
  result = std_vector_Sl_CString_Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::iterator arg2 ;
  std::vector< CString >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::vector< CString >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_erase" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString_erase" "', argument " "2"" of type '" "std::vector< CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CString >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString_erase" "', argument " "2"" of type '" "std::vector< CString >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString_erase" "', argument " "3"" of type '" "std::vector< CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CString >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString_erase" "', argument " "3"" of type '" "std::vector< CString >::iterator""'");
    }
  }
  result = std_vector_Sl_CString_Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VCString_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CString >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_VCString_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CString >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CString >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_VCString_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VCString_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CString >::erase(std::vector< CString >::iterator)\n"
    "    std::vector< CString >::erase(std::vector< CString >::iterator,std::vector< CString >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_VCString__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString >::size_type arg1 ;
  std::vector< CString >::value_type *arg2 = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  std::vector< CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VCString" "', argument " "1"" of type '" "std::vector< CString >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CString >::size_type >(val1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_VCString" "', argument " "2"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VCString" "', argument " "2"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (std::vector< CString > *)new std::vector< CString >(arg1,(std::vector< CString >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VCString(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_VCString", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_VCString__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_VCString__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_VCString__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_VCString__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_VCString'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CString >::vector()\n"
    "    std::vector< CString >::vector(std::vector< CString > const &)\n"
    "    std::vector< CString >::vector(std::vector< CString >::size_type)\n"
    "    std::vector< CString >::vector(std::vector< CString >::size_type,std::vector< CString >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VCString_push_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VCString_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_push_back" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VCString_push_back" "', argument " "2"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VCString_push_back" "', argument " "2"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->push_back((std::vector< CString >::value_type const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CString >::value_type *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_front" "', argument " "1"" of type '" "std::vector< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = (std::vector< CString >::value_type *) &((std::vector< CString > const *)arg1)->front();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  (void)swig::container_owner<swig::traits<std::vector< CString >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CString >::value_type *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_back" "', argument " "1"" of type '" "std::vector< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = (std::vector< CString >::value_type *) &((std::vector< CString > const *)arg1)->back();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  (void)swig::container_owner<swig::traits<std::vector< CString >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_assign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::size_type arg2 ;
  std::vector< CString >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VCString_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_assign" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VCString_assign" "', argument " "2"" of type '" "std::vector< CString >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CString >::size_type >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VCString_assign" "', argument " "3"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VCString_assign" "', argument " "3"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->assign(arg2,(std::vector< CString >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::size_type arg2 ;
  std::vector< CString >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_resize" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VCString_resize" "', argument " "2"" of type '" "std::vector< CString >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CString >::size_type >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VCString_resize" "', argument " "3"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VCString_resize" "', argument " "3"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->resize(arg2,(std::vector< CString >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VCString_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VCString_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VCString_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VCString_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CString >::resize(std::vector< CString >::size_type)\n"
    "    std::vector< CString >::resize(std::vector< CString >::size_type,std::vector< CString >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VCString_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::iterator arg2 ;
  std::vector< CString >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  int res3 = SWIG_OLDOBJ ;
  std::vector< CString >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_insert" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString_insert" "', argument " "2"" of type '" "std::vector< CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CString >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString_insert" "', argument " "2"" of type '" "std::vector< CString >::iterator""'");
    }
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VCString_insert" "', argument " "3"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VCString_insert" "', argument " "3"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  result = std_vector_Sl_CString_Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),(CString const &)*arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::iterator arg2 ;
  std::vector< CString >::size_type arg3 ;
  std::vector< CString >::value_type *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_insert" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString_insert" "', argument " "2"" of type '" "std::vector< CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CString >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VCString_insert" "', argument " "2"" of type '" "std::vector< CString >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VCString_insert" "', argument " "3"" of type '" "std::vector< CString >::size_type""'");
  } 
  arg3 = static_cast< std::vector< CString >::size_type >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VCString_insert" "', argument " "4"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VCString_insert" "', argument " "4"" of type '" "std::vector< CString >::value_type const &""'"); 
    }
    arg4 = ptr;
  }
  std_vector_Sl_CString_Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VCString_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CString >::iterator > *>(iter) != 0));
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VCString_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CString,std::allocator< CString > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CString >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VCString_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VCString_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CString >::insert(std::vector< CString >::iterator,std::vector< CString >::value_type const &)\n"
    "    std::vector< CString >::insert(std::vector< CString >::iterator,std::vector< CString >::size_type,std::vector< CString >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VCString_reserve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  std::vector< CString >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VCString_reserve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_reserve" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VCString_reserve" "', argument " "2"" of type '" "std::vector< CString >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CString >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VCString_capacity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CString >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VCString_capacity" "', argument " "1"" of type '" "std::vector< CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  result = ((std::vector< CString > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_VCString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CString > *arg1 = (std::vector< CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VCString" "', argument " "1"" of type '" "std::vector< CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CString > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *VCString_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_CString_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *VCString_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_PyMCString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::less< CString > *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::map< CString,CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__lessT_CString_t,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_PyMCString" "', argument " "1"" of type '" "std::less< CString > const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_PyMCString" "', argument " "1"" of type '" "std::less< CString > const &""'"); 
  }
  arg1 = reinterpret_cast< std::less< CString > * >(argp1);
  result = (std::map< CString,CString > *)new std::map< CString,CString >((std::less< CString > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_CString_CString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_iterator" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (swig::SwigPyIterator *)std_map_Sl_CString_Sc_CString_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString___nonzero__" "', argument " "1"" of type '" "std::map< CString,CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (bool)std_map_Sl_CString_Sc_CString_Sg____nonzero__((std::map< CString,CString > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString___bool__" "', argument " "1"" of type '" "std::map< CString,CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (bool)std_map_Sl_CString_Sc_CString_Sg____bool__((std::map< CString,CString > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CString >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString___len__" "', argument " "1"" of type '" "std::map< CString,CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = std_map_Sl_CString_Sc_CString_Sg____len__((std::map< CString,CString > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString___getitem__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::map< CString,CString >::mapped_type *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMCString___getitem__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString___getitem__" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMCString___getitem__" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMCString___getitem__" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  try {
    result = (std::map< CString,CString >::mapped_type *) &std_map_Sl_CString_Sc_CString_Sg____getitem__(arg1,(CString const &)*arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString___delitem__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMCString___delitem__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString___delitem__" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMCString___delitem__" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMCString___delitem__" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  try {
    std_map_Sl_CString_Sc_CString_Sg____delitem__(arg1,(CString const &)*arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_has_key(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMCString_has_key", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_has_key" "', argument " "1"" of type '" "std::map< CString,CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMCString_has_key" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMCString_has_key" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)std_map_Sl_CString_Sc_CString_Sg__has_key((std::map< CString,CString > const *)arg1,(CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_keys(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_keys" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (PyObject *)std_map_Sl_CString_Sc_CString_Sg__keys(arg1);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_values(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_values" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (PyObject *)std_map_Sl_CString_Sc_CString_Sg__values(arg1);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_items(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_items" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (PyObject *)std_map_Sl_CString_Sc_CString_Sg__items(arg1);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString___contains__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMCString___contains__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString___contains__" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMCString___contains__" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMCString___contains__" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)std_map_Sl_CString_Sc_CString_Sg____contains__(arg1,(CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_key_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_key_iterator" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (swig::SwigPyIterator *)std_map_Sl_CString_Sc_CString_Sg__key_iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_value_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_value_iterator" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (swig::SwigPyIterator *)std_map_Sl_CString_Sc_CString_Sg__value_iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString___setitem__" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMCString___setitem__" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMCString___setitem__" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  std_map_Sl_CString_Sc_CString_Sg____setitem____SWIG_0(arg1,(CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString >::key_type *arg2 = 0 ;
  std::map< CString,CString >::mapped_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString___setitem__" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMCString___setitem__" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMCString___setitem__" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "PyMCString___setitem__" "', argument " "3"" of type '" "std::map< CString,CString >::mapped_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMCString___setitem__" "', argument " "3"" of type '" "std::map< CString,CString >::mapped_type const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_map_Sl_CString_Sc_CString_Sg____setitem____SWIG_1(arg1,(CString const &)*arg2,(CString const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PyMCString___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_PyMCString___setitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_PyMCString___setitem____SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PyMCString___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::map< CString,CString >::__setitem__(std::map< CString,CString >::key_type const &)\n"
    "    std::map< CString,CString >::__setitem__(std::map< CString,CString >::key_type const &,std::map< CString,CString >::mapped_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PyMCString_asdict(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_asdict" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (PyObject *)std_map_Sl_CString_Sc_CString_Sg__asdict(arg1);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_PyMCString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::map< CString,CString > *)new std::map< CString,CString >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_CString_CString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_PyMCString__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::map< CString,CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > > *ptr = (std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_PyMCString" "', argument " "1"" of type '" "std::map< CString,CString > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_PyMCString" "', argument " "1"" of type '" "std::map< CString,CString > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::map< CString,CString > *)new std::map< CString,CString >((std::map< CString,CString > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_CString_CString_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_PyMCString(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_PyMCString", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_PyMCString__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_std__lessT_CString_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_PyMCString__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_PyMCString__SWIG_2(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_PyMCString'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::map< CString,CString >::map(std::less< CString > const &)\n"
    "    std::map< CString,CString >::map()\n"
    "    std::map< CString,CString >::map(std::map< CString,CString > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PyMCString_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_empty" "', argument " "1"" of type '" "std::map< CString,CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (bool)((std::map< CString,CString > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CString >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_size" "', argument " "1"" of type '" "std::map< CString,CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = ((std::map< CString,CString > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMCString_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_swap" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__mapT_CString_CString_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMCString_swap" "', argument " "2"" of type '" "std::map< CString,CString > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMCString_swap" "', argument " "2"" of type '" "std::map< CString,CString > &""'"); 
  }
  arg2 = reinterpret_cast< std::map< CString,CString > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CString >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_begin" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CString >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_end" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CString >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_rbegin" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CString >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CString >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_rend" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CString >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_clear" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< std::pair< CString const,CString > > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_get_allocator" "', argument " "1"" of type '" "std::map< CString,CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  result = ((std::map< CString,CString > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::map< CString,CString >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_std__pairT_CString_const_CString_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  std::map< CString,CString >::size_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_erase" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMCString_erase" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMCString_erase" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->erase((std::map< CString,CString >::key_type const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_count(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::map< CString,CString >::size_type result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMCString_count", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_count" "', argument " "1"" of type '" "std::map< CString,CString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMCString_count" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMCString_count" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((std::map< CString,CString > const *)arg1)->count((std::map< CString,CString >::key_type const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_erase" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyMCString_erase" "', argument " "2"" of type '" "std::map< CString,CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::map< CString,CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::map< CString,CString >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyMCString_erase" "', argument " "2"" of type '" "std::map< CString,CString >::iterator""'");
    }
  }
  std_map_Sl_CString_Sc_CString_Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_erase__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString >::iterator arg2 ;
  std::map< CString,CString >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_erase" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyMCString_erase" "', argument " "2"" of type '" "std::map< CString,CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::map< CString,CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::map< CString,CString >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyMCString_erase" "', argument " "2"" of type '" "std::map< CString,CString >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyMCString_erase" "', argument " "3"" of type '" "std::map< CString,CString >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::map< CString,CString >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::map< CString,CString >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyMCString_erase" "', argument " "3"" of type '" "std::map< CString,CString >::iterator""'");
    }
  }
  std_map_Sl_CString_Sc_CString_Sg__erase__SWIG_2(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PyMCString_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::map< CString,CString >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_PyMCString_erase__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_PyMCString_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::map< CString,CString >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::map< CString,CString >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_PyMCString_erase__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PyMCString_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::map< CString,CString >::erase(std::map< CString,CString >::key_type const &)\n"
    "    std::map< CString,CString >::erase(std::map< CString,CString >::iterator)\n"
    "    std::map< CString,CString >::erase(std::map< CString,CString >::iterator,std::map< CString,CString >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PyMCString_find(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::map< CString,CString >::iterator result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMCString_find", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_find" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMCString_find" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMCString_find" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->find((std::map< CString,CString >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_lower_bound(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::map< CString,CString >::iterator result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMCString_lower_bound", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_lower_bound" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMCString_lower_bound" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMCString_lower_bound" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->lower_bound((std::map< CString,CString >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMCString_upper_bound(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  std::map< CString,CString >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::map< CString,CString >::iterator result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMCString_upper_bound", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMCString_upper_bound" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMCString_upper_bound" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMCString_upper_bound" "', argument " "2"" of type '" "std::map< CString,CString >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->upper_bound((std::map< CString,CString >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,CString >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_PyMCString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,CString > *arg1 = (std::map< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_CString_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_PyMCString" "', argument " "1"" of type '" "std::map< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,CString > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *PyMCString_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__mapT_CString_CString_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *PyMCString_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_PyMStringVString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::less< CString > *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::map< CString,VCString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__lessT_CString_t,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_PyMStringVString" "', argument " "1"" of type '" "std::less< CString > const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_PyMStringVString" "', argument " "1"" of type '" "std::less< CString > const &""'"); 
  }
  arg1 = reinterpret_cast< std::less< CString > * >(argp1);
  result = (std::map< CString,VCString > *)new std::map< CString,VCString >((std::less< CString > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_CString_VCString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_iterator" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (swig::SwigPyIterator *)std_map_Sl_CString_Sc_VCString_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString___nonzero__" "', argument " "1"" of type '" "std::map< CString,VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (bool)std_map_Sl_CString_Sc_VCString_Sg____nonzero__((std::map< CString,std::vector< CString,std::allocator< CString > > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString___bool__" "', argument " "1"" of type '" "std::map< CString,VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (bool)std_map_Sl_CString_Sc_VCString_Sg____bool__((std::map< CString,std::vector< CString,std::allocator< CString > > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString___len__" "', argument " "1"" of type '" "std::map< CString,VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = std_map_Sl_CString_Sc_VCString_Sg____len__((std::map< CString,std::vector< CString,std::allocator< CString > > > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString___getitem__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::mapped_type *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMStringVString___getitem__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString___getitem__" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMStringVString___getitem__" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMStringVString___getitem__" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  try {
    result = (std::map< CString,std::vector< CString,std::allocator< CString > > >::mapped_type *) &std_map_Sl_CString_Sc_VCString_Sg____getitem__(arg1,(CString const &)*arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = swig::from(static_cast< std::vector< CString,std::allocator< CString > > >(*result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString___delitem__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMStringVString___delitem__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString___delitem__" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMStringVString___delitem__" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMStringVString___delitem__" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  try {
    std_map_Sl_CString_Sc_VCString_Sg____delitem__(arg1,(CString const &)*arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_has_key(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMStringVString_has_key", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_has_key" "', argument " "1"" of type '" "std::map< CString,VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMStringVString_has_key" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMStringVString_has_key" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)std_map_Sl_CString_Sc_VCString_Sg__has_key((std::map< CString,std::vector< CString,std::allocator< CString > > > const *)arg1,(CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_keys(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_keys" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (PyObject *)std_map_Sl_CString_Sc_VCString_Sg__keys(arg1);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_values(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_values" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (PyObject *)std_map_Sl_CString_Sc_VCString_Sg__values(arg1);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_items(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_items" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (PyObject *)std_map_Sl_CString_Sc_VCString_Sg__items(arg1);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString___contains__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMStringVString___contains__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString___contains__" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMStringVString___contains__" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMStringVString___contains__" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)std_map_Sl_CString_Sc_VCString_Sg____contains__(arg1,(CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_key_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_key_iterator" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (swig::SwigPyIterator *)std_map_Sl_CString_Sc_VCString_Sg__key_iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_value_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_value_iterator" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (swig::SwigPyIterator *)std_map_Sl_CString_Sc_VCString_Sg__value_iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString___setitem__" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMStringVString___setitem__" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMStringVString___setitem__" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  std_map_Sl_CString_Sc_VCString_Sg____setitem____SWIG_0(arg1,(CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type *arg2 = 0 ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::mapped_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString___setitem__" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMStringVString___setitem__" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMStringVString___setitem__" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "PyMStringVString___setitem__" "', argument " "3"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::mapped_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMStringVString___setitem__" "', argument " "3"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::mapped_type const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_map_Sl_CString_Sc_VCString_Sg____setitem____SWIG_1(arg1,(CString const &)*arg2,(std::vector< CString,std::allocator< CString > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PyMStringVString___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_PyMStringVString___setitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CString,std::allocator< CString > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_PyMStringVString___setitem____SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PyMStringVString___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::map< CString,VCString >::__setitem__(std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &)\n"
    "    std::map< CString,VCString >::__setitem__(std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &,std::map< CString,std::vector< CString,std::allocator< CString > > >::mapped_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_asdict(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_asdict" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (PyObject *)std_map_Sl_CString_Sc_VCString_Sg__asdict(arg1);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_PyMStringVString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::map< CString,VCString > *)new std::map< CString,VCString >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_CString_VCString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_PyMStringVString__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::map< CString,VCString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > > *ptr = (std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_PyMStringVString" "', argument " "1"" of type '" "std::map< CString,VCString > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_PyMStringVString" "', argument " "1"" of type '" "std::map< CString,VCString > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::map< CString,VCString > *)new std::map< CString,VCString >((std::map< CString,VCString > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_CString_VCString_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_PyMStringVString(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_PyMStringVString", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_PyMStringVString__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_std__lessT_CString_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_PyMStringVString__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_PyMStringVString__SWIG_2(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_PyMStringVString'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::map< CString,VCString >::map(std::less< CString > const &)\n"
    "    std::map< CString,VCString >::map()\n"
    "    std::map< CString,VCString >::map(std::map< CString,VCString > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_empty" "', argument " "1"" of type '" "std::map< CString,VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (bool)((std::map< CString,VCString > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_size" "', argument " "1"" of type '" "std::map< CString,VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = ((std::map< CString,VCString > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  std::map< CString,VCString > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMStringVString_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_swap" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__mapT_CString_VCString_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMStringVString_swap" "', argument " "2"" of type '" "std::map< CString,VCString > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMStringVString_swap" "', argument " "2"" of type '" "std::map< CString,VCString > &""'"); 
  }
  arg2 = reinterpret_cast< std::map< CString,VCString > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_begin" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_end" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::map< CString,std::vector< CString,std::allocator< CString > > >::reverse_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_rbegin" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,std::vector< CString,std::allocator< CString > > >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::map< CString,std::vector< CString,std::allocator< CString > > >::reverse_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_rend" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,std::vector< CString,std::allocator< CString > > >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_clear" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_get_allocator" "', argument " "1"" of type '" "std::map< CString,VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  result = ((std::map< CString,VCString > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::map< CString,std::vector< CString,std::allocator< CString > > >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_std__pairT_CString_const_std__vectorT_CString_t_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::size_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_erase" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMStringVString_erase" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMStringVString_erase" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->erase((std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_count(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::size_type result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMStringVString_count", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_count" "', argument " "1"" of type '" "std::map< CString,VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMStringVString_count" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMStringVString_count" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((std::map< CString,VCString > const *)arg1)->count((std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  SwigValueWrapper< std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_erase" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyMStringVString_erase" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyMStringVString_erase" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator""'");
    }
  }
  std_map_Sl_CString_Sc_VCString_Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_erase__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  SwigValueWrapper< std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > arg2 ;
  SwigValueWrapper< std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_erase" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyMStringVString_erase" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyMStringVString_erase" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyMStringVString_erase" "', argument " "3"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyMStringVString_erase" "', argument " "3"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator""'");
    }
  }
  std_map_Sl_CString_Sc_VCString_Sg__erase__SWIG_2(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PyMStringVString_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_PyMStringVString_erase__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_PyMStringVString_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_PyMStringVString_erase__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PyMStringVString_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::map< CString,VCString >::erase(std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &)\n"
    "    std::map< CString,VCString >::erase(std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator)\n"
    "    std::map< CString,VCString >::erase(std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator,std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_find(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  SwigValueWrapper< std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMStringVString_find", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_find" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMStringVString_find" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMStringVString_find" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->find((std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_lower_bound(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  SwigValueWrapper< std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMStringVString_lower_bound", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_lower_bound" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMStringVString_lower_bound" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMStringVString_lower_bound" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->lower_bound((std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyMStringVString_upper_bound(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  SwigValueWrapper< std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyMStringVString_upper_bound", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyMStringVString_upper_bound" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyMStringVString_upper_bound" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyMStringVString_upper_bound" "', argument " "2"" of type '" "std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->upper_bound((std::map< CString,std::vector< CString,std::allocator< CString > > >::key_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::map< CString,std::vector< CString,std::allocator< CString > > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_PyMStringVString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::map< CString,VCString > *arg1 = (std::map< CString,VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__mapT_CString_VCString_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_PyMStringVString" "', argument " "1"" of type '" "std::map< CString,VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::map< CString,VCString > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *PyMStringVString_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__mapT_CString_VCString_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *PyMStringVString_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_MCString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  MCString *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_MCString", 0, 0, 0)) SWIG_fail;
  result = (MCString *)new MCString();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_MCString, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_MCString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  MCString *arg1 = (MCString *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_MCString, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_MCString" "', argument " "1"" of type '" "MCString *""'"); 
  }
  arg1 = reinterpret_cast< MCString * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *MCString_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_MCString, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *MCString_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_PyModulesVector_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_iterator" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_CModule_Sm__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___nonzero__" "', argument " "1"" of type '" "std::vector< CModule * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = (bool)std_vector_Sl_CModule_Sm__Sg____nonzero__((std::vector< CModule * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___bool__" "', argument " "1"" of type '" "std::vector< CModule * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = (bool)std_vector_Sl_CModule_Sm__Sg____bool__((std::vector< CModule * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CModule * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___len__" "', argument " "1"" of type '" "std::vector< CModule * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = std_vector_Sl_CModule_Sm__Sg____len__((std::vector< CModule * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___getslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::difference_type arg2 ;
  std::vector< CModule * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  std::vector< CModule *,std::allocator< CModule * > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyModulesVector___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___getslice__" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PyModulesVector___getslice__" "', argument " "2"" of type '" "std::vector< CModule * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CModule * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PyModulesVector___getslice__" "', argument " "3"" of type '" "std::vector< CModule * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CModule * >::difference_type >(val3);
  try {
    result = (std::vector< CModule *,std::allocator< CModule * > > *)std_vector_Sl_CModule_Sm__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CModule_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::difference_type arg2 ;
  std::vector< CModule * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___setslice__" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PyModulesVector___setslice__" "', argument " "2"" of type '" "std::vector< CModule * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CModule * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PyModulesVector___setslice__" "', argument " "3"" of type '" "std::vector< CModule * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CModule * >::difference_type >(val3);
  try {
    std_vector_Sl_CModule_Sm__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::difference_type arg2 ;
  std::vector< CModule * >::difference_type arg3 ;
  std::vector< CModule *,std::allocator< CModule * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___setslice__" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PyModulesVector___setslice__" "', argument " "2"" of type '" "std::vector< CModule * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CModule * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PyModulesVector___setslice__" "', argument " "3"" of type '" "std::vector< CModule * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CModule * >::difference_type >(val3);
  {
    std::vector< CModule*,std::allocator< CModule * > > *ptr = (std::vector< CModule*,std::allocator< CModule * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "PyModulesVector___setslice__" "', argument " "4"" of type '" "std::vector< CModule *,std::allocator< CModule * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyModulesVector___setslice__" "', argument " "4"" of type '" "std::vector< CModule *,std::allocator< CModule * > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_CModule_Sm__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< CModule *,std::allocator< CModule * > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PyModulesVector___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_PyModulesVector___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_PyModulesVector___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PyModulesVector___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CModule * >::__setslice__(std::vector< CModule * >::difference_type,std::vector< CModule * >::difference_type)\n"
    "    std::vector< CModule * >::__setslice__(std::vector< CModule * >::difference_type,std::vector< CModule * >::difference_type,std::vector< CModule *,std::allocator< CModule * > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___delslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::difference_type arg2 ;
  std::vector< CModule * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyModulesVector___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___delslice__" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PyModulesVector___delslice__" "', argument " "2"" of type '" "std::vector< CModule * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CModule * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PyModulesVector___delslice__" "', argument " "3"" of type '" "std::vector< CModule * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CModule * >::difference_type >(val3);
  try {
    std_vector_Sl_CModule_Sm__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___delitem__" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PyModulesVector___delitem__" "', argument " "2"" of type '" "std::vector< CModule * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CModule * >::difference_type >(val2);
  try {
    std_vector_Sl_CModule_Sm__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< CModule *,std::allocator< CModule * > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___getitem__" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    result = (std::vector< CModule *,std::allocator< CModule * > > *)std_vector_Sl_CModule_Sm__Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CModule_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  std::vector< CModule *,std::allocator< CModule * > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___setitem__" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  {
    std::vector< CModule*,std::allocator< CModule * > > *ptr = (std::vector< CModule*,std::allocator< CModule * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "PyModulesVector___setitem__" "', argument " "3"" of type '" "std::vector< CModule *,std::allocator< CModule * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyModulesVector___setitem__" "', argument " "3"" of type '" "std::vector< CModule *,std::allocator< CModule * > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_CModule_Sm__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< CModule *,std::allocator< CModule * > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___setitem__" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CModule_Sm__Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___delitem__" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CModule_Sm__Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PyModulesVector___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_PyModulesVector___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_PyModulesVector___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PyModulesVector___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CModule * >::__delitem__(std::vector< CModule * >::difference_type)\n"
    "    std::vector< CModule * >::__delitem__(SWIGPY_SLICEOBJECT *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< CModule * >::value_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___getitem__" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PyModulesVector___getitem__" "', argument " "2"" of type '" "std::vector< CModule * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CModule * >::difference_type >(val2);
  try {
    result = (std::vector< CModule * >::value_type)std_vector_Sl_CModule_Sm__Sg____getitem____SWIG_1(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModule, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PyModulesVector___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_PyModulesVector___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_PyModulesVector___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PyModulesVector___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CModule * >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CModule * >::__getitem__(std::vector< CModule * >::difference_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::difference_type arg2 ;
  std::vector< CModule * >::value_type arg3 = (std::vector< CModule * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector___setitem__" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PyModulesVector___setitem__" "', argument " "2"" of type '" "std::vector< CModule * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CModule * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "PyModulesVector___setitem__" "', argument " "3"" of type '" "std::vector< CModule * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CModule * >::value_type >(argp3);
  try {
    std_vector_Sl_CModule_Sm__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PyModulesVector___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_PyModulesVector___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_PyModulesVector___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CModule, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_PyModulesVector___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PyModulesVector___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CModule * >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< CModule *,std::allocator< CModule * > > const &)\n"
    "    std::vector< CModule * >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CModule * >::__setitem__(std::vector< CModule * >::difference_type,std::vector< CModule * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CModule * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_pop" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  try {
    result = (std::vector< CModule * >::value_type)std_vector_Sl_CModule_Sm__Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModule, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::value_type arg2 = (std::vector< CModule * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyModulesVector_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_append" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyModulesVector_append" "', argument " "2"" of type '" "std::vector< CModule * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CModule * >::value_type >(argp2);
  std_vector_Sl_CModule_Sm__Sg__append(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_PyModulesVector__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< CModule * > *)new std::vector< CModule * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CModule_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_PyModulesVector__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< CModule * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< CModule*,std::allocator< CModule * > > *ptr = (std::vector< CModule*,std::allocator< CModule * > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_PyModulesVector" "', argument " "1"" of type '" "std::vector< CModule * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_PyModulesVector" "', argument " "1"" of type '" "std::vector< CModule * > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< CModule * > *)new std::vector< CModule * >((std::vector< CModule * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CModule_p_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_empty" "', argument " "1"" of type '" "std::vector< CModule * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = (bool)((std::vector< CModule * > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CModule * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_size" "', argument " "1"" of type '" "std::vector< CModule * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = ((std::vector< CModule * > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyModulesVector_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_swap" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_CModule_p_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyModulesVector_swap" "', argument " "2"" of type '" "std::vector< CModule * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PyModulesVector_swap" "', argument " "2"" of type '" "std::vector< CModule * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CModule * > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CModule * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_begin" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CModule * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CModule * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_end" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CModule * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CModule * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_rbegin" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CModule * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CModule * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_rend" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CModule * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_clear" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< CModule * > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_get_allocator" "', argument " "1"" of type '" "std::vector< CModule * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = ((std::vector< CModule * > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< CModule * >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_CModule_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_PyModulesVector__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< CModule * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_PyModulesVector" "', argument " "1"" of type '" "std::vector< CModule * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CModule * >::size_type >(val1);
  result = (std::vector< CModule * > *)new std::vector< CModule * >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CModule_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_pop_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_pop_back" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_resize" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PyModulesVector_resize" "', argument " "2"" of type '" "std::vector< CModule * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CModule * >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< CModule * >::iterator result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_erase" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector_erase" "', argument " "2"" of type '" "std::vector< CModule * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector_erase" "', argument " "2"" of type '" "std::vector< CModule * >::iterator""'");
    }
  }
  result = std_vector_Sl_CModule_Sm__Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CModule * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::iterator arg2 ;
  std::vector< CModule * >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::vector< CModule * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_erase" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector_erase" "', argument " "2"" of type '" "std::vector< CModule * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector_erase" "', argument " "2"" of type '" "std::vector< CModule * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector_erase" "', argument " "3"" of type '" "std::vector< CModule * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector_erase" "', argument " "3"" of type '" "std::vector< CModule * >::iterator""'");
    }
  }
  result = std_vector_Sl_CModule_Sm__Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CModule * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PyModulesVector_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_PyModulesVector_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_PyModulesVector_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PyModulesVector_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CModule * >::erase(std::vector< CModule * >::iterator)\n"
    "    std::vector< CModule * >::erase(std::vector< CModule * >::iterator,std::vector< CModule * >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_PyModulesVector__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * >::size_type arg1 ;
  std::vector< CModule * >::value_type arg2 = (std::vector< CModule * >::value_type) 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::vector< CModule * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_PyModulesVector" "', argument " "1"" of type '" "std::vector< CModule * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CModule * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_PyModulesVector" "', argument " "2"" of type '" "std::vector< CModule * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CModule * >::value_type >(argp2);
  result = (std::vector< CModule * > *)new std::vector< CModule * >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CModule_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_PyModulesVector(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_PyModulesVector", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_PyModulesVector__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_PyModulesVector__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_PyModulesVector__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CModule, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_PyModulesVector__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_PyModulesVector'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CModule * >::vector()\n"
    "    std::vector< CModule * >::vector(std::vector< CModule * > const &)\n"
    "    std::vector< CModule * >::vector(std::vector< CModule * >::size_type)\n"
    "    std::vector< CModule * >::vector(std::vector< CModule * >::size_type,std::vector< CModule * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_push_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::value_type arg2 = (std::vector< CModule * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyModulesVector_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_push_back" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PyModulesVector_push_back" "', argument " "2"" of type '" "std::vector< CModule * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CModule * >::value_type >(argp2);
  (arg1)->push_back(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CModule * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_front" "', argument " "1"" of type '" "std::vector< CModule * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = (std::vector< CModule * >::value_type)((std::vector< CModule * > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModule, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CModule * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_back" "', argument " "1"" of type '" "std::vector< CModule * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = (std::vector< CModule * >::value_type)((std::vector< CModule * > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModule, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_assign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::size_type arg2 ;
  std::vector< CModule * >::value_type arg3 = (std::vector< CModule * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyModulesVector_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_assign" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PyModulesVector_assign" "', argument " "2"" of type '" "std::vector< CModule * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CModule * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "PyModulesVector_assign" "', argument " "3"" of type '" "std::vector< CModule * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CModule * >::value_type >(argp3);
  (arg1)->assign(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::size_type arg2 ;
  std::vector< CModule * >::value_type arg3 = (std::vector< CModule * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_resize" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PyModulesVector_resize" "', argument " "2"" of type '" "std::vector< CModule * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CModule * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "PyModulesVector_resize" "', argument " "3"" of type '" "std::vector< CModule * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CModule * >::value_type >(argp3);
  (arg1)->resize(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PyModulesVector_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_PyModulesVector_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CModule, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_PyModulesVector_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PyModulesVector_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CModule * >::resize(std::vector< CModule * >::size_type)\n"
    "    std::vector< CModule * >::resize(std::vector< CModule * >::size_type,std::vector< CModule * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::iterator arg2 ;
  std::vector< CModule * >::value_type arg3 = (std::vector< CModule * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  std::vector< CModule * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_insert" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector_insert" "', argument " "2"" of type '" "std::vector< CModule * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector_insert" "', argument " "2"" of type '" "std::vector< CModule * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "PyModulesVector_insert" "', argument " "3"" of type '" "std::vector< CModule * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CModule * >::value_type >(argp3);
  result = std_vector_Sl_CModule_Sm__Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CModule * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::iterator arg2 ;
  std::vector< CModule * >::size_type arg3 ;
  std::vector< CModule * >::value_type arg4 = (std::vector< CModule * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_insert" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector_insert" "', argument " "2"" of type '" "std::vector< CModule * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "PyModulesVector_insert" "', argument " "2"" of type '" "std::vector< CModule * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PyModulesVector_insert" "', argument " "3"" of type '" "std::vector< CModule * >::size_type""'");
  } 
  arg3 = static_cast< std::vector< CModule * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "PyModulesVector_insert" "', argument " "4"" of type '" "std::vector< CModule * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector< CModule * >::value_type >(argp4);
  std_vector_Sl_CModule_Sm__Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PyModulesVector_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CModule, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_PyModulesVector_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CModule*,std::allocator< CModule * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CModule * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CModule, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_PyModulesVector_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PyModulesVector_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CModule * >::insert(std::vector< CModule * >::iterator,std::vector< CModule * >::value_type)\n"
    "    std::vector< CModule * >::insert(std::vector< CModule * >::iterator,std::vector< CModule * >::size_type,std::vector< CModule * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_reserve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  std::vector< CModule * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "PyModulesVector_reserve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_reserve" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PyModulesVector_reserve" "', argument " "2"" of type '" "std::vector< CModule * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CModule * >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PyModulesVector_capacity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CModule * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PyModulesVector_capacity" "', argument " "1"" of type '" "std::vector< CModule * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  result = ((std::vector< CModule * > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_PyModulesVector(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CModule * > *arg1 = (std::vector< CModule * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CModule_p_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_PyModulesVector" "', argument " "1"" of type '" "std::vector< CModule * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CModule * > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *PyModulesVector_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_CModule_p_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *PyModulesVector_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_VListeners_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_iterator" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_CListener_Sm__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___nonzero__" "', argument " "1"" of type '" "std::vector< CListener * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = (bool)std_vector_Sl_CListener_Sm__Sg____nonzero__((std::vector< CListener * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___bool__" "', argument " "1"" of type '" "std::vector< CListener * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = (bool)std_vector_Sl_CListener_Sm__Sg____bool__((std::vector< CListener * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CListener * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___len__" "', argument " "1"" of type '" "std::vector< CListener * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = std_vector_Sl_CListener_Sm__Sg____len__((std::vector< CListener * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___getslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::difference_type arg2 ;
  std::vector< CListener * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  std::vector< CListener *,std::allocator< CListener * > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VListeners___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___getslice__" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VListeners___getslice__" "', argument " "2"" of type '" "std::vector< CListener * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CListener * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VListeners___getslice__" "', argument " "3"" of type '" "std::vector< CListener * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CListener * >::difference_type >(val3);
  try {
    result = (std::vector< CListener *,std::allocator< CListener * > > *)std_vector_Sl_CListener_Sm__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CListener_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::difference_type arg2 ;
  std::vector< CListener * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___setslice__" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VListeners___setslice__" "', argument " "2"" of type '" "std::vector< CListener * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CListener * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VListeners___setslice__" "', argument " "3"" of type '" "std::vector< CListener * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CListener * >::difference_type >(val3);
  try {
    std_vector_Sl_CListener_Sm__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::difference_type arg2 ;
  std::vector< CListener * >::difference_type arg3 ;
  std::vector< CListener *,std::allocator< CListener * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___setslice__" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VListeners___setslice__" "', argument " "2"" of type '" "std::vector< CListener * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CListener * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VListeners___setslice__" "', argument " "3"" of type '" "std::vector< CListener * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CListener * >::difference_type >(val3);
  {
    std::vector< CListener*,std::allocator< CListener * > > *ptr = (std::vector< CListener*,std::allocator< CListener * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VListeners___setslice__" "', argument " "4"" of type '" "std::vector< CListener *,std::allocator< CListener * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VListeners___setslice__" "', argument " "4"" of type '" "std::vector< CListener *,std::allocator< CListener * > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_CListener_Sm__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< CListener *,std::allocator< CListener * > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VListeners___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_VListeners___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VListeners___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VListeners___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CListener * >::__setslice__(std::vector< CListener * >::difference_type,std::vector< CListener * >::difference_type)\n"
    "    std::vector< CListener * >::__setslice__(std::vector< CListener * >::difference_type,std::vector< CListener * >::difference_type,std::vector< CListener *,std::allocator< CListener * > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VListeners___delslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::difference_type arg2 ;
  std::vector< CListener * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VListeners___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___delslice__" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VListeners___delslice__" "', argument " "2"" of type '" "std::vector< CListener * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CListener * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VListeners___delslice__" "', argument " "3"" of type '" "std::vector< CListener * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CListener * >::difference_type >(val3);
  try {
    std_vector_Sl_CListener_Sm__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___delitem__" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VListeners___delitem__" "', argument " "2"" of type '" "std::vector< CListener * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CListener * >::difference_type >(val2);
  try {
    std_vector_Sl_CListener_Sm__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< CListener *,std::allocator< CListener * > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___getitem__" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    result = (std::vector< CListener *,std::allocator< CListener * > > *)std_vector_Sl_CListener_Sm__Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CListener_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  std::vector< CListener *,std::allocator< CListener * > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___setitem__" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  {
    std::vector< CListener*,std::allocator< CListener * > > *ptr = (std::vector< CListener*,std::allocator< CListener * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VListeners___setitem__" "', argument " "3"" of type '" "std::vector< CListener *,std::allocator< CListener * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VListeners___setitem__" "', argument " "3"" of type '" "std::vector< CListener *,std::allocator< CListener * > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_CListener_Sm__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< CListener *,std::allocator< CListener * > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___setitem__" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CListener_Sm__Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___delitem__" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CListener_Sm__Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VListeners___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VListeners___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VListeners___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VListeners___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CListener * >::__delitem__(std::vector< CListener * >::difference_type)\n"
    "    std::vector< CListener * >::__delitem__(SWIGPY_SLICEOBJECT *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VListeners___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< CListener * >::value_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___getitem__" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VListeners___getitem__" "', argument " "2"" of type '" "std::vector< CListener * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CListener * >::difference_type >(val2);
  try {
    result = (std::vector< CListener * >::value_type)std_vector_Sl_CListener_Sm__Sg____getitem____SWIG_1(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CListener, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VListeners___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VListeners___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VListeners___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VListeners___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CListener * >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CListener * >::__getitem__(std::vector< CListener * >::difference_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VListeners___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::difference_type arg2 ;
  std::vector< CListener * >::value_type arg3 = (std::vector< CListener * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners___setitem__" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VListeners___setitem__" "', argument " "2"" of type '" "std::vector< CListener * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CListener * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VListeners___setitem__" "', argument " "3"" of type '" "std::vector< CListener * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CListener * >::value_type >(argp3);
  try {
    std_vector_Sl_CListener_Sm__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VListeners___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VListeners___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VListeners___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CListener, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VListeners___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VListeners___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CListener * >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< CListener *,std::allocator< CListener * > > const &)\n"
    "    std::vector< CListener * >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CListener * >::__setitem__(std::vector< CListener * >::difference_type,std::vector< CListener * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VListeners_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CListener * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_pop" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  try {
    result = (std::vector< CListener * >::value_type)std_vector_Sl_CListener_Sm__Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CListener, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::value_type arg2 = (std::vector< CListener * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VListeners_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_append" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VListeners_append" "', argument " "2"" of type '" "std::vector< CListener * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CListener * >::value_type >(argp2);
  std_vector_Sl_CListener_Sm__Sg__append(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VListeners__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< CListener * > *)new std::vector< CListener * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CListener_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VListeners__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< CListener * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< CListener*,std::allocator< CListener * > > *ptr = (std::vector< CListener*,std::allocator< CListener * > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_VListeners" "', argument " "1"" of type '" "std::vector< CListener * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VListeners" "', argument " "1"" of type '" "std::vector< CListener * > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< CListener * > *)new std::vector< CListener * >((std::vector< CListener * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CListener_p_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_empty" "', argument " "1"" of type '" "std::vector< CListener * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = (bool)((std::vector< CListener * > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CListener * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_size" "', argument " "1"" of type '" "std::vector< CListener * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = ((std::vector< CListener * > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VListeners_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_swap" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_CListener_p_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VListeners_swap" "', argument " "2"" of type '" "std::vector< CListener * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VListeners_swap" "', argument " "2"" of type '" "std::vector< CListener * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CListener * > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CListener * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_begin" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CListener * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CListener * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_end" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CListener * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CListener * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_rbegin" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CListener * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CListener * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_rend" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CListener * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_clear" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< CListener * > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_get_allocator" "', argument " "1"" of type '" "std::vector< CListener * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = ((std::vector< CListener * > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< CListener * >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_CListener_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VListeners__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< CListener * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VListeners" "', argument " "1"" of type '" "std::vector< CListener * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CListener * >::size_type >(val1);
  result = (std::vector< CListener * > *)new std::vector< CListener * >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CListener_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_pop_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_pop_back" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_resize" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VListeners_resize" "', argument " "2"" of type '" "std::vector< CListener * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CListener * >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< CListener * >::iterator result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_erase" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners_erase" "', argument " "2"" of type '" "std::vector< CListener * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners_erase" "', argument " "2"" of type '" "std::vector< CListener * >::iterator""'");
    }
  }
  result = std_vector_Sl_CListener_Sm__Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CListener * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::iterator arg2 ;
  std::vector< CListener * >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::vector< CListener * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_erase" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners_erase" "', argument " "2"" of type '" "std::vector< CListener * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners_erase" "', argument " "2"" of type '" "std::vector< CListener * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners_erase" "', argument " "3"" of type '" "std::vector< CListener * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners_erase" "', argument " "3"" of type '" "std::vector< CListener * >::iterator""'");
    }
  }
  result = std_vector_Sl_CListener_Sm__Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CListener * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VListeners_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_VListeners_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_VListeners_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VListeners_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CListener * >::erase(std::vector< CListener * >::iterator)\n"
    "    std::vector< CListener * >::erase(std::vector< CListener * >::iterator,std::vector< CListener * >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_VListeners__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * >::size_type arg1 ;
  std::vector< CListener * >::value_type arg2 = (std::vector< CListener * >::value_type) 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::vector< CListener * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VListeners" "', argument " "1"" of type '" "std::vector< CListener * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CListener * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_VListeners" "', argument " "2"" of type '" "std::vector< CListener * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CListener * >::value_type >(argp2);
  result = (std::vector< CListener * > *)new std::vector< CListener * >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CListener_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VListeners(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_VListeners", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_VListeners__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_VListeners__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_VListeners__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CListener, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_VListeners__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_VListeners'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CListener * >::vector()\n"
    "    std::vector< CListener * >::vector(std::vector< CListener * > const &)\n"
    "    std::vector< CListener * >::vector(std::vector< CListener * >::size_type)\n"
    "    std::vector< CListener * >::vector(std::vector< CListener * >::size_type,std::vector< CListener * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VListeners_push_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::value_type arg2 = (std::vector< CListener * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VListeners_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_push_back" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VListeners_push_back" "', argument " "2"" of type '" "std::vector< CListener * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CListener * >::value_type >(argp2);
  (arg1)->push_back(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CListener * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_front" "', argument " "1"" of type '" "std::vector< CListener * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = (std::vector< CListener * >::value_type)((std::vector< CListener * > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CListener, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CListener * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_back" "', argument " "1"" of type '" "std::vector< CListener * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = (std::vector< CListener * >::value_type)((std::vector< CListener * > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CListener, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_assign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::size_type arg2 ;
  std::vector< CListener * >::value_type arg3 = (std::vector< CListener * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VListeners_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_assign" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VListeners_assign" "', argument " "2"" of type '" "std::vector< CListener * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CListener * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VListeners_assign" "', argument " "3"" of type '" "std::vector< CListener * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CListener * >::value_type >(argp3);
  (arg1)->assign(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::size_type arg2 ;
  std::vector< CListener * >::value_type arg3 = (std::vector< CListener * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_resize" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VListeners_resize" "', argument " "2"" of type '" "std::vector< CListener * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CListener * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VListeners_resize" "', argument " "3"" of type '" "std::vector< CListener * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CListener * >::value_type >(argp3);
  (arg1)->resize(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VListeners_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VListeners_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CListener, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VListeners_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VListeners_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CListener * >::resize(std::vector< CListener * >::size_type)\n"
    "    std::vector< CListener * >::resize(std::vector< CListener * >::size_type,std::vector< CListener * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VListeners_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::iterator arg2 ;
  std::vector< CListener * >::value_type arg3 = (std::vector< CListener * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  std::vector< CListener * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_insert" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners_insert" "', argument " "2"" of type '" "std::vector< CListener * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners_insert" "', argument " "2"" of type '" "std::vector< CListener * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VListeners_insert" "', argument " "3"" of type '" "std::vector< CListener * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CListener * >::value_type >(argp3);
  result = std_vector_Sl_CListener_Sm__Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CListener * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::iterator arg2 ;
  std::vector< CListener * >::size_type arg3 ;
  std::vector< CListener * >::value_type arg4 = (std::vector< CListener * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_insert" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners_insert" "', argument " "2"" of type '" "std::vector< CListener * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VListeners_insert" "', argument " "2"" of type '" "std::vector< CListener * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VListeners_insert" "', argument " "3"" of type '" "std::vector< CListener * >::size_type""'");
  } 
  arg3 = static_cast< std::vector< CListener * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VListeners_insert" "', argument " "4"" of type '" "std::vector< CListener * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector< CListener * >::value_type >(argp4);
  std_vector_Sl_CListener_Sm__Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VListeners_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CListener, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VListeners_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CListener*,std::allocator< CListener * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CListener * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CListener, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VListeners_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VListeners_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CListener * >::insert(std::vector< CListener * >::iterator,std::vector< CListener * >::value_type)\n"
    "    std::vector< CListener * >::insert(std::vector< CListener * >::iterator,std::vector< CListener * >::size_type,std::vector< CListener * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VListeners_reserve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  std::vector< CListener * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VListeners_reserve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_reserve" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VListeners_reserve" "', argument " "2"" of type '" "std::vector< CListener * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CListener * >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VListeners_capacity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CListener * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VListeners_capacity" "', argument " "1"" of type '" "std::vector< CListener * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  result = ((std::vector< CListener * > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_VListeners(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CListener * > *arg1 = (std::vector< CListener * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CListener_p_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VListeners" "', argument " "1"" of type '" "std::vector< CListener * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CListener * > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *VListeners_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_CListener_p_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *VListeners_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_BufLines_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_iterator" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  result = (swig::SwigPyIterator *)std_deque_Sl_CBufLine_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___nonzero__" "', argument " "1"" of type '" "std::deque< CBufLine > const *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  result = (bool)std_deque_Sl_CBufLine_Sg____nonzero__((std::deque< CBufLine > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___bool__" "', argument " "1"" of type '" "std::deque< CBufLine > const *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  result = (bool)std_deque_Sl_CBufLine_Sg____bool__((std::deque< CBufLine > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___len__" "', argument " "1"" of type '" "std::deque< CBufLine > const *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  result = std_deque_Sl_CBufLine_Sg____len__((std::deque< CBufLine > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___getslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::difference_type arg2 ;
  std::deque< CBufLine >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  std::deque< CBufLine,std::allocator< CBufLine > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "BufLines___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___getslice__" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "BufLines___getslice__" "', argument " "2"" of type '" "std::deque< CBufLine >::difference_type""'");
  } 
  arg2 = static_cast< std::deque< CBufLine >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "BufLines___getslice__" "', argument " "3"" of type '" "std::deque< CBufLine >::difference_type""'");
  } 
  arg3 = static_cast< std::deque< CBufLine >::difference_type >(val3);
  try {
    result = (std::deque< CBufLine,std::allocator< CBufLine > > *)std_deque_Sl_CBufLine_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__dequeT_CBufLine_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::difference_type arg2 ;
  std::deque< CBufLine >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___setslice__" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "BufLines___setslice__" "', argument " "2"" of type '" "std::deque< CBufLine >::difference_type""'");
  } 
  arg2 = static_cast< std::deque< CBufLine >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "BufLines___setslice__" "', argument " "3"" of type '" "std::deque< CBufLine >::difference_type""'");
  } 
  arg3 = static_cast< std::deque< CBufLine >::difference_type >(val3);
  try {
    std_deque_Sl_CBufLine_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::difference_type arg2 ;
  std::deque< CBufLine >::difference_type arg3 ;
  std::deque< CBufLine,std::allocator< CBufLine > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___setslice__" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "BufLines___setslice__" "', argument " "2"" of type '" "std::deque< CBufLine >::difference_type""'");
  } 
  arg2 = static_cast< std::deque< CBufLine >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "BufLines___setslice__" "', argument " "3"" of type '" "std::deque< CBufLine >::difference_type""'");
  } 
  arg3 = static_cast< std::deque< CBufLine >::difference_type >(val3);
  {
    std::deque< CBufLine,std::allocator< CBufLine > > *ptr = (std::deque< CBufLine,std::allocator< CBufLine > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "BufLines___setslice__" "', argument " "4"" of type '" "std::deque< CBufLine,std::allocator< CBufLine > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "BufLines___setslice__" "', argument " "4"" of type '" "std::deque< CBufLine,std::allocator< CBufLine > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_deque_Sl_CBufLine_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::deque< CBufLine,std::allocator< CBufLine > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "BufLines___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_BufLines___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_BufLines___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'BufLines___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::deque< CBufLine >::__setslice__(std::deque< CBufLine >::difference_type,std::deque< CBufLine >::difference_type)\n"
    "    std::deque< CBufLine >::__setslice__(std::deque< CBufLine >::difference_type,std::deque< CBufLine >::difference_type,std::deque< CBufLine,std::allocator< CBufLine > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_BufLines___delslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::difference_type arg2 ;
  std::deque< CBufLine >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "BufLines___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___delslice__" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "BufLines___delslice__" "', argument " "2"" of type '" "std::deque< CBufLine >::difference_type""'");
  } 
  arg2 = static_cast< std::deque< CBufLine >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "BufLines___delslice__" "', argument " "3"" of type '" "std::deque< CBufLine >::difference_type""'");
  } 
  arg3 = static_cast< std::deque< CBufLine >::difference_type >(val3);
  try {
    std_deque_Sl_CBufLine_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___delitem__" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "BufLines___delitem__" "', argument " "2"" of type '" "std::deque< CBufLine >::difference_type""'");
  } 
  arg2 = static_cast< std::deque< CBufLine >::difference_type >(val2);
  try {
    std_deque_Sl_CBufLine_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::deque< CBufLine,std::allocator< CBufLine > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___getitem__" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    result = (std::deque< CBufLine,std::allocator< CBufLine > > *)std_deque_Sl_CBufLine_Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__dequeT_CBufLine_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  std::deque< CBufLine,std::allocator< CBufLine > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___setitem__" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  {
    std::deque< CBufLine,std::allocator< CBufLine > > *ptr = (std::deque< CBufLine,std::allocator< CBufLine > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "BufLines___setitem__" "', argument " "3"" of type '" "std::deque< CBufLine,std::allocator< CBufLine > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "BufLines___setitem__" "', argument " "3"" of type '" "std::deque< CBufLine,std::allocator< CBufLine > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_deque_Sl_CBufLine_Sg____setitem____SWIG_0(arg1,arg2,(std::deque< CBufLine,std::allocator< CBufLine > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___setitem__" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_deque_Sl_CBufLine_Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___delitem__" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_deque_Sl_CBufLine_Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "BufLines___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_BufLines___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_BufLines___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'BufLines___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::deque< CBufLine >::__delitem__(std::deque< CBufLine >::difference_type)\n"
    "    std::deque< CBufLine >::__delitem__(SWIGPY_SLICEOBJECT *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_BufLines___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::deque< CBufLine >::value_type *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___getitem__" "', argument " "1"" of type '" "std::deque< CBufLine > const *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "BufLines___getitem__" "', argument " "2"" of type '" "std::deque< CBufLine >::difference_type""'");
  } 
  arg2 = static_cast< std::deque< CBufLine >::difference_type >(val2);
  try {
    result = (std::deque< CBufLine >::value_type *) &std_deque_Sl_CBufLine_Sg____getitem____SWIG_1((std::deque< CBufLine > const *)arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBufLine, 0 |  0 );
  (void)swig::container_owner<swig::traits<std::deque< CBufLine >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "BufLines___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_BufLines___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_BufLines___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'BufLines___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::deque< CBufLine >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::deque< CBufLine >::__getitem__(std::deque< CBufLine >::difference_type) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_BufLines___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::difference_type arg2 ;
  std::deque< CBufLine >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines___setitem__" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "BufLines___setitem__" "', argument " "2"" of type '" "std::deque< CBufLine >::difference_type""'");
  } 
  arg2 = static_cast< std::deque< CBufLine >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CBufLine,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "BufLines___setitem__" "', argument " "3"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "BufLines___setitem__" "', argument " "3"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  arg3 = reinterpret_cast< std::deque< CBufLine >::value_type * >(argp3);
  try {
    std_deque_Sl_CBufLine_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(CBufLine const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "BufLines___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_BufLines___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_BufLines___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CBufLine, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_BufLines___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'BufLines___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::deque< CBufLine >::__setitem__(SWIGPY_SLICEOBJECT *,std::deque< CBufLine,std::allocator< CBufLine > > const &)\n"
    "    std::deque< CBufLine >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::deque< CBufLine >::__setitem__(std::deque< CBufLine >::difference_type,std::deque< CBufLine >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_BufLines_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::deque< CBufLine >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_pop" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  try {
    result = std_deque_Sl_CBufLine_Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj((new std::deque< CBufLine >::value_type(result)), SWIGTYPE_p_CBufLine, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "BufLines_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_append" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CBufLine,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "BufLines_append" "', argument " "2"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "BufLines_append" "', argument " "2"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::deque< CBufLine >::value_type * >(argp2);
  std_deque_Sl_CBufLine_Sg__append(arg1,(CBufLine const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_BufLines__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::deque< CBufLine > *)new std::deque< CBufLine >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__dequeT_CBufLine_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_BufLines__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::deque< CBufLine > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::deque< CBufLine,std::allocator< CBufLine > > *ptr = (std::deque< CBufLine,std::allocator< CBufLine > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_BufLines" "', argument " "1"" of type '" "std::deque< CBufLine > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_BufLines" "', argument " "1"" of type '" "std::deque< CBufLine > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::deque< CBufLine > *)new std::deque< CBufLine >((std::deque< CBufLine > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__dequeT_CBufLine_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_empty" "', argument " "1"" of type '" "std::deque< CBufLine > const *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  result = (bool)((std::deque< CBufLine > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_size" "', argument " "1"" of type '" "std::deque< CBufLine > const *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  result = ((std::deque< CBufLine > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "BufLines_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_swap" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__dequeT_CBufLine_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "BufLines_swap" "', argument " "2"" of type '" "std::deque< CBufLine > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "BufLines_swap" "', argument " "2"" of type '" "std::deque< CBufLine > &""'"); 
  }
  arg2 = reinterpret_cast< std::deque< CBufLine > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::deque< CBufLine >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_begin" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::deque< CBufLine >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::deque< CBufLine >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_end" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::deque< CBufLine >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::deque< CBufLine >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_rbegin" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::deque< CBufLine >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::deque< CBufLine >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_rend" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::deque< CBufLine >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_clear" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< CBufLine > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_get_allocator" "', argument " "1"" of type '" "std::deque< CBufLine > const *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  result = ((std::deque< CBufLine > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::deque< CBufLine >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_CBufLine_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_BufLines__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::deque< CBufLine > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_BufLines" "', argument " "1"" of type '" "std::deque< CBufLine >::size_type""'");
  } 
  arg1 = static_cast< std::deque< CBufLine >::size_type >(val1);
  result = (std::deque< CBufLine > *)new std::deque< CBufLine >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__dequeT_CBufLine_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_pop_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_pop_back" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_resize" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "BufLines_resize" "', argument " "2"" of type '" "std::deque< CBufLine >::size_type""'");
  } 
  arg2 = static_cast< std::deque< CBufLine >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::deque< CBufLine >::iterator result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_erase" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines_erase" "', argument " "2"" of type '" "std::deque< CBufLine >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines_erase" "', argument " "2"" of type '" "std::deque< CBufLine >::iterator""'");
    }
  }
  result = std_deque_Sl_CBufLine_Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::deque< CBufLine >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::iterator arg2 ;
  std::deque< CBufLine >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::deque< CBufLine >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_erase" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines_erase" "', argument " "2"" of type '" "std::deque< CBufLine >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines_erase" "', argument " "2"" of type '" "std::deque< CBufLine >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines_erase" "', argument " "3"" of type '" "std::deque< CBufLine >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines_erase" "', argument " "3"" of type '" "std::deque< CBufLine >::iterator""'");
    }
  }
  result = std_deque_Sl_CBufLine_Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::deque< CBufLine >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "BufLines_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_BufLines_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_BufLines_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'BufLines_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::deque< CBufLine >::erase(std::deque< CBufLine >::iterator)\n"
    "    std::deque< CBufLine >::erase(std::deque< CBufLine >::iterator,std::deque< CBufLine >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_BufLines__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine >::size_type arg1 ;
  std::deque< CBufLine >::value_type *arg2 = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::deque< CBufLine > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_BufLines" "', argument " "1"" of type '" "std::deque< CBufLine >::size_type""'");
  } 
  arg1 = static_cast< std::deque< CBufLine >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CBufLine,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_BufLines" "', argument " "2"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_BufLines" "', argument " "2"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::deque< CBufLine >::value_type * >(argp2);
  result = (std::deque< CBufLine > *)new std::deque< CBufLine >(arg1,(std::deque< CBufLine >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__dequeT_CBufLine_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_BufLines(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_BufLines", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_BufLines__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_BufLines__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_BufLines__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CBufLine, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_BufLines__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_BufLines'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::deque< CBufLine >::deque()\n"
    "    std::deque< CBufLine >::deque(std::deque< CBufLine > const &)\n"
    "    std::deque< CBufLine >::deque(std::deque< CBufLine >::size_type)\n"
    "    std::deque< CBufLine >::deque(std::deque< CBufLine >::size_type,std::deque< CBufLine >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_BufLines_push_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "BufLines_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_push_back" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CBufLine,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "BufLines_push_back" "', argument " "2"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "BufLines_push_back" "', argument " "2"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::deque< CBufLine >::value_type * >(argp2);
  (arg1)->push_back((std::deque< CBufLine >::value_type const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::deque< CBufLine >::value_type *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_front" "', argument " "1"" of type '" "std::deque< CBufLine > const *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  result = (std::deque< CBufLine >::value_type *) &((std::deque< CBufLine > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBufLine, 0 |  0 );
  (void)swig::container_owner<swig::traits<std::deque< CBufLine >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::deque< CBufLine >::value_type *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_back" "', argument " "1"" of type '" "std::deque< CBufLine > const *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  result = (std::deque< CBufLine >::value_type *) &((std::deque< CBufLine > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBufLine, 0 |  0 );
  (void)swig::container_owner<swig::traits<std::deque< CBufLine >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_assign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::size_type arg2 ;
  std::deque< CBufLine >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "BufLines_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_assign" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "BufLines_assign" "', argument " "2"" of type '" "std::deque< CBufLine >::size_type""'");
  } 
  arg2 = static_cast< std::deque< CBufLine >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CBufLine,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "BufLines_assign" "', argument " "3"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "BufLines_assign" "', argument " "3"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  arg3 = reinterpret_cast< std::deque< CBufLine >::value_type * >(argp3);
  (arg1)->assign(arg2,(std::deque< CBufLine >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::size_type arg2 ;
  std::deque< CBufLine >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_resize" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "BufLines_resize" "', argument " "2"" of type '" "std::deque< CBufLine >::size_type""'");
  } 
  arg2 = static_cast< std::deque< CBufLine >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CBufLine,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "BufLines_resize" "', argument " "3"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "BufLines_resize" "', argument " "3"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  arg3 = reinterpret_cast< std::deque< CBufLine >::value_type * >(argp3);
  (arg1)->resize(arg2,(std::deque< CBufLine >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "BufLines_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_BufLines_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CBufLine, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_BufLines_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'BufLines_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::deque< CBufLine >::resize(std::deque< CBufLine >::size_type)\n"
    "    std::deque< CBufLine >::resize(std::deque< CBufLine >::size_type,std::deque< CBufLine >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_BufLines_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::iterator arg2 ;
  std::deque< CBufLine >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  std::deque< CBufLine >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_insert" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines_insert" "', argument " "2"" of type '" "std::deque< CBufLine >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines_insert" "', argument " "2"" of type '" "std::deque< CBufLine >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CBufLine,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "BufLines_insert" "', argument " "3"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "BufLines_insert" "', argument " "3"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  arg3 = reinterpret_cast< std::deque< CBufLine >::value_type * >(argp3);
  result = std_deque_Sl_CBufLine_Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),(CBufLine const &)*arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::deque< CBufLine >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::iterator arg2 ;
  std::deque< CBufLine >::size_type arg3 ;
  std::deque< CBufLine >::value_type *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_insert" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines_insert" "', argument " "2"" of type '" "std::deque< CBufLine >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "BufLines_insert" "', argument " "2"" of type '" "std::deque< CBufLine >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "BufLines_insert" "', argument " "3"" of type '" "std::deque< CBufLine >::size_type""'");
  } 
  arg3 = static_cast< std::deque< CBufLine >::size_type >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CBufLine,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "BufLines_insert" "', argument " "4"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "BufLines_insert" "', argument " "4"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  arg4 = reinterpret_cast< std::deque< CBufLine >::value_type * >(argp4);
  std_deque_Sl_CBufLine_Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(CBufLine const &)*arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "BufLines_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *>(iter) != 0));
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CBufLine, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_BufLines_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::deque< CBufLine,std::allocator< CBufLine > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::deque< CBufLine >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_ConvertPtr(argv[3], 0, SWIGTYPE_p_CBufLine, SWIG_POINTER_NO_NULL | 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_BufLines_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'BufLines_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::deque< CBufLine >::insert(std::deque< CBufLine >::iterator,std::deque< CBufLine >::value_type const &)\n"
    "    std::deque< CBufLine >::insert(std::deque< CBufLine >::iterator,std::deque< CBufLine >::size_type,std::deque< CBufLine >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_BufLines_pop_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_pop_front" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  (arg1)->pop_front();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BufLines_push_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  std::deque< CBufLine >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "BufLines_push_front", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BufLines_push_front" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CBufLine,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "BufLines_push_front" "', argument " "2"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "BufLines_push_front" "', argument " "2"" of type '" "std::deque< CBufLine >::value_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::deque< CBufLine >::value_type * >(argp2);
  (arg1)->push_front((std::deque< CBufLine >::value_type const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_BufLines(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::deque< CBufLine > *arg1 = (std::deque< CBufLine > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__dequeT_CBufLine_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_BufLines" "', argument " "1"" of type '" "std::deque< CBufLine > *""'"); 
  }
  arg1 = reinterpret_cast< std::deque< CBufLine > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *BufLines_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__dequeT_CBufLine_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *BufLines_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_VVString_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_iterator" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_VCString_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___nonzero__" "', argument " "1"" of type '" "std::vector< VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = (bool)std_vector_Sl_VCString_Sg____nonzero__((std::vector< std::vector< CString,std::allocator< CString > > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___bool__" "', argument " "1"" of type '" "std::vector< VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = (bool)std_vector_Sl_VCString_Sg____bool__((std::vector< std::vector< CString,std::allocator< CString > > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::vector< CString,std::allocator< CString > > >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___len__" "', argument " "1"" of type '" "std::vector< VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = std_vector_Sl_VCString_Sg____len__((std::vector< std::vector< CString,std::allocator< CString > > > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___getslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::difference_type arg2 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VVString___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___getslice__" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VVString___getslice__" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VVString___getslice__" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::difference_type >(val3);
  try {
    result = (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *)std_vector_Sl_VCString_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_VCString_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::difference_type arg2 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___setslice__" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VVString___setslice__" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VVString___setslice__" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::difference_type >(val3);
  try {
    std_vector_Sl_VCString_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::difference_type arg2 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::difference_type arg3 ;
  std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___setslice__" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VVString___setslice__" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VVString___setslice__" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::difference_type >(val3);
  {
    std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *ptr = (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VVString___setslice__" "', argument " "4"" of type '" "std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VVString___setslice__" "', argument " "4"" of type '" "std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_VCString_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VVString___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_VVString___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VVString___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VVString___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< VCString >::__setslice__(std::vector< std::vector< CString,std::allocator< CString > > >::difference_type,std::vector< std::vector< CString,std::allocator< CString > > >::difference_type)\n"
    "    std::vector< VCString >::__setslice__(std::vector< std::vector< CString,std::allocator< CString > > >::difference_type,std::vector< std::vector< CString,std::allocator< CString > > >::difference_type,std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VVString___delslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::difference_type arg2 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VVString___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___delslice__" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VVString___delslice__" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VVString___delslice__" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::difference_type >(val3);
  try {
    std_vector_Sl_VCString_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___delitem__" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VVString___delitem__" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::difference_type >(val2);
  try {
    std_vector_Sl_VCString_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___getitem__" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    result = (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *)std_vector_Sl_VCString_Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_VCString_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___setitem__" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  {
    std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *ptr = (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VVString___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VVString___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_VCString_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___setitem__" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_VCString_Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___delitem__" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_VCString_Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VVString___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VVString___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VVString___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VVString___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< VCString >::__delitem__(std::vector< std::vector< CString,std::allocator< CString > > >::difference_type)\n"
    "    std::vector< VCString >::__delitem__(SWIGPY_SLICEOBJECT *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VVString___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::value_type *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___getitem__" "', argument " "1"" of type '" "std::vector< VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VVString___getitem__" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::difference_type >(val2);
  try {
    result = (std::vector< std::vector< CString,std::allocator< CString > > >::value_type *) &std_vector_Sl_VCString_Sg____getitem____SWIG_1((std::vector< std::vector< CString,std::allocator< CString > > > const *)arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = swig::from(static_cast< std::vector< CString,std::allocator< CString > > >(*result));
  (void)swig::container_owner<swig::traits<std::vector< std::vector< CString,std::allocator< CString > > >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VVString___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VVString___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VVString___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VVString___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< VCString >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< VCString >::__getitem__(std::vector< std::vector< CString,std::allocator< CString > > >::difference_type) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VVString___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::difference_type arg2 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString___setitem__" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VVString___setitem__" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::difference_type >(val2);
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VVString___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VVString___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_VCString_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(std::vector< CString,std::allocator< CString > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VVString___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VVString___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VVString___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CString,std::allocator< CString > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VVString___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VVString___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< VCString >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > const &)\n"
    "    std::vector< VCString >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< VCString >::__setitem__(std::vector< std::vector< CString,std::allocator< CString > > >::difference_type,std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VVString_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::vector< CString,std::allocator< CString > > >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_pop" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  try {
    result = std_vector_Sl_VCString_Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = swig::from(static_cast< std::vector< CString,std::allocator< CString > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VVString_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_append" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VVString_append" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VVString_append" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  std_vector_Sl_VCString_Sg__append(arg1,(std::vector< CString,std::allocator< CString > > const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VVString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< VCString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< VCString > *)new std::vector< VCString >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_VCString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VVString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< VCString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *ptr = (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_VVString" "', argument " "1"" of type '" "std::vector< VCString > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VVString" "', argument " "1"" of type '" "std::vector< VCString > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< VCString > *)new std::vector< VCString >((std::vector< VCString > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_VCString_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_empty" "', argument " "1"" of type '" "std::vector< VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = (bool)((std::vector< VCString > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::vector< CString,std::allocator< CString > > >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_size" "', argument " "1"" of type '" "std::vector< VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = ((std::vector< VCString > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< VCString > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VVString_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_swap" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_VCString_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VVString_swap" "', argument " "2"" of type '" "std::vector< VCString > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VVString_swap" "', argument " "2"" of type '" "std::vector< VCString > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< VCString > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< std::vector< CString,std::allocator< CString > > >::iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_begin" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< CString,std::allocator< CString > > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< std::vector< CString,std::allocator< CString > > >::iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_end" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< CString,std::allocator< CString > > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< std::vector< CString,std::allocator< CString > > >::reverse_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_rbegin" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< CString,std::allocator< CString > > >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< std::vector< CString,std::allocator< CString > > >::reverse_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_rend" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< CString,std::allocator< CString > > >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_clear" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< std::vector< CString,std::allocator< CString > > > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_get_allocator" "', argument " "1"" of type '" "std::vector< VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = ((std::vector< VCString > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< std::vector< CString,std::allocator< CString > > >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_std__vectorT_CString_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VVString__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::vector< CString,std::allocator< CString > > >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< VCString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VVString" "', argument " "1"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::size_type""'");
  } 
  arg1 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::size_type >(val1);
  result = (std::vector< VCString > *)new std::vector< VCString >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_VCString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_pop_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_pop_back" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_resize" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VVString_resize" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  SwigValueWrapper< std::vector< std::vector< CString,std::allocator< CString > > >::iterator > arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  SwigValueWrapper< std::vector< std::vector< CString,std::allocator< CString > > >::iterator > result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_erase" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString_erase" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString_erase" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::iterator""'");
    }
  }
  result = std_vector_Sl_VCString_Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< CString,std::allocator< CString > > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  SwigValueWrapper< std::vector< std::vector< CString,std::allocator< CString > > >::iterator > arg2 ;
  SwigValueWrapper< std::vector< std::vector< CString,std::allocator< CString > > >::iterator > arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  SwigValueWrapper< std::vector< std::vector< CString,std::allocator< CString > > >::iterator > result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_erase" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString_erase" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString_erase" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString_erase" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString_erase" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::iterator""'");
    }
  }
  result = std_vector_Sl_VCString_Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< CString,std::allocator< CString > > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VVString_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_VVString_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_VVString_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VVString_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< VCString >::erase(std::vector< std::vector< CString,std::allocator< CString > > >::iterator)\n"
    "    std::vector< VCString >::erase(std::vector< std::vector< CString,std::allocator< CString > > >::iterator,std::vector< std::vector< CString,std::allocator< CString > > >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_VVString__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::vector< CString,std::allocator< CString > > >::size_type arg1 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::value_type *arg2 = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  std::vector< VCString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VVString" "', argument " "1"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::size_type""'");
  } 
  arg1 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::size_type >(val1);
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_VVString" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VVString" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (std::vector< VCString > *)new std::vector< VCString >(arg1,(std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_VCString_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VVString(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_VVString", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_VVString__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_VVString__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_VVString__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      int res = swig::asptr(argv[1], (std::vector< CString,std::allocator< CString > >**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_VVString__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_VVString'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< VCString >::vector()\n"
    "    std::vector< VCString >::vector(std::vector< VCString > const &)\n"
    "    std::vector< VCString >::vector(std::vector< std::vector< CString,std::allocator< CString > > >::size_type)\n"
    "    std::vector< VCString >::vector(std::vector< std::vector< CString,std::allocator< CString > > >::size_type,std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VVString_push_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VVString_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_push_back" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VVString_push_back" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VVString_push_back" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->push_back((std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::vector< CString,std::allocator< CString > > >::value_type *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_front" "', argument " "1"" of type '" "std::vector< VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = (std::vector< std::vector< CString,std::allocator< CString > > >::value_type *) &((std::vector< VCString > const *)arg1)->front();
  resultobj = swig::from(static_cast< std::vector< CString,std::allocator< CString > > >(*result));
  (void)swig::container_owner<swig::traits<std::vector< std::vector< CString,std::allocator< CString > > >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::vector< CString,std::allocator< CString > > >::value_type *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_back" "', argument " "1"" of type '" "std::vector< VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = (std::vector< std::vector< CString,std::allocator< CString > > >::value_type *) &((std::vector< VCString > const *)arg1)->back();
  resultobj = swig::from(static_cast< std::vector< CString,std::allocator< CString > > >(*result));
  (void)swig::container_owner<swig::traits<std::vector< std::vector< CString,std::allocator< CString > > >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_assign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::size_type arg2 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VVString_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_assign" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VVString_assign" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::size_type >(val2);
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VVString_assign" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VVString_assign" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->assign(arg2,(std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::size_type arg2 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_resize" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VVString_resize" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::size_type >(val2);
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VVString_resize" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VVString_resize" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->resize(arg2,(std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VVString_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VVString_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CString,std::allocator< CString > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VVString_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VVString_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< VCString >::resize(std::vector< std::vector< CString,std::allocator< CString > > >::size_type)\n"
    "    std::vector< VCString >::resize(std::vector< std::vector< CString,std::allocator< CString > > >::size_type,std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VVString_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  SwigValueWrapper< std::vector< std::vector< CString,std::allocator< CString > > >::iterator > arg2 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  int res3 = SWIG_OLDOBJ ;
  SwigValueWrapper< std::vector< std::vector< CString,std::allocator< CString > > >::iterator > result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_insert" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString_insert" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString_insert" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::iterator""'");
    }
  }
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VVString_insert" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VVString_insert" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  result = std_vector_Sl_VCString_Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),(std::vector< CString,std::allocator< CString > > const &)*arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< CString,std::allocator< CString > > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  SwigValueWrapper< std::vector< std::vector< CString,std::allocator< CString > > >::iterator > arg2 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::size_type arg3 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::value_type *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_insert" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString_insert" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VVString_insert" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VVString_insert" "', argument " "3"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::size_type""'");
  } 
  arg3 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::size_type >(val3);
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VVString_insert" "', argument " "4"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VVString_insert" "', argument " "4"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &""'"); 
    }
    arg4 = ptr;
  }
  std_vector_Sl_VCString_Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< CString,std::allocator< CString > > const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VVString_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *>(iter) != 0));
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CString,std::allocator< CString > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VVString_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< CString,std::allocator< CString > > >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< CString,std::allocator< CString > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VVString_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VVString_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< VCString >::insert(std::vector< std::vector< CString,std::allocator< CString > > >::iterator,std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &)\n"
    "    std::vector< VCString >::insert(std::vector< std::vector< CString,std::allocator< CString > > >::iterator,std::vector< std::vector< CString,std::allocator< CString > > >::size_type,std::vector< std::vector< CString,std::allocator< CString > > >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VVString_reserve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VVString_reserve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_reserve" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VVString_reserve" "', argument " "2"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VVString_capacity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::vector< CString,std::allocator< CString > > >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VVString_capacity" "', argument " "1"" of type '" "std::vector< VCString > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  result = ((std::vector< VCString > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_VVString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< VCString > *arg1 = (std::vector< VCString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_VCString_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VVString" "', argument " "1"" of type '" "std::vector< VCString > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< VCString > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *VVString_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_VCString_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *VVString_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_VClients_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_iterator" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_CClient_Sm__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___nonzero__" "', argument " "1"" of type '" "std::vector< CClient * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = (bool)std_vector_Sl_CClient_Sm__Sg____nonzero__((std::vector< CClient * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___bool__" "', argument " "1"" of type '" "std::vector< CClient * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = (bool)std_vector_Sl_CClient_Sm__Sg____bool__((std::vector< CClient * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___len__" "', argument " "1"" of type '" "std::vector< CClient * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = std_vector_Sl_CClient_Sm__Sg____len__((std::vector< CClient * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___getslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::difference_type arg2 ;
  std::vector< CClient * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  std::vector< CClient *,std::allocator< CClient * > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VClients___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___getslice__" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VClients___getslice__" "', argument " "2"" of type '" "std::vector< CClient * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CClient * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VClients___getslice__" "', argument " "3"" of type '" "std::vector< CClient * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CClient * >::difference_type >(val3);
  try {
    result = (std::vector< CClient *,std::allocator< CClient * > > *)std_vector_Sl_CClient_Sm__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CClient_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::difference_type arg2 ;
  std::vector< CClient * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___setslice__" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VClients___setslice__" "', argument " "2"" of type '" "std::vector< CClient * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CClient * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VClients___setslice__" "', argument " "3"" of type '" "std::vector< CClient * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CClient * >::difference_type >(val3);
  try {
    std_vector_Sl_CClient_Sm__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::difference_type arg2 ;
  std::vector< CClient * >::difference_type arg3 ;
  std::vector< CClient *,std::allocator< CClient * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___setslice__" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VClients___setslice__" "', argument " "2"" of type '" "std::vector< CClient * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CClient * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VClients___setslice__" "', argument " "3"" of type '" "std::vector< CClient * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CClient * >::difference_type >(val3);
  {
    std::vector< CClient*,std::allocator< CClient * > > *ptr = (std::vector< CClient*,std::allocator< CClient * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VClients___setslice__" "', argument " "4"" of type '" "std::vector< CClient *,std::allocator< CClient * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VClients___setslice__" "', argument " "4"" of type '" "std::vector< CClient *,std::allocator< CClient * > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_CClient_Sm__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< CClient *,std::allocator< CClient * > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VClients___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_VClients___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VClients___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VClients___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CClient * >::__setslice__(std::vector< CClient * >::difference_type,std::vector< CClient * >::difference_type)\n"
    "    std::vector< CClient * >::__setslice__(std::vector< CClient * >::difference_type,std::vector< CClient * >::difference_type,std::vector< CClient *,std::allocator< CClient * > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VClients___delslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::difference_type arg2 ;
  std::vector< CClient * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VClients___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___delslice__" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VClients___delslice__" "', argument " "2"" of type '" "std::vector< CClient * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CClient * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VClients___delslice__" "', argument " "3"" of type '" "std::vector< CClient * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CClient * >::difference_type >(val3);
  try {
    std_vector_Sl_CClient_Sm__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___delitem__" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VClients___delitem__" "', argument " "2"" of type '" "std::vector< CClient * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CClient * >::difference_type >(val2);
  try {
    std_vector_Sl_CClient_Sm__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< CClient *,std::allocator< CClient * > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___getitem__" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    result = (std::vector< CClient *,std::allocator< CClient * > > *)std_vector_Sl_CClient_Sm__Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CClient_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  std::vector< CClient *,std::allocator< CClient * > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___setitem__" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  {
    std::vector< CClient*,std::allocator< CClient * > > *ptr = (std::vector< CClient*,std::allocator< CClient * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VClients___setitem__" "', argument " "3"" of type '" "std::vector< CClient *,std::allocator< CClient * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VClients___setitem__" "', argument " "3"" of type '" "std::vector< CClient *,std::allocator< CClient * > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_CClient_Sm__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< CClient *,std::allocator< CClient * > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___setitem__" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CClient_Sm__Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___delitem__" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CClient_Sm__Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VClients___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VClients___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VClients___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VClients___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CClient * >::__delitem__(std::vector< CClient * >::difference_type)\n"
    "    std::vector< CClient * >::__delitem__(SWIGPY_SLICEOBJECT *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VClients___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< CClient * >::value_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___getitem__" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VClients___getitem__" "', argument " "2"" of type '" "std::vector< CClient * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CClient * >::difference_type >(val2);
  try {
    result = (std::vector< CClient * >::value_type)std_vector_Sl_CClient_Sm__Sg____getitem____SWIG_1(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CClient, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VClients___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VClients___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VClients___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VClients___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CClient * >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CClient * >::__getitem__(std::vector< CClient * >::difference_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VClients___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::difference_type arg2 ;
  std::vector< CClient * >::value_type arg3 = (std::vector< CClient * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients___setitem__" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VClients___setitem__" "', argument " "2"" of type '" "std::vector< CClient * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CClient * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VClients___setitem__" "', argument " "3"" of type '" "std::vector< CClient * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CClient * >::value_type >(argp3);
  try {
    std_vector_Sl_CClient_Sm__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VClients___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VClients___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VClients___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VClients___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VClients___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CClient * >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< CClient *,std::allocator< CClient * > > const &)\n"
    "    std::vector< CClient * >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CClient * >::__setitem__(std::vector< CClient * >::difference_type,std::vector< CClient * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VClients_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_pop" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  try {
    result = (std::vector< CClient * >::value_type)std_vector_Sl_CClient_Sm__Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CClient, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::value_type arg2 = (std::vector< CClient * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VClients_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_append" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VClients_append" "', argument " "2"" of type '" "std::vector< CClient * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CClient * >::value_type >(argp2);
  std_vector_Sl_CClient_Sm__Sg__append(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VClients__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< CClient * > *)new std::vector< CClient * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CClient_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VClients__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< CClient * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< CClient*,std::allocator< CClient * > > *ptr = (std::vector< CClient*,std::allocator< CClient * > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_VClients" "', argument " "1"" of type '" "std::vector< CClient * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VClients" "', argument " "1"" of type '" "std::vector< CClient * > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< CClient * > *)new std::vector< CClient * >((std::vector< CClient * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CClient_p_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_empty" "', argument " "1"" of type '" "std::vector< CClient * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = (bool)((std::vector< CClient * > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_size" "', argument " "1"" of type '" "std::vector< CClient * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = ((std::vector< CClient * > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VClients_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_swap" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_CClient_p_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VClients_swap" "', argument " "2"" of type '" "std::vector< CClient * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VClients_swap" "', argument " "2"" of type '" "std::vector< CClient * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CClient * > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_begin" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CClient * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_end" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CClient * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_rbegin" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CClient * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_rend" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CClient * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_clear" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< CClient * > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_get_allocator" "', argument " "1"" of type '" "std::vector< CClient * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = ((std::vector< CClient * > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< CClient * >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_CClient_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VClients__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< CClient * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VClients" "', argument " "1"" of type '" "std::vector< CClient * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CClient * >::size_type >(val1);
  result = (std::vector< CClient * > *)new std::vector< CClient * >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CClient_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_pop_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_pop_back" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_resize" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VClients_resize" "', argument " "2"" of type '" "std::vector< CClient * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CClient * >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< CClient * >::iterator result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_erase" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients_erase" "', argument " "2"" of type '" "std::vector< CClient * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients_erase" "', argument " "2"" of type '" "std::vector< CClient * >::iterator""'");
    }
  }
  result = std_vector_Sl_CClient_Sm__Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CClient * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::iterator arg2 ;
  std::vector< CClient * >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::vector< CClient * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_erase" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients_erase" "', argument " "2"" of type '" "std::vector< CClient * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients_erase" "', argument " "2"" of type '" "std::vector< CClient * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients_erase" "', argument " "3"" of type '" "std::vector< CClient * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients_erase" "', argument " "3"" of type '" "std::vector< CClient * >::iterator""'");
    }
  }
  result = std_vector_Sl_CClient_Sm__Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CClient * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VClients_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_VClients_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_VClients_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VClients_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CClient * >::erase(std::vector< CClient * >::iterator)\n"
    "    std::vector< CClient * >::erase(std::vector< CClient * >::iterator,std::vector< CClient * >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_VClients__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * >::size_type arg1 ;
  std::vector< CClient * >::value_type arg2 = (std::vector< CClient * >::value_type) 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::vector< CClient * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VClients" "', argument " "1"" of type '" "std::vector< CClient * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CClient * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_VClients" "', argument " "2"" of type '" "std::vector< CClient * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CClient * >::value_type >(argp2);
  result = (std::vector< CClient * > *)new std::vector< CClient * >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CClient_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VClients(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_VClients", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_VClients__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_VClients__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_VClients__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CClient, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_VClients__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_VClients'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CClient * >::vector()\n"
    "    std::vector< CClient * >::vector(std::vector< CClient * > const &)\n"
    "    std::vector< CClient * >::vector(std::vector< CClient * >::size_type)\n"
    "    std::vector< CClient * >::vector(std::vector< CClient * >::size_type,std::vector< CClient * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VClients_push_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::value_type arg2 = (std::vector< CClient * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VClients_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_push_back" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VClients_push_back" "', argument " "2"" of type '" "std::vector< CClient * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CClient * >::value_type >(argp2);
  (arg1)->push_back(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_front" "', argument " "1"" of type '" "std::vector< CClient * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = (std::vector< CClient * >::value_type)((std::vector< CClient * > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CClient, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_back" "', argument " "1"" of type '" "std::vector< CClient * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = (std::vector< CClient * >::value_type)((std::vector< CClient * > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CClient, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_assign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::size_type arg2 ;
  std::vector< CClient * >::value_type arg3 = (std::vector< CClient * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VClients_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_assign" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VClients_assign" "', argument " "2"" of type '" "std::vector< CClient * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CClient * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VClients_assign" "', argument " "3"" of type '" "std::vector< CClient * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CClient * >::value_type >(argp3);
  (arg1)->assign(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::size_type arg2 ;
  std::vector< CClient * >::value_type arg3 = (std::vector< CClient * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_resize" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VClients_resize" "', argument " "2"" of type '" "std::vector< CClient * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CClient * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VClients_resize" "', argument " "3"" of type '" "std::vector< CClient * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CClient * >::value_type >(argp3);
  (arg1)->resize(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VClients_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VClients_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VClients_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VClients_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CClient * >::resize(std::vector< CClient * >::size_type)\n"
    "    std::vector< CClient * >::resize(std::vector< CClient * >::size_type,std::vector< CClient * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VClients_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::iterator arg2 ;
  std::vector< CClient * >::value_type arg3 = (std::vector< CClient * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  std::vector< CClient * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_insert" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients_insert" "', argument " "2"" of type '" "std::vector< CClient * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients_insert" "', argument " "2"" of type '" "std::vector< CClient * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VClients_insert" "', argument " "3"" of type '" "std::vector< CClient * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CClient * >::value_type >(argp3);
  result = std_vector_Sl_CClient_Sm__Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CClient * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::iterator arg2 ;
  std::vector< CClient * >::size_type arg3 ;
  std::vector< CClient * >::value_type arg4 = (std::vector< CClient * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_insert" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients_insert" "', argument " "2"" of type '" "std::vector< CClient * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VClients_insert" "', argument " "2"" of type '" "std::vector< CClient * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VClients_insert" "', argument " "3"" of type '" "std::vector< CClient * >::size_type""'");
  } 
  arg3 = static_cast< std::vector< CClient * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VClients_insert" "', argument " "4"" of type '" "std::vector< CClient * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector< CClient * >::value_type >(argp4);
  std_vector_Sl_CClient_Sm__Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VClients_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VClients_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CClient*,std::allocator< CClient * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CClient * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VClients_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VClients_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CClient * >::insert(std::vector< CClient * >::iterator,std::vector< CClient * >::value_type)\n"
    "    std::vector< CClient * >::insert(std::vector< CClient * >::iterator,std::vector< CClient * >::size_type,std::vector< CClient * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VClients_reserve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  std::vector< CClient * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VClients_reserve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_reserve" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VClients_reserve" "', argument " "2"" of type '" "std::vector< CClient * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CClient * >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VClients_capacity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VClients_capacity" "', argument " "1"" of type '" "std::vector< CClient * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  result = ((std::vector< CClient * > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_VClients(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CClient * > *arg1 = (std::vector< CClient * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CClient_p_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VClients" "', argument " "1"" of type '" "std::vector< CClient * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CClient * > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *VClients_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_CClient_p_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *VClients_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_VServers_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_iterator" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_CServer_Sm__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___nonzero__" "', argument " "1"" of type '" "std::vector< CServer * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = (bool)std_vector_Sl_CServer_Sm__Sg____nonzero__((std::vector< CServer * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___bool__" "', argument " "1"" of type '" "std::vector< CServer * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = (bool)std_vector_Sl_CServer_Sm__Sg____bool__((std::vector< CServer * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CServer * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___len__" "', argument " "1"" of type '" "std::vector< CServer * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = std_vector_Sl_CServer_Sm__Sg____len__((std::vector< CServer * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___getslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::difference_type arg2 ;
  std::vector< CServer * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  std::vector< CServer *,std::allocator< CServer * > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VServers___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___getslice__" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VServers___getslice__" "', argument " "2"" of type '" "std::vector< CServer * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CServer * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VServers___getslice__" "', argument " "3"" of type '" "std::vector< CServer * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CServer * >::difference_type >(val3);
  try {
    result = (std::vector< CServer *,std::allocator< CServer * > > *)std_vector_Sl_CServer_Sm__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CServer_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::difference_type arg2 ;
  std::vector< CServer * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___setslice__" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VServers___setslice__" "', argument " "2"" of type '" "std::vector< CServer * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CServer * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VServers___setslice__" "', argument " "3"" of type '" "std::vector< CServer * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CServer * >::difference_type >(val3);
  try {
    std_vector_Sl_CServer_Sm__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::difference_type arg2 ;
  std::vector< CServer * >::difference_type arg3 ;
  std::vector< CServer *,std::allocator< CServer * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___setslice__" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VServers___setslice__" "', argument " "2"" of type '" "std::vector< CServer * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CServer * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VServers___setslice__" "', argument " "3"" of type '" "std::vector< CServer * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CServer * >::difference_type >(val3);
  {
    std::vector< CServer*,std::allocator< CServer * > > *ptr = (std::vector< CServer*,std::allocator< CServer * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VServers___setslice__" "', argument " "4"" of type '" "std::vector< CServer *,std::allocator< CServer * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VServers___setslice__" "', argument " "4"" of type '" "std::vector< CServer *,std::allocator< CServer * > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_CServer_Sm__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< CServer *,std::allocator< CServer * > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VServers___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_VServers___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VServers___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VServers___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CServer * >::__setslice__(std::vector< CServer * >::difference_type,std::vector< CServer * >::difference_type)\n"
    "    std::vector< CServer * >::__setslice__(std::vector< CServer * >::difference_type,std::vector< CServer * >::difference_type,std::vector< CServer *,std::allocator< CServer * > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VServers___delslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::difference_type arg2 ;
  std::vector< CServer * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VServers___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___delslice__" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VServers___delslice__" "', argument " "2"" of type '" "std::vector< CServer * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CServer * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VServers___delslice__" "', argument " "3"" of type '" "std::vector< CServer * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CServer * >::difference_type >(val3);
  try {
    std_vector_Sl_CServer_Sm__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___delitem__" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VServers___delitem__" "', argument " "2"" of type '" "std::vector< CServer * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CServer * >::difference_type >(val2);
  try {
    std_vector_Sl_CServer_Sm__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< CServer *,std::allocator< CServer * > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___getitem__" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    result = (std::vector< CServer *,std::allocator< CServer * > > *)std_vector_Sl_CServer_Sm__Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CServer_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  std::vector< CServer *,std::allocator< CServer * > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___setitem__" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  {
    std::vector< CServer*,std::allocator< CServer * > > *ptr = (std::vector< CServer*,std::allocator< CServer * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VServers___setitem__" "', argument " "3"" of type '" "std::vector< CServer *,std::allocator< CServer * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VServers___setitem__" "', argument " "3"" of type '" "std::vector< CServer *,std::allocator< CServer * > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_CServer_Sm__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< CServer *,std::allocator< CServer * > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___setitem__" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CServer_Sm__Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___delitem__" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CServer_Sm__Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VServers___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VServers___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VServers___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VServers___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CServer * >::__delitem__(std::vector< CServer * >::difference_type)\n"
    "    std::vector< CServer * >::__delitem__(SWIGPY_SLICEOBJECT *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VServers___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< CServer * >::value_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___getitem__" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VServers___getitem__" "', argument " "2"" of type '" "std::vector< CServer * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CServer * >::difference_type >(val2);
  try {
    result = (std::vector< CServer * >::value_type)std_vector_Sl_CServer_Sm__Sg____getitem____SWIG_1(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CServer, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VServers___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VServers___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VServers___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VServers___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CServer * >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CServer * >::__getitem__(std::vector< CServer * >::difference_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VServers___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::difference_type arg2 ;
  std::vector< CServer * >::value_type arg3 = (std::vector< CServer * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers___setitem__" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VServers___setitem__" "', argument " "2"" of type '" "std::vector< CServer * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CServer * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VServers___setitem__" "', argument " "3"" of type '" "std::vector< CServer * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CServer * >::value_type >(argp3);
  try {
    std_vector_Sl_CServer_Sm__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VServers___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VServers___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VServers___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CServer, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VServers___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VServers___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CServer * >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< CServer *,std::allocator< CServer * > > const &)\n"
    "    std::vector< CServer * >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CServer * >::__setitem__(std::vector< CServer * >::difference_type,std::vector< CServer * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VServers_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CServer * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_pop" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  try {
    result = (std::vector< CServer * >::value_type)std_vector_Sl_CServer_Sm__Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CServer, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::value_type arg2 = (std::vector< CServer * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VServers_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_append" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VServers_append" "', argument " "2"" of type '" "std::vector< CServer * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CServer * >::value_type >(argp2);
  std_vector_Sl_CServer_Sm__Sg__append(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VServers__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< CServer * > *)new std::vector< CServer * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CServer_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VServers__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< CServer * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< CServer*,std::allocator< CServer * > > *ptr = (std::vector< CServer*,std::allocator< CServer * > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_VServers" "', argument " "1"" of type '" "std::vector< CServer * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VServers" "', argument " "1"" of type '" "std::vector< CServer * > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< CServer * > *)new std::vector< CServer * >((std::vector< CServer * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CServer_p_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_empty" "', argument " "1"" of type '" "std::vector< CServer * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = (bool)((std::vector< CServer * > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CServer * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_size" "', argument " "1"" of type '" "std::vector< CServer * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = ((std::vector< CServer * > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VServers_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_swap" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_CServer_p_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VServers_swap" "', argument " "2"" of type '" "std::vector< CServer * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VServers_swap" "', argument " "2"" of type '" "std::vector< CServer * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CServer * > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CServer * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_begin" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CServer * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CServer * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_end" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CServer * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CServer * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_rbegin" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CServer * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CServer * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_rend" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CServer * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_clear" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< CServer * > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_get_allocator" "', argument " "1"" of type '" "std::vector< CServer * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = ((std::vector< CServer * > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< CServer * >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_CServer_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VServers__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< CServer * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VServers" "', argument " "1"" of type '" "std::vector< CServer * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CServer * >::size_type >(val1);
  result = (std::vector< CServer * > *)new std::vector< CServer * >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CServer_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_pop_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_pop_back" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_resize" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VServers_resize" "', argument " "2"" of type '" "std::vector< CServer * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CServer * >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< CServer * >::iterator result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_erase" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers_erase" "', argument " "2"" of type '" "std::vector< CServer * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers_erase" "', argument " "2"" of type '" "std::vector< CServer * >::iterator""'");
    }
  }
  result = std_vector_Sl_CServer_Sm__Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CServer * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::iterator arg2 ;
  std::vector< CServer * >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::vector< CServer * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_erase" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers_erase" "', argument " "2"" of type '" "std::vector< CServer * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers_erase" "', argument " "2"" of type '" "std::vector< CServer * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers_erase" "', argument " "3"" of type '" "std::vector< CServer * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers_erase" "', argument " "3"" of type '" "std::vector< CServer * >::iterator""'");
    }
  }
  result = std_vector_Sl_CServer_Sm__Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CServer * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VServers_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_VServers_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_VServers_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VServers_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CServer * >::erase(std::vector< CServer * >::iterator)\n"
    "    std::vector< CServer * >::erase(std::vector< CServer * >::iterator,std::vector< CServer * >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_VServers__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * >::size_type arg1 ;
  std::vector< CServer * >::value_type arg2 = (std::vector< CServer * >::value_type) 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::vector< CServer * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VServers" "', argument " "1"" of type '" "std::vector< CServer * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CServer * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_VServers" "', argument " "2"" of type '" "std::vector< CServer * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CServer * >::value_type >(argp2);
  result = (std::vector< CServer * > *)new std::vector< CServer * >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CServer_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VServers(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_VServers", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_VServers__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_VServers__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_VServers__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CServer, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_VServers__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_VServers'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CServer * >::vector()\n"
    "    std::vector< CServer * >::vector(std::vector< CServer * > const &)\n"
    "    std::vector< CServer * >::vector(std::vector< CServer * >::size_type)\n"
    "    std::vector< CServer * >::vector(std::vector< CServer * >::size_type,std::vector< CServer * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VServers_push_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::value_type arg2 = (std::vector< CServer * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VServers_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_push_back" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VServers_push_back" "', argument " "2"" of type '" "std::vector< CServer * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CServer * >::value_type >(argp2);
  (arg1)->push_back(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CServer * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_front" "', argument " "1"" of type '" "std::vector< CServer * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = (std::vector< CServer * >::value_type)((std::vector< CServer * > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CServer, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CServer * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_back" "', argument " "1"" of type '" "std::vector< CServer * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = (std::vector< CServer * >::value_type)((std::vector< CServer * > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CServer, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_assign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::size_type arg2 ;
  std::vector< CServer * >::value_type arg3 = (std::vector< CServer * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VServers_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_assign" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VServers_assign" "', argument " "2"" of type '" "std::vector< CServer * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CServer * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VServers_assign" "', argument " "3"" of type '" "std::vector< CServer * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CServer * >::value_type >(argp3);
  (arg1)->assign(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::size_type arg2 ;
  std::vector< CServer * >::value_type arg3 = (std::vector< CServer * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_resize" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VServers_resize" "', argument " "2"" of type '" "std::vector< CServer * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CServer * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VServers_resize" "', argument " "3"" of type '" "std::vector< CServer * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CServer * >::value_type >(argp3);
  (arg1)->resize(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VServers_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VServers_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CServer, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VServers_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VServers_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CServer * >::resize(std::vector< CServer * >::size_type)\n"
    "    std::vector< CServer * >::resize(std::vector< CServer * >::size_type,std::vector< CServer * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VServers_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::iterator arg2 ;
  std::vector< CServer * >::value_type arg3 = (std::vector< CServer * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  std::vector< CServer * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_insert" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers_insert" "', argument " "2"" of type '" "std::vector< CServer * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers_insert" "', argument " "2"" of type '" "std::vector< CServer * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VServers_insert" "', argument " "3"" of type '" "std::vector< CServer * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CServer * >::value_type >(argp3);
  result = std_vector_Sl_CServer_Sm__Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CServer * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::iterator arg2 ;
  std::vector< CServer * >::size_type arg3 ;
  std::vector< CServer * >::value_type arg4 = (std::vector< CServer * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_insert" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers_insert" "', argument " "2"" of type '" "std::vector< CServer * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VServers_insert" "', argument " "2"" of type '" "std::vector< CServer * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VServers_insert" "', argument " "3"" of type '" "std::vector< CServer * >::size_type""'");
  } 
  arg3 = static_cast< std::vector< CServer * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VServers_insert" "', argument " "4"" of type '" "std::vector< CServer * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector< CServer * >::value_type >(argp4);
  std_vector_Sl_CServer_Sm__Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VServers_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CServer, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VServers_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CServer*,std::allocator< CServer * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CServer * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CServer, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VServers_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VServers_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CServer * >::insert(std::vector< CServer * >::iterator,std::vector< CServer * >::value_type)\n"
    "    std::vector< CServer * >::insert(std::vector< CServer * >::iterator,std::vector< CServer * >::size_type,std::vector< CServer * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VServers_reserve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  std::vector< CServer * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VServers_reserve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_reserve" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VServers_reserve" "', argument " "2"" of type '" "std::vector< CServer * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CServer * >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VServers_capacity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CServer * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VServers_capacity" "', argument " "1"" of type '" "std::vector< CServer * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  result = ((std::vector< CServer * > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_VServers(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CServer * > *arg1 = (std::vector< CServer * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CServer_p_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VServers" "', argument " "1"" of type '" "std::vector< CServer * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CServer * > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *VServers_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_CServer_p_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *VServers_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_VQueries_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_iterator" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_CQuery_Sm__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___nonzero__" "', argument " "1"" of type '" "std::vector< CQuery * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = (bool)std_vector_Sl_CQuery_Sm__Sg____nonzero__((std::vector< CQuery * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___bool__" "', argument " "1"" of type '" "std::vector< CQuery * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = (bool)std_vector_Sl_CQuery_Sm__Sg____bool__((std::vector< CQuery * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CQuery * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___len__" "', argument " "1"" of type '" "std::vector< CQuery * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = std_vector_Sl_CQuery_Sm__Sg____len__((std::vector< CQuery * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___getslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::difference_type arg2 ;
  std::vector< CQuery * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  std::vector< CQuery *,std::allocator< CQuery * > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VQueries___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___getslice__" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VQueries___getslice__" "', argument " "2"" of type '" "std::vector< CQuery * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CQuery * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VQueries___getslice__" "', argument " "3"" of type '" "std::vector< CQuery * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CQuery * >::difference_type >(val3);
  try {
    result = (std::vector< CQuery *,std::allocator< CQuery * > > *)std_vector_Sl_CQuery_Sm__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CQuery_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::difference_type arg2 ;
  std::vector< CQuery * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___setslice__" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VQueries___setslice__" "', argument " "2"" of type '" "std::vector< CQuery * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CQuery * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VQueries___setslice__" "', argument " "3"" of type '" "std::vector< CQuery * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CQuery * >::difference_type >(val3);
  try {
    std_vector_Sl_CQuery_Sm__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::difference_type arg2 ;
  std::vector< CQuery * >::difference_type arg3 ;
  std::vector< CQuery *,std::allocator< CQuery * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___setslice__" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VQueries___setslice__" "', argument " "2"" of type '" "std::vector< CQuery * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CQuery * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VQueries___setslice__" "', argument " "3"" of type '" "std::vector< CQuery * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CQuery * >::difference_type >(val3);
  {
    std::vector< CQuery*,std::allocator< CQuery * > > *ptr = (std::vector< CQuery*,std::allocator< CQuery * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VQueries___setslice__" "', argument " "4"" of type '" "std::vector< CQuery *,std::allocator< CQuery * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VQueries___setslice__" "', argument " "4"" of type '" "std::vector< CQuery *,std::allocator< CQuery * > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_CQuery_Sm__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< CQuery *,std::allocator< CQuery * > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VQueries___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_VQueries___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VQueries___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VQueries___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CQuery * >::__setslice__(std::vector< CQuery * >::difference_type,std::vector< CQuery * >::difference_type)\n"
    "    std::vector< CQuery * >::__setslice__(std::vector< CQuery * >::difference_type,std::vector< CQuery * >::difference_type,std::vector< CQuery *,std::allocator< CQuery * > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VQueries___delslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::difference_type arg2 ;
  std::vector< CQuery * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VQueries___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___delslice__" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VQueries___delslice__" "', argument " "2"" of type '" "std::vector< CQuery * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CQuery * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VQueries___delslice__" "', argument " "3"" of type '" "std::vector< CQuery * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< CQuery * >::difference_type >(val3);
  try {
    std_vector_Sl_CQuery_Sm__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___delitem__" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VQueries___delitem__" "', argument " "2"" of type '" "std::vector< CQuery * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CQuery * >::difference_type >(val2);
  try {
    std_vector_Sl_CQuery_Sm__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< CQuery *,std::allocator< CQuery * > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___getitem__" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    result = (std::vector< CQuery *,std::allocator< CQuery * > > *)std_vector_Sl_CQuery_Sm__Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CQuery_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  std::vector< CQuery *,std::allocator< CQuery * > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___setitem__" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  {
    std::vector< CQuery*,std::allocator< CQuery * > > *ptr = (std::vector< CQuery*,std::allocator< CQuery * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VQueries___setitem__" "', argument " "3"" of type '" "std::vector< CQuery *,std::allocator< CQuery * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VQueries___setitem__" "', argument " "3"" of type '" "std::vector< CQuery *,std::allocator< CQuery * > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_CQuery_Sm__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< CQuery *,std::allocator< CQuery * > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___setitem__" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CQuery_Sm__Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___delitem__" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_CQuery_Sm__Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VQueries___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VQueries___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VQueries___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VQueries___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CQuery * >::__delitem__(std::vector< CQuery * >::difference_type)\n"
    "    std::vector< CQuery * >::__delitem__(SWIGPY_SLICEOBJECT *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VQueries___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< CQuery * >::value_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___getitem__" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VQueries___getitem__" "', argument " "2"" of type '" "std::vector< CQuery * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CQuery * >::difference_type >(val2);
  try {
    result = (std::vector< CQuery * >::value_type)std_vector_Sl_CQuery_Sm__Sg____getitem____SWIG_1(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CQuery, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VQueries___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VQueries___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VQueries___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VQueries___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CQuery * >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CQuery * >::__getitem__(std::vector< CQuery * >::difference_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VQueries___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::difference_type arg2 ;
  std::vector< CQuery * >::value_type arg3 = (std::vector< CQuery * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries___setitem__" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VQueries___setitem__" "', argument " "2"" of type '" "std::vector< CQuery * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< CQuery * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VQueries___setitem__" "', argument " "3"" of type '" "std::vector< CQuery * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CQuery * >::value_type >(argp3);
  try {
    std_vector_Sl_CQuery_Sm__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VQueries___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VQueries___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VQueries___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CQuery, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VQueries___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VQueries___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CQuery * >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< CQuery *,std::allocator< CQuery * > > const &)\n"
    "    std::vector< CQuery * >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< CQuery * >::__setitem__(std::vector< CQuery * >::difference_type,std::vector< CQuery * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VQueries_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CQuery * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_pop" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  try {
    result = (std::vector< CQuery * >::value_type)std_vector_Sl_CQuery_Sm__Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CQuery, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::value_type arg2 = (std::vector< CQuery * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VQueries_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_append" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VQueries_append" "', argument " "2"" of type '" "std::vector< CQuery * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CQuery * >::value_type >(argp2);
  std_vector_Sl_CQuery_Sm__Sg__append(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VQueries__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< CQuery * > *)new std::vector< CQuery * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CQuery_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VQueries__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< CQuery * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< CQuery*,std::allocator< CQuery * > > *ptr = (std::vector< CQuery*,std::allocator< CQuery * > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_VQueries" "', argument " "1"" of type '" "std::vector< CQuery * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VQueries" "', argument " "1"" of type '" "std::vector< CQuery * > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< CQuery * > *)new std::vector< CQuery * >((std::vector< CQuery * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CQuery_p_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_empty" "', argument " "1"" of type '" "std::vector< CQuery * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = (bool)((std::vector< CQuery * > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CQuery * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_size" "', argument " "1"" of type '" "std::vector< CQuery * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = ((std::vector< CQuery * > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VQueries_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_swap" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_CQuery_p_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VQueries_swap" "', argument " "2"" of type '" "std::vector< CQuery * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VQueries_swap" "', argument " "2"" of type '" "std::vector< CQuery * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CQuery * > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CQuery * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_begin" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CQuery * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CQuery * >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_end" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CQuery * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CQuery * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_rbegin" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CQuery * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CQuery * >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_rend" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CQuery * >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_clear" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< CQuery * > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_get_allocator" "', argument " "1"" of type '" "std::vector< CQuery * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = ((std::vector< CQuery * > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< CQuery * >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_CQuery_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VQueries__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< CQuery * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VQueries" "', argument " "1"" of type '" "std::vector< CQuery * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CQuery * >::size_type >(val1);
  result = (std::vector< CQuery * > *)new std::vector< CQuery * >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CQuery_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_pop_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_pop_back" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_resize" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VQueries_resize" "', argument " "2"" of type '" "std::vector< CQuery * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CQuery * >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< CQuery * >::iterator result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_erase" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries_erase" "', argument " "2"" of type '" "std::vector< CQuery * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries_erase" "', argument " "2"" of type '" "std::vector< CQuery * >::iterator""'");
    }
  }
  result = std_vector_Sl_CQuery_Sm__Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CQuery * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::iterator arg2 ;
  std::vector< CQuery * >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::vector< CQuery * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_erase" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries_erase" "', argument " "2"" of type '" "std::vector< CQuery * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries_erase" "', argument " "2"" of type '" "std::vector< CQuery * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries_erase" "', argument " "3"" of type '" "std::vector< CQuery * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries_erase" "', argument " "3"" of type '" "std::vector< CQuery * >::iterator""'");
    }
  }
  result = std_vector_Sl_CQuery_Sm__Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CQuery * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VQueries_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_VQueries_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_VQueries_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VQueries_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CQuery * >::erase(std::vector< CQuery * >::iterator)\n"
    "    std::vector< CQuery * >::erase(std::vector< CQuery * >::iterator,std::vector< CQuery * >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_VQueries__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * >::size_type arg1 ;
  std::vector< CQuery * >::value_type arg2 = (std::vector< CQuery * >::value_type) 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::vector< CQuery * > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VQueries" "', argument " "1"" of type '" "std::vector< CQuery * >::size_type""'");
  } 
  arg1 = static_cast< std::vector< CQuery * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_VQueries" "', argument " "2"" of type '" "std::vector< CQuery * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CQuery * >::value_type >(argp2);
  result = (std::vector< CQuery * > *)new std::vector< CQuery * >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CQuery_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VQueries(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_VQueries", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_VQueries__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_VQueries__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_VQueries__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CQuery, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_VQueries__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_VQueries'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CQuery * >::vector()\n"
    "    std::vector< CQuery * >::vector(std::vector< CQuery * > const &)\n"
    "    std::vector< CQuery * >::vector(std::vector< CQuery * >::size_type)\n"
    "    std::vector< CQuery * >::vector(std::vector< CQuery * >::size_type,std::vector< CQuery * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VQueries_push_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::value_type arg2 = (std::vector< CQuery * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VQueries_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_push_back" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VQueries_push_back" "', argument " "2"" of type '" "std::vector< CQuery * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CQuery * >::value_type >(argp2);
  (arg1)->push_back(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CQuery * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_front" "', argument " "1"" of type '" "std::vector< CQuery * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = (std::vector< CQuery * >::value_type)((std::vector< CQuery * > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CQuery, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CQuery * >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_back" "', argument " "1"" of type '" "std::vector< CQuery * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = (std::vector< CQuery * >::value_type)((std::vector< CQuery * > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CQuery, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_assign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::size_type arg2 ;
  std::vector< CQuery * >::value_type arg3 = (std::vector< CQuery * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VQueries_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_assign" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VQueries_assign" "', argument " "2"" of type '" "std::vector< CQuery * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CQuery * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VQueries_assign" "', argument " "3"" of type '" "std::vector< CQuery * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CQuery * >::value_type >(argp3);
  (arg1)->assign(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::size_type arg2 ;
  std::vector< CQuery * >::value_type arg3 = (std::vector< CQuery * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_resize" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VQueries_resize" "', argument " "2"" of type '" "std::vector< CQuery * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CQuery * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VQueries_resize" "', argument " "3"" of type '" "std::vector< CQuery * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CQuery * >::value_type >(argp3);
  (arg1)->resize(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VQueries_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VQueries_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CQuery, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VQueries_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VQueries_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CQuery * >::resize(std::vector< CQuery * >::size_type)\n"
    "    std::vector< CQuery * >::resize(std::vector< CQuery * >::size_type,std::vector< CQuery * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VQueries_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::iterator arg2 ;
  std::vector< CQuery * >::value_type arg3 = (std::vector< CQuery * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  std::vector< CQuery * >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_insert" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries_insert" "', argument " "2"" of type '" "std::vector< CQuery * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries_insert" "', argument " "2"" of type '" "std::vector< CQuery * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VQueries_insert" "', argument " "3"" of type '" "std::vector< CQuery * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector< CQuery * >::value_type >(argp3);
  result = std_vector_Sl_CQuery_Sm__Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< CQuery * >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::iterator arg2 ;
  std::vector< CQuery * >::size_type arg3 ;
  std::vector< CQuery * >::value_type arg4 = (std::vector< CQuery * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_insert" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries_insert" "', argument " "2"" of type '" "std::vector< CQuery * >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VQueries_insert" "', argument " "2"" of type '" "std::vector< CQuery * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VQueries_insert" "', argument " "3"" of type '" "std::vector< CQuery * >::size_type""'");
  } 
  arg3 = static_cast< std::vector< CQuery * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VQueries_insert" "', argument " "4"" of type '" "std::vector< CQuery * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector< CQuery * >::value_type >(argp4);
  std_vector_Sl_CQuery_Sm__Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VQueries_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CQuery, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VQueries_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< CQuery*,std::allocator< CQuery * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< CQuery * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CQuery, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VQueries_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VQueries_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< CQuery * >::insert(std::vector< CQuery * >::iterator,std::vector< CQuery * >::value_type)\n"
    "    std::vector< CQuery * >::insert(std::vector< CQuery * >::iterator,std::vector< CQuery * >::size_type,std::vector< CQuery * >::value_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VQueries_reserve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  std::vector< CQuery * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VQueries_reserve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_reserve" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VQueries_reserve" "', argument " "2"" of type '" "std::vector< CQuery * >::size_type""'");
  } 
  arg2 = static_cast< std::vector< CQuery * >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VQueries_capacity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CQuery * >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VQueries_capacity" "', argument " "1"" of type '" "std::vector< CQuery * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  result = ((std::vector< CQuery * > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_VQueries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< CQuery * > *arg1 = (std::vector< CQuery * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_CQuery_p_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VQueries" "', argument " "1"" of type '" "std::vector< CQuery * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< CQuery * > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *VQueries_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_CQuery_p_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *VQueries_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CTranslationInfo_GetTranslations(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  SwigValueWrapper< std::map< CString,CTranslationInfo,std::less< CString >,std::allocator< std::pair< CString const,CTranslationInfo > > > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTranslationInfo_GetTranslations", 0, 0, 0)) SWIG_fail;
  result = CTranslationInfo::GetTranslations();
  resultobj = SWIG_NewPointerObj((new std::map< CString,CTranslationInfo,std::less< CString >,std::allocator< std::pair< CString const,CTranslationInfo > > >(result)), SWIGTYPE_p_std__mapT_CString_CTranslationInfo_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CTranslationInfo_t_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTranslationInfo_sSelfName_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslationInfo *arg1 = (CTranslationInfo *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTranslationInfo_sSelfName_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTranslationInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTranslationInfo_sSelfName_set" "', argument " "1"" of type '" "CTranslationInfo *""'"); 
  }
  arg1 = reinterpret_cast< CTranslationInfo * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTranslationInfo_sSelfName_set" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTranslationInfo_sSelfName_set" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  if (arg1) (arg1)->sSelfName = *arg2;
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTranslationInfo_sSelfName_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslationInfo *arg1 = (CTranslationInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTranslationInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTranslationInfo_sSelfName_get" "', argument " "1"" of type '" "CTranslationInfo *""'"); 
  }
  arg1 = reinterpret_cast< CTranslationInfo * >(argp1);
  result = (CString *) & ((arg1)->sSelfName);
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CTranslationInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslationInfo *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CTranslationInfo", 0, 0, 0)) SWIG_fail;
  result = (CTranslationInfo *)new CTranslationInfo();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTranslationInfo, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CTranslationInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslationInfo *arg1 = (CTranslationInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTranslationInfo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CTranslationInfo" "', argument " "1"" of type '" "CTranslationInfo *""'"); 
  }
  arg1 = reinterpret_cast< CTranslationInfo * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CTranslationInfo_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CTranslationInfo, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CTranslationInfo_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CTranslation_Get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslation *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTranslation_Get", 0, 0, 0)) SWIG_fail;
  result = (CTranslation *) &CTranslation::Get();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTranslation, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTranslation_Singular(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslation *arg1 = (CTranslation *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTranslation_Singular", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTranslation, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTranslation_Singular" "', argument " "1"" of type '" "CTranslation *""'"); 
  }
  arg1 = reinterpret_cast< CTranslation * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTranslation_Singular" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTranslation_Singular" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CTranslation_Singular" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTranslation_Singular" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CTranslation_Singular" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTranslation_Singular" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (arg1)->Singular((CString const &)*arg2,(CString const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTranslation_Plural(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslation *arg1 = (CTranslation *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  int val6 ;
  int ecode6 = 0 ;
  PyObject *swig_obj[6] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTranslation_Plural", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTranslation, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTranslation_Plural" "', argument " "1"" of type '" "CTranslation *""'"); 
  }
  arg1 = reinterpret_cast< CTranslation * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTranslation_Plural" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTranslation_Plural" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CTranslation_Plural" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTranslation_Plural" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CTranslation_Plural" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTranslation_Plural" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CTranslation_Plural" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTranslation_Plural" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CTranslation_Plural" "', argument " "6"" of type '" "int""'");
  } 
  arg6 = static_cast< int >(val6);
  result = (arg1)->Plural((CString const &)*arg2,(CString const &)*arg3,(CString const &)*arg4,(CString const &)*arg5,arg6);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTranslation_PushLanguage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslation *arg1 = (CTranslation *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTranslation_PushLanguage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTranslation, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTranslation_PushLanguage" "', argument " "1"" of type '" "CTranslation *""'"); 
  }
  arg1 = reinterpret_cast< CTranslation * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTranslation_PushLanguage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTranslation_PushLanguage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->PushLanguage((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTranslation_PopLanguage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslation *arg1 = (CTranslation *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTranslation, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTranslation_PopLanguage" "', argument " "1"" of type '" "CTranslation *""'"); 
  }
  arg1 = reinterpret_cast< CTranslation * >(argp1);
  (arg1)->PopLanguage();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTranslation_NewReference(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslation *arg1 = (CTranslation *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTranslation_NewReference", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTranslation, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTranslation_NewReference" "', argument " "1"" of type '" "CTranslation *""'"); 
  }
  arg1 = reinterpret_cast< CTranslation * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTranslation_NewReference" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTranslation_NewReference" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->NewReference((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTranslation_DelReference(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslation *arg1 = (CTranslation *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTranslation_DelReference", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTranslation, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTranslation_DelReference" "', argument " "1"" of type '" "CTranslation *""'"); 
  }
  arg1 = reinterpret_cast< CTranslation * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTranslation_DelReference" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTranslation_DelReference" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->DelReference((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CTranslation(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslation *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CTranslation", 0, 0, 0)) SWIG_fail;
  result = (CTranslation *)new CTranslation();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTranslation, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CTranslation(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslation *arg1 = (CTranslation *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTranslation, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CTranslation" "', argument " "1"" of type '" "CTranslation *""'"); 
  }
  arg1 = reinterpret_cast< CTranslation * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CTranslation_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CTranslation, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CTranslation_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CLanguageScope(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  CLanguageScope *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CLanguageScope" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CLanguageScope" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (CLanguageScope *)new CLanguageScope((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CLanguageScope, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CLanguageScope(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CLanguageScope *arg1 = (CLanguageScope *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CLanguageScope, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CLanguageScope" "', argument " "1"" of type '" "CLanguageScope *""'"); 
  }
  arg1 = reinterpret_cast< CLanguageScope * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CLanguageScope_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CLanguageScope, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CLanguageScope_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CTranslationDomainRefHolder(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  CTranslationDomainRefHolder *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CTranslationDomainRefHolder" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CTranslationDomainRefHolder" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (CTranslationDomainRefHolder *)new CTranslationDomainRefHolder((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTranslationDomainRefHolder, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CTranslationDomainRefHolder(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTranslationDomainRefHolder *arg1 = (CTranslationDomainRefHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTranslationDomainRefHolder, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CTranslationDomainRefHolder" "', argument " "1"" of type '" "CTranslationDomainRefHolder *""'"); 
  }
  arg1 = reinterpret_cast< CTranslationDomainRefHolder * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CTranslationDomainRefHolder_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CTranslationDomainRefHolder, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CTranslationDomainRefHolder_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CDelayedTranslation__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CDelayedTranslation *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CDelayedTranslation *)new CDelayedTranslation();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CDelayedTranslation, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CDelayedTranslation__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  CDelayedTranslation *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CDelayedTranslation" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CDelayedTranslation" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CDelayedTranslation" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CDelayedTranslation" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CDelayedTranslation" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CDelayedTranslation" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (CDelayedTranslation *)new CDelayedTranslation((CString const &)*arg1,(CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CDelayedTranslation, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CDelayedTranslation(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CDelayedTranslation", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CDelayedTranslation__SWIG_0(self, argc, argv);
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_new_CDelayedTranslation__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CDelayedTranslation'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CDelayedTranslation::CDelayedTranslation()\n"
    "    CDelayedTranslation::CDelayedTranslation(CString const &,CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CDelayedTranslation_Resolve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CDelayedTranslation *arg1 = (CDelayedTranslation *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CDelayedTranslation, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDelayedTranslation_Resolve" "', argument " "1"" of type '" "CDelayedTranslation const *""'"); 
  }
  arg1 = reinterpret_cast< CDelayedTranslation * >(argp1);
  result = ((CDelayedTranslation const *)arg1)->Resolve();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CDelayedTranslation(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CDelayedTranslation *arg1 = (CDelayedTranslation *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CDelayedTranslation, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CDelayedTranslation" "', argument " "1"" of type '" "CDelayedTranslation *""'"); 
  }
  arg1 = reinterpret_cast< CDelayedTranslation * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CDelayedTranslation_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CDelayedTranslation, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CDelayedTranslation_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_COptionalTranslation__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  COptionalTranslation *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_COptionalTranslation" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_COptionalTranslation" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (COptionalTranslation *)new COptionalTranslation((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_COptionalTranslation, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_COptionalTranslation__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  COptionalTranslation *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_COptionalTranslation" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = reinterpret_cast< char * >(buf1);
  result = (COptionalTranslation *)new COptionalTranslation((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_COptionalTranslation, SWIG_POINTER_NEW |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_COptionalTranslation__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CDelayedTranslation *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  COptionalTranslation *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CDelayedTranslation,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_COptionalTranslation" "', argument " "1"" of type '" "CDelayedTranslation const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_COptionalTranslation" "', argument " "1"" of type '" "CDelayedTranslation const &""'"); 
  }
  arg1 = reinterpret_cast< CDelayedTranslation * >(argp1);
  result = (COptionalTranslation *)new COptionalTranslation((CDelayedTranslation const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_COptionalTranslation, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_COptionalTranslation(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_COptionalTranslation", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_CDelayedTranslation, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_COptionalTranslation__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_COptionalTranslation__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsCharPtrAndSize(argv[0], 0, NULL, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_COptionalTranslation__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_COptionalTranslation'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    COptionalTranslation::COptionalTranslation(CString const &)\n"
    "    COptionalTranslation::COptionalTranslation(char const *)\n"
    "    COptionalTranslation::COptionalTranslation(CDelayedTranslation const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_COptionalTranslation_Resolve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  COptionalTranslation *arg1 = (COptionalTranslation *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_COptionalTranslation, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "COptionalTranslation_Resolve" "', argument " "1"" of type '" "COptionalTranslation const *""'"); 
  }
  arg1 = reinterpret_cast< COptionalTranslation * >(argp1);
  result = ((COptionalTranslation const *)arg1)->Resolve();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_COptionalTranslation(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  COptionalTranslation *arg1 = (COptionalTranslation *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_COptionalTranslation, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_COptionalTranslation" "', argument " "1"" of type '" "COptionalTranslation *""'"); 
  }
  arg1 = reinterpret_cast< COptionalTranslation * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *COptionalTranslation_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_COptionalTranslation, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *COptionalTranslation_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CCoreTranslationMixin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCoreTranslationMixin *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CCoreTranslationMixin", 0, 0, 0)) SWIG_fail;
  result = (CCoreTranslationMixin *)new CCoreTranslationMixin();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CCoreTranslationMixin, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CCoreTranslationMixin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCoreTranslationMixin *arg1 = (CCoreTranslationMixin *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCoreTranslationMixin, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CCoreTranslationMixin" "', argument " "1"" of type '" "CCoreTranslationMixin *""'"); 
  }
  arg1 = reinterpret_cast< CCoreTranslationMixin * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CCoreTranslationMixin_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CCoreTranslationMixin, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CCoreTranslationMixin_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_SetFdCloseOnExec(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "SetFdCloseOnExec" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = static_cast< int >(val1);
  SetFdCloseOnExec(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN int Swig_var_g_HexDigits_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable g_HexDigits is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_g_HexDigits_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_FromCharPtr(g_HexDigits);
  return pyobj;
}


SWIGINTERN PyObject *_wrap_new_CUtils(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUtils *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CUtils", 0, 0, 0)) SWIG_fail;
  result = (CUtils *)new CUtils();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CUtils, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CUtils(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUtils *arg1 = (CUtils *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUtils, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CUtils" "', argument " "1"" of type '" "CUtils *""'"); 
  }
  arg1 = reinterpret_cast< CUtils * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetIP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  unsigned long arg1 ;
  unsigned long val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_unsigned_SS_long(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "CUtils_GetIP" "', argument " "1"" of type '" "unsigned long""'");
  } 
  arg1 = static_cast< unsigned long >(val1);
  result = CUtils::GetIP(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetLongIP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  unsigned long result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_GetLongIP" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetLongIP" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (unsigned long)CUtils::GetLongIP((CString const &)*arg1);
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetHostName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUtils_GetHostName", 0, 0, 0)) SWIG_fail;
  result = CUtils::GetHostName();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_PrintError(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_PrintError" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_PrintError" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  CUtils::PrintError((CString const &)*arg1);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_PrintMessage__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  bool arg2 ;
  int res1 = SWIG_OLDOBJ ;
  bool val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_PrintMessage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_PrintMessage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUtils_PrintMessage" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  CUtils::PrintMessage((CString const &)*arg1,arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_PrintMessage__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_PrintMessage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_PrintMessage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  CUtils::PrintMessage((CString const &)*arg1);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_PrintMessage(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUtils_PrintMessage", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CUtils_PrintMessage__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CUtils_PrintMessage__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUtils_PrintMessage'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUtils::PrintMessage(CString const &,bool)\n"
    "    CUtils::PrintMessage(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUtils_PrintPrompt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_PrintPrompt" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_PrintPrompt" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  CUtils::PrintPrompt((CString const &)*arg1);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_PrintAction(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_PrintAction" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_PrintAction" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  CUtils::PrintAction((CString const &)*arg1);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_PrintStatus__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  bool arg1 ;
  CString *arg2 = 0 ;
  bool val1 ;
  int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_bool(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "CUtils_PrintStatus" "', argument " "1"" of type '" "bool""'");
  } 
  arg1 = static_cast< bool >(val1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_PrintStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_PrintStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  CUtils::PrintStatus(arg1,(CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_PrintStatus__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  bool arg1 ;
  bool val1 ;
  int ecode1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_bool(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "CUtils_PrintStatus" "', argument " "1"" of type '" "bool""'");
  } 
  arg1 = static_cast< bool >(val1);
  CUtils::PrintStatus(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_PrintStatus(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUtils_PrintStatus", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_bool(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_CUtils_PrintStatus__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_bool(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CUtils_PrintStatus__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUtils_PrintStatus'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUtils::PrintStatus(bool,CString const &)\n"
    "    CUtils::PrintStatus(bool)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUtils_AskSaltedHashPassForConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUtils_AskSaltedHashPassForConfig", 0, 0, 0)) SWIG_fail;
  result = CUtils::AskSaltedHashPassForConfig();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetSalt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUtils_GetSalt", 0, 0, 0)) SWIG_fail;
  result = CUtils::GetSalt();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_SaltedMD5Hash(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUtils_SaltedMD5Hash", 2, 2, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_SaltedMD5Hash" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_SaltedMD5Hash" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_SaltedMD5Hash" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_SaltedMD5Hash" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = CUtils::SaltedMD5Hash((CString const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_SaltedSHA256Hash(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUtils_SaltedSHA256Hash", 2, 2, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_SaltedSHA256Hash" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_SaltedSHA256Hash" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_SaltedSHA256Hash" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_SaltedSHA256Hash" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = CUtils::SaltedSHA256Hash((CString const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_SaltedHash(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUtils_SaltedHash", 2, 2, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_SaltedHash" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_SaltedHash" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_SaltedHash" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_SaltedHash" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = CUtils::SaltedHash((CString const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetPass(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_GetPass" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetPass" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = CUtils::GetPass((CString const &)*arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetInput__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_GetInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sRet");
    }
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUtils_GetInput" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetInput" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CUtils_GetInput" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetInput" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (bool)CUtils::GetInput((CString const &)*arg1,*arg2,(CString const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetInput__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_GetInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sRet");
    }
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUtils_GetInput" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetInput" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)CUtils::GetInput((CString const &)*arg1,*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetInput__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_GetInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sRet");
    }
  }
  result = (bool)CUtils::GetInput((CString const &)*arg1,*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetInput(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUtils_GetInput", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        String* p;
        _v = SWIG_IsOK(SWIG_ConvertPtr(argv[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
      }
      if (_v) {
        return _wrap_CUtils_GetInput__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        String* p;
        _v = SWIG_IsOK(SWIG_ConvertPtr(argv[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CUtils_GetInput__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        String* p;
        _v = SWIG_IsOK(SWIG_ConvertPtr(argv[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CUtils_GetInput__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUtils_GetInput'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUtils::GetInput(CString const &,CString &,CString const &,CString const &)\n"
    "    CUtils::GetInput(CString const &,CString &,CString const &)\n"
    "    CUtils::GetInput(CString const &,CString &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUtils_GetBoolInput__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  bool arg2 ;
  int res1 = SWIG_OLDOBJ ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_GetBoolInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetBoolInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUtils_GetBoolInput" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)CUtils::GetBoolInput((CString const &)*arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetBoolInput__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  bool *arg2 = (bool *) 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_GetBoolInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetBoolInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_bool, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_GetBoolInput" "', argument " "2"" of type '" "bool *""'"); 
  }
  arg2 = reinterpret_cast< bool * >(argp2);
  result = (bool)CUtils::GetBoolInput((CString const &)*arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetBoolInput__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_GetBoolInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetBoolInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CUtils::GetBoolInput((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetBoolInput(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUtils_GetBoolInput", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CUtils_GetBoolInput__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_bool, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CUtils_GetBoolInput__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CUtils_GetBoolInput__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUtils_GetBoolInput'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUtils::GetBoolInput(CString const &,bool)\n"
    "    CUtils::GetBoolInput(CString const &,bool *)\n"
    "    CUtils::GetBoolInput(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUtils_GetNumInput__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  unsigned int *arg2 = 0 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_GetNumInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetNumInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_unsigned_int,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_GetNumInput" "', argument " "2"" of type '" "unsigned int &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetNumInput" "', argument " "2"" of type '" "unsigned int &""'"); 
  }
  arg2 = reinterpret_cast< unsigned int * >(argp2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CUtils_GetNumInput" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CUtils_GetNumInput" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CUtils_GetNumInput" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  result = (bool)CUtils::GetNumInput((CString const &)*arg1,*arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetNumInput__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  unsigned int *arg2 = 0 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_GetNumInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetNumInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_unsigned_int,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_GetNumInput" "', argument " "2"" of type '" "unsigned int &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetNumInput" "', argument " "2"" of type '" "unsigned int &""'"); 
  }
  arg2 = reinterpret_cast< unsigned int * >(argp2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CUtils_GetNumInput" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CUtils_GetNumInput" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  result = (bool)CUtils::GetNumInput((CString const &)*arg1,*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetNumInput__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  unsigned int *arg2 = 0 ;
  unsigned int arg3 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_GetNumInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetNumInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_unsigned_int,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_GetNumInput" "', argument " "2"" of type '" "unsigned int &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetNumInput" "', argument " "2"" of type '" "unsigned int &""'"); 
  }
  arg2 = reinterpret_cast< unsigned int * >(argp2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CUtils_GetNumInput" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  result = (bool)CUtils::GetNumInput((CString const &)*arg1,*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetNumInput__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  unsigned int *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_GetNumInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetNumInput" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_unsigned_int,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_GetNumInput" "', argument " "2"" of type '" "unsigned int &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetNumInput" "', argument " "2"" of type '" "unsigned int &""'"); 
  }
  arg2 = reinterpret_cast< unsigned int * >(argp2);
  result = (bool)CUtils::GetNumInput((CString const &)*arg1,*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetNumInput(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUtils_GetNumInput", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_unsigned_int, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CUtils_GetNumInput__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_unsigned_int, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CUtils_GetNumInput__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_unsigned_int, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CUtils_GetNumInput__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_unsigned_int, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_unsigned_SS_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_CUtils_GetNumInput__SWIG_0(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUtils_GetNumInput'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUtils::GetNumInput(CString const &,unsigned int &,unsigned int,unsigned int,unsigned int)\n"
    "    CUtils::GetNumInput(CString const &,unsigned int &,unsigned int,unsigned int)\n"
    "    CUtils::GetNumInput(CString const &,unsigned int &,unsigned int)\n"
    "    CUtils::GetNumInput(CString const &,unsigned int &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUtils_GetTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  timeval result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUtils_GetTime", 0, 0, 0)) SWIG_fail;
  result = CUtils::GetTime();
  resultobj = SWIG_NewPointerObj((new timeval(result)), SWIGTYPE_p_timeval, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetMillTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  unsigned long long result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUtils_GetMillTime", 0, 0, 0)) SWIG_fail;
  result = (unsigned long long)CUtils::GetMillTime();
  resultobj = SWIG_From_unsigned_SS_long_SS_long(static_cast< unsigned long long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_CTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  time_t arg1 ;
  CString *arg2 = 0 ;
  void *argp1 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUtils_CTime", 2, 2, swig_obj)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_time_t,  0  | 0);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_CTime" "', argument " "1"" of type '" "time_t""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_CTime" "', argument " "1"" of type '" "time_t""'");
    } else {
      time_t * temp = reinterpret_cast< time_t * >(argp1);
      arg1 = *temp;
      if (SWIG_IsNewObj(res1)) delete temp;
    }
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_CTime" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_CTime" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = CUtils::CTime(SWIG_STD_MOVE(arg1),(CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_FormatTime__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  time_t arg1 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_time_t,  0  | 0);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_FormatTime" "', argument " "1"" of type '" "time_t""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_FormatTime" "', argument " "1"" of type '" "time_t""'");
    } else {
      time_t * temp = reinterpret_cast< time_t * >(argp1);
      arg1 = *temp;
      if (SWIG_IsNewObj(res1)) delete temp;
    }
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_FormatTime" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_FormatTime" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUtils_FormatTime" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_FormatTime" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = CUtils::FormatTime(SWIG_STD_MOVE(arg1),(CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_FormatTime__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  timeval *arg1 = 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_timeval,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_FormatTime" "', argument " "1"" of type '" "timeval const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_FormatTime" "', argument " "1"" of type '" "timeval const &""'"); 
  }
  arg1 = reinterpret_cast< timeval * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_FormatTime" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_FormatTime" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUtils_FormatTime" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_FormatTime" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = CUtils::FormatTime((timeval const &)*arg1,(CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_FormatTime(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUtils_FormatTime", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_time_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CUtils_FormatTime__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_timeval, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CUtils_FormatTime__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUtils_FormatTime'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUtils::FormatTime(time_t,CString const &,CString const &)\n"
    "    CUtils::FormatTime(timeval const &,CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUtils_FormatServerTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  timeval *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_timeval,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_FormatServerTime" "', argument " "1"" of type '" "timeval const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_FormatServerTime" "', argument " "1"" of type '" "timeval const &""'"); 
  }
  arg1 = reinterpret_cast< timeval * >(argp1);
  result = CUtils::FormatServerTime((timeval const &)*arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_ParseServerTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  timeval result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_ParseServerTime" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_ParseServerTime" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = CUtils::ParseServerTime((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj((new timeval(result)), SWIGTYPE_p_timeval, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetTimezones(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  SCString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUtils_GetTimezones", 0, 0, 0)) SWIG_fail;
  result = CUtils::GetTimezones();
  resultobj = swig::from(static_cast< std::set< CString,std::less< CString >,std::allocator< CString > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetEncodings(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  SCString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUtils_GetEncodings", 0, 0, 0)) SWIG_fail;
  result = CUtils::GetEncodings();
  resultobj = swig::from(static_cast< std::set< CString,std::less< CString >,std::allocator< CString > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_CheckCIDR(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUtils_CheckCIDR", 2, 2, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_CheckCIDR" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_CheckCIDR" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_CheckCIDR" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_CheckCIDR" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)CUtils::CheckCIDR((CString const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_GetMessageTags(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  MCString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUtils_GetMessageTags" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_GetMessageTags" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = CUtils::GetMessageTags((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj((new MCString(result)), SWIGTYPE_p_MCString, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUtils_SetMessageTags(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  MCString *arg2 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUtils_SetMessageTags", 2, 2, swig_obj)) SWIG_fail;
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[0], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg1 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 1 sLine");
    }
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_MCString,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUtils_SetMessageTags" "', argument " "2"" of type '" "MCString const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUtils_SetMessageTags" "', argument " "2"" of type '" "MCString const &""'"); 
  }
  arg2 = reinterpret_cast< MCString * >(argp2);
  CUtils::SetMessageTags(*arg1,(MCString const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CUtils_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CUtils, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CUtils_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CException(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CException::EType arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  CException *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_CException" "', argument " "1"" of type '" "CException::EType""'");
  } 
  arg1 = static_cast< CException::EType >(val1);
  result = (CException *)new CException(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CException, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CException(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CException *arg1 = (CException *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CException, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CException" "', argument " "1"" of type '" "CException *""'"); 
  }
  arg1 = reinterpret_cast< CException * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CException_GetType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CException *arg1 = (CException *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CException::EType result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CException, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CException_GetType" "', argument " "1"" of type '" "CException const *""'"); 
  }
  arg1 = reinterpret_cast< CException * >(argp1);
  result = (CException::EType)((CException const *)arg1)->GetType();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CException_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CException, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CException_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CTable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTable *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CTable", 0, 0, 0)) SWIG_fail;
  result = (CTable *)new CTable();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTable, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CTable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTable *arg1 = (CTable *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTable, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CTable" "', argument " "1"" of type '" "CTable *""'"); 
  }
  arg1 = reinterpret_cast< CTable * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTable_AddColumn(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTable *arg1 = (CTable *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTable_AddColumn", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTable_AddColumn" "', argument " "1"" of type '" "CTable *""'"); 
  }
  arg1 = reinterpret_cast< CTable * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTable_AddColumn" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTable_AddColumn" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->AddColumn((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTable_SetStyle(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTable *arg1 = (CTable *) 0 ;
  CTable::EStyle arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTable_SetStyle", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTable_SetStyle" "', argument " "1"" of type '" "CTable *""'"); 
  }
  arg1 = reinterpret_cast< CTable * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CTable_SetStyle" "', argument " "2"" of type '" "CTable::EStyle""'");
  } 
  arg2 = static_cast< CTable::EStyle >(val2);
  result = (bool)(arg1)->SetStyle(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTable_AddRow(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTable *arg1 = (CTable *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::vector< CString,std::allocator< CString > > >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTable_AddRow" "', argument " "1"" of type '" "CTable *""'"); 
  }
  arg1 = reinterpret_cast< CTable * >(argp1);
  result = (arg1)->AddRow();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTable_SetCell__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTable *arg1 = (CTable *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  std::vector< std::vector< CString,std::allocator< CString > > >::size_type arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  size_t val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTable_SetCell" "', argument " "1"" of type '" "CTable *""'"); 
  }
  arg1 = reinterpret_cast< CTable * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTable_SetCell" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTable_SetCell" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CTable_SetCell" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTable_SetCell" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_size_t(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CTable_SetCell" "', argument " "4"" of type '" "std::vector< std::vector< CString,std::allocator< CString > > >::size_type""'");
  } 
  arg4 = static_cast< std::vector< std::vector< CString,std::allocator< CString > > >::size_type >(val4);
  result = (bool)(arg1)->SetCell((CString const &)*arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTable_SetCell__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTable *arg1 = (CTable *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTable_SetCell" "', argument " "1"" of type '" "CTable *""'"); 
  }
  arg1 = reinterpret_cast< CTable * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTable_SetCell" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTable_SetCell" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CTable_SetCell" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTable_SetCell" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->SetCell((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTable_SetCell(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CTable_SetCell", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTable, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CTable_SetCell__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTable, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_size_t(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CTable_SetCell__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CTable_SetCell'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CTable::SetCell(CString const &,CString const &,std::vector< std::vector< CString,std::allocator< CString > > >::size_type)\n"
    "    CTable::SetCell(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CTable_GetLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTable *arg1 = (CTable *) 0 ;
  unsigned int arg2 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTable_GetLine", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTable_GetLine" "', argument " "1"" of type '" "CTable const *""'"); 
  }
  arg1 = reinterpret_cast< CTable * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CTable_GetLine" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sLine");
    }
  }
  result = (bool)((CTable const *)arg1)->GetLine(arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTable_GetColumnWidth(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTable *arg1 = (CTable *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  CString::size_type result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTable_GetColumnWidth", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTable_GetColumnWidth" "', argument " "1"" of type '" "CTable const *""'"); 
  }
  arg1 = reinterpret_cast< CTable * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CTable_GetColumnWidth" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = ((CTable const *)arg1)->GetColumnWidth(arg2);
  resultobj = SWIG_NewPointerObj((new CString::size_type(result)), SWIGTYPE_p_CString__size_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTable_Clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTable *arg1 = (CTable *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTable_Clear" "', argument " "1"" of type '" "CTable *""'"); 
  }
  arg1 = reinterpret_cast< CTable * >(argp1);
  (arg1)->Clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTable_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTable *arg1 = (CTable *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::vector< CString,std::allocator< CString > > >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTable_size" "', argument " "1"" of type '" "CTable const *""'"); 
  }
  arg1 = reinterpret_cast< CTable * >(argp1);
  result = ((CTable const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTable_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTable *arg1 = (CTable *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTable_empty" "', argument " "1"" of type '" "CTable const *""'"); 
  }
  arg1 = reinterpret_cast< CTable * >(argp1);
  result = (bool)((CTable const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CTable_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CTable, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CTable_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CConfigEntry__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CConfigEntry *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CConfigEntry *)new CConfigEntry();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CConfigEntry, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CConfigEntry__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CConfigEntry *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CConfig,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CConfigEntry" "', argument " "1"" of type '" "CConfig const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CConfigEntry" "', argument " "1"" of type '" "CConfig const &""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  result = (CConfigEntry *)new CConfigEntry((CConfig const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CConfigEntry, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CConfigEntry__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfigEntry *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CConfigEntry *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CConfigEntry,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CConfigEntry" "', argument " "1"" of type '" "CConfigEntry const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CConfigEntry" "', argument " "1"" of type '" "CConfigEntry const &""'"); 
  }
  arg1 = reinterpret_cast< CConfigEntry * >(argp1);
  result = (CConfigEntry *)new CConfigEntry((CConfigEntry const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CConfigEntry, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CConfigEntry(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CConfigEntry", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CConfigEntry__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_CConfig, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CConfigEntry__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_CConfigEntry, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CConfigEntry__SWIG_2(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CConfigEntry'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CConfigEntry::CConfigEntry()\n"
    "    CConfigEntry::CConfigEntry(CConfig const &)\n"
    "    CConfigEntry::CConfigEntry(CConfigEntry const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CConfigEntry(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfigEntry *arg1 = (CConfigEntry *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfigEntry, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CConfigEntry" "', argument " "1"" of type '" "CConfigEntry *""'"); 
  }
  arg1 = reinterpret_cast< CConfigEntry * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfigEntry_m_pSubConfig_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfigEntry *arg1 = (CConfigEntry *) 0 ;
  CConfig *arg2 = (CConfig *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CConfigEntry_m_pSubConfig_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfigEntry, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfigEntry_m_pSubConfig_set" "', argument " "1"" of type '" "CConfigEntry *""'"); 
  }
  arg1 = reinterpret_cast< CConfigEntry * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CConfig, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfigEntry_m_pSubConfig_set" "', argument " "2"" of type '" "CConfig *""'"); 
  }
  arg2 = reinterpret_cast< CConfig * >(argp2);
  if (arg1) (arg1)->m_pSubConfig = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfigEntry_m_pSubConfig_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfigEntry *arg1 = (CConfigEntry *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CConfig *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfigEntry, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfigEntry_m_pSubConfig_get" "', argument " "1"" of type '" "CConfigEntry *""'"); 
  }
  arg1 = reinterpret_cast< CConfigEntry * >(argp1);
  result = (CConfig *) ((arg1)->m_pSubConfig);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CConfig, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CConfigEntry_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CConfigEntry, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CConfigEntry_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfig *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CConfig", 0, 0, 0)) SWIG_fail;
  result = (CConfig *)new CConfig();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CConfig, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_BeginEntries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::map< CString,std::vector< CString,std::allocator< CString > > >::const_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_BeginEntries" "', argument " "1"" of type '" "CConfig const *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  result = ((CConfig const *)arg1)->BeginEntries();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const CConfig::EntryMapIterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_EndEntries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::map< CString,std::vector< CString,std::allocator< CString > > >::const_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_EndEntries" "', argument " "1"" of type '" "CConfig const *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  result = ((CConfig const *)arg1)->EndEntries();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const CConfig::EntryMapIterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_BeginSubConfigs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::map< CString,std::vector< std::pair< CString,CConfigEntry >,std::allocator< std::pair< CString,CConfigEntry > > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< std::pair< CString,CConfigEntry >,std::allocator< std::pair< CString,CConfigEntry > > > > > >::const_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_BeginSubConfigs" "', argument " "1"" of type '" "CConfig const *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  result = ((CConfig const *)arg1)->BeginSubConfigs();
  resultobj = SWIG_NewPointerObj((new CConfig::SubConfigMapIterator(result)), SWIGTYPE_p_std__mapT_CString_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_t_t_t__const_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_EndSubConfigs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::map< CString,std::vector< std::pair< CString,CConfigEntry >,std::allocator< std::pair< CString,CConfigEntry > > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< std::pair< CString,CConfigEntry >,std::allocator< std::pair< CString,CConfigEntry > > > > > >::const_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_EndSubConfigs" "', argument " "1"" of type '" "CConfig const *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  result = ((CConfig const *)arg1)->EndSubConfigs();
  resultobj = SWIG_NewPointerObj((new CConfig::SubConfigMapIterator(result)), SWIGTYPE_p_std__mapT_CString_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_t_t_t__const_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_AddKeyValuePair(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CConfig_AddKeyValuePair", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_AddKeyValuePair" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_AddKeyValuePair" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_AddKeyValuePair" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_AddKeyValuePair" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_AddKeyValuePair" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->AddKeyValuePair((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_AddSubConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CConfig arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CConfig_AddSubConfig", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_AddSubConfig" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_AddSubConfig" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_AddSubConfig" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_AddSubConfig" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_AddSubConfig" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CConfig,  0  | 0);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CConfig_AddSubConfig" "', argument " "4"" of type '" "CConfig""'"); 
    }  
    if (!argp4) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_AddSubConfig" "', argument " "4"" of type '" "CConfig""'");
    } else {
      CConfig * temp = reinterpret_cast< CConfig * >(argp4);
      arg4 = *temp;
      if (SWIG_IsNewObj(res4)) delete temp;
    }
  }
  result = (bool)(arg1)->AddSubConfig((CString const &)*arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindStringVector__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  VCString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindStringVector" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindStringVector" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindStringVector" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_CString_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_FindStringVector" "', argument " "3"" of type '" "VCString &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindStringVector" "', argument " "3"" of type '" "VCString &""'"); 
  }
  arg3 = reinterpret_cast< VCString * >(argp3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CConfig_FindStringVector" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->FindStringVector((CString const &)*arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindStringVector__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  VCString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindStringVector" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindStringVector" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindStringVector" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_CString_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_FindStringVector" "', argument " "3"" of type '" "VCString &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindStringVector" "', argument " "3"" of type '" "VCString &""'"); 
  }
  arg3 = reinterpret_cast< VCString * >(argp3);
  result = (bool)(arg1)->FindStringVector((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindStringVector(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CConfig_FindStringVector", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CConfig_FindStringVector__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CConfig_FindStringVector__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CConfig_FindStringVector'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CConfig::FindStringVector(CString const &,VCString &,bool)\n"
    "    CConfig::FindStringVector(CString const &,VCString &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CConfig_FindStringEntry__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindStringEntry" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindStringEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindStringEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sRes");
    }
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CConfig_FindStringEntry" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindStringEntry" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (bool)(arg1)->FindStringEntry((CString const &)*arg2,*arg3,(CString const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindStringEntry__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindStringEntry" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindStringEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindStringEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sRes");
    }
  }
  result = (bool)(arg1)->FindStringEntry((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindStringEntry(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CConfig_FindStringEntry", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          return _wrap_CConfig_FindStringEntry__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CConfig_FindStringEntry__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CConfig_FindStringEntry'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CConfig::FindStringEntry(CString const &,CString &,CString const &)\n"
    "    CConfig::FindStringEntry(CString const &,CString &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CConfig_FindBoolEntry__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  bool *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindBoolEntry" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindBoolEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindBoolEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_bool,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_FindBoolEntry" "', argument " "3"" of type '" "bool &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindBoolEntry" "', argument " "3"" of type '" "bool &""'"); 
  }
  arg3 = reinterpret_cast< bool * >(argp3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CConfig_FindBoolEntry" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->FindBoolEntry((CString const &)*arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindBoolEntry__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  bool *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindBoolEntry" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindBoolEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindBoolEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_bool,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_FindBoolEntry" "', argument " "3"" of type '" "bool &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindBoolEntry" "', argument " "3"" of type '" "bool &""'"); 
  }
  arg3 = reinterpret_cast< bool * >(argp3);
  result = (bool)(arg1)->FindBoolEntry((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindBoolEntry(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CConfig_FindBoolEntry", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_bool, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CConfig_FindBoolEntry__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_bool, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CConfig_FindBoolEntry__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CConfig_FindBoolEntry'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CConfig::FindBoolEntry(CString const &,bool &,bool)\n"
    "    CConfig::FindBoolEntry(CString const &,bool &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CConfig_FindUIntEntry__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  unsigned int *arg3 = 0 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindUIntEntry" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindUIntEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindUIntEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_unsigned_int,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_FindUIntEntry" "', argument " "3"" of type '" "unsigned int &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindUIntEntry" "', argument " "3"" of type '" "unsigned int &""'"); 
  }
  arg3 = reinterpret_cast< unsigned int * >(argp3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CConfig_FindUIntEntry" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  result = (bool)(arg1)->FindUIntEntry((CString const &)*arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindUIntEntry__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  unsigned int *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindUIntEntry" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindUIntEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindUIntEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_unsigned_int,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_FindUIntEntry" "', argument " "3"" of type '" "unsigned int &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindUIntEntry" "', argument " "3"" of type '" "unsigned int &""'"); 
  }
  arg3 = reinterpret_cast< unsigned int * >(argp3);
  result = (bool)(arg1)->FindUIntEntry((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindUIntEntry(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CConfig_FindUIntEntry", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_unsigned_int, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CConfig_FindUIntEntry__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_unsigned_int, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CConfig_FindUIntEntry__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CConfig_FindUIntEntry'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CConfig::FindUIntEntry(CString const &,unsigned int &,unsigned int)\n"
    "    CConfig::FindUIntEntry(CString const &,unsigned int &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CConfig_FindUShortEntry__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  unsigned short *arg3 = 0 ;
  unsigned short arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  unsigned short val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindUShortEntry" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindUShortEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindUShortEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_unsigned_short,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_FindUShortEntry" "', argument " "3"" of type '" "unsigned short &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindUShortEntry" "', argument " "3"" of type '" "unsigned short &""'"); 
  }
  arg3 = reinterpret_cast< unsigned short * >(argp3);
  ecode4 = SWIG_AsVal_unsigned_SS_short(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CConfig_FindUShortEntry" "', argument " "4"" of type '" "unsigned short""'");
  } 
  arg4 = static_cast< unsigned short >(val4);
  result = (bool)(arg1)->FindUShortEntry((CString const &)*arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindUShortEntry__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  unsigned short *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindUShortEntry" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindUShortEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindUShortEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_unsigned_short,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_FindUShortEntry" "', argument " "3"" of type '" "unsigned short &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindUShortEntry" "', argument " "3"" of type '" "unsigned short &""'"); 
  }
  arg3 = reinterpret_cast< unsigned short * >(argp3);
  result = (bool)(arg1)->FindUShortEntry((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindUShortEntry(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CConfig_FindUShortEntry", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_unsigned_short, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CConfig_FindUShortEntry__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_unsigned_short, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_short(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CConfig_FindUShortEntry__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CConfig_FindUShortEntry'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CConfig::FindUShortEntry(CString const &,unsigned short &,unsigned short)\n"
    "    CConfig::FindUShortEntry(CString const &,unsigned short &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CConfig_FindDoubleEntry__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  double *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindDoubleEntry" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindDoubleEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindDoubleEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_double,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_FindDoubleEntry" "', argument " "3"" of type '" "double &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindDoubleEntry" "', argument " "3"" of type '" "double &""'"); 
  }
  arg3 = reinterpret_cast< double * >(argp3);
  ecode4 = SWIG_AsVal_double(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CConfig_FindDoubleEntry" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  result = (bool)(arg1)->FindDoubleEntry((CString const &)*arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindDoubleEntry__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  double *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindDoubleEntry" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindDoubleEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindDoubleEntry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_double,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_FindDoubleEntry" "', argument " "3"" of type '" "double &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindDoubleEntry" "', argument " "3"" of type '" "double &""'"); 
  }
  arg3 = reinterpret_cast< double * >(argp3);
  result = (bool)(arg1)->FindDoubleEntry((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindDoubleEntry(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CConfig_FindDoubleEntry", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_double, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CConfig_FindDoubleEntry__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_double, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_double(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CConfig_FindDoubleEntry__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CConfig_FindDoubleEntry'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CConfig::FindDoubleEntry(CString const &,double &,double)\n"
    "    CConfig::FindDoubleEntry(CString const &,double &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CConfig_FindSubConfig__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  CConfig::SubConfig *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindSubConfig" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindSubConfig" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindSubConfig" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_FindSubConfig" "', argument " "3"" of type '" "CConfig::SubConfig &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindSubConfig" "', argument " "3"" of type '" "CConfig::SubConfig &""'"); 
  }
  arg3 = reinterpret_cast< CConfig::SubConfig * >(argp3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CConfig_FindSubConfig" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->FindSubConfig((CString const &)*arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindSubConfig__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CString *arg2 = 0 ;
  CConfig::SubConfig *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_FindSubConfig" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_FindSubConfig" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindSubConfig" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CConfig_FindSubConfig" "', argument " "3"" of type '" "CConfig::SubConfig &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_FindSubConfig" "', argument " "3"" of type '" "CConfig::SubConfig &""'"); 
  }
  arg3 = reinterpret_cast< CConfig::SubConfig * >(argp3);
  result = (bool)(arg1)->FindSubConfig((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_FindSubConfig(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CConfig_FindSubConfig", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CConfig_FindSubConfig__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CConfig_FindSubConfig__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CConfig_FindSubConfig'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CConfig::FindSubConfig(CString const &,CConfig::SubConfig &,bool)\n"
    "    CConfig::FindSubConfig(CString const &,CConfig::SubConfig &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CConfig_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_empty" "', argument " "1"" of type '" "CConfig const *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  result = (bool)((CConfig const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_Parse(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CFile *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CConfig_Parse", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_Parse" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CFile,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_Parse" "', argument " "2"" of type '" "CFile &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_Parse" "', argument " "2"" of type '" "CFile &""'"); 
  }
  arg2 = reinterpret_cast< CFile * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sErrorMsg");
    }
  }
  result = (bool)(arg1)->Parse(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_Write__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CFile *arg2 = 0 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_Write" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CFile,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_Write" "', argument " "2"" of type '" "CFile &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_Write" "', argument " "2"" of type '" "CFile &""'"); 
  }
  arg2 = reinterpret_cast< CFile * >(argp2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CConfig_Write" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  (arg1)->Write(*arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_Write__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  CFile *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CConfig_Write" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CFile,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CConfig_Write" "', argument " "2"" of type '" "CFile &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CConfig_Write" "', argument " "2"" of type '" "CFile &""'"); 
  }
  arg2 = reinterpret_cast< CFile * >(argp2);
  (arg1)->Write(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CConfig_Write(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CConfig_Write", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CFile, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CConfig_Write__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CConfig, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CFile, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CConfig_Write__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CConfig_Write'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CConfig::Write(CFile &,unsigned int)\n"
    "    CConfig::Write(CFile &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CConfig" "', argument " "1"" of type '" "CConfig *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CConfig_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CConfig, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CConfig_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CSCharBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  size_t arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  CSCharBuffer *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_CSCharBuffer" "', argument " "1"" of type '" "size_t""'");
  } 
  arg1 = static_cast< size_t >(val1);
  result = (CSCharBuffer *)new CSCharBuffer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSCharBuffer, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CSCharBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSCharBuffer *arg1 = (CSCharBuffer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSCharBuffer, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CSCharBuffer" "', argument " "1"" of type '" "CSCharBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CSCharBuffer * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSCharBuffer___call__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSCharBuffer *arg1 = (CSCharBuffer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSCharBuffer___call__", 1, 1, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSCharBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSCharBuffer___call__" "', argument " "1"" of type '" "CSCharBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CSCharBuffer * >(argp1);
  result = (char *)(arg1)->operator ()();
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CSCharBuffer_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CSCharBuffer, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CSCharBuffer_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CSSockAddr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSSockAddr *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CSSockAddr", 0, 0, 0)) SWIG_fail;
  result = (CSSockAddr *)new CSSockAddr();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSSockAddr, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CSSockAddr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSSockAddr *arg1 = (CSSockAddr *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSSockAddr, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CSSockAddr" "', argument " "1"" of type '" "CSSockAddr *""'"); 
  }
  arg1 = reinterpret_cast< CSSockAddr * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSSockAddr_SinFamily(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSSockAddr *arg1 = (CSSockAddr *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSSockAddr, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSSockAddr_SinFamily" "', argument " "1"" of type '" "CSSockAddr *""'"); 
  }
  arg1 = reinterpret_cast< CSSockAddr * >(argp1);
  (arg1)->SinFamily();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSSockAddr_SinPort(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSSockAddr *arg1 = (CSSockAddr *) 0 ;
  uint16_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSSockAddr_SinPort", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSSockAddr, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSSockAddr_SinPort" "', argument " "1"" of type '" "CSSockAddr *""'"); 
  }
  arg1 = reinterpret_cast< CSSockAddr * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSSockAddr_SinPort" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  (arg1)->SinPort(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSSockAddr_SetIPv6(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSSockAddr *arg1 = (CSSockAddr *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSSockAddr_SetIPv6", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSSockAddr, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSSockAddr_SetIPv6" "', argument " "1"" of type '" "CSSockAddr *""'"); 
  }
  arg1 = reinterpret_cast< CSSockAddr * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSSockAddr_SetIPv6" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetIPv6(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSSockAddr_GetIPv6(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSSockAddr *arg1 = (CSSockAddr *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSSockAddr, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSSockAddr_GetIPv6" "', argument " "1"" of type '" "CSSockAddr const *""'"); 
  }
  arg1 = reinterpret_cast< CSSockAddr * >(argp1);
  result = (bool)((CSSockAddr const *)arg1)->GetIPv6();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSSockAddr_GetSockAddrLen(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSSockAddr *arg1 = (CSSockAddr *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  socklen_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSSockAddr, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSSockAddr_GetSockAddrLen" "', argument " "1"" of type '" "CSSockAddr *""'"); 
  }
  arg1 = reinterpret_cast< CSSockAddr * >(argp1);
  result = (arg1)->GetSockAddrLen();
  resultobj = SWIG_NewPointerObj((new socklen_t(result)), SWIGTYPE_p_socklen_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSSockAddr_GetSockAddr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSSockAddr *arg1 = (CSSockAddr *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  sockaddr_in *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSSockAddr, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSSockAddr_GetSockAddr" "', argument " "1"" of type '" "CSSockAddr *""'"); 
  }
  arg1 = reinterpret_cast< CSSockAddr * >(argp1);
  result = (sockaddr_in *)(arg1)->GetSockAddr();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_sockaddr_in, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSSockAddr_GetAddr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSSockAddr *arg1 = (CSSockAddr *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  in_addr *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSSockAddr, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSSockAddr_GetAddr" "', argument " "1"" of type '" "CSSockAddr *""'"); 
  }
  arg1 = reinterpret_cast< CSSockAddr * >(argp1);
  result = (in_addr *)(arg1)->GetAddr();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_in_addr, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSSockAddr_SetAFRequire(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSSockAddr *arg1 = (CSSockAddr *) 0 ;
  CSSockAddr::EAFRequire arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSSockAddr_SetAFRequire", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSSockAddr, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSSockAddr_SetAFRequire" "', argument " "1"" of type '" "CSSockAddr *""'"); 
  }
  arg1 = reinterpret_cast< CSSockAddr * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSSockAddr_SetAFRequire" "', argument " "2"" of type '" "CSSockAddr::EAFRequire""'");
  } 
  arg2 = static_cast< CSSockAddr::EAFRequire >(val2);
  (arg1)->SetAFRequire(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSSockAddr_GetAFRequire(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSSockAddr *arg1 = (CSSockAddr *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CSSockAddr::EAFRequire result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSSockAddr, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSSockAddr_GetAFRequire" "', argument " "1"" of type '" "CSSockAddr const *""'"); 
  }
  arg1 = reinterpret_cast< CSSockAddr * >(argp1);
  result = (CSSockAddr::EAFRequire)((CSSockAddr const *)arg1)->GetAFRequire();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CSSockAddr_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CSSockAddr, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CSSockAddr_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CGetAddrInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  Csock *arg2 = (Csock *) 0 ;
  CSSockAddr *arg3 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  CGetAddrInfo *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CGetAddrInfo", 3, 3, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CGetAddrInfo" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CGetAddrInfo" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CGetAddrInfo" "', argument " "2"" of type '" "Csock *""'"); 
  }
  arg2 = reinterpret_cast< Csock * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CSSockAddr,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CGetAddrInfo" "', argument " "3"" of type '" "CSSockAddr &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CGetAddrInfo" "', argument " "3"" of type '" "CSSockAddr &""'"); 
  }
  arg3 = reinterpret_cast< CSSockAddr * >(argp3);
  result = (CGetAddrInfo *)new CGetAddrInfo((CString const &)*arg1,arg2,*arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CGetAddrInfo, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CGetAddrInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CGetAddrInfo *arg1 = (CGetAddrInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CGetAddrInfo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CGetAddrInfo" "', argument " "1"" of type '" "CGetAddrInfo *""'"); 
  }
  arg1 = reinterpret_cast< CGetAddrInfo * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CGetAddrInfo_Init(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CGetAddrInfo *arg1 = (CGetAddrInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CGetAddrInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CGetAddrInfo_Init" "', argument " "1"" of type '" "CGetAddrInfo *""'"); 
  }
  arg1 = reinterpret_cast< CGetAddrInfo * >(argp1);
  (arg1)->Init();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CGetAddrInfo_Process(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CGetAddrInfo *arg1 = (CGetAddrInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CGetAddrInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CGetAddrInfo_Process" "', argument " "1"" of type '" "CGetAddrInfo *""'"); 
  }
  arg1 = reinterpret_cast< CGetAddrInfo * >(argp1);
  result = (int)(arg1)->Process();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CGetAddrInfo_Finish(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CGetAddrInfo *arg1 = (CGetAddrInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CGetAddrInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CGetAddrInfo_Finish" "', argument " "1"" of type '" "CGetAddrInfo *""'"); 
  }
  arg1 = reinterpret_cast< CGetAddrInfo * >(argp1);
  result = (int)(arg1)->Finish();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CGetAddrInfo_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CGetAddrInfo, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CGetAddrInfo_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CS_GetAddrInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  Csock *arg2 = (Csock *) 0 ;
  CSSockAddr *arg3 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CS_GetAddrInfo", 3, 3, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CS_GetAddrInfo" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CS_GetAddrInfo" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CS_GetAddrInfo" "', argument " "2"" of type '" "Csock *""'"); 
  }
  arg2 = reinterpret_cast< Csock * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CSSockAddr,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CS_GetAddrInfo" "', argument " "3"" of type '" "CSSockAddr &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CS_GetAddrInfo" "', argument " "3"" of type '" "CSSockAddr &""'"); 
  }
  arg3 = reinterpret_cast< CSSockAddr * >(argp3);
  result = (int)CS_GetAddrInfo((CString const &)*arg1,arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_GetCsockSSLIdx(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "GetCsockSSLIdx", 0, 0, 0)) SWIG_fail;
  result = (int)GetCsockSSLIdx();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN int Swig_var_CS_BLOCKSIZE_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable CS_BLOCKSIZE is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_CS_BLOCKSIZE_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_From_long(static_cast< long >(CS_BLOCKSIZE));
  return pyobj;
}


SWIGINTERN PyObject *_wrap_InitCsocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "InitCsocket", 0, 0, 0)) SWIG_fail;
  result = (bool)InitCsocket();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ShutdownCsocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "ShutdownCsocket", 0, 0, 0)) SWIG_fail;
  ShutdownCsocket();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GetSockError(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "GetSockError", 0, 0, 0)) SWIG_fail;
  result = (int)GetSockError();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TFD_ZERO(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  fd_set *arg1 = (fd_set *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_fd_set, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TFD_ZERO" "', argument " "1"" of type '" "fd_set *""'"); 
  }
  arg1 = reinterpret_cast< fd_set * >(argp1);
  TFD_ZERO(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TFD_SET(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cs_sock_t arg1 ;
  fd_set *arg2 = (fd_set *) 0 ;
  int val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "TFD_SET", 2, 2, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "TFD_SET" "', argument " "1"" of type '" "cs_sock_t""'");
  } 
  arg1 = static_cast< cs_sock_t >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_fd_set, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TFD_SET" "', argument " "2"" of type '" "fd_set *""'"); 
  }
  arg2 = reinterpret_cast< fd_set * >(argp2);
  TFD_SET(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TFD_ISSET(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cs_sock_t arg1 ;
  fd_set *arg2 = (fd_set *) 0 ;
  int val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "TFD_ISSET", 2, 2, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "TFD_ISSET" "', argument " "1"" of type '" "cs_sock_t""'");
  } 
  arg1 = static_cast< cs_sock_t >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_fd_set, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TFD_ISSET" "', argument " "2"" of type '" "fd_set *""'"); 
  }
  arg2 = reinterpret_cast< fd_set * >(argp2);
  result = (bool)TFD_ISSET(arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TFD_CLR(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  cs_sock_t arg1 ;
  fd_set *arg2 = (fd_set *) 0 ;
  int val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "TFD_CLR", 2, 2, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "TFD_CLR" "', argument " "1"" of type '" "cs_sock_t""'");
  } 
  arg1 = static_cast< cs_sock_t >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_fd_set, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TFD_CLR" "', argument " "2"" of type '" "fd_set *""'"); 
  }
  arg2 = reinterpret_cast< fd_set * >(argp2);
  TFD_CLR(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap___Perror(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  char *arg2 = (char *) 0 ;
  uint32_t arg3 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "__Perror", 3, 3, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__Perror" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "__Perror" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__Perror" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "__Perror" "', argument " "3"" of type '" "uint32_t""'");
  } 
  arg3 = static_cast< uint32_t >(val3);
  __Perror((CString const &)*arg1,(char const *)arg2,SWIG_STD_MOVE(arg3));
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CCron(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CCron", 0, 0, 0)) SWIG_fail;
  result = (CCron *)new CCron();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CCron, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CCron(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CCron" "', argument " "1"" of type '" "CCron *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_run(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  timeval *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CCron_run", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_run" "', argument " "1"" of type '" "CCron *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_timeval,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CCron_run" "', argument " "2"" of type '" "timeval &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CCron_run" "', argument " "2"" of type '" "timeval &""'"); 
  }
  arg2 = reinterpret_cast< timeval * >(argp2);
  (arg1)->run(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_StartMaxCycles__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  double arg2 ;
  uint32_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_StartMaxCycles" "', argument " "1"" of type '" "CCron *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CCron_StartMaxCycles" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CCron_StartMaxCycles" "', argument " "3"" of type '" "uint32_t""'");
  } 
  arg3 = static_cast< uint32_t >(val3);
  (arg1)->StartMaxCycles(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_StartMaxCycles__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  timeval *arg2 = 0 ;
  uint32_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_StartMaxCycles" "', argument " "1"" of type '" "CCron *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_timeval,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CCron_StartMaxCycles" "', argument " "2"" of type '" "timeval const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CCron_StartMaxCycles" "', argument " "2"" of type '" "timeval const &""'"); 
  }
  arg2 = reinterpret_cast< timeval * >(argp2);
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CCron_StartMaxCycles" "', argument " "3"" of type '" "uint32_t""'");
  } 
  arg3 = static_cast< uint32_t >(val3);
  (arg1)->StartMaxCycles((timeval const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_StartMaxCycles(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CCron_StartMaxCycles", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CCron, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_timeval, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_long(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CCron_StartMaxCycles__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CCron, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_double(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_long(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CCron_StartMaxCycles__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CCron_StartMaxCycles'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CCron::StartMaxCycles(double,uint32_t)\n"
    "    CCron::StartMaxCycles(timeval const &,uint32_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CCron_Start__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_Start" "', argument " "1"" of type '" "CCron *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CCron_Start" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  (arg1)->Start(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_Start__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  timeval *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_Start" "', argument " "1"" of type '" "CCron *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_timeval,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CCron_Start" "', argument " "2"" of type '" "timeval const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CCron_Start" "', argument " "2"" of type '" "timeval const &""'"); 
  }
  arg2 = reinterpret_cast< timeval * >(argp2);
  (arg1)->Start((timeval const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_Start(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CCron_Start", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CCron, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_timeval, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CCron_Start__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CCron, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_double(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CCron_Start__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CCron_Start'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CCron::Start(double)\n"
    "    CCron::Start(timeval const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CCron_Stop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_Stop" "', argument " "1"" of type '" "CCron *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  (arg1)->Stop();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_Pause(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_Pause" "', argument " "1"" of type '" "CCron *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  (arg1)->Pause();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_UnPause(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_UnPause" "', argument " "1"" of type '" "CCron *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  (arg1)->UnPause();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_Reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_Reset" "', argument " "1"" of type '" "CCron *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  (arg1)->Reset();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_GetInterval(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  timeval result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_GetInterval" "', argument " "1"" of type '" "CCron const *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  result = ((CCron const *)arg1)->GetInterval();
  resultobj = SWIG_NewPointerObj((new timeval(result)), SWIGTYPE_p_timeval, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_GetMaxCycles(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint32_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_GetMaxCycles" "', argument " "1"" of type '" "CCron const *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  result = ((CCron const *)arg1)->GetMaxCycles();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_GetCyclesLeft(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint32_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_GetCyclesLeft" "', argument " "1"" of type '" "CCron const *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  result = ((CCron const *)arg1)->GetCyclesLeft();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_isValid(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_isValid" "', argument " "1"" of type '" "CCron const *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  result = (bool)((CCron const *)arg1)->isValid();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_GetName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_GetName" "', argument " "1"" of type '" "CCron const *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  result = (CString *) &((CCron const *)arg1)->GetName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_SetName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CCron_SetName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_SetName" "', argument " "1"" of type '" "CCron *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CCron_SetName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CCron_SetName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_GetNextRun(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  timeval result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_GetNextRun" "', argument " "1"" of type '" "CCron const *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  result = ((CCron const *)arg1)->GetNextRun();
  resultobj = SWIG_NewPointerObj((new timeval(result)), SWIGTYPE_p_timeval, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCron_RunJob(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCron *arg1 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCron_RunJob" "', argument " "1"" of type '" "CCron *""'"); 
  }
  arg1 = reinterpret_cast< CCron * >(argp1);
  (arg1)->RunJob();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CCron_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CCron, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CCron_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CSMonitorFD(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSMonitorFD *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CSMonitorFD", 0, 0, 0)) SWIG_fail;
  result = (CSMonitorFD *)new CSMonitorFD();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSMonitorFD, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CSMonitorFD(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSMonitorFD *arg1 = (CSMonitorFD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSMonitorFD, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CSMonitorFD" "', argument " "1"" of type '" "CSMonitorFD *""'"); 
  }
  arg1 = reinterpret_cast< CSMonitorFD * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSMonitorFD_GatherFDsForSelect(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSMonitorFD *arg1 = (CSMonitorFD *) 0 ;
  std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > *arg2 = 0 ;
  long *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSMonitorFD_GatherFDsForSelect", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSMonitorFD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSMonitorFD_GatherFDsForSelect" "', argument " "1"" of type '" "CSMonitorFD *""'"); 
  }
  arg1 = reinterpret_cast< CSMonitorFD * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSMonitorFD_GatherFDsForSelect" "', argument " "2"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSMonitorFD_GatherFDsForSelect" "', argument " "2"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > &""'"); 
  }
  arg2 = reinterpret_cast< std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_long,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSMonitorFD_GatherFDsForSelect" "', argument " "3"" of type '" "long &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSMonitorFD_GatherFDsForSelect" "', argument " "3"" of type '" "long &""'"); 
  }
  arg3 = reinterpret_cast< long * >(argp3);
  result = (bool)(arg1)->GatherFDsForSelect(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSMonitorFD_FDsThatTriggered(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSMonitorFD *arg1 = (CSMonitorFD *) 0 ;
  std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSMonitorFD_FDsThatTriggered", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSMonitorFD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSMonitorFD_FDsThatTriggered" "', argument " "1"" of type '" "CSMonitorFD *""'"); 
  }
  arg1 = reinterpret_cast< CSMonitorFD * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSMonitorFD_FDsThatTriggered" "', argument " "2"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSMonitorFD_FDsThatTriggered" "', argument " "2"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > const &""'"); 
  }
  arg2 = reinterpret_cast< std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > * >(argp2);
  result = (bool)(arg1)->FDsThatTriggered((std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSMonitorFD_CheckFDs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSMonitorFD *arg1 = (CSMonitorFD *) 0 ;
  std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSMonitorFD_CheckFDs", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSMonitorFD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSMonitorFD_CheckFDs" "', argument " "1"" of type '" "CSMonitorFD *""'"); 
  }
  arg1 = reinterpret_cast< CSMonitorFD * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSMonitorFD_CheckFDs" "', argument " "2"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSMonitorFD_CheckFDs" "', argument " "2"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > const &""'"); 
  }
  arg2 = reinterpret_cast< std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > * >(argp2);
  result = (bool)(arg1)->CheckFDs((std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSMonitorFD_Add(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSMonitorFD *arg1 = (CSMonitorFD *) 0 ;
  cs_sock_t arg2 ;
  short arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  short val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSMonitorFD_Add", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSMonitorFD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSMonitorFD_Add" "', argument " "1"" of type '" "CSMonitorFD *""'"); 
  }
  arg1 = reinterpret_cast< CSMonitorFD * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSMonitorFD_Add" "', argument " "2"" of type '" "cs_sock_t""'");
  } 
  arg2 = static_cast< cs_sock_t >(val2);
  ecode3 = SWIG_AsVal_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSMonitorFD_Add" "', argument " "3"" of type '" "short""'");
  } 
  arg3 = static_cast< short >(val3);
  (arg1)->Add(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSMonitorFD_Remove(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSMonitorFD *arg1 = (CSMonitorFD *) 0 ;
  cs_sock_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSMonitorFD_Remove", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSMonitorFD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSMonitorFD_Remove" "', argument " "1"" of type '" "CSMonitorFD *""'"); 
  }
  arg1 = reinterpret_cast< CSMonitorFD * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSMonitorFD_Remove" "', argument " "2"" of type '" "cs_sock_t""'");
  } 
  arg2 = static_cast< cs_sock_t >(val2);
  (arg1)->Remove(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSMonitorFD_DisableMonitor(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSMonitorFD *arg1 = (CSMonitorFD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSMonitorFD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSMonitorFD_DisableMonitor" "', argument " "1"" of type '" "CSMonitorFD *""'"); 
  }
  arg1 = reinterpret_cast< CSMonitorFD * >(argp1);
  (arg1)->DisableMonitor();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSMonitorFD_IsEnabled(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSMonitorFD *arg1 = (CSMonitorFD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSMonitorFD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSMonitorFD_IsEnabled" "', argument " "1"" of type '" "CSMonitorFD const *""'"); 
  }
  arg1 = reinterpret_cast< CSMonitorFD * >(argp1);
  result = (bool)((CSMonitorFD const *)arg1)->IsEnabled();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CSMonitorFD_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CSMonitorFD, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CSMonitorFD_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CSockCommon(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockCommon *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CSockCommon", 0, 0, 0)) SWIG_fail;
  result = (CSockCommon *)new CSockCommon();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSockCommon, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CSockCommon(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CSockCommon" "', argument " "1"" of type '" "CSockCommon *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockCommon_CleanupCrons(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockCommon_CleanupCrons" "', argument " "1"" of type '" "CSockCommon *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  (arg1)->CleanupCrons();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockCommon_CleanupFDMonitors(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockCommon_CleanupFDMonitors" "', argument " "1"" of type '" "CSockCommon *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  (arg1)->CleanupFDMonitors();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockCommon_GetCrons(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CCron *,std::allocator< CCron * > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockCommon_GetCrons" "', argument " "1"" of type '" "CSockCommon const *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  result = (std::vector< CCron *,std::allocator< CCron * > > *) &((CSockCommon const *)arg1)->GetCrons();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CCron_p_std__allocatorT_CCron_p_t_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockCommon_Cron(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockCommon_Cron" "', argument " "1"" of type '" "CSockCommon *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  (arg1)->Cron();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockCommon_AddCron(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  CCron *arg2 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSockCommon_AddCron", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockCommon_AddCron" "', argument " "1"" of type '" "CSockCommon *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockCommon_AddCron" "', argument " "2"" of type '" "CCron *""'"); 
  }
  arg2 = reinterpret_cast< CCron * >(argp2);
  (arg1)->AddCron(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockCommon_DelCron__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockCommon_DelCron" "', argument " "1"" of type '" "CSockCommon *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockCommon_DelCron" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockCommon_DelCron" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSockCommon_DelCron" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockCommon_DelCron" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  (arg1)->DelCron((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockCommon_DelCron__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockCommon_DelCron" "', argument " "1"" of type '" "CSockCommon *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockCommon_DelCron" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockCommon_DelCron" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSockCommon_DelCron" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->DelCron((CString const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockCommon_DelCron__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockCommon_DelCron" "', argument " "1"" of type '" "CSockCommon *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockCommon_DelCron" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockCommon_DelCron" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->DelCron((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockCommon_DelCron__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  uint32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockCommon_DelCron" "', argument " "1"" of type '" "CSockCommon *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSockCommon_DelCron" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = static_cast< uint32_t >(val2);
  (arg1)->DelCron(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockCommon_DelCron(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSockCommon_DelCron", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockCommon, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CSockCommon_DelCron__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockCommon, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CSockCommon_DelCron__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockCommon, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CSockCommon_DelCron__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockCommon, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CSockCommon_DelCron__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSockCommon_DelCron'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSockCommon::DelCron(CString const &,bool,bool)\n"
    "    CSockCommon::DelCron(CString const &,bool)\n"
    "    CSockCommon::DelCron(CString const &)\n"
    "    CSockCommon::DelCron(uint32_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSockCommon_DelCronByAddr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  CCron *arg2 = (CCron *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSockCommon_DelCronByAddr", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockCommon_DelCronByAddr" "', argument " "1"" of type '" "CSockCommon *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CCron, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockCommon_DelCronByAddr" "', argument " "2"" of type '" "CCron *""'"); 
  }
  arg2 = reinterpret_cast< CCron * >(argp2);
  (arg1)->DelCronByAddr(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockCommon_CheckFDs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSockCommon_CheckFDs", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockCommon_CheckFDs" "', argument " "1"" of type '" "CSockCommon *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockCommon_CheckFDs" "', argument " "2"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockCommon_CheckFDs" "', argument " "2"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > const &""'"); 
  }
  arg2 = reinterpret_cast< std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > * >(argp2);
  (arg1)->CheckFDs((std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockCommon_AssignFDs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > *arg2 = 0 ;
  timeval *arg3 = (timeval *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSockCommon_AssignFDs", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockCommon_AssignFDs" "', argument " "1"" of type '" "CSockCommon *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockCommon_AssignFDs" "', argument " "2"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockCommon_AssignFDs" "', argument " "2"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > &""'"); 
  }
  arg2 = reinterpret_cast< std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_timeval, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockCommon_AssignFDs" "', argument " "3"" of type '" "timeval *""'"); 
  }
  arg3 = reinterpret_cast< timeval * >(argp3);
  (arg1)->AssignFDs(*arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockCommon_MonitorFD(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockCommon *arg1 = (CSockCommon *) 0 ;
  CSMonitorFD *arg2 = (CSMonitorFD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSockCommon_MonitorFD", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockCommon, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockCommon_MonitorFD" "', argument " "1"" of type '" "CSockCommon *""'"); 
  }
  arg1 = reinterpret_cast< CSockCommon * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CSMonitorFD, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockCommon_MonitorFD" "', argument " "2"" of type '" "CSMonitorFD *""'"); 
  }
  arg2 = reinterpret_cast< CSMonitorFD * >(argp2);
  (arg1)->MonitorFD(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CSockCommon_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CSockCommon, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CSockCommon_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_Csock__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Csock" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = static_cast< int >(val1);
  result = (Csock *)new Csock(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Csock__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (Csock *)new Csock();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Csock__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  uint16_t arg2 ;
  int arg3 ;
  int res1 = SWIG_OLDOBJ ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Csock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Csock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Csock" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Csock" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  result = (Csock *)new Csock((CString const &)*arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Csock__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  uint16_t arg2 ;
  int res1 = SWIG_OLDOBJ ;
  long val2 ;
  int ecode2 = 0 ;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Csock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Csock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Csock" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  result = (Csock *)new Csock((CString const &)*arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Csock(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_Csock", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_Csock__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_int(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_Csock__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_new_Csock__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_new_Csock__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_Csock'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::Csock(int)\n"
    "    Csock::Csock()\n"
    "    Csock::Csock(CString const &,uint16_t,int)\n"
    "    Csock::Csock(CString const &,uint16_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_GetSockObj(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  uint16_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  long val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Csock *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_GetSockObj", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetSockObj" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_GetSockObj" "', argument " "3"" of type '" "uint16_t""'");
  } 
  arg3 = static_cast< uint16_t >(val3);
  result = (Csock *)(arg1)->GetSockObj((CString const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Csock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Csock" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Dereference(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Dereference" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->Dereference();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Copy(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  Csock *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_Copy", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Copy" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_Csock,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_Copy" "', argument " "2"" of type '" "Csock const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_Copy" "', argument " "2"" of type '" "Csock const &""'"); 
  }
  arg2 = reinterpret_cast< Csock * >(argp2);
  (arg1)->Copy((Csock const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock___lshift____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock___lshift__" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock___lshift__" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock___lshift__" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (Csock *) &(arg1)->operator <<((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_Csock___lshift____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  std::ostream &(*arg2)(std::ostream &) = (std::ostream &(*)(std::ostream &)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock___lshift__" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_r_std__ostream__r_std__ostream);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "Csock___lshift__" "', argument " "2"" of type '" "std::ostream &(*)(std::ostream &)""'"); 
    }
  }
  result = (Csock *) &(arg1)->operator <<(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_Csock___lshift____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  int32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock___lshift__" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_int32_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock___lshift__" "', argument " "2"" of type '" "int32_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock___lshift__" "', argument " "2"" of type '" "int32_t""'");
    } else {
      int32_t * temp = reinterpret_cast< int32_t * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  result = (Csock *) &(arg1)->operator <<(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_Csock___lshift____SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock___lshift__" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock___lshift__" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = static_cast< uint32_t >(val2);
  result = (Csock *) &(arg1)->operator <<(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_Csock___lshift____SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  int64_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock___lshift__" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_int64_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock___lshift__" "', argument " "2"" of type '" "int64_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock___lshift__" "', argument " "2"" of type '" "int64_t""'");
    } else {
      int64_t * temp = reinterpret_cast< int64_t * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  result = (Csock *) &(arg1)->operator <<(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_Csock___lshift____SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint64_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock___lshift__" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock___lshift__" "', argument " "2"" of type '" "uint64_t""'");
  } 
  arg2 = static_cast< uint64_t >(val2);
  result = (Csock *) &(arg1)->operator <<(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_Csock___lshift____SWIG_6(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock___lshift__" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock___lshift__" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = static_cast< float >(val2);
  result = (Csock *) &(arg1)->operator <<(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_Csock___lshift____SWIG_7(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock___lshift__" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock___lshift__" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  result = (Csock *) &(arg1)->operator <<(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_Csock___lshift__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock___lshift__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *ptr = 0;
      int res = SWIG_ConvertFunctionPtr(argv[1], &ptr, SWIGTYPE_p_f_r_std__ostream__r_std__ostream);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Csock___lshift____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_int32_t, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Csock___lshift____SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_int64_t, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Csock___lshift____SWIG_4(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Csock___lshift____SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Csock___lshift____SWIG_5(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_float(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Csock___lshift____SWIG_6(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_double(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Csock___lshift____SWIG_7(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Csock___lshift____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_Csock_Connect(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Connect" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)(arg1)->Connect();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ConnectUnixInternal(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_ConnectUnixInternal", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ConnectUnixInternal" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_ConnectUnixInternal" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_ConnectUnixInternal" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->ConnectUnixInternal((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ListenUnixInternal__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  int arg3 ;
  uint32_t arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int val3 ;
  int ecode3 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ListenUnixInternal" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_ListenUnixInternal" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_ListenUnixInternal" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_ListenUnixInternal" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  ecode4 = SWIG_AsVal_long(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Csock_ListenUnixInternal" "', argument " "4"" of type '" "uint32_t""'");
  } 
  arg4 = static_cast< uint32_t >(val4);
  result = (bool)(arg1)->ListenUnixInternal((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ListenUnixInternal__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ListenUnixInternal" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_ListenUnixInternal" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_ListenUnixInternal" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_ListenUnixInternal" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  result = (bool)(arg1)->ListenUnixInternal((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ListenUnixInternal__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ListenUnixInternal" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_ListenUnixInternal" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_ListenUnixInternal" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->ListenUnixInternal((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ListenUnixInternal(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_ListenUnixInternal", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Csock_ListenUnixInternal__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Csock_ListenUnixInternal__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_long(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Csock_ListenUnixInternal__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_ListenUnixInternal'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::ListenUnixInternal(CString const &,int,uint32_t)\n"
    "    Csock::ListenUnixInternal(CString const &,int)\n"
    "    Csock::ListenUnixInternal(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_Listen__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint16_t arg2 ;
  int arg3 ;
  CString *arg4 = 0 ;
  uint32_t arg5 ;
  bool arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  long val5 ;
  int ecode5 = 0 ;
  bool val6 ;
  int ecode6 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Listen" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_Listen" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_Listen" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Csock_Listen" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_Listen" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_long(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Csock_Listen" "', argument " "5"" of type '" "uint32_t""'");
  } 
  arg5 = static_cast< uint32_t >(val5);
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Csock_Listen" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  result = (bool)(arg1)->Listen(arg2,arg3,(CString const &)*arg4,arg5,arg6);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Listen__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint16_t arg2 ;
  int arg3 ;
  CString *arg4 = 0 ;
  uint32_t arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  long val5 ;
  int ecode5 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Listen" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_Listen" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_Listen" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Csock_Listen" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_Listen" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_long(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Csock_Listen" "', argument " "5"" of type '" "uint32_t""'");
  } 
  arg5 = static_cast< uint32_t >(val5);
  result = (bool)(arg1)->Listen(arg2,arg3,(CString const &)*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Listen__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint16_t arg2 ;
  int arg3 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Listen" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_Listen" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_Listen" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Csock_Listen" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_Listen" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (bool)(arg1)->Listen(arg2,arg3,(CString const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Listen__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint16_t arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Listen" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_Listen" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_Listen" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  result = (bool)(arg1)->Listen(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Listen__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint16_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Listen" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_Listen" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  result = (bool)(arg1)->Listen(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Listen(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[7] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_Listen", 0, 6, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Csock_Listen__SWIG_4(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Csock_Listen__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_Csock_Listen__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_long(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_Csock_Listen__SWIG_1(self, argc, argv);
            }
          }
        }
      }
    }
  }
  if (argc == 6) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_long(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              {
                int res = SWIG_AsVal_bool(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                return _wrap_Csock_Listen__SWIG_0(self, argc, argv);
              }
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_Listen'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::Listen(uint16_t,int,CString const &,uint32_t,bool)\n"
    "    Csock::Listen(uint16_t,int,CString const &,uint32_t)\n"
    "    Csock::Listen(uint16_t,int,CString const &)\n"
    "    Csock::Listen(uint16_t,int)\n"
    "    Csock::Listen(uint16_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_Accept(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  uint16_t *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  cs_sock_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_Accept", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Accept" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sHost");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_uint16_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Csock_Accept" "', argument " "3"" of type '" "uint16_t &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_Accept" "', argument " "3"" of type '" "uint16_t &""'"); 
  }
  arg3 = reinterpret_cast< uint16_t * >(argp3);
  result = (cs_sock_t)(arg1)->Accept(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_AcceptSSL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_AcceptSSL" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)(arg1)->AcceptSSL();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SSLClientSetup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SSLClientSetup" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)(arg1)->SSLClientSetup();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SSLServerSetup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SSLServerSetup" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)(arg1)->SSLServerSetup();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ConnectSSL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ConnectSSL" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)(arg1)->ConnectSSL();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_StartTLS(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_StartTLS" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)(arg1)->StartTLS();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Write__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  char *arg2 = (char *) 0 ;
  size_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Write" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_Write" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_Write" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = static_cast< size_t >(val3);
  result = (bool)(arg1)->Write((char const *)arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Write__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Write" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_Write" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_Write" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->Write((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Write(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_Write", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Csock_Write__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Csock_Write__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_Write'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::Write(char const *,size_t)\n"
    "    Csock::Write(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_Read(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  char *arg2 = (char *) 0 ;
  size_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  cs_ssize_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_Read", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Read" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_Read" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_Read" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = static_cast< size_t >(val3);
  result = (arg1)->Read(arg2,arg3);
  resultobj = SWIG_NewPointerObj((new cs_ssize_t(result)), SWIGTYPE_p_ssize_t, SWIG_POINTER_OWN |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetLocalIP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetLocalIP" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetLocalIP();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetRemoteIP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetRemoteIP" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetRemoteIP();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_IsConnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_IsConnected" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)((Csock const *)arg1)->IsConnected();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetIsConnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetIsConnected", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetIsConnected" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetIsConnected" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetIsConnected(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetRSock__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cs_sock_t *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetRSock" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (cs_sock_t *) &(arg1)->GetRSock();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetRSock__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cs_sock_t *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetRSock" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (cs_sock_t *) &((Csock const *)arg1)->GetRSock();
  resultobj = SWIG_From_int(static_cast< int >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetRSock(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_GetRSock", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Csock_GetRSock__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Csock_GetRSock__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_GetRSock'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::GetRSock()\n"
    "    Csock::GetRSock() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_SetRSock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  cs_sock_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetRSock", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetRSock" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetRSock" "', argument " "2"" of type '" "cs_sock_t""'");
  } 
  arg2 = static_cast< cs_sock_t >(val2);
  (arg1)->SetRSock(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetWSock__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cs_sock_t *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetWSock" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (cs_sock_t *) &(arg1)->GetWSock();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetWSock__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cs_sock_t *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetWSock" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (cs_sock_t *) &((Csock const *)arg1)->GetWSock();
  resultobj = SWIG_From_int(static_cast< int >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetWSock(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_GetWSock", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Csock_GetWSock__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Csock_GetWSock__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_GetWSock'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::GetWSock()\n"
    "    Csock::GetWSock() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_SetWSock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  cs_sock_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetWSock", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetWSock" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetWSock" "', argument " "2"" of type '" "cs_sock_t""'");
  } 
  arg2 = static_cast< cs_sock_t >(val2);
  (arg1)->SetWSock(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetSock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  cs_sock_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetSock", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetSock" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetSock" "', argument " "2"" of type '" "cs_sock_t""'");
  } 
  arg2 = static_cast< cs_sock_t >(val2);
  (arg1)->SetSock(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetSock__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cs_sock_t *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetSock" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (cs_sock_t *) &(arg1)->GetSock();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetSock__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  cs_sock_t *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetSock" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (cs_sock_t *) &((Csock const *)arg1)->GetSock();
  resultobj = SWIG_From_int(static_cast< int >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetSock(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_GetSock", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Csock_GetSock__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Csock_GetSock__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_GetSock'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::GetSock()\n"
    "    Csock::GetSock() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_CallSockError__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  int arg2 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_CallSockError" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_CallSockError" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Csock_CallSockError" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_CallSockError" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->CallSockError(arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_CallSockError__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_CallSockError" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_CallSockError" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  (arg1)->CallSockError(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_CallSockError(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_CallSockError", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Csock_CallSockError__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Csock_CallSockError__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_CallSockError'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::CallSockError(int,CString const &)\n"
    "    Csock::CallSockError(int)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_ResetTimer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ResetTimer" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->ResetTimer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_PauseRead(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_PauseRead" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->PauseRead();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_UnPauseRead(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_UnPauseRead" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->UnPauseRead();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_IsReadPaused(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_IsReadPaused" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)((Csock const *)arg1)->IsReadPaused();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetTimeout__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  int arg2 ;
  uint32_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetTimeout" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetTimeout" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_SetTimeout" "', argument " "3"" of type '" "uint32_t""'");
  } 
  arg3 = static_cast< uint32_t >(val3);
  (arg1)->SetTimeout(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetTimeout__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetTimeout" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetTimeout" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  (arg1)->SetTimeout(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetTimeout(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_SetTimeout", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Csock_SetTimeout__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_long(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Csock_SetTimeout__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_SetTimeout'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::SetTimeout(int,uint32_t)\n"
    "    Csock::SetTimeout(int)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_SetTimeoutType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetTimeoutType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetTimeoutType" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetTimeoutType" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = static_cast< uint32_t >(val2);
  (arg1)->SetTimeoutType(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetTimeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetTimeout" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (int)((Csock const *)arg1)->GetTimeout();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetTimeoutType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint32_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetTimeoutType" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetTimeoutType();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_CheckTimeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_CheckTimeout", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_CheckTimeout" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_time_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_CheckTimeout" "', argument " "2"" of type '" "time_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_CheckTimeout" "', argument " "2"" of type '" "time_t""'");
    } else {
      time_t * temp = reinterpret_cast< time_t * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  result = (bool)(arg1)->CheckTimeout(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_PushBuff__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  char *arg2 = (char *) 0 ;
  size_t arg3 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_PushBuff" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_PushBuff" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_PushBuff" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = static_cast< size_t >(val3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Csock_PushBuff" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  (arg1)->PushBuff((char const *)arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_PushBuff__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  char *arg2 = (char *) 0 ;
  size_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_PushBuff" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_PushBuff" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_PushBuff" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = static_cast< size_t >(val3);
  (arg1)->PushBuff((char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_PushBuff(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_PushBuff", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Csock_PushBuff__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_Csock_PushBuff__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_PushBuff'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::PushBuff(char const *,size_t,bool)\n"
    "    Csock::PushBuff(char const *,size_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_GetInternalReadBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetInternalReadBuffer" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (CString *) &(arg1)->GetInternalReadBuffer();
  {
    if (result) {
      resultobj = CPyRetString::wrap(*result);
    } else {
      resultobj = Py_None;
      Py_INCREF(Py_None);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetInternalWriteBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetInternalWriteBuffer" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (CString *) &(arg1)->GetInternalWriteBuffer();
  {
    if (result) {
      resultobj = CPyRetString::wrap(*result);
    } else {
      resultobj = Py_None;
      Py_INCREF(Py_None);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetMaxBufferThreshold(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetMaxBufferThreshold", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetMaxBufferThreshold" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetMaxBufferThreshold" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = static_cast< uint32_t >(val2);
  (arg1)->SetMaxBufferThreshold(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetMaxBufferThreshold(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint32_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetMaxBufferThreshold" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetMaxBufferThreshold();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetType" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (int)((Csock const *)arg1)->GetType();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetType" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetType" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  (arg1)->SetType(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetSockName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetSockName" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (CString *) &((Csock const *)arg1)->GetSockName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetSockName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetSockName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetSockName" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_SetSockName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_SetSockName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetSockName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetHostName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetHostName" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (CString *) &((Csock const *)arg1)->GetHostName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetHostName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetHostName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetHostName" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_SetHostName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_SetHostName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetHostName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetStartTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint64_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetStartTime" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetStartTime();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ResetStartTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ResetStartTime" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->ResetStartTime();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetBytesRead(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint64_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetBytesRead" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetBytesRead();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ResetBytesRead(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ResetBytesRead" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->ResetBytesRead();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetBytesWritten(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint64_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetBytesWritten" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetBytesWritten();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ResetBytesWritten(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ResetBytesWritten" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->ResetBytesWritten();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetAvgRead__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint64_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  double result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetAvgRead" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_GetAvgRead" "', argument " "2"" of type '" "uint64_t""'");
  } 
  arg2 = static_cast< uint64_t >(val2);
  result = (double)((Csock const *)arg1)->GetAvgRead(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetAvgRead__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetAvgRead" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (double)((Csock const *)arg1)->GetAvgRead();
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetAvgRead(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_GetAvgRead", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Csock_GetAvgRead__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Csock_GetAvgRead__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_GetAvgRead'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::GetAvgRead(uint64_t) const\n"
    "    Csock::GetAvgRead() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_GetAvgWrite__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint64_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  double result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetAvgWrite" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_GetAvgWrite" "', argument " "2"" of type '" "uint64_t""'");
  } 
  arg2 = static_cast< uint64_t >(val2);
  result = (double)((Csock const *)arg1)->GetAvgWrite(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetAvgWrite__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetAvgWrite" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (double)((Csock const *)arg1)->GetAvgWrite();
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetAvgWrite(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_GetAvgWrite", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Csock_GetAvgWrite__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Csock_GetAvgWrite__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_GetAvgWrite'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::GetAvgWrite(uint64_t) const\n"
    "    Csock::GetAvgWrite() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_GetRemotePort(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint16_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetRemotePort" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetRemotePort();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetLocalPort(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint16_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetLocalPort" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetLocalPort();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetPort(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint16_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetPort" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetPort();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetPort(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint16_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetPort", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetPort" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetPort" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  (arg1)->SetPort(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Close__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  Csock::ECloseType arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Close" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_Close" "', argument " "2"" of type '" "Csock::ECloseType""'");
  } 
  arg2 = static_cast< Csock::ECloseType >(val2);
  (arg1)->Close(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Close__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Close" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->Close();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Close(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_Close", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Csock_Close__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Csock_Close__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_Close'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::Close(Csock::ECloseType)\n"
    "    Csock::Close()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_GetCloseType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Csock::ECloseType result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetCloseType" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (Csock::ECloseType)((Csock const *)arg1)->GetCloseType();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_IsClosed(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_IsClosed" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)((Csock const *)arg1)->IsClosed();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_NonBlockingIO(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_NonBlockingIO" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->NonBlockingIO();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetSSL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetSSL" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)((Csock const *)arg1)->GetSSL();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetSSL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetSSL", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetSSL" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetSSL" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetSSL(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_HasWriteBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_HasWriteBuffer" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)((Csock const *)arg1)->HasWriteBuffer();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ClearWriteBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ClearWriteBuffer" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->ClearWriteBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SslIsEstablished(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SslIsEstablished" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)((Csock const *)arg1)->SslIsEstablished();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ConnectInetd__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  bool arg2 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ConnectInetd" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_ConnectInetd" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Csock_ConnectInetd" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_ConnectInetd" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->ConnectInetd(arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ConnectInetd__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ConnectInetd" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_ConnectInetd" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)(arg1)->ConnectInetd(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ConnectInetd__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ConnectInetd" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)(arg1)->ConnectInetd();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ConnectInetd(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_ConnectInetd", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Csock_ConnectInetd__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_Csock_ConnectInetd__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_Csock_ConnectInetd__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_ConnectInetd'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::ConnectInetd(bool,CString const &)\n"
    "    Csock::ConnectInetd(bool)\n"
    "    Csock::ConnectInetd()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_ConnectFD__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  int arg2 ;
  int arg3 ;
  CString *arg4 = 0 ;
  bool arg5 ;
  Csock::ETConn arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  bool val5 ;
  int ecode5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ConnectFD" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_ConnectFD" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_ConnectFD" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Csock_ConnectFD" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_ConnectFD" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Csock_ConnectFD" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Csock_ConnectFD" "', argument " "6"" of type '" "Csock::ETConn""'");
  } 
  arg6 = static_cast< Csock::ETConn >(val6);
  result = (bool)(arg1)->ConnectFD(arg2,arg3,(CString const &)*arg4,arg5,arg6);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ConnectFD__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  int arg2 ;
  int arg3 ;
  CString *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  bool val5 ;
  int ecode5 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ConnectFD" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_ConnectFD" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_ConnectFD" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Csock_ConnectFD" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_ConnectFD" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Csock_ConnectFD" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  result = (bool)(arg1)->ConnectFD(arg2,arg3,(CString const &)*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ConnectFD__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  int arg2 ;
  int arg3 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ConnectFD" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_ConnectFD" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_ConnectFD" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Csock_ConnectFD" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_ConnectFD" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (bool)(arg1)->ConnectFD(arg2,arg3,(CString const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ConnectFD(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[7] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_ConnectFD", 0, 6, argv))) SWIG_fail;
  --argc;
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_Csock_ConnectFD__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_bool(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_Csock_ConnectFD__SWIG_1(self, argc, argv);
            }
          }
        }
      }
    }
  }
  if (argc == 6) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_bool(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              {
                int res = SWIG_AsVal_int(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                return _wrap_Csock_ConnectFD__SWIG_0(self, argc, argv);
              }
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_ConnectFD'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::ConnectFD(int,int,CString const &,bool,Csock::ETConn)\n"
    "    Csock::ConnectFD(int,int,CString const &,bool)\n"
    "    Csock::ConnectFD(int,int,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_SetParentSockName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetParentSockName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetParentSockName" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_SetParentSockName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_SetParentSockName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetParentSockName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetParentSockName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetParentSockName" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (CString *) &((Csock const *)arg1)->GetParentSockName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetRate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint32_t arg2 ;
  uint64_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetRate", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetRate" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetRate" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = static_cast< uint32_t >(val2);
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_SetRate" "', argument " "3"" of type '" "uint64_t""'");
  } 
  arg3 = static_cast< uint64_t >(val3);
  (arg1)->SetRate(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetRateBytes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint32_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetRateBytes" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetRateBytes();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetRateTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint64_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetRateTime" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetRateTime();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Connected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Connected" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->Connected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Disconnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Disconnected" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->Disconnected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Timeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Timeout" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->Timeout();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ReadData(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  char *arg2 = (char *) 0 ;
  size_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_ReadData", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ReadData" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_ReadData" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_ReadData" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = static_cast< size_t >(val3);
  (arg1)->ReadData((char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ReadLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_ReadLine", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ReadLine" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_ReadLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_ReadLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->ReadLine((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_EnableReadLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_EnableReadLine" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->EnableReadLine();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_DisableReadLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_DisableReadLine" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->DisableReadLine();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_HasReadLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_HasReadLine" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)((Csock const *)arg1)->HasReadLine();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ReachedMaxBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ReachedMaxBuffer" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->ReachedMaxBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SockError(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  int arg2 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SockError", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SockError" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SockError" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Csock_SockError" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_SockError" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->SockError(arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ConnectionFrom(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  uint16_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  long val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_ConnectionFrom", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ConnectionFrom" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_ConnectionFrom" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_ConnectionFrom" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_ConnectionFrom" "', argument " "3"" of type '" "uint16_t""'");
  } 
  arg3 = static_cast< uint16_t >(val3);
  result = (bool)(arg1)->ConnectionFrom((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_Listening(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  uint16_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  long val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_Listening", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_Listening" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_Listening" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_Listening" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Csock_Listening" "', argument " "3"" of type '" "uint16_t""'");
  } 
  arg3 = static_cast< uint16_t >(val3);
  (arg1)->Listening((CString const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ConnectionRefused(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ConnectionRefused" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->ConnectionRefused();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ReadPaused(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ReadPaused" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->ReadPaused();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetTimeSinceLastDataTransaction__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  time_t result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetTimeSinceLastDataTransaction" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_time_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_GetTimeSinceLastDataTransaction" "', argument " "2"" of type '" "time_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_GetTimeSinceLastDataTransaction" "', argument " "2"" of type '" "time_t""'");
    } else {
      time_t * temp = reinterpret_cast< time_t * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  result = ((Csock const *)arg1)->GetTimeSinceLastDataTransaction(arg2);
  resultobj = SWIG_NewPointerObj((new time_t(result)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetTimeSinceLastDataTransaction__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  time_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetTimeSinceLastDataTransaction" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetTimeSinceLastDataTransaction();
  resultobj = SWIG_NewPointerObj((new time_t(result)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetTimeSinceLastDataTransaction(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_GetTimeSinceLastDataTransaction", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Csock_GetTimeSinceLastDataTransaction__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_time_t, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Csock_GetTimeSinceLastDataTransaction__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_GetTimeSinceLastDataTransaction'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::GetTimeSinceLastDataTransaction(time_t) const\n"
    "    Csock::GetTimeSinceLastDataTransaction() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_GetLastCheckTimeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  time_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetLastCheckTimeout" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetLastCheckTimeout();
  resultobj = SWIG_NewPointerObj((new time_t(result)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetNextCheckTimeout__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  time_t result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetNextCheckTimeout" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_time_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_GetNextCheckTimeout" "', argument " "2"" of type '" "time_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_GetNextCheckTimeout" "', argument " "2"" of type '" "time_t""'");
    } else {
      time_t * temp = reinterpret_cast< time_t * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  result = ((Csock const *)arg1)->GetNextCheckTimeout(arg2);
  resultobj = SWIG_NewPointerObj((new time_t(result)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetNextCheckTimeout__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  time_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetNextCheckTimeout" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = ((Csock const *)arg1)->GetNextCheckTimeout();
  resultobj = SWIG_NewPointerObj((new time_t(result)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetNextCheckTimeout(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Csock_GetNextCheckTimeout", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_Csock_GetNextCheckTimeout__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Csock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_time_t, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Csock_GetNextCheckTimeout__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Csock_GetNextCheckTimeout'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Csock::GetNextCheckTimeout(time_t) const\n"
    "    Csock::GetNextCheckTimeout() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Csock_GetPending(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetPending" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (int)((Csock const *)arg1)->GetPending();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetConState(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Csock::ECONState result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetConState" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (Csock::ECONState)((Csock const *)arg1)->GetConState();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetConState(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  Csock::ECONState arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetConState", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetConState" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetConState" "', argument " "2"" of type '" "Csock::ECONState""'");
  } 
  arg2 = static_cast< Csock::ECONState >(val2);
  (arg1)->SetConState(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_CreateSocksFD(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_CreateSocksFD" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)(arg1)->CreateSocksFD();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_CloseSocksFD(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_CloseSocksFD" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  (arg1)->CloseSocksFD();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetBindHost" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (CString *) &((Csock const *)arg1)->GetBindHost();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetBindHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetBindHost" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_SetBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_SetBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetBindHost((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_DNSLookup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  Csock::EDNSLType arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_DNSLookup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_DNSLookup" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_DNSLookup" "', argument " "2"" of type '" "Csock::EDNSLType""'");
  } 
  arg2 = static_cast< Csock::EDNSLType >(val2);
  result = (int)(arg1)->DNSLookup(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetupVHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetupVHost" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)(arg1)->SetupVHost();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetIPv6(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetIPv6" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (bool)((Csock const *)arg1)->GetIPv6();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetIPv6(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetIPv6", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetIPv6" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetIPv6" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetIPv6(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetAFRequire(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CSSockAddr::EAFRequire arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetAFRequire", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetAFRequire" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetAFRequire" "', argument " "2"" of type '" "CSSockAddr::EAFRequire""'");
  } 
  arg2 = static_cast< CSSockAddr::EAFRequire >(val2);
  (arg1)->SetAFRequire(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_AllowWrite(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  uint64_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_AllowWrite", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_AllowWrite" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_uint64_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_AllowWrite" "', argument " "2"" of type '" "uint64_t &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_AllowWrite" "', argument " "2"" of type '" "uint64_t &""'"); 
  }
  arg2 = reinterpret_cast< uint64_t * >(argp2);
  result = (bool)((Csock const *)arg1)->AllowWrite(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_SetSkipConnect(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_SetSkipConnect", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_SetSkipConnect" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Csock_SetSkipConnect" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetSkipConnect(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetAddrInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  CString *arg2 = 0 ;
  CSSockAddr *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_GetAddrInfo", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetAddrInfo" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_GetAddrInfo" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_GetAddrInfo" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CSSockAddr,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Csock_GetAddrInfo" "', argument " "3"" of type '" "CSSockAddr &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_GetAddrInfo" "', argument " "3"" of type '" "CSSockAddr &""'"); 
  }
  arg3 = reinterpret_cast< CSSockAddr * >(argp3);
  result = (int)(arg1)->GetAddrInfo((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_ConvertAddress(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  sockaddr_storage *arg2 = (sockaddr_storage *) 0 ;
  socklen_t arg3 ;
  CString *arg4 = 0 ;
  uint16_t *arg5 = (uint16_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject *swig_obj[5] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_ConvertAddress", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_ConvertAddress" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_sockaddr_storage, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Csock_ConvertAddress" "', argument " "2"" of type '" "sockaddr_storage const *""'"); 
  }
  arg2 = reinterpret_cast< sockaddr_storage * >(argp2);
  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_socklen_t,  0  | 0);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Csock_ConvertAddress" "', argument " "3"" of type '" "socklen_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Csock_ConvertAddress" "', argument " "3"" of type '" "socklen_t""'");
    } else {
      socklen_t * temp = reinterpret_cast< socklen_t * >(argp3);
      arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp;
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sIP");
    }
  }
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_uint16_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Csock_ConvertAddress" "', argument " "5"" of type '" "uint16_t *""'"); 
  }
  arg5 = reinterpret_cast< uint16_t * >(argp5);
  result = (int)((Csock const *)arg1)->ConvertAddress((sockaddr_storage const *)arg2,arg3,*arg4,arg5);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_GetMaxConns(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_GetMaxConns" "', argument " "1"" of type '" "Csock const *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  result = (int)((Csock const *)arg1)->GetMaxConns();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Csock_WriteBytes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Csock *arg1 = (Csock *) 0 ;
  PyObject *arg2 = (PyObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Csock_WriteBytes", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Csock_WriteBytes" "', argument " "1"" of type '" "Csock *""'"); 
  }
  arg1 = reinterpret_cast< Csock * >(argp1);
  arg2 = swig_obj[1];
  result = (PyObject *)Csock_WriteBytes(arg1,arg2);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Csock_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Csock, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Csock_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CSConnection__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  uint16_t arg2 ;
  int arg3 ;
  int res1 = SWIG_OLDOBJ ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  CSConnection *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CSConnection" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CSConnection" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CSConnection" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CSConnection" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  result = (CSConnection *)new CSConnection((CString const &)*arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSConnection, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CSConnection__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  uint16_t arg2 ;
  int res1 = SWIG_OLDOBJ ;
  long val2 ;
  int ecode2 = 0 ;
  CSConnection *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CSConnection" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CSConnection" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CSConnection" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  result = (CSConnection *)new CSConnection((CString const &)*arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSConnection, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CSConnection(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CSConnection", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_new_CSConnection__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_new_CSConnection__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CSConnection'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSConnection::CSConnection(CString const &,uint16_t,int)\n"
    "    CSConnection::CSConnection(CString const &,uint16_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CSConnection(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CSConnection" "', argument " "1"" of type '" "CSConnection *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_GetHostname(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_GetHostname" "', argument " "1"" of type '" "CSConnection const *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  result = (CString *) &((CSConnection const *)arg1)->GetHostname();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_GetSockName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_GetSockName" "', argument " "1"" of type '" "CSConnection const *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  result = (CString *) &((CSConnection const *)arg1)->GetSockName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_GetBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_GetBindHost" "', argument " "1"" of type '" "CSConnection const *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  result = (CString *) &((CSConnection const *)arg1)->GetBindHost();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_GetPort(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint16_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_GetPort" "', argument " "1"" of type '" "CSConnection const *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  result = ((CSConnection const *)arg1)->GetPort();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_GetTimeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_GetTimeout" "', argument " "1"" of type '" "CSConnection const *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  result = (int)((CSConnection const *)arg1)->GetTimeout();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_GetIsSSL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_GetIsSSL" "', argument " "1"" of type '" "CSConnection const *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  result = (bool)((CSConnection const *)arg1)->GetIsSSL();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_GetAFRequire(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CSSockAddr::EAFRequire result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_GetAFRequire" "', argument " "1"" of type '" "CSConnection const *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  result = (CSSockAddr::EAFRequire)((CSConnection const *)arg1)->GetAFRequire();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_SetHostname(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSConnection_SetHostname", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_SetHostname" "', argument " "1"" of type '" "CSConnection *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSConnection_SetHostname" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSConnection_SetHostname" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetHostname((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_SetSockName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSConnection_SetSockName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_SetSockName" "', argument " "1"" of type '" "CSConnection *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSConnection_SetSockName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSConnection_SetSockName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetSockName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_SetBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSConnection_SetBindHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_SetBindHost" "', argument " "1"" of type '" "CSConnection *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSConnection_SetBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSConnection_SetBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetBindHost((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_SetPort(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  uint16_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSConnection_SetPort", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_SetPort" "', argument " "1"" of type '" "CSConnection *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSConnection_SetPort" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  (arg1)->SetPort(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_SetTimeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSConnection_SetTimeout", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_SetTimeout" "', argument " "1"" of type '" "CSConnection *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSConnection_SetTimeout" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  (arg1)->SetTimeout(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_SetIsSSL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSConnection_SetIsSSL", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_SetIsSSL" "', argument " "1"" of type '" "CSConnection *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSConnection_SetIsSSL" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetIsSSL(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSConnection_SetAFRequire(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSConnection *arg1 = (CSConnection *) 0 ;
  CSSockAddr::EAFRequire arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSConnection_SetAFRequire", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSConnection_SetAFRequire" "', argument " "1"" of type '" "CSConnection *""'"); 
  }
  arg1 = reinterpret_cast< CSConnection * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSConnection_SetAFRequire" "', argument " "2"" of type '" "CSSockAddr::EAFRequire""'");
  } 
  arg2 = static_cast< CSSockAddr::EAFRequire >(val2);
  (arg1)->SetAFRequire(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CSConnection_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CSConnection, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CSConnection_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CSSSLConnection__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  uint16_t arg2 ;
  int arg3 ;
  int res1 = SWIG_OLDOBJ ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  CSSSLConnection *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CSSSLConnection" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CSSSLConnection" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CSSSLConnection" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CSSSLConnection" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  result = (CSSSLConnection *)new CSSSLConnection((CString const &)*arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSSSLConnection, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CSSSLConnection__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  uint16_t arg2 ;
  int res1 = SWIG_OLDOBJ ;
  long val2 ;
  int ecode2 = 0 ;
  CSSSLConnection *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CSSSLConnection" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CSSSLConnection" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CSSSLConnection" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  result = (CSSSLConnection *)new CSSSLConnection((CString const &)*arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSSSLConnection, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CSSSLConnection(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CSSSLConnection", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_new_CSSSLConnection__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_new_CSSSLConnection__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CSSSLConnection'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSSSLConnection::CSSSLConnection(CString const &,uint16_t,int)\n"
    "    CSSSLConnection::CSSSLConnection(CString const &,uint16_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CSSSLConnection(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSSSLConnection *arg1 = (CSSSLConnection *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSSSLConnection, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CSSSLConnection" "', argument " "1"" of type '" "CSSSLConnection *""'"); 
  }
  arg1 = reinterpret_cast< CSSSLConnection * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CSSSLConnection_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CSSSLConnection, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CSSSLConnection_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CSListener__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  uint16_t arg1 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  long val1 ;
  int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  CSListener *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  ecode1 = SWIG_AsVal_long(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_CSListener" "', argument " "1"" of type '" "uint16_t""'");
  } 
  arg1 = static_cast< uint16_t >(val1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CSListener" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CSListener" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CSListener" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (CSListener *)new CSListener(arg1,(CString const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSListener, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CSListener__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  uint16_t arg1 ;
  CString *arg2 = 0 ;
  long val1 ;
  int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CSListener *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_long(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_CSListener" "', argument " "1"" of type '" "uint16_t""'");
  } 
  arg1 = static_cast< uint16_t >(val1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CSListener" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CSListener" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CSListener *)new CSListener(arg1,(CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSListener, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CSListener__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  uint16_t arg1 ;
  long val1 ;
  int ecode1 = 0 ;
  CSListener *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_long(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_CSListener" "', argument " "1"" of type '" "uint16_t""'");
  } 
  arg1 = static_cast< uint16_t >(val1);
  result = (CSListener *)new CSListener(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSListener, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CSListener(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CSListener", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_long(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_CSListener__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_long(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_CSListener__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    {
      int res = SWIG_AsVal_long(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_new_CSListener__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CSListener'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSListener::CSListener(uint16_t,CString const &,bool)\n"
    "    CSListener::CSListener(uint16_t,CString const &)\n"
    "    CSListener::CSListener(uint16_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CSListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CSListener" "', argument " "1"" of type '" "CSListener *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_SetDetach(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSListener_SetDetach", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_SetDetach" "', argument " "1"" of type '" "CSListener *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSListener_SetDetach" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetDetach(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_GetDetach(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_GetDetach" "', argument " "1"" of type '" "CSListener const *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  result = (bool)((CSListener const *)arg1)->GetDetach();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_GetPort(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint16_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_GetPort" "', argument " "1"" of type '" "CSListener const *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  result = ((CSListener const *)arg1)->GetPort();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_GetSockName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_GetSockName" "', argument " "1"" of type '" "CSListener const *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  result = (CString *) &((CSListener const *)arg1)->GetSockName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_GetBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_GetBindHost" "', argument " "1"" of type '" "CSListener const *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  result = (CString *) &((CSListener const *)arg1)->GetBindHost();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_GetIsSSL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_GetIsSSL" "', argument " "1"" of type '" "CSListener const *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  result = (bool)((CSListener const *)arg1)->GetIsSSL();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_GetMaxConns(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_GetMaxConns" "', argument " "1"" of type '" "CSListener const *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  result = (int)((CSListener const *)arg1)->GetMaxConns();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_GetTimeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint32_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_GetTimeout" "', argument " "1"" of type '" "CSListener const *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  result = ((CSListener const *)arg1)->GetTimeout();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_GetAFRequire(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CSSockAddr::EAFRequire result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_GetAFRequire" "', argument " "1"" of type '" "CSListener const *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  result = (CSSockAddr::EAFRequire)((CSListener const *)arg1)->GetAFRequire();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_SetPort(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  uint16_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSListener_SetPort", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_SetPort" "', argument " "1"" of type '" "CSListener *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSListener_SetPort" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  (arg1)->SetPort(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_SetSockName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSListener_SetSockName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_SetSockName" "', argument " "1"" of type '" "CSListener *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSListener_SetSockName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSListener_SetSockName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetSockName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_SetBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSListener_SetBindHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_SetBindHost" "', argument " "1"" of type '" "CSListener *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSListener_SetBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSListener_SetBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetBindHost((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_SetIsSSL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSListener_SetIsSSL", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_SetIsSSL" "', argument " "1"" of type '" "CSListener *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSListener_SetIsSSL" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetIsSSL(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_SetMaxConns(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSListener_SetMaxConns", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_SetMaxConns" "', argument " "1"" of type '" "CSListener *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSListener_SetMaxConns" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  (arg1)->SetMaxConns(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_SetTimeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  uint32_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSListener_SetTimeout", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_SetTimeout" "', argument " "1"" of type '" "CSListener *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSListener_SetTimeout" "', argument " "2"" of type '" "uint32_t""'");
  } 
  arg2 = static_cast< uint32_t >(val2);
  (arg1)->SetTimeout(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSListener_SetAFRequire(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSListener *arg1 = (CSListener *) 0 ;
  CSSockAddr::EAFRequire arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSListener_SetAFRequire", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSListener_SetAFRequire" "', argument " "1"" of type '" "CSListener *""'"); 
  }
  arg1 = reinterpret_cast< CSListener * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSListener_SetAFRequire" "', argument " "2"" of type '" "CSSockAddr::EAFRequire""'");
  } 
  arg2 = static_cast< CSSockAddr::EAFRequire >(val2);
  (arg1)->SetAFRequire(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CSListener_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CSListener, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CSListener_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CSocketManager(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CSocketManager", 0, 0, 0)) SWIG_fail;
  result = (CSocketManager *)new CSocketManager();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSocketManager, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CSocketManager(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CSocketManager" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_clear" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_Cleanup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_Cleanup" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  (arg1)->Cleanup();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_GetSockObj__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  CString *arg2 = 0 ;
  uint16_t arg3 ;
  int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  long val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_GetSockObj" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocketManager_GetSockObj" "', argument " "3"" of type '" "uint16_t""'");
  } 
  arg3 = static_cast< uint16_t >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSocketManager_GetSockObj" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = static_cast< int >(val4);
  result = (Csock *)(arg1)->GetSockObj((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_GetSockObj__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  CString *arg2 = 0 ;
  uint16_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  long val3 ;
  int ecode3 = 0 ;
  Csock *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_GetSockObj" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocketManager_GetSockObj" "', argument " "3"" of type '" "uint16_t""'");
  } 
  arg3 = static_cast< uint16_t >(val3);
  result = (Csock *)(arg1)->GetSockObj((CString const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_GetSockObj(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSocketManager_GetSockObj", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocketManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_long(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CSocketManager_GetSockObj__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocketManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_long(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CSocketManager_GetSockObj__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSocketManager_GetSockObj'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSocketManager::GetSockObj(CString const &,uint16_t,int)\n"
    "    CSocketManager::GetSockObj(CString const &,uint16_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSocketManager_Connect__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  CSConnection *arg2 = 0 ;
  Csock *arg3 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_Connect" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CSConnection,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_Connect" "', argument " "2"" of type '" "CSConnection const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_Connect" "', argument " "2"" of type '" "CSConnection const &""'"); 
  }
  arg2 = reinterpret_cast< CSConnection * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSocketManager_Connect" "', argument " "3"" of type '" "Csock *""'"); 
  }
  arg3 = reinterpret_cast< Csock * >(argp3);
  (arg1)->Connect((CSConnection const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_Connect__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  CSConnection *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_Connect" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CSConnection,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_Connect" "', argument " "2"" of type '" "CSConnection const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_Connect" "', argument " "2"" of type '" "CSConnection const &""'"); 
  }
  arg2 = reinterpret_cast< CSConnection * >(argp2);
  (arg1)->Connect((CSConnection const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_Connect(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSocketManager_Connect", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocketManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CSConnection, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CSocketManager_Connect__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocketManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CSConnection, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_Csock, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CSocketManager_Connect__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSocketManager_Connect'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSocketManager::Connect(CSConnection const &,Csock *)\n"
    "    CSocketManager::Connect(CSConnection const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSocketManager_Listen__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  CSListener *arg2 = 0 ;
  Csock *arg3 = (Csock *) 0 ;
  uint16_t *arg4 = (uint16_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_Listen" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CSListener,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_Listen" "', argument " "2"" of type '" "CSListener const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_Listen" "', argument " "2"" of type '" "CSListener const &""'"); 
  }
  arg2 = reinterpret_cast< CSListener * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSocketManager_Listen" "', argument " "3"" of type '" "Csock *""'"); 
  }
  arg3 = reinterpret_cast< Csock * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_uint16_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSocketManager_Listen" "', argument " "4"" of type '" "uint16_t *""'"); 
  }
  arg4 = reinterpret_cast< uint16_t * >(argp4);
  result = (bool)(arg1)->Listen((CSListener const &)*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_Listen__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  CSListener *arg2 = 0 ;
  Csock *arg3 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_Listen" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CSListener,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_Listen" "', argument " "2"" of type '" "CSListener const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_Listen" "', argument " "2"" of type '" "CSListener const &""'"); 
  }
  arg2 = reinterpret_cast< CSListener * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSocketManager_Listen" "', argument " "3"" of type '" "Csock *""'"); 
  }
  arg3 = reinterpret_cast< Csock * >(argp3);
  result = (bool)(arg1)->Listen((CSListener const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_Listen__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  CSListener *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_Listen" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CSListener,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_Listen" "', argument " "2"" of type '" "CSListener const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_Listen" "', argument " "2"" of type '" "CSListener const &""'"); 
  }
  arg2 = reinterpret_cast< CSListener * >(argp2);
  result = (bool)(arg1)->Listen((CSListener const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_Listen(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSocketManager_Listen", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocketManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CSListener, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CSocketManager_Listen__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocketManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CSListener, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_Csock, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CSocketManager_Listen__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocketManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CSListener, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_Csock, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_uint16_t, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CSocketManager_Listen__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSocketManager_Listen'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSocketManager::Listen(CSListener const &,Csock *,uint16_t *)\n"
    "    CSocketManager::Listen(CSListener const &,Csock *)\n"
    "    CSocketManager::Listen(CSListener const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSocketManager_HasFDs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_HasFDs" "', argument " "1"" of type '" "CSocketManager const *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  result = (bool)((CSocketManager const *)arg1)->HasFDs();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_Loop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_Loop" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  (arg1)->Loop();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_DynamicSelectLoop__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  uint64_t arg2 ;
  uint64_t arg3 ;
  time_t arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  void *argp4 ;
  int res4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_DynamicSelectLoop" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocketManager_DynamicSelectLoop" "', argument " "2"" of type '" "uint64_t""'");
  } 
  arg2 = static_cast< uint64_t >(val2);
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocketManager_DynamicSelectLoop" "', argument " "3"" of type '" "uint64_t""'");
  } 
  arg3 = static_cast< uint64_t >(val3);
  {
    res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_time_t,  0  | 0);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSocketManager_DynamicSelectLoop" "', argument " "4"" of type '" "time_t""'"); 
    }  
    if (!argp4) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_DynamicSelectLoop" "', argument " "4"" of type '" "time_t""'");
    } else {
      time_t * temp = reinterpret_cast< time_t * >(argp4);
      arg4 = *temp;
      if (SWIG_IsNewObj(res4)) delete temp;
    }
  }
  (arg1)->DynamicSelectLoop(arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_DynamicSelectLoop__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  uint64_t arg2 ;
  uint64_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_DynamicSelectLoop" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocketManager_DynamicSelectLoop" "', argument " "2"" of type '" "uint64_t""'");
  } 
  arg2 = static_cast< uint64_t >(val2);
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocketManager_DynamicSelectLoop" "', argument " "3"" of type '" "uint64_t""'");
  } 
  arg3 = static_cast< uint64_t >(val3);
  (arg1)->DynamicSelectLoop(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_DynamicSelectLoop(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSocketManager_DynamicSelectLoop", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocketManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_long(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CSocketManager_DynamicSelectLoop__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocketManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_long(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_ConvertPtr(argv[3], 0, SWIGTYPE_p_time_t, SWIG_POINTER_NO_NULL | 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CSocketManager_DynamicSelectLoop__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSocketManager_DynamicSelectLoop'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSocketManager::DynamicSelectLoop(uint64_t,uint64_t,time_t)\n"
    "    CSocketManager::DynamicSelectLoop(uint64_t,uint64_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSocketManager_AddSock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  Csock *arg2 = (Csock *) 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_AddSock", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_AddSock" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_AddSock" "', argument " "2"" of type '" "Csock *""'"); 
  }
  arg2 = reinterpret_cast< Csock * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSocketManager_AddSock" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_AddSock" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->AddSock(arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_FindSockByRemotePort(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  uint16_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Csock *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_FindSockByRemotePort", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_FindSockByRemotePort" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocketManager_FindSockByRemotePort" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  result = (Csock *)(arg1)->FindSockByRemotePort(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_FindSockByLocalPort(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  uint16_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Csock *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_FindSockByLocalPort", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_FindSockByLocalPort" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocketManager_FindSockByLocalPort" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  result = (Csock *)(arg1)->FindSockByLocalPort(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_FindSockByName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  Csock *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_FindSockByName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_FindSockByName" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_FindSockByName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_FindSockByName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (Csock *)(arg1)->FindSockByName((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_FindSockByFD(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  cs_sock_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Csock *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_FindSockByFD", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_FindSockByFD" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocketManager_FindSockByFD" "', argument " "2"" of type '" "cs_sock_t""'");
  } 
  arg2 = static_cast< cs_sock_t >(val2);
  result = (Csock *)(arg1)->FindSockByFD(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_FindSocksByName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  SwigValueWrapper< std::vector< Csock *,std::allocator< Csock * > > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_FindSocksByName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_FindSocksByName" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_FindSocksByName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_FindSocksByName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->FindSocksByName((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj((new std::vector< Csock *,std::allocator< Csock * > >(result)), SWIGTYPE_p_std__vectorT_Csock_p_std__allocatorT_Csock_p_t_t, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_FindSocksByRemoteHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  SwigValueWrapper< std::vector< Csock *,std::allocator< Csock * > > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_FindSocksByRemoteHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_FindSocksByRemoteHost" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_FindSocksByRemoteHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_FindSocksByRemoteHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->FindSocksByRemoteHost((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj((new std::vector< Csock *,std::allocator< Csock * > >(result)), SWIGTYPE_p_std__vectorT_Csock_p_std__allocatorT_Csock_p_t_t, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_GetErrno(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_GetErrno" "', argument " "1"" of type '" "CSocketManager const *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  result = (int)((CSocketManager const *)arg1)->GetErrno();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_GetSelectTimeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint64_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_GetSelectTimeout" "', argument " "1"" of type '" "CSocketManager const *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  result = ((CSocketManager const *)arg1)->GetSelectTimeout();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_SetSelectTimeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  uint64_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_SetSelectTimeout", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_SetSelectTimeout" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocketManager_SetSelectTimeout" "', argument " "2"" of type '" "uint64_t""'");
  } 
  arg2 = static_cast< uint64_t >(val2);
  (arg1)->SetSelectTimeout(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_DelSockByAddr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  Csock *arg2 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_DelSockByAddr", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_DelSockByAddr" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_DelSockByAddr" "', argument " "2"" of type '" "Csock *""'"); 
  }
  arg2 = reinterpret_cast< Csock * >(argp2);
  (arg1)->DelSockByAddr(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_DelSock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_DelSock", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_DelSock" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocketManager_DelSock" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = static_cast< size_t >(val2);
  (arg1)->DelSock(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_SwapSockByIdx(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  Csock *arg2 = (Csock *) 0 ;
  size_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_SwapSockByIdx", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_SwapSockByIdx" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_SwapSockByIdx" "', argument " "2"" of type '" "Csock *""'"); 
  }
  arg2 = reinterpret_cast< Csock * >(argp2);
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocketManager_SwapSockByIdx" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = static_cast< size_t >(val3);
  result = (bool)(arg1)->SwapSockByIdx(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_SwapSockByAddr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  Csock *arg2 = (Csock *) 0 ;
  Csock *arg3 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_SwapSockByAddr", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_SwapSockByAddr" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocketManager_SwapSockByAddr" "', argument " "2"" of type '" "Csock *""'"); 
  }
  arg2 = reinterpret_cast< Csock * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSocketManager_SwapSockByAddr" "', argument " "3"" of type '" "Csock *""'"); 
  }
  arg3 = reinterpret_cast< Csock * >(argp3);
  result = (bool)(arg1)->SwapSockByAddr(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_GetBytesRead(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint64_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_GetBytesRead" "', argument " "1"" of type '" "CSocketManager const *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  result = ((CSocketManager const *)arg1)->GetBytesRead();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_GetBytesWritten(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  uint64_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_GetBytesWritten" "', argument " "1"" of type '" "CSocketManager const *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  result = ((CSocketManager const *)arg1)->GetBytesWritten();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_FDSetCheck(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  cs_sock_t arg2 ;
  std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > *arg3 = 0 ;
  CSocketManager::ECheckType arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_FDSetCheck", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_FDSetCheck" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocketManager_FDSetCheck" "', argument " "2"" of type '" "cs_sock_t""'");
  } 
  arg2 = static_cast< cs_sock_t >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSocketManager_FDSetCheck" "', argument " "3"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_FDSetCheck" "', argument " "3"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > &""'"); 
  }
  arg3 = reinterpret_cast< std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > * >(argp3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSocketManager_FDSetCheck" "', argument " "4"" of type '" "CSocketManager::ECheckType""'");
  } 
  arg4 = static_cast< CSocketManager::ECheckType >(val4);
  (arg1)->FDSetCheck(arg2,*arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocketManager_FDHasCheck(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocketManager *arg1 = (CSocketManager *) 0 ;
  cs_sock_t arg2 ;
  std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > *arg3 = 0 ;
  CSocketManager::ECheckType arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocketManager_FDHasCheck", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocketManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocketManager_FDHasCheck" "', argument " "1"" of type '" "CSocketManager *""'"); 
  }
  arg1 = reinterpret_cast< CSocketManager * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocketManager_FDHasCheck" "', argument " "2"" of type '" "cs_sock_t""'");
  } 
  arg2 = static_cast< cs_sock_t >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSocketManager_FDHasCheck" "', argument " "3"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocketManager_FDHasCheck" "', argument " "3"" of type '" "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > &""'"); 
  }
  arg3 = reinterpret_cast< std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > * >(argp3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSocketManager_FDHasCheck" "', argument " "4"" of type '" "CSocketManager::ECheckType""'");
  } 
  arg4 = static_cast< CSocketManager::ECheckType >(val4);
  result = (bool)(arg1)->FDHasCheck(arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CSocketManager_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CSocketManager, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CSocketManager_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_ZNCSocketManager(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  TSocketManager< CZNCSock > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_ZNCSocketManager", 0, 0, 0)) SWIG_fail;
  result = (TSocketManager< CZNCSock > *)new TSocketManager< CZNCSock >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TSocketManagerT_CZNCSock_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_ZNCSocketManager(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  TSocketManager< CZNCSock > *arg1 = (TSocketManager< CZNCSock > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_TSocketManagerT_CZNCSock_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_ZNCSocketManager" "', argument " "1"" of type '" "TSocketManager< CZNCSock > *""'"); 
  }
  arg1 = reinterpret_cast< TSocketManager< CZNCSock > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ZNCSocketManager_GetSockObj__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  TSocketManager< CZNCSock > *arg1 = (TSocketManager< CZNCSock > *) 0 ;
  CString *arg2 = 0 ;
  uint16_t arg3 ;
  int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  long val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  CZNCSock *result = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_TSocketManagerT_CZNCSock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ZNCSocketManager_GetSockObj" "', argument " "1"" of type '" "TSocketManager< CZNCSock > *""'"); 
  }
  arg1 = reinterpret_cast< TSocketManager< CZNCSock > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ZNCSocketManager_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "ZNCSocketManager_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "ZNCSocketManager_GetSockObj" "', argument " "3"" of type '" "uint16_t""'");
  } 
  arg3 = static_cast< uint16_t >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "ZNCSocketManager_GetSockObj" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = static_cast< int >(val4);
  result = (CZNCSock *)(arg1)->GetSockObj((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_ZNCSocketManager_GetSockObj__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  TSocketManager< CZNCSock > *arg1 = (TSocketManager< CZNCSock > *) 0 ;
  CString *arg2 = 0 ;
  uint16_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  long val3 ;
  int ecode3 = 0 ;
  CZNCSock *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_TSocketManagerT_CZNCSock_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ZNCSocketManager_GetSockObj" "', argument " "1"" of type '" "TSocketManager< CZNCSock > *""'"); 
  }
  arg1 = reinterpret_cast< TSocketManager< CZNCSock > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ZNCSocketManager_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "ZNCSocketManager_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "ZNCSocketManager_GetSockObj" "', argument " "3"" of type '" "uint16_t""'");
  } 
  arg3 = static_cast< uint16_t >(val3);
  result = (CZNCSock *)(arg1)->GetSockObj((CString const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_ZNCSocketManager_GetSockObj(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "ZNCSocketManager_GetSockObj", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_TSocketManagerT_CZNCSock_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_long(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_ZNCSocketManager_GetSockObj__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_TSocketManagerT_CZNCSock_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_long(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_ZNCSocketManager_GetSockObj__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'ZNCSocketManager_GetSockObj'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    TSocketManager< CZNCSock >::GetSockObj(CString const &,uint16_t,int)\n"
    "    TSocketManager< CZNCSock >::GetSockObj(CString const &,uint16_t)\n");
  return 0;
}


SWIGINTERN PyObject *ZNCSocketManager_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TSocketManagerT_CZNCSock_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *ZNCSocketManager_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CZNCSock__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  CZNCSock *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_CZNCSock" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = static_cast< int >(val1);
  result = (CZNCSock *)new CZNCSock(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CZNCSock, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CZNCSock__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CZNCSock *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CZNCSock *)new CZNCSock();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CZNCSock, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CZNCSock__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  unsigned short arg2 ;
  int arg3 ;
  int res1 = SWIG_OLDOBJ ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  CZNCSock *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CZNCSock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CZNCSock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CZNCSock" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CZNCSock" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  result = (CZNCSock *)new CZNCSock((CString const &)*arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CZNCSock, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CZNCSock__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  unsigned short arg2 ;
  int res1 = SWIG_OLDOBJ ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  CZNCSock *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CZNCSock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CZNCSock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CZNCSock" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  result = (CZNCSock *)new CZNCSock((CString const &)*arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CZNCSock, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CZNCSock(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CZNCSock", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CZNCSock__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_int(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_CZNCSock__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_new_CZNCSock__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_new_CZNCSock__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CZNCSock'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CZNCSock::CZNCSock(int)\n"
    "    CZNCSock::CZNCSock()\n"
    "    CZNCSock::CZNCSock(CString const &,unsigned short,int)\n"
    "    CZNCSock::CZNCSock(CString const &,unsigned short)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CZNCSock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNCSock *arg1 = (CZNCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNCSock, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CZNCSock" "', argument " "1"" of type '" "CZNCSock *""'"); 
  }
  arg1 = reinterpret_cast< CZNCSock * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNCSock_ConvertAddress(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNCSock *arg1 = (CZNCSock *) 0 ;
  sockaddr_storage *arg2 = (sockaddr_storage *) 0 ;
  socklen_t arg3 ;
  CString *arg4 = 0 ;
  unsigned short *arg5 = (unsigned short *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject *swig_obj[5] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNCSock_ConvertAddress", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNCSock_ConvertAddress" "', argument " "1"" of type '" "CZNCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CZNCSock * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_sockaddr_storage, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNCSock_ConvertAddress" "', argument " "2"" of type '" "sockaddr_storage const *""'"); 
  }
  arg2 = reinterpret_cast< sockaddr_storage * >(argp2);
  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_socklen_t,  0  | 0);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CZNCSock_ConvertAddress" "', argument " "3"" of type '" "socklen_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNCSock_ConvertAddress" "', argument " "3"" of type '" "socklen_t""'");
    } else {
      socklen_t * temp = reinterpret_cast< socklen_t * >(argp3);
      arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp;
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sIP");
    }
  }
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_unsigned_short, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CZNCSock_ConvertAddress" "', argument " "5"" of type '" "unsigned short *""'"); 
  }
  arg5 = reinterpret_cast< unsigned short * >(argp5);
  result = (int)((CZNCSock const *)arg1)->ConvertAddress((sockaddr_storage const *)arg2,arg3,*arg4,arg5);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNCSock_GetSSLPeerFingerprint(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNCSock *arg1 = (CZNCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNCSock_GetSSLPeerFingerprint" "', argument " "1"" of type '" "CZNCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CZNCSock * >(argp1);
  result = ((CZNCSock const *)arg1)->GetSSLPeerFingerprint();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNCSock_SetHostToVerifySSL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNCSock *arg1 = (CZNCSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNCSock_SetHostToVerifySSL", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNCSock_SetHostToVerifySSL" "', argument " "1"" of type '" "CZNCSock *""'"); 
  }
  arg1 = reinterpret_cast< CZNCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNCSock_SetHostToVerifySSL" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNCSock_SetHostToVerifySSL" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetHostToVerifySSL((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNCSock_SetSSLTrustedPeerFingerprints(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNCSock *arg1 = (CZNCSock *) 0 ;
  SCString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNCSock_SetSSLTrustedPeerFingerprints", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNCSock_SetSSLTrustedPeerFingerprints" "', argument " "1"" of type '" "CZNCSock *""'"); 
  }
  arg1 = reinterpret_cast< CZNCSock * >(argp1);
  {
    std::set< CString,std::less< CString >,std::allocator< CString > > *ptr = (std::set< CString,std::less< CString >,std::allocator< CString > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNCSock_SetSSLTrustedPeerFingerprints" "', argument " "2"" of type '" "SCString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNCSock_SetSSLTrustedPeerFingerprints" "', argument " "2"" of type '" "SCString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetSSLTrustedPeerFingerprints((SCString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNCSock_SetTrustAllCerts(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNCSock *arg1 = (CZNCSock *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNCSock_SetTrustAllCerts", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNCSock_SetTrustAllCerts" "', argument " "1"" of type '" "CZNCSock *""'"); 
  }
  arg1 = reinterpret_cast< CZNCSock * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNCSock_SetTrustAllCerts" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetTrustAllCerts(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNCSock_GetTrustAllCerts(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNCSock *arg1 = (CZNCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNCSock_GetTrustAllCerts" "', argument " "1"" of type '" "CZNCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CZNCSock * >(argp1);
  result = (bool)((CZNCSock const *)arg1)->GetTrustAllCerts();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNCSock_SetTrustPKI(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNCSock *arg1 = (CZNCSock *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNCSock_SetTrustPKI", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNCSock_SetTrustPKI" "', argument " "1"" of type '" "CZNCSock *""'"); 
  }
  arg1 = reinterpret_cast< CZNCSock * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNCSock_SetTrustPKI" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetTrustPKI(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNCSock_GetTrustPKI(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNCSock *arg1 = (CZNCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNCSock_GetTrustPKI" "', argument " "1"" of type '" "CZNCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CZNCSock * >(argp1);
  result = (bool)((CZNCSock const *)arg1)->GetTrustPKI();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNCSock_SetEncoding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNCSock *arg1 = (CZNCSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNCSock_SetEncoding", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNCSock_SetEncoding" "', argument " "1"" of type '" "CZNCSock *""'"); 
  }
  arg1 = reinterpret_cast< CZNCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNCSock_SetEncoding" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNCSock_SetEncoding" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetEncoding((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNCSock_GetRemoteIP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNCSock *arg1 = (CZNCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNCSock_GetRemoteIP" "', argument " "1"" of type '" "CZNCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CZNCSock * >(argp1);
  result = ((CZNCSock const *)arg1)->GetRemoteIP();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CZNCSock_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CZNCSock, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CZNCSock_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CSockManager(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockManager *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CSockManager", 0, 0, 0)) SWIG_fail;
  result = (CSockManager *)new CSockManager();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSockManager, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CSockManager(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CSockManager" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenHost__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  bool arg5 ;
  int arg6 ;
  CZNCSock *arg7 = (CZNCSock *) 0 ;
  unsigned int arg8 ;
  EAddrType arg9 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  bool val5 ;
  int ecode5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  unsigned int val8 ;
  int ecode8 = 0 ;
  int val9 ;
  int ecode9 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 9) || (nobjs > 9)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenHost" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSockManager_ListenHost" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenHost" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenHost" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSockManager_ListenHost" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenHost" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_ListenHost" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CSockManager_ListenHost" "', argument " "6"" of type '" "int""'");
  } 
  arg6 = static_cast< int >(val6);
  res7 = SWIG_ConvertPtr(swig_obj[6], &argp7,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "CSockManager_ListenHost" "', argument " "7"" of type '" "CZNCSock *""'"); 
  }
  arg7 = reinterpret_cast< CZNCSock * >(argp7);
  ecode8 = SWIG_AsVal_unsigned_SS_int(swig_obj[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "CSockManager_ListenHost" "', argument " "8"" of type '" "unsigned int""'");
  } 
  arg8 = static_cast< unsigned int >(val8);
  ecode9 = SWIG_AsVal_int(swig_obj[8], &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), "in method '" "CSockManager_ListenHost" "', argument " "9"" of type '" "EAddrType""'");
  } 
  arg9 = static_cast< EAddrType >(val9);
  result = (bool)(arg1)->ListenHost(arg2,(CString const &)*arg3,(CString const &)*arg4,arg5,arg6,arg7,arg8,arg9);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenHost__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  bool arg5 ;
  int arg6 ;
  CZNCSock *arg7 = (CZNCSock *) 0 ;
  unsigned int arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  bool val5 ;
  int ecode5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  unsigned int val8 ;
  int ecode8 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 8) || (nobjs > 8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenHost" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSockManager_ListenHost" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenHost" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenHost" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSockManager_ListenHost" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenHost" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_ListenHost" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CSockManager_ListenHost" "', argument " "6"" of type '" "int""'");
  } 
  arg6 = static_cast< int >(val6);
  res7 = SWIG_ConvertPtr(swig_obj[6], &argp7,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "CSockManager_ListenHost" "', argument " "7"" of type '" "CZNCSock *""'"); 
  }
  arg7 = reinterpret_cast< CZNCSock * >(argp7);
  ecode8 = SWIG_AsVal_unsigned_SS_int(swig_obj[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "CSockManager_ListenHost" "', argument " "8"" of type '" "unsigned int""'");
  } 
  arg8 = static_cast< unsigned int >(val8);
  result = (bool)(arg1)->ListenHost(arg2,(CString const &)*arg3,(CString const &)*arg4,arg5,arg6,arg7,arg8);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenHost__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  bool arg5 ;
  int arg6 ;
  CZNCSock *arg7 = (CZNCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  bool val5 ;
  int ecode5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 7) || (nobjs > 7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenHost" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSockManager_ListenHost" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenHost" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenHost" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSockManager_ListenHost" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenHost" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_ListenHost" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CSockManager_ListenHost" "', argument " "6"" of type '" "int""'");
  } 
  arg6 = static_cast< int >(val6);
  res7 = SWIG_ConvertPtr(swig_obj[6], &argp7,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "CSockManager_ListenHost" "', argument " "7"" of type '" "CZNCSock *""'"); 
  }
  arg7 = reinterpret_cast< CZNCSock * >(argp7);
  result = (bool)(arg1)->ListenHost(arg2,(CString const &)*arg3,(CString const &)*arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenHost__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  bool arg5 ;
  int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  bool val5 ;
  int ecode5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenHost" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSockManager_ListenHost" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenHost" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenHost" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSockManager_ListenHost" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenHost" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_ListenHost" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CSockManager_ListenHost" "', argument " "6"" of type '" "int""'");
  } 
  arg6 = static_cast< int >(val6);
  result = (bool)(arg1)->ListenHost(arg2,(CString const &)*arg3,(CString const &)*arg4,arg5,arg6);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenHost__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  bool val5 ;
  int ecode5 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenHost" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSockManager_ListenHost" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenHost" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenHost" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSockManager_ListenHost" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenHost" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_ListenHost" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  result = (bool)(arg1)->ListenHost(arg2,(CString const &)*arg3,(CString const &)*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenHost__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenHost" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSockManager_ListenHost" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenHost" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenHost" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSockManager_ListenHost" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenHost" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (bool)(arg1)->ListenHost(arg2,(CString const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenHost(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[10] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSockManager_ListenHost", 0, 9, argv))) SWIG_fail;
  --argc;
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CSockManager_ListenHost__SWIG_5(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_bool(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_CSockManager_ListenHost__SWIG_4(self, argc, argv);
            }
          }
        }
      }
    }
  }
  if (argc == 6) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_bool(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              {
                int res = SWIG_AsVal_int(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                return _wrap_CSockManager_ListenHost__SWIG_3(self, argc, argv);
              }
            }
          }
        }
      }
    }
  }
  if (argc == 7) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_bool(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              {
                int res = SWIG_AsVal_int(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                void *vptr = 0;
                int res = SWIG_ConvertPtr(argv[6], &vptr, SWIGTYPE_p_CZNCSock, 0);
                _v = SWIG_CheckState(res);
                if (_v) {
                  return _wrap_CSockManager_ListenHost__SWIG_2(self, argc, argv);
                }
              }
            }
          }
        }
      }
    }
  }
  if (argc == 8) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_bool(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              {
                int res = SWIG_AsVal_int(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                void *vptr = 0;
                int res = SWIG_ConvertPtr(argv[6], &vptr, SWIGTYPE_p_CZNCSock, 0);
                _v = SWIG_CheckState(res);
                if (_v) {
                  {
                    int res = SWIG_AsVal_unsigned_SS_int(argv[7], NULL);
                    _v = SWIG_CheckState(res);
                  }
                  if (_v) {
                    return _wrap_CSockManager_ListenHost__SWIG_1(self, argc, argv);
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  if (argc == 9) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_bool(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              {
                int res = SWIG_AsVal_int(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                void *vptr = 0;
                int res = SWIG_ConvertPtr(argv[6], &vptr, SWIGTYPE_p_CZNCSock, 0);
                _v = SWIG_CheckState(res);
                if (_v) {
                  {
                    int res = SWIG_AsVal_unsigned_SS_int(argv[7], NULL);
                    _v = SWIG_CheckState(res);
                  }
                  if (_v) {
                    {
                      int res = SWIG_AsVal_int(argv[8], NULL);
                      _v = SWIG_CheckState(res);
                    }
                    if (_v) {
                      return _wrap_CSockManager_ListenHost__SWIG_0(self, argc, argv);
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSockManager_ListenHost'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSockManager::ListenHost(unsigned short,CString const &,CString const &,bool,int,CZNCSock *,unsigned int,EAddrType)\n"
    "    CSockManager::ListenHost(unsigned short,CString const &,CString const &,bool,int,CZNCSock *,unsigned int)\n"
    "    CSockManager::ListenHost(unsigned short,CString const &,CString const &,bool,int,CZNCSock *)\n"
    "    CSockManager::ListenHost(unsigned short,CString const &,CString const &,bool,int)\n"
    "    CSockManager::ListenHost(unsigned short,CString const &,CString const &,bool)\n"
    "    CSockManager::ListenHost(unsigned short,CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAll__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  int arg5 ;
  CZNCSock *arg6 = (CZNCSock *) 0 ;
  unsigned int arg7 ;
  EAddrType arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  int val8 ;
  int ecode8 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 8) || (nobjs > 8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenAll" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSockManager_ListenAll" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenAll" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenAll" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenAll" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_ListenAll" "', argument " "5"" of type '" "int""'");
  } 
  arg5 = static_cast< int >(val5);
  res6 = SWIG_ConvertPtr(swig_obj[5], &argp6,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "CSockManager_ListenAll" "', argument " "6"" of type '" "CZNCSock *""'"); 
  }
  arg6 = reinterpret_cast< CZNCSock * >(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CSockManager_ListenAll" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = static_cast< unsigned int >(val7);
  ecode8 = SWIG_AsVal_int(swig_obj[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "CSockManager_ListenAll" "', argument " "8"" of type '" "EAddrType""'");
  } 
  arg8 = static_cast< EAddrType >(val8);
  result = (bool)(arg1)->ListenAll(arg2,(CString const &)*arg3,arg4,arg5,arg6,arg7,arg8);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAll__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  int arg5 ;
  CZNCSock *arg6 = (CZNCSock *) 0 ;
  unsigned int arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 7) || (nobjs > 7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenAll" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSockManager_ListenAll" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenAll" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenAll" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenAll" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_ListenAll" "', argument " "5"" of type '" "int""'");
  } 
  arg5 = static_cast< int >(val5);
  res6 = SWIG_ConvertPtr(swig_obj[5], &argp6,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "CSockManager_ListenAll" "', argument " "6"" of type '" "CZNCSock *""'"); 
  }
  arg6 = reinterpret_cast< CZNCSock * >(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CSockManager_ListenAll" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = static_cast< unsigned int >(val7);
  result = (bool)(arg1)->ListenAll(arg2,(CString const &)*arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAll__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  int arg5 ;
  CZNCSock *arg6 = (CZNCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenAll" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSockManager_ListenAll" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenAll" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenAll" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenAll" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_ListenAll" "', argument " "5"" of type '" "int""'");
  } 
  arg5 = static_cast< int >(val5);
  res6 = SWIG_ConvertPtr(swig_obj[5], &argp6,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "CSockManager_ListenAll" "', argument " "6"" of type '" "CZNCSock *""'"); 
  }
  arg6 = reinterpret_cast< CZNCSock * >(argp6);
  result = (bool)(arg1)->ListenAll(arg2,(CString const &)*arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAll__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenAll" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSockManager_ListenAll" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenAll" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenAll" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenAll" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_ListenAll" "', argument " "5"" of type '" "int""'");
  } 
  arg5 = static_cast< int >(val5);
  result = (bool)(arg1)->ListenAll(arg2,(CString const &)*arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAll__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenAll" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSockManager_ListenAll" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenAll" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenAll" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenAll" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->ListenAll(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAll__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenAll" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSockManager_ListenAll" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenAll" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenAll" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->ListenAll(arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAll(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[9] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSockManager_ListenAll", 0, 8, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CSockManager_ListenAll__SWIG_5(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CSockManager_ListenAll__SWIG_4(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_CSockManager_ListenAll__SWIG_3(self, argc, argv);
            }
          }
        }
      }
    }
  }
  if (argc == 6) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              void *vptr = 0;
              int res = SWIG_ConvertPtr(argv[5], &vptr, SWIGTYPE_p_CZNCSock, 0);
              _v = SWIG_CheckState(res);
              if (_v) {
                return _wrap_CSockManager_ListenAll__SWIG_2(self, argc, argv);
              }
            }
          }
        }
      }
    }
  }
  if (argc == 7) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              void *vptr = 0;
              int res = SWIG_ConvertPtr(argv[5], &vptr, SWIGTYPE_p_CZNCSock, 0);
              _v = SWIG_CheckState(res);
              if (_v) {
                {
                  int res = SWIG_AsVal_unsigned_SS_int(argv[6], NULL);
                  _v = SWIG_CheckState(res);
                }
                if (_v) {
                  return _wrap_CSockManager_ListenAll__SWIG_1(self, argc, argv);
                }
              }
            }
          }
        }
      }
    }
  }
  if (argc == 8) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              void *vptr = 0;
              int res = SWIG_ConvertPtr(argv[5], &vptr, SWIGTYPE_p_CZNCSock, 0);
              _v = SWIG_CheckState(res);
              if (_v) {
                {
                  int res = SWIG_AsVal_unsigned_SS_int(argv[6], NULL);
                  _v = SWIG_CheckState(res);
                }
                if (_v) {
                  {
                    int res = SWIG_AsVal_int(argv[7], NULL);
                    _v = SWIG_CheckState(res);
                  }
                  if (_v) {
                    return _wrap_CSockManager_ListenAll__SWIG_0(self, argc, argv);
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSockManager_ListenAll'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSockManager::ListenAll(unsigned short,CString const &,bool,int,CZNCSock *,unsigned int,EAddrType)\n"
    "    CSockManager::ListenAll(unsigned short,CString const &,bool,int,CZNCSock *,unsigned int)\n"
    "    CSockManager::ListenAll(unsigned short,CString const &,bool,int,CZNCSock *)\n"
    "    CSockManager::ListenAll(unsigned short,CString const &,bool,int)\n"
    "    CSockManager::ListenAll(unsigned short,CString const &,bool)\n"
    "    CSockManager::ListenAll(unsigned short,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenRand__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  int arg5 ;
  CZNCSock *arg6 = (CZNCSock *) 0 ;
  unsigned int arg7 ;
  EAddrType arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  int val8 ;
  int ecode8 = 0 ;
  unsigned short result;
  
  (void)self;
  if ((nobjs < 8) || (nobjs > 8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenRand" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenRand" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenRand" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenRand" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_ListenRand" "', argument " "5"" of type '" "int""'");
  } 
  arg5 = static_cast< int >(val5);
  res6 = SWIG_ConvertPtr(swig_obj[5], &argp6,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "CSockManager_ListenRand" "', argument " "6"" of type '" "CZNCSock *""'"); 
  }
  arg6 = reinterpret_cast< CZNCSock * >(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CSockManager_ListenRand" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = static_cast< unsigned int >(val7);
  ecode8 = SWIG_AsVal_int(swig_obj[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "CSockManager_ListenRand" "', argument " "8"" of type '" "EAddrType""'");
  } 
  arg8 = static_cast< EAddrType >(val8);
  result = (unsigned short)(arg1)->ListenRand((CString const &)*arg2,(CString const &)*arg3,arg4,arg5,arg6,arg7,arg8);
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenRand__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  int arg5 ;
  CZNCSock *arg6 = (CZNCSock *) 0 ;
  unsigned int arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  unsigned short result;
  
  (void)self;
  if ((nobjs < 7) || (nobjs > 7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenRand" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenRand" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenRand" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenRand" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_ListenRand" "', argument " "5"" of type '" "int""'");
  } 
  arg5 = static_cast< int >(val5);
  res6 = SWIG_ConvertPtr(swig_obj[5], &argp6,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "CSockManager_ListenRand" "', argument " "6"" of type '" "CZNCSock *""'"); 
  }
  arg6 = reinterpret_cast< CZNCSock * >(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CSockManager_ListenRand" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = static_cast< unsigned int >(val7);
  result = (unsigned short)(arg1)->ListenRand((CString const &)*arg2,(CString const &)*arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenRand__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  int arg5 ;
  CZNCSock *arg6 = (CZNCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned short result;
  
  (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenRand" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenRand" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenRand" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenRand" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_ListenRand" "', argument " "5"" of type '" "int""'");
  } 
  arg5 = static_cast< int >(val5);
  res6 = SWIG_ConvertPtr(swig_obj[5], &argp6,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "CSockManager_ListenRand" "', argument " "6"" of type '" "CZNCSock *""'"); 
  }
  arg6 = reinterpret_cast< CZNCSock * >(argp6);
  result = (unsigned short)(arg1)->ListenRand((CString const &)*arg2,(CString const &)*arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenRand__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  unsigned short result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenRand" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenRand" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenRand" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenRand" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_ListenRand" "', argument " "5"" of type '" "int""'");
  } 
  arg5 = static_cast< int >(val5);
  result = (unsigned short)(arg1)->ListenRand((CString const &)*arg2,(CString const &)*arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenRand__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  unsigned short result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenRand" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenRand" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenRand" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenRand" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (unsigned short)(arg1)->ListenRand((CString const &)*arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenRand__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  unsigned short result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenRand" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenRand" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenRand" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (unsigned short)(arg1)->ListenRand((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenRand(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[9] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSockManager_ListenRand", 0, 8, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CSockManager_ListenRand__SWIG_5(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CSockManager_ListenRand__SWIG_4(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_CSockManager_ListenRand__SWIG_3(self, argc, argv);
            }
          }
        }
      }
    }
  }
  if (argc == 6) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              void *vptr = 0;
              int res = SWIG_ConvertPtr(argv[5], &vptr, SWIGTYPE_p_CZNCSock, 0);
              _v = SWIG_CheckState(res);
              if (_v) {
                return _wrap_CSockManager_ListenRand__SWIG_2(self, argc, argv);
              }
            }
          }
        }
      }
    }
  }
  if (argc == 7) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              void *vptr = 0;
              int res = SWIG_ConvertPtr(argv[5], &vptr, SWIGTYPE_p_CZNCSock, 0);
              _v = SWIG_CheckState(res);
              if (_v) {
                {
                  int res = SWIG_AsVal_unsigned_SS_int(argv[6], NULL);
                  _v = SWIG_CheckState(res);
                }
                if (_v) {
                  return _wrap_CSockManager_ListenRand__SWIG_1(self, argc, argv);
                }
              }
            }
          }
        }
      }
    }
  }
  if (argc == 8) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              void *vptr = 0;
              int res = SWIG_ConvertPtr(argv[5], &vptr, SWIGTYPE_p_CZNCSock, 0);
              _v = SWIG_CheckState(res);
              if (_v) {
                {
                  int res = SWIG_AsVal_unsigned_SS_int(argv[6], NULL);
                  _v = SWIG_CheckState(res);
                }
                if (_v) {
                  {
                    int res = SWIG_AsVal_int(argv[7], NULL);
                    _v = SWIG_CheckState(res);
                  }
                  if (_v) {
                    return _wrap_CSockManager_ListenRand__SWIG_0(self, argc, argv);
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSockManager_ListenRand'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSockManager::ListenRand(CString const &,CString const &,bool,int,CZNCSock *,unsigned int,EAddrType)\n"
    "    CSockManager::ListenRand(CString const &,CString const &,bool,int,CZNCSock *,unsigned int)\n"
    "    CSockManager::ListenRand(CString const &,CString const &,bool,int,CZNCSock *)\n"
    "    CSockManager::ListenRand(CString const &,CString const &,bool,int)\n"
    "    CSockManager::ListenRand(CString const &,CString const &,bool)\n"
    "    CSockManager::ListenRand(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAllRand__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  int arg4 ;
  CZNCSock *arg5 = (CZNCSock *) 0 ;
  unsigned int arg6 ;
  EAddrType arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  int val7 ;
  int ecode7 = 0 ;
  unsigned short result;
  
  (void)self;
  if ((nobjs < 7) || (nobjs > 7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenAllRand" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenAllRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenAllRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSockManager_ListenAllRand" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenAllRand" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = static_cast< int >(val4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CSockManager_ListenAllRand" "', argument " "5"" of type '" "CZNCSock *""'"); 
  }
  arg5 = reinterpret_cast< CZNCSock * >(argp5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CSockManager_ListenAllRand" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = static_cast< unsigned int >(val6);
  ecode7 = SWIG_AsVal_int(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CSockManager_ListenAllRand" "', argument " "7"" of type '" "EAddrType""'");
  } 
  arg7 = static_cast< EAddrType >(val7);
  result = (unsigned short)(arg1)->ListenAllRand((CString const &)*arg2,arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAllRand__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  int arg4 ;
  CZNCSock *arg5 = (CZNCSock *) 0 ;
  unsigned int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  unsigned short result;
  
  (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenAllRand" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenAllRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenAllRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSockManager_ListenAllRand" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenAllRand" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = static_cast< int >(val4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CSockManager_ListenAllRand" "', argument " "5"" of type '" "CZNCSock *""'"); 
  }
  arg5 = reinterpret_cast< CZNCSock * >(argp5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CSockManager_ListenAllRand" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = static_cast< unsigned int >(val6);
  result = (unsigned short)(arg1)->ListenAllRand((CString const &)*arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAllRand__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  int arg4 ;
  CZNCSock *arg5 = (CZNCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  unsigned short result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenAllRand" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenAllRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenAllRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSockManager_ListenAllRand" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenAllRand" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = static_cast< int >(val4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CSockManager_ListenAllRand" "', argument " "5"" of type '" "CZNCSock *""'"); 
  }
  arg5 = reinterpret_cast< CZNCSock * >(argp5);
  result = (unsigned short)(arg1)->ListenAllRand((CString const &)*arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAllRand__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  unsigned short result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenAllRand" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenAllRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenAllRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSockManager_ListenAllRand" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSockManager_ListenAllRand" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = static_cast< int >(val4);
  result = (unsigned short)(arg1)->ListenAllRand((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAllRand__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  unsigned short result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenAllRand" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenAllRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenAllRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSockManager_ListenAllRand" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (unsigned short)(arg1)->ListenAllRand((CString const &)*arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAllRand__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenAllRand" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenAllRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenAllRand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (unsigned short)(arg1)->ListenAllRand((CString const &)*arg2);
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenAllRand(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[8] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSockManager_ListenAllRand", 0, 7, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CSockManager_ListenAllRand__SWIG_5(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CSockManager_ListenAllRand__SWIG_4(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CSockManager_ListenAllRand__SWIG_3(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            void *vptr = 0;
            int res = SWIG_ConvertPtr(argv[4], &vptr, SWIGTYPE_p_CZNCSock, 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_CSockManager_ListenAllRand__SWIG_2(self, argc, argv);
            }
          }
        }
      }
    }
  }
  if (argc == 6) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            void *vptr = 0;
            int res = SWIG_ConvertPtr(argv[4], &vptr, SWIGTYPE_p_CZNCSock, 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              {
                int res = SWIG_AsVal_unsigned_SS_int(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                return _wrap_CSockManager_ListenAllRand__SWIG_1(self, argc, argv);
              }
            }
          }
        }
      }
    }
  }
  if (argc == 7) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            void *vptr = 0;
            int res = SWIG_ConvertPtr(argv[4], &vptr, SWIGTYPE_p_CZNCSock, 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              {
                int res = SWIG_AsVal_unsigned_SS_int(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                {
                  int res = SWIG_AsVal_int(argv[6], NULL);
                  _v = SWIG_CheckState(res);
                }
                if (_v) {
                  return _wrap_CSockManager_ListenAllRand__SWIG_0(self, argc, argv);
                }
              }
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSockManager_ListenAllRand'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSockManager::ListenAllRand(CString const &,bool,int,CZNCSock *,unsigned int,EAddrType)\n"
    "    CSockManager::ListenAllRand(CString const &,bool,int,CZNCSock *,unsigned int)\n"
    "    CSockManager::ListenAllRand(CString const &,bool,int,CZNCSock *)\n"
    "    CSockManager::ListenAllRand(CString const &,bool,int)\n"
    "    CSockManager::ListenAllRand(CString const &,bool)\n"
    "    CSockManager::ListenAllRand(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSockManager_Connect__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  CString *arg4 = 0 ;
  int arg5 ;
  bool arg6 ;
  CString *arg7 = 0 ;
  CZNCSock *arg8 = (CZNCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int val5 ;
  int ecode5 = 0 ;
  bool val6 ;
  int ecode6 = 0 ;
  int res7 = SWIG_OLDOBJ ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  
  (void)self;
  if ((nobjs < 8) || (nobjs > 8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_Connect" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSockManager_Connect" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSockManager_Connect" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_Connect" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_Connect" "', argument " "5"" of type '" "int""'");
  } 
  arg5 = static_cast< int >(val5);
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CSockManager_Connect" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  {
    CString *ptr = (CString *)0;
    res7 = SWIG_AsPtr_CString(swig_obj[6], &ptr);
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "CSockManager_Connect" "', argument " "7"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_Connect" "', argument " "7"" of type '" "CString const &""'"); 
    }
    arg7 = ptr;
  }
  res8 = SWIG_ConvertPtr(swig_obj[7], &argp8,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "CSockManager_Connect" "', argument " "8"" of type '" "CZNCSock *""'"); 
  }
  arg8 = reinterpret_cast< CZNCSock * >(argp8);
  (arg1)->Connect((CString const &)*arg2,arg3,(CString const &)*arg4,arg5,arg6,(CString const &)*arg7,arg8);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res7)) delete arg7;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res7)) delete arg7;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_Connect__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  CString *arg4 = 0 ;
  int arg5 ;
  bool arg6 ;
  CString *arg7 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int val5 ;
  int ecode5 = 0 ;
  bool val6 ;
  int ecode6 = 0 ;
  int res7 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 7) || (nobjs > 7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_Connect" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSockManager_Connect" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSockManager_Connect" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_Connect" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_Connect" "', argument " "5"" of type '" "int""'");
  } 
  arg5 = static_cast< int >(val5);
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CSockManager_Connect" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  {
    CString *ptr = (CString *)0;
    res7 = SWIG_AsPtr_CString(swig_obj[6], &ptr);
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "CSockManager_Connect" "', argument " "7"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_Connect" "', argument " "7"" of type '" "CString const &""'"); 
    }
    arg7 = ptr;
  }
  (arg1)->Connect((CString const &)*arg2,arg3,(CString const &)*arg4,arg5,arg6,(CString const &)*arg7);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res7)) delete arg7;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res7)) delete arg7;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_Connect__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  CString *arg4 = 0 ;
  int arg5 ;
  bool arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int val5 ;
  int ecode5 = 0 ;
  bool val6 ;
  int ecode6 = 0 ;
  
  (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_Connect" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSockManager_Connect" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSockManager_Connect" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_Connect" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_Connect" "', argument " "5"" of type '" "int""'");
  } 
  arg5 = static_cast< int >(val5);
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CSockManager_Connect" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  (arg1)->Connect((CString const &)*arg2,arg3,(CString const &)*arg4,arg5,arg6);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_Connect__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  CString *arg4 = 0 ;
  int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int val5 ;
  int ecode5 = 0 ;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_Connect" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSockManager_Connect" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSockManager_Connect" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_Connect" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSockManager_Connect" "', argument " "5"" of type '" "int""'");
  } 
  arg5 = static_cast< int >(val5);
  (arg1)->Connect((CString const &)*arg2,arg3,(CString const &)*arg4,arg5);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_Connect__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_Connect" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSockManager_Connect" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSockManager_Connect" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_Connect" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->Connect((CString const &)*arg2,arg3,(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_Connect(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[9] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSockManager_Connect", 0, 8, argv))) SWIG_fail;
  --argc;
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CSockManager_Connect__SWIG_4(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_CSockManager_Connect__SWIG_3(self, argc, argv);
            }
          }
        }
      }
    }
  }
  if (argc == 6) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              {
                int res = SWIG_AsVal_bool(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                return _wrap_CSockManager_Connect__SWIG_2(self, argc, argv);
              }
            }
          }
        }
      }
    }
  }
  if (argc == 7) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              {
                int res = SWIG_AsVal_bool(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                int res = SWIG_AsPtr_CString(argv[6], (CString**)(0));
                _v = SWIG_CheckState(res);
                if (_v) {
                  return _wrap_CSockManager_Connect__SWIG_1(self, argc, argv);
                }
              }
            }
          }
        }
      }
    }
  }
  if (argc == 8) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              {
                int res = SWIG_AsVal_bool(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                int res = SWIG_AsPtr_CString(argv[6], (CString**)(0));
                _v = SWIG_CheckState(res);
                if (_v) {
                  void *vptr = 0;
                  int res = SWIG_ConvertPtr(argv[7], &vptr, SWIGTYPE_p_CZNCSock, 0);
                  _v = SWIG_CheckState(res);
                  if (_v) {
                    return _wrap_CSockManager_Connect__SWIG_0(self, argc, argv);
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSockManager_Connect'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSockManager::Connect(CString const &,unsigned short,CString const &,int,bool,CString const &,CZNCSock *)\n"
    "    CSockManager::Connect(CString const &,unsigned short,CString const &,int,bool,CString const &)\n"
    "    CSockManager::Connect(CString const &,unsigned short,CString const &,int,bool)\n"
    "    CSockManager::Connect(CString const &,unsigned short,CString const &,int)\n"
    "    CSockManager::Connect(CString const &,unsigned short,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenUnix__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CZNCSock *arg4 = (CZNCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenUnix" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenUnix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenUnix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenUnix" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenUnix" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSockManager_ListenUnix" "', argument " "4"" of type '" "CZNCSock *""'"); 
  }
  arg4 = reinterpret_cast< CZNCSock * >(argp4);
  result = (bool)(arg1)->ListenUnix((CString const &)*arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenUnix__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ListenUnix" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ListenUnix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenUnix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ListenUnix" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ListenUnix" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->ListenUnix((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ListenUnix(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSockManager_ListenUnix", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CSockManager_ListenUnix__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CZNCSock, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CSockManager_ListenUnix__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSockManager_ListenUnix'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSockManager::ListenUnix(CString const &,CString const &,CZNCSock *)\n"
    "    CSockManager::ListenUnix(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSockManager_ConnectUnix__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CZNCSock *arg4 = (CZNCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ConnectUnix" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ConnectUnix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ConnectUnix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ConnectUnix" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ConnectUnix" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSockManager_ConnectUnix" "', argument " "4"" of type '" "CZNCSock *""'"); 
  }
  arg4 = reinterpret_cast< CZNCSock * >(argp4);
  result = (bool)(arg1)->ConnectUnix((CString const &)*arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ConnectUnix__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_ConnectUnix" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_ConnectUnix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ConnectUnix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSockManager_ConnectUnix" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_ConnectUnix" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->ConnectUnix((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_ConnectUnix(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSockManager_ConnectUnix", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CSockManager_ConnectUnix__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSockManager, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CZNCSock, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CSockManager_ConnectUnix__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSockManager_ConnectUnix'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSockManager::ConnectUnix(CString const &,CString const &,CZNCSock *)\n"
    "    CSockManager::ConnectUnix(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSockManager_GetAnonConnectionCount(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSockManager_GetAnonConnectionCount", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_GetAnonConnectionCount" "', argument " "1"" of type '" "CSockManager const *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_GetAnonConnectionCount" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSockManager_GetAnonConnectionCount" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (unsigned int)((CSockManager const *)arg1)->GetAnonConnectionCount((CString const &)*arg2);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSockManager_DelSockByAddr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSockManager *arg1 = (CSockManager *) 0 ;
  Csock *arg2 = (Csock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSockManager_DelSockByAddr", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSockManager, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSockManager_DelSockByAddr" "', argument " "1"" of type '" "CSockManager *""'"); 
  }
  arg1 = reinterpret_cast< CSockManager * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Csock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSockManager_DelSockByAddr" "', argument " "2"" of type '" "Csock *""'"); 
  }
  arg2 = reinterpret_cast< Csock * >(argp2);
  (arg1)->DelSockByAddr(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CSockManager_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CSockManager, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CSockManager_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CSocket__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CSocket *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CSocket" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (CSocket *)new CSocket(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSocket, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CSocket__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  CSocket *result = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CSocket" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CSocket" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CSocket" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CSocket" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_CSocket" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = static_cast< int >(val4);
  result = (CSocket *)new CSocket(arg1,(CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSocket, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CSocket__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  CSocket *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CSocket" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CSocket" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CSocket" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CSocket" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  result = (CSocket *)new CSocket(arg1,(CString const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSocket, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CSocket(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CSocket", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CSocket__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_new_CSocket__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_new_CSocket__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CSocket'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSocket::CSocket(CModule *)\n"
    "    CSocket::CSocket(CModule *,CString const &,unsigned short,int)\n"
    "    CSocket::CSocket(CModule *,CString const &,unsigned short)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CSocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CSocket" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_Connect__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_Connect" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  result = (bool)(arg1)->Connect();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_Listen__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  uint16_t arg2 ;
  int arg3 ;
  CString *arg4 = 0 ;
  uint32_t arg5 ;
  bool arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  long val5 ;
  int ecode5 = 0 ;
  bool val6 ;
  int ecode6 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_Listen" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocket_Listen" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocket_Listen" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSocket_Listen" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocket_Listen" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_long(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSocket_Listen" "', argument " "5"" of type '" "uint32_t""'");
  } 
  arg5 = static_cast< uint32_t >(val5);
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CSocket_Listen" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  result = (bool)(arg1)->Listen(arg2,arg3,(CString const &)*arg4,arg5,arg6);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_Listen__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  uint16_t arg2 ;
  int arg3 ;
  CString *arg4 = 0 ;
  uint32_t arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  long val5 ;
  int ecode5 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_Listen" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocket_Listen" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocket_Listen" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSocket_Listen" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocket_Listen" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_long(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSocket_Listen" "', argument " "5"" of type '" "uint32_t""'");
  } 
  arg5 = static_cast< uint32_t >(val5);
  result = (bool)(arg1)->Listen(arg2,arg3,(CString const &)*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_Listen__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  uint16_t arg2 ;
  int arg3 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_Listen" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocket_Listen" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocket_Listen" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CSocket_Listen" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocket_Listen" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (bool)(arg1)->Listen(arg2,arg3,(CString const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_Listen__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  uint16_t arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_Listen" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocket_Listen" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocket_Listen" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  result = (bool)(arg1)->Listen(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_Listen__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  uint16_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_Listen" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocket_Listen" "', argument " "2"" of type '" "uint16_t""'");
  } 
  arg2 = static_cast< uint16_t >(val2);
  result = (bool)(arg1)->Listen(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_ReachedMaxBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_ReachedMaxBuffer" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  (arg1)->ReachedMaxBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_SockError(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  int arg2 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocket_SockError", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_SockError" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocket_SockError" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CSocket_SockError" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocket_SockError" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->SockError(arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_ConnectionFrom(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocket_ConnectionFrom", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_ConnectionFrom" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocket_ConnectionFrom" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocket_ConnectionFrom" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocket_ConnectionFrom" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  result = (bool)(arg1)->ConnectionFrom((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_Connect__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  bool arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_Connect" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocket_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocket_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocket_Connect" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSocket_Connect" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CSocket_Connect" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  result = (bool)(arg1)->Connect((CString const &)*arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_Connect__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_Connect" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocket_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocket_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocket_Connect" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSocket_Connect" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->Connect((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_Connect__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_Connect" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocket_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocket_Connect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocket_Connect" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  result = (bool)(arg1)->Connect((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_Connect(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSocket_Connect", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocket, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CSocket_Connect__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocket, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CSocket_Connect__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocket, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CSocket_Connect__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocket, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_unsigned_SS_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_CSocket_Connect__SWIG_1(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSocket_Connect'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSocket::Connect()\n"
    "    CSocket::Connect(CString const &,unsigned short,bool,unsigned int)\n"
    "    CSocket::Connect(CString const &,unsigned short,bool)\n"
    "    CSocket::Connect(CString const &,unsigned short)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSocket_Listen__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  unsigned short arg2 ;
  bool arg3 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_Listen" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocket_Listen" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocket_Listen" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CSocket_Listen" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  result = (bool)(arg1)->Listen(arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_Listen__SWIG_6(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  unsigned short arg2 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_Listen" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CSocket_Listen" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CSocket_Listen" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)(arg1)->Listen(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_Listen(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[7] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CSocket_Listen", 0, 6, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocket, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CSocket_Listen__SWIG_4(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocket, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CSocket_Listen__SWIG_6(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocket, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CSocket_Listen__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocket, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CSocket_Listen__SWIG_5(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocket, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CSocket_Listen__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocket, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_long(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_CSocket_Listen__SWIG_1(self, argc, argv);
            }
          }
        }
      }
    }
  }
  if (argc == 6) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CSocket, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_long(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              {
                int res = SWIG_AsVal_bool(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                return _wrap_CSocket_Listen__SWIG_0(self, argc, argv);
              }
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CSocket_Listen'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CSocket::Listen(uint16_t,int,CString const &,uint32_t,bool)\n"
    "    CSocket::Listen(uint16_t,int,CString const &,uint32_t)\n"
    "    CSocket::Listen(uint16_t,int,CString const &)\n"
    "    CSocket::Listen(uint16_t,int)\n"
    "    CSocket::Listen(uint16_t)\n"
    "    CSocket::Listen(unsigned short,bool,unsigned int)\n"
    "    CSocket::Listen(unsigned short,bool)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CSocket_ConnectUnix(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocket_ConnectUnix", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_ConnectUnix" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocket_ConnectUnix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocket_ConnectUnix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->ConnectUnix((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_ListenUnix(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocket_ListenUnix", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_ListenUnix" "', argument " "1"" of type '" "CSocket *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocket_ListenUnix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocket_ListenUnix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->ListenUnix((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_ConstructSockName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CSocket_ConstructSockName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_ConstructSockName" "', argument " "1"" of type '" "CSocket const *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CSocket_ConstructSockName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CSocket_ConstructSockName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CSocket const *)arg1)->ConstructSockName((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CSocket_GetModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CSocket *arg1 = (CSocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CModule *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CSocket_GetModule" "', argument " "1"" of type '" "CSocket const *""'"); 
  }
  arg1 = reinterpret_cast< CSocket * >(argp1);
  result = (CModule *)((CSocket const *)arg1)->GetModule();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModule, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CSocket_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CSocket, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CSocket_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CIRCSocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSocket *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CIRCSocket", 0, 0, 0)) SWIG_fail;
  result = (CIRCSocket *)new CIRCSocket();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCSocket, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CIRCSocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSocket *arg1 = (CIRCSocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSocket, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CIRCSocket" "', argument " "1"" of type '" "CIRCSocket *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSocket * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CIRCSocket_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CIRCSocket, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CIRCSocket_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CFile__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CFile *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CFile *)new CFile();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CFile, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CFile__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  CFile *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CFile" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CFile" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (CFile *)new CFile((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CFile, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CFile(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CFile", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CFile__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CFile__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CFile'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::CFile()\n"
    "    CFile::CFile(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CFile(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CFile" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_SetFileName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CFile_SetFileName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_SetFileName" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_SetFileName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_SetFileName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetFileName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsReg__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  bool arg2 ;
  int res1 = SWIG_OLDOBJ ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsReg" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsReg" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsReg" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)CFile::IsReg((CString const &)*arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsReg__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsReg" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsReg" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CFile::IsReg((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsDir__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  bool arg2 ;
  int res1 = SWIG_OLDOBJ ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsDir" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)CFile::IsDir((CString const &)*arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsDir__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CFile::IsDir((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsChr__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  bool arg2 ;
  int res1 = SWIG_OLDOBJ ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsChr" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsChr" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsChr" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)CFile::IsChr((CString const &)*arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsChr__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsChr" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsChr" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CFile::IsChr((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsBlk__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  bool arg2 ;
  int res1 = SWIG_OLDOBJ ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsBlk" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsBlk" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsBlk" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)CFile::IsBlk((CString const &)*arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsBlk__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsBlk" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsBlk" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CFile::IsBlk((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsFifo__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  bool arg2 ;
  int res1 = SWIG_OLDOBJ ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsFifo" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsFifo" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsFifo" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)CFile::IsFifo((CString const &)*arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsFifo__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsFifo" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsFifo" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CFile::IsFifo((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsLnk__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  bool arg2 ;
  int res1 = SWIG_OLDOBJ ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsLnk" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsLnk" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsLnk" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)CFile::IsLnk((CString const &)*arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsLnk__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsLnk" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsLnk" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CFile::IsLnk((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsSock__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  bool arg2 ;
  int res1 = SWIG_OLDOBJ ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsSock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsSock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsSock" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)CFile::IsSock((CString const &)*arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsSock__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsSock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_IsSock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CFile::IsSock((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsReg__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsReg" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsReg" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)((CFile const *)arg1)->IsReg(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsReg__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsReg" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)((CFile const *)arg1)->IsReg();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsReg(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_IsReg", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsReg__SWIG_3(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsReg__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsReg__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsReg__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_IsReg'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::IsReg(CString const &,bool)\n"
    "    CFile::IsReg(CString const &)\n"
    "    CFile::IsReg(bool) const\n"
    "    CFile::IsReg() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_IsDir__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsDir" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsDir" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)((CFile const *)arg1)->IsDir(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsDir__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsDir" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)((CFile const *)arg1)->IsDir();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsDir(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_IsDir", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsDir__SWIG_3(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsDir__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsDir__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsDir__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_IsDir'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::IsDir(CString const &,bool)\n"
    "    CFile::IsDir(CString const &)\n"
    "    CFile::IsDir(bool) const\n"
    "    CFile::IsDir() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_IsChr__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsChr" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsChr" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)((CFile const *)arg1)->IsChr(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsChr__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsChr" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)((CFile const *)arg1)->IsChr();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsChr(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_IsChr", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsChr__SWIG_3(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsChr__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsChr__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsChr__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_IsChr'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::IsChr(CString const &,bool)\n"
    "    CFile::IsChr(CString const &)\n"
    "    CFile::IsChr(bool) const\n"
    "    CFile::IsChr() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_IsBlk__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsBlk" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsBlk" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)((CFile const *)arg1)->IsBlk(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsBlk__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsBlk" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)((CFile const *)arg1)->IsBlk();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsBlk(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_IsBlk", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsBlk__SWIG_3(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsBlk__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsBlk__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsBlk__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_IsBlk'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::IsBlk(CString const &,bool)\n"
    "    CFile::IsBlk(CString const &)\n"
    "    CFile::IsBlk(bool) const\n"
    "    CFile::IsBlk() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_IsFifo__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsFifo" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsFifo" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)((CFile const *)arg1)->IsFifo(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsFifo__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsFifo" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)((CFile const *)arg1)->IsFifo();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsFifo(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_IsFifo", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsFifo__SWIG_3(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsFifo__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsFifo__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsFifo__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_IsFifo'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::IsFifo(CString const &,bool)\n"
    "    CFile::IsFifo(CString const &)\n"
    "    CFile::IsFifo(bool) const\n"
    "    CFile::IsFifo() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_IsLnk__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsLnk" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsLnk" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)((CFile const *)arg1)->IsLnk(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsLnk__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsLnk" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)((CFile const *)arg1)->IsLnk();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsLnk(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_IsLnk", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsLnk__SWIG_3(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsLnk__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsLnk__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsLnk__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_IsLnk'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::IsLnk(CString const &,bool)\n"
    "    CFile::IsLnk(CString const &)\n"
    "    CFile::IsLnk(bool) const\n"
    "    CFile::IsLnk() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_IsSock__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsSock" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_IsSock" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)((CFile const *)arg1)->IsSock(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsSock__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsSock" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)((CFile const *)arg1)->IsSock();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsSock(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_IsSock", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsSock__SWIG_3(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_IsSock__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsSock__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_IsSock__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_IsSock'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::IsSock(CString const &,bool)\n"
    "    CFile::IsSock(CString const &)\n"
    "    CFile::IsSock(bool) const\n"
    "    CFile::IsSock() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_FType__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CFile::EFileTypes arg2 ;
  bool arg3 ;
  int res1 = SWIG_OLDOBJ ;
  int val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_FType" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_FType" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_FType" "', argument " "2"" of type '" "CFile::EFileTypes""'");
  } 
  arg2 = static_cast< CFile::EFileTypes >(val2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CFile_FType" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)CFile::FType((CString const &)*arg1,arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_FType__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CFile::EFileTypes arg2 ;
  int res1 = SWIG_OLDOBJ ;
  int val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_FType" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_FType" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_FType" "', argument " "2"" of type '" "CFile::EFileTypes""'");
  } 
  arg2 = static_cast< CFile::EFileTypes >(val2);
  result = (bool)CFile::FType((CString const &)*arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_FType(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_FType", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_FType__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CFile_FType__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_FType'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::FType(CString const &,CFile::EFileTypes,bool)\n"
    "    CFile::FType(CString const &,CFile::EFileTypes)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_Exists__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Exists" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)((CFile const *)arg1)->Exists();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetSize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  off_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetSize" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = ((CFile const *)arg1)->GetSize();
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetATime__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  time_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetATime" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = ((CFile const *)arg1)->GetATime();
  resultobj = SWIG_NewPointerObj((new time_t(result)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetMTime__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  time_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetMTime" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = ((CFile const *)arg1)->GetMTime();
  resultobj = SWIG_NewPointerObj((new time_t(result)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetCTime__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  time_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetCTime" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = ((CFile const *)arg1)->GetCTime();
  resultobj = SWIG_NewPointerObj((new time_t(result)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetUID__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  uid_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetUID" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = ((CFile const *)arg1)->GetUID();
  resultobj = SWIG_NewPointerObj((new uid_t(result)), SWIGTYPE_p_uid_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetGID__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  gid_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetGID" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = ((CFile const *)arg1)->GetGID();
  resultobj = SWIG_NewPointerObj((new gid_t(result)), SWIGTYPE_p_gid_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Exists__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Exists" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Exists" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CFile::Exists((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Exists(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_Exists", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_Exists__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_Exists__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_Exists'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::Exists() const\n"
    "    CFile::Exists(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_GetSize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  off_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetSize" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_GetSize" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = CFile::GetSize((CString const &)*arg1);
  resultobj = SWIG_From_long(static_cast< long >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetSize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_GetSize", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_GetSize__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_GetSize__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_GetSize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::GetSize() const\n"
    "    CFile::GetSize(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_GetATime__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  time_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetATime" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_GetATime" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = CFile::GetATime((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj((new time_t(result)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetATime(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_GetATime", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_GetATime__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_GetATime__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_GetATime'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::GetATime() const\n"
    "    CFile::GetATime(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_GetMTime__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  time_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetMTime" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_GetMTime" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = CFile::GetMTime((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj((new time_t(result)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetMTime(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_GetMTime", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_GetMTime__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_GetMTime__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_GetMTime'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::GetMTime() const\n"
    "    CFile::GetMTime(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_GetCTime__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  time_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetCTime" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_GetCTime" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = CFile::GetCTime((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj((new time_t(result)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetCTime(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_GetCTime", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_GetCTime__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_GetCTime__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_GetCTime'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::GetCTime() const\n"
    "    CFile::GetCTime(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_GetUID__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  uid_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetUID" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_GetUID" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = CFile::GetUID((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj((new uid_t(result)), SWIGTYPE_p_uid_t, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetUID(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_GetUID", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_GetUID__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_GetUID__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_GetUID'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::GetUID() const\n"
    "    CFile::GetUID(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_GetGID__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  gid_t result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetGID" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_GetGID" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = CFile::GetGID((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj((new gid_t(result)), SWIGTYPE_p_gid_t, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetGID(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_GetGID", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_GetGID__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_GetGID__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_GetGID'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::GetGID() const\n"
    "    CFile::GetGID(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_GetInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  stat *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CFile_GetInfo", 2, 2, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetInfo" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_GetInfo" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_stat,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_GetInfo" "', argument " "2"" of type '" "stat &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_GetInfo" "', argument " "2"" of type '" "stat &""'"); 
  }
  arg2 = reinterpret_cast< stat * >(argp2);
  result = (int)CFile::GetInfo((CString const &)*arg1,*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Delete__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Delete" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)(arg1)->Delete();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Move__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Move" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Move" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Move" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CFile_Move" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)(arg1)->Move((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Move__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Move" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Move" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Move" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->Move((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Copy__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Copy" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Copy" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Copy" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CFile_Copy" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)(arg1)->Copy((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Copy__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Copy" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Copy" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Copy" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->Copy((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Delete__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Delete" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Delete" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CFile::Delete((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Delete(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_Delete", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_Delete__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_Delete__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_Delete'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::Delete()\n"
    "    CFile::Delete(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_Move__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Move" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Move" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Move" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Move" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CFile_Move" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)CFile::Move((CString const &)*arg1,(CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Move__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Move" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Move" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Move" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Move" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)CFile::Move((CString const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Move(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_Move", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CFile_Move__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CFile_Move__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CFile_Move__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CFile_Move__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_Move'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::Move(CString const &,bool)\n"
    "    CFile::Move(CString const &)\n"
    "    CFile::Move(CString const &,CString const &,bool)\n"
    "    CFile::Move(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_Copy__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Copy" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Copy" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Copy" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Copy" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CFile_Copy" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)CFile::Copy((CString const &)*arg1,(CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Copy__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Copy" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Copy" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Copy" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Copy" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)CFile::Copy((CString const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Copy(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_Copy", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CFile_Copy__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CFile_Copy__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CFile_Copy__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CFile_Copy__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_Copy'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::Copy(CString const &,bool)\n"
    "    CFile::Copy(CString const &)\n"
    "    CFile::Copy(CString const &,CString const &,bool)\n"
    "    CFile::Copy(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_Chmod__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  mode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Chmod" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_mode_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Chmod" "', argument " "2"" of type '" "mode_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Chmod" "', argument " "2"" of type '" "mode_t""'");
    } else {
      mode_t * temp = reinterpret_cast< mode_t * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  result = (bool)(arg1)->Chmod(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Chmod__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  mode_t arg2 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Chmod" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Chmod" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_mode_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Chmod" "', argument " "2"" of type '" "mode_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Chmod" "', argument " "2"" of type '" "mode_t""'");
    } else {
      mode_t * temp = reinterpret_cast< mode_t * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  result = (bool)CFile::Chmod((CString const &)*arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Chmod(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_Chmod", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_mode_t, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CFile_Chmod__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_mode_t, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CFile_Chmod__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_Chmod'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::Chmod(mode_t)\n"
    "    CFile::Chmod(CString const &,mode_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_Seek(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  off_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CFile_Seek", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Seek" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_Seek" "', argument " "2"" of type '" "off_t""'");
  } 
  arg2 = static_cast< off_t >(val2);
  result = (bool)(arg1)->Seek(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Truncate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Truncate" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)(arg1)->Truncate();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Sync(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Sync" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)(arg1)->Sync();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Open__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  int arg3 ;
  mode_t arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int val3 ;
  int ecode3 = 0 ;
  void *argp4 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Open" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Open" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Open" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CFile_Open" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_mode_t,  0  | 0);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CFile_Open" "', argument " "4"" of type '" "mode_t""'"); 
    }  
    if (!argp4) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Open" "', argument " "4"" of type '" "mode_t""'");
    } else {
      mode_t * temp = reinterpret_cast< mode_t * >(argp4);
      arg4 = *temp;
      if (SWIG_IsNewObj(res4)) delete temp;
    }
  }
  result = (bool)(arg1)->Open((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Open__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Open" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Open" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Open" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CFile_Open" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  result = (bool)(arg1)->Open((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Open__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Open" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Open" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Open" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->Open((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Open__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  int arg2 ;
  mode_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  void *argp3 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Open" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_Open" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_mode_t,  0  | 0);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CFile_Open" "', argument " "3"" of type '" "mode_t""'"); 
    }  
    if (!argp3) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Open" "', argument " "3"" of type '" "mode_t""'");
    } else {
      mode_t * temp = reinterpret_cast< mode_t * >(argp3);
      arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp;
    }
  }
  result = (bool)(arg1)->Open(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Open__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Open" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CFile_Open" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  result = (bool)(arg1)->Open(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Open__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Open" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)(arg1)->Open();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Open(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_Open", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_Open__SWIG_5(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CFile_Open__SWIG_4(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CFile_Open__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_mode_t, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CFile_Open__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CFile_Open__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_ConvertPtr(argv[3], 0, SWIGTYPE_p_mode_t, SWIG_POINTER_NO_NULL | 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CFile_Open__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_Open'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::Open(CString const &,int,mode_t)\n"
    "    CFile::Open(CString const &,int)\n"
    "    CFile::Open(CString const &)\n"
    "    CFile::Open(int,mode_t)\n"
    "    CFile::Open(int)\n"
    "    CFile::Open()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_Read(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  char *arg2 = (char *) 0 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  ssize_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CFile_Read", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Read" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Read" "', argument " "2"" of type '" "char *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CFile_Read" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  result = (arg1)->Read(arg2,arg3);
  resultobj = SWIG_NewPointerObj((new ssize_t(result)), SWIGTYPE_p_ssize_t, SWIG_POINTER_OWN |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_ReadLine__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_ReadLine" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sData");
    }
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CFile_ReadLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_ReadLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->ReadLine(*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_ReadLine__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_ReadLine" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sData");
    }
  }
  result = (bool)(arg1)->ReadLine(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_ReadLine(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_ReadLine", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        String* p;
        _v = SWIG_IsOK(SWIG_ConvertPtr(argv[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
      }
      if (_v) {
        return _wrap_CFile_ReadLine__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        String* p;
        _v = SWIG_IsOK(SWIG_ConvertPtr(argv[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CFile_ReadLine__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_ReadLine'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::ReadLine(CString &,CString const &)\n"
    "    CFile::ReadLine(CString &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_ReadFile__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  size_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_ReadFile" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sData");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CFile_ReadFile" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = static_cast< size_t >(val3);
  result = (bool)(arg1)->ReadFile(*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_ReadFile__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_ReadFile" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sData");
    }
  }
  result = (bool)(arg1)->ReadFile(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_ReadFile(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_ReadFile", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        String* p;
        _v = SWIG_IsOK(SWIG_ConvertPtr(argv[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
      }
      if (_v) {
        return _wrap_CFile_ReadFile__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        String* p;
        _v = SWIG_IsOK(SWIG_ConvertPtr(argv[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
      }
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CFile_ReadFile__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_ReadFile'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::ReadFile(CString &,size_t)\n"
    "    CFile::ReadFile(CString &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_Write__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  char *arg2 = (char *) 0 ;
  size_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  ssize_t result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Write" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Write" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CFile_Write" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = static_cast< size_t >(val3);
  result = (arg1)->Write((char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj((new ssize_t(result)), SWIGTYPE_p_ssize_t, SWIG_POINTER_OWN |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Write__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  ssize_t result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Write" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_Write" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_Write" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->Write((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj((new ssize_t(result)), SWIGTYPE_p_ssize_t, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_Write(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_Write", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CFile_Write__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CFile_Write__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_Write'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::Write(char const *,size_t)\n"
    "    CFile::Write(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_Close(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_Close" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  (arg1)->Close();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_ClearBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_ClearBuffer" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  (arg1)->ClearBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_TryExLock__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_TryExLock" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_TryExLock" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_TryExLock" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CFile_TryExLock" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  result = (bool)(arg1)->TryExLock((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_TryExLock__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_TryExLock" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CFile_TryExLock" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_TryExLock" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->TryExLock((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_TryExLock__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_TryExLock" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)(arg1)->TryExLock();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_TryExLock(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CFile_TryExLock", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CFile_TryExLock__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CFile_TryExLock__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CFile, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CFile_TryExLock__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CFile_TryExLock'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CFile::TryExLock(CString const &,int)\n"
    "    CFile::TryExLock(CString const &)\n"
    "    CFile::TryExLock()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CFile_ExLock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_ExLock" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)(arg1)->ExLock();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_UnLock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_UnLock" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)(arg1)->UnLock();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_IsOpen(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_IsOpen" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)((CFile const *)arg1)->IsOpen();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetLongName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetLongName" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = ((CFile const *)arg1)->GetLongName();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetShortName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetShortName" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = ((CFile const *)arg1)->GetShortName();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetDir(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_GetDir" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = ((CFile const *)arg1)->GetDir();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_HadError(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_HadError" "', argument " "1"" of type '" "CFile const *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  result = (bool)((CFile const *)arg1)->HadError();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_ResetError(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFile *arg1 = (CFile *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFile, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_ResetError" "', argument " "1"" of type '" "CFile *""'"); 
  }
  arg1 = reinterpret_cast< CFile * >(argp1);
  (arg1)->ResetError();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_InitHomePath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFile_InitHomePath" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CFile_InitHomePath" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  CFile::InitHomePath((CString const &)*arg1);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFile_GetHomePath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CFile_GetHomePath", 0, 0, 0)) SWIG_fail;
  result = (CString *) &CFile::GetHomePath();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CFile_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CFile, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CFile_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CDir__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  CDir *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (CDir *)new CDir((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CDir, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CDir__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CDir *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CDir *)new CDir();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CDir, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CDir(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CDir", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CDir__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CDir__SWIG_0(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CDir'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CDir::CDir(CString const &)\n"
    "    CDir::CDir()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CDir(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CDir *arg1 = (CDir *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CDir, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CDir" "', argument " "1"" of type '" "CDir *""'"); 
  }
  arg1 = reinterpret_cast< CDir * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_CleanUp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CDir *arg1 = (CDir *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CDir, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_CleanUp" "', argument " "1"" of type '" "CDir *""'"); 
  }
  arg1 = reinterpret_cast< CDir * >(argp1);
  (arg1)->CleanUp();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_Fill(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CDir *arg1 = (CDir *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  size_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CDir_Fill", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CDir, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_Fill" "', argument " "1"" of type '" "CDir *""'"); 
  }
  arg1 = reinterpret_cast< CDir * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CDir_Fill" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_Fill" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->Fill((CString const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_FillByWildcard(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CDir *arg1 = (CDir *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  size_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CDir_FillByWildcard", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CDir, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_FillByWildcard" "', argument " "1"" of type '" "CDir *""'"); 
  }
  arg1 = reinterpret_cast< CDir * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CDir_FillByWildcard" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_FillByWildcard" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CDir_FillByWildcard" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_FillByWildcard" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (arg1)->FillByWildcard((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_Chmod__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  mode_t arg1 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  unsigned int result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_mode_t,  0  | 0);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_Chmod" "', argument " "1"" of type '" "mode_t""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_Chmod" "', argument " "1"" of type '" "mode_t""'");
    } else {
      mode_t * temp = reinterpret_cast< mode_t * >(argp1);
      arg1 = *temp;
      if (SWIG_IsNewObj(res1)) delete temp;
    }
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CDir_Chmod" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_Chmod" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CDir_Chmod" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_Chmod" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (unsigned int)CDir::Chmod(SWIG_STD_MOVE(arg1),(CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_Chmod__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  mode_t arg1 ;
  CString *arg2 = 0 ;
  void *argp1 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned int result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_mode_t,  0  | 0);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_Chmod" "', argument " "1"" of type '" "mode_t""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_Chmod" "', argument " "1"" of type '" "mode_t""'");
    } else {
      mode_t * temp = reinterpret_cast< mode_t * >(argp1);
      arg1 = *temp;
      if (SWIG_IsNewObj(res1)) delete temp;
    }
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CDir_Chmod" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_Chmod" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (unsigned int)CDir::Chmod(SWIG_STD_MOVE(arg1),(CString const &)*arg2);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_Chmod__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CDir *arg1 = (CDir *) 0 ;
  mode_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  unsigned int result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CDir, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_Chmod" "', argument " "1"" of type '" "CDir *""'"); 
  }
  arg1 = reinterpret_cast< CDir * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_mode_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CDir_Chmod" "', argument " "2"" of type '" "mode_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_Chmod" "', argument " "2"" of type '" "mode_t""'");
    } else {
      mode_t * temp = reinterpret_cast< mode_t * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  result = (unsigned int)(arg1)->Chmod(arg2);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_Chmod(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CDir_Chmod", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_mode_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CDir_Chmod__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CDir, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_mode_t, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CDir_Chmod__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_mode_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CDir_Chmod__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CDir_Chmod'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CDir::Chmod(mode_t,CString const &,CString const &)\n"
    "    CDir::Chmod(mode_t,CString const &)\n"
    "    CDir::Chmod(mode_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CDir_Delete__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  unsigned int result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_Delete" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_Delete" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CDir_Delete" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_Delete" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (unsigned int)CDir::Delete((CString const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_Delete__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  unsigned int result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_Delete" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_Delete" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (unsigned int)CDir::Delete((CString const &)*arg1);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_Delete__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CDir *arg1 = (CDir *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CDir, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_Delete" "', argument " "1"" of type '" "CDir *""'"); 
  }
  arg1 = reinterpret_cast< CDir * >(argp1);
  result = (unsigned int)(arg1)->Delete();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_Delete(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CDir_Delete", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CDir, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CDir_Delete__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CDir_Delete__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CDir_Delete__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CDir_Delete'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CDir::Delete(CString const &,CString const &)\n"
    "    CDir::Delete(CString const &)\n"
    "    CDir::Delete()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CDir_GetSortAttr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CDir *arg1 = (CDir *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CFile::EFileAttr result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CDir, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_GetSortAttr" "', argument " "1"" of type '" "CDir const *""'"); 
  }
  arg1 = reinterpret_cast< CDir * >(argp1);
  result = (CFile::EFileAttr)((CDir const *)arg1)->GetSortAttr();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_IsDescending(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CDir *arg1 = (CDir *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CDir, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_IsDescending" "', argument " "1"" of type '" "CDir const *""'"); 
  }
  arg1 = reinterpret_cast< CDir * >(argp1);
  result = (bool)((CDir const *)arg1)->IsDescending();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_CheckPathPrefix__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_CheckPathPrefix" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_CheckPathPrefix" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CDir_CheckPathPrefix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_CheckPathPrefix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CDir_CheckPathPrefix" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_CheckPathPrefix" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = CDir::CheckPathPrefix((CString const &)*arg1,(CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_CheckPathPrefix__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_CheckPathPrefix" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_CheckPathPrefix" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CDir_CheckPathPrefix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_CheckPathPrefix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = CDir::CheckPathPrefix((CString const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_CheckPathPrefix(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CDir_CheckPathPrefix", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CDir_CheckPathPrefix__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CDir_CheckPathPrefix__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CDir_CheckPathPrefix'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CDir::CheckPathPrefix(CString const &,CString const &,CString const &)\n"
    "    CDir::CheckPathPrefix(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CDir_ChangeDir__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_ChangeDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_ChangeDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CDir_ChangeDir" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_ChangeDir" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CDir_ChangeDir" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_ChangeDir" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = CDir::ChangeDir((CString const &)*arg1,(CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_ChangeDir__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_ChangeDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_ChangeDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CDir_ChangeDir" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_ChangeDir" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = CDir::ChangeDir((CString const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_ChangeDir(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CDir_ChangeDir", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CDir_ChangeDir__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CDir_ChangeDir__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CDir_ChangeDir'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CDir::ChangeDir(CString const &,CString const &,CString const &)\n"
    "    CDir::ChangeDir(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CDir_MakeDir__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  mode_t arg2 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_MakeDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_MakeDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_mode_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CDir_MakeDir" "', argument " "2"" of type '" "mode_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_MakeDir" "', argument " "2"" of type '" "mode_t""'");
    } else {
      mode_t * temp = reinterpret_cast< mode_t * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  result = (bool)CDir::MakeDir((CString const &)*arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_MakeDir__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDir_MakeDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDir_MakeDir" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CDir::MakeDir((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDir_MakeDir(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CDir_MakeDir", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CDir_MakeDir__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_mode_t, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CDir_MakeDir__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CDir_MakeDir'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CDir::MakeDir(CString const &,mode_t)\n"
    "    CDir::MakeDir(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CDir_GetCWD(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CDir_GetCWD", 0, 0, 0)) SWIG_fail;
  result = CDir::GetCWD();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CDir_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CDir, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CDir_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CMessage__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  CMessage *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CMessage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CMessage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (CMessage *)new CMessage((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CMessage, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CMessage__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CMessage *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CMessage *)new CMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CMessage__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CNick *arg1 = 0 ;
  CString *arg2 = 0 ;
  VCString *arg3 = 0 ;
  MCString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  CMessage *result = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CMessage" "', argument " "1"" of type '" "CNick const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CMessage" "', argument " "1"" of type '" "CNick const &""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CMessage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CMessage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CMessage" "', argument " "3"" of type '" "VCString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CMessage" "', argument " "3"" of type '" "VCString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_MCString,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CMessage" "', argument " "4"" of type '" "MCString const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CMessage" "', argument " "4"" of type '" "MCString const &""'"); 
  }
  arg4 = reinterpret_cast< MCString * >(argp4);
  result = (CMessage *)new CMessage((CNick const &)*arg1,(CString const &)*arg2,(VCString const &)*arg3,(MCString const &)*arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CMessage, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CMessage__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CNick *arg1 = 0 ;
  CString *arg2 = 0 ;
  VCString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  CMessage *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CMessage" "', argument " "1"" of type '" "CNick const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CMessage" "', argument " "1"" of type '" "CNick const &""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CMessage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CMessage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CMessage" "', argument " "3"" of type '" "VCString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CMessage" "', argument " "3"" of type '" "VCString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (CMessage *)new CMessage((CNick const &)*arg1,(CString const &)*arg2,(VCString const &)*arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CMessage, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CMessage__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CNick *arg1 = 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CMessage *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CMessage" "', argument " "1"" of type '" "CNick const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CMessage" "', argument " "1"" of type '" "CNick const &""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CMessage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CMessage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CMessage *)new CMessage((CNick const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CMessage, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CMessage(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CMessage", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CMessage__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CMessage__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_CNick, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_CMessage__SWIG_4(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_CNick, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CString,std::allocator< CString > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_new_CMessage__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_CNick, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CString,std::allocator< CString > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_ConvertPtr(argv[3], 0, SWIGTYPE_p_MCString, SWIG_POINTER_NO_NULL | 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_new_CMessage__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CMessage'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CMessage::CMessage(CString const &)\n"
    "    CMessage::CMessage()\n"
    "    CMessage::CMessage(CNick const &,CString const &,VCString const &,MCString const &)\n"
    "    CMessage::CMessage(CNick const &,CString const &,VCString const &)\n"
    "    CMessage::CMessage(CNick const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CMessage_GetType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CMessage::Type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetType" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CMessage::Type)((CMessage const *)arg1)->GetType();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_Equals(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_Equals", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_Equals" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_Equals" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CMessage_Equals" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (bool)((CMessage const *)arg1)->Equals((CMessage const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_Clone(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_Clone", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_Clone" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_Clone" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CMessage_Clone" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  (arg1)->Clone((CMessage const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CIRCNetwork *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetNetwork" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CIRCNetwork *)((CMessage const *)arg1)->GetNetwork();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_SetNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_SetNetwork", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_SetNetwork" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_SetNetwork" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  (arg1)->SetNetwork(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetClient(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CClient *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetClient" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CClient *)((CMessage const *)arg1)->GetClient();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CClient, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_SetClient(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_SetClient", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_SetClient" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_SetClient" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  (arg1)->SetClient(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetChan(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CChan *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetChan" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CChan *)((CMessage const *)arg1)->GetChan();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CChan, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_SetChan(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  CChan *arg2 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_SetChan", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_SetChan" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_SetChan" "', argument " "2"" of type '" "CChan *""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  (arg1)->SetChan(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetNick__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CNick *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetNick" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CNick *) &(arg1)->GetNick();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNick, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetNick__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CNick *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetNick" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CNick *) &((CMessage const *)arg1)->GetNick();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNick, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetNick(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CMessage_GetNick", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CMessage_GetNick__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CMessage_GetNick__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CMessage_GetNick'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CMessage::GetNick()\n"
    "    CMessage::GetNick() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CMessage_SetNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  CNick *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_SetNick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_SetNick" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_SetNick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CMessage_SetNick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  (arg1)->SetNick((CNick const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetCommand" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CString *) &((CMessage const *)arg1)->GetCommand();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_SetCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_SetCommand", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_SetCommand" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_SetCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CMessage_SetCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetCommand((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetParams__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  VCString *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetParams" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (VCString *) &((CMessage const *)arg1)->GetParams();
  resultobj = swig::from(static_cast< std::vector< CString,std::allocator< CString > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetParamsSplit__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  VCString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetParamsSplit" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CMessage_GetParamsSplit" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CMessage_GetParamsSplit" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  result = ((CMessage const *)arg1)->GetParamsSplit(arg2,arg3);
  resultobj = swig::from(static_cast< std::vector< CString,std::allocator< CString > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetParamsSplit__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  VCString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetParamsSplit" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CMessage_GetParamsSplit" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = ((CMessage const *)arg1)->GetParamsSplit(arg2);
  resultobj = swig::from(static_cast< std::vector< CString,std::allocator< CString > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetParamsSplit(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CMessage_GetParamsSplit", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CMessage_GetParamsSplit__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CMessage_GetParamsSplit__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CMessage_GetParamsSplit'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CMessage::GetParamsSplit(unsigned int,unsigned int) const\n"
    "    CMessage::GetParamsSplit(unsigned int) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CMessage_SetParams__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  VCString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_SetParams" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_SetParams" "', argument " "2"" of type '" "VCString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CMessage_SetParams" "', argument " "2"" of type '" "VCString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetParams((VCString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_SetParams__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  VCString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::unique_ptr< VCString > rvrdeleter2 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_SetParams" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_RELEASE |  0 );
  if (!SWIG_IsOK(res2)) {
    if (res2 == SWIG_ERROR_RELEASE_NOT_OWNED) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_SetParams" "', cannot release ownership as memory is not owned for argument " "2"" of type '" "VCString &&""'");
    } else {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_SetParams" "', argument " "2"" of type '" "VCString &&""'"); 
    }
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CMessage_SetParams" "', argument " "2"" of type '" "VCString &&""'"); 
  }
  arg2 = reinterpret_cast< VCString * >(argp2);
  rvrdeleter2.reset(arg2);
  (arg1)->SetParams((VCString &&)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_SetParams(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CMessage_SetParams", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CMessage_SetParams__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = swig::asptr(argv[1], (std::vector< CString,std::allocator< CString > >**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CMessage_SetParams__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CMessage_SetParams'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CMessage::SetParams(VCString const &)\n"
    "    CMessage::SetParams(VCString &&)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CMessage_GetParams__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetParams" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CMessage_GetParams" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CMessage_GetParams" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  result = ((CMessage const *)arg1)->GetParams(arg2,arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetParams__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetParams" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CMessage_GetParams" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = ((CMessage const *)arg1)->GetParams(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetParams(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CMessage_GetParams", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CMessage_GetParams__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CMessage_GetParams__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CMessage_GetParams__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CMessage_GetParams'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CMessage::GetParams() const\n"
    "    CMessage::GetParams(unsigned int,unsigned int) const\n"
    "    CMessage::GetParams(unsigned int) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CMessage_GetParamsColon__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetParamsColon" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CMessage_GetParamsColon" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CMessage_GetParamsColon" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  result = ((CMessage const *)arg1)->GetParamsColon(arg2,arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetParamsColon__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetParamsColon" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CMessage_GetParamsColon" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = ((CMessage const *)arg1)->GetParamsColon(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetParamsColon(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CMessage_GetParamsColon", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CMessage_GetParamsColon__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CMessage_GetParamsColon__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CMessage_GetParamsColon'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CMessage::GetParamsColon(unsigned int,unsigned int) const\n"
    "    CMessage::GetParamsColon(unsigned int) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CMessage_GetParam(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_GetParam", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetParam" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CMessage_GetParam" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = ((CMessage const *)arg1)->GetParam(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_SetParam(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  unsigned int arg2 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_SetParam", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_SetParam" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CMessage_SetParam" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CMessage_SetParam" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CMessage_SetParam" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->SetParam(arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  timeval *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetTime" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (timeval *) &((CMessage const *)arg1)->GetTime();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_timeval, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_SetTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  timeval *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_SetTime", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_SetTime" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_timeval,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_SetTime" "', argument " "2"" of type '" "timeval const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CMessage_SetTime" "', argument " "2"" of type '" "timeval const &""'"); 
  }
  arg2 = reinterpret_cast< timeval * >(argp2);
  (arg1)->SetTime((timeval const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetTags__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  MCString *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetTags" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (MCString *) &((CMessage const *)arg1)->GetTags();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_MCString, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetTags__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  MCString *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetTags" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (MCString *) &(arg1)->GetTags();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_MCString, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetTags(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CMessage_GetTags", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CMessage_GetTags__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CMessage_GetTags__SWIG_0(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CMessage_GetTags'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CMessage::GetTags() const\n"
    "    CMessage::GetTags()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CMessage_SetTags(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  MCString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_SetTags", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_SetTags" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_MCString,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_SetTags" "', argument " "2"" of type '" "MCString const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CMessage_SetTags" "', argument " "2"" of type '" "MCString const &""'"); 
  }
  arg2 = reinterpret_cast< MCString * >(argp2);
  (arg1)->SetTags((MCString const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_GetTag(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_GetTag", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_GetTag" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_GetTag" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CMessage_GetTag" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CMessage const *)arg1)->GetTag((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_SetTag(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_SetTag", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_SetTag" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_SetTag" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CMessage_SetTag" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CMessage_SetTag" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CMessage_SetTag" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->SetTag((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_ToString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_ToString" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CMessage_ToString" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = ((CMessage const *)arg1)->ToString(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_ToString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_ToString" "', argument " "1"" of type '" "CMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = ((CMessage const *)arg1)->ToString();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_ToString(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CMessage_ToString", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CMessage_ToString__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CMessage, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CMessage_ToString__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CMessage_ToString'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CMessage::ToString(unsigned int) const\n"
    "    CMessage::ToString() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CMessage_Parse(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CMessage_Parse", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_Parse" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CMessage_Parse" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CMessage_Parse" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->Parse((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CTargetMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CTargetMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CTargetMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CTargetMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CTargetMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTargetMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CActionMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CActionMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CActionMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CActionMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CActionMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CActionMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CAuthenticateMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CAuthenticateMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CAuthenticateMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CAuthenticateMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CAuthenticateMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CAuthenticateMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CCTCPMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CCTCPMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CCTCPMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CCTCPMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CCTCPMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CCTCPMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CJoinMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CJoinMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CJoinMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CJoinMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CJoinMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CJoinMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CModeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CModeMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CModeMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CModeMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CModeMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModeMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CNickMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CNickMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CNickMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CNickMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CNickMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNickMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CNoticeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CNoticeMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CNoticeMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CNoticeMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CNoticeMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNoticeMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CNumericMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CNumericMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CNumericMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CNumericMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CNumericMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNumericMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CKickMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CKickMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CKickMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CKickMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CKickMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CKickMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CInviteMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CInviteMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CInviteMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CInviteMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CInviteMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CInviteMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CPartMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CPartMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CPartMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CPartMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CPartMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CPartMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CQuitMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CQuitMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CQuitMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CQuitMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CQuitMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CQuitMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CTextMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CTextMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CTextMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CTextMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CTextMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTextMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CTopicMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CTopicMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CTopicMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CTopicMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CTopicMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTopicMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage_As_CChgHostMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CChgHostMessage *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage_As_CChgHostMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CChgHostMessage *) &(arg1)->SWIGTEMPLATEDISAMBIGUATOR As< CChgHostMessage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CChgHostMessage, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage___str__" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = CMessage___str__(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CMessage___repr__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CMessage___repr__" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = CMessage___repr__(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CMessage *arg1 = (CMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CMessage" "', argument " "1"" of type '" "CMessage *""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CTargetMessage_GetTarget(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTargetMessage *arg1 = (CTargetMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTargetMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTargetMessage_GetTarget" "', argument " "1"" of type '" "CTargetMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CTargetMessage * >(argp1);
  result = ((CTargetMessage const *)arg1)->GetTarget();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTargetMessage_SetTarget(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTargetMessage *arg1 = (CTargetMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTargetMessage_SetTarget", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTargetMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTargetMessage_SetTarget" "', argument " "1"" of type '" "CTargetMessage *""'"); 
  }
  arg1 = reinterpret_cast< CTargetMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTargetMessage_SetTarget" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTargetMessage_SetTarget" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetTarget((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CTargetMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTargetMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CTargetMessage", 0, 0, 0)) SWIG_fail;
  result = (CTargetMessage *)new CTargetMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTargetMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CTargetMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTargetMessage *arg1 = (CTargetMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTargetMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CTargetMessage" "', argument " "1"" of type '" "CTargetMessage *""'"); 
  }
  arg1 = reinterpret_cast< CTargetMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CTargetMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CTargetMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CTargetMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CActionMessage_GetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CActionMessage *arg1 = (CActionMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CActionMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CActionMessage_GetText" "', argument " "1"" of type '" "CActionMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CActionMessage * >(argp1);
  result = ((CActionMessage const *)arg1)->GetText();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CActionMessage_SetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CActionMessage *arg1 = (CActionMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CActionMessage_SetText", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CActionMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CActionMessage_SetText" "', argument " "1"" of type '" "CActionMessage *""'"); 
  }
  arg1 = reinterpret_cast< CActionMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CActionMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CActionMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetText((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CActionMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CActionMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CActionMessage", 0, 0, 0)) SWIG_fail;
  result = (CActionMessage *)new CActionMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CActionMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CActionMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CActionMessage *arg1 = (CActionMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CActionMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CActionMessage" "', argument " "1"" of type '" "CActionMessage *""'"); 
  }
  arg1 = reinterpret_cast< CActionMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CActionMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CActionMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CActionMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CAuthenticateMessage_GetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CAuthenticateMessage *arg1 = (CAuthenticateMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CAuthenticateMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CAuthenticateMessage_GetText" "', argument " "1"" of type '" "CAuthenticateMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CAuthenticateMessage * >(argp1);
  result = ((CAuthenticateMessage const *)arg1)->GetText();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CAuthenticateMessage_SetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CAuthenticateMessage *arg1 = (CAuthenticateMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CAuthenticateMessage_SetText", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CAuthenticateMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CAuthenticateMessage_SetText" "', argument " "1"" of type '" "CAuthenticateMessage *""'"); 
  }
  arg1 = reinterpret_cast< CAuthenticateMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CAuthenticateMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CAuthenticateMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetText((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CAuthenticateMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CAuthenticateMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CAuthenticateMessage", 0, 0, 0)) SWIG_fail;
  result = (CAuthenticateMessage *)new CAuthenticateMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CAuthenticateMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CAuthenticateMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CAuthenticateMessage *arg1 = (CAuthenticateMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CAuthenticateMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CAuthenticateMessage" "', argument " "1"" of type '" "CAuthenticateMessage *""'"); 
  }
  arg1 = reinterpret_cast< CAuthenticateMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CAuthenticateMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CAuthenticateMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CAuthenticateMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CCTCPMessage_IsReply(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCTCPMessage *arg1 = (CCTCPMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCTCPMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCTCPMessage_IsReply" "', argument " "1"" of type '" "CCTCPMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CCTCPMessage * >(argp1);
  result = (bool)((CCTCPMessage const *)arg1)->IsReply();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCTCPMessage_GetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCTCPMessage *arg1 = (CCTCPMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCTCPMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCTCPMessage_GetText" "', argument " "1"" of type '" "CCTCPMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CCTCPMessage * >(argp1);
  result = ((CCTCPMessage const *)arg1)->GetText();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCTCPMessage_SetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCTCPMessage *arg1 = (CCTCPMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CCTCPMessage_SetText", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCTCPMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCTCPMessage_SetText" "', argument " "1"" of type '" "CCTCPMessage *""'"); 
  }
  arg1 = reinterpret_cast< CCTCPMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CCTCPMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CCTCPMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetText((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CCTCPMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCTCPMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CCTCPMessage", 0, 0, 0)) SWIG_fail;
  result = (CCTCPMessage *)new CCTCPMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CCTCPMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CCTCPMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCTCPMessage *arg1 = (CCTCPMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCTCPMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CCTCPMessage" "', argument " "1"" of type '" "CCTCPMessage *""'"); 
  }
  arg1 = reinterpret_cast< CCTCPMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CCTCPMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CCTCPMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CCTCPMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CJoinMessage_GetKey(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CJoinMessage *arg1 = (CJoinMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CJoinMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CJoinMessage_GetKey" "', argument " "1"" of type '" "CJoinMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CJoinMessage * >(argp1);
  result = ((CJoinMessage const *)arg1)->GetKey();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CJoinMessage_SetKey(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CJoinMessage *arg1 = (CJoinMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CJoinMessage_SetKey", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CJoinMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CJoinMessage_SetKey" "', argument " "1"" of type '" "CJoinMessage *""'"); 
  }
  arg1 = reinterpret_cast< CJoinMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CJoinMessage_SetKey" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CJoinMessage_SetKey" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetKey((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CJoinMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CJoinMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CJoinMessage", 0, 0, 0)) SWIG_fail;
  result = (CJoinMessage *)new CJoinMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CJoinMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CJoinMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CJoinMessage *arg1 = (CJoinMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CJoinMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CJoinMessage" "', argument " "1"" of type '" "CJoinMessage *""'"); 
  }
  arg1 = reinterpret_cast< CJoinMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CJoinMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CJoinMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CJoinMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CModeMessage_GetModes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModeMessage *arg1 = (CModeMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModeMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModeMessage_GetModes" "', argument " "1"" of type '" "CModeMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CModeMessage * >(argp1);
  result = ((CModeMessage const *)arg1)->GetModes();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModeMessage_GetModeList(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModeMessage *arg1 = (CModeMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModeMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModeMessage_GetModeList" "', argument " "1"" of type '" "CModeMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CModeMessage * >(argp1);
  result = ((CModeMessage const *)arg1)->GetModeList();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModeMessage_GetModeParams(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModeMessage *arg1 = (CModeMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  VCString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModeMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModeMessage_GetModeParams" "', argument " "1"" of type '" "CModeMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CModeMessage * >(argp1);
  result = ((CModeMessage const *)arg1)->GetModeParams();
  resultobj = swig::from(static_cast< std::vector< CString,std::allocator< CString > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModeMessage_HasModes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModeMessage *arg1 = (CModeMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModeMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModeMessage_HasModes" "', argument " "1"" of type '" "CModeMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CModeMessage * >(argp1);
  result = (bool)((CModeMessage const *)arg1)->HasModes();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CModeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModeMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CModeMessage", 0, 0, 0)) SWIG_fail;
  result = (CModeMessage *)new CModeMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModeMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CModeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModeMessage *arg1 = (CModeMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModeMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CModeMessage" "', argument " "1"" of type '" "CModeMessage *""'"); 
  }
  arg1 = reinterpret_cast< CModeMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CModeMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CModeMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CModeMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CNickMessage_GetOldNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNickMessage *arg1 = (CNickMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNickMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNickMessage_GetOldNick" "', argument " "1"" of type '" "CNickMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CNickMessage * >(argp1);
  result = ((CNickMessage const *)arg1)->GetOldNick();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNickMessage_GetNewNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNickMessage *arg1 = (CNickMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNickMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNickMessage_GetNewNick" "', argument " "1"" of type '" "CNickMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CNickMessage * >(argp1);
  result = ((CNickMessage const *)arg1)->GetNewNick();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNickMessage_SetNewNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNickMessage *arg1 = (CNickMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CNickMessage_SetNewNick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNickMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNickMessage_SetNewNick" "', argument " "1"" of type '" "CNickMessage *""'"); 
  }
  arg1 = reinterpret_cast< CNickMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CNickMessage_SetNewNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CNickMessage_SetNewNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetNewNick((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CNickMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNickMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CNickMessage", 0, 0, 0)) SWIG_fail;
  result = (CNickMessage *)new CNickMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNickMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CNickMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNickMessage *arg1 = (CNickMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNickMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CNickMessage" "', argument " "1"" of type '" "CNickMessage *""'"); 
  }
  arg1 = reinterpret_cast< CNickMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CNickMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CNickMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CNickMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CNoticeMessage_GetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNoticeMessage *arg1 = (CNoticeMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNoticeMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNoticeMessage_GetText" "', argument " "1"" of type '" "CNoticeMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CNoticeMessage * >(argp1);
  result = ((CNoticeMessage const *)arg1)->GetText();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNoticeMessage_SetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNoticeMessage *arg1 = (CNoticeMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CNoticeMessage_SetText", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNoticeMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNoticeMessage_SetText" "', argument " "1"" of type '" "CNoticeMessage *""'"); 
  }
  arg1 = reinterpret_cast< CNoticeMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CNoticeMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CNoticeMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetText((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CNoticeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNoticeMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CNoticeMessage", 0, 0, 0)) SWIG_fail;
  result = (CNoticeMessage *)new CNoticeMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNoticeMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CNoticeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNoticeMessage *arg1 = (CNoticeMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNoticeMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CNoticeMessage" "', argument " "1"" of type '" "CNoticeMessage *""'"); 
  }
  arg1 = reinterpret_cast< CNoticeMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CNoticeMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CNoticeMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CNoticeMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CNumericMessage_GetCode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNumericMessage *arg1 = (CNumericMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNumericMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNumericMessage_GetCode" "', argument " "1"" of type '" "CNumericMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CNumericMessage * >(argp1);
  result = (unsigned int)((CNumericMessage const *)arg1)->GetCode();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CNumericMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNumericMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CNumericMessage", 0, 0, 0)) SWIG_fail;
  result = (CNumericMessage *)new CNumericMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNumericMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CNumericMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNumericMessage *arg1 = (CNumericMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNumericMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CNumericMessage" "', argument " "1"" of type '" "CNumericMessage *""'"); 
  }
  arg1 = reinterpret_cast< CNumericMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CNumericMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CNumericMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CNumericMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CKickMessage_GetKickedNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CKickMessage *arg1 = (CKickMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CKickMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CKickMessage_GetKickedNick" "', argument " "1"" of type '" "CKickMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CKickMessage * >(argp1);
  result = ((CKickMessage const *)arg1)->GetKickedNick();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CKickMessage_SetKickedNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CKickMessage *arg1 = (CKickMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CKickMessage_SetKickedNick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CKickMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CKickMessage_SetKickedNick" "', argument " "1"" of type '" "CKickMessage *""'"); 
  }
  arg1 = reinterpret_cast< CKickMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CKickMessage_SetKickedNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CKickMessage_SetKickedNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetKickedNick((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CKickMessage_GetReason(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CKickMessage *arg1 = (CKickMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CKickMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CKickMessage_GetReason" "', argument " "1"" of type '" "CKickMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CKickMessage * >(argp1);
  result = ((CKickMessage const *)arg1)->GetReason();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CKickMessage_SetReason(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CKickMessage *arg1 = (CKickMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CKickMessage_SetReason", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CKickMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CKickMessage_SetReason" "', argument " "1"" of type '" "CKickMessage *""'"); 
  }
  arg1 = reinterpret_cast< CKickMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CKickMessage_SetReason" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CKickMessage_SetReason" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetReason((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CKickMessage_GetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CKickMessage *arg1 = (CKickMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CKickMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CKickMessage_GetText" "', argument " "1"" of type '" "CKickMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CKickMessage * >(argp1);
  result = ((CKickMessage const *)arg1)->GetText();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CKickMessage_SetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CKickMessage *arg1 = (CKickMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CKickMessage_SetText", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CKickMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CKickMessage_SetText" "', argument " "1"" of type '" "CKickMessage *""'"); 
  }
  arg1 = reinterpret_cast< CKickMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CKickMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CKickMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetText((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CKickMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CKickMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CKickMessage", 0, 0, 0)) SWIG_fail;
  result = (CKickMessage *)new CKickMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CKickMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CKickMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CKickMessage *arg1 = (CKickMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CKickMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CKickMessage" "', argument " "1"" of type '" "CKickMessage *""'"); 
  }
  arg1 = reinterpret_cast< CKickMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CKickMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CKickMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CKickMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CInviteMessage_GetInvitedNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CInviteMessage *arg1 = (CInviteMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CInviteMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CInviteMessage_GetInvitedNick" "', argument " "1"" of type '" "CInviteMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CInviteMessage * >(argp1);
  result = ((CInviteMessage const *)arg1)->GetInvitedNick();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CInviteMessage_SetInvitedNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CInviteMessage *arg1 = (CInviteMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CInviteMessage_SetInvitedNick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CInviteMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CInviteMessage_SetInvitedNick" "', argument " "1"" of type '" "CInviteMessage *""'"); 
  }
  arg1 = reinterpret_cast< CInviteMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CInviteMessage_SetInvitedNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CInviteMessage_SetInvitedNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetInvitedNick((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CInviteMessage_GetChannel(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CInviteMessage *arg1 = (CInviteMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CInviteMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CInviteMessage_GetChannel" "', argument " "1"" of type '" "CInviteMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CInviteMessage * >(argp1);
  result = ((CInviteMessage const *)arg1)->GetChannel();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CInviteMessage_SetChannel(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CInviteMessage *arg1 = (CInviteMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CInviteMessage_SetChannel", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CInviteMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CInviteMessage_SetChannel" "', argument " "1"" of type '" "CInviteMessage *""'"); 
  }
  arg1 = reinterpret_cast< CInviteMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CInviteMessage_SetChannel" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CInviteMessage_SetChannel" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetChannel((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CInviteMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CInviteMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CInviteMessage", 0, 0, 0)) SWIG_fail;
  result = (CInviteMessage *)new CInviteMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CInviteMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CInviteMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CInviteMessage *arg1 = (CInviteMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CInviteMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CInviteMessage" "', argument " "1"" of type '" "CInviteMessage *""'"); 
  }
  arg1 = reinterpret_cast< CInviteMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CInviteMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CInviteMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CInviteMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CPartMessage_GetReason(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPartMessage *arg1 = (CPartMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPartMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPartMessage_GetReason" "', argument " "1"" of type '" "CPartMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CPartMessage * >(argp1);
  result = ((CPartMessage const *)arg1)->GetReason();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPartMessage_SetReason(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPartMessage *arg1 = (CPartMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPartMessage_SetReason", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPartMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPartMessage_SetReason" "', argument " "1"" of type '" "CPartMessage *""'"); 
  }
  arg1 = reinterpret_cast< CPartMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPartMessage_SetReason" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPartMessage_SetReason" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetReason((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPartMessage_GetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPartMessage *arg1 = (CPartMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPartMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPartMessage_GetText" "', argument " "1"" of type '" "CPartMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CPartMessage * >(argp1);
  result = ((CPartMessage const *)arg1)->GetText();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPartMessage_SetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPartMessage *arg1 = (CPartMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPartMessage_SetText", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPartMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPartMessage_SetText" "', argument " "1"" of type '" "CPartMessage *""'"); 
  }
  arg1 = reinterpret_cast< CPartMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPartMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPartMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetText((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CPartMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPartMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CPartMessage", 0, 0, 0)) SWIG_fail;
  result = (CPartMessage *)new CPartMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CPartMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CPartMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPartMessage *arg1 = (CPartMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPartMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CPartMessage" "', argument " "1"" of type '" "CPartMessage *""'"); 
  }
  arg1 = reinterpret_cast< CPartMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CPartMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CPartMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CPartMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CQuitMessage_GetReason(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CQuitMessage *arg1 = (CQuitMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuitMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuitMessage_GetReason" "', argument " "1"" of type '" "CQuitMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CQuitMessage * >(argp1);
  result = ((CQuitMessage const *)arg1)->GetReason();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuitMessage_SetReason(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CQuitMessage *arg1 = (CQuitMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CQuitMessage_SetReason", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuitMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuitMessage_SetReason" "', argument " "1"" of type '" "CQuitMessage *""'"); 
  }
  arg1 = reinterpret_cast< CQuitMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CQuitMessage_SetReason" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuitMessage_SetReason" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetReason((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuitMessage_GetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CQuitMessage *arg1 = (CQuitMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuitMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuitMessage_GetText" "', argument " "1"" of type '" "CQuitMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CQuitMessage * >(argp1);
  result = ((CQuitMessage const *)arg1)->GetText();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuitMessage_SetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CQuitMessage *arg1 = (CQuitMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CQuitMessage_SetText", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuitMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuitMessage_SetText" "', argument " "1"" of type '" "CQuitMessage *""'"); 
  }
  arg1 = reinterpret_cast< CQuitMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CQuitMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuitMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetText((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CQuitMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CQuitMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CQuitMessage", 0, 0, 0)) SWIG_fail;
  result = (CQuitMessage *)new CQuitMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CQuitMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CQuitMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CQuitMessage *arg1 = (CQuitMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuitMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CQuitMessage" "', argument " "1"" of type '" "CQuitMessage *""'"); 
  }
  arg1 = reinterpret_cast< CQuitMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CQuitMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CQuitMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CQuitMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CTextMessage_GetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTextMessage *arg1 = (CTextMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTextMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTextMessage_GetText" "', argument " "1"" of type '" "CTextMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CTextMessage * >(argp1);
  result = ((CTextMessage const *)arg1)->GetText();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTextMessage_SetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTextMessage *arg1 = (CTextMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTextMessage_SetText", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTextMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTextMessage_SetText" "', argument " "1"" of type '" "CTextMessage *""'"); 
  }
  arg1 = reinterpret_cast< CTextMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTextMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTextMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetText((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CTextMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTextMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CTextMessage", 0, 0, 0)) SWIG_fail;
  result = (CTextMessage *)new CTextMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTextMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CTextMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTextMessage *arg1 = (CTextMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTextMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CTextMessage" "', argument " "1"" of type '" "CTextMessage *""'"); 
  }
  arg1 = reinterpret_cast< CTextMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CTextMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CTextMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CTextMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CTopicMessage_GetTopic(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTopicMessage *arg1 = (CTopicMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTopicMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTopicMessage_GetTopic" "', argument " "1"" of type '" "CTopicMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CTopicMessage * >(argp1);
  result = ((CTopicMessage const *)arg1)->GetTopic();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTopicMessage_SetTopic(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTopicMessage *arg1 = (CTopicMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTopicMessage_SetTopic", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTopicMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTopicMessage_SetTopic" "', argument " "1"" of type '" "CTopicMessage *""'"); 
  }
  arg1 = reinterpret_cast< CTopicMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTopicMessage_SetTopic" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTopicMessage_SetTopic" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetTopic((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTopicMessage_GetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTopicMessage *arg1 = (CTopicMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTopicMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTopicMessage_GetText" "', argument " "1"" of type '" "CTopicMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CTopicMessage * >(argp1);
  result = ((CTopicMessage const *)arg1)->GetText();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTopicMessage_SetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTopicMessage *arg1 = (CTopicMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTopicMessage_SetText", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTopicMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTopicMessage_SetText" "', argument " "1"" of type '" "CTopicMessage *""'"); 
  }
  arg1 = reinterpret_cast< CTopicMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTopicMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTopicMessage_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetText((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CTopicMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTopicMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CTopicMessage", 0, 0, 0)) SWIG_fail;
  result = (CTopicMessage *)new CTopicMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTopicMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CTopicMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTopicMessage *arg1 = (CTopicMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTopicMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CTopicMessage" "', argument " "1"" of type '" "CTopicMessage *""'"); 
  }
  arg1 = reinterpret_cast< CTopicMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CTopicMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CTopicMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CTopicMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CChgHostMessage_GetNewIdent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChgHostMessage *arg1 = (CChgHostMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChgHostMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChgHostMessage_GetNewIdent" "', argument " "1"" of type '" "CChgHostMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CChgHostMessage * >(argp1);
  result = ((CChgHostMessage const *)arg1)->GetNewIdent();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChgHostMessage_SetNewIdent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChgHostMessage *arg1 = (CChgHostMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChgHostMessage_SetNewIdent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChgHostMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChgHostMessage_SetNewIdent" "', argument " "1"" of type '" "CChgHostMessage *""'"); 
  }
  arg1 = reinterpret_cast< CChgHostMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChgHostMessage_SetNewIdent" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChgHostMessage_SetNewIdent" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetNewIdent((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChgHostMessage_GetNewHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChgHostMessage *arg1 = (CChgHostMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChgHostMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChgHostMessage_GetNewHost" "', argument " "1"" of type '" "CChgHostMessage const *""'"); 
  }
  arg1 = reinterpret_cast< CChgHostMessage * >(argp1);
  result = ((CChgHostMessage const *)arg1)->GetNewHost();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChgHostMessage_SetNewHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChgHostMessage *arg1 = (CChgHostMessage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChgHostMessage_SetNewHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChgHostMessage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChgHostMessage_SetNewHost" "', argument " "1"" of type '" "CChgHostMessage *""'"); 
  }
  arg1 = reinterpret_cast< CChgHostMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChgHostMessage_SetNewHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChgHostMessage_SetNewHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetNewHost((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CChgHostMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChgHostMessage *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CChgHostMessage", 0, 0, 0)) SWIG_fail;
  result = (CChgHostMessage *)new CChgHostMessage();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CChgHostMessage, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CChgHostMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChgHostMessage *arg1 = (CChgHostMessage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChgHostMessage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CChgHostMessage" "', argument " "1"" of type '" "CChgHostMessage *""'"); 
  }
  arg1 = reinterpret_cast< CChgHostMessage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CChgHostMessage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CChgHostMessage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CChgHostMessage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CModuleEntry_pcVersion_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModuleEntry *arg1 = (CModuleEntry *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModuleEntry_pcVersion_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModuleEntry, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModuleEntry_pcVersion_set" "', argument " "1"" of type '" "CModuleEntry *""'"); 
  }
  arg1 = reinterpret_cast< CModuleEntry * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModuleEntry_pcVersion_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  if (arg2) {
    size_t size = strlen(reinterpret_cast< const char * >(reinterpret_cast< const char * >(arg2))) + 1;
    arg1->pcVersion = (char const *)reinterpret_cast< char* >(memcpy(new char[size], arg2, sizeof(char)*(size)));
  } else {
    arg1->pcVersion = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModuleEntry_pcVersion_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModuleEntry *arg1 = (CModuleEntry *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModuleEntry, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModuleEntry_pcVersion_get" "', argument " "1"" of type '" "CModuleEntry *""'"); 
  }
  arg1 = reinterpret_cast< CModuleEntry * >(argp1);
  result = (char *) ((arg1)->pcVersion);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModuleEntry_pcVersionExtra_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModuleEntry *arg1 = (CModuleEntry *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModuleEntry_pcVersionExtra_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModuleEntry, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModuleEntry_pcVersionExtra_set" "', argument " "1"" of type '" "CModuleEntry *""'"); 
  }
  arg1 = reinterpret_cast< CModuleEntry * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModuleEntry_pcVersionExtra_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  if (arg2) {
    size_t size = strlen(reinterpret_cast< const char * >(reinterpret_cast< const char * >(arg2))) + 1;
    arg1->pcVersionExtra = (char const *)reinterpret_cast< char* >(memcpy(new char[size], arg2, sizeof(char)*(size)));
  } else {
    arg1->pcVersionExtra = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModuleEntry_pcVersionExtra_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModuleEntry *arg1 = (CModuleEntry *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModuleEntry, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModuleEntry_pcVersionExtra_get" "', argument " "1"" of type '" "CModuleEntry *""'"); 
  }
  arg1 = reinterpret_cast< CModuleEntry * >(argp1);
  result = (char *) ((arg1)->pcVersionExtra);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModuleEntry_pcCompileOptions_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModuleEntry *arg1 = (CModuleEntry *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModuleEntry_pcCompileOptions_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModuleEntry, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModuleEntry_pcCompileOptions_set" "', argument " "1"" of type '" "CModuleEntry *""'"); 
  }
  arg1 = reinterpret_cast< CModuleEntry * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModuleEntry_pcCompileOptions_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  if (arg2) {
    size_t size = strlen(reinterpret_cast< const char * >(reinterpret_cast< const char * >(arg2))) + 1;
    arg1->pcCompileOptions = (char const *)reinterpret_cast< char* >(memcpy(new char[size], arg2, sizeof(char)*(size)));
  } else {
    arg1->pcCompileOptions = 0;
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModuleEntry_pcCompileOptions_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModuleEntry *arg1 = (CModuleEntry *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModuleEntry, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModuleEntry_pcCompileOptions_get" "', argument " "1"" of type '" "CModuleEntry *""'"); 
  }
  arg1 = reinterpret_cast< CModuleEntry * >(argp1);
  result = (char *) ((arg1)->pcCompileOptions);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModuleEntry_fpFillModInfo_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModuleEntry *arg1 = (CModuleEntry *) 0 ;
  void (*arg2)(CModInfo &) = (void (*)(CModInfo &)) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModuleEntry_fpFillModInfo_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModuleEntry, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModuleEntry_fpFillModInfo_set" "', argument " "1"" of type '" "CModuleEntry *""'"); 
  }
  arg1 = reinterpret_cast< CModuleEntry * >(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_r_CModInfo__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "CModuleEntry_fpFillModInfo_set" "', argument " "2"" of type '" "void (*)(CModInfo &)""'"); 
    }
  }
  if (arg1) (arg1)->fpFillModInfo = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModuleEntry_fpFillModInfo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModuleEntry *arg1 = (CModuleEntry *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  void (*result)(CModInfo &) = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModuleEntry, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModuleEntry_fpFillModInfo_get" "', argument " "1"" of type '" "CModuleEntry *""'"); 
  }
  arg1 = reinterpret_cast< CModuleEntry * >(argp1);
  result = (void (*)(CModInfo &)) ((arg1)->fpFillModInfo);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_r_CModInfo__void);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CModuleEntry(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModuleEntry *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CModuleEntry", 0, 0, 0)) SWIG_fail;
  result = (CModuleEntry *)new CModuleEntry();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModuleEntry, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CModuleEntry(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModuleEntry *arg1 = (CModuleEntry *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModuleEntry, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CModuleEntry" "', argument " "1"" of type '" "CModuleEntry *""'"); 
  }
  arg1 = reinterpret_cast< CModuleEntry * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CModuleEntry_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CModuleEntry, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CModuleEntry_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_delete_CCapability(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCapability *arg1 = (CCapability *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCapability, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CCapability" "', argument " "1"" of type '" "CCapability *""'"); 
  }
  arg1 = reinterpret_cast< CCapability * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCapability_OnServerChangedSupport(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCapability *arg1 = (CCapability *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CCapability_OnServerChangedSupport", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCapability, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCapability_OnServerChangedSupport" "', argument " "1"" of type '" "CCapability *""'"); 
  }
  arg1 = reinterpret_cast< CCapability * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CCapability_OnServerChangedSupport" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CCapability_OnServerChangedSupport" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->OnServerChangedSupport(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCapability_OnClientChangedSupport(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCapability *arg1 = (CCapability *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CCapability_OnClientChangedSupport", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCapability, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCapability_OnClientChangedSupport" "', argument " "1"" of type '" "CCapability *""'"); 
  }
  arg1 = reinterpret_cast< CCapability * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CCapability_OnClientChangedSupport" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CCapability_OnClientChangedSupport" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->OnClientChangedSupport(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCapability_GetModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCapability *arg1 = (CCapability *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CModule *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCapability, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCapability_GetModule" "', argument " "1"" of type '" "CCapability *""'"); 
  }
  arg1 = reinterpret_cast< CCapability * >(argp1);
  result = (CModule *)(arg1)->GetModule();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModule, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CCapability_SetModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCapability *arg1 = (CCapability *) 0 ;
  CModule *arg2 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CCapability_SetModule", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CCapability, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CCapability_SetModule" "', argument " "1"" of type '" "CCapability *""'"); 
  }
  arg1 = reinterpret_cast< CCapability * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CCapability_SetModule" "', argument " "2"" of type '" "CModule *""'"); 
  }
  arg2 = reinterpret_cast< CModule * >(argp2);
  (arg1)->SetModule(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CCapability(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CCapability *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CCapability", 0, 0, 0)) SWIG_fail;
  result = (CCapability *)new CCapability();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CCapability, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CCapability_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CCapability, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CCapability_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CTimer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  CTimer *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CTimer", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CTimer" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CTimer" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CTimer" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CTimer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CTimer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_CTimer" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CTimer" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  result = (CTimer *)new CTimer(arg1,arg2,arg3,(CString const &)*arg4,(CString const &)*arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTimer, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CTimer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTimer *arg1 = (CTimer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTimer, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CTimer" "', argument " "1"" of type '" "CTimer *""'"); 
  }
  arg1 = reinterpret_cast< CTimer * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTimer_SetModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTimer *arg1 = (CTimer *) 0 ;
  CModule *arg2 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTimer_SetModule", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTimer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTimer_SetModule" "', argument " "1"" of type '" "CTimer *""'"); 
  }
  arg1 = reinterpret_cast< CTimer * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTimer_SetModule" "', argument " "2"" of type '" "CModule *""'"); 
  }
  arg2 = reinterpret_cast< CModule * >(argp2);
  (arg1)->SetModule(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTimer_SetDescription(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTimer *arg1 = (CTimer *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTimer_SetDescription", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTimer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTimer_SetDescription" "', argument " "1"" of type '" "CTimer *""'"); 
  }
  arg1 = reinterpret_cast< CTimer * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTimer_SetDescription" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTimer_SetDescription" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetDescription((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTimer_GetModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTimer *arg1 = (CTimer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CModule *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTimer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTimer_GetModule" "', argument " "1"" of type '" "CTimer const *""'"); 
  }
  arg1 = reinterpret_cast< CTimer * >(argp1);
  result = (CModule *)((CTimer const *)arg1)->GetModule();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModule, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTimer_GetDescription(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTimer *arg1 = (CTimer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTimer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTimer_GetDescription" "', argument " "1"" of type '" "CTimer const *""'"); 
  }
  arg1 = reinterpret_cast< CTimer * >(argp1);
  result = (CString *) &((CTimer const *)arg1)->GetDescription();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CTimer_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CTimer, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CTimer_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CFPTimer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  CFPTimer *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CFPTimer", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CFPTimer" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CFPTimer" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CFPTimer" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CFPTimer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CFPTimer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_CFPTimer" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CFPTimer" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  result = (CFPTimer *)new CFPTimer(arg1,arg2,arg3,(CString const &)*arg4,(CString const &)*arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CFPTimer, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CFPTimer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFPTimer *arg1 = (CFPTimer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFPTimer, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CFPTimer" "', argument " "1"" of type '" "CFPTimer *""'"); 
  }
  arg1 = reinterpret_cast< CFPTimer * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CFPTimer_SetFPCallback(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CFPTimer *arg1 = (CFPTimer *) 0 ;
  FPTimer_t arg2 = (FPTimer_t) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CFPTimer_SetFPCallback", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CFPTimer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CFPTimer_SetFPCallback" "', argument " "1"" of type '" "CFPTimer *""'"); 
  }
  arg1 = reinterpret_cast< CFPTimer * >(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_CModule_p_CFPTimer__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "CFPTimer_SetFPCallback" "', argument " "2"" of type '" "FPTimer_t""'"); 
    }
  }
  (arg1)->SetFPCallback(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CFPTimer_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CFPTimer, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CFPTimer_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CModInfo__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CModInfo *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CModInfo *)new CModInfo();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModInfo, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CModInfo__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  CModInfo::EModuleType arg3 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  int val3 ;
  int ecode3 = 0 ;
  CModInfo *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CModInfo" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModInfo" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CModInfo" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModInfo" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CModInfo" "', argument " "3"" of type '" "CModInfo::EModuleType""'");
  } 
  arg3 = static_cast< CModInfo::EModuleType >(val3);
  result = (CModInfo *)new CModInfo((CString const &)*arg1,(CString const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModInfo, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CModInfo(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CModInfo", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CModInfo__SWIG_0(self, argc, argv);
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_new_CModInfo__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CModInfo'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModInfo::CModInfo()\n"
    "    CModInfo::CModInfo(CString const &,CString const &,CModInfo::EModuleType)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CModInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CModInfo" "', argument " "1"" of type '" "CModInfo *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo___lt__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  CModInfo *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModInfo___lt__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo___lt__" "', argument " "1"" of type '" "CModInfo const *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModInfo___lt__" "', argument " "2"" of type '" "CModInfo const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModInfo___lt__" "', argument " "2"" of type '" "CModInfo const &""'"); 
  }
  arg2 = reinterpret_cast< CModInfo * >(argp2);
  result = (bool)((CModInfo const *)arg1)->operator <((CModInfo const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_CModInfo_SupportsType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  CModInfo::EModuleType arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModInfo_SupportsType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_SupportsType" "', argument " "1"" of type '" "CModInfo const *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CModInfo_SupportsType" "', argument " "2"" of type '" "CModInfo::EModuleType""'");
  } 
  arg2 = static_cast< CModInfo::EModuleType >(val2);
  result = (bool)((CModInfo const *)arg1)->SupportsType(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_AddType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  CModInfo::EModuleType arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModInfo_AddType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_AddType" "', argument " "1"" of type '" "CModInfo *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CModInfo_AddType" "', argument " "2"" of type '" "CModInfo::EModuleType""'");
  } 
  arg2 = static_cast< CModInfo::EModuleType >(val2);
  (arg1)->AddType(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_ModuleTypeToString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo::EModuleType arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "CModInfo_ModuleTypeToString" "', argument " "1"" of type '" "CModInfo::EModuleType""'");
  } 
  arg1 = static_cast< CModInfo::EModuleType >(val1);
  result = CModInfo::ModuleTypeToString(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_GetName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_GetName" "', argument " "1"" of type '" "CModInfo const *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  result = (CString *) &((CModInfo const *)arg1)->GetName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_GetPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_GetPath" "', argument " "1"" of type '" "CModInfo const *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  result = (CString *) &((CModInfo const *)arg1)->GetPath();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_GetDescription(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_GetDescription" "', argument " "1"" of type '" "CModInfo const *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  result = (CString *) &((CModInfo const *)arg1)->GetDescription();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_GetWikiPage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_GetWikiPage" "', argument " "1"" of type '" "CModInfo const *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  result = (CString *) &((CModInfo const *)arg1)->GetWikiPage();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_GetArgsHelpText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_GetArgsHelpText" "', argument " "1"" of type '" "CModInfo const *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  result = (CString *) &((CModInfo const *)arg1)->GetArgsHelpText();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_GetHasArgs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_GetHasArgs" "', argument " "1"" of type '" "CModInfo const *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  result = (bool)((CModInfo const *)arg1)->GetHasArgs();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_GetLoader(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CModInfo::ModLoader result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_GetLoader" "', argument " "1"" of type '" "CModInfo const *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  result = (CModInfo::ModLoader)((CModInfo const *)arg1)->GetLoader();
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModule);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_GetDefaultType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CModInfo::EModuleType result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_GetDefaultType" "', argument " "1"" of type '" "CModInfo const *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  result = (CModInfo::EModuleType)((CModInfo const *)arg1)->GetDefaultType();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_SetName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModInfo_SetName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_SetName" "', argument " "1"" of type '" "CModInfo *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModInfo_SetName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModInfo_SetName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_SetPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModInfo_SetPath", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_SetPath" "', argument " "1"" of type '" "CModInfo *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModInfo_SetPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModInfo_SetPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetPath((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_SetDescription(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModInfo_SetDescription", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_SetDescription" "', argument " "1"" of type '" "CModInfo *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModInfo_SetDescription" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModInfo_SetDescription" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetDescription((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_SetWikiPage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModInfo_SetWikiPage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_SetWikiPage" "', argument " "1"" of type '" "CModInfo *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModInfo_SetWikiPage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModInfo_SetWikiPage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetWikiPage((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_SetArgsHelpText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModInfo_SetArgsHelpText", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_SetArgsHelpText" "', argument " "1"" of type '" "CModInfo *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModInfo_SetArgsHelpText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModInfo_SetArgsHelpText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetArgsHelpText((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_SetHasArgs__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_SetHasArgs" "', argument " "1"" of type '" "CModInfo *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CModInfo_SetHasArgs" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetHasArgs(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_SetHasArgs__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_SetHasArgs" "', argument " "1"" of type '" "CModInfo *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  (arg1)->SetHasArgs();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_SetHasArgs(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModInfo_SetHasArgs", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModInfo, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CModInfo_SetHasArgs__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModInfo, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CModInfo_SetHasArgs__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModInfo_SetHasArgs'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModInfo::SetHasArgs(bool)\n"
    "    CModInfo::SetHasArgs()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModInfo_SetLoader(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  CModInfo::ModLoader arg2 = (CModInfo::ModLoader) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModInfo_SetLoader", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_SetLoader" "', argument " "1"" of type '" "CModInfo *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModule);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "CModInfo_SetLoader" "', argument " "2"" of type '" "CModInfo::ModLoader""'"); 
    }
  }
  (arg1)->SetLoader(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_SetDefaultType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  CModInfo::EModuleType arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModInfo_SetDefaultType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_SetDefaultType" "', argument " "1"" of type '" "CModInfo *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CModInfo_SetDefaultType" "', argument " "2"" of type '" "CModInfo::EModuleType""'");
  } 
  arg2 = static_cast< CModInfo::EModuleType >(val2);
  (arg1)->SetDefaultType(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_t_s__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_t_s" "', argument " "1"" of type '" "CModInfo const *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModInfo_t_s" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModInfo_t_s" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModInfo_t_s" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModInfo_t_s" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = ((CModInfo const *)arg1)->t_s((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_t_s__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = (CModInfo *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModInfo_t_s" "', argument " "1"" of type '" "CModInfo const *""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModInfo_t_s" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModInfo_t_s" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CModInfo const *)arg1)->t_s((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModInfo_t_s(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModInfo_t_s", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModInfo, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModInfo_t_s__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModInfo, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CModInfo_t_s__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModInfo_t_s'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModInfo::t_s(CString const &,CString const &) const\n"
    "    CModInfo::t_s(CString const &) const\n");
  return 0;
}


SWIGINTERN PyObject *CModInfo_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CModInfo, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CModInfo_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CModCommand__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CModCommand *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CModCommand *)new CModCommand();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModCommand, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CModCommand__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CModule *arg2 = (CModule *) 0 ;
  CModCommand::ModCmdFunc arg3 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  CModCommand *result = 0 ;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CModCommand" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModCommand" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CModCommand" "', argument " "2"" of type '" "CModule *""'"); 
  }
  arg2 = reinterpret_cast< CModule * >(argp2);
  {
    int res = SWIG_ConvertMember(swig_obj[2], SWIG_as_voidptr(&arg3), sizeof(arg3),SWIGTYPE_m_CModule__f_r_q_const__CString__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "new_CModCommand" "', argument " "3"" of type '" "CModCommand::ModCmdFunc""'"); 
    }
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CModCommand" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModCommand" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_CModCommand" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModCommand" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  result = (CModCommand *)new CModCommand((CString const &)*arg1,arg2,arg3,(CString const &)*arg4,(CString const &)*arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModCommand, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CModCommand__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  SwigValueWrapper< std::function< void (CString const &) > > arg2 ;
  COptionalTranslation *arg3 = 0 ;
  COptionalTranslation *arg4 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  CModCommand *result = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CModCommand" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModCommand" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__functionT_void_fCString_const_RF_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CModCommand" "', argument " "2"" of type '" "CModCommand::CmdFunc""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModCommand" "', argument " "2"" of type '" "CModCommand::CmdFunc""'");
    } else {
      CModCommand::CmdFunc * temp = reinterpret_cast< CModCommand::CmdFunc * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_COptionalTranslation,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CModCommand" "', argument " "3"" of type '" "COptionalTranslation const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModCommand" "', argument " "3"" of type '" "COptionalTranslation const &""'"); 
  }
  arg3 = reinterpret_cast< COptionalTranslation * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_COptionalTranslation,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CModCommand" "', argument " "4"" of type '" "COptionalTranslation const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModCommand" "', argument " "4"" of type '" "COptionalTranslation const &""'"); 
  }
  arg4 = reinterpret_cast< COptionalTranslation * >(argp4);
  result = (CModCommand *)new CModCommand((CString const &)*arg1,arg2,(COptionalTranslation const &)*arg3,(COptionalTranslation const &)*arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModCommand, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CModCommand__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModCommand *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CModCommand *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CModCommand,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CModCommand" "', argument " "1"" of type '" "CModCommand const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModCommand" "', argument " "1"" of type '" "CModCommand const &""'"); 
  }
  arg1 = reinterpret_cast< CModCommand * >(argp1);
  result = (CModCommand *)new CModCommand((CModCommand const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModCommand, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CModCommand(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CModCommand", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CModCommand__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_CModCommand, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CModCommand__SWIG_3(self, argc, argv);
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_std__functionT_void_fCString_const_RF_t, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_COptionalTranslation, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_ConvertPtr(argv[3], 0, SWIGTYPE_p_COptionalTranslation, SWIG_POINTER_NO_NULL | 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_new_CModCommand__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CModule, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_m_CModule__f_r_q_const__CString__void, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            int res = SWIG_AsPtr_CString(argv[4], (CString**)(0));
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_new_CModCommand__SWIG_1(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CModCommand'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModCommand::CModCommand()\n"
    "    CModCommand::CModCommand(CString const &,CModule *,CModCommand::ModCmdFunc,CString const &,CString const &)\n"
    "    CModCommand::CModCommand(CString const &,CModCommand::CmdFunc,COptionalTranslation const &,COptionalTranslation const &)\n"
    "    CModCommand::CModCommand(CModCommand const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModCommand_InitHelp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CTable,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModCommand_InitHelp" "', argument " "1"" of type '" "CTable &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModCommand_InitHelp" "', argument " "1"" of type '" "CTable &""'"); 
  }
  arg1 = reinterpret_cast< CTable * >(argp1);
  CModCommand::InitHelp(*arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModCommand_AddHelp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModCommand *arg1 = (CModCommand *) 0 ;
  CTable *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModCommand_AddHelp", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModCommand, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModCommand_AddHelp" "', argument " "1"" of type '" "CModCommand const *""'"); 
  }
  arg1 = reinterpret_cast< CModCommand * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTable,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModCommand_AddHelp" "', argument " "2"" of type '" "CTable &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModCommand_AddHelp" "', argument " "2"" of type '" "CTable &""'"); 
  }
  arg2 = reinterpret_cast< CTable * >(argp2);
  ((CModCommand const *)arg1)->AddHelp(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModCommand_GetCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModCommand *arg1 = (CModCommand *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModCommand, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModCommand_GetCommand" "', argument " "1"" of type '" "CModCommand const *""'"); 
  }
  arg1 = reinterpret_cast< CModCommand * >(argp1);
  result = (CString *) &((CModCommand const *)arg1)->GetCommand();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModCommand_GetFunction(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModCommand *arg1 = (CModCommand *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::function< void (CString const &) > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModCommand, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModCommand_GetFunction" "', argument " "1"" of type '" "CModCommand const *""'"); 
  }
  arg1 = reinterpret_cast< CModCommand * >(argp1);
  result = ((CModCommand const *)arg1)->GetFunction();
  resultobj = SWIG_NewPointerObj((new CModCommand::CmdFunc(result)), SWIGTYPE_p_std__functionT_void_fCString_const_RF_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModCommand_GetArgs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModCommand *arg1 = (CModCommand *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModCommand, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModCommand_GetArgs" "', argument " "1"" of type '" "CModCommand const *""'"); 
  }
  arg1 = reinterpret_cast< CModCommand * >(argp1);
  result = ((CModCommand const *)arg1)->GetArgs();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModCommand_GetDescription(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModCommand *arg1 = (CModCommand *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModCommand, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModCommand_GetDescription" "', argument " "1"" of type '" "CModCommand const *""'"); 
  }
  arg1 = reinterpret_cast< CModCommand * >(argp1);
  result = ((CModCommand const *)arg1)->GetDescription();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModCommand_Call(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModCommand *arg1 = (CModCommand *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModCommand_Call", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModCommand, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModCommand_Call" "', argument " "1"" of type '" "CModCommand const *""'"); 
  }
  arg1 = reinterpret_cast< CModCommand * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModCommand_Call" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModCommand_Call" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ((CModCommand const *)arg1)->Call((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CModCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModCommand *arg1 = (CModCommand *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModCommand, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CModCommand" "', argument " "1"" of type '" "CModCommand *""'"); 
  }
  arg1 = reinterpret_cast< CModCommand * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CModCommand_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CModCommand, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CModCommand_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CModule__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  ModHandle arg1 = (ModHandle) 0 ;
  CUser *arg2 = (CUser *) 0 ;
  CIRCNetwork *arg3 = (CIRCNetwork *) 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  CModInfo::EModuleType arg6 ;
  int res1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  int val6 ;
  int ecode6 = 0 ;
  CModule *result = 0 ;
  
  (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CModule" "', argument " "1"" of type '" "ModHandle""'"); 
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CModule" "', argument " "2"" of type '" "CUser *""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CModule" "', argument " "3"" of type '" "CIRCNetwork *""'"); 
  }
  arg3 = reinterpret_cast< CIRCNetwork * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CModule" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModule" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_CModule" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModule" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_CModule" "', argument " "6"" of type '" "CModInfo::EModuleType""'");
  } 
  arg6 = static_cast< CModInfo::EModuleType >(val6);
  result = (CModule *)new CModule(arg1,arg2,arg3,(CString const &)*arg4,(CString const &)*arg5,arg6);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModule, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CModule__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  ModHandle arg1 = (ModHandle) 0 ;
  CUser *arg2 = (CUser *) 0 ;
  CIRCNetwork *arg3 = (CIRCNetwork *) 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  int res1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  CModule *result = 0 ;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CModule" "', argument " "1"" of type '" "ModHandle""'"); 
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CModule" "', argument " "2"" of type '" "CUser *""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CModule" "', argument " "3"" of type '" "CIRCNetwork *""'"); 
  }
  arg3 = reinterpret_cast< CIRCNetwork * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CModule" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModule" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_CModule" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModule" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  result = (CModule *)new CModule(arg1,arg2,arg3,(CString const &)*arg4,(CString const &)*arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModule, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CModule(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[7] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CModule", 0, 6, argv))) SWIG_fail;
  --argc;
  if (argc == 5) {
    int _v = 0;
    void *ptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &ptr, 0, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CUser, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            int res = SWIG_AsPtr_CString(argv[4], (CString**)(0));
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_new_CModule__SWIG_1(self, argc, argv);
            }
          }
        }
      }
    }
  }
  if (argc == 6) {
    int _v = 0;
    void *ptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &ptr, 0, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CUser, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            int res = SWIG_AsPtr_CString(argv[4], (CString**)(0));
            _v = SWIG_CheckState(res);
            if (_v) {
              {
                int res = SWIG_AsVal_int(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                return _wrap_new_CModule__SWIG_0(self, argc, argv);
              }
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CModule'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModule::CModule(ModHandle,CUser *,CIRCNetwork *,CString const &,CString const &,CModInfo::EModuleType)\n"
    "    CModule::CModule(ModHandle,CUser *,CIRCNetwork *,CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CModule" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_SetUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CUser *arg2 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_SetUser", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_SetUser" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_SetUser" "', argument " "2"" of type '" "CUser *""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  (arg1)->SetUser(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_SetNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_SetNetwork", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_SetNetwork" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_SetNetwork" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  (arg1)->SetNetwork(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_SetClient(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_SetClient", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_SetClient" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_SetClient" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  (arg1)->SetClient(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_Unload(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_Unload" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->Unload();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnLoad(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnLoad", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnLoad" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnLoad" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnLoad" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (bool)(arg1)->OnLoad((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnBoot(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnBoot" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (bool)(arg1)->OnBoot();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_WebRequiresLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_WebRequiresLogin" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (bool)(arg1)->WebRequiresLogin();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_WebRequiresAdmin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_WebRequiresAdmin" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (bool)(arg1)->WebRequiresAdmin();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetWebMenuTitle(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetWebMenuTitle" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (arg1)->GetWebMenuTitle();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetWebPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetWebPath" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (arg1)->GetWebPath();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetWebFilesPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetWebFilesPath" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (arg1)->GetWebFilesPath();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnWebPreRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CWebSock *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnWebPreRequest", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnWebPreRequest" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CWebSock,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnWebPreRequest" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnWebPreRequest" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  arg2 = reinterpret_cast< CWebSock * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnWebPreRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnWebPreRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->OnWebPreRequest(*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnWebRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CWebSock *arg2 = 0 ;
  CString *arg3 = 0 ;
  CTemplate *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnWebRequest", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnWebRequest" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CWebSock,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnWebRequest" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnWebRequest" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  arg2 = reinterpret_cast< CWebSock * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnWebRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnWebRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CTemplate,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnWebRequest" "', argument " "4"" of type '" "CTemplate &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnWebRequest" "', argument " "4"" of type '" "CTemplate &""'"); 
  }
  arg4 = reinterpret_cast< CTemplate * >(argp4);
  result = (bool)(arg1)->OnWebRequest(*arg2,(CString const &)*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_ValidateWebRequestCSRFCheck(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CWebSock *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_ValidateWebRequestCSRFCheck", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_ValidateWebRequestCSRFCheck" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CWebSock,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_ValidateWebRequestCSRFCheck" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_ValidateWebRequestCSRFCheck" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  arg2 = reinterpret_cast< CWebSock * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_ValidateWebRequestCSRFCheck" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_ValidateWebRequestCSRFCheck" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->ValidateWebRequestCSRFCheck(*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddSubPage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  SwigValueWrapper< std::shared_ptr< CWebSubPage > > arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_AddSubPage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_AddSubPage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_AddSubPage" "', argument " "2"" of type '" "TWebSubPage""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddSubPage" "', argument " "2"" of type '" "TWebSubPage""'");
    } else {
      TWebSubPage * temp = reinterpret_cast< TWebSubPage * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  (arg1)->AddSubPage(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_ClearSubPages(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_ClearSubPages" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->ClearSubPages();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetSubPages(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  VWebSubPages *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetSubPages" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (VWebSubPages *) &(arg1)->GetSubPages();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnEmbeddedWebRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CWebSock *arg2 = 0 ;
  CString *arg3 = 0 ;
  CTemplate *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnEmbeddedWebRequest", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnEmbeddedWebRequest" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CWebSock,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnEmbeddedWebRequest" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnEmbeddedWebRequest" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  arg2 = reinterpret_cast< CWebSock * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnEmbeddedWebRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnEmbeddedWebRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CTemplate,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnEmbeddedWebRequest" "', argument " "4"" of type '" "CTemplate &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnEmbeddedWebRequest" "', argument " "4"" of type '" "CTemplate &""'"); 
  }
  arg4 = reinterpret_cast< CTemplate * >(argp4);
  result = (bool)(arg1)->OnEmbeddedWebRequest(*arg2,(CString const &)*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPreRehash(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPreRehash" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->OnPreRehash();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPostRehash(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPostRehash" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->OnPostRehash();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnIRCDisconnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnIRCDisconnected" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->OnIRCDisconnected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnIRCConnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnIRCConnected" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->OnIRCConnected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnIRCConnecting(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CIRCSock *arg2 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnIRCConnecting", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnIRCConnecting" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnIRCConnecting" "', argument " "2"" of type '" "CIRCSock *""'"); 
  }
  arg2 = reinterpret_cast< CIRCSock * >(argp2);
  result = (CModule::EModRet)(arg1)->OnIRCConnecting(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnIRCConnectionError(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CIRCSock *arg2 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnIRCConnectionError", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnIRCConnectionError" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnIRCConnectionError" "', argument " "2"" of type '" "CIRCSock *""'"); 
  }
  arg2 = reinterpret_cast< CIRCSock * >(argp2);
  (arg1)->OnIRCConnectionError(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnIRCRegistration(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[5] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnIRCRegistration", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnIRCRegistration" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sPass");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sNick");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sIdent");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[4], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg5 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 5 sRealName");
    }
  }
  result = (CModule::EModRet)(arg1)->OnIRCRegistration(*arg2,*arg3,*arg4,*arg5);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnBroadcast(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnBroadcast", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnBroadcast" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnBroadcast(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanPermission3(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  char arg5 ;
  bool arg6 ;
  bool arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  char val5 ;
  int ecode5 = 0 ;
  bool val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[7] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanPermission3", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanPermission3" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanPermission3" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnChanPermission3" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanPermission3" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnChanPermission3" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanPermission3" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_char(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModule_OnChanPermission3" "', argument " "5"" of type '" "char""'");
  } 
  arg5 = static_cast< char >(val5);
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CModule_OnChanPermission3" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  ecode7 = SWIG_AsVal_bool(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CModule_OnChanPermission3" "', argument " "7"" of type '" "bool""'");
  } 
  arg7 = static_cast< bool >(val7);
  (arg1)->OnChanPermission3((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5,arg6,arg7);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanPermission2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  unsigned char arg5 ;
  bool arg6 ;
  bool arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  unsigned char val5 ;
  int ecode5 = 0 ;
  bool val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[7] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanPermission2", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanPermission2" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanPermission2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnChanPermission2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanPermission2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnChanPermission2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanPermission2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_unsigned_SS_char(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModule_OnChanPermission2" "', argument " "5"" of type '" "unsigned char""'");
  } 
  arg5 = static_cast< unsigned char >(val5);
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CModule_OnChanPermission2" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  ecode7 = SWIG_AsVal_bool(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CModule_OnChanPermission2" "', argument " "7"" of type '" "bool""'");
  } 
  arg7 = static_cast< bool >(val7);
  (arg1)->OnChanPermission2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5,arg6,arg7);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanPermission(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  unsigned char arg5 ;
  bool arg6 ;
  bool arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  unsigned char val5 ;
  int ecode5 = 0 ;
  bool val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[7] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanPermission", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanPermission" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanPermission" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanPermission" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnChanPermission" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanPermission" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnChanPermission" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanPermission" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_unsigned_SS_char(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModule_OnChanPermission" "', argument " "5"" of type '" "unsigned char""'");
  } 
  arg5 = static_cast< unsigned char >(val5);
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CModule_OnChanPermission" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  ecode7 = SWIG_AsVal_bool(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CModule_OnChanPermission" "', argument " "7"" of type '" "bool""'");
  } 
  arg7 = static_cast< bool >(val7);
  (arg1)->OnChanPermission((CNick const &)*arg2,(CNick const &)*arg3,*arg4,arg5,arg6,arg7);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnOp2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnOp2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnOp2" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnOp2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnOp2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnOp2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnOp2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnOp2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModule_OnOp2" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  (arg1)->OnOp2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnOp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnOp", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnOp" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnOp" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnOp" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnOp" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnOp" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnOp" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnOp" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModule_OnOp" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  (arg1)->OnOp((CNick const &)*arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnDeop2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnDeop2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnDeop2" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnDeop2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnDeop2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnDeop2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnDeop2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnDeop2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModule_OnDeop2" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  (arg1)->OnDeop2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnDeop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnDeop", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnDeop" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnDeop" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnDeop" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnDeop" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnDeop" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnDeop" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnDeop" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModule_OnDeop" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  (arg1)->OnDeop((CNick const &)*arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnVoice2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnVoice2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnVoice2" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnVoice2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnVoice2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnVoice2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnVoice2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnVoice2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModule_OnVoice2" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  (arg1)->OnVoice2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnVoice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnVoice", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnVoice" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnVoice" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnVoice" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnVoice" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnVoice" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnVoice" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnVoice" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModule_OnVoice" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  (arg1)->OnVoice((CNick const &)*arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnDevoice2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnDevoice2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnDevoice2" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnDevoice2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnDevoice2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnDevoice2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnDevoice2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnDevoice2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModule_OnDevoice2" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  (arg1)->OnDevoice2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnDevoice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnDevoice", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnDevoice" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnDevoice" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnDevoice" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnDevoice" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnDevoice" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnDevoice" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnDevoice" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModule_OnDevoice" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  (arg1)->OnDevoice((CNick const &)*arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnMode2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CChan *arg3 = 0 ;
  char arg4 ;
  CString *arg5 = 0 ;
  bool arg6 ;
  bool arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  char val4 ;
  int ecode4 = 0 ;
  int res5 = SWIG_OLDOBJ ;
  bool val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[7] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnMode2", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnMode2" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnMode2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnMode2" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnMode2" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  ecode4 = SWIG_AsVal_char(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModule_OnMode2" "', argument " "4"" of type '" "char""'");
  } 
  arg4 = static_cast< char >(val4);
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModule_OnMode2" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnMode2" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CModule_OnMode2" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  ecode7 = SWIG_AsVal_bool(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CModule_OnMode2" "', argument " "7"" of type '" "bool""'");
  } 
  arg7 = static_cast< bool >(val7);
  (arg1)->OnMode2((CNick const *)arg2,*arg3,arg4,(CString const &)*arg5,arg6,arg7);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnMode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  char arg4 ;
  CString *arg5 = 0 ;
  bool arg6 ;
  bool arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  char val4 ;
  int ecode4 = 0 ;
  int res5 = SWIG_OLDOBJ ;
  bool val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[7] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnMode", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnMode" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnMode" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnMode" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnMode" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnMode" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  ecode4 = SWIG_AsVal_char(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModule_OnMode" "', argument " "4"" of type '" "char""'");
  } 
  arg4 = static_cast< char >(val4);
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModule_OnMode" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnMode" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CModule_OnMode" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  ecode7 = SWIG_AsVal_bool(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CModule_OnMode" "', argument " "7"" of type '" "bool""'");
  } 
  arg7 = static_cast< bool >(val7);
  (arg1)->OnMode((CNick const &)*arg2,*arg3,arg4,(CString const &)*arg5,arg6,arg7);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnRawMode2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnRawMode2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnRawMode2" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnRawMode2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnRawMode2" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnRawMode2" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnRawMode2" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnRawMode2" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModule_OnRawMode2" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnRawMode2" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  (arg1)->OnRawMode2((CNick const *)arg2,*arg3,(CString const &)*arg4,(CString const &)*arg5);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnRawMode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnRawMode", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnRawMode" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnRawMode" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnRawMode" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnRawMode" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnRawMode" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnRawMode" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnRawMode" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModule_OnRawMode" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnRawMode" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  (arg1)->OnRawMode((CNick const &)*arg2,*arg3,(CString const &)*arg4,(CString const &)*arg5);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnRaw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnRaw", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnRaw" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  result = (CModule::EModRet)(arg1)->OnRaw(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnRawMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnRawMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnRawMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnRawMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnNumericMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNumericMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnNumericMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnNumericMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNumericMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnNumericMessage" "', argument " "2"" of type '" "CNumericMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnNumericMessage" "', argument " "2"" of type '" "CNumericMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNumericMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnNumericMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnStatusCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnStatusCommand", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnStatusCommand" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sCommand");
    }
  }
  result = (CModule::EModRet)(arg1)->OnStatusCommand(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnModCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnModCommand", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnModCommand" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnModCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnModCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->OnModCommand((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUnknownModCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUnknownModCommand", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUnknownModCommand" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUnknownModCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnUnknownModCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->OnUnknownModCommand((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnModNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnModNotice", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnModNotice" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->OnModNotice((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnModCTCP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnModCTCP", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnModCTCP" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnModCTCP" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnModCTCP" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->OnModCTCP((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnQuitMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CQuitMessage *arg2 = 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnQuitMessage", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnQuitMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CQuitMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnQuitMessage" "', argument " "2"" of type '" "CQuitMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnQuitMessage" "', argument " "2"" of type '" "CQuitMessage &""'"); 
  }
  arg2 = reinterpret_cast< CQuitMessage * >(argp2);
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnQuitMessage" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnQuitMessage" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->OnQuitMessage(*arg2,(std::vector< CChan *,std::allocator< CChan * > > const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnQuit(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnQuit", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnQuit" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnQuit" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnQuit" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnQuit" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnQuit" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnQuit" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnQuit" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->OnQuit((CNick const &)*arg2,(CString const &)*arg3,(std::vector< CChan *,std::allocator< CChan * > > const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnNickMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNickMessage *arg2 = 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnNickMessage", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnNickMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNickMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnNickMessage" "', argument " "2"" of type '" "CNickMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnNickMessage" "', argument " "2"" of type '" "CNickMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNickMessage * >(argp2);
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnNickMessage" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnNickMessage" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->OnNickMessage(*arg2,(std::vector< CChan *,std::allocator< CChan * > > const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnNick", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnNick" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnNick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnNick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnNick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnNick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnNick" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnNick" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->OnNick((CNick const &)*arg2,(CString const &)*arg3,(std::vector< CChan *,std::allocator< CChan * > > const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnKickMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CKickMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnKickMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnKickMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CKickMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnKickMessage" "', argument " "2"" of type '" "CKickMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnKickMessage" "', argument " "2"" of type '" "CKickMessage &""'"); 
  }
  arg2 = reinterpret_cast< CKickMessage * >(argp2);
  (arg1)->OnKickMessage(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnKick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  CChan *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnKick", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnKick" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnKick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnKick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnKick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnKick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnKick" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnKick" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModule_OnKick" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnKick" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  (arg1)->OnKick((CNick const &)*arg2,(CString const &)*arg3,*arg4,(CString const &)*arg5);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnJoining(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CChan *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnJoining", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnJoining" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnJoining" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnJoining" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  result = (CModule::EModRet)(arg1)->OnJoining(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnJoinMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CJoinMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnJoinMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnJoinMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CJoinMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnJoinMessage" "', argument " "2"" of type '" "CJoinMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnJoinMessage" "', argument " "2"" of type '" "CJoinMessage &""'"); 
  }
  arg2 = reinterpret_cast< CJoinMessage * >(argp2);
  (arg1)->OnJoinMessage(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnJoin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnJoin", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnJoin" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnJoin" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnJoin" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnJoin" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnJoin" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  (arg1)->OnJoin((CNick const &)*arg2,*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPartMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CPartMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPartMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPartMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CPartMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPartMessage" "', argument " "2"" of type '" "CPartMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPartMessage" "', argument " "2"" of type '" "CPartMessage &""'"); 
  }
  arg2 = reinterpret_cast< CPartMessage * >(argp2);
  (arg1)->OnPartMessage(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPart(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPart", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPart" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPart" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPart" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnPart" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPart" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnPart" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPart" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->OnPart((CNick const &)*arg2,*arg3,(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnInviteMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CInviteMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnInviteMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnInviteMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CInviteMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnInviteMessage" "', argument " "2"" of type '" "CInviteMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnInviteMessage" "', argument " "2"" of type '" "CInviteMessage &""'"); 
  }
  arg2 = reinterpret_cast< CInviteMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnInviteMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnInvite(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnInvite", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnInvite" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnInvite" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnInvite" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnInvite" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnInvite" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (CModule::EModRet)(arg1)->OnInvite((CNick const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanBufferStarting(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CChan *arg2 = 0 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanBufferStarting", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanBufferStarting" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanBufferStarting" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanBufferStarting" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnChanBufferStarting" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanBufferStarting" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (CModule::EModRet)(arg1)->OnChanBufferStarting(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanBufferEnding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CChan *arg2 = 0 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanBufferEnding", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanBufferEnding" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanBufferEnding" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanBufferEnding" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnChanBufferEnding" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanBufferEnding" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (CModule::EModRet)(arg1)->OnChanBufferEnding(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanBufferPlayMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanBufferPlayMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanBufferPlayMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanBufferPlayMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanBufferPlayMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnChanBufferPlayMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanBufferPlayLine2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CChan *arg2 = 0 ;
  CClient *arg3 = 0 ;
  CString *arg4 = 0 ;
  timeval *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject *swig_obj[5] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanBufferPlayLine2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanBufferPlayLine2" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanBufferPlayLine2" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanBufferPlayLine2" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnChanBufferPlayLine2" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanBufferPlayLine2" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sLine");
    }
  }
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_timeval,  0  | 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModule_OnChanBufferPlayLine2" "', argument " "5"" of type '" "timeval const &""'"); 
  }
  if (!argp5) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanBufferPlayLine2" "', argument " "5"" of type '" "timeval const &""'"); 
  }
  arg5 = reinterpret_cast< timeval * >(argp5);
  result = (CModule::EModRet)(arg1)->OnChanBufferPlayLine2(*arg2,*arg3,*arg4,(timeval const &)*arg5);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanBufferPlayLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CChan *arg2 = 0 ;
  CClient *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanBufferPlayLine", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanBufferPlayLine" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanBufferPlayLine" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanBufferPlayLine" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnChanBufferPlayLine" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanBufferPlayLine" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sLine");
    }
  }
  result = (CModule::EModRet)(arg1)->OnChanBufferPlayLine(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivBufferStarting(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CQuery *arg2 = 0 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivBufferStarting", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivBufferStarting" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CQuery,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivBufferStarting" "', argument " "2"" of type '" "CQuery &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivBufferStarting" "', argument " "2"" of type '" "CQuery &""'"); 
  }
  arg2 = reinterpret_cast< CQuery * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnPrivBufferStarting" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivBufferStarting" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (CModule::EModRet)(arg1)->OnPrivBufferStarting(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivBufferEnding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CQuery *arg2 = 0 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivBufferEnding", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivBufferEnding" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CQuery,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivBufferEnding" "', argument " "2"" of type '" "CQuery &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivBufferEnding" "', argument " "2"" of type '" "CQuery &""'"); 
  }
  arg2 = reinterpret_cast< CQuery * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnPrivBufferEnding" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivBufferEnding" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (CModule::EModRet)(arg1)->OnPrivBufferEnding(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivBufferPlayMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivBufferPlayMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivBufferPlayMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivBufferPlayMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivBufferPlayMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnPrivBufferPlayMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivBufferPlayLine2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CClient *arg2 = 0 ;
  CString *arg3 = 0 ;
  timeval *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivBufferPlayLine2", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivBufferPlayLine2" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivBufferPlayLine2" "', argument " "2"" of type '" "CClient &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivBufferPlayLine2" "', argument " "2"" of type '" "CClient &""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sLine");
    }
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_timeval,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnPrivBufferPlayLine2" "', argument " "4"" of type '" "timeval const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivBufferPlayLine2" "', argument " "4"" of type '" "timeval const &""'"); 
  }
  arg4 = reinterpret_cast< timeval * >(argp4);
  result = (CModule::EModRet)(arg1)->OnPrivBufferPlayLine2(*arg2,*arg3,(timeval const &)*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivBufferPlayLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CClient *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivBufferPlayLine", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivBufferPlayLine" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivBufferPlayLine" "', argument " "2"" of type '" "CClient &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivBufferPlayLine" "', argument " "2"" of type '" "CClient &""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sLine");
    }
  }
  result = (CModule::EModRet)(arg1)->OnPrivBufferPlayLine(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnClientLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnClientLogin" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->OnClientLogin();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnClientDisconnect(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnClientDisconnect" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->OnClientDisconnect();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserRaw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserRaw", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserRaw" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserRaw(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserRawMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserRawMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserRawMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUserRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnUserRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserRawMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserCTCPReplyMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserCTCPReplyMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserCTCPReplyMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUserCTCPReplyMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnUserCTCPReplyMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserCTCPReplyMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserCTCPReply(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserCTCPReply", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserCTCPReply" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserCTCPReply(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserCTCPMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserCTCPMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserCTCPMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUserCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnUserCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserCTCPMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserCTCP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserCTCP", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserCTCP" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserCTCP(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserActionMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CActionMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserActionMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserActionMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CActionMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUserActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnUserActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  arg2 = reinterpret_cast< CActionMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserActionMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserAction(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserAction", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserAction" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserAction(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserTextMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CTextMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserTextMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserTextMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTextMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUserTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnUserTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTextMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserTextMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserMsg", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserMsg" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserMsg(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserNoticeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNoticeMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserNoticeMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserNoticeMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNoticeMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUserNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnUserNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNoticeMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserNoticeMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserNotice", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserNotice" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserNotice(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserJoinMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CJoinMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserJoinMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserJoinMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CJoinMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUserJoinMessage" "', argument " "2"" of type '" "CJoinMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnUserJoinMessage" "', argument " "2"" of type '" "CJoinMessage &""'"); 
  }
  arg2 = reinterpret_cast< CJoinMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserJoinMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserJoin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserJoin", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserJoin" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sChannel");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sKey");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserJoin(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserPartMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CPartMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserPartMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserPartMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CPartMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUserPartMessage" "', argument " "2"" of type '" "CPartMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnUserPartMessage" "', argument " "2"" of type '" "CPartMessage &""'"); 
  }
  arg2 = reinterpret_cast< CPartMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserPartMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserPart(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserPart", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserPart" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sChannel");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserPart(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserTopicMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CTopicMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserTopicMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserTopicMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTopicMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUserTopicMessage" "', argument " "2"" of type '" "CTopicMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnUserTopicMessage" "', argument " "2"" of type '" "CTopicMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTopicMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserTopicMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserTopic(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserTopic", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserTopic" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sChannel");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sTopic");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserTopic(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserTopicRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserTopicRequest", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserTopicRequest" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sChannel");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserTopicRequest(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserQuitMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CQuitMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserQuitMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserQuitMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CQuitMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUserQuitMessage" "', argument " "2"" of type '" "CQuitMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnUserQuitMessage" "', argument " "2"" of type '" "CQuitMessage &""'"); 
  }
  arg2 = reinterpret_cast< CQuitMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserQuitMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserQuit(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserQuit", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserQuit" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserQuit(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnCTCPReplyMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnCTCPReplyMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnCTCPReplyMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnCTCPReplyMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnCTCPReplyMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnCTCPReplyMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnCTCPReply(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnCTCPReply", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnCTCPReply" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnCTCPReply" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnCTCPReply" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnCTCPReply(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivCTCPMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivCTCPMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivCTCPMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnPrivCTCPMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivCTCP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivCTCP", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivCTCP" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivCTCP" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivCTCP" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnPrivCTCP(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanCTCPMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanCTCPMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanCTCPMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnChanCTCPMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanCTCP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanCTCP", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanCTCP" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanCTCP" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanCTCP" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnChanCTCP" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanCTCP" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnChanCTCP(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivActionMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CActionMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivActionMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivActionMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CActionMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  arg2 = reinterpret_cast< CActionMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnPrivActionMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivAction(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivAction", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivAction" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivAction" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivAction" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnPrivAction(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanActionMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CActionMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanActionMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanActionMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CActionMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  arg2 = reinterpret_cast< CActionMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnChanActionMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanAction(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanAction", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanAction" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanAction" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanAction" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnChanAction" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanAction" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnChanAction(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivTextMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CTextMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivTextMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivTextMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTextMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTextMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnPrivTextMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivMsg", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivMsg" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivMsg" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivMsg" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnPrivMsg(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanTextMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CTextMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanTextMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanTextMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTextMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTextMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnChanTextMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanMsg", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanMsg" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanMsg" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanMsg" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnChanMsg" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanMsg" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnChanMsg(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivNoticeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNoticeMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivNoticeMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivNoticeMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNoticeMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNoticeMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnPrivNoticeMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivNotice", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivNotice" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivNotice" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivNotice" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnPrivNotice(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanNoticeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNoticeMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanNoticeMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanNoticeMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNoticeMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNoticeMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnChanNoticeMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanNotice", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanNotice" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanNotice" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanNotice" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnChanNotice" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanNotice" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnChanNotice(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnTopicMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CTopicMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnTopicMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnTopicMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTopicMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnTopicMessage" "', argument " "2"" of type '" "CTopicMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnTopicMessage" "', argument " "2"" of type '" "CTopicMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTopicMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnTopicMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnTopic(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnTopic", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnTopic" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnTopic" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnTopic" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnTopic" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnTopic" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sTopic");
    }
  }
  result = (CModule::EModRet)(arg1)->OnTopic(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnServerCapAvailable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnServerCapAvailable", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnServerCapAvailable" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnServerCapAvailable" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnServerCapAvailable" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->OnServerCapAvailable((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnServerCap302Available(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnServerCap302Available", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnServerCap302Available" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnServerCap302Available" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnServerCap302Available" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnServerCap302Available" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnServerCap302Available" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->OnServerCap302Available((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnServerCapResult(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnServerCapResult", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnServerCapResult" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnServerCapResult" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnServerCapResult" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CModule_OnServerCapResult" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->OnServerCapResult((CString const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnTimerAutoJoin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CChan *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnTimerAutoJoin", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnTimerAutoJoin" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnTimerAutoJoin" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnTimerAutoJoin" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  result = (CModule::EModRet)(arg1)->OnTimerAutoJoin(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnAddNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CIRCNetwork *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnAddNetwork", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnAddNetwork" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CIRCNetwork,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnAddNetwork" "', argument " "2"" of type '" "CIRCNetwork &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnAddNetwork" "', argument " "2"" of type '" "CIRCNetwork &""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sErrorRet");
    }
  }
  result = (CModule::EModRet)(arg1)->OnAddNetwork(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnDeleteNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CIRCNetwork *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnDeleteNetwork", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnDeleteNetwork" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CIRCNetwork,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnDeleteNetwork" "', argument " "2"" of type '" "CIRCNetwork &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnDeleteNetwork" "', argument " "2"" of type '" "CIRCNetwork &""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  result = (CModule::EModRet)(arg1)->OnDeleteNetwork(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnSendToClientMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnSendToClientMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnSendToClientMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnSendToClientMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnSendToClientMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnSendToClientMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnSendToClient(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnSendToClient", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnSendToClient" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnSendToClient" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnSendToClient" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (CModule::EModRet)(arg1)->OnSendToClient(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnSendToIRCMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnSendToIRCMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnSendToIRCMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnSendToIRCMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnSendToIRCMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnSendToIRCMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnSendToIRC(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnSendToIRC", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnSendToIRC" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  result = (CModule::EModRet)(arg1)->OnSendToIRC(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUserTagMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CTargetMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUserTagMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUserTagMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTargetMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUserTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnUserTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTargetMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserTagMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnChanTagMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CTargetMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnChanTagMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnChanTagMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTargetMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnChanTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnChanTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTargetMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnChanTagMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnPrivTagMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CTargetMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnPrivTagMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnPrivTagMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTargetMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnPrivTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnPrivTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTargetMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnPrivTagMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetDLL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  ModHandle result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetDLL" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (ModHandle)(arg1)->GetDLL();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_PutIRC__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_PutIRC" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_PutIRC" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_PutIRC" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PutIRC((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_PutIRC__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_PutIRC" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_PutIRC" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_PutIRC" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (bool)(arg1)->PutIRC((CMessage const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_PutIRC(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModule_PutIRC", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModule_PutIRC__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModule_PutIRC__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModule_PutIRC'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModule::PutIRC(CString const &)\n"
    "    CModule::PutIRC(CMessage const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModule_PutUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_PutUser", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_PutUser" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PutUser((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_PutStatus(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_PutStatus", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_PutStatus" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PutStatus((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_PutModule__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_PutModule" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PutModule((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_PutModule__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CTable *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned int result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_PutModule" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTable,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_PutModule" "', argument " "2"" of type '" "CTable const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_PutModule" "', argument " "2"" of type '" "CTable const &""'"); 
  }
  arg2 = reinterpret_cast< CTable * >(argp2);
  result = (unsigned int)(arg1)->PutModule((CTable const &)*arg2);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_PutModule(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModule_PutModule", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CTable, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModule_PutModule__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModule_PutModule__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModule_PutModule'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModule::PutModule(CString const &)\n"
    "    CModule::PutModule(CTable const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModule_PutModNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_PutModNotice", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_PutModNotice" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_PutModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_PutModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PutModNotice((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetModName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetModName" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (CString *) &((CModule const *)arg1)->GetModName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetModNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetModNick" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = ((CModule const *)arg1)->GetModNick();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetModDataDir(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetModDataDir" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (CString *) &((CModule const *)arg1)->GetModDataDir();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddTimer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CTimer *arg2 = (CTimer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_AddTimer" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CTimer, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_AddTimer" "', argument " "2"" of type '" "CTimer *""'"); 
  }
  arg2 = reinterpret_cast< CTimer * >(argp2);
  result = (bool)(arg1)->AddTimer(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddTimer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  FPTimer_t arg2 = (FPTimer_t) 0 ;
  CString *arg3 = 0 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  CString *arg6 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  int res6 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_AddTimer" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_CModule_p_CFPTimer__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "CModule_AddTimer" "', argument " "2"" of type '" "FPTimer_t""'"); 
    }
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_AddTimer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddTimer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModule_AddTimer" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModule_AddTimer" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  {
    CString *ptr = (CString *)0;
    res6 = SWIG_AsPtr_CString(swig_obj[5], &ptr);
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "CModule_AddTimer" "', argument " "6"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddTimer" "', argument " "6"" of type '" "CString const &""'"); 
    }
    arg6 = ptr;
  }
  result = (bool)(arg1)->AddTimer(arg2,(CString const &)*arg3,arg4,arg5,(CString const &)*arg6);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res6)) delete arg6;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res6)) delete arg6;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddTimer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  FPTimer_t arg2 = (FPTimer_t) 0 ;
  CString *arg3 = 0 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_AddTimer" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_CModule_p_CFPTimer__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "CModule_AddTimer" "', argument " "2"" of type '" "FPTimer_t""'"); 
    }
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_AddTimer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddTimer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModule_AddTimer" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModule_AddTimer" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  result = (bool)(arg1)->AddTimer(arg2,(CString const &)*arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddTimer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  FPTimer_t arg2 = (FPTimer_t) 0 ;
  CString *arg3 = 0 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_AddTimer" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(swig_obj[1], (void**)(&arg2), SWIGTYPE_p_f_p_CModule_p_CFPTimer__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "CModule_AddTimer" "', argument " "2"" of type '" "FPTimer_t""'"); 
    }
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_AddTimer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddTimer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModule_AddTimer" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  result = (bool)(arg1)->AddTimer(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddTimer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[7] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModule_AddTimer", 0, 6, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CTimer, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModule_AddTimer__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *ptr = 0;
      int res = SWIG_ConvertFunctionPtr(argv[1], &ptr, SWIGTYPE_p_f_p_CModule_p_CFPTimer__void);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CModule_AddTimer__SWIG_3(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *ptr = 0;
      int res = SWIG_ConvertFunctionPtr(argv[1], &ptr, SWIGTYPE_p_f_p_CModule_p_CFPTimer__void);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_unsigned_SS_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_CModule_AddTimer__SWIG_2(self, argc, argv);
            }
          }
        }
      }
    }
  }
  if (argc == 6) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *ptr = 0;
      int res = SWIG_ConvertFunctionPtr(argv[1], &ptr, SWIGTYPE_p_f_p_CModule_p_CFPTimer__void);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_unsigned_SS_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              int res = SWIG_AsPtr_CString(argv[5], (CString**)(0));
              _v = SWIG_CheckState(res);
              if (_v) {
                return _wrap_CModule_AddTimer__SWIG_1(self, argc, argv);
              }
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModule_AddTimer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModule::AddTimer(CTimer *)\n"
    "    CModule::AddTimer(FPTimer_t,CString const &,unsigned int,unsigned int,CString const &)\n"
    "    CModule::AddTimer(FPTimer_t,CString const &,unsigned int,unsigned int)\n"
    "    CModule::AddTimer(FPTimer_t,CString const &,unsigned int)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModule_RemTimer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CTimer *arg2 = (CTimer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_RemTimer" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CTimer, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_RemTimer" "', argument " "2"" of type '" "CTimer *""'"); 
  }
  arg2 = reinterpret_cast< CTimer * >(argp2);
  result = (bool)(arg1)->RemTimer(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_RemTimer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_RemTimer" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_RemTimer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_RemTimer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->RemTimer((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_RemTimer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModule_RemTimer", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CTimer, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModule_RemTimer__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModule_RemTimer__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModule_RemTimer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModule::RemTimer(CTimer *)\n"
    "    CModule::RemTimer(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModule_UnlinkTimer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CTimer *arg2 = (CTimer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_UnlinkTimer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_UnlinkTimer" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CTimer, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_UnlinkTimer" "', argument " "2"" of type '" "CTimer *""'"); 
  }
  arg2 = reinterpret_cast< CTimer * >(argp2);
  result = (bool)(arg1)->UnlinkTimer(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_FindTimer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CTimer *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_FindTimer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_FindTimer" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_FindTimer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_FindTimer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CTimer *)(arg1)->FindTimer((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTimer, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_BeginTimers(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::set< CTimer *,std::less< CTimer * >,std::allocator< CTimer * > >::const_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_BeginTimers" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = ((CModule const *)arg1)->BeginTimers();
  resultobj = SWIG_NewPointerObj((new std::set< CTimer *,std::less< CTimer * >,std::allocator< CTimer * > >::const_iterator(result)), SWIGTYPE_p_std__setT_CTimer_p_std__lessT_CTimer_p_t_std__allocatorT_CTimer_p_t_t__const_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_EndTimers(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::set< CTimer *,std::less< CTimer * >,std::allocator< CTimer * > >::const_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_EndTimers" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = ((CModule const *)arg1)->EndTimers();
  resultobj = SWIG_NewPointerObj((new std::set< CTimer *,std::less< CTimer * >,std::allocator< CTimer * > >::const_iterator(result)), SWIGTYPE_p_std__setT_CTimer_p_std__lessT_CTimer_p_t_std__allocatorT_CTimer_p_t_t__const_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_ListTimers(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_ListTimers" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->ListTimers();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddSocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CSocket *arg2 = (CSocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_AddSocket", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_AddSocket" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_AddSocket" "', argument " "2"" of type '" "CSocket *""'"); 
  }
  arg2 = reinterpret_cast< CSocket * >(argp2);
  result = (bool)(arg1)->AddSocket(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_RemSocket__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CSocket *arg2 = (CSocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_RemSocket" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_RemSocket" "', argument " "2"" of type '" "CSocket *""'"); 
  }
  arg2 = reinterpret_cast< CSocket * >(argp2);
  result = (bool)(arg1)->RemSocket(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_RemSocket__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_RemSocket" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_RemSocket" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_RemSocket" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->RemSocket((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_RemSocket(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModule_RemSocket", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CSocket, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModule_RemSocket__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModule_RemSocket__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModule_RemSocket'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModule::RemSocket(CSocket *)\n"
    "    CModule::RemSocket(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModule_UnlinkSocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CSocket *arg2 = (CSocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_UnlinkSocket", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_UnlinkSocket" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CSocket, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_UnlinkSocket" "', argument " "2"" of type '" "CSocket *""'"); 
  }
  arg2 = reinterpret_cast< CSocket * >(argp2);
  result = (bool)(arg1)->UnlinkSocket(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_FindSocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CSocket *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_FindSocket", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_FindSocket" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_FindSocket" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_FindSocket" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CSocket *)(arg1)->FindSocket((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSocket, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_BeginSockets(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::set< CSocket *,std::less< CSocket * >,std::allocator< CSocket * > >::const_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_BeginSockets" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = ((CModule const *)arg1)->BeginSockets();
  resultobj = SWIG_NewPointerObj((new std::set< CSocket *,std::less< CSocket * >,std::allocator< CSocket * > >::const_iterator(result)), SWIGTYPE_p_std__setT_CSocket_p_std__lessT_CSocket_p_t_std__allocatorT_CSocket_p_t_t__const_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_EndSockets(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::set< CSocket *,std::less< CSocket * >,std::allocator< CSocket * > >::const_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_EndSockets" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = ((CModule const *)arg1)->EndSockets();
  resultobj = SWIG_NewPointerObj((new std::set< CSocket *,std::less< CSocket * >,std::allocator< CSocket * > >::const_iterator(result)), SWIGTYPE_p_std__setT_CSocket_p_std__lessT_CSocket_p_t_std__allocatorT_CSocket_p_t_t__const_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_ListSockets(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_ListSockets" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->ListSockets();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddHelpCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_AddHelpCommand" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->AddHelpCommand();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddCommand__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CModCommand *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_AddCommand" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModCommand,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_AddCommand" "', argument " "2"" of type '" "CModCommand const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddCommand" "', argument " "2"" of type '" "CModCommand const &""'"); 
  }
  arg2 = reinterpret_cast< CModCommand * >(argp2);
  result = (bool)(arg1)->AddCommand((CModCommand const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddCommand__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CModCommand::ModCmdFunc arg3 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_AddCommand" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_AddCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    int res = SWIG_ConvertMember(swig_obj[2], SWIG_as_voidptr(&arg3), sizeof(arg3),SWIGTYPE_m_CModule__f_r_q_const__CString__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "CModule_AddCommand" "', argument " "3"" of type '" "CModCommand::ModCmdFunc""'"); 
    }
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_AddCommand" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddCommand" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModule_AddCommand" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddCommand" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  result = (bool)(arg1)->AddCommand((CString const &)*arg2,arg3,(CString const &)*arg4,(CString const &)*arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddCommand__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CModCommand::ModCmdFunc arg3 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_AddCommand" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_AddCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    int res = SWIG_ConvertMember(swig_obj[2], SWIG_as_voidptr(&arg3), sizeof(arg3),SWIGTYPE_m_CModule__f_r_q_const__CString__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "CModule_AddCommand" "', argument " "3"" of type '" "CModCommand::ModCmdFunc""'"); 
    }
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_AddCommand" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddCommand" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (bool)(arg1)->AddCommand((CString const &)*arg2,arg3,(CString const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddCommand__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CModCommand::ModCmdFunc arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_AddCommand" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_AddCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    int res = SWIG_ConvertMember(swig_obj[2], SWIG_as_voidptr(&arg3), sizeof(arg3),SWIGTYPE_m_CModule__f_r_q_const__CString__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "CModule_AddCommand" "', argument " "3"" of type '" "CModCommand::ModCmdFunc""'"); 
    }
  }
  result = (bool)(arg1)->AddCommand((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddCommand__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  COptionalTranslation *arg3 = 0 ;
  COptionalTranslation *arg4 = 0 ;
  SwigValueWrapper< std::function< void (CString const &) > > arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 ;
  int res5 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_AddCommand" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_AddCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_COptionalTranslation,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_AddCommand" "', argument " "3"" of type '" "COptionalTranslation const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddCommand" "', argument " "3"" of type '" "COptionalTranslation const &""'"); 
  }
  arg3 = reinterpret_cast< COptionalTranslation * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_COptionalTranslation,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_AddCommand" "', argument " "4"" of type '" "COptionalTranslation const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddCommand" "', argument " "4"" of type '" "COptionalTranslation const &""'"); 
  }
  arg4 = reinterpret_cast< COptionalTranslation * >(argp4);
  {
    res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_std__functionT_void_fCString_const_RF_t,  0  | 0);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModule_AddCommand" "', argument " "5"" of type '" "std::function< void (CString const &) >""'"); 
    }  
    if (!argp5) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddCommand" "', argument " "5"" of type '" "std::function< void (CString const &) >""'");
    } else {
      std::function< void (CString const &) > * temp = reinterpret_cast< std::function< void (CString const &) > * >(argp5);
      arg5 = *temp;
      if (SWIG_IsNewObj(res5)) delete temp;
    }
  }
  result = (bool)(arg1)->AddCommand((CString const &)*arg2,(COptionalTranslation const &)*arg3,(COptionalTranslation const &)*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddCommand(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModule_AddCommand", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CModCommand, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModule_AddCommand__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_m_CModule__f_r_q_const__CString__void, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CModule_AddCommand__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_m_CModule__f_r_q_const__CString__void, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CModule_AddCommand__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_m_CModule__f_r_q_const__CString__void, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            int res = SWIG_AsPtr_CString(argv[4], (CString**)(0));
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_CModule_AddCommand__SWIG_1(self, argc, argv);
            }
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_COptionalTranslation, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_ConvertPtr(argv[3], 0, SWIGTYPE_p_COptionalTranslation, SWIG_POINTER_NO_NULL | 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            int res = SWIG_ConvertPtr(argv[4], 0, SWIGTYPE_p_std__functionT_void_fCString_const_RF_t, SWIG_POINTER_NO_NULL | 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_CModule_AddCommand__SWIG_4(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModule_AddCommand'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModule::AddCommand(CModCommand const &)\n"
    "    CModule::AddCommand(CString const &,CModCommand::ModCmdFunc,CString const &,CString const &)\n"
    "    CModule::AddCommand(CString const &,CModCommand::ModCmdFunc,CString const &)\n"
    "    CModule::AddCommand(CString const &,CModCommand::ModCmdFunc)\n"
    "    CModule::AddCommand(CString const &,COptionalTranslation const &,COptionalTranslation const &,std::function< void (CString const &) >)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModule_RemCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_RemCommand", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_RemCommand" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_RemCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_RemCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->RemCommand((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_FindCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CModCommand *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_FindCommand", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_FindCommand" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_FindCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_FindCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CModCommand *)((CModule const *)arg1)->FindCommand((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModCommand, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_HandleCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_HandleCommand", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_HandleCommand" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_HandleCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_HandleCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->HandleCommand((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_HandleHelpCommand__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_HandleHelpCommand" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_HandleHelpCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_HandleHelpCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->HandleHelpCommand((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_HandleHelpCommand__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_HandleHelpCommand" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->HandleHelpCommand();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_HandleHelpCommand(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModule_HandleHelpCommand", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CModule_HandleHelpCommand__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModule_HandleHelpCommand__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModule_HandleHelpCommand'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModule::HandleHelpCommand(CString const &)\n"
    "    CModule::HandleHelpCommand()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModule_LoadRegistry(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_LoadRegistry" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (bool)(arg1)->LoadRegistry();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_SaveRegistry(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_SaveRegistry" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (bool)((CModule const *)arg1)->SaveRegistry();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_MoveRegistry(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_MoveRegistry", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_MoveRegistry" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_MoveRegistry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_MoveRegistry" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->MoveRegistry((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_SetNV__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_SetNV" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_SetNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_SetNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_SetNV" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_SetNV" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModule_SetNV" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->SetNV((CString const &)*arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_SetNV__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_SetNV" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_SetNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_SetNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_SetNV" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_SetNV" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->SetNV((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_SetNV(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModule_SetNV", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CModule_SetNV__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CModule_SetNV__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModule_SetNV'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModule::SetNV(CString const &,CString const &,bool)\n"
    "    CModule::SetNV(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModule_GetNV(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_GetNV", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetNV" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_GetNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_GetNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CModule const *)arg1)->GetNV((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_HasNV(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_HasNV", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_HasNV" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_HasNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_HasNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)((CModule const *)arg1)->HasNV((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_DelNV__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_DelNV" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_DelNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_DelNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CModule_DelNV" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)(arg1)->DelNV((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_DelNV__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_DelNV" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_DelNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_DelNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->DelNV((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_FindNV(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  MCString::iterator result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_FindNV", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_FindNV" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_FindNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_FindNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->FindNV((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const MCString::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_EndNV(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  MCString::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_EndNV" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (arg1)->EndNV();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const MCString::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_BeginNV(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  MCString::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_BeginNV" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (arg1)->BeginNV();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const MCString::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_DelNV__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  MCString::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_DelNV" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "CModule_DelNV" "', argument " "2"" of type '" "MCString::iterator""'");
  } else {
    swig::SwigPyIterator_T<MCString::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<MCString::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "CModule_DelNV" "', argument " "2"" of type '" "MCString::iterator""'");
    }
  }
  (arg1)->DelNV(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_DelNV(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModule_DelNV", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<MCString::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_CModule_DelNV__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModule_DelNV__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CModule_DelNV__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModule_DelNV'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModule::DelNV(CString const &,bool)\n"
    "    CModule::DelNV(CString const &)\n"
    "    CModule::DelNV(MCString::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModule_ClearNV__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_ClearNV" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CModule_ClearNV" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (bool)(arg1)->ClearNV(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_ClearNV__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_ClearNV" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (bool)(arg1)->ClearNV();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_ClearNV(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModule_ClearNV", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CModule_ClearNV__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CModule_ClearNV__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModule_ClearNV'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModule::ClearNV(bool)\n"
    "    CModule::ClearNV()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModule_GetSavePath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetSavePath" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (CString *) &((CModule const *)arg1)->GetSavePath();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_ExpandString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_ExpandString" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_ExpandString" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_ExpandString" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CModule const *)arg1)->ExpandString((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_ExpandString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_ExpandString" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_ExpandString" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_ExpandString" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sRet");
    }
  }
  result = (CString *) &((CModule const *)arg1)->ExpandString((CString const &)*arg2,*arg3);
  {
    if (result) {
      resultobj = CPyRetString::wrap(*result);
    } else {
      resultobj = Py_None;
      Py_INCREF(Py_None);
    }
  }
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_ExpandString(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModule_ExpandString", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModule_ExpandString__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModule, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          return _wrap_CModule_ExpandString__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModule_ExpandString'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModule::ExpandString(CString const &) const\n"
    "    CModule::ExpandString(CString const &,CString &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModule_SetType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CModInfo::EModuleType arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_SetType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_SetType" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CModule_SetType" "', argument " "2"" of type '" "CModInfo::EModuleType""'");
  } 
  arg2 = static_cast< CModInfo::EModuleType >(val2);
  (arg1)->SetType(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_SetDescription(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_SetDescription", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_SetDescription" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_SetDescription" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_SetDescription" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetDescription((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_SetModPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_SetModPath", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_SetModPath" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_SetModPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_SetModPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetModPath((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_SetArgs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_SetArgs", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_SetArgs" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_SetArgs" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_SetArgs" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetArgs((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CModInfo::EModuleType result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetType" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (CModInfo::EModuleType)((CModule const *)arg1)->GetType();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetDescription(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetDescription" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (CString *) &((CModule const *)arg1)->GetDescription();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetArgs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetArgs" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (CString *) &((CModule const *)arg1)->GetArgs();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetModPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetModPath" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (CString *) &((CModule const *)arg1)->GetModPath();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CUser *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetUser" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (CUser *)((CModule const *)arg1)->GetUser();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CUser, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CIRCNetwork *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetNetwork" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (CIRCNetwork *)((CModule const *)arg1)->GetNetwork();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetClient(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CClient *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetClient" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (CClient *)((CModule const *)arg1)->GetClient();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CClient, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_GetManager(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CSockManager *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_GetManager" "', argument " "1"" of type '" "CModule const *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (CSockManager *)((CModule const *)arg1)->GetManager();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSockManager, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnAddUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CUser *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnAddUser", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnAddUser" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnAddUser" "', argument " "2"" of type '" "CUser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnAddUser" "', argument " "2"" of type '" "CUser &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sErrorRet");
    }
  }
  result = (CModule::EModRet)(arg1)->OnAddUser(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnDeleteUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CUser *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnDeleteUser", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnDeleteUser" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnDeleteUser" "', argument " "2"" of type '" "CUser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnDeleteUser" "', argument " "2"" of type '" "CUser &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  result = (CModule::EModRet)(arg1)->OnDeleteUser(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnClientConnect(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CZNCSock *arg2 = (CZNCSock *) 0 ;
  CString *arg3 = 0 ;
  unsigned short arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  unsigned short val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnClientConnect", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnClientConnect" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnClientConnect" "', argument " "2"" of type '" "CZNCSock *""'"); 
  }
  arg2 = reinterpret_cast< CZNCSock * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnClientConnect" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnClientConnect" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_unsigned_SS_short(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModule_OnClientConnect" "', argument " "4"" of type '" "unsigned short""'");
  } 
  arg4 = static_cast< unsigned short >(val4);
  (arg1)->OnClientConnect(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnLoginAttempt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  std::shared_ptr< CAuthBase > arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnLoginAttempt", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnLoginAttempt" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    int newmem = 0;
    res2 = SWIG_ConvertPtrAndOwn(swig_obj[1], &argp2, SWIGTYPE_p_std__shared_ptrT_CAuthBase_t,  0 , &newmem);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnLoginAttempt" "', argument " "2"" of type '" "std::shared_ptr< CAuthBase >""'");
    }
    if (argp2) arg2 = *(reinterpret_cast< std::shared_ptr< CAuthBase > * >(argp2));
    if (newmem & SWIG_CAST_NEW_MEMORY) delete reinterpret_cast< std::shared_ptr< CAuthBase > * >(argp2);
  }
  result = (CModule::EModRet)(arg1)->OnLoginAttempt(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnFailedLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnFailedLogin", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnFailedLogin" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnFailedLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnFailedLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnFailedLogin" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnFailedLogin" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->OnFailedLogin((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUnknownUserRaw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUnknownUserRaw", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUnknownUserRaw" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUnknownUserRaw" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sLine");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUnknownUserRaw(arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnUnknownUserRawMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnUnknownUserRawMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnUnknownUserRawMessage" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnUnknownUserRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnUnknownUserRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUnknownUserRawMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnClientAttached(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnClientAttached" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->OnClientAttached();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnClientDetached(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnClientDetached" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->OnClientDetached();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnClientCapLs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  SCString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnClientCapLs", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnClientCapLs" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnClientCapLs" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__setT_CString_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnClientCapLs" "', argument " "3"" of type '" "SCString &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnClientCapLs" "', argument " "3"" of type '" "SCString &""'"); 
  }
  arg3 = reinterpret_cast< SCString * >(argp3);
  (arg1)->OnClientCapLs(arg2,*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_IsClientCapSupported(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_IsClientCapSupported", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_IsClientCapSupported" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_IsClientCapSupported" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_IsClientCapSupported" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_IsClientCapSupported" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModule_IsClientCapSupported" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->IsClientCapSupported(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnClientCapRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnClientCapRequest", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnClientCapRequest" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnClientCapRequest" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnClientCapRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnClientCapRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModule_OnClientCapRequest" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  (arg1)->OnClientCapRequest(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnClientGetSASLMechanisms(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  SCString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnClientGetSASLMechanisms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnClientGetSASLMechanisms" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__setT_CString_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnClientGetSASLMechanisms" "', argument " "2"" of type '" "SCString &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnClientGetSASLMechanisms" "', argument " "2"" of type '" "SCString &""'"); 
  }
  arg2 = reinterpret_cast< SCString * >(argp2);
  (arg1)->OnClientGetSASLMechanisms(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnClientSASLServerInitialChallenge(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnClientSASLServerInitialChallenge", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnClientSASLServerInitialChallenge" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnClientSASLServerInitialChallenge" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnClientSASLServerInitialChallenge" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sResponse");
    }
  }
  result = (CModule::EModRet)(arg1)->OnClientSASLServerInitialChallenge((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnClientSASLAuthenticate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnClientSASLAuthenticate", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnClientSASLAuthenticate" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnClientSASLAuthenticate" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnClientSASLAuthenticate" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnClientSASLAuthenticate" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnClientSASLAuthenticate" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (CModule::EModRet)(arg1)->OnClientSASLAuthenticate((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnClientSASLAborted(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnClientSASLAborted" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->OnClientSASLAborted();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnModuleLoading(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CModInfo::EModuleType arg4 ;
  bool *arg5 = 0 ;
  CString *arg6 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject *swig_obj[6] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnModuleLoading", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnModuleLoading" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnModuleLoading" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnModuleLoading" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnModuleLoading" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnModuleLoading" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModule_OnModuleLoading" "', argument " "4"" of type '" "CModInfo::EModuleType""'");
  } 
  arg4 = static_cast< CModInfo::EModuleType >(val4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_bool,  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModule_OnModuleLoading" "', argument " "5"" of type '" "bool &""'"); 
  }
  if (!argp5) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnModuleLoading" "', argument " "5"" of type '" "bool &""'"); 
  }
  arg5 = reinterpret_cast< bool * >(argp5);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[5], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg6 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 6 sRetMsg");
    }
  }
  result = (CModule::EModRet)(arg1)->OnModuleLoading((CString const &)*arg2,(CString const &)*arg3,arg4,*arg5,*arg6);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnModuleUnloading(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CModule *arg2 = (CModule *) 0 ;
  bool *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnModuleUnloading", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnModuleUnloading" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnModuleUnloading" "', argument " "2"" of type '" "CModule *""'"); 
  }
  arg2 = reinterpret_cast< CModule * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_bool,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnModuleUnloading" "', argument " "3"" of type '" "bool &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnModuleUnloading" "', argument " "3"" of type '" "bool &""'"); 
  }
  arg3 = reinterpret_cast< bool * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sRetMsg");
    }
  }
  result = (CModule::EModRet)(arg1)->OnModuleUnloading(arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnGetModInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CModInfo *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[5] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnGetModInfo", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnGetModInfo" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnGetModInfo" "', argument " "2"" of type '" "CModInfo &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnGetModInfo" "', argument " "2"" of type '" "CModInfo &""'"); 
  }
  arg2 = reinterpret_cast< CModInfo * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_OnGetModInfo" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnGetModInfo" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_bool,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModule_OnGetModInfo" "', argument " "4"" of type '" "bool &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnGetModInfo" "', argument " "4"" of type '" "bool &""'"); 
  }
  arg4 = reinterpret_cast< bool * >(argp4);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[4], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg5 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 5 sRetMsg");
    }
  }
  result = (CModule::EModRet)(arg1)->OnGetModInfo(*arg2,(CString const &)*arg3,*arg4,*arg5);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_OnGetAvailableMods(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > *arg2 = 0 ;
  CModInfo::EModuleType arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_OnGetAvailableMods", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_OnGetAvailableMods" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__setT_CModInfo_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_OnGetAvailableMods" "', argument " "2"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_OnGetAvailableMods" "', argument " "2"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  arg2 = reinterpret_cast< std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > * >(argp2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CModule_OnGetAvailableMods" "', argument " "3"" of type '" "CModInfo::EModuleType""'");
  } 
  arg3 = static_cast< CModInfo::EModuleType >(val3);
  (arg1)->OnGetAvailableMods(*arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_InternalServerDependentCapsOnServerCap302Available(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_InternalServerDependentCapsOnServerCap302Available", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_InternalServerDependentCapsOnServerCap302Available" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_InternalServerDependentCapsOnServerCap302Available" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_InternalServerDependentCapsOnServerCap302Available" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_InternalServerDependentCapsOnServerCap302Available" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_InternalServerDependentCapsOnServerCap302Available" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->InternalServerDependentCapsOnServerCap302Available((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_InternalServerDependentCapsOnServerCapResult(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_InternalServerDependentCapsOnServerCapResult", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_InternalServerDependentCapsOnServerCapResult" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_InternalServerDependentCapsOnServerCapResult" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_InternalServerDependentCapsOnServerCapResult" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CModule_InternalServerDependentCapsOnServerCapResult" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->InternalServerDependentCapsOnServerCapResult((CString const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_InternalServerDependentCapsOnClientCapLs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  SCString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_InternalServerDependentCapsOnClientCapLs", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_InternalServerDependentCapsOnClientCapLs" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_InternalServerDependentCapsOnClientCapLs" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__setT_CString_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_InternalServerDependentCapsOnClientCapLs" "', argument " "3"" of type '" "SCString &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_InternalServerDependentCapsOnClientCapLs" "', argument " "3"" of type '" "SCString &""'"); 
  }
  arg3 = reinterpret_cast< SCString * >(argp3);
  (arg1)->InternalServerDependentCapsOnClientCapLs(arg2,*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_InternalServerDependentCapsIsClientCapSupported(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_InternalServerDependentCapsIsClientCapSupported", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_InternalServerDependentCapsIsClientCapSupported" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_InternalServerDependentCapsIsClientCapSupported" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_InternalServerDependentCapsIsClientCapSupported" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_InternalServerDependentCapsIsClientCapSupported" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModule_InternalServerDependentCapsIsClientCapSupported" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->InternalServerDependentCapsIsClientCapSupported(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_InternalServerDependentCapsOnClientCapRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_InternalServerDependentCapsOnClientCapRequest", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_InternalServerDependentCapsOnClientCapRequest" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_InternalServerDependentCapsOnClientCapRequest" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModule_InternalServerDependentCapsOnClientCapRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_InternalServerDependentCapsOnClientCapRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModule_InternalServerDependentCapsOnClientCapRequest" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  (arg1)->InternalServerDependentCapsOnClientCapRequest(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_InternalServerDependentCapsOnClientAttached(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_InternalServerDependentCapsOnClientAttached" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->InternalServerDependentCapsOnClientAttached();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_InternalServerDependentCapsOnClientDetached(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_InternalServerDependentCapsOnClientDetached" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->InternalServerDependentCapsOnClientDetached();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_InternalServerDependentCapsOnIRCConnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_InternalServerDependentCapsOnIRCConnected" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->InternalServerDependentCapsOnIRCConnected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_InternalServerDependentCapsOnIRCDisconnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_InternalServerDependentCapsOnIRCDisconnected" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  (arg1)->InternalServerDependentCapsOnIRCDisconnected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule___str__" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = CModule___str__(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_BeginNV_(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  MCString_iter result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_BeginNV_" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = CModule_BeginNV_(arg1);
  resultobj = SWIG_NewPointerObj((new MCString_iter(result)), SWIGTYPE_p_MCString_iter, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_ExistsNV(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_ExistsNV", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_ExistsNV" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_ExistsNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_ExistsNV" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)CModule_ExistsNV(arg1,(CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModule_AddServerDependentCapability(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  CString *arg2 = 0 ;
  PyObject *arg3 = (PyObject *) 0 ;
  PyObject *arg4 = (PyObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModule_AddServerDependentCapability", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModule_AddServerDependentCapability" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModule_AddServerDependentCapability" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModule_AddServerDependentCapability" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  arg3 = swig_obj[2];
  arg4 = swig_obj[3];
  CModule_AddServerDependentCapability(arg1,(CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *CModule_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CModule, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CModule_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CModules__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CModules *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CModules *)new CModules();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModules, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CModules(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CModules" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CModules__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModules *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CModules *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CModules,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CModules" "', argument " "1"" of type '" "CModules const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CModules" "', argument " "1"" of type '" "CModules const &""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (CModules *)new CModules((CModules const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModules, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CModules(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CModules", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CModules__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_CModules, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CModules__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CModules'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModules::CModules()\n"
    "    CModules::CModules(CModules const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModules_SetUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CUser *arg2 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_SetUser", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_SetUser" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_SetUser" "', argument " "2"" of type '" "CUser *""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  (arg1)->SetUser(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_SetNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_SetNetwork", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_SetNetwork" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_SetNetwork" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  (arg1)->SetNetwork(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_SetClient(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_SetClient", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_SetClient" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_SetClient" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  (arg1)->SetClient(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_GetUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CUser *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_GetUser" "', argument " "1"" of type '" "CModules const *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (CUser *)((CModules const *)arg1)->GetUser();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CUser, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_GetNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CIRCNetwork *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_GetNetwork" "', argument " "1"" of type '" "CModules const *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (CIRCNetwork *)((CModules const *)arg1)->GetNetwork();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_GetClient(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CClient *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_GetClient" "', argument " "1"" of type '" "CModules const *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (CClient *)((CModules const *)arg1)->GetClient();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CClient, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_UnloadAll(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_UnloadAll" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  (arg1)->UnloadAll();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnBoot(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnBoot" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (bool)(arg1)->OnBoot();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPreRehash(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPreRehash" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (bool)(arg1)->OnPreRehash();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPostRehash(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPostRehash" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (bool)(arg1)->OnPostRehash();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnIRCDisconnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnIRCDisconnected" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (bool)(arg1)->OnIRCDisconnected();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnIRCConnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnIRCConnected" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (bool)(arg1)->OnIRCConnected();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnIRCConnecting(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CIRCSock *arg2 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnIRCConnecting", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnIRCConnecting" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnIRCConnecting" "', argument " "2"" of type '" "CIRCSock *""'"); 
  }
  arg2 = reinterpret_cast< CIRCSock * >(argp2);
  result = (bool)(arg1)->OnIRCConnecting(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnIRCConnectionError(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CIRCSock *arg2 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnIRCConnectionError", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnIRCConnectionError" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnIRCConnectionError" "', argument " "2"" of type '" "CIRCSock *""'"); 
  }
  arg2 = reinterpret_cast< CIRCSock * >(argp2);
  result = (bool)(arg1)->OnIRCConnectionError(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnIRCRegistration(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnIRCRegistration", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnIRCRegistration" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sPass");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sNick");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sIdent");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[4], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg5 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 5 sRealName");
    }
  }
  result = (bool)(arg1)->OnIRCRegistration(*arg2,*arg3,*arg4,*arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnBroadcast(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnBroadcast", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnBroadcast" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sMessage");
    }
  }
  result = (bool)(arg1)->OnBroadcast(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanPermission3(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  char arg5 ;
  bool arg6 ;
  bool arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  char val5 ;
  int ecode5 = 0 ;
  bool val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[7] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanPermission3", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanPermission3" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanPermission3" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnChanPermission3" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanPermission3" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnChanPermission3" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanPermission3" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_char(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModules_OnChanPermission3" "', argument " "5"" of type '" "char""'");
  } 
  arg5 = static_cast< char >(val5);
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CModules_OnChanPermission3" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  ecode7 = SWIG_AsVal_bool(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CModules_OnChanPermission3" "', argument " "7"" of type '" "bool""'");
  } 
  arg7 = static_cast< bool >(val7);
  result = (bool)(arg1)->OnChanPermission3((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanPermission2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  unsigned char arg5 ;
  bool arg6 ;
  bool arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  unsigned char val5 ;
  int ecode5 = 0 ;
  bool val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[7] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanPermission2", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanPermission2" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanPermission2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnChanPermission2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanPermission2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnChanPermission2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanPermission2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_unsigned_SS_char(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModules_OnChanPermission2" "', argument " "5"" of type '" "unsigned char""'");
  } 
  arg5 = static_cast< unsigned char >(val5);
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CModules_OnChanPermission2" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  ecode7 = SWIG_AsVal_bool(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CModules_OnChanPermission2" "', argument " "7"" of type '" "bool""'");
  } 
  arg7 = static_cast< bool >(val7);
  result = (bool)(arg1)->OnChanPermission2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanPermission(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  unsigned char arg5 ;
  bool arg6 ;
  bool arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  unsigned char val5 ;
  int ecode5 = 0 ;
  bool val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[7] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanPermission", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanPermission" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanPermission" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanPermission" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnChanPermission" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanPermission" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnChanPermission" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanPermission" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_unsigned_SS_char(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModules_OnChanPermission" "', argument " "5"" of type '" "unsigned char""'");
  } 
  arg5 = static_cast< unsigned char >(val5);
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CModules_OnChanPermission" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  ecode7 = SWIG_AsVal_bool(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CModules_OnChanPermission" "', argument " "7"" of type '" "bool""'");
  } 
  arg7 = static_cast< bool >(val7);
  result = (bool)(arg1)->OnChanPermission((CNick const &)*arg2,(CNick const &)*arg3,*arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnOp2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnOp2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnOp2" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnOp2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnOp2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnOp2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnOp2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnOp2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModules_OnOp2" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  result = (bool)(arg1)->OnOp2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnOp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnOp", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnOp" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnOp" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnOp" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnOp" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnOp" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnOp" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnOp" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModules_OnOp" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  result = (bool)(arg1)->OnOp((CNick const &)*arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnDeop2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnDeop2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnDeop2" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnDeop2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnDeop2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnDeop2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnDeop2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnDeop2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModules_OnDeop2" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  result = (bool)(arg1)->OnDeop2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnDeop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnDeop", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnDeop" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnDeop" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnDeop" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnDeop" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnDeop" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnDeop" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnDeop" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModules_OnDeop" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  result = (bool)(arg1)->OnDeop((CNick const &)*arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnVoice2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnVoice2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnVoice2" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnVoice2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnVoice2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnVoice2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnVoice2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnVoice2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModules_OnVoice2" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  result = (bool)(arg1)->OnVoice2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnVoice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnVoice", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnVoice" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnVoice" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnVoice" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnVoice" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnVoice" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnVoice" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnVoice" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModules_OnVoice" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  result = (bool)(arg1)->OnVoice((CNick const &)*arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnDevoice2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnDevoice2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnDevoice2" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnDevoice2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnDevoice2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnDevoice2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnDevoice2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnDevoice2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModules_OnDevoice2" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  result = (bool)(arg1)->OnDevoice2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnDevoice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnDevoice", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnDevoice" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnDevoice" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnDevoice" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnDevoice" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnDevoice" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnDevoice" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnDevoice" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CModules_OnDevoice" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  result = (bool)(arg1)->OnDevoice((CNick const &)*arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnRawMode2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnRawMode2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnRawMode2" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnRawMode2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnRawMode2" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnRawMode2" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnRawMode2" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnRawMode2" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModules_OnRawMode2" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnRawMode2" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  result = (bool)(arg1)->OnRawMode2((CNick const *)arg2,*arg3,(CString const &)*arg4,(CString const &)*arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnRawMode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnRawMode", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnRawMode" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnRawMode" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnRawMode" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnRawMode" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnRawMode" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnRawMode" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnRawMode" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModules_OnRawMode" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnRawMode" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  result = (bool)(arg1)->OnRawMode((CNick const &)*arg2,*arg3,(CString const &)*arg4,(CString const &)*arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnMode2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CChan *arg3 = 0 ;
  char arg4 ;
  CString *arg5 = 0 ;
  bool arg6 ;
  bool arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  char val4 ;
  int ecode4 = 0 ;
  int res5 = SWIG_OLDOBJ ;
  bool val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[7] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnMode2", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnMode2" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnMode2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnMode2" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnMode2" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  ecode4 = SWIG_AsVal_char(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModules_OnMode2" "', argument " "4"" of type '" "char""'");
  } 
  arg4 = static_cast< char >(val4);
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModules_OnMode2" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnMode2" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CModules_OnMode2" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  ecode7 = SWIG_AsVal_bool(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CModules_OnMode2" "', argument " "7"" of type '" "bool""'");
  } 
  arg7 = static_cast< bool >(val7);
  result = (bool)(arg1)->OnMode2((CNick const *)arg2,*arg3,arg4,(CString const &)*arg5,arg6,arg7);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnMode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  char arg4 ;
  CString *arg5 = 0 ;
  bool arg6 ;
  bool arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  char val4 ;
  int ecode4 = 0 ;
  int res5 = SWIG_OLDOBJ ;
  bool val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[7] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnMode", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnMode" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnMode" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnMode" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnMode" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnMode" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  ecode4 = SWIG_AsVal_char(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModules_OnMode" "', argument " "4"" of type '" "char""'");
  } 
  arg4 = static_cast< char >(val4);
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModules_OnMode" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnMode" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CModules_OnMode" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  ecode7 = SWIG_AsVal_bool(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CModules_OnMode" "', argument " "7"" of type '" "bool""'");
  } 
  arg7 = static_cast< bool >(val7);
  result = (bool)(arg1)->OnMode((CNick const &)*arg2,*arg3,arg4,(CString const &)*arg5,arg6,arg7);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnRaw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnRaw", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnRaw" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  result = (bool)(arg1)->OnRaw(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnRawMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnRawMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnRawMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (bool)(arg1)->OnRawMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnNumericMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNumericMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnNumericMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnNumericMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNumericMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnNumericMessage" "', argument " "2"" of type '" "CNumericMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnNumericMessage" "', argument " "2"" of type '" "CNumericMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNumericMessage * >(argp2);
  result = (bool)(arg1)->OnNumericMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnStatusCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnStatusCommand", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnStatusCommand" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sCommand");
    }
  }
  result = (bool)(arg1)->OnStatusCommand(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnModCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnModCommand", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnModCommand" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnModCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnModCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->OnModCommand((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnModNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnModNotice", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnModNotice" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->OnModNotice((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnModCTCP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnModCTCP", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnModCTCP" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnModCTCP" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnModCTCP" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->OnModCTCP((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnQuit(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnQuit", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnQuit" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnQuit" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnQuit" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnQuit" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnQuit" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnQuit" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnQuit" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg4 = ptr;
  }
  result = (bool)(arg1)->OnQuit((CNick const &)*arg2,(CString const &)*arg3,(std::vector< CChan *,std::allocator< CChan * > > const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnQuitMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CQuitMessage *arg2 = 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnQuitMessage", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnQuitMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CQuitMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnQuitMessage" "', argument " "2"" of type '" "CQuitMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnQuitMessage" "', argument " "2"" of type '" "CQuitMessage &""'"); 
  }
  arg2 = reinterpret_cast< CQuitMessage * >(argp2);
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnQuitMessage" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnQuitMessage" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->OnQuitMessage(*arg2,(std::vector< CChan *,std::allocator< CChan * > > const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnNick", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnNick" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnNick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnNick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnNick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnNick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnNick" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnNick" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg4 = ptr;
  }
  result = (bool)(arg1)->OnNick((CNick const &)*arg2,(CString const &)*arg3,(std::vector< CChan *,std::allocator< CChan * > > const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnNickMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNickMessage *arg2 = 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnNickMessage", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnNickMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNickMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnNickMessage" "', argument " "2"" of type '" "CNickMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnNickMessage" "', argument " "2"" of type '" "CNickMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNickMessage * >(argp2);
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnNickMessage" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnNickMessage" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->OnNickMessage(*arg2,(std::vector< CChan *,std::allocator< CChan * > > const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnKick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  CChan *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnKick", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnKick" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnKick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnKick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnKick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnKick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnKick" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnKick" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModules_OnKick" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnKick" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  result = (bool)(arg1)->OnKick((CNick const &)*arg2,(CString const &)*arg3,*arg4,(CString const &)*arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnKickMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CKickMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnKickMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnKickMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CKickMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnKickMessage" "', argument " "2"" of type '" "CKickMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnKickMessage" "', argument " "2"" of type '" "CKickMessage &""'"); 
  }
  arg2 = reinterpret_cast< CKickMessage * >(argp2);
  result = (bool)(arg1)->OnKickMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnJoining(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CChan *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnJoining", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnJoining" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnJoining" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnJoining" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  result = (bool)(arg1)->OnJoining(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnJoin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnJoin", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnJoin" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnJoin" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnJoin" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnJoin" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnJoin" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  result = (bool)(arg1)->OnJoin((CNick const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnJoinMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CJoinMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnJoinMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnJoinMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CJoinMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnJoinMessage" "', argument " "2"" of type '" "CJoinMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnJoinMessage" "', argument " "2"" of type '" "CJoinMessage &""'"); 
  }
  arg2 = reinterpret_cast< CJoinMessage * >(argp2);
  result = (bool)(arg1)->OnJoinMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPart(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPart", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPart" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPart" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPart" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnPart" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPart" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnPart" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPart" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (bool)(arg1)->OnPart((CNick const &)*arg2,*arg3,(CString const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPartMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CPartMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPartMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPartMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CPartMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPartMessage" "', argument " "2"" of type '" "CPartMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPartMessage" "', argument " "2"" of type '" "CPartMessage &""'"); 
  }
  arg2 = reinterpret_cast< CPartMessage * >(argp2);
  result = (bool)(arg1)->OnPartMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnInvite(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnInvite", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnInvite" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnInvite" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnInvite" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnInvite" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnInvite" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->OnInvite((CNick const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnInviteMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CInviteMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnInviteMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnInviteMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CInviteMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnInviteMessage" "', argument " "2"" of type '" "CInviteMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnInviteMessage" "', argument " "2"" of type '" "CInviteMessage &""'"); 
  }
  arg2 = reinterpret_cast< CInviteMessage * >(argp2);
  result = (bool)(arg1)->OnInviteMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanBufferStarting(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CChan *arg2 = 0 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanBufferStarting", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanBufferStarting" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanBufferStarting" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanBufferStarting" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnChanBufferStarting" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanBufferStarting" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (bool)(arg1)->OnChanBufferStarting(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanBufferEnding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CChan *arg2 = 0 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanBufferEnding", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanBufferEnding" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanBufferEnding" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanBufferEnding" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnChanBufferEnding" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanBufferEnding" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (bool)(arg1)->OnChanBufferEnding(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanBufferPlayLine2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CChan *arg2 = 0 ;
  CClient *arg3 = 0 ;
  CString *arg4 = 0 ;
  timeval *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanBufferPlayLine2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanBufferPlayLine2" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanBufferPlayLine2" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanBufferPlayLine2" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnChanBufferPlayLine2" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanBufferPlayLine2" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sLine");
    }
  }
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_timeval,  0  | 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModules_OnChanBufferPlayLine2" "', argument " "5"" of type '" "timeval const &""'"); 
  }
  if (!argp5) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanBufferPlayLine2" "', argument " "5"" of type '" "timeval const &""'"); 
  }
  arg5 = reinterpret_cast< timeval * >(argp5);
  result = (bool)(arg1)->OnChanBufferPlayLine2(*arg2,*arg3,*arg4,(timeval const &)*arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanBufferPlayLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CChan *arg2 = 0 ;
  CClient *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanBufferPlayLine", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanBufferPlayLine" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanBufferPlayLine" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanBufferPlayLine" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnChanBufferPlayLine" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanBufferPlayLine" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sLine");
    }
  }
  result = (bool)(arg1)->OnChanBufferPlayLine(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivBufferStarting(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CQuery *arg2 = 0 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivBufferStarting", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivBufferStarting" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CQuery,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivBufferStarting" "', argument " "2"" of type '" "CQuery &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivBufferStarting" "', argument " "2"" of type '" "CQuery &""'"); 
  }
  arg2 = reinterpret_cast< CQuery * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnPrivBufferStarting" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivBufferStarting" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (bool)(arg1)->OnPrivBufferStarting(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivBufferEnding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CQuery *arg2 = 0 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivBufferEnding", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivBufferEnding" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CQuery,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivBufferEnding" "', argument " "2"" of type '" "CQuery &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivBufferEnding" "', argument " "2"" of type '" "CQuery &""'"); 
  }
  arg2 = reinterpret_cast< CQuery * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnPrivBufferEnding" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivBufferEnding" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (bool)(arg1)->OnPrivBufferEnding(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivBufferPlayLine2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CClient *arg2 = 0 ;
  CString *arg3 = 0 ;
  timeval *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivBufferPlayLine2", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivBufferPlayLine2" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivBufferPlayLine2" "', argument " "2"" of type '" "CClient &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivBufferPlayLine2" "', argument " "2"" of type '" "CClient &""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sLine");
    }
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_timeval,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnPrivBufferPlayLine2" "', argument " "4"" of type '" "timeval const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivBufferPlayLine2" "', argument " "4"" of type '" "timeval const &""'"); 
  }
  arg4 = reinterpret_cast< timeval * >(argp4);
  result = (bool)(arg1)->OnPrivBufferPlayLine2(*arg2,*arg3,(timeval const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivBufferPlayLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CClient *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivBufferPlayLine", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivBufferPlayLine" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivBufferPlayLine" "', argument " "2"" of type '" "CClient &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivBufferPlayLine" "', argument " "2"" of type '" "CClient &""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sLine");
    }
  }
  result = (bool)(arg1)->OnPrivBufferPlayLine(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanBufferPlayMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanBufferPlayMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanBufferPlayMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanBufferPlayMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanBufferPlayMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (bool)(arg1)->OnChanBufferPlayMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivBufferPlayMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivBufferPlayMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivBufferPlayMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivBufferPlayMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivBufferPlayMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (bool)(arg1)->OnPrivBufferPlayMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnClientLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnClientLogin" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (bool)(arg1)->OnClientLogin();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnClientDisconnect(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnClientDisconnect" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (bool)(arg1)->OnClientDisconnect();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserRaw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserRaw", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserRaw" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  result = (bool)(arg1)->OnUserRaw(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserRawMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserRawMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserRawMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnUserRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnUserRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (bool)(arg1)->OnUserRawMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserCTCPReply(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserCTCPReply", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserCTCPReply" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (bool)(arg1)->OnUserCTCPReply(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserCTCPReplyMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserCTCPReplyMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserCTCPReplyMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnUserCTCPReplyMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnUserCTCPReplyMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (bool)(arg1)->OnUserCTCPReplyMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserCTCP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserCTCP", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserCTCP" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (bool)(arg1)->OnUserCTCP(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserCTCPMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserCTCPMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserCTCPMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnUserCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnUserCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (bool)(arg1)->OnUserCTCPMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserAction(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserAction", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserAction" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (bool)(arg1)->OnUserAction(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserActionMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CActionMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserActionMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserActionMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CActionMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnUserActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnUserActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  arg2 = reinterpret_cast< CActionMessage * >(argp2);
  result = (bool)(arg1)->OnUserActionMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserMsg", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserMsg" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (bool)(arg1)->OnUserMsg(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserTextMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CTextMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserTextMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserTextMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTextMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnUserTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnUserTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTextMessage * >(argp2);
  result = (bool)(arg1)->OnUserTextMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserNotice", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserNotice" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (bool)(arg1)->OnUserNotice(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserNoticeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNoticeMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserNoticeMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserNoticeMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNoticeMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnUserNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnUserNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNoticeMessage * >(argp2);
  result = (bool)(arg1)->OnUserNoticeMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserJoin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserJoin", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserJoin" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sChannel");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sKey");
    }
  }
  result = (bool)(arg1)->OnUserJoin(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserJoinMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CJoinMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserJoinMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserJoinMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CJoinMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnUserJoinMessage" "', argument " "2"" of type '" "CJoinMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnUserJoinMessage" "', argument " "2"" of type '" "CJoinMessage &""'"); 
  }
  arg2 = reinterpret_cast< CJoinMessage * >(argp2);
  result = (bool)(arg1)->OnUserJoinMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserPart(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserPart", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserPart" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sChannel");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (bool)(arg1)->OnUserPart(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserPartMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CPartMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserPartMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserPartMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CPartMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnUserPartMessage" "', argument " "2"" of type '" "CPartMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnUserPartMessage" "', argument " "2"" of type '" "CPartMessage &""'"); 
  }
  arg2 = reinterpret_cast< CPartMessage * >(argp2);
  result = (bool)(arg1)->OnUserPartMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserTopic(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserTopic", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserTopic" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sChannel");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sTopic");
    }
  }
  result = (bool)(arg1)->OnUserTopic(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserTopicMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CTopicMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserTopicMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserTopicMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTopicMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnUserTopicMessage" "', argument " "2"" of type '" "CTopicMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnUserTopicMessage" "', argument " "2"" of type '" "CTopicMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTopicMessage * >(argp2);
  result = (bool)(arg1)->OnUserTopicMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserTopicRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserTopicRequest", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserTopicRequest" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sChannel");
    }
  }
  result = (bool)(arg1)->OnUserTopicRequest(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserQuit(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserQuit", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserQuit" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sMessage");
    }
  }
  result = (bool)(arg1)->OnUserQuit(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserQuitMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CQuitMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserQuitMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserQuitMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CQuitMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnUserQuitMessage" "', argument " "2"" of type '" "CQuitMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnUserQuitMessage" "', argument " "2"" of type '" "CQuitMessage &""'"); 
  }
  arg2 = reinterpret_cast< CQuitMessage * >(argp2);
  result = (bool)(arg1)->OnUserQuitMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUserTagMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CTargetMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUserTagMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUserTagMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTargetMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnUserTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnUserTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTargetMessage * >(argp2);
  result = (bool)(arg1)->OnUserTagMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanTagMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CTargetMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanTagMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanTagMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTargetMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTargetMessage * >(argp2);
  result = (bool)(arg1)->OnChanTagMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivTagMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CTargetMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivTagMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivTagMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTargetMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTargetMessage * >(argp2);
  result = (bool)(arg1)->OnPrivTagMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnCTCPReply(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnCTCPReply", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnCTCPReply" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnCTCPReply" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnCTCPReply" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (bool)(arg1)->OnCTCPReply(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnCTCPReplyMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnCTCPReplyMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnCTCPReplyMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnCTCPReplyMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnCTCPReplyMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (bool)(arg1)->OnCTCPReplyMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivCTCP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivCTCP", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivCTCP" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivCTCP" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivCTCP" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (bool)(arg1)->OnPrivCTCP(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivCTCPMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivCTCPMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivCTCPMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (bool)(arg1)->OnPrivCTCPMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanCTCP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanCTCP", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanCTCP" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanCTCP" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanCTCP" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnChanCTCP" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanCTCP" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sMessage");
    }
  }
  result = (bool)(arg1)->OnChanCTCP(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanCTCPMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanCTCPMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanCTCPMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (bool)(arg1)->OnChanCTCPMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivAction(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivAction", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivAction" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivAction" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivAction" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (bool)(arg1)->OnPrivAction(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivActionMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CActionMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivActionMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivActionMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CActionMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  arg2 = reinterpret_cast< CActionMessage * >(argp2);
  result = (bool)(arg1)->OnPrivActionMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanAction(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanAction", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanAction" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanAction" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanAction" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnChanAction" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanAction" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sMessage");
    }
  }
  result = (bool)(arg1)->OnChanAction(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanActionMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CActionMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanActionMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanActionMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CActionMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  arg2 = reinterpret_cast< CActionMessage * >(argp2);
  result = (bool)(arg1)->OnChanActionMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivMsg", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivMsg" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivMsg" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivMsg" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (bool)(arg1)->OnPrivMsg(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivTextMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CTextMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivTextMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivTextMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTextMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTextMessage * >(argp2);
  result = (bool)(arg1)->OnPrivTextMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanMsg", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanMsg" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanMsg" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanMsg" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnChanMsg" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanMsg" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sMessage");
    }
  }
  result = (bool)(arg1)->OnChanMsg(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanTextMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CTextMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanTextMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanTextMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTextMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTextMessage * >(argp2);
  result = (bool)(arg1)->OnChanTextMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivNotice", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivNotice" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivNotice" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivNotice" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (bool)(arg1)->OnPrivNotice(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnPrivNoticeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNoticeMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnPrivNoticeMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnPrivNoticeMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNoticeMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnPrivNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnPrivNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNoticeMessage * >(argp2);
  result = (bool)(arg1)->OnPrivNoticeMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanNotice", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanNotice" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanNotice" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanNotice" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnChanNotice" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanNotice" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sMessage");
    }
  }
  result = (bool)(arg1)->OnChanNotice(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnChanNoticeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNoticeMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnChanNoticeMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnChanNoticeMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNoticeMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnChanNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnChanNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNoticeMessage * >(argp2);
  result = (bool)(arg1)->OnChanNoticeMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnTopic(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnTopic", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnTopic" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnTopic" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnTopic" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnTopic" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnTopic" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sTopic");
    }
  }
  result = (bool)(arg1)->OnTopic(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnTopicMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CTopicMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnTopicMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnTopicMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTopicMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnTopicMessage" "', argument " "2"" of type '" "CTopicMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnTopicMessage" "', argument " "2"" of type '" "CTopicMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTopicMessage * >(argp2);
  result = (bool)(arg1)->OnTopicMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnTimerAutoJoin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CChan *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnTimerAutoJoin", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnTimerAutoJoin" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnTimerAutoJoin" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnTimerAutoJoin" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  result = (bool)(arg1)->OnTimerAutoJoin(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnAddNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CIRCNetwork *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnAddNetwork", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnAddNetwork" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CIRCNetwork,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnAddNetwork" "', argument " "2"" of type '" "CIRCNetwork &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnAddNetwork" "', argument " "2"" of type '" "CIRCNetwork &""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sErrorRet");
    }
  }
  result = (bool)(arg1)->OnAddNetwork(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnDeleteNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CIRCNetwork *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnDeleteNetwork", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnDeleteNetwork" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CIRCNetwork,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnDeleteNetwork" "', argument " "2"" of type '" "CIRCNetwork &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnDeleteNetwork" "', argument " "2"" of type '" "CIRCNetwork &""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  result = (bool)(arg1)->OnDeleteNetwork(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnSendToClient(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnSendToClient", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnSendToClient" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnSendToClient" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnSendToClient" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (bool)(arg1)->OnSendToClient(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnSendToClientMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnSendToClientMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnSendToClientMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnSendToClientMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnSendToClientMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (bool)(arg1)->OnSendToClientMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnSendToIRC(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnSendToIRC", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnSendToIRC" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  result = (bool)(arg1)->OnSendToIRC(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnSendToIRCMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnSendToIRCMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnSendToIRCMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnSendToIRCMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnSendToIRCMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (bool)(arg1)->OnSendToIRCMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnClientAttached(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnClientAttached" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (bool)(arg1)->OnClientAttached();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnClientDetached(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnClientDetached" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (bool)(arg1)->OnClientDetached();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnServerCapAvailable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnServerCapAvailable", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnServerCapAvailable" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnServerCapAvailable" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnServerCapAvailable" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnServerCapAvailable" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnServerCapAvailable" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->OnServerCapAvailable((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnServerCapResult(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnServerCapResult", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnServerCapResult" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnServerCapResult" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnServerCapResult" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CModules_OnServerCapResult" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)(arg1)->OnServerCapResult((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_FindModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CModule *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_FindModule", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_FindModule" "', argument " "1"" of type '" "CModules const *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_FindModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_FindModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CModule *)((CModules const *)arg1)->FindModule((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModule, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_LoadModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CModInfo::EModuleType arg4 ;
  CUser *arg5 = (CUser *) 0 ;
  CIRCNetwork *arg6 = (CIRCNetwork *) 0 ;
  CString *arg7 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject *swig_obj[7] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_LoadModule", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_LoadModule" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_LoadModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_LoadModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_LoadModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_LoadModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModules_LoadModule" "', argument " "4"" of type '" "CModInfo::EModuleType""'");
  } 
  arg4 = static_cast< CModInfo::EModuleType >(val4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModules_LoadModule" "', argument " "5"" of type '" "CUser *""'"); 
  }
  arg5 = reinterpret_cast< CUser * >(argp5);
  res6 = SWIG_ConvertPtr(swig_obj[5], &argp6,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "CModules_LoadModule" "', argument " "6"" of type '" "CIRCNetwork *""'"); 
  }
  arg6 = reinterpret_cast< CIRCNetwork * >(argp6);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[6], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg7 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 7 sRetMsg");
    }
  }
  result = (bool)(arg1)->LoadModule((CString const &)*arg2,(CString const &)*arg3,arg4,arg5,arg6,*arg7);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_UnloadModule__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_UnloadModule" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_UnloadModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_UnloadModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->UnloadModule((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_UnloadModule__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_UnloadModule" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_UnloadModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_UnloadModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sRetMsg");
    }
  }
  result = (bool)(arg1)->UnloadModule((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_UnloadModule(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModules_UnloadModule", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModules, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CModules_UnloadModule__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CModules, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          return _wrap_CModules_UnloadModule__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModules_UnloadModule'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModules::UnloadModule(CString const &)\n"
    "    CModules::UnloadModule(CString const &,CString &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModules_ReloadModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CUser *arg4 = (CUser *) 0 ;
  CIRCNetwork *arg5 = (CIRCNetwork *) 0 ;
  CString *arg6 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject *swig_obj[6] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_ReloadModule", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_ReloadModule" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_ReloadModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_ReloadModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_ReloadModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_ReloadModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_ReloadModule" "', argument " "4"" of type '" "CUser *""'"); 
  }
  arg4 = reinterpret_cast< CUser * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModules_ReloadModule" "', argument " "5"" of type '" "CIRCNetwork *""'"); 
  }
  arg5 = reinterpret_cast< CIRCNetwork * >(argp5);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[5], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg6 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 6 sRetMsg");
    }
  }
  result = (bool)(arg1)->ReloadModule((CString const &)*arg2,(CString const &)*arg3,arg4,arg5,*arg6);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_GetModInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_GetModInfo", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CModInfo,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_GetModInfo" "', argument " "1"" of type '" "CModInfo &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_GetModInfo" "', argument " "1"" of type '" "CModInfo &""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_GetModInfo" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_GetModInfo" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sRetMsg");
    }
  }
  result = (bool)CModules::GetModInfo(*arg1,(CString const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_GetModPathInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModInfo *arg1 = 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_GetModPathInfo", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CModInfo,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_GetModPathInfo" "', argument " "1"" of type '" "CModInfo &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_GetModPathInfo" "', argument " "1"" of type '" "CModInfo &""'"); 
  }
  arg1 = reinterpret_cast< CModInfo * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_GetModPathInfo" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_GetModPathInfo" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_GetModPathInfo" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_GetModPathInfo" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sRetMsg");
    }
  }
  result = (bool)CModules::GetModPathInfo(*arg1,(CString const &)*arg2,(CString const &)*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_GetAvailableMods__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > *arg1 = 0 ;
  CModInfo::EModuleType arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__setT_CModInfo_t,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_GetAvailableMods" "', argument " "1"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_GetAvailableMods" "', argument " "1"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CModules_GetAvailableMods" "', argument " "2"" of type '" "CModInfo::EModuleType""'");
  } 
  arg2 = static_cast< CModInfo::EModuleType >(val2);
  CModules::GetAvailableMods(*arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_GetAvailableMods__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__setT_CModInfo_t,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_GetAvailableMods" "', argument " "1"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_GetAvailableMods" "', argument " "1"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > * >(argp1);
  CModules::GetAvailableMods(*arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_GetAvailableMods(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModules_GetAvailableMods", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_std__setT_CModInfo_t, SWIG_POINTER_NO_NULL);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CModules_GetAvailableMods__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_std__setT_CModInfo_t, SWIG_POINTER_NO_NULL);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CModules_GetAvailableMods__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModules_GetAvailableMods'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModules::GetAvailableMods(std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &,CModInfo::EModuleType)\n"
    "    CModules::GetAvailableMods(std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModules_GetDefaultMods__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > *arg1 = 0 ;
  CModInfo::EModuleType arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__setT_CModInfo_t,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_GetDefaultMods" "', argument " "1"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_GetDefaultMods" "', argument " "1"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CModules_GetDefaultMods" "', argument " "2"" of type '" "CModInfo::EModuleType""'");
  } 
  arg2 = static_cast< CModInfo::EModuleType >(val2);
  CModules::GetDefaultMods(*arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_GetDefaultMods__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__setT_CModInfo_t,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_GetDefaultMods" "', argument " "1"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_GetDefaultMods" "', argument " "1"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  arg1 = reinterpret_cast< std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > * >(argp1);
  CModules::GetDefaultMods(*arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_GetDefaultMods(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CModules_GetDefaultMods", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_std__setT_CModInfo_t, SWIG_POINTER_NO_NULL);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CModules_GetDefaultMods__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_std__setT_CModInfo_t, SWIG_POINTER_NO_NULL);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CModules_GetDefaultMods__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CModules_GetDefaultMods'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CModules::GetDefaultMods(std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &,CModInfo::EModuleType)\n"
    "    CModules::GetDefaultMods(std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CModules_FindModPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_FindModPath", 3, 3, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_FindModPath" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_FindModPath" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sModPath");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sDataPath");
    }
  }
  result = (bool)CModules::FindModPath((CString const &)*arg1,*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_GetModDirs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  SwigValueWrapper< std::queue< std::pair< CString,CString > > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_GetModDirs", 0, 0, 0)) SWIG_fail;
  result = CModules::GetModDirs();
  
  resultobj = PyList_New((&result)->size());
  if (resultobj) {
    for (size_t i = 0; !(&result)->empty(); (&result)->pop(), ++i) {
      PyList_SetItem(resultobj, i, Py_BuildValue("ss", (&result)->front().first.c_str(), (&result)->front().second.c_str()));
    }
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnAddUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CUser *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnAddUser", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnAddUser" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnAddUser" "', argument " "2"" of type '" "CUser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnAddUser" "', argument " "2"" of type '" "CUser &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sErrorRet");
    }
  }
  result = (bool)(arg1)->OnAddUser(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnDeleteUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CUser *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnDeleteUser", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnDeleteUser" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnDeleteUser" "', argument " "2"" of type '" "CUser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnDeleteUser" "', argument " "2"" of type '" "CUser &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  result = (bool)(arg1)->OnDeleteUser(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnClientConnect(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CZNCSock *arg2 = (CZNCSock *) 0 ;
  CString *arg3 = 0 ;
  unsigned short arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  unsigned short val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnClientConnect", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnClientConnect" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnClientConnect" "', argument " "2"" of type '" "CZNCSock *""'"); 
  }
  arg2 = reinterpret_cast< CZNCSock * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnClientConnect" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnClientConnect" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_unsigned_SS_short(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModules_OnClientConnect" "', argument " "4"" of type '" "unsigned short""'");
  } 
  arg4 = static_cast< unsigned short >(val4);
  result = (bool)(arg1)->OnClientConnect(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnLoginAttempt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  std::shared_ptr< CAuthBase > arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnLoginAttempt", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnLoginAttempt" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    int newmem = 0;
    res2 = SWIG_ConvertPtrAndOwn(swig_obj[1], &argp2, SWIGTYPE_p_std__shared_ptrT_CAuthBase_t,  0 , &newmem);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnLoginAttempt" "', argument " "2"" of type '" "std::shared_ptr< CAuthBase >""'");
    }
    if (argp2) arg2 = *(reinterpret_cast< std::shared_ptr< CAuthBase > * >(argp2));
    if (newmem & SWIG_CAST_NEW_MEMORY) delete reinterpret_cast< std::shared_ptr< CAuthBase > * >(argp2);
  }
  result = (bool)(arg1)->OnLoginAttempt(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnFailedLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnFailedLogin", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnFailedLogin" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnFailedLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnFailedLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnFailedLogin" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnFailedLogin" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->OnFailedLogin((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUnknownUserRaw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUnknownUserRaw", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUnknownUserRaw" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnUnknownUserRaw" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sLine");
    }
  }
  result = (bool)(arg1)->OnUnknownUserRaw(arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnUnknownUserRawMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnUnknownUserRawMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnUnknownUserRawMessage" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnUnknownUserRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnUnknownUserRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (bool)(arg1)->OnUnknownUserRawMessage(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnClientCapLs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  SCString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnClientCapLs", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnClientCapLs" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnClientCapLs" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__setT_CString_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnClientCapLs" "', argument " "3"" of type '" "SCString &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnClientCapLs" "', argument " "3"" of type '" "SCString &""'"); 
  }
  arg3 = reinterpret_cast< SCString * >(argp3);
  result = (bool)(arg1)->OnClientCapLs(arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_IsClientCapSupported(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_IsClientCapSupported", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_IsClientCapSupported" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_IsClientCapSupported" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_IsClientCapSupported" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_IsClientCapSupported" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModules_IsClientCapSupported" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->IsClientCapSupported(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnClientCapRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnClientCapRequest", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnClientCapRequest" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnClientCapRequest" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnClientCapRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnClientCapRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModules_OnClientCapRequest" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->OnClientCapRequest(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnClientGetSASLMechanisms(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  SCString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnClientGetSASLMechanisms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnClientGetSASLMechanisms" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__setT_CString_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnClientGetSASLMechanisms" "', argument " "2"" of type '" "SCString &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnClientGetSASLMechanisms" "', argument " "2"" of type '" "SCString &""'"); 
  }
  arg2 = reinterpret_cast< SCString * >(argp2);
  result = (bool)(arg1)->OnClientGetSASLMechanisms(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnClientSASLAborted(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnClientSASLAborted" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (bool)(arg1)->OnClientSASLAborted();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnClientSASLServerInitialChallenge(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnClientSASLServerInitialChallenge", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnClientSASLServerInitialChallenge" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnClientSASLServerInitialChallenge" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnClientSASLServerInitialChallenge" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sResponse");
    }
  }
  result = (bool)(arg1)->OnClientSASLServerInitialChallenge((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnClientSASLAuthenticate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnClientSASLAuthenticate", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnClientSASLAuthenticate" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnClientSASLAuthenticate" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnClientSASLAuthenticate" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnClientSASLAuthenticate" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnClientSASLAuthenticate" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->OnClientSASLAuthenticate((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnModuleLoading(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CModInfo::EModuleType arg4 ;
  bool *arg5 = 0 ;
  CString *arg6 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject *swig_obj[6] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnModuleLoading", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnModuleLoading" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnModuleLoading" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnModuleLoading" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnModuleLoading" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnModuleLoading" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CModules_OnModuleLoading" "', argument " "4"" of type '" "CModInfo::EModuleType""'");
  } 
  arg4 = static_cast< CModInfo::EModuleType >(val4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_bool,  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CModules_OnModuleLoading" "', argument " "5"" of type '" "bool &""'"); 
  }
  if (!argp5) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnModuleLoading" "', argument " "5"" of type '" "bool &""'"); 
  }
  arg5 = reinterpret_cast< bool * >(argp5);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[5], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg6 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 6 sRetMsg");
    }
  }
  result = (bool)(arg1)->OnModuleLoading((CString const &)*arg2,(CString const &)*arg3,arg4,*arg5,*arg6);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnModuleUnloading(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CModule *arg2 = (CModule *) 0 ;
  bool *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnModuleUnloading", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnModuleUnloading" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnModuleUnloading" "', argument " "2"" of type '" "CModule *""'"); 
  }
  arg2 = reinterpret_cast< CModule * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_bool,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnModuleUnloading" "', argument " "3"" of type '" "bool &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnModuleUnloading" "', argument " "3"" of type '" "bool &""'"); 
  }
  arg3 = reinterpret_cast< bool * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sRetMsg");
    }
  }
  result = (bool)(arg1)->OnModuleUnloading(arg2,*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnGetModInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CModInfo *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnGetModInfo", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnGetModInfo" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnGetModInfo" "', argument " "2"" of type '" "CModInfo &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnGetModInfo" "', argument " "2"" of type '" "CModInfo &""'"); 
  }
  arg2 = reinterpret_cast< CModInfo * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CModules_OnGetModInfo" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnGetModInfo" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_bool,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CModules_OnGetModInfo" "', argument " "4"" of type '" "bool &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnGetModInfo" "', argument " "4"" of type '" "bool &""'"); 
  }
  arg4 = reinterpret_cast< bool * >(argp4);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[4], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg5 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 5 sRetMsg");
    }
  }
  result = (bool)(arg1)->OnGetModInfo(*arg2,(CString const &)*arg3,*arg4,*arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_OnGetAvailableMods(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > *arg2 = 0 ;
  CModInfo::EModuleType arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_OnGetAvailableMods", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_OnGetAvailableMods" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__setT_CModInfo_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_OnGetAvailableMods" "', argument " "2"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CModules_OnGetAvailableMods" "', argument " "2"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  arg2 = reinterpret_cast< std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > * >(argp2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CModules_OnGetAvailableMods" "', argument " "3"" of type '" "CModInfo::EModuleType""'");
  } 
  arg3 = static_cast< CModInfo::EModuleType >(val3);
  result = (bool)(arg1)->OnGetAvailableMods(*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModules_removeModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  CModule *arg2 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModules_removeModule", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModules_removeModule" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModules_removeModule" "', argument " "2"" of type '" "CModule *""'"); 
  }
  arg2 = reinterpret_cast< CModule * >(argp2);
  result = (bool)CModules_removeModule(arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CModules_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CModules, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CModules_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CNick__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CNick *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CNick *)new CNick();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNick, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CNick__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  CNick *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CNick" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CNick" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (CNick *)new CNick((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNick, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CNick" "', argument " "1"" of type '" "CNick *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CNick__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CNick *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CNick *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CNick" "', argument " "1"" of type '" "CNick const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CNick" "', argument " "1"" of type '" "CNick const &""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  result = (CNick *)new CNick((CNick const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNick, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CNick(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CNick", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CNick__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_CNick, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CNick__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CNick__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CNick'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CNick::CNick()\n"
    "    CNick::CNick(CString const &)\n"
    "    CNick::CNick(CNick const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CNick_Reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_Reset" "', argument " "1"" of type '" "CNick *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  (arg1)->Reset();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_Parse(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CNick_Parse", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_Parse" "', argument " "1"" of type '" "CNick *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CNick_Parse" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CNick_Parse" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->Parse((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_GetHostMask(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_GetHostMask" "', argument " "1"" of type '" "CNick const *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  result = ((CNick const *)arg1)->GetHostMask();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_GetCommonChans(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg2 = 0 ;
  CIRCNetwork *arg3 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  size_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CNick_GetCommonChans", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_GetCommonChans" "', argument " "1"" of type '" "CNick const *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_CChan_p_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CNick_GetCommonChans" "', argument " "2"" of type '" "std::vector< CChan *,std::allocator< CChan * > > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CNick_GetCommonChans" "', argument " "2"" of type '" "std::vector< CChan *,std::allocator< CChan * > > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< CChan *,std::allocator< CChan * > > * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CNick_GetCommonChans" "', argument " "3"" of type '" "CIRCNetwork *""'"); 
  }
  arg3 = reinterpret_cast< CIRCNetwork * >(argp3);
  result = ((CNick const *)arg1)->GetCommonChans(*arg2,arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_NickEquals(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CNick_NickEquals", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_NickEquals" "', argument " "1"" of type '" "CNick const *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CNick_NickEquals" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CNick_NickEquals" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)((CNick const *)arg1)->NickEquals((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_SetNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CNick_SetNetwork", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_SetNetwork" "', argument " "1"" of type '" "CNick *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CNick_SetNetwork" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  (arg1)->SetNetwork(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_SetNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CNick_SetNick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_SetNick" "', argument " "1"" of type '" "CNick *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CNick_SetNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CNick_SetNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetNick((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_SetIdent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CNick_SetIdent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_SetIdent" "', argument " "1"" of type '" "CNick *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CNick_SetIdent" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CNick_SetIdent" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetIdent((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_SetHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CNick_SetHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_SetHost" "', argument " "1"" of type '" "CNick *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CNick_SetHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CNick_SetHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetHost((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_AddPerm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CNick_AddPerm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_AddPerm" "', argument " "1"" of type '" "CNick *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CNick_AddPerm" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = (bool)(arg1)->AddPerm(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_RemPerm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CNick_RemPerm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_RemPerm" "', argument " "1"" of type '" "CNick *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CNick_RemPerm" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = (bool)(arg1)->RemPerm(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_GetPermStr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_GetPermStr" "', argument " "1"" of type '" "CNick const *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  result = ((CNick const *)arg1)->GetPermStr();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_GetPermChar(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_GetPermChar" "', argument " "1"" of type '" "CNick const *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  result = (char)((CNick const *)arg1)->GetPermChar();
  resultobj = SWIG_From_char(static_cast< char >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_HasPerm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CNick_HasPerm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_HasPerm" "', argument " "1"" of type '" "CNick const *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CNick_HasPerm" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = (bool)((CNick const *)arg1)->HasPerm(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_GetNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_GetNick" "', argument " "1"" of type '" "CNick const *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  result = (CString *) &((CNick const *)arg1)->GetNick();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_GetIdent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_GetIdent" "', argument " "1"" of type '" "CNick const *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  result = (CString *) &((CNick const *)arg1)->GetIdent();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_GetHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_GetHost" "', argument " "1"" of type '" "CNick const *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  result = (CString *) &((CNick const *)arg1)->GetHost();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_GetNickMask(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_GetNickMask" "', argument " "1"" of type '" "CNick const *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  result = ((CNick const *)arg1)->GetNickMask();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick_Clone(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  CNick *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CNick_Clone", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick_Clone" "', argument " "1"" of type '" "CNick *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CNick_Clone" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CNick_Clone" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  (arg1)->Clone((CNick const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick___str__" "', argument " "1"" of type '" "CNick *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  result = CNick___str__(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CNick___repr__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CNick *arg1 = (CNick *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CNick___repr__" "', argument " "1"" of type '" "CNick *""'"); 
  }
  arg1 = reinterpret_cast< CNick * >(argp1);
  result = CNick___repr__(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CNick_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CNick, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CNick_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CChan__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  bool arg3 ;
  CConfig *arg4 = (CConfig *) 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  CChan *result = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CChan" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CChan" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CChan" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CChan" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CChan" "', argument " "4"" of type '" "CConfig *""'"); 
  }
  arg4 = reinterpret_cast< CConfig * >(argp4);
  result = (CChan *)new CChan((CString const &)*arg1,arg2,arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CChan, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CChan__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  bool arg3 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  CChan *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CChan" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CChan" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CChan" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CChan" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (CChan *)new CChan((CString const &)*arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CChan, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CChan(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CChan", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_new_CChan__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CConfig, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_new_CChan__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CChan'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CChan::CChan(CString const &,CIRCNetwork *,bool,CConfig *)\n"
    "    CChan::CChan(CString const &,CIRCNetwork *,bool)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CChan(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CChan" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_Reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_Reset" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  (arg1)->Reset();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_ToConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CConfig result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_ToConfig" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = ((CChan const *)arg1)->ToConfig();
  resultobj = SWIG_NewPointerObj((new CConfig(result)), SWIGTYPE_p_CConfig, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_Clone(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CChan *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_Clone", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_Clone" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_Clone" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_Clone" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  (arg1)->Clone(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_Cycle(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_Cycle" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ((CChan const *)arg1)->Cycle();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_JoinUser__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_JoinUser" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_JoinUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_JoinUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->JoinUser((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_JoinUser__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_JoinUser" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  (arg1)->JoinUser();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_JoinUser(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CChan_JoinUser", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CChan_JoinUser__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CChan_JoinUser__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CChan_JoinUser'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CChan::JoinUser(CString const &)\n"
    "    CChan::JoinUser()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CChan_AttachUser__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_AttachUser" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_AttachUser" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  (arg1)->AttachUser(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_AttachUser__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_AttachUser" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  (arg1)->AttachUser();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_AttachUser(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CChan_AttachUser", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CChan_AttachUser__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CClient, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CChan_AttachUser__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CChan_AttachUser'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CChan::AttachUser(CClient *)\n"
    "    CChan::AttachUser()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CChan_DetachUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_DetachUser" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  (arg1)->DetachUser();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_OnWho(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_OnWho", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_OnWho" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_OnWho" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_OnWho" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CChan_OnWho" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_OnWho" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CChan_OnWho" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_OnWho" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->OnWho((CString const &)*arg2,(CString const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetModes__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetModes" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_SetModes" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_SetModes" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetModes((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetModes__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  VCString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetModes" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_SetModes" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_SetModes" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CChan_SetModes" "', argument " "3"" of type '" "VCString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_SetModes" "', argument " "3"" of type '" "VCString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->SetModes((CString const &)*arg2,(VCString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetModes(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CChan_SetModes", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CChan_SetModes__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CString,std::allocator< CString > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CChan_SetModes__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CChan_SetModes'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CChan::SetModes(CString const &)\n"
    "    CChan::SetModes(CString const &,VCString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CChan_ModeChange__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  CNick *arg3 = (CNick *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_ModeChange" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_ModeChange" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_ModeChange" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CChan_ModeChange" "', argument " "3"" of type '" "CNick const *""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  (arg1)->ModeChange((CString const &)*arg2,(CNick const *)arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_ModeChange__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_ModeChange" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_ModeChange" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_ModeChange" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->ModeChange((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_ModeChange__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  VCString *arg3 = 0 ;
  CNick *arg4 = (CNick *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_ModeChange" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_ModeChange" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_ModeChange" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CChan_ModeChange" "', argument " "3"" of type '" "VCString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_ModeChange" "', argument " "3"" of type '" "VCString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CChan_ModeChange" "', argument " "4"" of type '" "CNick const *""'"); 
  }
  arg4 = reinterpret_cast< CNick * >(argp4);
  (arg1)->ModeChange((CString const &)*arg2,(VCString const &)*arg3,(CNick const *)arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_ModeChange__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  VCString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_ModeChange" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_ModeChange" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_ModeChange" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::vector< CString,std::allocator< CString > > *ptr = (std::vector< CString,std::allocator< CString > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CChan_ModeChange" "', argument " "3"" of type '" "VCString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_ModeChange" "', argument " "3"" of type '" "VCString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->ModeChange((CString const &)*arg2,(VCString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_ModeChange(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CChan_ModeChange", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CChan_ModeChange__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CNick, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CChan_ModeChange__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CString,std::allocator< CString > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CChan_ModeChange__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< CString,std::allocator< CString > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CNick, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CChan_ModeChange__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CChan_ModeChange'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CChan::ModeChange(CString const &,CNick const *)\n"
    "    CChan::ModeChange(CString const &)\n"
    "    CChan::ModeChange(CString const &,VCString const &,CNick const *)\n"
    "    CChan::ModeChange(CString const &,VCString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CChan_AddMode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  char arg2 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_AddMode", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_AddMode" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_AddMode" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CChan_AddMode" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddMode" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->AddMode(arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_RemMode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_RemMode", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_RemMode" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_RemMode" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = (bool)(arg1)->RemMode(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetModeString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetModeString" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = ((CChan const *)arg1)->GetModeString();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetModeArg__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetModeArg" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sArgs");
    }
  }
  result = ((CChan const *)arg1)->GetModeArg(*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetModeForNames(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetModeForNames" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = ((CChan const *)arg1)->GetModeForNames();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_ClearNicks(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_ClearNicks" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  (arg1)->ClearNicks();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_FindNick__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CNick *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_FindNick" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_FindNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_FindNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CNick *)((CChan const *)arg1)->FindNick((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNick, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_FindNick__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CNick *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_FindNick" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_FindNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_FindNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CNick *)(arg1)->FindNick((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNick, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_FindNick(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CChan_FindNick", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CChan_FindNick__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CChan_FindNick__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CChan_FindNick'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CChan::FindNick(CString const &) const\n"
    "    CChan::FindNick(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CChan_AddNicks(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_AddNicks", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_AddNicks" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_AddNicks" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddNicks" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (int)(arg1)->AddNicks((CString const &)*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_AddNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_AddNick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_AddNick" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_AddNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->AddNick((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_RemNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_RemNick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_RemNick" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_RemNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_RemNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->RemNick((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_ChangeNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_ChangeNick", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_ChangeNick" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_ChangeNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_ChangeNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CChan_ChangeNick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_ChangeNick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->ChangeNick((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CBuffer *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetBuffer" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (CBuffer *) &((CChan const *)arg1)->GetBuffer();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBuffer, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetBufferCount(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetBufferCount" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (unsigned int)((CChan const *)arg1)->GetBufferCount();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetBufferCount__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  unsigned int arg2 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetBufferCount" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_SetBufferCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CChan_SetBufferCount" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)(arg1)->SetBufferCount(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetBufferCount__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetBufferCount" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_SetBufferCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = (bool)(arg1)->SetBufferCount(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetBufferCount(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CChan_SetBufferCount", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CChan_SetBufferCount__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CChan_SetBufferCount__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CChan_SetBufferCount'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CChan::SetBufferCount(unsigned int,bool)\n"
    "    CChan::SetBufferCount(unsigned int)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CChan_InheritBufferCount__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  unsigned int arg2 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_InheritBufferCount" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_InheritBufferCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CChan_InheritBufferCount" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->InheritBufferCount(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_InheritBufferCount__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_InheritBufferCount" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_InheritBufferCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  (arg1)->InheritBufferCount(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_InheritBufferCount(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CChan_InheritBufferCount", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CChan_InheritBufferCount__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CChan_InheritBufferCount__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CChan_InheritBufferCount'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CChan::InheritBufferCount(unsigned int,bool)\n"
    "    CChan::InheritBufferCount(unsigned int)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CChan_ResetBufferCount(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_ResetBufferCount" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  (arg1)->ResetBufferCount();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_AddBuffer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CMessage *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  size_t result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_AddBuffer" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_AddBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CChan_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (arg1)->AddBuffer((CMessage const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_AddBuffer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  size_t result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_AddBuffer" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_AddBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (arg1)->AddBuffer((CMessage const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_AddBuffer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  timeval *arg4 = (timeval *) 0 ;
  MCString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  size_t result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_AddBuffer" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CChan_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_timeval, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CChan_AddBuffer" "', argument " "4"" of type '" "timeval const *""'"); 
  }
  arg4 = reinterpret_cast< timeval * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_MCString,  0  | 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CChan_AddBuffer" "', argument " "5"" of type '" "MCString const &""'"); 
  }
  if (!argp5) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddBuffer" "', argument " "5"" of type '" "MCString const &""'"); 
  }
  arg5 = reinterpret_cast< MCString * >(argp5);
  result = (arg1)->AddBuffer((CString const &)*arg2,(CString const &)*arg3,(timeval const *)arg4,(MCString const &)*arg5);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_AddBuffer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  timeval *arg4 = (timeval *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  size_t result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_AddBuffer" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CChan_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_timeval, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CChan_AddBuffer" "', argument " "4"" of type '" "timeval const *""'"); 
  }
  arg4 = reinterpret_cast< timeval * >(argp4);
  result = (arg1)->AddBuffer((CString const &)*arg2,(CString const &)*arg3,(timeval const *)arg4);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_AddBuffer__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  size_t result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_AddBuffer" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CChan_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (arg1)->AddBuffer((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_AddBuffer__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  size_t result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_AddBuffer" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->AddBuffer((CString const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_AddBuffer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CChan_AddBuffer", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CChan_AddBuffer__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CChan_AddBuffer__SWIG_5(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CChan_AddBuffer__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CChan_AddBuffer__SWIG_4(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_timeval, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CChan_AddBuffer__SWIG_3(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_timeval, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            int res = SWIG_ConvertPtr(argv[4], 0, SWIGTYPE_p_MCString, SWIG_POINTER_NO_NULL | 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_CChan_AddBuffer__SWIG_2(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CChan_AddBuffer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CChan::AddBuffer(CMessage const &,CString const &)\n"
    "    CChan::AddBuffer(CMessage const &)\n"
    "    CChan::AddBuffer(CString const &,CString const &,timeval const *,MCString const &)\n"
    "    CChan::AddBuffer(CString const &,CString const &,timeval const *)\n"
    "    CChan::AddBuffer(CString const &,CString const &)\n"
    "    CChan::AddBuffer(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CChan_ClearBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_ClearBuffer" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  (arg1)->ClearBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SendBuffer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SendBuffer" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_SendBuffer" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  (arg1)->SendBuffer(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SendBuffer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  CBuffer *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SendBuffer" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_SendBuffer" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CBuffer,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CChan_SendBuffer" "', argument " "3"" of type '" "CBuffer const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_SendBuffer" "', argument " "3"" of type '" "CBuffer const &""'"); 
  }
  arg3 = reinterpret_cast< CBuffer * >(argp3);
  (arg1)->SendBuffer(arg2,(CBuffer const &)*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SendBuffer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CChan_SendBuffer", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CClient, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CChan_SendBuffer__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CClient, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CBuffer, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CChan_SendBuffer__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CChan_SendBuffer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CChan::SendBuffer(CClient *)\n"
    "    CChan::SendBuffer(CClient *,CBuffer const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CChan_GetPermStr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetPermStr" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = ((CChan const *)arg1)->GetPermStr();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_HasPerm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_HasPerm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_HasPerm" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_HasPerm" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = (bool)((CChan const *)arg1)->HasPerm(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_AddPerm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_AddPerm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_AddPerm" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_AddPerm" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = (bool)(arg1)->AddPerm(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_RemPerm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_RemPerm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_RemPerm" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_RemPerm" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = (bool)(arg1)->RemPerm(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetModeKnown(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_SetModeKnown", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetModeKnown" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_SetModeKnown" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetModeKnown(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetIsOn(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_SetIsOn", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetIsOn" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_SetIsOn" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetIsOn(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetKey(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_SetKey", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetKey" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_SetKey" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_SetKey" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetKey((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetTopic(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_SetTopic", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetTopic" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_SetTopic" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_SetTopic" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetTopic((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetTopicOwner(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_SetTopicOwner", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetTopicOwner" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_SetTopicOwner" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_SetTopicOwner" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetTopicOwner((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetTopicDate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  unsigned long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_SetTopicDate", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetTopicDate" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_SetTopicDate" "', argument " "2"" of type '" "unsigned long""'");
  } 
  arg2 = static_cast< unsigned long >(val2);
  (arg1)->SetTopicDate(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetDefaultModes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_SetDefaultModes", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetDefaultModes" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CChan_SetDefaultModes" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CChan_SetDefaultModes" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetDefaultModes((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetAutoClearChanBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_SetAutoClearChanBuffer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetAutoClearChanBuffer" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_SetAutoClearChanBuffer" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetAutoClearChanBuffer(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_InheritAutoClearChanBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_InheritAutoClearChanBuffer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_InheritAutoClearChanBuffer" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_InheritAutoClearChanBuffer" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->InheritAutoClearChanBuffer(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_ResetAutoClearChanBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_ResetAutoClearChanBuffer" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  (arg1)->ResetAutoClearChanBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetDetached__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetDetached" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_SetDetached" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetDetached(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetDetached__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetDetached" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  (arg1)->SetDetached();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetDetached(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CChan_SetDetached", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CChan_SetDetached__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CChan_SetDetached__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CChan_SetDetached'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CChan::SetDetached(bool)\n"
    "    CChan::SetDetached()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CChan_SetInConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_SetInConfig", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetInConfig" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_SetInConfig" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetInConfig(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_SetCreationDate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  unsigned long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_SetCreationDate", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_SetCreationDate" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_SetCreationDate" "', argument " "2"" of type '" "unsigned long""'");
  } 
  arg2 = static_cast< unsigned long >(val2);
  (arg1)->SetCreationDate(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_Disable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_Disable" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  (arg1)->Disable();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_Enable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_Enable" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  (arg1)->Enable();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_IncJoinTries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_IncJoinTries" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  (arg1)->IncJoinTries();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_ResetJoinTries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_ResetJoinTries" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  (arg1)->ResetJoinTries();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CIRCNetwork *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetNetwork" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (CIRCNetwork *)((CChan const *)arg1)->GetNetwork();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_IsModeKnown(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_IsModeKnown" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (bool)((CChan const *)arg1)->IsModeKnown();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_HasMode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CChan_HasMode", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_HasMode" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_HasMode" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = (bool)((CChan const *)arg1)->HasMode(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetOptions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetOptions" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = ((CChan const *)arg1)->GetOptions();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetModeArg__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetModeArg" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CChan_GetModeArg" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = ((CChan const *)arg1)->GetModeArg(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetModeArg(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CChan_GetModeArg", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_char(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CChan_GetModeArg__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CChan, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        String* p;
        _v = SWIG_IsOK(SWIG_ConvertPtr(argv[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
      }
      if (_v) {
        return _wrap_CChan_GetModeArg__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CChan_GetModeArg'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CChan::GetModeArg(CString &) const\n"
    "    CChan::GetModeArg(char) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CChan_GetPermCounts(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::map< char,unsigned int,std::less< char >,std::allocator< std::pair< char const,unsigned int > > > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetPermCounts" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = ((CChan const *)arg1)->GetPermCounts();
  resultobj = SWIG_NewPointerObj((new std::map< char,unsigned int,std::less< char >,std::allocator< std::pair< char const,unsigned int > > >(result)), SWIGTYPE_p_std__mapT_char_unsigned_int_std__lessT_char_t_std__allocatorT_std__pairT_char_const_unsigned_int_t_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_IsOn(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_IsOn" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (bool)((CChan const *)arg1)->IsOn();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetName" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (CString *) &((CChan const *)arg1)->GetName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetModes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< char,CString,std::less< char >,std::allocator< std::pair< char const,CString > > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetModes" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (std::map< char,CString,std::less< char >,std::allocator< std::pair< char const,CString > > > *) &((CChan const *)arg1)->GetModes();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_char_CString_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CString_t_t_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetKey(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetKey" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (CString *) &((CChan const *)arg1)->GetKey();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetTopic(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetTopic" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (CString *) &((CChan const *)arg1)->GetTopic();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetTopicOwner(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetTopicOwner" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (CString *) &((CChan const *)arg1)->GetTopicOwner();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetTopicDate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetTopicDate" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (unsigned long)((CChan const *)arg1)->GetTopicDate();
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetDefaultModes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetDefaultModes" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (CString *) &((CChan const *)arg1)->GetDefaultModes();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetNicks(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetNicks" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > > *) &((CChan const *)arg1)->GetNicks();
  resultobj = swig::from(static_cast< std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetNickCount(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  size_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetNickCount" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = ((CChan const *)arg1)->GetNickCount();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_AutoClearChanBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_AutoClearChanBuffer" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (bool)((CChan const *)arg1)->AutoClearChanBuffer();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_IsDetached(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_IsDetached" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (bool)((CChan const *)arg1)->IsDetached();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_InConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_InConfig" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (bool)((CChan const *)arg1)->InConfig();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetCreationDate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetCreationDate" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (unsigned long)((CChan const *)arg1)->GetCreationDate();
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_IsDisabled(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_IsDisabled" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (bool)((CChan const *)arg1)->IsDisabled();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetJoinTries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetJoinTries" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (unsigned int)((CChan const *)arg1)->GetJoinTries();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_HasBufferCountSet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_HasBufferCountSet" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (bool)((CChan const *)arg1)->HasBufferCountSet();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_HasAutoClearChanBufferSet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_HasAutoClearChanBufferSet" "', argument " "1"" of type '" "CChan const *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = (bool)((CChan const *)arg1)->HasAutoClearChanBufferSet();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan___str__" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = CChan___str__(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan___repr__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan___repr__" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = CChan___repr__(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CChan_GetNicks_(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CChan *arg1 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CChan_GetNicks_" "', argument " "1"" of type '" "CChan *""'"); 
  }
  arg1 = reinterpret_cast< CChan * >(argp1);
  result = CChan_GetNicks_(arg1);
  resultobj = swig::from(static_cast< std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CChan_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CChan, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CChan_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  CUser *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CUser" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CUser" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (CUser *)new CUser((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CUser, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CUser" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_ParseConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CConfig *arg2 = (CConfig *) 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_ParseConfig", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_ParseConfig" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_ParseConfig" "', argument " "2"" of type '" "CConfig *""'"); 
  }
  arg2 = reinterpret_cast< CConfig * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sError");
    }
  }
  result = (bool)(arg1)->ParseConfig(arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SaltedHash(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SaltedHash", 2, 2, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SaltedHash" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SaltedHash" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SaltedHash" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SaltedHash" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = CUser::SaltedHash((CString const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_ToConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CConfig result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_ToConfig" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = ((CUser const *)arg1)->ToConfig();
  resultobj = SWIG_NewPointerObj((new CConfig(result)), SWIGTYPE_p_CConfig, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_CheckPass(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_CheckPass", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_CheckPass" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_CheckPass" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_CheckPass" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->CheckPass((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_AddAllowedHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_AddAllowedHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_AddAllowedHost" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_AddAllowedHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_AddAllowedHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->AddAllowedHost((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_RemAllowedHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_RemAllowedHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_RemAllowedHost" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_RemAllowedHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_RemAllowedHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->RemAllowedHost((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_ClearAllowedHosts(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_ClearAllowedHosts" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  (arg1)->ClearAllowedHosts();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_IsHostAllowed(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_IsHostAllowed", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_IsHostAllowed" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_IsHostAllowed" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_IsHostAllowed" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)((CUser const *)arg1)->IsHostAllowed((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_IsValid__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_IsValid" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sErrMsg");
    }
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CUser_IsValid" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)((CUser const *)arg1)->IsValid(*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_IsValid__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_IsValid" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sErrMsg");
    }
  }
  result = (bool)((CUser const *)arg1)->IsValid(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_IsValid(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_IsValid", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        String* p;
        _v = SWIG_IsOK(SWIG_ConvertPtr(argv[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
      }
      if (_v) {
        return _wrap_CUser_IsValid__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        String* p;
        _v = SWIG_IsOK(SWIG_ConvertPtr(argv[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CUser_IsValid__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_IsValid'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::IsValid(CString &,bool) const\n"
    "    CUser::IsValid(CString &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_IsValidUsername(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_IsValidUsername" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_IsValidUsername" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CUser::IsValidUsername((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_IsValidUserName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_IsValidUserName" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_IsValidUserName" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CUser::IsValidUserName((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_MakeCleanUsername(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_MakeCleanUsername" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_MakeCleanUsername" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = CUser::MakeCleanUsername((CString const &)*arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_MakeCleanUserName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_MakeCleanUserName" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_MakeCleanUserName" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = CUser::MakeCleanUserName((CString const &)*arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetModules__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CModules *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetModules" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CModules *) &(arg1)->GetModules();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModules, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetModules__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CModules *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetModules" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CModules *) &((CUser const *)arg1)->GetModules();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModules, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetModules(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_GetModules", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CUser_GetModules__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CUser_GetModules__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_GetModules'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::GetModules()\n"
    "    CUser::GetModules() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_AddNetwork__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CIRCNetwork *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_AddNetwork" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_AddNetwork" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_AddNetwork" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sErrorRet");
    }
  }
  result = (CIRCNetwork *)(arg1)->AddNetwork((CString const &)*arg2,*arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_DeleteNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_DeleteNetwork", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_DeleteNetwork" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_DeleteNetwork" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_DeleteNetwork" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->DeleteNetwork((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_AddNetwork__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_AddNetwork" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_AddNetwork" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  result = (bool)(arg1)->AddNetwork(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_AddNetwork(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_AddNetwork", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CUser_AddNetwork__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          return _wrap_CUser_AddNetwork__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_AddNetwork'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::AddNetwork(CString const &,CString &)\n"
    "    CUser::AddNetwork(CIRCNetwork *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_RemoveNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_RemoveNetwork", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_RemoveNetwork" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_RemoveNetwork" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  (arg1)->RemoveNetwork(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_FindNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CIRCNetwork *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_FindNetwork", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_FindNetwork" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_FindNetwork" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_FindNetwork" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CIRCNetwork *)((CUser const *)arg1)->FindNetwork((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetNetworks(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetNetworks" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > *) &((CUser const *)arg1)->GetNetworks();
  resultobj = swig::from(static_cast< std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_HasSpaceForNewNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_HasSpaceForNewNetwork" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->HasSpaceForNewNetwork();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutUser__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  CClient *arg4 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutUser" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutUser" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CUser_PutUser" "', argument " "4"" of type '" "CClient *""'"); 
  }
  arg4 = reinterpret_cast< CClient * >(argp4);
  result = (bool)(arg1)->PutUser((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutUser__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutUser" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutUser" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (bool)(arg1)->PutUser((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutUser__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutUser" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PutUser((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutUser(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_PutUser", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CUser_PutUser__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CUser_PutUser__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CUser_PutUser__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_PutUser'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::PutUser(CString const &,CClient *,CClient *)\n"
    "    CUser::PutUser(CString const &,CClient *)\n"
    "    CUser::PutUser(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_PutAllUser__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  CClient *arg4 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutAllUser" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutAllUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutAllUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutAllUser" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CUser_PutAllUser" "', argument " "4"" of type '" "CClient *""'"); 
  }
  arg4 = reinterpret_cast< CClient * >(argp4);
  result = (bool)(arg1)->PutAllUser((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutAllUser__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutAllUser" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutAllUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutAllUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutAllUser" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (bool)(arg1)->PutAllUser((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutAllUser__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutAllUser" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutAllUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutAllUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PutAllUser((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutAllUser(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_PutAllUser", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CUser_PutAllUser__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CUser_PutAllUser__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CUser_PutAllUser__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_PutAllUser'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::PutAllUser(CString const &,CClient *,CClient *)\n"
    "    CUser::PutAllUser(CString const &,CClient *)\n"
    "    CUser::PutAllUser(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_PutStatus__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  CClient *arg4 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutStatus" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutStatus" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CUser_PutStatus" "', argument " "4"" of type '" "CClient *""'"); 
  }
  arg4 = reinterpret_cast< CClient * >(argp4);
  result = (bool)(arg1)->PutStatus((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutStatus__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutStatus" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutStatus" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (bool)(arg1)->PutStatus((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutStatus__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutStatus" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PutStatus((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutStatus(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_PutStatus", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CUser_PutStatus__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CUser_PutStatus__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CUser_PutStatus__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_PutStatus'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::PutStatus(CString const &,CClient *,CClient *)\n"
    "    CUser::PutStatus(CString const &,CClient *)\n"
    "    CUser::PutStatus(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_PutStatusNotice__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  CClient *arg4 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutStatusNotice" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutStatusNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutStatusNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutStatusNotice" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CUser_PutStatusNotice" "', argument " "4"" of type '" "CClient *""'"); 
  }
  arg4 = reinterpret_cast< CClient * >(argp4);
  result = (bool)(arg1)->PutStatusNotice((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutStatusNotice__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutStatusNotice" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutStatusNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutStatusNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutStatusNotice" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (bool)(arg1)->PutStatusNotice((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutStatusNotice__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutStatusNotice" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutStatusNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutStatusNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PutStatusNotice((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutStatusNotice(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_PutStatusNotice", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CUser_PutStatusNotice__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CUser_PutStatusNotice__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CUser_PutStatusNotice__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_PutStatusNotice'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::PutStatusNotice(CString const &,CClient *,CClient *)\n"
    "    CUser::PutStatusNotice(CString const &,CClient *)\n"
    "    CUser::PutStatusNotice(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_PutModule__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CClient *arg4 = (CClient *) 0 ;
  CClient *arg5 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutModule" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CUser_PutModule" "', argument " "4"" of type '" "CClient *""'"); 
  }
  arg4 = reinterpret_cast< CClient * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CUser_PutModule" "', argument " "5"" of type '" "CClient *""'"); 
  }
  arg5 = reinterpret_cast< CClient * >(argp5);
  result = (bool)(arg1)->PutModule((CString const &)*arg2,(CString const &)*arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutModule__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CClient *arg4 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutModule" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CUser_PutModule" "', argument " "4"" of type '" "CClient *""'"); 
  }
  arg4 = reinterpret_cast< CClient * >(argp4);
  result = (bool)(arg1)->PutModule((CString const &)*arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutModule__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutModule" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->PutModule((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutModule(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_PutModule", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CUser_PutModule__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CUser_PutModule__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            void *vptr = 0;
            int res = SWIG_ConvertPtr(argv[4], &vptr, SWIGTYPE_p_CClient, 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_CUser_PutModule__SWIG_0(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_PutModule'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::PutModule(CString const &,CString const &,CClient *,CClient *)\n"
    "    CUser::PutModule(CString const &,CString const &,CClient *)\n"
    "    CUser::PutModule(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_PutModNotice__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CClient *arg4 = (CClient *) 0 ;
  CClient *arg5 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutModNotice" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutModNotice" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutModNotice" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CUser_PutModNotice" "', argument " "4"" of type '" "CClient *""'"); 
  }
  arg4 = reinterpret_cast< CClient * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CUser_PutModNotice" "', argument " "5"" of type '" "CClient *""'"); 
  }
  arg5 = reinterpret_cast< CClient * >(argp5);
  result = (bool)(arg1)->PutModNotice((CString const &)*arg2,(CString const &)*arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutModNotice__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CClient *arg4 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutModNotice" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutModNotice" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutModNotice" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CUser_PutModNotice" "', argument " "4"" of type '" "CClient *""'"); 
  }
  arg4 = reinterpret_cast< CClient * >(argp4);
  result = (bool)(arg1)->PutModNotice((CString const &)*arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutModNotice__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_PutModNotice" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_PutModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_PutModNotice" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_PutModNotice" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->PutModNotice((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_PutModNotice(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_PutModNotice", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CUser_PutModNotice__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CUser_PutModNotice__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            void *vptr = 0;
            int res = SWIG_ConvertPtr(argv[4], &vptr, SWIGTYPE_p_CClient, 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_CUser_PutModNotice__SWIG_0(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_PutModNotice'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::PutModNotice(CString const &,CString const &,CClient *,CClient *)\n"
    "    CUser::PutModNotice(CString const &,CString const &,CClient *)\n"
    "    CUser::PutModNotice(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_IsUserAttached(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_IsUserAttached" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->IsUserAttached();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_UserConnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_UserConnected", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_UserConnected" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_UserConnected" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  (arg1)->UserConnected(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_UserDisconnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_UserDisconnected", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_UserDisconnected" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_UserDisconnected" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  (arg1)->UserDisconnected(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetLocalDCCIP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetLocalDCCIP" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = ((CUser const *)arg1)->GetLocalDCCIP();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_ExpandString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_ExpandString" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_ExpandString" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_ExpandString" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CUser const *)arg1)->ExpandString((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_ExpandString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_ExpandString" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_ExpandString" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_ExpandString" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sRet");
    }
  }
  result = (CString *) &((CUser const *)arg1)->ExpandString((CString const &)*arg2,*arg3);
  {
    if (result) {
      resultobj = CPyRetString::wrap(*result);
    } else {
      resultobj = Py_None;
      Py_INCREF(Py_None);
    }
  }
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_ExpandString(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_ExpandString", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CUser_ExpandString__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          return _wrap_CUser_ExpandString__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_ExpandString'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::ExpandString(CString const &) const\n"
    "    CUser::ExpandString(CString const &,CString &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_AddTimestamp__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_AddTimestamp" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_AddTimestamp" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_AddTimestamp" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CUser const *)arg1)->AddTimestamp((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_AddTimestamp__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  time_t arg2 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_AddTimestamp" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_time_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_AddTimestamp" "', argument " "2"" of type '" "time_t""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_AddTimestamp" "', argument " "2"" of type '" "time_t""'");
    } else {
      time_t * temp = reinterpret_cast< time_t * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_AddTimestamp" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_AddTimestamp" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = ((CUser const *)arg1)->AddTimestamp(arg2,(CString const &)*arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_AddTimestamp__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  timeval arg2 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_AddTimestamp" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_timeval,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_AddTimestamp" "', argument " "2"" of type '" "timeval""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_AddTimestamp" "', argument " "2"" of type '" "timeval""'");
    } else {
      timeval * temp = reinterpret_cast< timeval * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_AddTimestamp" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_AddTimestamp" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = ((CUser const *)arg1)->AddTimestamp(arg2,(CString const &)*arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_AddTimestamp(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_AddTimestamp", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CUser_AddTimestamp__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_time_t, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CUser_AddTimestamp__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_timeval, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CUser_AddTimestamp__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_AddTimestamp'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::AddTimestamp(CString const &) const\n"
    "    CUser::AddTimestamp(time_t,CString const &) const\n"
    "    CUser::AddTimestamp(timeval,CString const &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_CloneNetworks(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CUser *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_CloneNetworks", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_CloneNetworks" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_CloneNetworks" "', argument " "2"" of type '" "CUser const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_CloneNetworks" "', argument " "2"" of type '" "CUser const &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  (arg1)->CloneNetworks((CUser const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_Clone__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CUser *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_Clone" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_Clone" "', argument " "2"" of type '" "CUser const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_Clone" "', argument " "2"" of type '" "CUser const &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sErrorRet");
    }
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CUser_Clone" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->Clone((CUser const &)*arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_Clone__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CUser *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_Clone" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_Clone" "', argument " "2"" of type '" "CUser const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_Clone" "', argument " "2"" of type '" "CUser const &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sErrorRet");
    }
  }
  result = (bool)(arg1)->Clone((CUser const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_Clone(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_Clone", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CUser, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          return _wrap_CUser_Clone__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CUser, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CUser_Clone__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_Clone'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::Clone(CUser const &,CString &,bool)\n"
    "    CUser::Clone(CUser const &,CString &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_BounceAllClients(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_BounceAllClients" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  (arg1)->BounceAllClients();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_AddBytesRead(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  unsigned long long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_AddBytesRead", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_AddBytesRead" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_AddBytesRead" "', argument " "2"" of type '" "unsigned long long""'");
  } 
  arg2 = static_cast< unsigned long long >(val2);
  (arg1)->AddBytesRead(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_AddBytesWritten(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  unsigned long long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_AddBytesWritten", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_AddBytesWritten" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_AddBytesWritten" "', argument " "2"" of type '" "unsigned long long""'");
  } 
  arg2 = static_cast< unsigned long long >(val2);
  (arg1)->AddBytesWritten(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetNick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetNick" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetNick((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetAltNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetAltNick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetAltNick" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetAltNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetAltNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetAltNick((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetIdent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetIdent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetIdent" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetIdent" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetIdent" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetIdent((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetRealName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetRealName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetRealName" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetRealName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetRealName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetRealName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetBindHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetBindHost" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetBindHost((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetDCCBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetDCCBindHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetDCCBindHost" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetDCCBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetDCCBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetDCCBindHost((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetPass__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CUser::eHashType arg3 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetPass" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetPass" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetPass" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CUser_SetPass" "', argument " "3"" of type '" "CUser::eHashType""'");
  } 
  arg3 = static_cast< CUser::eHashType >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CUser_SetPass" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetPass" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->SetPass((CString const &)*arg2,arg3,(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetPass__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CUser::eHashType arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetPass" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetPass" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetPass" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CUser_SetPass" "', argument " "3"" of type '" "CUser::eHashType""'");
  } 
  arg3 = static_cast< CUser::eHashType >(val3);
  (arg1)->SetPass((CString const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetPass(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_SetPass", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CUser_SetPass__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CUser_SetPass__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_SetPass'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::SetPass(CString const &,CUser::eHashType,CString const &)\n"
    "    CUser::SetPass(CString const &,CUser::eHashType)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_SetMultiClients(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetMultiClients", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetMultiClients" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetMultiClients" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetMultiClients(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetDenyLoadMod(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetDenyLoadMod", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetDenyLoadMod" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetDenyLoadMod" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetDenyLoadMod(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetAdmin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetAdmin", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetAdmin" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetAdmin" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetAdmin(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetDenySetBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetDenySetBindHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetDenySetBindHost" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetDenySetBindHost" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetDenySetBindHost(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetDenySetIdent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetDenySetIdent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetDenySetIdent" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetDenySetIdent" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetDenySetIdent(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetDenySetNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetDenySetNetwork", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetDenySetNetwork" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetDenySetNetwork" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetDenySetNetwork(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetDenySetRealName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetDenySetRealName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetDenySetRealName" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetDenySetRealName" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetDenySetRealName(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetDenySetQuitMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetDenySetQuitMsg", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetDenySetQuitMsg" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetDenySetQuitMsg" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetDenySetQuitMsg(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetDenySetCTCPReplies(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetDenySetCTCPReplies", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetDenySetCTCPReplies" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetDenySetCTCPReplies" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetDenySetCTCPReplies(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetStatusPrefix(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetStatusPrefix", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetStatusPrefix" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetStatusPrefix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetStatusPrefix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->SetStatusPrefix((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetDefaultChanModes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetDefaultChanModes", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetDefaultChanModes" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetDefaultChanModes" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetDefaultChanModes" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetDefaultChanModes((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetClientEncoding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetClientEncoding", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetClientEncoding" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetClientEncoding" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetClientEncoding" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetClientEncoding((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetQuitMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetQuitMsg", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetQuitMsg" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetQuitMsg" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetQuitMsg" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetQuitMsg((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_AddCTCPReply(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_AddCTCPReply", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_AddCTCPReply" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_AddCTCPReply" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_AddCTCPReply" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CUser_AddCTCPReply" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_AddCTCPReply" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->AddCTCPReply((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_DelCTCPReply(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_DelCTCPReply", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_DelCTCPReply" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_DelCTCPReply" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_DelCTCPReply" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->DelCTCPReply((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetBufferCount__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  unsigned int arg2 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetBufferCount" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetBufferCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CUser_SetBufferCount" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)(arg1)->SetBufferCount(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetBufferCount__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetBufferCount" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetBufferCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = (bool)(arg1)->SetBufferCount(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetBufferCount(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_SetBufferCount", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CUser_SetBufferCount__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CUser_SetBufferCount__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_SetBufferCount'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::SetBufferCount(unsigned int,bool)\n"
    "    CUser::SetBufferCount(unsigned int)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_SetChanBufferSize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  unsigned int arg2 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetChanBufferSize" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetChanBufferSize" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CUser_SetChanBufferSize" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)(arg1)->SetChanBufferSize(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetChanBufferSize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetChanBufferSize" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetChanBufferSize" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = (bool)(arg1)->SetChanBufferSize(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetChanBufferSize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_SetChanBufferSize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CUser_SetChanBufferSize__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CUser_SetChanBufferSize__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_SetChanBufferSize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::SetChanBufferSize(unsigned int,bool)\n"
    "    CUser::SetChanBufferSize(unsigned int)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_SetQueryBufferSize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  unsigned int arg2 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetQueryBufferSize" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetQueryBufferSize" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CUser_SetQueryBufferSize" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)(arg1)->SetQueryBufferSize(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetQueryBufferSize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetQueryBufferSize" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetQueryBufferSize" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = (bool)(arg1)->SetQueryBufferSize(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetQueryBufferSize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_SetQueryBufferSize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CUser_SetQueryBufferSize__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CUser_SetQueryBufferSize__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_SetQueryBufferSize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::SetQueryBufferSize(unsigned int,bool)\n"
    "    CUser::SetQueryBufferSize(unsigned int)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_SetAutoClearChanBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetAutoClearChanBuffer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetAutoClearChanBuffer" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetAutoClearChanBuffer" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetAutoClearChanBuffer(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetAutoClearQueryBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetAutoClearQueryBuffer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetAutoClearQueryBuffer" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetAutoClearQueryBuffer" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetAutoClearQueryBuffer(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetLanguage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetLanguage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetLanguage" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetLanguage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetLanguage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->SetLanguage((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetBeingDeleted(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetBeingDeleted", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetBeingDeleted" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetBeingDeleted" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetBeingDeleted(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetTimestampFormat(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetTimestampFormat", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetTimestampFormat" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetTimestampFormat" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetTimestampFormat" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetTimestampFormat((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetTimestampAppend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetTimestampAppend", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetTimestampAppend" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetTimestampAppend" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetTimestampAppend(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetTimestampPrepend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetTimestampPrepend", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetTimestampPrepend" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetTimestampPrepend" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetTimestampPrepend(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetAuthOnlyViaModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetAuthOnlyViaModule", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetAuthOnlyViaModule" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetAuthOnlyViaModule" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetAuthOnlyViaModule(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetTimezone(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetTimezone", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetTimezone" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetTimezone" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetTimezone" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetTimezone((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetJoinTries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetJoinTries", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetJoinTries" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetJoinTries" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  (arg1)->SetJoinTries(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetMaxJoins(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetMaxJoins", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetMaxJoins" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetMaxJoins" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  (arg1)->SetMaxJoins(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetSkinName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetSkinName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetSkinName" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CUser_SetSkinName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CUser_SetSkinName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetSkinName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetMaxNetworks(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetMaxNetworks", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetMaxNetworks" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetMaxNetworks" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  (arg1)->SetMaxNetworks(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetMaxQueryBuffers(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetMaxQueryBuffers", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetMaxQueryBuffers" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetMaxQueryBuffers" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  (arg1)->SetMaxQueryBuffers(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_SetNoTrafficTimeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CUser_SetNoTrafficTimeout", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_SetNoTrafficTimeout" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_SetNoTrafficTimeout" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  (arg1)->SetNoTrafficTimeout(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetUserClients(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient *,std::allocator< CClient * > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetUserClients" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (std::vector< CClient *,std::allocator< CClient * > > *) &((CUser const *)arg1)->GetUserClients();
  resultobj = swig::from(static_cast< std::vector< CClient*,std::allocator< CClient * > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetAllClients(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient *,std::allocator< CClient * > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetAllClients" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = ((CUser const *)arg1)->GetAllClients();
  resultobj = swig::from(static_cast< std::vector< CClient*,std::allocator< CClient * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetUserName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetUserName" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetUserName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetUsername(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetUsername" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetUsername();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetCleanUserName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetCleanUserName" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetCleanUserName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetNick__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetNick" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_GetNick" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (CString *) &((CUser const *)arg1)->GetNick(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetNick__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetNick" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetNick();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetNick(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_GetNick", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CUser_GetNick__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CUser_GetNick__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_GetNick'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::GetNick(bool) const\n"
    "    CUser::GetNick() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_GetAltNick__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetAltNick" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_GetAltNick" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (CString *) &((CUser const *)arg1)->GetAltNick(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetAltNick__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetAltNick" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetAltNick();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetAltNick(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_GetAltNick", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CUser_GetAltNick__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CUser_GetAltNick__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_GetAltNick'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::GetAltNick(bool) const\n"
    "    CUser::GetAltNick() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_GetIdent__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetIdent" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CUser_GetIdent" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (CString *) &((CUser const *)arg1)->GetIdent(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetIdent__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetIdent" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetIdent();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetIdent(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CUser_GetIdent", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CUser_GetIdent__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CUser_GetIdent__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CUser_GetIdent'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CUser::GetIdent(bool) const\n"
    "    CUser::GetIdent() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CUser_GetRealName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetRealName" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = ((CUser const *)arg1)->GetRealName();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetBindHost" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetBindHost();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetDCCBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetDCCBindHost" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetDCCBindHost();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetPass(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetPass" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetPass();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetPassHashType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CUser::eHashType result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetPassHashType" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CUser::eHashType)((CUser const *)arg1)->GetPassHashType();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetPassSalt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetPassSalt" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetPassSalt();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetAllowedHosts(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< CString,std::less< CString >,std::allocator< CString > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetAllowedHosts" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (std::set< CString,std::less< CString >,std::allocator< CString > > *) &((CUser const *)arg1)->GetAllowedHosts();
  resultobj = swig::from(static_cast< std::set< CString,std::less< CString >,std::allocator< CString > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetTimestampFormat(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetTimestampFormat" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetTimestampFormat();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetClientEncoding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetClientEncoding" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetClientEncoding();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetTimestampAppend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetTimestampAppend" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->GetTimestampAppend();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetTimestampPrepend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetTimestampPrepend" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->GetTimestampPrepend();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetUserPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetUserPath" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetUserPath();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_DenyLoadMod(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_DenyLoadMod" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->DenyLoadMod();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_IsAdmin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_IsAdmin" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->IsAdmin();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_DenySetBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_DenySetBindHost" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->DenySetBindHost();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_DenySetIdent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_DenySetIdent" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->DenySetIdent();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_DenySetNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_DenySetNetwork" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->DenySetNetwork();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_DenySetRealName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_DenySetRealName" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->DenySetRealName();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_DenySetQuitMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_DenySetQuitMsg" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->DenySetQuitMsg();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_DenySetCTCPReplies(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_DenySetCTCPReplies" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->DenySetCTCPReplies();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_MultiClients(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_MultiClients" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->MultiClients();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_AuthOnlyViaModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_AuthOnlyViaModule" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->AuthOnlyViaModule();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetStatusPrefix(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetStatusPrefix" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetStatusPrefix();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetDefaultChanModes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetDefaultChanModes" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (CString *) &((CUser const *)arg1)->GetDefaultChanModes();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetPingFrequency(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetPingFrequency" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (unsigned int)((CUser const *)arg1)->GetPingFrequency();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetPingSlack(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetPingSlack" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (unsigned int)((CUser const *)arg1)->GetPingSlack();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetNoTrafficTimeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetNoTrafficTimeout" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (unsigned int)((CUser const *)arg1)->GetNoTrafficTimeout();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetQuitMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetQuitMsg" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = ((CUser const *)arg1)->GetQuitMsg();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetCTCPReplies(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  MCString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetCTCPReplies" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (MCString *) &((CUser const *)arg1)->GetCTCPReplies();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_MCString, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetBufferCount(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetBufferCount" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (unsigned int)((CUser const *)arg1)->GetBufferCount();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetChanBufferSize(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetChanBufferSize" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (unsigned int)((CUser const *)arg1)->GetChanBufferSize();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetQueryBufferSize(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetQueryBufferSize" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (unsigned int)((CUser const *)arg1)->GetQueryBufferSize();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_AutoClearChanBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_AutoClearChanBuffer" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->AutoClearChanBuffer();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_AutoClearQueryBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_AutoClearQueryBuffer" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->AutoClearQueryBuffer();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_IsBeingDeleted(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_IsBeingDeleted" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (bool)((CUser const *)arg1)->IsBeingDeleted();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetTimezone(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetTimezone" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = ((CUser const *)arg1)->GetTimezone();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_BytesRead(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long long result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_BytesRead" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (unsigned long long)((CUser const *)arg1)->BytesRead();
  resultobj = SWIG_From_unsigned_SS_long_SS_long(static_cast< unsigned long long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_BytesWritten(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long long result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_BytesWritten" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (unsigned long long)((CUser const *)arg1)->BytesWritten();
  resultobj = SWIG_From_unsigned_SS_long_SS_long(static_cast< unsigned long long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_JoinTries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_JoinTries" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (unsigned int)((CUser const *)arg1)->JoinTries();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_MaxJoins(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_MaxJoins" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (unsigned int)((CUser const *)arg1)->MaxJoins();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetSkinName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetSkinName" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = ((CUser const *)arg1)->GetSkinName();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetLanguage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetLanguage" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = ((CUser const *)arg1)->GetLanguage();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_MaxNetworks(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_MaxNetworks" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (unsigned int)((CUser const *)arg1)->MaxNetworks();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_MaxQueryBuffers(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_MaxQueryBuffers" "', argument " "1"" of type '" "CUser const *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = (unsigned int)((CUser const *)arg1)->MaxQueryBuffers();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser___str__" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = CUser___str__(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser___repr__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser___repr__" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = CUser___repr__(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUser_GetNetworks_(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUser_GetNetworks_" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  result = CUser_GetNetworks_(arg1);
  resultobj = swig::from(static_cast< std::vector< CIRCNetwork*,std::allocator< CIRCNetwork * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CUser_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CUser, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CUser_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CIRCNetwork_IsValidNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_IsValidNetwork" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_IsValidNetwork" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CIRCNetwork::IsValidNetwork((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CIRCNetwork__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CIRCNetwork *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CIRCNetwork" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CIRCNetwork" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CIRCNetwork" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CIRCNetwork *)new CIRCNetwork(arg1,(CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CIRCNetwork__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CIRCNetwork *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  CIRCNetwork *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CIRCNetwork" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CIRCNetwork,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CIRCNetwork" "', argument " "2"" of type '" "CIRCNetwork const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CIRCNetwork" "', argument " "2"" of type '" "CIRCNetwork const &""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  result = (CIRCNetwork *)new CIRCNetwork(arg1,(CIRCNetwork const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CIRCNetwork(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CIRCNetwork", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CIRCNetwork, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_CIRCNetwork__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CUser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_CIRCNetwork__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CIRCNetwork'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::CIRCNetwork(CUser *,CString const &)\n"
    "    CIRCNetwork::CIRCNetwork(CUser *,CIRCNetwork const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CIRCNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CIRCNetwork" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_Clone__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CIRCNetwork *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_Clone" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CIRCNetwork,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_Clone" "', argument " "2"" of type '" "CIRCNetwork const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_Clone" "', argument " "2"" of type '" "CIRCNetwork const &""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CIRCNetwork_Clone" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->Clone((CIRCNetwork const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_Clone__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CIRCNetwork *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_Clone" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CIRCNetwork,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_Clone" "', argument " "2"" of type '" "CIRCNetwork const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_Clone" "', argument " "2"" of type '" "CIRCNetwork const &""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  (arg1)->Clone((CIRCNetwork const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_Clone(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_Clone", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CIRCNetwork, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_Clone__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CIRCNetwork, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CIRCNetwork_Clone__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_Clone'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::Clone(CIRCNetwork const &,bool)\n"
    "    CIRCNetwork::Clone(CIRCNetwork const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetNetworkPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetNetworkPath" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = ((CIRCNetwork const *)arg1)->GetNetworkPath();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_DelServers(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_DelServers" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  (arg1)->DelServers();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ParseConfig__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CConfig *arg2 = (CConfig *) 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_ParseConfig" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_ParseConfig" "', argument " "2"" of type '" "CConfig *""'"); 
  }
  arg2 = reinterpret_cast< CConfig * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sError");
    }
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CIRCNetwork_ParseConfig" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->ParseConfig(arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ParseConfig__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CConfig *arg2 = (CConfig *) 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_ParseConfig" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_ParseConfig" "', argument " "2"" of type '" "CConfig *""'"); 
  }
  arg2 = reinterpret_cast< CConfig * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sError");
    }
  }
  result = (bool)(arg1)->ParseConfig(arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ParseConfig(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_ParseConfig", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CConfig, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          return _wrap_CIRCNetwork_ParseConfig__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CConfig, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CIRCNetwork_ParseConfig__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_ParseConfig'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::ParseConfig(CConfig *,CString &,bool)\n"
    "    CIRCNetwork::ParseConfig(CConfig *,CString &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ToConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CConfig result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_ToConfig" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = ((CIRCNetwork const *)arg1)->ToConfig();
  resultobj = SWIG_NewPointerObj((new CConfig(result)), SWIGTYPE_p_CConfig, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_BounceAllClients(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_BounceAllClients" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  (arg1)->BounceAllClients();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_IsUserAttached(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_IsUserAttached" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (bool)((CIRCNetwork const *)arg1)->IsUserAttached();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_IsUserOnline(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_IsUserOnline" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (bool)((CIRCNetwork const *)arg1)->IsUserOnline();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ClientConnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_ClientConnected", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_ClientConnected" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_ClientConnected" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  (arg1)->ClientConnected(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ClientDisconnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_ClientDisconnected", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_ClientDisconnected" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_ClientDisconnected" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  (arg1)->ClientDisconnected(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CUser *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetUser" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CUser *)((CIRCNetwork const *)arg1)->GetUser();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CUser, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetName" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CString *) &((CIRCNetwork const *)arg1)->GetName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_IsNetworkAttached(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_IsNetworkAttached" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (bool)((CIRCNetwork const *)arg1)->IsNetworkAttached();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetClients(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient *,std::allocator< CClient * > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetClients" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (std::vector< CClient *,std::allocator< CClient * > > *) &((CIRCNetwork const *)arg1)->GetClients();
  resultobj = swig::from(static_cast< std::vector< CClient*,std::allocator< CClient * > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_FindClients(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::vector< CClient *,std::allocator< CClient * > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_FindClients", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_FindClients" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_FindClients" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_FindClients" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CIRCNetwork const *)arg1)->FindClients((CString const &)*arg2);
  resultobj = swig::from(static_cast< std::vector< CClient*,std::allocator< CClient * > > >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CUser *arg2 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetUser", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetUser" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetUser" "', argument " "2"" of type '" "CUser *""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  (arg1)->SetUser(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetName" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_SetName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->SetName((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetModules__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CModules *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetModules" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CModules *) &(arg1)->GetModules();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModules, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetModules__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CModules *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetModules" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CModules *) &((CIRCNetwork const *)arg1)->GetModules();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModules, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetModules(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_GetModules", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CIRCNetwork_GetModules__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CIRCNetwork_GetModules__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_GetModules'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::GetModules()\n"
    "    CIRCNetwork::GetModules() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutUser__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  CClient *arg4 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutUser" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_PutUser" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_PutUser" "', argument " "4"" of type '" "CClient *""'"); 
  }
  arg4 = reinterpret_cast< CClient * >(argp4);
  result = (bool)(arg1)->PutUser((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutUser__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutUser" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_PutUser" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (bool)(arg1)->PutUser((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutUser__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutUser" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PutUser((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutUser__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CMessage *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  CClient *arg4 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutUser" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutUser" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutUser" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_PutUser" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_PutUser" "', argument " "4"" of type '" "CClient *""'"); 
  }
  arg4 = reinterpret_cast< CClient * >(argp4);
  result = (bool)(arg1)->PutUser((CMessage const &)*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutUser__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CMessage *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutUser" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutUser" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutUser" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_PutUser" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (bool)(arg1)->PutUser((CMessage const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutUser__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutUser" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutUser" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutUser" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (bool)(arg1)->PutUser((CMessage const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutUser(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_PutUser", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_PutUser__SWIG_5(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_PutUser__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_PutUser__SWIG_4(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_PutUser__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CIRCNetwork_PutUser__SWIG_3(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CIRCNetwork_PutUser__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_PutUser'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::PutUser(CString const &,CClient *,CClient *)\n"
    "    CIRCNetwork::PutUser(CString const &,CClient *)\n"
    "    CIRCNetwork::PutUser(CString const &)\n"
    "    CIRCNetwork::PutUser(CMessage const &,CClient *,CClient *)\n"
    "    CIRCNetwork::PutUser(CMessage const &,CClient *)\n"
    "    CIRCNetwork::PutUser(CMessage const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutStatus__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  CClient *arg4 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutStatus" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_PutStatus" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_PutStatus" "', argument " "4"" of type '" "CClient *""'"); 
  }
  arg4 = reinterpret_cast< CClient * >(argp4);
  result = (bool)(arg1)->PutStatus((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutStatus__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutStatus" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_PutStatus" "', argument " "3"" of type '" "CClient *""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (bool)(arg1)->PutStatus((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutStatus__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutStatus" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PutStatus((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutStatus(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_PutStatus", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_PutStatus__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_PutStatus__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CClient, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CIRCNetwork_PutStatus__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_PutStatus'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::PutStatus(CString const &,CClient *,CClient *)\n"
    "    CIRCNetwork::PutStatus(CString const &,CClient *)\n"
    "    CIRCNetwork::PutStatus(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutModule__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CClient *arg4 = (CClient *) 0 ;
  CClient *arg5 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutModule" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_PutModule" "', argument " "4"" of type '" "CClient *""'"); 
  }
  arg4 = reinterpret_cast< CClient * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CIRCNetwork_PutModule" "', argument " "5"" of type '" "CClient *""'"); 
  }
  arg5 = reinterpret_cast< CClient * >(argp5);
  result = (bool)(arg1)->PutModule((CString const &)*arg2,(CString const &)*arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutModule__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CClient *arg4 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutModule" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_PutModule" "', argument " "4"" of type '" "CClient *""'"); 
  }
  arg4 = reinterpret_cast< CClient * >(argp4);
  result = (bool)(arg1)->PutModule((CString const &)*arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutModule__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutModule" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->PutModule((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutModule(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_PutModule", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_PutModule__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CIRCNetwork_PutModule__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CClient, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            void *vptr = 0;
            int res = SWIG_ConvertPtr(argv[4], &vptr, SWIGTYPE_p_CClient, 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_CIRCNetwork_PutModule__SWIG_0(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_PutModule'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::PutModule(CString const &,CString const &,CClient *,CClient *)\n"
    "    CIRCNetwork::PutModule(CString const &,CString const &,CClient *)\n"
    "    CIRCNetwork::PutModule(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetChans(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CChan *,std::allocator< CChan * > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetChans" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (std::vector< CChan *,std::allocator< CChan * > > *) &((CIRCNetwork const *)arg1)->GetChans();
  resultobj = swig::from(static_cast< std::vector< CChan*,std::allocator< CChan * > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_FindChan(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CChan *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_FindChan", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_FindChan" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    int res = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "CIRCNetwork_FindChan" "', argument " "2"" of type '" "CString""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  result = (CChan *)((CIRCNetwork const *)arg1)->FindChan(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CChan, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_FindChans(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::vector< CChan *,std::allocator< CChan * > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_FindChans", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_FindChans" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_FindChans" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_FindChans" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CIRCNetwork const *)arg1)->FindChans((CString const &)*arg2);
  resultobj = swig::from(static_cast< std::vector< CChan*,std::allocator< CChan * > > >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddChan__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CChan *arg2 = (CChan *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddChan" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CChan, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddChan" "', argument " "2"" of type '" "CChan *""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  result = (bool)(arg1)->AddChan(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddChan__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddChan" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddChan" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddChan" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CIRCNetwork_AddChan" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)(arg1)->AddChan((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddChan(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_AddChan", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CChan, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_AddChan__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CIRCNetwork_AddChan__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_AddChan'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::AddChan(CChan *)\n"
    "    CIRCNetwork::AddChan(CString const &,bool)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_DelChan(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_DelChan", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_DelChan" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_DelChan" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_DelChan" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->DelChan((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_MoveChan(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  unsigned int arg3 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_MoveChan", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_MoveChan" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_MoveChan" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_MoveChan" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CIRCNetwork_MoveChan" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sError");
    }
  }
  result = (bool)(arg1)->MoveChan((CString const &)*arg2,arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SwapChans(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SwapChans", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SwapChans" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SwapChans" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_SwapChans" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_SwapChans" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_SwapChans" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sError");
    }
  }
  result = (bool)(arg1)->SwapChans((CString const &)*arg2,(CString const &)*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_JoinChans__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_JoinChans" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  (arg1)->JoinChans();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_JoinChans__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  std::set< CChan *,std::less< CChan * >,std::allocator< CChan * > > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_JoinChans" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__setT_CChan_p_std__lessT_CChan_p_t_std__allocatorT_CChan_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_JoinChans" "', argument " "2"" of type '" "std::set< CChan *,std::less< CChan * >,std::allocator< CChan * > > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_JoinChans" "', argument " "2"" of type '" "std::set< CChan *,std::less< CChan * >,std::allocator< CChan * > > &""'"); 
  }
  arg2 = reinterpret_cast< std::set< CChan *,std::less< CChan * >,std::allocator< CChan * > > * >(argp2);
  (arg1)->JoinChans(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_JoinChans(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_JoinChans", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CIRCNetwork_JoinChans__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__setT_CChan_p_std__lessT_CChan_p_t_std__allocatorT_CChan_p_t_t, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_JoinChans__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_JoinChans'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::JoinChans()\n"
    "    CIRCNetwork::JoinChans(std::set< CChan *,std::less< CChan * >,std::allocator< CChan * > > &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetQueries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CQuery *,std::allocator< CQuery * > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetQueries" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (std::vector< CQuery *,std::allocator< CQuery * > > *) &((CIRCNetwork const *)arg1)->GetQueries();
  resultobj = swig::from(static_cast< std::vector< CQuery*,std::allocator< CQuery * > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_FindQuery(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CQuery *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_FindQuery", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_FindQuery" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_FindQuery" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_FindQuery" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CQuery *)((CIRCNetwork const *)arg1)->FindQuery((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CQuery, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_FindQueries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::vector< CQuery *,std::allocator< CQuery * > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_FindQueries", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_FindQueries" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_FindQueries" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_FindQueries" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CIRCNetwork const *)arg1)->FindQueries((CString const &)*arg2);
  resultobj = swig::from(static_cast< std::vector< CQuery*,std::allocator< CQuery * > > >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddQuery(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CQuery *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_AddQuery", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddQuery" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddQuery" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddQuery" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CQuery *)(arg1)->AddQuery((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CQuery, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_DelQuery(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_DelQuery", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_DelQuery" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_DelQuery" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_DelQuery" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->DelQuery((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetChanPrefixes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetChanPrefixes" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CString *) &((CIRCNetwork const *)arg1)->GetChanPrefixes();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetChanPrefixes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetChanPrefixes", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetChanPrefixes" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetChanPrefixes" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_SetChanPrefixes" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetChanPrefixes((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_IsChan(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_IsChan", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_IsChan" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_IsChan" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_IsChan" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)((CIRCNetwork const *)arg1)->IsChan((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetServers(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CServer *,std::allocator< CServer * > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetServers" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (std::vector< CServer *,std::allocator< CServer * > > *) &((CIRCNetwork const *)arg1)->GetServers();
  resultobj = swig::from(static_cast< std::vector< CServer*,std::allocator< CServer * > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_HasServers(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_HasServers" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (bool)((CIRCNetwork const *)arg1)->HasServers();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_FindServer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CServer *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_FindServer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_FindServer" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_FindServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_FindServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CServer *)((CIRCNetwork const *)arg1)->FindServer((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CServer, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_DelServer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_DelServer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_DelServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_DelServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CIRCNetwork_DelServer" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_DelServer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_DelServer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (bool)(arg1)->DelServer((CString const &)*arg2,arg3,(CString const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_DelServer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CServer *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_DelServer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CServer,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_DelServer" "', argument " "2"" of type '" "CServer const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_DelServer" "', argument " "2"" of type '" "CServer const &""'"); 
  }
  arg2 = reinterpret_cast< CServer * >(argp2);
  result = (bool)(arg1)->DelServer((CServer const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_DelServer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_DelServer", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CServer, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_DelServer__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CIRCNetwork_DelServer__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_DelServer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::DelServer(CString const &,unsigned short,CString const &)\n"
    "    CIRCNetwork::DelServer(CServer const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddServer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  SwigValueWrapper< CServer > arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddServer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CServer,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddServer" "', argument " "2"" of type '" "CServer""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddServer" "', argument " "2"" of type '" "CServer""'");
    } else {
      CServer * temp = reinterpret_cast< CServer * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  result = (bool)(arg1)->AddServer(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddServer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddServer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->AddServer((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddServer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  CString *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  bool val5 ;
  int ecode5 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddServer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CIRCNetwork_AddServer" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_AddServer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddServer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CIRCNetwork_AddServer" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  result = (bool)(arg1)->AddServer((CString const &)*arg2,arg3,(CString const &)*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddServer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddServer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CIRCNetwork_AddServer" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_AddServer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddServer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (bool)(arg1)->AddServer((CString const &)*arg2,arg3,(CString const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddServer__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddServer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CIRCNetwork_AddServer" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  result = (bool)(arg1)->AddServer((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddServer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_AddServer", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CServer, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_AddServer__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_AddServer__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CIRCNetwork_AddServer__SWIG_4(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CIRCNetwork_AddServer__SWIG_3(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_short(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_bool(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_CIRCNetwork_AddServer__SWIG_2(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_AddServer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::AddServer(CServer)\n"
    "    CIRCNetwork::AddServer(CString const &)\n"
    "    CIRCNetwork::AddServer(CString const &,unsigned short,CString const &,bool)\n"
    "    CIRCNetwork::AddServer(CString const &,unsigned short,CString const &)\n"
    "    CIRCNetwork::AddServer(CString const &,unsigned short)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetNextServer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  CServer *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetNextServer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCNetwork_GetNextServer" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (CServer *)(arg1)->GetNextServer(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CServer, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetNextServer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CServer *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetNextServer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CServer *)(arg1)->GetNextServer();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CServer, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetNextServer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_GetNextServer", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CIRCNetwork_GetNextServer__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CIRCNetwork_GetNextServer__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_GetNextServer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::GetNextServer(bool)\n"
    "    CIRCNetwork::GetNextServer()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetCurrentServer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CServer *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetCurrentServer" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CServer *)((CIRCNetwork const *)arg1)->GetCurrentServer();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CServer, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetIRCServer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetIRCServer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetIRCServer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetIRCServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_SetIRCServer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetIRCServer((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetNextServer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CServer *arg2 = (CServer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetNextServer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetNextServer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetNextServer" "', argument " "2"" of type '" "CServer const *""'"); 
  }
  arg2 = reinterpret_cast< CServer * >(argp2);
  result = (bool)(arg1)->SetNextServer((CServer const *)arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_IsLastServer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_IsLastServer" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (bool)((CIRCNetwork const *)arg1)->IsLastServer();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetTrustedFingerprints(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SCString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetTrustedFingerprints" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (SCString *) &((CIRCNetwork const *)arg1)->GetTrustedFingerprints();
  resultobj = swig::from(static_cast< std::set< CString,std::less< CString >,std::allocator< CString > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddTrustedFingerprint(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_AddTrustedFingerprint", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddTrustedFingerprint" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddTrustedFingerprint" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddTrustedFingerprint" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->AddTrustedFingerprint((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_DelTrustedFingerprint(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_DelTrustedFingerprint", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_DelTrustedFingerprint" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_DelTrustedFingerprint" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_DelTrustedFingerprint" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->DelTrustedFingerprint((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ClearTrustedFingerprints(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_ClearTrustedFingerprints" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  (arg1)->ClearTrustedFingerprints();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetIRCConnectEnabled(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetIRCConnectEnabled", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetIRCConnectEnabled" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCNetwork_SetIRCConnectEnabled" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetIRCConnectEnabled(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetIRCConnectEnabled(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetIRCConnectEnabled" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (bool)((CIRCNetwork const *)arg1)->GetIRCConnectEnabled();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetIRCSock__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CIRCSock *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetIRCSock" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CIRCSock *)(arg1)->GetIRCSock();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCSock, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetIRCSock__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CIRCSock *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetIRCSock" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CIRCSock *)((CIRCNetwork const *)arg1)->GetIRCSock();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCSock, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetIRCSock(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_GetIRCSock", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CIRCNetwork_GetIRCSock__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CIRCNetwork_GetIRCSock__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_GetIRCSock'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::GetIRCSock()\n"
    "    CIRCNetwork::GetIRCSock() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetIRCServer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetIRCServer" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CString *) &((CIRCNetwork const *)arg1)->GetIRCServer();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetIRCNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CNick *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetIRCNick" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CNick *) &((CIRCNetwork const *)arg1)->GetIRCNick();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CNick, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetIRCNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CNick *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetIRCNick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetIRCNick" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetIRCNick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_SetIRCNick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  (arg1)->SetIRCNick((CNick const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetCurNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetCurNick" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = ((CIRCNetwork const *)arg1)->GetCurNick();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_IsIRCAway(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_IsIRCAway" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (bool)((CIRCNetwork const *)arg1)->IsIRCAway();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetIRCAway(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetIRCAway", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetIRCAway" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCNetwork_SetIRCAway" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetIRCAway(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_Connect(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_Connect" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (bool)(arg1)->Connect();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_IsIRCConnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_IsIRCConnected" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (bool)((CIRCNetwork const *)arg1)->IsIRCConnected();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetIRCSocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CIRCSock *arg2 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetIRCSocket", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetIRCSocket" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetIRCSocket" "', argument " "2"" of type '" "CIRCSock *""'"); 
  }
  arg2 = reinterpret_cast< CIRCSock * >(argp2);
  (arg1)->SetIRCSocket(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_IRCConnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_IRCConnected" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  (arg1)->IRCConnected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_IRCDisconnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_IRCDisconnected" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  (arg1)->IRCDisconnected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_CheckIRCConnect(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_CheckIRCConnect" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  (arg1)->CheckIRCConnect();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_NotifyClientsAboutServerDependentCap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_NotifyClientsAboutServerDependentCap", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_NotifyClientsAboutServerDependentCap" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_NotifyClientsAboutServerDependentCap" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_NotifyClientsAboutServerDependentCap" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CIRCNetwork_NotifyClientsAboutServerDependentCap" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->NotifyClientsAboutServerDependentCap((CString const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_IsServerCapAccepted(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_IsServerCapAccepted", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_IsServerCapAccepted" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_IsServerCapAccepted" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_IsServerCapAccepted" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)((CIRCNetwork const *)arg1)->IsServerCapAccepted((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutIRC__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutIRC" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutIRC" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutIRC" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PutIRC((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutIRC__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_PutIRC" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_PutIRC" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_PutIRC" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (bool)(arg1)->PutIRC((CMessage const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_PutIRC(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_PutIRC", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_PutIRC__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_PutIRC__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_PutIRC'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::PutIRC(CString const &)\n"
    "    CIRCNetwork::PutIRC(CMessage const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddRawBuffer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CMessage *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->AddRawBuffer((CMessage const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddRawBuffer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  (arg1)->AddRawBuffer((CMessage const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateRawBuffer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CMessage *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  arg3 = reinterpret_cast< CMessage * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->UpdateRawBuffer((CString const &)*arg2,(CMessage const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateRawBuffer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CMessage *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  arg3 = reinterpret_cast< CMessage * >(argp3);
  (arg1)->UpdateRawBuffer((CString const &)*arg2,(CMessage const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateExactRawBuffer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CMessage *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->UpdateExactRawBuffer((CMessage const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateExactRawBuffer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  (arg1)->UpdateExactRawBuffer((CMessage const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ClearRawBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_ClearRawBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  (arg1)->ClearRawBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddRawBuffer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->AddRawBuffer((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddRawBuffer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->AddRawBuffer((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddRawBuffer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_AddRawBuffer", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_AddRawBuffer__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_AddRawBuffer__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_AddRawBuffer__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_AddRawBuffer__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_AddRawBuffer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::AddRawBuffer(CMessage const &,CString const &)\n"
    "    CIRCNetwork::AddRawBuffer(CMessage const &)\n"
    "    CIRCNetwork::AddRawBuffer(CString const &,CString const &)\n"
    "    CIRCNetwork::AddRawBuffer(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateRawBuffer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->UpdateRawBuffer((CString const &)*arg2,(CString const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateRawBuffer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateRawBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->UpdateRawBuffer((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateRawBuffer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_UpdateRawBuffer", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_UpdateRawBuffer__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_UpdateRawBuffer__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CIRCNetwork_UpdateRawBuffer__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CIRCNetwork_UpdateRawBuffer__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_UpdateRawBuffer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::UpdateRawBuffer(CString const &,CMessage const &,CString const &)\n"
    "    CIRCNetwork::UpdateRawBuffer(CString const &,CMessage const &)\n"
    "    CIRCNetwork::UpdateRawBuffer(CString const &,CString const &,CString const &)\n"
    "    CIRCNetwork::UpdateRawBuffer(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateExactRawBuffer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->UpdateExactRawBuffer((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateExactRawBuffer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateExactRawBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->UpdateExactRawBuffer((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateExactRawBuffer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_UpdateExactRawBuffer", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_UpdateExactRawBuffer__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_UpdateExactRawBuffer__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_UpdateExactRawBuffer__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_UpdateExactRawBuffer__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_UpdateExactRawBuffer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::UpdateExactRawBuffer(CMessage const &,CString const &)\n"
    "    CIRCNetwork::UpdateExactRawBuffer(CMessage const &)\n"
    "    CIRCNetwork::UpdateExactRawBuffer(CString const &,CString const &)\n"
    "    CIRCNetwork::UpdateExactRawBuffer(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddMotdBuffer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CMessage *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->AddMotdBuffer((CMessage const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddMotdBuffer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  (arg1)->AddMotdBuffer((CMessage const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateMotdBuffer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CMessage *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  arg3 = reinterpret_cast< CMessage * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->UpdateMotdBuffer((CString const &)*arg2,(CMessage const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateMotdBuffer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CMessage *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  arg3 = reinterpret_cast< CMessage * >(argp3);
  (arg1)->UpdateMotdBuffer((CString const &)*arg2,(CMessage const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ClearMotdBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_ClearMotdBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  (arg1)->ClearMotdBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddMotdBuffer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->AddMotdBuffer((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddMotdBuffer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddMotdBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->AddMotdBuffer((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddMotdBuffer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_AddMotdBuffer", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_AddMotdBuffer__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_AddMotdBuffer__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_AddMotdBuffer__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_AddMotdBuffer__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_AddMotdBuffer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::AddMotdBuffer(CMessage const &,CString const &)\n"
    "    CIRCNetwork::AddMotdBuffer(CMessage const &)\n"
    "    CIRCNetwork::AddMotdBuffer(CString const &,CString const &)\n"
    "    CIRCNetwork::AddMotdBuffer(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateMotdBuffer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->UpdateMotdBuffer((CString const &)*arg2,(CString const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateMotdBuffer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateMotdBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->UpdateMotdBuffer((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateMotdBuffer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_UpdateMotdBuffer", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_UpdateMotdBuffer__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_UpdateMotdBuffer__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CIRCNetwork_UpdateMotdBuffer__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CIRCNetwork_UpdateMotdBuffer__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_UpdateMotdBuffer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::UpdateMotdBuffer(CString const &,CMessage const &,CString const &)\n"
    "    CIRCNetwork::UpdateMotdBuffer(CString const &,CMessage const &)\n"
    "    CIRCNetwork::UpdateMotdBuffer(CString const &,CString const &,CString const &)\n"
    "    CIRCNetwork::UpdateMotdBuffer(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddNoticeBuffer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CMessage *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->AddNoticeBuffer((CMessage const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddNoticeBuffer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  (arg1)->AddNoticeBuffer((CMessage const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateNoticeBuffer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CMessage *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  arg3 = reinterpret_cast< CMessage * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->UpdateNoticeBuffer((CString const &)*arg2,(CMessage const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateNoticeBuffer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CMessage *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  arg3 = reinterpret_cast< CMessage * >(argp3);
  (arg1)->UpdateNoticeBuffer((CString const &)*arg2,(CMessage const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ClearNoticeBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_ClearNoticeBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  (arg1)->ClearNoticeBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddNoticeBuffer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->AddNoticeBuffer((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddNoticeBuffer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_AddNoticeBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->AddNoticeBuffer((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddNoticeBuffer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_AddNoticeBuffer", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_AddNoticeBuffer__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_AddNoticeBuffer__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_AddNoticeBuffer__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_AddNoticeBuffer__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_AddNoticeBuffer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::AddNoticeBuffer(CMessage const &,CString const &)\n"
    "    CIRCNetwork::AddNoticeBuffer(CMessage const &)\n"
    "    CIRCNetwork::AddNoticeBuffer(CString const &,CString const &)\n"
    "    CIRCNetwork::AddNoticeBuffer(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateNoticeBuffer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->UpdateNoticeBuffer((CString const &)*arg2,(CString const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateNoticeBuffer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_UpdateNoticeBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->UpdateNoticeBuffer((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_UpdateNoticeBuffer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_UpdateNoticeBuffer", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_UpdateNoticeBuffer__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCNetwork_UpdateNoticeBuffer__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CIRCNetwork_UpdateNoticeBuffer__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CIRCNetwork_UpdateNoticeBuffer__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_UpdateNoticeBuffer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::UpdateNoticeBuffer(CString const &,CMessage const &,CString const &)\n"
    "    CIRCNetwork::UpdateNoticeBuffer(CString const &,CMessage const &)\n"
    "    CIRCNetwork::UpdateNoticeBuffer(CString const &,CString const &,CString const &)\n"
    "    CIRCNetwork::UpdateNoticeBuffer(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ClearQueryBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_ClearQueryBuffer" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  (arg1)->ClearQueryBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetNick__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetNick" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCNetwork_GetNick" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (CString *) &((CIRCNetwork const *)arg1)->GetNick(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetNick__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetNick" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CString *) &((CIRCNetwork const *)arg1)->GetNick();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetNick(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_GetNick", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CIRCNetwork_GetNick__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CIRCNetwork_GetNick__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_GetNick'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::GetNick(bool const) const\n"
    "    CIRCNetwork::GetNick() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetAltNick__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetAltNick" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCNetwork_GetAltNick" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (CString *) &((CIRCNetwork const *)arg1)->GetAltNick(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetAltNick__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetAltNick" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CString *) &((CIRCNetwork const *)arg1)->GetAltNick();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetAltNick(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_GetAltNick", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CIRCNetwork_GetAltNick__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CIRCNetwork_GetAltNick__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_GetAltNick'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::GetAltNick(bool const) const\n"
    "    CIRCNetwork::GetAltNick() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetIdent__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetIdent" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCNetwork_GetIdent" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (CString *) &((CIRCNetwork const *)arg1)->GetIdent(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetIdent__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetIdent" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CString *) &((CIRCNetwork const *)arg1)->GetIdent();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetIdent(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_GetIdent", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CIRCNetwork_GetIdent__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CIRCNetwork_GetIdent__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_GetIdent'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::GetIdent(bool const) const\n"
    "    CIRCNetwork::GetIdent() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetRealName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetRealName" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = ((CIRCNetwork const *)arg1)->GetRealName();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetBindHost" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CString *) &((CIRCNetwork const *)arg1)->GetBindHost();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetEncoding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetEncoding" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CString *) &((CIRCNetwork const *)arg1)->GetEncoding();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetQuitMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetQuitMsg" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = ((CIRCNetwork const *)arg1)->GetQuitMsg();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetNick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetNick" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_SetNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetNick((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetAltNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetAltNick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetAltNick" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetAltNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_SetAltNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetAltNick((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetIdent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetIdent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetIdent" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetIdent" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_SetIdent" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetIdent((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetRealName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetRealName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetRealName" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetRealName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_SetRealName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetRealName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetBindHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetBindHost" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_SetBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetBindHost((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetEncoding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetEncoding", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetEncoding" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetEncoding" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_SetEncoding" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetEncoding((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetQuitMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetQuitMsg", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetQuitMsg" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_SetQuitMsg" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_SetQuitMsg" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetQuitMsg((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetFloodRate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetFloodRate" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (double)((CIRCNetwork const *)arg1)->GetFloodRate();
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetFloodBurst(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned short result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetFloodBurst" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (unsigned short)((CIRCNetwork const *)arg1)->GetFloodBurst();
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetFloodRate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetFloodRate", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetFloodRate" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCNetwork_SetFloodRate" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  (arg1)->SetFloodRate(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetFloodBurst(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  unsigned short arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetFloodBurst", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetFloodBurst" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCNetwork_SetFloodBurst" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  (arg1)->SetFloodBurst(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetJoinDelay(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned short result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetJoinDelay" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (unsigned short)((CIRCNetwork const *)arg1)->GetJoinDelay();
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetJoinDelay(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  unsigned short arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetJoinDelay", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetJoinDelay" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCNetwork_SetJoinDelay" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  (arg1)->SetJoinDelay(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetTrustAllCerts__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetTrustAllCerts" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCNetwork_SetTrustAllCerts" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetTrustAllCerts(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetTrustAllCerts__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetTrustAllCerts" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  (arg1)->SetTrustAllCerts();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetTrustAllCerts(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetTrustAllCerts", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CIRCNetwork_SetTrustAllCerts__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CIRCNetwork_SetTrustAllCerts__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_SetTrustAllCerts'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::SetTrustAllCerts(bool const)\n"
    "    CIRCNetwork::SetTrustAllCerts()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetTrustAllCerts(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetTrustAllCerts" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (bool)((CIRCNetwork const *)arg1)->GetTrustAllCerts();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetTrustPKI__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetTrustPKI" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCNetwork_SetTrustPKI" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetTrustPKI(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetTrustPKI__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_SetTrustPKI" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  (arg1)->SetTrustPKI();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_SetTrustPKI(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_SetTrustPKI", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CIRCNetwork_SetTrustPKI__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CIRCNetwork_SetTrustPKI__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_SetTrustPKI'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::SetTrustPKI(bool const)\n"
    "    CIRCNetwork::SetTrustPKI()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetTrustPKI(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetTrustPKI" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (bool)((CIRCNetwork const *)arg1)->GetTrustPKI();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_BytesRead(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long long result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_BytesRead" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (unsigned long long)((CIRCNetwork const *)arg1)->BytesRead();
  resultobj = SWIG_From_unsigned_SS_long_SS_long(static_cast< unsigned long long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_BytesWritten(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long long result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_BytesWritten" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (unsigned long long)((CIRCNetwork const *)arg1)->BytesWritten();
  resultobj = SWIG_From_unsigned_SS_long_SS_long(static_cast< unsigned long long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddBytesRead(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  unsigned long long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_AddBytesRead", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddBytesRead" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCNetwork_AddBytesRead" "', argument " "2"" of type '" "unsigned long long""'");
  } 
  arg2 = static_cast< unsigned long long >(val2);
  (arg1)->AddBytesRead(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_AddBytesWritten(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  unsigned long long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCNetwork_AddBytesWritten", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_AddBytesWritten" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCNetwork_AddBytesWritten" "', argument " "2"" of type '" "unsigned long long""'");
  } 
  arg2 = static_cast< unsigned long long >(val2);
  (arg1)->AddBytesWritten(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ExpandString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_ExpandString" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_ExpandString" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_ExpandString" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CIRCNetwork const *)arg1)->ExpandString((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ExpandString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_ExpandString" "', argument " "1"" of type '" "CIRCNetwork const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCNetwork_ExpandString" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCNetwork_ExpandString" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sRet");
    }
  }
  result = (CString *) &((CIRCNetwork const *)arg1)->ExpandString((CString const &)*arg2,*arg3);
  {
    if (result) {
      resultobj = CPyRetString::wrap(*result);
    } else {
      resultobj = Py_None;
      Py_INCREF(Py_None);
    }
  }
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_ExpandString(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCNetwork_ExpandString", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCNetwork_ExpandString__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          return _wrap_CIRCNetwork_ExpandString__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCNetwork_ExpandString'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCNetwork::ExpandString(CString const &) const\n"
    "    CIRCNetwork::ExpandString(CString const &,CString &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork___str__" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = CIRCNetwork___str__(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork___repr__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork___repr__" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = CIRCNetwork___repr__(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetChans_(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CChan *,std::allocator< CChan * > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetChans_" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = CIRCNetwork_GetChans_(arg1);
  resultobj = swig::from(static_cast< std::vector< CChan*,std::allocator< CChan * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetServers_(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CServer *,std::allocator< CServer * > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetServers_" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = CIRCNetwork_GetServers_(arg1);
  resultobj = swig::from(static_cast< std::vector< CServer*,std::allocator< CServer * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCNetwork_GetQueries_(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CQuery *,std::allocator< CQuery * > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCNetwork_GetQueries_" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = CIRCNetwork_GetQueries_(arg1);
  resultobj = swig::from(static_cast< std::vector< CQuery*,std::allocator< CQuery * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CIRCNetwork_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CIRCNetwork, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CIRCNetwork_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CQuery(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CQuery *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CQuery", 2, 2, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CQuery" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CQuery" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CQuery" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  result = (CQuery *)new CQuery((CString const &)*arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CQuery, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CQuery(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CQuery" "', argument " "1"" of type '" "CQuery *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_GetBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CBuffer *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_GetBuffer" "', argument " "1"" of type '" "CQuery const *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  result = (CBuffer *) &((CQuery const *)arg1)->GetBuffer();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBuffer, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_GetBufferCount(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_GetBufferCount" "', argument " "1"" of type '" "CQuery const *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  result = (unsigned int)((CQuery const *)arg1)->GetBufferCount();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_SetBufferCount__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  unsigned int arg2 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_SetBufferCount" "', argument " "1"" of type '" "CQuery *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CQuery_SetBufferCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CQuery_SetBufferCount" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)(arg1)->SetBufferCount(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_SetBufferCount__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_SetBufferCount" "', argument " "1"" of type '" "CQuery *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CQuery_SetBufferCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = (bool)(arg1)->SetBufferCount(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_SetBufferCount(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CQuery_SetBufferCount", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CQuery, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CQuery_SetBufferCount__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CQuery, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CQuery_SetBufferCount__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CQuery_SetBufferCount'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CQuery::SetBufferCount(unsigned int,bool)\n"
    "    CQuery::SetBufferCount(unsigned int)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CQuery_AddBuffer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  CMessage *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  size_t result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_AddBuffer" "', argument " "1"" of type '" "CQuery *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CQuery_AddBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuery_AddBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CQuery_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuery_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (arg1)->AddBuffer((CMessage const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_AddBuffer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  size_t result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_AddBuffer" "', argument " "1"" of type '" "CQuery *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CQuery_AddBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuery_AddBuffer" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (arg1)->AddBuffer((CMessage const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_AddBuffer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  timeval *arg4 = (timeval *) 0 ;
  MCString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  size_t result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_AddBuffer" "', argument " "1"" of type '" "CQuery *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CQuery_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuery_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CQuery_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuery_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_timeval, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CQuery_AddBuffer" "', argument " "4"" of type '" "timeval const *""'"); 
  }
  arg4 = reinterpret_cast< timeval * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_MCString,  0  | 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CQuery_AddBuffer" "', argument " "5"" of type '" "MCString const &""'"); 
  }
  if (!argp5) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuery_AddBuffer" "', argument " "5"" of type '" "MCString const &""'"); 
  }
  arg5 = reinterpret_cast< MCString * >(argp5);
  result = (arg1)->AddBuffer((CString const &)*arg2,(CString const &)*arg3,(timeval const *)arg4,(MCString const &)*arg5);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_AddBuffer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  timeval *arg4 = (timeval *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  size_t result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_AddBuffer" "', argument " "1"" of type '" "CQuery *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CQuery_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuery_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CQuery_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuery_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_timeval, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CQuery_AddBuffer" "', argument " "4"" of type '" "timeval const *""'"); 
  }
  arg4 = reinterpret_cast< timeval * >(argp4);
  result = (arg1)->AddBuffer((CString const &)*arg2,(CString const &)*arg3,(timeval const *)arg4);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_AddBuffer__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  size_t result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_AddBuffer" "', argument " "1"" of type '" "CQuery *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CQuery_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuery_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CQuery_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuery_AddBuffer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (arg1)->AddBuffer((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_AddBuffer__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  size_t result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_AddBuffer" "', argument " "1"" of type '" "CQuery *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CQuery_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuery_AddBuffer" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->AddBuffer((CString const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_AddBuffer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CQuery_AddBuffer", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CQuery, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CQuery_AddBuffer__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CQuery, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CQuery_AddBuffer__SWIG_5(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CQuery, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CQuery_AddBuffer__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CQuery, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CQuery_AddBuffer__SWIG_4(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CQuery, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_timeval, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CQuery_AddBuffer__SWIG_3(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CQuery, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_timeval, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            int res = SWIG_ConvertPtr(argv[4], 0, SWIGTYPE_p_MCString, SWIG_POINTER_NO_NULL | 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_CQuery_AddBuffer__SWIG_2(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CQuery_AddBuffer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CQuery::AddBuffer(CMessage const &,CString const &)\n"
    "    CQuery::AddBuffer(CMessage const &)\n"
    "    CQuery::AddBuffer(CString const &,CString const &,timeval const *,MCString const &)\n"
    "    CQuery::AddBuffer(CString const &,CString const &,timeval const *)\n"
    "    CQuery::AddBuffer(CString const &,CString const &)\n"
    "    CQuery::AddBuffer(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CQuery_ClearBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_ClearBuffer" "', argument " "1"" of type '" "CQuery *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  (arg1)->ClearBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_SendBuffer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_SendBuffer" "', argument " "1"" of type '" "CQuery *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CQuery_SendBuffer" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  (arg1)->SendBuffer(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_SendBuffer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  CBuffer *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_SendBuffer" "', argument " "1"" of type '" "CQuery *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CQuery_SendBuffer" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CBuffer,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CQuery_SendBuffer" "', argument " "3"" of type '" "CBuffer const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CQuery_SendBuffer" "', argument " "3"" of type '" "CBuffer const &""'"); 
  }
  arg3 = reinterpret_cast< CBuffer * >(argp3);
  (arg1)->SendBuffer(arg2,(CBuffer const &)*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CQuery_SendBuffer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CQuery_SendBuffer", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CQuery, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CClient, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CQuery_SendBuffer__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CQuery, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CClient, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CBuffer, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CQuery_SendBuffer__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CQuery_SendBuffer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CQuery::SendBuffer(CClient *)\n"
    "    CQuery::SendBuffer(CClient *,CBuffer const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CQuery_GetName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CQuery *arg1 = (CQuery *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CQuery_GetName" "', argument " "1"" of type '" "CQuery const *""'"); 
  }
  arg1 = reinterpret_cast< CQuery * >(argp1);
  result = (CString *) &((CQuery const *)arg1)->GetName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CQuery_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CQuery, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CQuery_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_delete_CAuthBase(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CAuthBase *arg1 = (CAuthBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CAuthBase > tempshared1 ;
  std::shared_ptr< CAuthBase > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CAuthBase_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CAuthBase" "', argument " "1"" of type '" "CAuthBase *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  (void)arg1; delete smartarg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CAuthBase_SetLoginInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CAuthBase *arg1 = (CAuthBase *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CZNCSock *arg4 = (CZNCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CAuthBase > tempshared1 ;
  std::shared_ptr< CAuthBase > *smartarg1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CAuthBase_SetLoginInfo", 4, 4, swig_obj)) SWIG_fail;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CAuthBase_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CAuthBase_SetLoginInfo" "', argument " "1"" of type '" "CAuthBase *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CAuthBase_SetLoginInfo" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CAuthBase_SetLoginInfo" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CAuthBase_SetLoginInfo" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CAuthBase_SetLoginInfo" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CAuthBase_SetLoginInfo" "', argument " "4"" of type '" "CZNCSock *""'"); 
  }
  arg4 = reinterpret_cast< CZNCSock * >(argp4);
  (arg1)->SetLoginInfo((CString const &)*arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CAuthBase_AcceptLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CAuthBase *arg1 = (CAuthBase *) 0 ;
  CUser *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CAuthBase > tempshared1 ;
  std::shared_ptr< CAuthBase > *smartarg1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CAuthBase_AcceptLogin", 2, 2, swig_obj)) SWIG_fail;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CAuthBase_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CAuthBase_AcceptLogin" "', argument " "1"" of type '" "CAuthBase *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CAuthBase_AcceptLogin" "', argument " "2"" of type '" "CUser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CAuthBase_AcceptLogin" "', argument " "2"" of type '" "CUser &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  (arg1)->AcceptLogin(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CAuthBase_RefuseLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CAuthBase *arg1 = (CAuthBase *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CAuthBase > tempshared1 ;
  std::shared_ptr< CAuthBase > *smartarg1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CAuthBase_RefuseLogin", 2, 2, swig_obj)) SWIG_fail;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CAuthBase_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CAuthBase_RefuseLogin" "', argument " "1"" of type '" "CAuthBase *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CAuthBase_RefuseLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CAuthBase_RefuseLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->RefuseLogin((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CAuthBase_GetUsername(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CAuthBase *arg1 = (CAuthBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CAuthBase const > tempshared1 ;
  std::shared_ptr< CAuthBase const > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CAuthBase_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CAuthBase_GetUsername" "', argument " "1"" of type '" "CAuthBase const *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr< const CAuthBase > * >(argp1);
      delete reinterpret_cast< std::shared_ptr< const CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr< const CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  result = (CString *) &((CAuthBase const *)arg1)->GetUsername();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CAuthBase_GetPassword(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CAuthBase *arg1 = (CAuthBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CAuthBase const > tempshared1 ;
  std::shared_ptr< CAuthBase const > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CAuthBase_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CAuthBase_GetPassword" "', argument " "1"" of type '" "CAuthBase const *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr< const CAuthBase > * >(argp1);
      delete reinterpret_cast< std::shared_ptr< const CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr< const CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  result = (CString *) &((CAuthBase const *)arg1)->GetPassword();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CAuthBase_GetSocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CAuthBase *arg1 = (CAuthBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CAuthBase const > tempshared1 ;
  std::shared_ptr< CAuthBase const > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  Csock *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CAuthBase_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CAuthBase_GetSocket" "', argument " "1"" of type '" "CAuthBase const *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr< const CAuthBase > * >(argp1);
      delete reinterpret_cast< std::shared_ptr< const CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr< const CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  result = (Csock *)((CAuthBase const *)arg1)->GetSocket();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CAuthBase_GetRemoteIP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CAuthBase *arg1 = (CAuthBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CAuthBase const > tempshared1 ;
  std::shared_ptr< CAuthBase const > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CAuthBase_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CAuthBase_GetRemoteIP" "', argument " "1"" of type '" "CAuthBase const *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr< const CAuthBase > * >(argp1);
      delete reinterpret_cast< std::shared_ptr< const CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr< const CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  result = ((CAuthBase const *)arg1)->GetRemoteIP();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CAuthBase_Invalidate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CAuthBase *arg1 = (CAuthBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CAuthBase > tempshared1 ;
  std::shared_ptr< CAuthBase > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CAuthBase_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CAuthBase_Invalidate" "', argument " "1"" of type '" "CAuthBase *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CAuthBase > * >(argp1);
      arg1 = const_cast< CAuthBase * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  (arg1)->Invalidate();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CAuthBase_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__shared_ptrT_CAuthBase_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_CClientAuth(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  CClientAuth *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CClientAuth", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CClientAuth" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CClientAuth" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CClientAuth" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CClientAuth" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CClientAuth" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (CClientAuth *)new CClientAuth(arg1,(CString const &)*arg2,(CString const &)*arg3);
  {
    std::shared_ptr<  CClientAuth > *smartresult = result ? new std::shared_ptr<  CClientAuth >(result SWIG_NO_NULL_DELETER_SWIG_POINTER_NEW) : 0;
    resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(smartresult), SWIGTYPE_p_std__shared_ptrT_CClientAuth_t, SWIG_POINTER_NEW | SWIG_POINTER_OWN);
  }
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CClientAuth(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClientAuth *arg1 = (CClientAuth *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CClientAuth > tempshared1 ;
  std::shared_ptr< CClientAuth > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CClientAuth_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CClientAuth" "', argument " "1"" of type '" "CClientAuth *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CClientAuth > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CClientAuth > * >(argp1);
      arg1 = const_cast< CClientAuth * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CClientAuth > * >(argp1);
      arg1 = const_cast< CClientAuth * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  (void)arg1; delete smartarg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClientAuth_Invalidate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClientAuth *arg1 = (CClientAuth *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CClientAuth > tempshared1 ;
  std::shared_ptr< CClientAuth > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CClientAuth_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClientAuth_Invalidate" "', argument " "1"" of type '" "CClientAuth *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CClientAuth > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CClientAuth > * >(argp1);
      arg1 = const_cast< CClientAuth * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CClientAuth > * >(argp1);
      arg1 = const_cast< CClientAuth * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  (arg1)->Invalidate();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClientAuth_AcceptedLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClientAuth *arg1 = (CClientAuth *) 0 ;
  CUser *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CClientAuth > tempshared1 ;
  std::shared_ptr< CClientAuth > *smartarg1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClientAuth_AcceptedLogin", 2, 2, swig_obj)) SWIG_fail;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CClientAuth_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClientAuth_AcceptedLogin" "', argument " "1"" of type '" "CClientAuth *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CClientAuth > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CClientAuth > * >(argp1);
      arg1 = const_cast< CClientAuth * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CClientAuth > * >(argp1);
      arg1 = const_cast< CClientAuth * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClientAuth_AcceptedLogin" "', argument " "2"" of type '" "CUser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClientAuth_AcceptedLogin" "', argument " "2"" of type '" "CUser &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  (arg1)->AcceptedLogin(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClientAuth_RefusedLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClientAuth *arg1 = (CClientAuth *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CClientAuth > tempshared1 ;
  std::shared_ptr< CClientAuth > *smartarg1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClientAuth_RefusedLogin", 2, 2, swig_obj)) SWIG_fail;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CClientAuth_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClientAuth_RefusedLogin" "', argument " "1"" of type '" "CClientAuth *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CClientAuth > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CClientAuth > * >(argp1);
      arg1 = const_cast< CClientAuth * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CClientAuth > * >(argp1);
      arg1 = const_cast< CClientAuth * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClientAuth_RefusedLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClientAuth_RefusedLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->RefusedLogin((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *CClientAuth_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__shared_ptrT_CClientAuth_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CClientAuth_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CClient(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CClient", 0, 0, 0)) SWIG_fail;
  result = (CClient *)new CClient();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CClient, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CClient(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CClient" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_SendRequiredPasswordNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_SendRequiredPasswordNotice" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  (arg1)->SendRequiredPasswordNotice();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_AcceptLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CUser *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_AcceptLogin", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_AcceptLogin" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_AcceptLogin" "', argument " "2"" of type '" "CUser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_AcceptLogin" "', argument " "2"" of type '" "CUser &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  (arg1)->AcceptLogin(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_RefuseLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_RefuseLogin", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_RefuseLogin" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_RefuseLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_RefuseLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->RefuseLogin((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_GetNick__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_GetNick" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CClient_GetNick" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = ((CClient const *)arg1)->GetNick(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_GetNick__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_GetNick" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = ((CClient const *)arg1)->GetNick();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_GetNick(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CClient_GetNick", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CClient, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CClient_GetNick__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CClient, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CClient_GetNick__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CClient_GetNick'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CClient::GetNick(bool) const\n"
    "    CClient::GetNick() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CClient_GetNickMask(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_GetNickMask" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = ((CClient const *)arg1)->GetNickMask();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_GetIdentifier(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_GetIdentifier" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = ((CClient const *)arg1)->GetIdentifier();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_CapVersion(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned short result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_CapVersion" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (unsigned short)((CClient const *)arg1)->CapVersion();
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HasCap302(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HasCap302" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->HasCap302();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HasCapNotify(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HasCapNotify" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->HasCapNotify();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HasAwayNotify(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HasAwayNotify" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->HasAwayNotify();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HasAccountNotify(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HasAccountNotify" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->HasAccountNotify();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HasInviteNotify(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HasInviteNotify" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->HasInviteNotify();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HasExtendedJoin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HasExtendedJoin" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->HasExtendedJoin();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HasNamesx(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HasNamesx" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->HasNamesx();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HasUHNames(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HasUHNames" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->HasUHNames();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HasChgHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HasChgHost" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->HasChgHost();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_IsAway(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_IsAway" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->IsAway();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HasServerTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HasServerTime" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->HasServerTime();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HasBatch(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HasBatch" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->HasBatch();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HasEchoMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HasEchoMessage" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->HasEchoMessage();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HasSelfMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HasSelfMessage" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->HasSelfMessage();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_IsValidIdentifier(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_IsValidIdentifier" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_IsValidIdentifier" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CClient::IsValidIdentifier((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_UserCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_UserCommand", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_UserCommand" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  (arg1)->UserCommand(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_UserPortCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_UserPortCommand", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_UserPortCommand" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  (arg1)->UserPortCommand(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_StatusCTCP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_StatusCTCP", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_StatusCTCP" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_StatusCTCP" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_StatusCTCP" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->StatusCTCP((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_BouncedOff(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_BouncedOff" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  (arg1)->BouncedOff();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_IsAttached(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_IsAttached" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->IsAttached();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_IsPlaybackActive(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_IsPlaybackActive" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)((CClient const *)arg1)->IsPlaybackActive();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_SetPlaybackActive(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_SetPlaybackActive", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_SetPlaybackActive" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CClient_SetPlaybackActive" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetPlaybackActive(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_PutIRC(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_PutIRC", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_PutIRC" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_PutIRC" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_PutIRC" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->PutIRC((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_PutIRCStripping(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CMessage arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_PutIRCStripping", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_PutIRCStripping" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_PutIRCStripping" "', argument " "2"" of type '" "CMessage""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_PutIRCStripping" "', argument " "2"" of type '" "CMessage""'");
    } else {
      CMessage * temp = reinterpret_cast< CMessage * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  (arg1)->PutIRCStripping(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_PutClientRaw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_PutClientRaw", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_PutClientRaw" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_PutClientRaw" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_PutClientRaw" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PutClientRaw((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_PutClient__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_PutClient" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_PutClient" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_PutClient" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->PutClient((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_PutClient__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_PutClient" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_PutClient" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_PutClient" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (bool)(arg1)->PutClient((CMessage const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_PutClient(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CClient_PutClient", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CClient, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CClient_PutClient__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CClient, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CClient_PutClient__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CClient_PutClient'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CClient::PutClient(CString const &)\n"
    "    CClient::PutClient(CMessage const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CClient_PutStatus__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CTable *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned int result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_PutStatus" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTable,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_PutStatus" "', argument " "2"" of type '" "CTable const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_PutStatus" "', argument " "2"" of type '" "CTable const &""'"); 
  }
  arg2 = reinterpret_cast< CTable * >(argp2);
  result = (unsigned int)(arg1)->PutStatus((CTable const &)*arg2);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_PutStatus__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_PutStatus" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_PutStatus" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->PutStatus((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_PutStatus(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CClient_PutStatus", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CClient, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CTable, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CClient_PutStatus__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CClient, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CClient_PutStatus__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CClient_PutStatus'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CClient::PutStatus(CTable const &)\n"
    "    CClient::PutStatus(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CClient_PutStatusNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_PutStatusNotice", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_PutStatusNotice" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_PutStatusNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_PutStatusNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->PutStatusNotice((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_PutModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_PutModule", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_PutModule" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_PutModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CClient_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_PutModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->PutModule((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_PutModNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_PutModNotice", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_PutModNotice" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_PutModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_PutModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CClient_PutModNotice" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_PutModNotice" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->PutModNotice((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_IsCapEnabled(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_IsCapEnabled", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_IsCapEnabled" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_IsCapEnabled" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_IsCapEnabled" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)((CClient const *)arg1)->IsCapEnabled((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_IsTagEnabled(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_IsTagEnabled", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_IsTagEnabled" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_IsTagEnabled" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_IsTagEnabled" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)((CClient const *)arg1)->IsTagEnabled((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_SetTagSupport(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_SetTagSupport", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_SetTagSupport" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_SetTagSupport" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_SetTagSupport" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CClient_SetTagSupport" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->SetTagSupport((CString const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_NotifyServerDependentCap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_NotifyServerDependentCap", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_NotifyServerDependentCap" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_NotifyServerDependentCap" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_NotifyServerDependentCap" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CClient_NotifyServerDependentCap" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CClient_NotifyServerDependentCap" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_NotifyServerDependentCap" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->NotifyServerDependentCap((CString const &)*arg2,arg3,(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_ReadLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_ReadLine", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_ReadLine" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_ReadLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_ReadLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->ReadLine((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_SendMotd(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_SendMotd" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (bool)(arg1)->SendMotd();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HelpUser__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HelpUser" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_HelpUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_HelpUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->HelpUser((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HelpUser__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_HelpUser" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  (arg1)->HelpUser();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_HelpUser(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CClient_HelpUser", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CClient, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CClient_HelpUser__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CClient, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CClient_HelpUser__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CClient_HelpUser'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CClient::HelpUser(CString const &)\n"
    "    CClient::HelpUser()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CClient_AuthUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_AuthUser" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  (arg1)->AuthUser();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_Connected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_Connected" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  (arg1)->Connected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_Timeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_Timeout" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  (arg1)->Timeout();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_Disconnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_Disconnected" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  (arg1)->Disconnected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_ConnectionRefused(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_ConnectionRefused" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  (arg1)->ConnectionRefused();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_ReachedMaxBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_ReachedMaxBuffer" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  (arg1)->ReachedMaxBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_SetNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_SetNick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_SetNick" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_SetNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_SetNick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetNick((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_SetAway(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_SetAway", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_SetAway" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CClient_SetAway" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetAway(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_GetUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CUser *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_GetUser" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (CUser *)((CClient const *)arg1)->GetUser();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CUser, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_SetNetwork__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  bool arg3 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_SetNetwork" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_SetNetwork" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CClient_SetNetwork" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CClient_SetNetwork" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  (arg1)->SetNetwork(arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_SetNetwork__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_SetNetwork" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_SetNetwork" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CClient_SetNetwork" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->SetNetwork(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_SetNetwork__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_SetNetwork" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_SetNetwork" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  (arg1)->SetNetwork(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_SetNetwork(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CClient_SetNetwork", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CClient, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CClient_SetNetwork__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CClient, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CClient_SetNetwork__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CClient, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CIRCNetwork, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CClient_SetNetwork__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CClient_SetNetwork'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CClient::SetNetwork(CIRCNetwork *,bool,bool)\n"
    "    CClient::SetNetwork(CIRCNetwork *,bool)\n"
    "    CClient::SetNetwork(CIRCNetwork *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CClient_GetNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CIRCNetwork *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_GetNetwork" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (CIRCNetwork *)((CClient const *)arg1)->GetNetwork();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_GetClients(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CClient *,std::allocator< CClient * > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_GetClients" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (std::vector< CClient *,std::allocator< CClient * > > *) &((CClient const *)arg1)->GetClients();
  resultobj = swig::from(static_cast< std::vector< CClient*,std::allocator< CClient * > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_GetIRCSock__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CIRCSock *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_GetIRCSock" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (CIRCSock *)((CClient const *)arg1)->GetIRCSock();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCSock, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_GetIRCSock__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CIRCSock *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_GetIRCSock" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = (CIRCSock *)(arg1)->GetIRCSock();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCSock, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_GetIRCSock(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CClient_GetIRCSock", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CClient, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CClient_GetIRCSock__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CClient, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CClient_GetIRCSock__SWIG_0(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CClient_GetIRCSock'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CClient::GetIRCSock() const\n"
    "    CClient::GetIRCSock()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CClient_GetFullName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_GetFullName" "', argument " "1"" of type '" "CClient const *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  result = ((CClient const *)arg1)->GetFullName();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_SendSASLChallenge(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_SendSASLChallenge", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_SendSASLChallenge" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    int res = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "CClient_SendSASLChallenge" "', argument " "2"" of type '" "CString""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  (arg1)->SendSASLChallenge(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_RefuseSASLLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_RefuseSASLLogin", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_RefuseSASLLogin" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_RefuseSASLLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_RefuseSASLLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->RefuseSASLLogin((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_AcceptSASLLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CUser *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_AcceptSASLLogin", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_AcceptSASLLogin" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_AcceptSASLLogin" "', argument " "2"" of type '" "CUser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_AcceptSASLLogin" "', argument " "2"" of type '" "CUser &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  (arg1)->AcceptSASLLogin(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_StartSASLPasswordCheck(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_StartSASLPasswordCheck", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_StartSASLPasswordCheck" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_StartSASLPasswordCheck" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_StartSASLPasswordCheck" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CClient_StartSASLPasswordCheck" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_StartSASLPasswordCheck" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CClient_StartSASLPasswordCheck" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_StartSASLPasswordCheck" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->StartSASLPasswordCheck((CString const &)*arg2,(CString const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CClient_ParseUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CClient *arg1 = (CClient *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CClient_ParseUser", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CClient_ParseUser" "', argument " "1"" of type '" "CClient *""'"); 
  }
  arg1 = reinterpret_cast< CClient * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CClient_ParseUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CClient_ParseUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->ParseUser((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *CClient_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CClient, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CClient_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CIRCSock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCNetwork *arg1 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CIRCSock *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CIRCSock" "', argument " "1"" of type '" "CIRCNetwork *""'"); 
  }
  arg1 = reinterpret_cast< CIRCNetwork * >(argp1);
  result = (CIRCSock *)new CIRCSock(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCSock, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CIRCSock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CIRCSock" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_ReadLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCSock_ReadLine", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_ReadLine" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_ReadLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_ReadLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->ReadLine((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_Connected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_Connected" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  (arg1)->Connected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_Disconnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_Disconnected" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  (arg1)->Disconnected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_ConnectionRefused(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_ConnectionRefused" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  (arg1)->ConnectionRefused();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_SockError(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  int arg2 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCSock_SockError", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_SockError" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCSock_SockError" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCSock_SockError" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_SockError" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->SockError(arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_Timeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_Timeout" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  (arg1)->Timeout();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_ReachedMaxBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_ReachedMaxBuffer" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  (arg1)->ReachedMaxBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_PutIRCRaw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCSock_PutIRCRaw", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_PutIRCRaw" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_PutIRCRaw" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_PutIRCRaw" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->PutIRCRaw((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_PutIRC__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_PutIRC" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_PutIRC" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_PutIRC" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->PutIRC((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_PutIRC__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_PutIRC" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_PutIRC" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_PutIRC" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  (arg1)->PutIRC((CMessage const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_PutIRC(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCSock_PutIRC", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCSock_PutIRC__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCSock_PutIRC__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCSock_PutIRC'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCSock::PutIRC(CString const &)\n"
    "    CIRCSock::PutIRC(CMessage const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCSock_PutIRCQuick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCSock_PutIRCQuick", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_PutIRCQuick" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_PutIRCQuick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_PutIRCQuick" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->PutIRCQuick((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_ResetChans(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_ResetChans" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  (arg1)->ResetChans();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_Quit__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_Quit" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_Quit" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_Quit" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->Quit((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_Quit__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_Quit" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  (arg1)->Quit();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_Quit(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCSock_Quit", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CIRCSock_Quit__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCSock_Quit__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCSock_Quit'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCSock::Quit(CString const &)\n"
    "    CIRCSock::Quit()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCSock_PauseCap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_PauseCap" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  (arg1)->PauseCap();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_ResumeCap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_ResumeCap" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  (arg1)->ResumeCap();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_IsTagEnabled(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCSock_IsTagEnabled", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_IsTagEnabled" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_IsTagEnabled" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_IsTagEnabled" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)((CIRCSock const *)arg1)->IsTagEnabled((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_SetTagSupport(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCSock_SetTagSupport", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_SetTagSupport" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_SetTagSupport" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_SetTagSupport" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CIRCSock_SetTagSupport" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->SetTagSupport((CString const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_SetPass(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCSock_SetPass", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_SetPass" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_SetPass" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_SetPass" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetPass((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetMaxNickLen(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetMaxNickLen" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (unsigned int)((CIRCSock const *)arg1)->GetMaxNickLen();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetModeType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  CIRCSock::EChanModeArgs result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCSock_GetModeType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetModeType" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCSock_GetModeType" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = (CIRCSock::EChanModeArgs)((CIRCSock const *)arg1)->GetModeType(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetPermFromMode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCSock_GetPermFromMode", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetPermFromMode" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCSock_GetPermFromMode" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = (char)((CIRCSock const *)arg1)->GetPermFromMode(arg2);
  resultobj = SWIG_From_char(static_cast< char >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetModeFromPerm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCSock_GetModeFromPerm", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetModeFromPerm" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCSock_GetModeFromPerm" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = (char)((CIRCSock const *)arg1)->GetModeFromPerm(arg2);
  resultobj = SWIG_From_char(static_cast< char >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetChanModes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< char,CIRCSock::EChanModeArgs,std::less< char >,std::allocator< std::pair< char const,CIRCSock::EChanModeArgs > > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetChanModes" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (std::map< char,CIRCSock::EChanModeArgs,std::less< char >,std::allocator< std::pair< char const,CIRCSock::EChanModeArgs > > > *) &((CIRCSock const *)arg1)->GetChanModes();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_char_CIRCSock__EChanModeArgs_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CIRCSock__EChanModeArgs_t_t_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_IsPermChar(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCSock_IsPermChar", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_IsPermChar" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCSock_IsPermChar" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = (bool)((CIRCSock const *)arg1)->IsPermChar(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_IsPermMode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCSock_IsPermMode", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_IsPermMode" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CIRCSock_IsPermMode" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  result = (bool)((CIRCSock const *)arg1)->IsPermMode(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetPerms(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetPerms" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (CString *) &((CIRCSock const *)arg1)->GetPerms();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetPermModes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetPermModes" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (CString *) &((CIRCSock const *)arg1)->GetPermModes();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetNickMask(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetNickMask" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = ((CIRCSock const *)arg1)->GetNickMask();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetNick" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (CString *) &((CIRCSock const *)arg1)->GetNick();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetPass(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetPass" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (CString *) &((CIRCSock const *)arg1)->GetPass();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CIRCNetwork *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetNetwork" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (CIRCNetwork *)((CIRCSock const *)arg1)->GetNetwork();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_HasNamesx(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_HasNamesx" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (bool)((CIRCSock const *)arg1)->HasNamesx();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_HasUHNames(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_HasUHNames" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (bool)((CIRCSock const *)arg1)->HasUHNames();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_HasAwayNotify(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_HasAwayNotify" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (bool)((CIRCSock const *)arg1)->HasAwayNotify();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_HasAccountNotify(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_HasAccountNotify" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (bool)((CIRCSock const *)arg1)->HasAccountNotify();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_HasExtendedJoin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_HasExtendedJoin" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (bool)((CIRCSock const *)arg1)->HasExtendedJoin();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_HasServerTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_HasServerTime" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (bool)((CIRCSock const *)arg1)->HasServerTime();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_HasMessageTagCap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_HasMessageTagCap" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (bool)((CIRCSock const *)arg1)->HasMessageTagCap();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetUserModes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::set< char,std::less< char >,std::allocator< char > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetUserModes" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (std::set< char,std::less< char >,std::allocator< char > > *) &((CIRCSock const *)arg1)->GetUserModes();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__setT_char_std__lessT_char_t_std__allocatorT_char_t_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_IsAuthed(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_IsAuthed" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (bool)((CIRCSock const *)arg1)->IsAuthed();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetAcceptedCaps(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SCString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetAcceptedCaps" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (SCString *) &((CIRCSock const *)arg1)->GetAcceptedCaps();
  resultobj = swig::from(static_cast< std::set< CString,std::less< CString >,std::allocator< CString > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_IsCapAccepted(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIRCSock_IsCapAccepted", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_IsCapAccepted" "', argument " "1"" of type '" "CIRCSock *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_IsCapAccepted" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_IsCapAccepted" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->IsCapAccepted((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetCapLsValue__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetCapLsValue" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_GetCapLsValue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_GetCapLsValue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCSock_GetCapLsValue" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_GetCapLsValue" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = ((CIRCSock const *)arg1)->GetCapLsValue((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetCapLsValue__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetCapLsValue" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_GetCapLsValue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_GetCapLsValue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CIRCSock const *)arg1)->GetCapLsValue((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetCapLsValue(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCSock_GetCapLsValue", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCSock_GetCapLsValue__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCSock_GetCapLsValue__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCSock_GetCapLsValue'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCSock::GetCapLsValue(CString const &,CString const &) const\n"
    "    CIRCSock::GetCapLsValue(CString const &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetISupport__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  MCString *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetISupport" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  result = (MCString *) &((CIRCSock const *)arg1)->GetISupport();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_MCString, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetISupport__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetISupport" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_GetISupport" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_GetISupport" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CIRCSock_GetISupport" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_GetISupport" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = ((CIRCSock const *)arg1)->GetISupport((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetISupport__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CIRCSock *arg1 = (CIRCSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIRCSock_GetISupport" "', argument " "1"" of type '" "CIRCSock const *""'"); 
  }
  arg1 = reinterpret_cast< CIRCSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIRCSock_GetISupport" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIRCSock_GetISupport" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CIRCSock const *)arg1)->GetISupport((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIRCSock_GetISupport(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CIRCSock_GetISupport", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CIRCSock_GetISupport__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CIRCSock_GetISupport__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CIRCSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CIRCSock_GetISupport__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CIRCSock_GetISupport'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CIRCSock::GetISupport() const\n"
    "    CIRCSock::GetISupport(CString const &,CString const &) const\n"
    "    CIRCSock::GetISupport(CString const &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CIRCSock_IsFloodProtected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  double arg1 ;
  double val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_double(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "CIRCSock_IsFloodProtected" "', argument " "1"" of type '" "double""'");
  } 
  arg1 = static_cast< double >(val1);
  result = (bool)CIRCSock::IsFloodProtected(arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CIRCSock_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CIRCSock, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CIRCSock_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_delete_CListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CListener *arg1 = (CListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CListener, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CListener" "', argument " "1"" of type '" "CListener *""'"); 
  }
  arg1 = reinterpret_cast< CListener * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CListener_IsSSL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CListener *arg1 = (CListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CListener_IsSSL" "', argument " "1"" of type '" "CListener const *""'"); 
  }
  arg1 = reinterpret_cast< CListener * >(argp1);
  result = (bool)((CListener const *)arg1)->IsSSL();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CListener_GetRealListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CListener *arg1 = (CListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CRealListener *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CListener_GetRealListener" "', argument " "1"" of type '" "CListener const *""'"); 
  }
  arg1 = reinterpret_cast< CListener * >(argp1);
  result = (CRealListener *)((CListener const *)arg1)->GetRealListener();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CRealListener, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CListener_GetURIPrefix(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CListener *arg1 = (CListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CListener_GetURIPrefix" "', argument " "1"" of type '" "CListener const *""'"); 
  }
  arg1 = reinterpret_cast< CListener * >(argp1);
  result = (CString *) &((CListener const *)arg1)->GetURIPrefix();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CListener_GetAcceptType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CListener *arg1 = (CListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CListener::EAcceptType result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CListener_GetAcceptType" "', argument " "1"" of type '" "CListener const *""'"); 
  }
  arg1 = reinterpret_cast< CListener * >(argp1);
  result = (CListener::EAcceptType)((CListener const *)arg1)->GetAcceptType();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CListener_SetAcceptType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CListener *arg1 = (CListener *) 0 ;
  CListener::EAcceptType arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CListener_SetAcceptType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CListener_SetAcceptType" "', argument " "1"" of type '" "CListener *""'"); 
  }
  arg1 = reinterpret_cast< CListener * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CListener_SetAcceptType" "', argument " "2"" of type '" "CListener::EAcceptType""'");
  } 
  arg2 = static_cast< CListener::EAcceptType >(val2);
  (arg1)->SetAcceptType(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CListener_Listen(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CListener *arg1 = (CListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CListener_Listen" "', argument " "1"" of type '" "CListener *""'"); 
  }
  arg1 = reinterpret_cast< CListener * >(argp1);
  result = (bool)(arg1)->Listen();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CListener_ResetRealListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CListener *arg1 = (CListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CListener_ResetRealListener" "', argument " "1"" of type '" "CListener *""'"); 
  }
  arg1 = reinterpret_cast< CListener * >(argp1);
  (arg1)->ResetRealListener();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CListener_ToConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CListener *arg1 = (CListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CConfig result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CListener_ToConfig" "', argument " "1"" of type '" "CListener const *""'"); 
  }
  arg1 = reinterpret_cast< CListener * >(argp1);
  result = ((CListener const *)arg1)->ToConfig();
  resultobj = SWIG_NewPointerObj((new CConfig(result)), SWIGTYPE_p_CConfig, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CListener_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CListener, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_CTCPListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  unsigned short arg1 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  EAddrType arg5 ;
  CListener::EAcceptType arg6 ;
  unsigned short val1 ;
  int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  PyObject *swig_obj[6] ;
  CTCPListener *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CTCPListener", 6, 6, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_short(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_CTCPListener" "', argument " "1"" of type '" "unsigned short""'");
  } 
  arg1 = static_cast< unsigned short >(val1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CTCPListener" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CTCPListener" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CTCPListener" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CTCPListener" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_CTCPListener" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_CTCPListener" "', argument " "5"" of type '" "EAddrType""'");
  } 
  arg5 = static_cast< EAddrType >(val5);
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_CTCPListener" "', argument " "6"" of type '" "CListener::EAcceptType""'");
  } 
  arg6 = static_cast< CListener::EAcceptType >(val6);
  result = (CTCPListener *)new CTCPListener(arg1,(CString const &)*arg2,(CString const &)*arg3,arg4,arg5,arg6);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTCPListener, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CTCPListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTCPListener *arg1 = (CTCPListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTCPListener, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CTCPListener" "', argument " "1"" of type '" "CTCPListener *""'"); 
  }
  arg1 = reinterpret_cast< CTCPListener * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTCPListener_GetAddrType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTCPListener *arg1 = (CTCPListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  EAddrType result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTCPListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTCPListener_GetAddrType" "', argument " "1"" of type '" "CTCPListener const *""'"); 
  }
  arg1 = reinterpret_cast< CTCPListener * >(argp1);
  result = (EAddrType)((CTCPListener const *)arg1)->GetAddrType();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTCPListener_GetPort(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTCPListener *arg1 = (CTCPListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned short result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTCPListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTCPListener_GetPort" "', argument " "1"" of type '" "CTCPListener const *""'"); 
  }
  arg1 = reinterpret_cast< CTCPListener * >(argp1);
  result = (unsigned short)((CTCPListener const *)arg1)->GetPort();
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTCPListener_GetBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTCPListener *arg1 = (CTCPListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTCPListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTCPListener_GetBindHost" "', argument " "1"" of type '" "CTCPListener const *""'"); 
  }
  arg1 = reinterpret_cast< CTCPListener * >(argp1);
  result = (CString *) &((CTCPListener const *)arg1)->GetBindHost();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTCPListener_Listen(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTCPListener *arg1 = (CTCPListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTCPListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTCPListener_Listen" "', argument " "1"" of type '" "CTCPListener *""'"); 
  }
  arg1 = reinterpret_cast< CTCPListener * >(argp1);
  result = (bool)(arg1)->Listen();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTCPListener_ToConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTCPListener *arg1 = (CTCPListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CConfig result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTCPListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTCPListener_ToConfig" "', argument " "1"" of type '" "CTCPListener const *""'"); 
  }
  arg1 = reinterpret_cast< CTCPListener * >(argp1);
  result = ((CTCPListener const *)arg1)->ToConfig();
  resultobj = SWIG_NewPointerObj((new CConfig(result)), SWIGTYPE_p_CConfig, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CTCPListener_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CTCPListener, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CTCPListener_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CUnixListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  CListener::EAcceptType arg4 ;
  CString *arg5 = 0 ;
  CString *arg6 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int res5 = SWIG_OLDOBJ ;
  int res6 = SWIG_OLDOBJ ;
  PyObject *swig_obj[6] ;
  CUnixListener *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CUnixListener", 6, 6, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CUnixListener" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CUnixListener" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CUnixListener" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CUnixListener" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CUnixListener" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_CUnixListener" "', argument " "4"" of type '" "CListener::EAcceptType""'");
  } 
  arg4 = static_cast< CListener::EAcceptType >(val4);
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_CUnixListener" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CUnixListener" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res6 = SWIG_AsPtr_CString(swig_obj[5], &ptr);
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "new_CUnixListener" "', argument " "6"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CUnixListener" "', argument " "6"" of type '" "CString const &""'"); 
    }
    arg6 = ptr;
  }
  result = (CUnixListener *)new CUnixListener((CString const &)*arg1,(CString const &)*arg2,arg3,arg4,(CString const &)*arg5,(CString const &)*arg6);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CUnixListener, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res5)) delete arg5;
  if (SWIG_IsNewObj(res6)) delete arg6;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res5)) delete arg5;
  if (SWIG_IsNewObj(res6)) delete arg6;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CUnixListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUnixListener *arg1 = (CUnixListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUnixListener, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CUnixListener" "', argument " "1"" of type '" "CUnixListener *""'"); 
  }
  arg1 = reinterpret_cast< CUnixListener * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUnixListener_GetPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUnixListener *arg1 = (CUnixListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUnixListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUnixListener_GetPath" "', argument " "1"" of type '" "CUnixListener const *""'"); 
  }
  arg1 = reinterpret_cast< CUnixListener * >(argp1);
  result = (CString *) &((CUnixListener const *)arg1)->GetPath();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUnixListener_GetGroup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUnixListener *arg1 = (CUnixListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUnixListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUnixListener_GetGroup" "', argument " "1"" of type '" "CUnixListener const *""'"); 
  }
  arg1 = reinterpret_cast< CUnixListener * >(argp1);
  result = (CString *) &((CUnixListener const *)arg1)->GetGroup();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUnixListener_GetMode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUnixListener *arg1 = (CUnixListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUnixListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUnixListener_GetMode" "', argument " "1"" of type '" "CUnixListener const *""'"); 
  }
  arg1 = reinterpret_cast< CUnixListener * >(argp1);
  result = ((CUnixListener const *)arg1)->GetMode();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUnixListener_Listen(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUnixListener *arg1 = (CUnixListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUnixListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUnixListener_Listen" "', argument " "1"" of type '" "CUnixListener *""'"); 
  }
  arg1 = reinterpret_cast< CUnixListener * >(argp1);
  result = (bool)(arg1)->Listen();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CUnixListener_ToConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUnixListener *arg1 = (CUnixListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CConfig result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUnixListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CUnixListener_ToConfig" "', argument " "1"" of type '" "CUnixListener const *""'"); 
  }
  arg1 = reinterpret_cast< CUnixListener * >(argp1);
  result = ((CUnixListener const *)arg1)->ToConfig();
  resultobj = SWIG_NewPointerObj((new CConfig(result)), SWIGTYPE_p_CConfig, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CUnixListener_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CUnixListener, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CUnixListener_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CRealListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CListener *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CRealListener *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CListener,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CRealListener" "', argument " "1"" of type '" "CListener &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CRealListener" "', argument " "1"" of type '" "CListener &""'"); 
  }
  arg1 = reinterpret_cast< CListener * >(argp1);
  result = (CRealListener *)new CRealListener(*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CRealListener, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CRealListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CRealListener *arg1 = (CRealListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CRealListener, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CRealListener" "', argument " "1"" of type '" "CRealListener *""'"); 
  }
  arg1 = reinterpret_cast< CRealListener * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CRealListener_ConnectionFrom(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CRealListener *arg1 = (CRealListener *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CRealListener_ConnectionFrom", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CRealListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CRealListener_ConnectionFrom" "', argument " "1"" of type '" "CRealListener *""'"); 
  }
  arg1 = reinterpret_cast< CRealListener * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CRealListener_ConnectionFrom" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CRealListener_ConnectionFrom" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CRealListener_ConnectionFrom" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  result = (bool)(arg1)->ConnectionFrom((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CRealListener_GetSockObj(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CRealListener *arg1 = (CRealListener *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Csock *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CRealListener_GetSockObj", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CRealListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CRealListener_GetSockObj" "', argument " "1"" of type '" "CRealListener *""'"); 
  }
  arg1 = reinterpret_cast< CRealListener * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CRealListener_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CRealListener_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CRealListener_GetSockObj" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  result = (Csock *)(arg1)->GetSockObj((CString const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CRealListener_SockError(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CRealListener *arg1 = (CRealListener *) 0 ;
  int arg2 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CRealListener_SockError", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CRealListener, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CRealListener_SockError" "', argument " "1"" of type '" "CRealListener *""'"); 
  }
  arg1 = reinterpret_cast< CRealListener * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CRealListener_SockError" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CRealListener_SockError" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CRealListener_SockError" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->SockError(arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *CRealListener_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CRealListener, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CRealListener_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CIncomingConnection(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  unsigned short arg2 ;
  CListener::EAcceptType arg3 ;
  CString *arg4 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  CIncomingConnection *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CIncomingConnection", 4, 4, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CIncomingConnection" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CIncomingConnection" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CIncomingConnection" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CIncomingConnection" "', argument " "3"" of type '" "CListener::EAcceptType""'");
  } 
  arg3 = static_cast< CListener::EAcceptType >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CIncomingConnection" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CIncomingConnection" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (CIncomingConnection *)new CIncomingConnection((CString const &)*arg1,arg2,arg3,(CString const &)*arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CIncomingConnection, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CIncomingConnection(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIncomingConnection *arg1 = (CIncomingConnection *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIncomingConnection, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CIncomingConnection" "', argument " "1"" of type '" "CIncomingConnection *""'"); 
  }
  arg1 = reinterpret_cast< CIncomingConnection * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIncomingConnection_ReadLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIncomingConnection *arg1 = (CIncomingConnection *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CIncomingConnection_ReadLine", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIncomingConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIncomingConnection_ReadLine" "', argument " "1"" of type '" "CIncomingConnection *""'"); 
  }
  arg1 = reinterpret_cast< CIncomingConnection * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CIncomingConnection_ReadLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CIncomingConnection_ReadLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->ReadLine((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CIncomingConnection_ReachedMaxBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CIncomingConnection *arg1 = (CIncomingConnection *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CIncomingConnection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CIncomingConnection_ReachedMaxBuffer" "', argument " "1"" of type '" "CIncomingConnection *""'"); 
  }
  arg1 = reinterpret_cast< CIncomingConnection * >(argp1);
  (arg1)->ReachedMaxBuffer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CIncomingConnection_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CIncomingConnection, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CIncomingConnection_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_delete_CHTTPSock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CHTTPSock" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_ReadData(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  char *arg2 = (char *) 0 ;
  size_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_ReadData", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_ReadData" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_ReadData" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CHTTPSock_ReadData" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = static_cast< size_t >(val3);
  (arg1)->ReadData((char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_ReadLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_ReadLine", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_ReadLine" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_ReadLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_ReadLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->ReadLine((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_Connected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_Connected" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  (arg1)->Connected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetSockObj(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Csock *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_GetSockObj", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetSockObj" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CHTTPSock_GetSockObj" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  result = (Csock *)(arg1)->GetSockObj((CString const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_ForceLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_ForceLogin" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = (bool)(arg1)->ForceLogin();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_OnLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_OnLogin", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_OnLogin" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_OnLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_OnLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CHTTPSock_OnLogin" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_OnLogin" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CHTTPSock_OnLogin" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->OnLogin((CString const &)*arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_OnPageRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_OnPageRequest", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_OnPageRequest" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_OnPageRequest" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_OnPageRequest" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->OnPageRequest((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_PrintFile__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  CString arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_PrintFile" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_PrintFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_PrintFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    int res = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "CHTTPSock_PrintFile" "', argument " "3"" of type '" "CString""'"); 
    }
    arg3 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  result = (bool)(arg1)->PrintFile((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_PrintFile__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_PrintFile" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_PrintFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_PrintFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->PrintFile((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_PrintFile(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CHTTPSock_PrintFile", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CHTTPSock_PrintFile__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CHTTPSock_PrintFile__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CHTTPSock_PrintFile'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CHTTPSock::PrintFile(CString const &,CString)\n"
    "    CHTTPSock::PrintFile(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_CheckPost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_CheckPost" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  (arg1)->CheckPost();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_SentHeader(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_SentHeader" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = (bool)((CHTTPSock const *)arg1)->SentHeader();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_PrintHeader__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  off_t arg2 ;
  CString *arg3 = 0 ;
  unsigned int arg4 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  int res5 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_PrintHeader" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CHTTPSock_PrintHeader" "', argument " "2"" of type '" "off_t""'");
  } 
  arg2 = static_cast< off_t >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CHTTPSock_PrintHeader" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_PrintHeader" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CHTTPSock_PrintHeader" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CHTTPSock_PrintHeader" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_PrintHeader" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  result = (bool)(arg1)->PrintHeader(arg2,(CString const &)*arg3,arg4,(CString const &)*arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_PrintHeader__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  off_t arg2 ;
  CString *arg3 = 0 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_PrintHeader" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CHTTPSock_PrintHeader" "', argument " "2"" of type '" "off_t""'");
  } 
  arg2 = static_cast< off_t >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CHTTPSock_PrintHeader" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_PrintHeader" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CHTTPSock_PrintHeader" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  result = (bool)(arg1)->PrintHeader(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_PrintHeader__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  off_t arg2 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_PrintHeader" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CHTTPSock_PrintHeader" "', argument " "2"" of type '" "off_t""'");
  } 
  arg2 = static_cast< off_t >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CHTTPSock_PrintHeader" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_PrintHeader" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->PrintHeader(arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_PrintHeader__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  off_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_PrintHeader" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  ecode2 = SWIG_AsVal_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CHTTPSock_PrintHeader" "', argument " "2"" of type '" "off_t""'");
  } 
  arg2 = static_cast< off_t >(val2);
  result = (bool)(arg1)->PrintHeader(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_PrintHeader(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CHTTPSock_PrintHeader", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CHTTPSock_PrintHeader__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CHTTPSock_PrintHeader__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CHTTPSock_PrintHeader__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_long(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            int res = SWIG_AsPtr_CString(argv[4], (CString**)(0));
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_CHTTPSock_PrintHeader__SWIG_0(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CHTTPSock_PrintHeader'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CHTTPSock::PrintHeader(off_t,CString const &,unsigned int,CString const &)\n"
    "    CHTTPSock::PrintHeader(off_t,CString const &,unsigned int)\n"
    "    CHTTPSock::PrintHeader(off_t,CString const &)\n"
    "    CHTTPSock::PrintHeader(off_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_AddHeader(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_AddHeader", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_AddHeader" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_AddHeader" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_AddHeader" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CHTTPSock_AddHeader" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_AddHeader" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->AddHeader((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_SetContentType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_SetContentType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_SetContentType" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_SetContentType" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_SetContentType" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetContentType((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_PrintNotFound(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_PrintNotFound" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = (bool)(arg1)->PrintNotFound();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_Redirect(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_Redirect", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_Redirect" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_Redirect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_Redirect" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->Redirect((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_PrintErrorPage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  unsigned int arg2 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_PrintErrorPage", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_PrintErrorPage" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CHTTPSock_PrintErrorPage" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CHTTPSock_PrintErrorPage" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_PrintErrorPage" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CHTTPSock_PrintErrorPage" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_PrintErrorPage" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (bool)(arg1)->PrintErrorPage(arg2,(CString const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_ParseParams(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  std::map< CString,VCString,std::less< CString >,std::allocator< std::pair< CString const,VCString > > > *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_ParseParams", 2, 2, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_ParseParams" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_ParseParams" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__mapT_CString_VCString_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_ParseParams" "', argument " "2"" of type '" "std::map< CString,VCString,std::less< CString >,std::allocator< std::pair< CString const,VCString > > > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_ParseParams" "', argument " "2"" of type '" "std::map< CString,VCString,std::less< CString >,std::allocator< std::pair< CString const,VCString > > > &""'"); 
  }
  arg2 = reinterpret_cast< std::map< CString,VCString,std::less< CString >,std::allocator< std::pair< CString const,VCString > > > * >(argp2);
  CHTTPSock::ParseParams((CString const &)*arg1,*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_ParseURI(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_ParseURI" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  (arg1)->ParseURI();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetPage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetPage" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  (arg1)->GetPage();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetDate__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  time_t arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_time_t,  0  | 0);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetDate" "', argument " "1"" of type '" "time_t""'"); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetDate" "', argument " "1"" of type '" "time_t""'");
    } else {
      time_t * temp = reinterpret_cast< time_t * >(argp1);
      arg1 = *temp;
      if (SWIG_IsNewObj(res1)) delete temp;
    }
  }
  result = CHTTPSock::GetDate(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetDate__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CString result;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = CHTTPSock::GetDate();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetDate(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CHTTPSock_GetDate", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_CHTTPSock_GetDate__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_time_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CHTTPSock_GetDate__SWIG_0(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CHTTPSock_GetDate'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CHTTPSock::GetDate(time_t)\n"
    "    CHTTPSock::GetDate()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetRemoteIP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetRemoteIP" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = ((CHTTPSock const *)arg1)->GetRemoteIP();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetRequestCookie(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_GetRequestCookie", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetRequestCookie" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_GetRequestCookie" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetRequestCookie" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CHTTPSock const *)arg1)->GetRequestCookie((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_SendCookie(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_SendCookie", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_SendCookie" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_SendCookie" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_SendCookie" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CHTTPSock_SendCookie" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_SendCookie" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->SendCookie((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_SetDocRoot(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_SetDocRoot", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_SetDocRoot" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_SetDocRoot" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_SetDocRoot" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetDocRoot((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_SetLoggedIn(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CHTTPSock_SetLoggedIn", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_SetLoggedIn" "', argument " "1"" of type '" "CHTTPSock *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CHTTPSock_SetLoggedIn" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetLoggedIn(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetPath" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = ((CHTTPSock const *)arg1)->GetPath();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_IsLoggedIn(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_IsLoggedIn" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = (bool)((CHTTPSock const *)arg1)->IsLoggedIn();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetDocRoot(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetDocRoot" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = (CString *) &((CHTTPSock const *)arg1)->GetDocRoot();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetUser" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = (CString *) &((CHTTPSock const *)arg1)->GetUser();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetPass(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetPass" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = (CString *) &((CHTTPSock const *)arg1)->GetPass();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParamString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetParamString" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = (CString *) &((CHTTPSock const *)arg1)->GetParamString();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetContentType(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetContentType" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = (CString *) &((CHTTPSock const *)arg1)->GetContentType();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetURI(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetURI" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = (CString *) &((CHTTPSock const *)arg1)->GetURI();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetURIPrefix(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetURIPrefix" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = (CString *) &((CHTTPSock const *)arg1)->GetURIPrefix();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_IsPost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_IsPost" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = (bool)((CHTTPSock const *)arg1)->IsPost();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParam__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetParam" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_GetParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CHTTPSock_GetParam" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CHTTPSock_GetParam" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParam" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = ((CHTTPSock const *)arg1)->GetParam((CString const &)*arg2,arg3,(CString const &)*arg4);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParam__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetParam" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_GetParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CHTTPSock_GetParam" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = ((CHTTPSock const *)arg1)->GetParam((CString const &)*arg2,arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParam__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetParam" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_GetParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CHTTPSock const *)arg1)->GetParam((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParam(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CHTTPSock_GetParam", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CHTTPSock_GetParam__SWIG_2(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CHTTPSock_GetParam__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CHTTPSock_GetParam__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CHTTPSock_GetParam'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CHTTPSock::GetParam(CString const &,bool,CString const &) const\n"
    "    CHTTPSock::GetParam(CString const &,bool) const\n"
    "    CHTTPSock::GetParam(CString const &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetRawParam__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetRawParam" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_GetRawParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetRawParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CHTTPSock_GetRawParam" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = ((CHTTPSock const *)arg1)->GetRawParam((CString const &)*arg2,arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetRawParam__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetRawParam" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_GetRawParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetRawParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CHTTPSock const *)arg1)->GetRawParam((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetRawParam(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CHTTPSock_GetRawParam", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CHTTPSock_GetRawParam__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CHTTPSock_GetRawParam__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CHTTPSock_GetRawParam'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CHTTPSock::GetRawParam(CString const &,bool) const\n"
    "    CHTTPSock::GetRawParam(CString const &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_HasParam__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_HasParam" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_HasParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_HasParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CHTTPSock_HasParam" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)((CHTTPSock const *)arg1)->HasParam((CString const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_HasParam__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_HasParam" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_HasParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_HasParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)((CHTTPSock const *)arg1)->HasParam((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_HasParam(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CHTTPSock_HasParam", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CHTTPSock_HasParam__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CHTTPSock_HasParam__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CHTTPSock_HasParam'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CHTTPSock::HasParam(CString const &,bool) const\n"
    "    CHTTPSock::HasParam(CString const &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParams__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  std::map< CString,VCString,std::less< CString >,std::allocator< std::pair< CString const,VCString > > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetParams" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CHTTPSock_GetParams" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (std::map< CString,VCString,std::less< CString >,std::allocator< std::pair< CString const,VCString > > > *) &((CHTTPSock const *)arg1)->GetParams(arg2);
  resultobj = swig::from(static_cast< std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParams__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::map< CString,VCString,std::less< CString >,std::allocator< std::pair< CString const,VCString > > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetParams" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  result = (std::map< CString,VCString,std::less< CString >,std::allocator< std::pair< CString const,VCString > > > *) &((CHTTPSock const *)arg1)->GetParams();
  resultobj = swig::from(static_cast< std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParams(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CHTTPSock_GetParams", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CHTTPSock_GetParams__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CHTTPSock_GetParams__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CHTTPSock_GetParams'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CHTTPSock::GetParams(bool) const\n"
    "    CHTTPSock::GetParams() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParamValues__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  VCString *arg3 = 0 ;
  bool arg4 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  int res5 = SWIG_OLDOBJ ;
  size_t result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetParamValues" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_GetParamValues" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_CString_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CHTTPSock_GetParamValues" "', argument " "3"" of type '" "VCString &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "3"" of type '" "VCString &""'"); 
  }
  arg3 = reinterpret_cast< VCString * >(argp3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CHTTPSock_GetParamValues" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CHTTPSock_GetParamValues" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  result = ((CHTTPSock const *)arg1)->GetParamValues((CString const &)*arg2,*arg3,arg4,(CString const &)*arg5);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParamValues__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  VCString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  size_t result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetParamValues" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_GetParamValues" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_CString_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CHTTPSock_GetParamValues" "', argument " "3"" of type '" "VCString &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "3"" of type '" "VCString &""'"); 
  }
  arg3 = reinterpret_cast< VCString * >(argp3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CHTTPSock_GetParamValues" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = ((CHTTPSock const *)arg1)->GetParamValues((CString const &)*arg2,*arg3,arg4);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParamValues__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  VCString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  size_t result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetParamValues" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_GetParamValues" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_CString_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CHTTPSock_GetParamValues" "', argument " "3"" of type '" "VCString &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "3"" of type '" "VCString &""'"); 
  }
  arg3 = reinterpret_cast< VCString * >(argp3);
  result = ((CHTTPSock const *)arg1)->GetParamValues((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParamValues__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  std::set< CString,std::less< CString >,std::allocator< CString > > *arg3 = 0 ;
  bool arg4 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  int res5 = SWIG_OLDOBJ ;
  size_t result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetParamValues" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_GetParamValues" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__setT_CString_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CHTTPSock_GetParamValues" "', argument " "3"" of type '" "std::set< CString,std::less< CString >,std::allocator< CString > > &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "3"" of type '" "std::set< CString,std::less< CString >,std::allocator< CString > > &""'"); 
  }
  arg3 = reinterpret_cast< std::set< CString,std::less< CString >,std::allocator< CString > > * >(argp3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CHTTPSock_GetParamValues" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CHTTPSock_GetParamValues" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  result = ((CHTTPSock const *)arg1)->GetParamValues((CString const &)*arg2,*arg3,arg4,(CString const &)*arg5);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParamValues__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  std::set< CString,std::less< CString >,std::allocator< CString > > *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  size_t result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetParamValues" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_GetParamValues" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__setT_CString_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CHTTPSock_GetParamValues" "', argument " "3"" of type '" "std::set< CString,std::less< CString >,std::allocator< CString > > &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "3"" of type '" "std::set< CString,std::less< CString >,std::allocator< CString > > &""'"); 
  }
  arg3 = reinterpret_cast< std::set< CString,std::less< CString >,std::allocator< CString > > * >(argp3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CHTTPSock_GetParamValues" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = ((CHTTPSock const *)arg1)->GetParamValues((CString const &)*arg2,*arg3,arg4);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParamValues__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CHTTPSock *arg1 = (CHTTPSock *) 0 ;
  CString *arg2 = 0 ;
  std::set< CString,std::less< CString >,std::allocator< CString > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  size_t result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CHTTPSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CHTTPSock_GetParamValues" "', argument " "1"" of type '" "CHTTPSock const *""'"); 
  }
  arg1 = reinterpret_cast< CHTTPSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CHTTPSock_GetParamValues" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__setT_CString_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CHTTPSock_GetParamValues" "', argument " "3"" of type '" "std::set< CString,std::less< CString >,std::allocator< CString > > &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CHTTPSock_GetParamValues" "', argument " "3"" of type '" "std::set< CString,std::less< CString >,std::allocator< CString > > &""'"); 
  }
  arg3 = reinterpret_cast< std::set< CString,std::less< CString >,std::allocator< CString > > * >(argp3);
  result = ((CHTTPSock const *)arg1)->GetParamValues((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CHTTPSock_GetParamValues(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CHTTPSock_GetParamValues", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CHTTPSock_GetParamValues__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_std__setT_CString_t, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CHTTPSock_GetParamValues__SWIG_5(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_std__setT_CString_t, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CHTTPSock_GetParamValues__SWIG_4(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CHTTPSock_GetParamValues__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_std__vectorT_CString_t, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            int res = SWIG_AsPtr_CString(argv[4], (CString**)(0));
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_CHTTPSock_GetParamValues__SWIG_0(self, argc, argv);
            }
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CHTTPSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_std__setT_CString_t, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            int res = SWIG_AsPtr_CString(argv[4], (CString**)(0));
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_CHTTPSock_GetParamValues__SWIG_3(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CHTTPSock_GetParamValues'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CHTTPSock::GetParamValues(CString const &,VCString &,bool,CString const &) const\n"
    "    CHTTPSock::GetParamValues(CString const &,VCString &,bool) const\n"
    "    CHTTPSock::GetParamValues(CString const &,VCString &) const\n"
    "    CHTTPSock::GetParamValues(CString const &,std::set< CString,std::less< CString >,std::allocator< CString > > &,bool,CString const &) const\n"
    "    CHTTPSock::GetParamValues(CString const &,std::set< CString,std::less< CString >,std::allocator< CString > > &,bool) const\n"
    "    CHTTPSock::GetParamValues(CString const &,std::set< CString,std::less< CString >,std::allocator< CString > > &) const\n");
  return 0;
}


SWIGINTERN PyObject *CHTTPSock_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CHTTPSock, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_CTemplateTagHandler(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateTagHandler *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CTemplateTagHandler", 0, 0, 0)) SWIG_fail;
  result = (CTemplateTagHandler *)new CTemplateTagHandler();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplateTagHandler, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CTemplateTagHandler(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateTagHandler *arg1 = (CTemplateTagHandler *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateTagHandler, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CTemplateTagHandler" "', argument " "1"" of type '" "CTemplateTagHandler *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateTagHandler * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateTagHandler_HandleVar(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateTagHandler *arg1 = (CTemplateTagHandler *) 0 ;
  CTemplate *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplateTagHandler_HandleVar", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateTagHandler, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateTagHandler_HandleVar" "', argument " "1"" of type '" "CTemplateTagHandler *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateTagHandler * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTemplate,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplateTagHandler_HandleVar" "', argument " "2"" of type '" "CTemplate &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateTagHandler_HandleVar" "', argument " "2"" of type '" "CTemplate &""'"); 
  }
  arg2 = reinterpret_cast< CTemplate * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CTemplateTagHandler_HandleVar" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateTagHandler_HandleVar" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CTemplateTagHandler_HandleVar" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateTagHandler_HandleVar" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[4], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg5 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 5 sOutput");
    }
  }
  result = (bool)(arg1)->HandleVar(*arg2,(CString const &)*arg3,(CString const &)*arg4,*arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateTagHandler_HandleTag(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateTagHandler *arg1 = (CTemplateTagHandler *) 0 ;
  CTemplate *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplateTagHandler_HandleTag", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateTagHandler, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateTagHandler_HandleTag" "', argument " "1"" of type '" "CTemplateTagHandler *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateTagHandler * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTemplate,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplateTagHandler_HandleTag" "', argument " "2"" of type '" "CTemplate &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateTagHandler_HandleTag" "', argument " "2"" of type '" "CTemplate &""'"); 
  }
  arg2 = reinterpret_cast< CTemplate * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CTemplateTagHandler_HandleTag" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateTagHandler_HandleTag" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CTemplateTagHandler_HandleTag" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateTagHandler_HandleTag" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[4], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg5 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 5 sOutput");
    }
  }
  result = (bool)(arg1)->HandleTag(*arg2,(CString const &)*arg3,(CString const &)*arg4,*arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateTagHandler_HandleIf(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateTagHandler *arg1 = (CTemplateTagHandler *) 0 ;
  CTemplate *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplateTagHandler_HandleIf", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateTagHandler, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateTagHandler_HandleIf" "', argument " "1"" of type '" "CTemplateTagHandler *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateTagHandler * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTemplate,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplateTagHandler_HandleIf" "', argument " "2"" of type '" "CTemplate &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateTagHandler_HandleIf" "', argument " "2"" of type '" "CTemplate &""'"); 
  }
  arg2 = reinterpret_cast< CTemplate * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CTemplateTagHandler_HandleIf" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateTagHandler_HandleIf" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CTemplateTagHandler_HandleIf" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateTagHandler_HandleIf" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[4], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg5 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 5 sOutput");
    }
  }
  result = (bool)(arg1)->HandleIf(*arg2,(CString const &)*arg3,(CString const &)*arg4,*arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateTagHandler_HandleValue(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateTagHandler *arg1 = (CTemplateTagHandler *) 0 ;
  CTemplate *arg2 = 0 ;
  CString *arg3 = 0 ;
  MCString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplateTagHandler_HandleValue", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateTagHandler, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateTagHandler_HandleValue" "', argument " "1"" of type '" "CTemplateTagHandler *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateTagHandler * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTemplate,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplateTagHandler_HandleValue" "', argument " "2"" of type '" "CTemplate &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateTagHandler_HandleValue" "', argument " "2"" of type '" "CTemplate &""'"); 
  }
  arg2 = reinterpret_cast< CTemplate * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sValue");
    }
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_MCString,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CTemplateTagHandler_HandleValue" "', argument " "4"" of type '" "MCString const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateTagHandler_HandleValue" "', argument " "4"" of type '" "MCString const &""'"); 
  }
  arg4 = reinterpret_cast< MCString * >(argp4);
  result = (bool)(arg1)->HandleValue(*arg2,*arg3,(MCString const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CTemplateTagHandler_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CTemplateTagHandler, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CTemplateTagHandler_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CTemplateOptions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateOptions *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CTemplateOptions", 0, 0, 0)) SWIG_fail;
  result = (CTemplateOptions *)new CTemplateOptions();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplateOptions, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CTemplateOptions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateOptions *arg1 = (CTemplateOptions *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateOptions, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CTemplateOptions" "', argument " "1"" of type '" "CTemplateOptions *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateOptions * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateOptions_Parse(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateOptions *arg1 = (CTemplateOptions *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplateOptions_Parse", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateOptions, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateOptions_Parse" "', argument " "1"" of type '" "CTemplateOptions *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateOptions * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplateOptions_Parse" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateOptions_Parse" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->Parse((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateOptions_GetEscapeFrom(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateOptions *arg1 = (CTemplateOptions *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString::EEscape result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateOptions, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateOptions_GetEscapeFrom" "', argument " "1"" of type '" "CTemplateOptions const *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateOptions * >(argp1);
  result = ((CTemplateOptions const *)arg1)->GetEscapeFrom();
  resultobj = SWIG_NewPointerObj((new CString::EEscape(result)), SWIGTYPE_p_CString__EEscape, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateOptions_GetEscapeTo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateOptions *arg1 = (CTemplateOptions *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString::EEscape result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateOptions, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateOptions_GetEscapeTo" "', argument " "1"" of type '" "CTemplateOptions const *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateOptions * >(argp1);
  result = ((CTemplateOptions const *)arg1)->GetEscapeTo();
  resultobj = SWIG_NewPointerObj((new CString::EEscape(result)), SWIGTYPE_p_CString__EEscape, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CTemplateOptions_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CTemplateOptions, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CTemplateOptions_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CTemplateLoopContext__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  unsigned long arg1 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  std::vector< CTemplate *,std::allocator< CTemplate * > > *arg4 = (std::vector< CTemplate *,std::allocator< CTemplate * > > *) 0 ;
  unsigned long val1 ;
  int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  CTemplateLoopContext *result = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_long(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_CTemplateLoopContext" "', argument " "1"" of type '" "unsigned long""'");
  } 
  arg1 = static_cast< unsigned long >(val1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CTemplateLoopContext" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CTemplateLoopContext" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CTemplateLoopContext" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_std__vectorT_CTemplate_p_std__allocatorT_CTemplate_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CTemplateLoopContext" "', argument " "4"" of type '" "std::vector< CTemplate *,std::allocator< CTemplate * > > *""'"); 
  }
  arg4 = reinterpret_cast< std::vector< CTemplate *,std::allocator< CTemplate * > > * >(argp4);
  result = (CTemplateLoopContext *)new CTemplateLoopContext(arg1,(CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplateLoopContext, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CTemplateLoopContext(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CTemplateLoopContext" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CTemplateLoopContext__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CTemplateLoopContext *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CTemplateLoopContext,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CTemplateLoopContext" "', argument " "1"" of type '" "CTemplateLoopContext const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CTemplateLoopContext" "', argument " "1"" of type '" "CTemplateLoopContext const &""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  result = (CTemplateLoopContext *)new CTemplateLoopContext((CTemplateLoopContext const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplateLoopContext, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CTemplateLoopContext(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CTemplateLoopContext", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_CTemplateLoopContext, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CTemplateLoopContext__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 4) {
    int _v = 0;
    {
      int res = SWIG_AsVal_unsigned_SS_long(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_std__vectorT_CTemplate_p_std__allocatorT_CTemplate_p_t_t, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_new_CTemplateLoopContext__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CTemplateLoopContext'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CTemplateLoopContext::CTemplateLoopContext(unsigned long,CString const &,bool,std::vector< CTemplate *,std::allocator< CTemplate * > > *)\n"
    "    CTemplateLoopContext::CTemplateLoopContext(CTemplateLoopContext const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_SetHasData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_SetHasData" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CTemplateLoopContext_SetHasData" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetHasData(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_SetHasData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_SetHasData" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  (arg1)->SetHasData();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_SetHasData(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CTemplateLoopContext_SetHasData", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplateLoopContext, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CTemplateLoopContext_SetHasData__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplateLoopContext, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CTemplateLoopContext_SetHasData__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CTemplateLoopContext_SetHasData'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CTemplateLoopContext::SetHasData(bool)\n"
    "    CTemplateLoopContext::SetHasData()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_SetName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplateLoopContext_SetName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_SetName" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplateLoopContext_SetName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateLoopContext_SetName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_SetRowIndex(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplateLoopContext_SetRowIndex", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_SetRowIndex" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CTemplateLoopContext_SetRowIndex" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  (arg1)->SetRowIndex(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_IncRowIndex(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_IncRowIndex" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  result = (unsigned int)(arg1)->IncRowIndex();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_DecRowIndex(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_DecRowIndex" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  result = (unsigned int)(arg1)->DecRowIndex();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_SetFilePosition(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplateLoopContext_SetFilePosition", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_SetFilePosition" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CTemplateLoopContext_SetFilePosition" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  (arg1)->SetFilePosition(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_HasData(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_HasData" "', argument " "1"" of type '" "CTemplateLoopContext const *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  result = (bool)((CTemplateLoopContext const *)arg1)->HasData();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_GetName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_GetName" "', argument " "1"" of type '" "CTemplateLoopContext const *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  result = (CString *) &((CTemplateLoopContext const *)arg1)->GetName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_GetFilePosition(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_GetFilePosition" "', argument " "1"" of type '" "CTemplateLoopContext const *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  result = (unsigned long)((CTemplateLoopContext const *)arg1)->GetFilePosition();
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_GetRowIndex(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_GetRowIndex" "', argument " "1"" of type '" "CTemplateLoopContext const *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  result = (unsigned int)((CTemplateLoopContext const *)arg1)->GetRowIndex();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_GetRowCount(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  size_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_GetRowCount" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  result = (arg1)->GetRowCount();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_GetRows(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CTemplate *,std::allocator< CTemplate * > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_GetRows" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  result = (std::vector< CTemplate *,std::allocator< CTemplate * > > *)(arg1)->GetRows();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CTemplate_p_std__allocatorT_CTemplate_p_t_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_GetNextRow(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CTemplate *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_GetNextRow" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  result = (CTemplate *)(arg1)->GetNextRow();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplate, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_GetCurRow(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CTemplate *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_GetCurRow" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  result = (CTemplate *)(arg1)->GetCurRow();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplate, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_GetRow(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  CTemplate *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplateLoopContext_GetRow", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_GetRow" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CTemplateLoopContext_GetRow" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = (CTemplate *)(arg1)->GetRow(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplate, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_GetValue__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_GetValue" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplateLoopContext_GetValue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateLoopContext_GetValue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CTemplateLoopContext_GetValue" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (arg1)->GetValue((CString const &)*arg2,arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_GetValue__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplateLoopContext *arg1 = (CTemplateLoopContext *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplateLoopContext_GetValue" "', argument " "1"" of type '" "CTemplateLoopContext *""'"); 
  }
  arg1 = reinterpret_cast< CTemplateLoopContext * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplateLoopContext_GetValue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplateLoopContext_GetValue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->GetValue((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplateLoopContext_GetValue(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CTemplateLoopContext_GetValue", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplateLoopContext, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CTemplateLoopContext_GetValue__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplateLoopContext, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CTemplateLoopContext_GetValue__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CTemplateLoopContext_GetValue'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CTemplateLoopContext::GetValue(CString const &,bool)\n"
    "    CTemplateLoopContext::GetValue(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *CTemplateLoopContext_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CTemplateLoopContext, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CTemplateLoopContext_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CTemplate__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CTemplate *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CTemplate *)new CTemplate();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplate, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CTemplate__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  CTemplate *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CTemplate" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CTemplate" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (CTemplate *)new CTemplate((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplate, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CTemplate__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::shared_ptr< CTemplateOptions > *arg1 = 0 ;
  CTemplate *arg2 = (CTemplate *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  CTemplate *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CTemplateOptions_t,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CTemplate" "', argument " "1"" of type '" "std::shared_ptr< CTemplateOptions > const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CTemplate" "', argument " "1"" of type '" "std::shared_ptr< CTemplateOptions > const &""'"); 
  }
  arg1 = reinterpret_cast< std::shared_ptr< CTemplateOptions > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CTemplate" "', argument " "2"" of type '" "CTemplate *""'"); 
  }
  arg2 = reinterpret_cast< CTemplate * >(argp2);
  result = (CTemplate *)new CTemplate((std::shared_ptr< CTemplateOptions > const &)*arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplate, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CTemplate__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::shared_ptr< CTemplateOptions > *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CTemplate *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CTemplateOptions_t,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CTemplate" "', argument " "1"" of type '" "std::shared_ptr< CTemplateOptions > const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CTemplate" "', argument " "1"" of type '" "std::shared_ptr< CTemplateOptions > const &""'"); 
  }
  arg1 = reinterpret_cast< std::shared_ptr< CTemplateOptions > * >(argp1);
  result = (CTemplate *)new CTemplate((std::shared_ptr< CTemplateOptions > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplate, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CTemplate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CTemplate" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CTemplate__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CTemplate *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CTemplate,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CTemplate" "', argument " "1"" of type '" "CTemplate const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CTemplate" "', argument " "1"" of type '" "CTemplate const &""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  result = (CTemplate *)new CTemplate((CTemplate const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplate, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CTemplate(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CTemplate", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CTemplate__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_std__shared_ptrT_CTemplateOptions_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CTemplate__SWIG_3(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_CTemplate, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CTemplate__SWIG_4(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CTemplate__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_std__shared_ptrT_CTemplateOptions_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CTemplate, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_CTemplate__SWIG_2(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CTemplate'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CTemplate::CTemplate()\n"
    "    CTemplate::CTemplate(CString const &)\n"
    "    CTemplate::CTemplate(std::shared_ptr< CTemplateOptions > const &,CTemplate *)\n"
    "    CTemplate::CTemplate(std::shared_ptr< CTemplateOptions > const &)\n"
    "    CTemplate::CTemplate(CTemplate const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CTemplate_AddTagHandler(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  SwigValueWrapper< std::shared_ptr< CTemplateTagHandler > > arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_AddTagHandler", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_AddTagHandler" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__shared_ptrT_CTemplateTagHandler_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_AddTagHandler" "', argument " "2"" of type '" "std::shared_ptr< CTemplateTagHandler >""'"); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_AddTagHandler" "', argument " "2"" of type '" "std::shared_ptr< CTemplateTagHandler >""'");
    } else {
      std::shared_ptr< CTemplateTagHandler > * temp = reinterpret_cast< std::shared_ptr< CTemplateTagHandler > * >(argp2);
      arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp;
    }
  }
  (arg1)->AddTagHandler(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_GetTagHandlers(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::shared_ptr< CTemplateTagHandler >,std::allocator< std::shared_ptr< CTemplateTagHandler > > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_GetTagHandlers" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  result = (std::vector< std::shared_ptr< CTemplateTagHandler >,std::allocator< std::shared_ptr< CTemplateTagHandler > > > *) &(arg1)->GetTagHandlers();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__shared_ptrT_CTemplateTagHandler_t_std__allocatorT_std__shared_ptrT_CTemplateTagHandler_t_t_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_ResolveLiteral(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_ResolveLiteral", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_ResolveLiteral" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_ResolveLiteral" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_ResolveLiteral" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->ResolveLiteral((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_Init(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_Init" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  (arg1)->Init();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_GetParent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  CTemplate *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_GetParent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_GetParent" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CTemplate_GetParent" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = (CTemplate *)(arg1)->GetParent(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplate, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_ExpandFile__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_ExpandFile" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_ExpandFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_ExpandFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CTemplate_ExpandFile" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (arg1)->ExpandFile((CString const &)*arg2,arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_ExpandFile__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_ExpandFile" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_ExpandFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_ExpandFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->ExpandFile((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_ExpandFile(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CTemplate_ExpandFile", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplate, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CTemplate_ExpandFile__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplate, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CTemplate_ExpandFile__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CTemplate_ExpandFile'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CTemplate::ExpandFile(CString const &,bool)\n"
    "    CTemplate::ExpandFile(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CTemplate_SetFile(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_SetFile", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_SetFile" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_SetFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_SetFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->SetFile((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_SetPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_SetPath", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_SetPath" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_SetPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_SetPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetPath((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_MakePath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_MakePath", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_MakePath" "', argument " "1"" of type '" "CTemplate const *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_MakePath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_MakePath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CTemplate const *)arg1)->MakePath((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_PrependPath__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_PrependPath" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_PrependPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_PrependPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CTemplate_PrependPath" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->PrependPath((CString const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_PrependPath__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_PrependPath" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_PrependPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_PrependPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->PrependPath((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_PrependPath(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CTemplate_PrependPath", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplate, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CTemplate_PrependPath__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplate, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CTemplate_PrependPath__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CTemplate_PrependPath'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CTemplate::PrependPath(CString const &,bool)\n"
    "    CTemplate::PrependPath(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CTemplate_AppendPath__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_AppendPath" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_AppendPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_AppendPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CTemplate_AppendPath" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->AppendPath((CString const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_AppendPath__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_AppendPath" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_AppendPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_AppendPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->AppendPath((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_AppendPath(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CTemplate_AppendPath", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplate, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CTemplate_AppendPath__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplate, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CTemplate_AppendPath__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CTemplate_AppendPath'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CTemplate::AppendPath(CString const &,bool)\n"
    "    CTemplate::AppendPath(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CTemplate_RemovePath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_RemovePath", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_RemovePath" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_RemovePath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_RemovePath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->RemovePath((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_ClearPaths(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_ClearPaths" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  (arg1)->ClearPaths();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_PrintString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_PrintString", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_PrintString" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sRet");
    }
  }
  result = (bool)(arg1)->PrintString(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_Print__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  std::ostream *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_Print" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__ostream,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_Print" "', argument " "2"" of type '" "std::ostream &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_Print" "', argument " "2"" of type '" "std::ostream &""'"); 
  }
  arg2 = reinterpret_cast< std::ostream * >(argp2);
  result = (bool)(arg1)->Print(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_Print__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  std::ostream *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_Print" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_Print" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_Print" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__ostream,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CTemplate_Print" "', argument " "3"" of type '" "std::ostream &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_Print" "', argument " "3"" of type '" "std::ostream &""'"); 
  }
  arg3 = reinterpret_cast< std::ostream * >(argp3);
  result = (bool)(arg1)->Print((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_Print(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CTemplate_Print", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplate, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__ostream, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CTemplate_Print__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplate, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_std__ostream, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CTemplate_Print__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CTemplate_Print'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CTemplate::Print(std::ostream &)\n"
    "    CTemplate::Print(CString const &,std::ostream &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CTemplate_ValidIf(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_ValidIf", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_ValidIf" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_ValidIf" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_ValidIf" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->ValidIf((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_ValidExpr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_ValidExpr", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_ValidExpr" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_ValidExpr" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_ValidExpr" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->ValidExpr((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_IsTrue(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_IsTrue", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_IsTrue" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_IsTrue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_IsTrue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->IsTrue((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_HasLoop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_HasLoop", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_HasLoop" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_HasLoop" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_HasLoop" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->HasLoop((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_GetValue__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_GetValue" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_GetValue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_GetValue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CTemplate_GetValue" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (arg1)->GetValue((CString const &)*arg2,arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_GetValue__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_GetValue" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_GetValue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_GetValue" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->GetValue((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_GetValue(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CTemplate_GetValue", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplate, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CTemplate_GetValue__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CTemplate, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CTemplate_GetValue__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CTemplate_GetValue'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CTemplate::GetValue(CString const &,bool)\n"
    "    CTemplate::GetValue(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CTemplate_AddRow(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CTemplate *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_AddRow", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_AddRow" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_AddRow" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_AddRow" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CTemplate *) &(arg1)->AddRow((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplate, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_GetRow(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  CTemplate *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_GetRow", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_GetRow" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_GetRow" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_GetRow" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CTemplate_GetRow" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  result = (CTemplate *)(arg1)->GetRow((CString const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplate, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_GetLoop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  std::vector< CTemplate *,std::allocator< CTemplate * > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_GetLoop", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_GetLoop" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_GetLoop" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_GetLoop" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (std::vector< CTemplate *,std::allocator< CTemplate * > > *)(arg1)->GetLoop((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CTemplate_p_std__allocatorT_CTemplate_p_t_t, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_DelCurLoopContext(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_DelCurLoopContext" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  (arg1)->DelCurLoopContext();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_GetCurLoopContext(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CTemplateLoopContext *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_GetCurLoopContext" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  result = (CTemplateLoopContext *)(arg1)->GetCurLoopContext();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplateLoopContext, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_GetCurTemplate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CTemplate *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_GetCurTemplate" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  result = (CTemplate *)(arg1)->GetCurTemplate();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CTemplate, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_GetFileName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_GetFileName" "', argument " "1"" of type '" "CTemplate const *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  result = (CString *) &((CTemplate const *)arg1)->GetFileName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CTemplate_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CTemplate *arg1 = (CTemplate *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CTemplate_set", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CTemplate, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CTemplate_set" "', argument " "1"" of type '" "CTemplate *""'"); 
  }
  arg1 = reinterpret_cast< CTemplate * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CTemplate_set" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_set" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CTemplate_set" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CTemplate_set" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  CTemplate_set(arg1,(CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *CTemplate_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CTemplate, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CTemplate_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CZNCTagHandler(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CZNCTagHandler *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CWebSock,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CZNCTagHandler" "', argument " "1"" of type '" "CWebSock &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CZNCTagHandler" "', argument " "1"" of type '" "CWebSock &""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  result = (CZNCTagHandler *)new CZNCTagHandler(*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CZNCTagHandler, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CZNCTagHandler(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNCTagHandler *arg1 = (CZNCTagHandler *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNCTagHandler, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CZNCTagHandler" "', argument " "1"" of type '" "CZNCTagHandler *""'"); 
  }
  arg1 = reinterpret_cast< CZNCTagHandler * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNCTagHandler_HandleTag(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNCTagHandler *arg1 = (CZNCTagHandler *) 0 ;
  CTemplate *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNCTagHandler_HandleTag", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNCTagHandler, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNCTagHandler_HandleTag" "', argument " "1"" of type '" "CZNCTagHandler *""'"); 
  }
  arg1 = reinterpret_cast< CZNCTagHandler * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTemplate,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNCTagHandler_HandleTag" "', argument " "2"" of type '" "CTemplate &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNCTagHandler_HandleTag" "', argument " "2"" of type '" "CTemplate &""'"); 
  }
  arg2 = reinterpret_cast< CTemplate * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CZNCTagHandler_HandleTag" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNCTagHandler_HandleTag" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CZNCTagHandler_HandleTag" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNCTagHandler_HandleTag" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[4], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg5 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 5 sOutput");
    }
  }
  result = (bool)(arg1)->HandleTag(*arg2,(CString const &)*arg3,(CString const &)*arg4,*arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *CZNCTagHandler_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CZNCTagHandler, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CZNCTagHandler_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CWebSession(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CWebSession *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CWebSession", 2, 2, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CWebSession" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSession" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CWebSession" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSession" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CWebSession *)new CWebSession((CString const &)*arg1,(CString const &)*arg2);
  {
    std::shared_ptr<  CWebSession > *smartresult = result ? new std::shared_ptr<  CWebSession >(result SWIG_NO_NULL_DELETER_SWIG_POINTER_NEW) : 0;
    resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(smartresult), SWIGTYPE_p_std__shared_ptrT_CWebSession_t, SWIG_POINTER_NEW | SWIG_POINTER_OWN);
  }
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CWebSession(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSession *arg1 = (CWebSession *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CWebSession > tempshared1 ;
  std::shared_ptr< CWebSession > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CWebSession_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CWebSession" "', argument " "1"" of type '" "CWebSession *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  (void)arg1; delete smartarg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSession_GetId(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSession *arg1 = (CWebSession *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CWebSession const > tempshared1 ;
  std::shared_ptr< CWebSession const > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CWebSession_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSession_GetId" "', argument " "1"" of type '" "CWebSession const *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      delete reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  result = (CString *) &((CWebSession const *)arg1)->GetId();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSession_GetIP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSession *arg1 = (CWebSession *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CWebSession const > tempshared1 ;
  std::shared_ptr< CWebSession const > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CWebSession_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSession_GetIP" "', argument " "1"" of type '" "CWebSession const *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      delete reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  result = (CString *) &((CWebSession const *)arg1)->GetIP();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSession_GetUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSession *arg1 = (CWebSession *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CWebSession const > tempshared1 ;
  std::shared_ptr< CWebSession const > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  CUser *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CWebSession_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSession_GetUser" "', argument " "1"" of type '" "CWebSession const *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      delete reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  result = (CUser *)((CWebSession const *)arg1)->GetUser();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CUser, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSession_GetLastActive(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSession *arg1 = (CWebSession *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CWebSession const > tempshared1 ;
  std::shared_ptr< CWebSession const > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  time_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CWebSession_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSession_GetLastActive" "', argument " "1"" of type '" "CWebSession const *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      delete reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  result = ((CWebSession const *)arg1)->GetLastActive();
  resultobj = SWIG_NewPointerObj((new time_t(result)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSession_IsLoggedIn(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSession *arg1 = (CWebSession *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CWebSession const > tempshared1 ;
  std::shared_ptr< CWebSession const > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CWebSession_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSession_IsLoggedIn" "', argument " "1"" of type '" "CWebSession const *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      delete reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  result = (bool)((CWebSession const *)arg1)->IsLoggedIn();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSession_IsAdmin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSession *arg1 = (CWebSession *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CWebSession const > tempshared1 ;
  std::shared_ptr< CWebSession const > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CWebSession_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSession_IsAdmin" "', argument " "1"" of type '" "CWebSession const *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      delete reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr< const CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  result = (bool)((CWebSession const *)arg1)->IsAdmin();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSession_UpdateLastActive(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSession *arg1 = (CWebSession *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CWebSession > tempshared1 ;
  std::shared_ptr< CWebSession > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CWebSession_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSession_UpdateLastActive" "', argument " "1"" of type '" "CWebSession *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  (arg1)->UpdateLastActive();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSession_SetUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSession *arg1 = (CWebSession *) 0 ;
  CUser *arg2 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CWebSession > tempshared1 ;
  std::shared_ptr< CWebSession > *smartarg1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CUser *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSession_SetUser", 2, 2, swig_obj)) SWIG_fail;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CWebSession_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSession_SetUser" "', argument " "1"" of type '" "CWebSession *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSession_SetUser" "', argument " "2"" of type '" "CUser *""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  result = (CUser *)(arg1)->SetUser(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CUser, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSession_ClearMessageLoops(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSession *arg1 = (CWebSession *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CWebSession > tempshared1 ;
  std::shared_ptr< CWebSession > *smartarg1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CWebSession_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSession_ClearMessageLoops" "', argument " "1"" of type '" "CWebSession *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  (arg1)->ClearMessageLoops();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSession_FillMessageLoops(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSession *arg1 = (CWebSession *) 0 ;
  CTemplate *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CWebSession > tempshared1 ;
  std::shared_ptr< CWebSession > *smartarg1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSession_FillMessageLoops", 2, 2, swig_obj)) SWIG_fail;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CWebSession_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSession_FillMessageLoops" "', argument " "1"" of type '" "CWebSession *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTemplate,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSession_FillMessageLoops" "', argument " "2"" of type '" "CTemplate &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSession_FillMessageLoops" "', argument " "2"" of type '" "CTemplate &""'"); 
  }
  arg2 = reinterpret_cast< CTemplate * >(argp2);
  (arg1)->FillMessageLoops(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSession_AddError(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSession *arg1 = (CWebSession *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CWebSession > tempshared1 ;
  std::shared_ptr< CWebSession > *smartarg1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  size_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSession_AddError", 2, 2, swig_obj)) SWIG_fail;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CWebSession_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSession_AddError" "', argument " "1"" of type '" "CWebSession *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSession_AddError" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSession_AddError" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->AddError((CString const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSession_AddSuccess(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSession *arg1 = (CWebSession *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::shared_ptr< CWebSession > tempshared1 ;
  std::shared_ptr< CWebSession > *smartarg1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  size_t result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSession_AddSuccess", 2, 2, swig_obj)) SWIG_fail;
  {
    int newmem = 0;
    res1 = SWIG_ConvertPtrAndOwn(swig_obj[0], &argp1, SWIGTYPE_p_std__shared_ptrT_CWebSession_t, 0 |  0 , &newmem);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSession_AddSuccess" "', argument " "1"" of type '" "CWebSession *""'");
    }
    if (newmem & SWIG_CAST_NEW_MEMORY) {
      tempshared1 = *reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      delete reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >(tempshared1.get());
    } else {
      smartarg1 = reinterpret_cast< std::shared_ptr<  CWebSession > * >(argp1);
      arg1 = const_cast< CWebSession * >((smartarg1 ? smartarg1->get() : 0));
    }
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSession_AddSuccess" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSession_AddSuccess" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->AddSuccess((CString const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *CWebSession_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__shared_ptrT_CWebSession_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CWebSession_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CWebSubPage__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  unsigned int arg3 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  CWebSubPage *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CWebSubPage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSubPage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CWebSubPage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSubPage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CWebSubPage" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  result = (CWebSubPage *)new CWebSubPage((CString const &)*arg1,(CString const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CWebSubPage, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CWebSubPage__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  CWebSubPage *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CWebSubPage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSubPage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CWebSubPage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSubPage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CWebSubPage *)new CWebSubPage((CString const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CWebSubPage, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CWebSubPage__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  CWebSubPage *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CWebSubPage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSubPage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (CWebSubPage *)new CWebSubPage((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CWebSubPage, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CWebSubPage__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  COptionalTranslation *arg2 = 0 ;
  VPair *arg3 = 0 ;
  unsigned int arg4 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  CWebSubPage *result = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CWebSubPage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSubPage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_COptionalTranslation,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CWebSubPage" "', argument " "2"" of type '" "COptionalTranslation const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSubPage" "', argument " "2"" of type '" "COptionalTranslation const &""'"); 
  }
  arg2 = reinterpret_cast< COptionalTranslation * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CWebSubPage" "', argument " "3"" of type '" "VPair const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSubPage" "', argument " "3"" of type '" "VPair const &""'"); 
  }
  arg3 = reinterpret_cast< VPair * >(argp3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_CWebSubPage" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  result = (CWebSubPage *)new CWebSubPage((CString const &)*arg1,(COptionalTranslation const &)*arg2,(VPair const &)*arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CWebSubPage, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CWebSubPage__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  COptionalTranslation *arg2 = 0 ;
  VPair *arg3 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  CWebSubPage *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CWebSubPage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSubPage" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_COptionalTranslation,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CWebSubPage" "', argument " "2"" of type '" "COptionalTranslation const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSubPage" "', argument " "2"" of type '" "COptionalTranslation const &""'"); 
  }
  arg2 = reinterpret_cast< COptionalTranslation * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CWebSubPage" "', argument " "3"" of type '" "VPair const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSubPage" "', argument " "3"" of type '" "VPair const &""'"); 
  }
  arg3 = reinterpret_cast< VPair * >(argp3);
  result = (CWebSubPage *)new CWebSubPage((CString const &)*arg1,(COptionalTranslation const &)*arg2,(VPair const &)*arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CWebSubPage, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CWebSubPage(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CWebSubPage", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CWebSubPage__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_CWebSubPage__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_COptionalTranslation, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_new_CWebSubPage__SWIG_4(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_new_CWebSubPage__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_COptionalTranslation, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_new_CWebSubPage__SWIG_3(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CWebSubPage'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CWebSubPage::CWebSubPage(CString const &,CString const &,unsigned int)\n"
    "    CWebSubPage::CWebSubPage(CString const &,CString const &)\n"
    "    CWebSubPage::CWebSubPage(CString const &)\n"
    "    CWebSubPage::CWebSubPage(CString const &,COptionalTranslation const &,VPair const &,unsigned int)\n"
    "    CWebSubPage::CWebSubPage(CString const &,COptionalTranslation const &,VPair const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CWebSubPage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSubPage *arg1 = (CWebSubPage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSubPage, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CWebSubPage" "', argument " "1"" of type '" "CWebSubPage *""'"); 
  }
  arg1 = reinterpret_cast< CWebSubPage * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSubPage_SetName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSubPage *arg1 = (CWebSubPage *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSubPage_SetName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSubPage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSubPage_SetName" "', argument " "1"" of type '" "CWebSubPage *""'"); 
  }
  arg1 = reinterpret_cast< CWebSubPage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSubPage_SetName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSubPage_SetName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSubPage_SetTitle(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSubPage *arg1 = (CWebSubPage *) 0 ;
  COptionalTranslation *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSubPage_SetTitle", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSubPage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSubPage_SetTitle" "', argument " "1"" of type '" "CWebSubPage *""'"); 
  }
  arg1 = reinterpret_cast< CWebSubPage * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_COptionalTranslation,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSubPage_SetTitle" "', argument " "2"" of type '" "COptionalTranslation const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSubPage_SetTitle" "', argument " "2"" of type '" "COptionalTranslation const &""'"); 
  }
  arg2 = reinterpret_cast< COptionalTranslation * >(argp2);
  (arg1)->SetTitle((COptionalTranslation const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSubPage_AddParam(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSubPage *arg1 = (CWebSubPage *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSubPage_AddParam", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSubPage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSubPage_AddParam" "', argument " "1"" of type '" "CWebSubPage *""'"); 
  }
  arg1 = reinterpret_cast< CWebSubPage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSubPage_AddParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSubPage_AddParam" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CWebSubPage_AddParam" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSubPage_AddParam" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->AddParam((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSubPage_RequiresAdmin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSubPage *arg1 = (CWebSubPage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSubPage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSubPage_RequiresAdmin" "', argument " "1"" of type '" "CWebSubPage const *""'"); 
  }
  arg1 = reinterpret_cast< CWebSubPage * >(argp1);
  result = (bool)((CWebSubPage const *)arg1)->RequiresAdmin();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSubPage_GetName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSubPage *arg1 = (CWebSubPage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSubPage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSubPage_GetName" "', argument " "1"" of type '" "CWebSubPage const *""'"); 
  }
  arg1 = reinterpret_cast< CWebSubPage * >(argp1);
  result = (CString *) &((CWebSubPage const *)arg1)->GetName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSubPage_GetTitle(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSubPage *arg1 = (CWebSubPage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSubPage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSubPage_GetTitle" "', argument " "1"" of type '" "CWebSubPage const *""'"); 
  }
  arg1 = reinterpret_cast< CWebSubPage * >(argp1);
  result = ((CWebSubPage const *)arg1)->GetTitle();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSubPage_GetParams(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSubPage *arg1 = (CWebSubPage *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  VPair *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSubPage, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSubPage_GetParams" "', argument " "1"" of type '" "CWebSubPage const *""'"); 
  }
  arg1 = reinterpret_cast< CWebSubPage * >(argp1);
  result = (VPair *) &((CWebSubPage const *)arg1)->GetParams();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CWebSubPage_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CWebSubPage, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CWebSubPage_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CWebSessionMap__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  CWebSessionMap *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_CWebSessionMap" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  result = (CWebSessionMap *)new CWebSessionMap(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CWebSessionMap, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CWebSessionMap__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CWebSessionMap *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CWebSessionMap *)new CWebSessionMap();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CWebSessionMap, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CWebSessionMap(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CWebSessionMap", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CWebSessionMap__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_unsigned_SS_int(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_CWebSessionMap__SWIG_0(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CWebSessionMap'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CWebSessionMap::CWebSessionMap(unsigned int)\n"
    "    CWebSessionMap::CWebSessionMap()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CWebSessionMap_FinishUserSessions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSessionMap *arg1 = (CWebSessionMap *) 0 ;
  CUser *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSessionMap_FinishUserSessions", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSessionMap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSessionMap_FinishUserSessions" "', argument " "1"" of type '" "CWebSessionMap *""'"); 
  }
  arg1 = reinterpret_cast< CWebSessionMap * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSessionMap_FinishUserSessions" "', argument " "2"" of type '" "CUser const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSessionMap_FinishUserSessions" "', argument " "2"" of type '" "CUser const &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  (arg1)->FinishUserSessions((CUser const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CWebSessionMap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSessionMap *arg1 = (CWebSessionMap *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSessionMap, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CWebSessionMap" "', argument " "1"" of type '" "CWebSessionMap *""'"); 
  }
  arg1 = reinterpret_cast< CWebSessionMap * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CWebSessionMap_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CWebSessionMap, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CWebSessionMap_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CWebSock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  CWebSock *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CWebSock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CWebSock" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (CWebSock *)new CWebSock((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CWebSock, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CWebSock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CWebSock" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_ForceLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_ForceLogin" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  result = (bool)(arg1)->ForceLogin();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_OnLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSock_OnLogin", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_OnLogin" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSock_OnLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_OnLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CWebSock_OnLogin" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_OnLogin" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CWebSock_OnLogin" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->OnLogin((CString const &)*arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_OnPageRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSock_OnPageRequest", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_OnPageRequest" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSock_OnPageRequest" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_OnPageRequest" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->OnPageRequest((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_PrintTemplate__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CModule *arg4 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  CWebSock::EPageReqResult result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_PrintTemplate" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSock_PrintTemplate" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_PrintTemplate" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sPageRet");
    }
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CWebSock_PrintTemplate" "', argument " "4"" of type '" "CModule *""'"); 
  }
  arg4 = reinterpret_cast< CModule * >(argp4);
  result = (CWebSock::EPageReqResult)(arg1)->PrintTemplate((CString const &)*arg2,*arg3,arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_PrintTemplate__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CWebSock::EPageReqResult result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_PrintTemplate" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSock_PrintTemplate" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_PrintTemplate" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sPageRet");
    }
  }
  result = (CWebSock::EPageReqResult)(arg1)->PrintTemplate((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_PrintTemplate(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CWebSock_PrintTemplate", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CWebSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          return _wrap_CWebSock_PrintTemplate__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CWebSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CModule, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CWebSock_PrintTemplate__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CWebSock_PrintTemplate'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CWebSock::PrintTemplate(CString const &,CString &,CModule *)\n"
    "    CWebSock::PrintTemplate(CString const &,CString &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CWebSock_PrintStaticFile__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CModule *arg4 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  CWebSock::EPageReqResult result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_PrintStaticFile" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSock_PrintStaticFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_PrintStaticFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sPageRet");
    }
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CWebSock_PrintStaticFile" "', argument " "4"" of type '" "CModule *""'"); 
  }
  arg4 = reinterpret_cast< CModule * >(argp4);
  result = (CWebSock::EPageReqResult)(arg1)->PrintStaticFile((CString const &)*arg2,*arg3,arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_PrintStaticFile__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CWebSock::EPageReqResult result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_PrintStaticFile" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSock_PrintStaticFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_PrintStaticFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sPageRet");
    }
  }
  result = (CWebSock::EPageReqResult)(arg1)->PrintStaticFile((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_PrintStaticFile(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CWebSock_PrintStaticFile", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CWebSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          return _wrap_CWebSock_PrintStaticFile__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CWebSock, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CModule, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CWebSock_PrintStaticFile__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CWebSock_PrintStaticFile'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CWebSock::PrintStaticFile(CString const &,CString &,CModule *)\n"
    "    CWebSock::PrintStaticFile(CString const &,CString &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CWebSock_FindTmpl(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  CModule *arg2 = (CModule *) 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSock_FindTmpl", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_FindTmpl" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSock_FindTmpl" "', argument " "2"" of type '" "CModule *""'"); 
  }
  arg2 = reinterpret_cast< CModule * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CWebSock_FindTmpl" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_FindTmpl" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (arg1)->FindTmpl(arg2,(CString const &)*arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_PrintErrorPage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSock_PrintErrorPage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_PrintErrorPage" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSock_PrintErrorPage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_PrintErrorPage" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->PrintErrorPage((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_GetSession(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::shared_ptr< CWebSession > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_GetSession" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  result = (arg1)->GetSession();
  {
    std::shared_ptr<  CWebSession > *smartresult = result ? new std::shared_ptr<  CWebSession >(result) : 0;
    resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(smartresult), SWIGTYPE_p_std__shared_ptrT_CWebSession_t, SWIG_POINTER_OWN);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_GetSockObj(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Csock *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSock_GetSockObj", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_GetSockObj" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSock_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CWebSock_GetSockObj" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  result = (Csock *)(arg1)->GetSockObj((CString const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_GetSkinPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_GetSkinPath" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_GetSkinPath" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = CWebSock::GetSkinPath((CString const &)*arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_GetAvailSkins(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  VCString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSock_GetAvailSkins", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_GetAvailSkins" "', argument " "1"" of type '" "CWebSock const *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_CString_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSock_GetAvailSkins" "', argument " "2"" of type '" "VCString &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_GetAvailSkins" "', argument " "2"" of type '" "VCString &""'"); 
  }
  arg2 = reinterpret_cast< VCString * >(argp2);
  ((CWebSock const *)arg1)->GetAvailSkins(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_GetSkinName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_GetSkinName" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  result = (arg1)->GetSkinName();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_GetRequestCookie(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSock_GetRequestCookie", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_GetRequestCookie" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSock_GetRequestCookie" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_GetRequestCookie" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->GetRequestCookie((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_SendCookie(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSock_SendCookie", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_SendCookie" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSock_SendCookie" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_SendCookie" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CWebSock_SendCookie" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_SendCookie" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->SendCookie((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_FinishUserSessions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CUser,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_FinishUserSessions" "', argument " "1"" of type '" "CUser const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_FinishUserSessions" "', argument " "1"" of type '" "CUser const &""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  CWebSock::FinishUserSessions((CUser const &)*arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_GetCSRFCheck(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_GetCSRFCheck" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  result = (arg1)->GetCSRFCheck();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CWebSock_ValidateCSRFCheck(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CWebSock *arg1 = (CWebSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CWebSock_ValidateCSRFCheck", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CWebSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CWebSock_ValidateCSRFCheck" "', argument " "1"" of type '" "CWebSock *""'"); 
  }
  arg1 = reinterpret_cast< CWebSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CWebSock_ValidateCSRFCheck" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CWebSock_ValidateCSRFCheck" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->ValidateCSRFCheck((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *CWebSock_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CWebSock, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CWebSock_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CZNC(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CZNC", 0, 0, 0)) SWIG_fail;
  result = (CZNC *)new CZNC();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CZNC, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CZNC(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CZNC" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_DeleteUsers(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_DeleteUsers" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  (arg1)->DeleteUsers();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_Loop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_Loop" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  (arg1)->Loop();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_WritePidFile(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_WritePidFile", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_WritePidFile" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_WritePidFile" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  result = (bool)(arg1)->WritePidFile(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_DeletePidFile(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_DeletePidFile" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (bool)(arg1)->DeletePidFile();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_WaitForChildLock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_WaitForChildLock" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (bool)(arg1)->WaitForChildLock();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_IsHostAllowed(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_IsHostAllowed", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_IsHostAllowed" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_IsHostAllowed" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_IsHostAllowed" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)((CZNC const *)arg1)->IsHostAllowed((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AllowConnectionFrom(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_AllowConnectionFrom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AllowConnectionFrom" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_AllowConnectionFrom" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_AllowConnectionFrom" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)((CZNC const *)arg1)->AllowConnectionFrom((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_InitDirs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_InitDirs", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_InitDirs" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_InitDirs" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_InitDirs" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CZNC_InitDirs" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_InitDirs" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->InitDirs((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_OnBoot(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_OnBoot" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (bool)(arg1)->OnBoot();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_ExpandConfigPath__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_ExpandConfigPath" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_ExpandConfigPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_ExpandConfigPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CZNC_ExpandConfigPath" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (arg1)->ExpandConfigPath((CString const &)*arg2,arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_ExpandConfigPath__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_ExpandConfigPath" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_ExpandConfigPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_ExpandConfigPath" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->ExpandConfigPath((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_ExpandConfigPath(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CZNC_ExpandConfigPath", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CZNC_ExpandConfigPath__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CZNC_ExpandConfigPath__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CZNC_ExpandConfigPath'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CZNC::ExpandConfigPath(CString const &,bool)\n"
    "    CZNC::ExpandConfigPath(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CZNC_WriteNewConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_WriteNewConfig", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_WriteNewConfig" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_WriteNewConfig" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_WriteNewConfig" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->WriteNewConfig((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_WriteConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_WriteConfig" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (bool)(arg1)->WriteConfig();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_ParseConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_ParseConfig", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_ParseConfig" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_ParseConfig" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_ParseConfig" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sError");
    }
  }
  result = (bool)(arg1)->ParseConfig((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_RehashConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_RehashConfig", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_RehashConfig" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sError");
    }
  }
  result = (bool)(arg1)->RehashConfig(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_BackupConfigOnce(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_BackupConfigOnce", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_BackupConfigOnce" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_BackupConfigOnce" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_BackupConfigOnce" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->BackupConfigOnce((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetVersion(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_GetVersion", 0, 0, 0)) SWIG_fail;
  result = CZNC::GetVersion();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetTag__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  bool arg1 ;
  bool arg2 ;
  bool val1 ;
  int ecode1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_bool(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "CZNC_GetTag" "', argument " "1"" of type '" "bool""'");
  } 
  arg1 = static_cast< bool >(val1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_GetTag" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = CZNC::GetTag(arg1,arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetTag__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  bool arg1 ;
  bool val1 ;
  int ecode1 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_bool(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "CZNC_GetTag" "', argument " "1"" of type '" "bool""'");
  } 
  arg1 = static_cast< bool >(val1);
  result = CZNC::GetTag(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetTag__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CString result;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = CZNC::GetTag();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetTag(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CZNC_GetTag", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_CZNC_GetTag__SWIG_2(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_bool(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_CZNC_GetTag__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_bool(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CZNC_GetTag__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CZNC_GetTag'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CZNC::GetTag(bool,bool)\n"
    "    CZNC::GetTag(bool)\n"
    "    CZNC::GetTag()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CZNC_GetCompileOptionsString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_GetCompileOptionsString", 0, 0, 0)) SWIG_fail;
  result = CZNC::GetCompileOptionsString();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetUptime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetUptime" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = ((CZNC const *)arg1)->GetUptime();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_ClearBindHosts(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_ClearBindHosts" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  (arg1)->ClearBindHosts();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AddBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_AddBindHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AddBindHost" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_AddBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_AddBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->AddBindHost((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_RemBindHost(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_RemBindHost", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_RemBindHost" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_RemBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_RemBindHost" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->RemBindHost((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_ClearTrustedProxies(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_ClearTrustedProxies" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  (arg1)->ClearTrustedProxies();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AddTrustedProxy(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_AddTrustedProxy", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AddTrustedProxy" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_AddTrustedProxy" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_AddTrustedProxy" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->AddTrustedProxy((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_RemTrustedProxy(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_RemTrustedProxy", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_RemTrustedProxy" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_RemTrustedProxy" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_RemTrustedProxy" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->RemTrustedProxy((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetClientCapBlacklist(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SCString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetClientCapBlacklist" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (SCString *) &((CZNC const *)arg1)->GetClientCapBlacklist();
  resultobj = swig::from(static_cast< std::set< CString,std::less< CString >,std::allocator< CString > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetServerCapBlacklist(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SCString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetServerCapBlacklist" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (SCString *) &((CZNC const *)arg1)->GetServerCapBlacklist();
  resultobj = swig::from(static_cast< std::set< CString,std::less< CString >,std::allocator< CString > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_Broadcast__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  CUser *arg4 = (CUser *) 0 ;
  CClient *arg5 = (CClient *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_Broadcast" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_Broadcast" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_Broadcast" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CZNC_Broadcast" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CZNC_Broadcast" "', argument " "4"" of type '" "CUser *""'"); 
  }
  arg4 = reinterpret_cast< CUser * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CZNC_Broadcast" "', argument " "5"" of type '" "CClient *""'"); 
  }
  arg5 = reinterpret_cast< CClient * >(argp5);
  (arg1)->Broadcast((CString const &)*arg2,arg3,arg4,arg5);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_Broadcast__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  CUser *arg4 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_Broadcast" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_Broadcast" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_Broadcast" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CZNC_Broadcast" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CZNC_Broadcast" "', argument " "4"" of type '" "CUser *""'"); 
  }
  arg4 = reinterpret_cast< CUser * >(argp4);
  (arg1)->Broadcast((CString const &)*arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_Broadcast__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_Broadcast" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_Broadcast" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_Broadcast" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CZNC_Broadcast" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->Broadcast((CString const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_Broadcast__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_Broadcast" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_Broadcast" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_Broadcast" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->Broadcast((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_Broadcast(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CZNC_Broadcast", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CZNC_Broadcast__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CZNC_Broadcast__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CUser, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CZNC_Broadcast__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_CUser, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            void *vptr = 0;
            int res = SWIG_ConvertPtr(argv[4], &vptr, SWIGTYPE_p_CClient, 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_CZNC_Broadcast__SWIG_0(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CZNC_Broadcast'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CZNC::Broadcast(CString const &,bool,CUser *,CClient *)\n"
    "    CZNC::Broadcast(CString const &,bool,CUser *)\n"
    "    CZNC::Broadcast(CString const &,bool)\n"
    "    CZNC::Broadcast(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CZNC_AddBytesRead(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  unsigned long long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_AddBytesRead", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AddBytesRead" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_AddBytesRead" "', argument " "2"" of type '" "unsigned long long""'");
  } 
  arg2 = static_cast< unsigned long long >(val2);
  (arg1)->AddBytesRead(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AddBytesWritten(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  unsigned long long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_AddBytesWritten", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AddBytesWritten" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_AddBytesWritten" "', argument " "2"" of type '" "unsigned long long""'");
  } 
  arg2 = static_cast< unsigned long long >(val2);
  (arg1)->AddBytesWritten(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_BytesRead(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long long result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_BytesRead" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (unsigned long long)((CZNC const *)arg1)->BytesRead();
  resultobj = SWIG_From_unsigned_SS_long_SS_long(static_cast< unsigned long long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_BytesWritten(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long long result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_BytesWritten" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (unsigned long long)((CZNC const *)arg1)->BytesWritten();
  resultobj = SWIG_From_unsigned_SS_long_SS_long(static_cast< unsigned long long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetTrafficStats(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CZNC::TrafficStatsPair *arg2 = 0 ;
  CZNC::TrafficStatsPair *arg3 = 0 ;
  CZNC::TrafficStatsPair *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  SwigValueWrapper< std::map< CString,std::pair< unsigned long long,unsigned long long >,std::less< CString >,std::allocator< std::pair< CString const,std::pair< unsigned long long,unsigned long long > > > > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_GetTrafficStats", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetTrafficStats" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__pairT_unsigned_long_long_unsigned_long_long_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_GetTrafficStats" "', argument " "2"" of type '" "CZNC::TrafficStatsPair &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_GetTrafficStats" "', argument " "2"" of type '" "CZNC::TrafficStatsPair &""'"); 
  }
  arg2 = reinterpret_cast< CZNC::TrafficStatsPair * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__pairT_unsigned_long_long_unsigned_long_long_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CZNC_GetTrafficStats" "', argument " "3"" of type '" "CZNC::TrafficStatsPair &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_GetTrafficStats" "', argument " "3"" of type '" "CZNC::TrafficStatsPair &""'"); 
  }
  arg3 = reinterpret_cast< CZNC::TrafficStatsPair * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_std__pairT_unsigned_long_long_unsigned_long_long_t,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CZNC_GetTrafficStats" "', argument " "4"" of type '" "CZNC::TrafficStatsPair &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_GetTrafficStats" "', argument " "4"" of type '" "CZNC::TrafficStatsPair &""'"); 
  }
  arg4 = reinterpret_cast< CZNC::TrafficStatsPair * >(argp4);
  result = (arg1)->GetTrafficStats(*arg2,*arg3,*arg4);
  resultobj = SWIG_NewPointerObj((new CZNC::TrafficStatsMap(result)), SWIGTYPE_p_std__mapT_CString_std__pairT_unsigned_long_long_unsigned_long_long_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__pairT_unsigned_long_long_unsigned_long_long_t_t_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetNetworkTrafficStats(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  CZNC::TrafficStatsPair *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  SwigValueWrapper< std::map< CString,std::pair< unsigned long long,unsigned long long >,std::less< CString >,std::allocator< std::pair< CString const,std::pair< unsigned long long,unsigned long long > > > > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_GetNetworkTrafficStats", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetNetworkTrafficStats" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_GetNetworkTrafficStats" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_GetNetworkTrafficStats" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__pairT_unsigned_long_long_unsigned_long_long_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CZNC_GetNetworkTrafficStats" "', argument " "3"" of type '" "CZNC::TrafficStatsPair &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_GetNetworkTrafficStats" "', argument " "3"" of type '" "CZNC::TrafficStatsPair &""'"); 
  }
  arg3 = reinterpret_cast< CZNC::TrafficStatsPair * >(argp3);
  result = (arg1)->GetNetworkTrafficStats((CString const &)*arg2,*arg3);
  resultobj = SWIG_NewPointerObj((new CZNC::TrafficStatsMap(result)), SWIGTYPE_p_std__mapT_CString_std__pairT_unsigned_long_long_unsigned_long_long_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__pairT_unsigned_long_long_unsigned_long_long_t_t_t_t, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AuthUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  std::shared_ptr< CAuthBase > arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_AuthUser", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AuthUser" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    int newmem = 0;
    res2 = SWIG_ConvertPtrAndOwn(swig_obj[1], &argp2, SWIGTYPE_p_std__shared_ptrT_CAuthBase_t,  0 , &newmem);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_AuthUser" "', argument " "2"" of type '" "std::shared_ptr< CAuthBase >""'");
    }
    if (argp2) arg2 = *(reinterpret_cast< std::shared_ptr< CAuthBase > * >(argp2));
    if (newmem & SWIG_CAST_NEW_MEMORY) delete reinterpret_cast< std::shared_ptr< CAuthBase > * >(argp2);
  }
  (arg1)->AuthUser(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetConfigState(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  enum CZNC::ConfigState arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetConfigState", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetConfigState" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_SetConfigState" "', argument " "2"" of type '" "enum CZNC::ConfigState""'");
  } 
  arg2 = static_cast< enum CZNC::ConfigState >(val2);
  (arg1)->SetConfigState(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetSkinName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetSkinName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetSkinName" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_SetSkinName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_SetSkinName" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetSkinName((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetStatusPrefix(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetStatusPrefix", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetStatusPrefix" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_SetStatusPrefix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_SetStatusPrefix" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetStatusPrefix((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetMaxBufferSize(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetMaxBufferSize", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetMaxBufferSize" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_SetMaxBufferSize" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  (arg1)->SetMaxBufferSize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetAnonIPLimit(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetAnonIPLimit", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetAnonIPLimit" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_SetAnonIPLimit" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  (arg1)->SetAnonIPLimit(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetServerThrottle(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetServerThrottle", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetServerThrottle" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_SetServerThrottle" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  (arg1)->SetServerThrottle(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetProtectWebSessions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetProtectWebSessions", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetProtectWebSessions" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_SetProtectWebSessions" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetProtectWebSessions(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetHideVersion(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetHideVersion", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetHideVersion" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_SetHideVersion" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetHideVersion(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetAuthOnlyViaModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetAuthOnlyViaModule", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetAuthOnlyViaModule" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_SetAuthOnlyViaModule" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  (arg1)->SetAuthOnlyViaModule(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetConnectDelay(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetConnectDelay", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetConnectDelay" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_SetConnectDelay" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  (arg1)->SetConnectDelay(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetSSLCiphers(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetSSLCiphers", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetSSLCiphers" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_SetSSLCiphers" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_SetSSLCiphers" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetSSLCiphers((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetSSLProtocols(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetSSLProtocols", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetSSLProtocols" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_SetSSLProtocols" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_SetSSLProtocols" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->SetSSLProtocols((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetSSLCertFile(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetSSLCertFile", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetSSLCertFile" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_SetSSLCertFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_SetSSLCertFile" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetSSLCertFile((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_SetConfigWriteDelay(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetConfigWriteDelay", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetConfigWriteDelay" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_SetConfigWriteDelay" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  (arg1)->SetConfigWriteDelay(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetConfigState(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  enum CZNC::ConfigState result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetConfigState" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (enum CZNC::ConfigState)(arg1)->GetConfigState();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetManager__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CSockManager *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetManager" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (CSockManager *) &(arg1)->GetManager();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSockManager, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetManager__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CSockManager *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetManager" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (CSockManager *) &((CZNC const *)arg1)->GetManager();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CSockManager, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetManager(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CZNC_GetManager", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CZNC_GetManager__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CZNC_GetManager__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CZNC_GetManager'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CZNC::GetManager()\n"
    "    CZNC::GetManager() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CZNC_GetModules(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CModules *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetModules" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (CModules *) &(arg1)->GetModules();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModules, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetSkinName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetSkinName" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = ((CZNC const *)arg1)->GetSkinName();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetStatusPrefix(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetStatusPrefix" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (CString *) &((CZNC const *)arg1)->GetStatusPrefix();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetCurPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetCurPath" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (CString *) &((CZNC const *)arg1)->GetCurPath();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetHomePath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetHomePath" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (CString *) &((CZNC const *)arg1)->GetHomePath();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetZNCPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetZNCPath" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (CString *) &((CZNC const *)arg1)->GetZNCPath();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetConfPath__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetConfPath" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_GetConfPath" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = ((CZNC const *)arg1)->GetConfPath(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetConfPath__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetConfPath" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = ((CZNC const *)arg1)->GetConfPath();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetConfPath(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CZNC_GetConfPath", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CZNC_GetConfPath__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CZNC_GetConfPath__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CZNC_GetConfPath'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CZNC::GetConfPath(bool) const\n"
    "    CZNC::GetConfPath() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CZNC_GetUserPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetUserPath" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = ((CZNC const *)arg1)->GetUserPath();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetModPath(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetModPath" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = ((CZNC const *)arg1)->GetModPath();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetPemLocation(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetPemLocation" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = ((CZNC const *)arg1)->GetPemLocation();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetKeyLocation(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetKeyLocation" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = ((CZNC const *)arg1)->GetKeyLocation();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetDHParamLocation(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetDHParamLocation" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = ((CZNC const *)arg1)->GetDHParamLocation();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetConfigFile(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetConfigFile" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (CString *) &((CZNC const *)arg1)->GetConfigFile();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_WritePemFile(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_WritePemFile" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (bool)(arg1)->WritePemFile();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetBindHosts(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  VCString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetBindHosts" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (VCString *) &((CZNC const *)arg1)->GetBindHosts();
  resultobj = swig::from(static_cast< std::vector< CString,std::allocator< CString > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetTrustedProxies(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  VCString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetTrustedProxies" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (VCString *) &((CZNC const *)arg1)->GetTrustedProxies();
  resultobj = swig::from(static_cast< std::vector< CString,std::allocator< CString > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetListeners(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< CListener *,std::allocator< CListener * > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetListeners" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (std::vector< CListener *,std::allocator< CListener * > > *) &((CZNC const *)arg1)->GetListeners();
  resultobj = swig::from(static_cast< std::vector< CListener*,std::allocator< CListener * > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_TimeStarted(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  time_t result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_TimeStarted" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = ((CZNC const *)arg1)->TimeStarted();
  resultobj = SWIG_NewPointerObj((new time_t(result)), SWIGTYPE_p_time_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetMaxBufferSize(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetMaxBufferSize" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (unsigned int)((CZNC const *)arg1)->GetMaxBufferSize();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetAnonIPLimit(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetAnonIPLimit" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (unsigned int)((CZNC const *)arg1)->GetAnonIPLimit();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetServerThrottle__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetServerThrottle" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (unsigned int)((CZNC const *)arg1)->GetServerThrottle();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetConnectDelay(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetConnectDelay" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (unsigned int)((CZNC const *)arg1)->GetConnectDelay();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetProtectWebSessions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetProtectWebSessions" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (bool)((CZNC const *)arg1)->GetProtectWebSessions();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetHideVersion(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetHideVersion" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (bool)((CZNC const *)arg1)->GetHideVersion();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetAuthOnlyViaModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetAuthOnlyViaModule" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (bool)((CZNC const *)arg1)->GetAuthOnlyViaModule();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetSSLCiphers(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetSSLCiphers" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = ((CZNC const *)arg1)->GetSSLCiphers();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetSSLProtocols(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetSSLProtocols" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = ((CZNC const *)arg1)->GetSSLProtocols();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetDisabledSSLProtocols(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Csock::EDisableProtocol result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetDisabledSSLProtocols" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (Csock::EDisableProtocol)((CZNC const *)arg1)->GetDisabledSSLProtocols();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetSSLCertFile(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetSSLCertFile" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = ((CZNC const *)arg1)->GetSSLCertFile();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetAvailableSSLProtocols(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  VCString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_GetAvailableSSLProtocols", 0, 0, 0)) SWIG_fail;
  result = CZNC::GetAvailableSSLProtocols();
  resultobj = swig::from(static_cast< std::vector< CString,std::allocator< CString > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetConfigWriteDelay(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetConfigWriteDelay" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (unsigned int)((CZNC const *)arg1)->GetConfigWriteDelay();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_CreateInstance(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_CreateInstance", 0, 0, 0)) SWIG_fail;
  CZNC::CreateInstance();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_Get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_Get", 0, 0, 0)) SWIG_fail;
  result = (CZNC *) &CZNC::Get();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CZNC, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_DestroyInstance(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_DestroyInstance", 0, 0, 0)) SWIG_fail;
  CZNC::DestroyInstance();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_FindUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CUser *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_FindUser", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_FindUser" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_FindUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_FindUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CUser *)(arg1)->FindUser((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CUser, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_FindModule__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  CModule *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_FindModule" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_FindModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_FindModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CZNC_FindModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_FindModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (CModule *)(arg1)->FindModule((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModule, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_FindModule__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  CUser *arg3 = (CUser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  CModule *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_FindModule" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_FindModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_FindModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CZNC_FindModule" "', argument " "3"" of type '" "CUser *""'"); 
  }
  arg3 = reinterpret_cast< CUser * >(argp3);
  result = (CModule *)(arg1)->FindModule((CString const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModule, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_FindModule(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CZNC_FindModule", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_CUser, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CZNC_FindModule__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CZNC_FindModule__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CZNC_FindModule'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CZNC::FindModule(CString const &,CString const &)\n"
    "    CZNC::FindModule(CString const &,CUser *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CZNC_UpdateModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_UpdateModule", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_UpdateModule" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_UpdateModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_UpdateModule" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->UpdateModule((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_DeleteUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_DeleteUser", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_DeleteUser" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_DeleteUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_DeleteUser" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->DeleteUser((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AddUser__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CUser *arg2 = (CUser *) 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AddUser" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_AddUser" "', argument " "2"" of type '" "CUser *""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sErrorRet");
    }
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CZNC_AddUser" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->AddUser(arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AddUser__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CUser *arg2 = (CUser *) 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AddUser" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_AddUser" "', argument " "2"" of type '" "CUser *""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sErrorRet");
    }
  }
  result = (bool)(arg1)->AddUser(arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AddUser(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CZNC_AddUser", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CUser, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          return _wrap_CZNC_AddUser__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CUser, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          String* p;
          _v = SWIG_IsOK(SWIG_ConvertPtr(argv[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
        }
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_CZNC_AddUser__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CZNC_AddUser'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CZNC::AddUser(CUser *,CString &,bool)\n"
    "    CZNC::AddUser(CUser *,CString &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CZNC_GetUserMap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::map< CString,CUser *,std::less< CString >,std::allocator< std::pair< CString const,CUser * > > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetUserMap" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (std::map< CString,CUser *,std::less< CString >,std::allocator< std::pair< CString const,CUser * > > > *) &((CZNC const *)arg1)->GetUserMap();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_CString_CUser_p_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CUser_p_t_t_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_FindListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  EAddrType arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  CListener *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_FindListener", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_FindListener" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_FindListener" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CZNC_FindListener" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_FindListener" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CZNC_FindListener" "', argument " "4"" of type '" "EAddrType""'");
  } 
  arg4 = static_cast< EAddrType >(val4);
  result = (CListener *)(arg1)->FindListener(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CListener, 0 |  0 );
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_FindUnixListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CListener *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_FindUnixListener", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_FindUnixListener" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_FindUnixListener" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_FindUnixListener" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CListener *)(arg1)->FindUnixListener((CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CListener, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AddListener__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CListener *arg2 = (CListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AddListener" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_AddListener" "', argument " "2"" of type '" "CListener *""'"); 
  }
  arg2 = reinterpret_cast< CListener * >(argp2);
  result = (bool)(arg1)->AddListener(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AddTCPListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  bool arg5 ;
  EAddrType arg6 ;
  CListener::EAcceptType arg7 ;
  CString *arg8 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  bool val5 ;
  int ecode5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  int val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[8] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_AddTCPListener", 8, 8, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AddTCPListener" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_AddTCPListener" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CZNC_AddTCPListener" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_AddTCPListener" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CZNC_AddTCPListener" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_AddTCPListener" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CZNC_AddTCPListener" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CZNC_AddTCPListener" "', argument " "6"" of type '" "EAddrType""'");
  } 
  arg6 = static_cast< EAddrType >(val6);
  ecode7 = SWIG_AsVal_int(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CZNC_AddTCPListener" "', argument " "7"" of type '" "CListener::EAcceptType""'");
  } 
  arg7 = static_cast< CListener::EAcceptType >(val7);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[7], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg8 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 8 sError");
    }
  }
  result = (bool)(arg1)->AddTCPListener(arg2,(CString const &)*arg3,(CString const &)*arg4,arg5,arg6,arg7,*arg8);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AddUnixListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  CListener::EAcceptType arg5 ;
  CString *arg6 = 0 ;
  CString *arg7 = 0 ;
  CString *arg8 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  int res6 = SWIG_OLDOBJ ;
  int res7 = SWIG_OLDOBJ ;
  PyObject *swig_obj[8] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_AddUnixListener", 8, 8, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AddUnixListener" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_AddUnixListener" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_AddUnixListener" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CZNC_AddUnixListener" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_AddUnixListener" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CZNC_AddUnixListener" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CZNC_AddUnixListener" "', argument " "5"" of type '" "CListener::EAcceptType""'");
  } 
  arg5 = static_cast< CListener::EAcceptType >(val5);
  {
    CString *ptr = (CString *)0;
    res6 = SWIG_AsPtr_CString(swig_obj[5], &ptr);
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "CZNC_AddUnixListener" "', argument " "6"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_AddUnixListener" "', argument " "6"" of type '" "CString const &""'"); 
    }
    arg6 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res7 = SWIG_AsPtr_CString(swig_obj[6], &ptr);
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "CZNC_AddUnixListener" "', argument " "7"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_AddUnixListener" "', argument " "7"" of type '" "CString const &""'"); 
    }
    arg7 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[7], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg8 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 8 sError");
    }
  }
  result = (bool)(arg1)->AddUnixListener((CString const &)*arg2,(CString const &)*arg3,arg4,arg5,(CString const &)*arg6,(CString const &)*arg7,*arg8);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res6)) delete arg6;
  if (SWIG_IsNewObj(res7)) delete arg7;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res6)) delete arg6;
  if (SWIG_IsNewObj(res7)) delete arg7;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_DelListener(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CListener *arg2 = (CListener *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_DelListener", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_DelListener" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CListener, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_DelListener" "', argument " "2"" of type '" "CListener *""'"); 
  }
  arg2 = reinterpret_cast< CListener * >(argp2);
  result = (bool)(arg1)->DelListener(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AddListener__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  bool arg5 ;
  EAddrType arg6 ;
  CListener::EAcceptType arg7 ;
  CString *arg8 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  bool val5 ;
  int ecode5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  int val7 ;
  int ecode7 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 8) || (nobjs > 8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AddListener" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CZNC_AddListener" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CZNC_AddListener" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_AddListener" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CZNC_AddListener" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_AddListener" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CZNC_AddListener" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CZNC_AddListener" "', argument " "6"" of type '" "EAddrType""'");
  } 
  arg6 = static_cast< EAddrType >(val6);
  ecode7 = SWIG_AsVal_int(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CZNC_AddListener" "', argument " "7"" of type '" "CListener::EAcceptType""'");
  } 
  arg7 = static_cast< CListener::EAcceptType >(val7);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[7], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg8 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 8 sError");
    }
  }
  result = (bool)(arg1)->AddListener(arg2,(CString const &)*arg3,(CString const &)*arg4,arg5,arg6,arg7,*arg8);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AddListener(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[9] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CZNC_AddListener", 0, 8, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_CListener, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CZNC_AddListener__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 8) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            {
              int res = SWIG_AsVal_bool(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              {
                int res = SWIG_AsVal_int(argv[5], NULL);
                _v = SWIG_CheckState(res);
              }
              if (_v) {
                {
                  int res = SWIG_AsVal_int(argv[6], NULL);
                  _v = SWIG_CheckState(res);
                }
                if (_v) {
                  {
                    String* p;
                    _v = SWIG_IsOK(SWIG_ConvertPtr(argv[7], (void**)&p, SWIG_TypeQuery("String*"), 0));
                  }
                  if (_v) {
                    return _wrap_CZNC_AddListener__SWIG_1(self, argc, argv);
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CZNC_AddListener'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CZNC::AddListener(CListener *)\n"
    "    CZNC::AddListener(unsigned short,CString const &,CString const &,bool,EAddrType,CListener::EAcceptType,CString &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CZNC_SetMotd(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_SetMotd", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_SetMotd" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_SetMotd" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_SetMotd" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetMotd((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AddMotd(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_AddMotd", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AddMotd" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_AddMotd" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_AddMotd" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->AddMotd((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_ClearMotd(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_ClearMotd" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  (arg1)->ClearMotd();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetMotd(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  VCString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetMotd" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (VCString *) &((CZNC const *)arg1)->GetMotd();
  resultobj = swig::from(static_cast< std::vector< CString,std::allocator< CString > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_AddServerThrottle(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_AddServerThrottle", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AddServerThrottle" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    int res = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "CZNC_AddServerThrottle" "', argument " "2"" of type '" "CString""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  (arg1)->AddServerThrottle(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetServerThrottle__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetServerThrottle" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    int res = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "CZNC_GetServerThrottle" "', argument " "2"" of type '" "CString""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  result = (bool)(arg1)->GetServerThrottle(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetServerThrottle(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CZNC_GetServerThrottle", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CZNC_GetServerThrottle__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CZNC, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CZNC_GetServerThrottle__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CZNC_GetServerThrottle'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CZNC::GetServerThrottle() const\n"
    "    CZNC::GetServerThrottle(CString)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CZNC_AddNetworkToQueue(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_AddNetworkToQueue", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_AddNetworkToQueue" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_AddNetworkToQueue" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  (arg1)->AddNetworkToQueue(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetConnectionQueue(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::list< CIRCNetwork *,std::allocator< CIRCNetwork * > > *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetConnectionQueue" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (std::list< CIRCNetwork *,std::allocator< CIRCNetwork * > > *) &(arg1)->GetConnectionQueue();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__listT_CIRCNetwork_p_std__allocatorT_CIRCNetwork_p_t_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_EnableConnectQueue(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_EnableConnectQueue" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  (arg1)->EnableConnectQueue();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_DisableConnectQueue(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_DisableConnectQueue" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  (arg1)->DisableConnectQueue();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_PauseConnectQueue(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_PauseConnectQueue" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  (arg1)->PauseConnectQueue();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_ResumeConnectQueue(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_ResumeConnectQueue" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  (arg1)->ResumeConnectQueue();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_ForceEncoding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_ForceEncoding" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  (arg1)->ForceEncoding();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_UnforceEncoding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_UnforceEncoding" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  (arg1)->UnforceEncoding();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_IsForcingEncoding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_IsForcingEncoding" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (bool)((CZNC const *)arg1)->IsForcingEncoding();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_FixupEncoding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_FixupEncoding", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_FixupEncoding" "', argument " "1"" of type '" "CZNC const *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_FixupEncoding" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CZNC_FixupEncoding" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = ((CZNC const *)arg1)->FixupEncoding((CString const &)*arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_LeakConnectQueueTimer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  CConnectQueueTimer *arg2 = (CConnectQueueTimer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CZNC_LeakConnectQueueTimer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_LeakConnectQueueTimer" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CConnectQueueTimer, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CZNC_LeakConnectQueueTimer" "', argument " "2"" of type '" "CConnectQueueTimer *""'"); 
  }
  arg2 = reinterpret_cast< CConnectQueueTimer * >(argp2);
  (arg1)->LeakConnectQueueTimer(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_DisableConfigTimer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_DisableConfigTimer" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  (arg1)->DisableConfigTimer();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_DumpConfig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CConfig *arg1 = (CConfig *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CConfig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_DumpConfig" "', argument " "1"" of type '" "CConfig const *""'"); 
  }
  arg1 = reinterpret_cast< CConfig * >(argp1);
  CZNC::DumpConfig((CConfig const *)arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CZNC_GetUserMap_(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CZNC *arg1 = (CZNC *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CZNC, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CZNC_GetUserMap_" "', argument " "1"" of type '" "CZNC *""'"); 
  }
  arg1 = reinterpret_cast< CZNC * >(argp1);
  result = (PyObject *)CZNC_GetUserMap_(arg1);
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CZNC_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CZNC, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CZNC_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CServer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  bool arg5 ;
  int res1 = SWIG_OLDOBJ ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  CServer *result = 0 ;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CServer" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CServer" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CServer" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CServer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CServer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_CServer" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_CServer" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  result = (CServer *)new CServer((CString const &)*arg1,arg2,(CString const &)*arg3,arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CServer, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CServer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  int res1 = SWIG_OLDOBJ ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  CServer *result = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CServer" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CServer" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CServer" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CServer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CServer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_CServer" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (CServer *)new CServer((CString const &)*arg1,arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CServer, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CServer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  unsigned short arg2 ;
  CString *arg3 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  CServer *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CServer" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CServer" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CServer" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CServer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CServer" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (CServer *)new CServer((CString const &)*arg1,arg2,(CString const &)*arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CServer, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CServer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  unsigned short arg2 ;
  int res1 = SWIG_OLDOBJ ;
  unsigned short val2 ;
  int ecode2 = 0 ;
  CServer *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CServer" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CServer" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CServer" "', argument " "2"" of type '" "unsigned short""'");
  } 
  arg2 = static_cast< unsigned short >(val2);
  result = (CServer *)new CServer((CString const &)*arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CServer, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CServer__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  CServer *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CServer" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CServer" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (CServer *)new CServer((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CServer, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CServer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CServer", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CServer__SWIG_4(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_new_CServer__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_new_CServer__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_new_CServer__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_short(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_bool(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_new_CServer__SWIG_0(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CServer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CServer::CServer(CString const &,unsigned short,CString const &,bool,bool)\n"
    "    CServer::CServer(CString const &,unsigned short,CString const &,bool)\n"
    "    CServer::CServer(CString const &,unsigned short,CString const &)\n"
    "    CServer::CServer(CString const &,unsigned short)\n"
    "    CServer::CServer(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CServer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CServer *arg1 = (CServer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CServer, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CServer" "', argument " "1"" of type '" "CServer *""'"); 
  }
  arg1 = reinterpret_cast< CServer * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CServer___eq__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CServer *arg1 = (CServer *) 0 ;
  CServer *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CServer___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CServer___eq__" "', argument " "1"" of type '" "CServer const *""'"); 
  }
  arg1 = reinterpret_cast< CServer * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CServer,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CServer___eq__" "', argument " "2"" of type '" "CServer const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CServer___eq__" "', argument " "2"" of type '" "CServer const &""'"); 
  }
  arg2 = reinterpret_cast< CServer * >(argp2);
  result = (bool)((CServer const *)arg1)->operator ==((CServer const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_CServer___lt__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CServer *arg1 = (CServer *) 0 ;
  CServer *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CServer___lt__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CServer___lt__" "', argument " "1"" of type '" "CServer const *""'"); 
  }
  arg1 = reinterpret_cast< CServer * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CServer,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CServer___lt__" "', argument " "2"" of type '" "CServer const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CServer___lt__" "', argument " "2"" of type '" "CServer const &""'"); 
  }
  arg2 = reinterpret_cast< CServer * >(argp2);
  result = (bool)((CServer const *)arg1)->operator <((CServer const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) {
    return NULL;
  }
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_CServer_Parse(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString arg1 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< CServer > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    int res = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "CServer_Parse" "', argument " "1"" of type '" "CString""'"); 
    }
    arg1 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  result = CServer::Parse(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj((new CServer(result)), SWIGTYPE_p_CServer, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CServer_GetName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CServer *arg1 = (CServer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CServer_GetName" "', argument " "1"" of type '" "CServer const *""'"); 
  }
  arg1 = reinterpret_cast< CServer * >(argp1);
  result = (CString *) &((CServer const *)arg1)->GetName();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CServer_GetPort(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CServer *arg1 = (CServer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned short result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CServer_GetPort" "', argument " "1"" of type '" "CServer const *""'"); 
  }
  arg1 = reinterpret_cast< CServer * >(argp1);
  result = (unsigned short)((CServer const *)arg1)->GetPort();
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CServer_GetPass(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CServer *arg1 = (CServer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CServer_GetPass" "', argument " "1"" of type '" "CServer const *""'"); 
  }
  arg1 = reinterpret_cast< CServer * >(argp1);
  result = (CString *) &((CServer const *)arg1)->GetPass();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CServer_IsSSL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CServer *arg1 = (CServer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CServer_IsSSL" "', argument " "1"" of type '" "CServer const *""'"); 
  }
  arg1 = reinterpret_cast< CServer * >(argp1);
  result = (bool)((CServer const *)arg1)->IsSSL();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CServer_IsUnixSocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CServer *arg1 = (CServer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CServer_IsUnixSocket" "', argument " "1"" of type '" "CServer const *""'"); 
  }
  arg1 = reinterpret_cast< CServer * >(argp1);
  result = (bool)((CServer const *)arg1)->IsUnixSocket();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CServer_GetString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CServer *arg1 = (CServer *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CServer_GetString" "', argument " "1"" of type '" "CServer const *""'"); 
  }
  arg1 = reinterpret_cast< CServer * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CServer_GetString" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  result = ((CServer const *)arg1)->GetString(arg2);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CServer_GetString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CServer *arg1 = (CServer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CServer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CServer_GetString" "', argument " "1"" of type '" "CServer const *""'"); 
  }
  arg1 = reinterpret_cast< CServer * >(argp1);
  result = ((CServer const *)arg1)->GetString();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CServer_GetString(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CServer_GetString", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CServer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_CServer_GetString__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CServer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CServer_GetString__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CServer_GetString'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CServer::GetString(bool) const\n"
    "    CServer::GetString() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CServer_IsValidHostName(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CServer_IsValidHostName" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CServer_IsValidHostName" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (bool)CServer::IsValidHostName((CString const &)*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *CServer_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CServer, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CServer_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CDebug_SetStdoutIsTTY(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  bool arg1 ;
  bool val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_bool(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "CDebug_SetStdoutIsTTY" "', argument " "1"" of type '" "bool""'");
  } 
  arg1 = static_cast< bool >(val1);
  CDebug::SetStdoutIsTTY(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDebug_StdoutIsTTY(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CDebug_StdoutIsTTY", 0, 0, 0)) SWIG_fail;
  result = (bool)CDebug::StdoutIsTTY();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDebug_SetDebug(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  bool arg1 ;
  bool val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsVal_bool(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "CDebug_SetDebug" "', argument " "1"" of type '" "bool""'");
  } 
  arg1 = static_cast< bool >(val1);
  CDebug::SetDebug(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDebug_Debug(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CDebug_Debug", 0, 0, 0)) SWIG_fail;
  result = (bool)CDebug::Debug();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CDebug_Filter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CDebug_Filter" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CDebug_Filter" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = CDebug::Filter((CString const &)*arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CDebug(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CDebug *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CDebug", 0, 0, 0)) SWIG_fail;
  result = (CDebug *)new CDebug();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CDebug, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CDebug(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CDebug *arg1 = (CDebug *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CDebug, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CDebug" "', argument " "1"" of type '" "CDebug *""'"); 
  }
  arg1 = reinterpret_cast< CDebug * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CDebug_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CDebug, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CDebug_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_delete_CDebugStream(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CDebugStream *arg1 = (CDebugStream *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CDebugStream, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CDebugStream" "', argument " "1"" of type '" "CDebugStream *""'"); 
  }
  arg1 = reinterpret_cast< CDebugStream * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CDebugStream(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CDebugStream *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CDebugStream", 0, 0, 0)) SWIG_fail;
  result = (CDebugStream *)new CDebugStream();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CDebugStream, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CDebugStream_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CDebugStream, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CDebugStream_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CExecSock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CExecSock *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CExecSock", 0, 0, 0)) SWIG_fail;
  result = (CExecSock *)new CExecSock();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CExecSock, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CExecSock_Execute(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CExecSock *arg1 = (CExecSock *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CExecSock_Execute", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CExecSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CExecSock_Execute" "', argument " "1"" of type '" "CExecSock *""'"); 
  }
  arg1 = reinterpret_cast< CExecSock * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CExecSock_Execute" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CExecSock_Execute" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (int)(arg1)->Execute((CString const &)*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CExecSock_Kill(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CExecSock *arg1 = (CExecSock *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CExecSock_Kill", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CExecSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CExecSock_Kill" "', argument " "1"" of type '" "CExecSock *""'"); 
  }
  arg1 = reinterpret_cast< CExecSock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CExecSock_Kill" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  (arg1)->Kill(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CExecSock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CExecSock *arg1 = (CExecSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CExecSock, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CExecSock" "', argument " "1"" of type '" "CExecSock *""'"); 
  }
  arg1 = reinterpret_cast< CExecSock * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CExecSock_popen2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CExecSock *arg1 = (CExecSock *) 0 ;
  int *arg2 = 0 ;
  int *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CExecSock_popen2", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CExecSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CExecSock_popen2" "', argument " "1"" of type '" "CExecSock *""'"); 
  }
  arg1 = reinterpret_cast< CExecSock * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_int,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CExecSock_popen2" "', argument " "2"" of type '" "int &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CExecSock_popen2" "', argument " "2"" of type '" "int &""'"); 
  }
  arg2 = reinterpret_cast< int * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_int,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CExecSock_popen2" "', argument " "3"" of type '" "int &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CExecSock_popen2" "', argument " "3"" of type '" "int &""'"); 
  }
  arg3 = reinterpret_cast< int * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CExecSock_popen2" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CExecSock_popen2" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (int)(arg1)->popen2(*arg2,*arg3,(CString const &)*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CExecSock_close2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CExecSock *arg1 = (CExecSock *) 0 ;
  int arg2 ;
  int arg3 ;
  int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CExecSock_close2", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CExecSock, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CExecSock_close2" "', argument " "1"" of type '" "CExecSock *""'"); 
  }
  arg1 = reinterpret_cast< CExecSock * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CExecSock_close2" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CExecSock_close2" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CExecSock_close2" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = static_cast< int >(val4);
  (arg1)->close2(arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CExecSock_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CExecSock, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CExecSock_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CBufLine__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CBufLine *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CBufLine *)new CBufLine();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBufLine, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CBufLine__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  CBufLine *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CBufLine" "', argument " "1"" of type '" "CMessage const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CBufLine" "', argument " "1"" of type '" "CMessage const &""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CBufLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CBufLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CBufLine *)new CBufLine((CMessage const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBufLine, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CBufLine__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CMessage *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  CBufLine *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CBufLine" "', argument " "1"" of type '" "CMessage const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CBufLine" "', argument " "1"" of type '" "CMessage const &""'"); 
  }
  arg1 = reinterpret_cast< CMessage * >(argp1);
  result = (CBufLine *)new CBufLine((CMessage const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBufLine, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CBufLine__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  timeval *arg3 = (timeval *) 0 ;
  MCString *arg4 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  CBufLine *result = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CBufLine" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CBufLine" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CBufLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CBufLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_timeval, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CBufLine" "', argument " "3"" of type '" "timeval const *""'"); 
  }
  arg3 = reinterpret_cast< timeval * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_MCString,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CBufLine" "', argument " "4"" of type '" "MCString const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CBufLine" "', argument " "4"" of type '" "MCString const &""'"); 
  }
  arg4 = reinterpret_cast< MCString * >(argp4);
  result = (CBufLine *)new CBufLine((CString const &)*arg1,(CString const &)*arg2,(timeval const *)arg3,(MCString const &)*arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBufLine, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CBufLine__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  timeval *arg3 = (timeval *) 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  CBufLine *result = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CBufLine" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CBufLine" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CBufLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CBufLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_timeval, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CBufLine" "', argument " "3"" of type '" "timeval const *""'"); 
  }
  arg3 = reinterpret_cast< timeval * >(argp3);
  result = (CBufLine *)new CBufLine((CString const &)*arg1,(CString const &)*arg2,(timeval const *)arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBufLine, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CBufLine__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  CBufLine *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CBufLine" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CBufLine" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CBufLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CBufLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (CBufLine *)new CBufLine((CString const &)*arg1,(CString const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBufLine, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CBufLine__SWIG_6(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  CBufLine *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CBufLine" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CBufLine" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (CBufLine *)new CBufLine((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBufLine, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CBufLine(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CBufLine", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CBufLine__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CBufLine__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_CBufLine__SWIG_6(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_CBufLine__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_CBufLine__SWIG_5(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_timeval, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_new_CBufLine__SWIG_4(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_timeval, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_ConvertPtr(argv[3], 0, SWIGTYPE_p_MCString, SWIG_POINTER_NO_NULL | 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_new_CBufLine__SWIG_3(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CBufLine'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CBufLine::CBufLine()\n"
    "    CBufLine::CBufLine(CMessage const &,CString const &)\n"
    "    CBufLine::CBufLine(CMessage const &)\n"
    "    CBufLine::CBufLine(CString const &,CString const &,timeval const *,MCString const &)\n"
    "    CBufLine::CBufLine(CString const &,CString const &,timeval const *)\n"
    "    CBufLine::CBufLine(CString const &,CString const &)\n"
    "    CBufLine::CBufLine(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CBufLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CBufLine" "', argument " "1"" of type '" "CBufLine *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBufLine_ToMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  CClient *arg2 = 0 ;
  MCString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  CMessage result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CBufLine_ToMessage", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBufLine_ToMessage" "', argument " "1"" of type '" "CBufLine const *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CClient,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBufLine_ToMessage" "', argument " "2"" of type '" "CClient const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBufLine_ToMessage" "', argument " "2"" of type '" "CClient const &""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_MCString,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBufLine_ToMessage" "', argument " "3"" of type '" "MCString const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBufLine_ToMessage" "', argument " "3"" of type '" "MCString const &""'"); 
  }
  arg3 = reinterpret_cast< MCString * >(argp3);
  result = ((CBufLine const *)arg1)->ToMessage((CClient const &)*arg2,(MCString const &)*arg3);
  resultobj = SWIG_NewPointerObj((new CMessage(result)), SWIGTYPE_p_CMessage, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBufLine_GetLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  CClient *arg2 = 0 ;
  MCString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CBufLine_GetLine", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBufLine_GetLine" "', argument " "1"" of type '" "CBufLine const *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CClient,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBufLine_GetLine" "', argument " "2"" of type '" "CClient const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBufLine_GetLine" "', argument " "2"" of type '" "CClient const &""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_MCString,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBufLine_GetLine" "', argument " "3"" of type '" "MCString const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBufLine_GetLine" "', argument " "3"" of type '" "MCString const &""'"); 
  }
  arg3 = reinterpret_cast< MCString * >(argp3);
  result = ((CBufLine const *)arg1)->GetLine((CClient const &)*arg2,(MCString const &)*arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBufLine_UpdateTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBufLine_UpdateTime" "', argument " "1"" of type '" "CBufLine *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  (arg1)->UpdateTime();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBufLine_Equals(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CBufLine_Equals", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBufLine_Equals" "', argument " "1"" of type '" "CBufLine const *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBufLine_Equals" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBufLine_Equals" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (bool)((CBufLine const *)arg1)->Equals((CMessage const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBufLine_SetFormat(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CBufLine_SetFormat", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBufLine_SetFormat" "', argument " "1"" of type '" "CBufLine *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBufLine_SetFormat" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBufLine_SetFormat" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetFormat((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBufLine_SetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CBufLine_SetText", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBufLine_SetText" "', argument " "1"" of type '" "CBufLine *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBufLine_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBufLine_SetText" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->SetText((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBufLine_SetTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  timeval *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CBufLine_SetTime", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBufLine_SetTime" "', argument " "1"" of type '" "CBufLine *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_timeval,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBufLine_SetTime" "', argument " "2"" of type '" "timeval const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBufLine_SetTime" "', argument " "2"" of type '" "timeval const &""'"); 
  }
  arg2 = reinterpret_cast< timeval * >(argp2);
  (arg1)->SetTime((timeval const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBufLine_SetTags(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  MCString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CBufLine_SetTags", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBufLine_SetTags" "', argument " "1"" of type '" "CBufLine *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_MCString,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBufLine_SetTags" "', argument " "2"" of type '" "MCString const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBufLine_SetTags" "', argument " "2"" of type '" "MCString const &""'"); 
  }
  arg2 = reinterpret_cast< MCString * >(argp2);
  (arg1)->SetTags((MCString const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBufLine_GetCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBufLine_GetCommand" "', argument " "1"" of type '" "CBufLine const *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  result = (CString *) &((CBufLine const *)arg1)->GetCommand();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBufLine_GetFormat(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBufLine_GetFormat" "', argument " "1"" of type '" "CBufLine const *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  result = ((CBufLine const *)arg1)->GetFormat();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBufLine_GetText(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBufLine_GetText" "', argument " "1"" of type '" "CBufLine const *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  result = (CString *) &((CBufLine const *)arg1)->GetText();
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBufLine_GetTime(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  timeval result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBufLine_GetTime" "', argument " "1"" of type '" "CBufLine const *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  result = ((CBufLine const *)arg1)->GetTime();
  resultobj = SWIG_NewPointerObj((new timeval(result)), SWIGTYPE_p_timeval, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBufLine_GetTags(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBufLine *arg1 = (CBufLine *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  MCString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBufLine, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBufLine_GetTags" "', argument " "1"" of type '" "CBufLine const *""'"); 
  }
  arg1 = reinterpret_cast< CBufLine * >(argp1);
  result = (MCString *) &((CBufLine const *)arg1)->GetTags();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_MCString, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CBufLine_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CBufLine, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CBufLine_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CBuffer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  CBuffer *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_CBuffer" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  result = (CBuffer *)new CBuffer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBuffer, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CBuffer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  CBuffer *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (CBuffer *)new CBuffer();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBuffer, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CBuffer(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_CBuffer", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_CBuffer__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_unsigned_SS_int(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_CBuffer__SWIG_0(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_CBuffer'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CBuffer::CBuffer(unsigned int)\n"
    "    CBuffer::CBuffer()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_delete_CBuffer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CBuffer" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_AddLine__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CMessage *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_AddLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_AddLine" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_AddLine" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBuffer_AddLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_AddLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (arg1)->AddLine((CMessage const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_AddLine__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_AddLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_AddLine" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_AddLine" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (arg1)->AddLine((CMessage const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_UpdateLine__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CString *arg2 = 0 ;
  CMessage *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_UpdateLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_UpdateLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBuffer_UpdateLine" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateLine" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  arg3 = reinterpret_cast< CMessage * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CBuffer_UpdateLine" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateLine" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (arg1)->UpdateLine((CString const &)*arg2,(CMessage const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_UpdateLine__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CString *arg2 = 0 ;
  CMessage *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_UpdateLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_UpdateLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBuffer_UpdateLine" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateLine" "', argument " "3"" of type '" "CMessage const &""'"); 
  }
  arg3 = reinterpret_cast< CMessage * >(argp3);
  result = (arg1)->UpdateLine((CString const &)*arg2,(CMessage const &)*arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_UpdateExactLine__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CMessage *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_UpdateExactLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_UpdateExactLine" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateExactLine" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBuffer_UpdateExactLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateExactLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (arg1)->UpdateExactLine((CMessage const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_UpdateExactLine__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_UpdateExactLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_UpdateExactLine" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateExactLine" "', argument " "2"" of type '" "CMessage const &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (arg1)->UpdateExactLine((CMessage const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_AddLine__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  timeval *arg4 = (timeval *) 0 ;
  MCString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_AddLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_AddLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_AddLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBuffer_AddLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_AddLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_timeval, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CBuffer_AddLine" "', argument " "4"" of type '" "timeval const *""'"); 
  }
  arg4 = reinterpret_cast< timeval * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_MCString,  0  | 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CBuffer_AddLine" "', argument " "5"" of type '" "MCString const &""'"); 
  }
  if (!argp5) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_AddLine" "', argument " "5"" of type '" "MCString const &""'"); 
  }
  arg5 = reinterpret_cast< MCString * >(argp5);
  result = (arg1)->AddLine((CString const &)*arg2,(CString const &)*arg3,(timeval const *)arg4,(MCString const &)*arg5);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_AddLine__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  timeval *arg4 = (timeval *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_AddLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_AddLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_AddLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBuffer_AddLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_AddLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_timeval, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CBuffer_AddLine" "', argument " "4"" of type '" "timeval const *""'"); 
  }
  arg4 = reinterpret_cast< timeval * >(argp4);
  result = (arg1)->AddLine((CString const &)*arg2,(CString const &)*arg3,(timeval const *)arg4);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_AddLine__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_AddLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_AddLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_AddLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBuffer_AddLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_AddLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (arg1)->AddLine((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_AddLine__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_AddLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_AddLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_AddLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->AddLine((CString const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_AddLine(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CBuffer_AddLine", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CBuffer_AddLine__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CBuffer_AddLine__SWIG_5(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CBuffer_AddLine__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CBuffer_AddLine__SWIG_4(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_timeval, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CBuffer_AddLine__SWIG_3(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_timeval, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            int res = SWIG_ConvertPtr(argv[4], 0, SWIGTYPE_p_MCString, SWIG_POINTER_NO_NULL | 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              return _wrap_CBuffer_AddLine__SWIG_2(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CBuffer_AddLine'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CBuffer::AddLine(CMessage const &,CString const &)\n"
    "    CBuffer::AddLine(CMessage const &)\n"
    "    CBuffer::AddLine(CString const &,CString const &,timeval const *,MCString const &)\n"
    "    CBuffer::AddLine(CString const &,CString const &,timeval const *)\n"
    "    CBuffer::AddLine(CString const &,CString const &)\n"
    "    CBuffer::AddLine(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CBuffer_UpdateLine__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_UpdateLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_UpdateLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBuffer_UpdateLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CBuffer_UpdateLine" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateLine" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  result = (arg1)->UpdateLine((CString const &)*arg2,(CString const &)*arg3,(CString const &)*arg4);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_UpdateLine__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_UpdateLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_UpdateLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBuffer_UpdateLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (arg1)->UpdateLine((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_UpdateLine(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CBuffer_UpdateLine", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CBuffer_UpdateLine__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CBuffer_UpdateLine__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CBuffer_UpdateLine__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsPtr_CString(argv[3], (CString**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CBuffer_UpdateLine__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CBuffer_UpdateLine'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CBuffer::UpdateLine(CString const &,CMessage const &,CString const &)\n"
    "    CBuffer::UpdateLine(CString const &,CMessage const &)\n"
    "    CBuffer::UpdateLine(CString const &,CString const &,CString const &)\n"
    "    CBuffer::UpdateLine(CString const &,CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CBuffer_UpdateExactLine__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_UpdateExactLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_UpdateExactLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateExactLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBuffer_UpdateExactLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateExactLine" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (arg1)->UpdateExactLine((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_UpdateExactLine__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_UpdateExactLine" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CBuffer_UpdateExactLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_UpdateExactLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (arg1)->UpdateExactLine((CString const &)*arg2);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_UpdateExactLine(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CBuffer_UpdateExactLine", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CBuffer_UpdateExactLine__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_CBuffer_UpdateExactLine__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_CMessage, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CBuffer_UpdateExactLine__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsPtr_CString(argv[2], (CString**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CBuffer_UpdateExactLine__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CBuffer_UpdateExactLine'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CBuffer::UpdateExactLine(CMessage const &,CString const &)\n"
    "    CBuffer::UpdateExactLine(CMessage const &)\n"
    "    CBuffer::UpdateExactLine(CString const &,CString const &)\n"
    "    CBuffer::UpdateExactLine(CString const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CBuffer_GetBufLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  CBufLine *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CBuffer_GetBufLine", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_GetBufLine" "', argument " "1"" of type '" "CBuffer const *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CBuffer_GetBufLine" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = (CBufLine *) &((CBuffer const *)arg1)->GetBufLine(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CBufLine, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_GetLine__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  std::deque< CBufLine >::size_type arg2 ;
  CClient *arg3 = 0 ;
  MCString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_GetLine" "', argument " "1"" of type '" "CBuffer const *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CBuffer_GetLine" "', argument " "2"" of type '" "std::deque< CBufLine >::size_type""'");
  } 
  arg2 = static_cast< std::deque< CBufLine >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBuffer_GetLine" "', argument " "3"" of type '" "CClient const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_GetLine" "', argument " "3"" of type '" "CClient const &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_MCString,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CBuffer_GetLine" "', argument " "4"" of type '" "MCString const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_GetLine" "', argument " "4"" of type '" "MCString const &""'"); 
  }
  arg4 = reinterpret_cast< MCString * >(argp4);
  result = ((CBuffer const *)arg1)->GetLine(arg2,(CClient const &)*arg3,(MCString const &)*arg4);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_GetLine__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  std::deque< CBufLine >::size_type arg2 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  CString result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_GetLine" "', argument " "1"" of type '" "CBuffer const *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CBuffer_GetLine" "', argument " "2"" of type '" "std::deque< CBufLine >::size_type""'");
  } 
  arg2 = static_cast< std::deque< CBufLine >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CBuffer_GetLine" "', argument " "3"" of type '" "CClient const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CBuffer_GetLine" "', argument " "3"" of type '" "CClient const &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = ((CBuffer const *)arg1)->GetLine(arg2,(CClient const &)*arg3);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_GetLine(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CBuffer_GetLine", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CClient, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_CBuffer_GetLine__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_CClient, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_ConvertPtr(argv[3], 0, SWIGTYPE_p_MCString, SWIG_POINTER_NO_NULL | 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_CBuffer_GetLine__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CBuffer_GetLine'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CBuffer::GetLine(std::deque< CBufLine >::size_type,CClient const &,MCString const &) const\n"
    "    CBuffer::GetLine(std::deque< CBufLine >::size_type,CClient const &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CBuffer_Size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::deque< CBufLine >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_Size" "', argument " "1"" of type '" "CBuffer const *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  result = ((CBuffer const *)arg1)->Size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_IsEmpty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_IsEmpty" "', argument " "1"" of type '" "CBuffer const *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  result = (bool)((CBuffer const *)arg1)->IsEmpty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_Clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_Clear" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  (arg1)->Clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_SetLineCount__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  unsigned int arg2 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_SetLineCount" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CBuffer_SetLineCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CBuffer_SetLineCount" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  result = (bool)(arg1)->SetLineCount(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_SetLineCount__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_SetLineCount" "', argument " "1"" of type '" "CBuffer *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CBuffer_SetLineCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  result = (bool)(arg1)->SetLineCount(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CBuffer_SetLineCount(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "CBuffer_SetLineCount", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_CBuffer_SetLineCount__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_CBuffer, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_CBuffer_SetLineCount__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CBuffer_SetLineCount'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    CBuffer::SetLineCount(unsigned int,bool)\n"
    "    CBuffer::SetLineCount(unsigned int)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_CBuffer_GetLineCount(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CBuffer *arg1 = (CBuffer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CBuffer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CBuffer_GetLineCount" "', argument " "1"" of type '" "CBuffer const *""'"); 
  }
  arg1 = reinterpret_cast< CBuffer * >(argp1);
  result = (unsigned int)((CBuffer const *)arg1)->GetLineCount();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CBuffer_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CBuffer, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CBuffer_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_String_s_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  String *arg1 = (String *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "String_s_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_String, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "String_s_set" "', argument " "1"" of type '" "String *""'"); 
  }
  arg1 = reinterpret_cast< String * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "String_s_set" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "String_s_set" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  if (arg1) (arg1)->s = *arg2;
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_String_s_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  String *arg1 = (String *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_String, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "String_s_get" "', argument " "1"" of type '" "String *""'"); 
  }
  arg1 = reinterpret_cast< String * >(argp1);
  result = (CString *) & ((arg1)->s);
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_String__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  String *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_String" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_String" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  result = (String *)new String((CString const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_String, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_String__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  String *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (String *)new String();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_String, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_String(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_String", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_String__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_String__SWIG_0(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_String'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    String::String(CString const &)\n"
    "    String::String()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_String___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  String *arg1 = (String *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_String, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "String___str__" "', argument " "1"" of type '" "String *""'"); 
  }
  arg1 = reinterpret_cast< String * >(argp1);
  result = String___str__(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_String(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  String *arg1 = (String *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_String, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_String" "', argument " "1"" of type '" "String *""'"); 
  }
  arg1 = reinterpret_cast< String * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *String_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_String, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *String_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CPyModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  CModInfo::EModuleType arg5 ;
  PyObject *arg6 = (PyObject *) 0 ;
  CModPython *arg7 = (CModPython *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  int val5 ;
  int ecode5 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject *swig_obj[7] ;
  CPyModule *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CPyModule", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CPyModule" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CPyModule" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CPyModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CPyModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CPyModule" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CPyModule" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_CPyModule" "', argument " "5"" of type '" "CModInfo::EModuleType""'");
  } 
  arg5 = static_cast< CModInfo::EModuleType >(val5);
  arg6 = swig_obj[5];
  res7 = SWIG_ConvertPtr(swig_obj[6], &argp7,SWIGTYPE_p_CModPython, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "new_CPyModule" "', argument " "7"" of type '" "CModPython *""'"); 
  }
  arg7 = reinterpret_cast< CModPython * >(argp7);
  result = (CPyModule *)new CPyModule(arg1,arg2,(CString const &)*arg3,(CString const &)*arg4,arg5,arg6,arg7);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CPyModule, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_GetPyObj(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_GetPyObj" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  result = (PyObject *)(arg1)->GetPyObj();
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_GetNewPyObj(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_GetNewPyObj" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  result = (PyObject *)(arg1)->GetNewPyObj();
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_DeletePyModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_DeletePyModule" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  (arg1)->DeletePyModule();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_GetPyExceptionStr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_GetPyExceptionStr" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  result = (arg1)->GetPyExceptionStr();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_GetModPython(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CModPython *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_GetModPython" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  result = (CModPython *)(arg1)->GetModPython();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModPython, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnBoot(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnBoot" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  result = (bool)(arg1)->OnBoot();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_WebRequiresLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_WebRequiresLogin" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  result = (bool)(arg1)->WebRequiresLogin();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_WebRequiresAdmin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_WebRequiresAdmin" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  result = (bool)(arg1)->WebRequiresAdmin();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_GetWebMenuTitle(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_GetWebMenuTitle" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  result = (arg1)->GetWebMenuTitle();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnWebPreRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CWebSock *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnWebPreRequest", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnWebPreRequest" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CWebSock,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnWebPreRequest" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnWebPreRequest" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  arg2 = reinterpret_cast< CWebSock * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnWebPreRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnWebPreRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->OnWebPreRequest(*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnWebRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CWebSock *arg2 = 0 ;
  CString *arg3 = 0 ;
  CTemplate *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnWebRequest", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnWebRequest" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CWebSock,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnWebRequest" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnWebRequest" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  arg2 = reinterpret_cast< CWebSock * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnWebRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnWebRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CTemplate,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CPyModule_OnWebRequest" "', argument " "4"" of type '" "CTemplate &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnWebRequest" "', argument " "4"" of type '" "CTemplate &""'"); 
  }
  arg4 = reinterpret_cast< CTemplate * >(argp4);
  result = (bool)(arg1)->OnWebRequest(*arg2,(CString const &)*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_ValidateWebRequestCSRFCheck(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CWebSock *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_ValidateWebRequestCSRFCheck", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_ValidateWebRequestCSRFCheck" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CWebSock,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_ValidateWebRequestCSRFCheck" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_ValidateWebRequestCSRFCheck" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  arg2 = reinterpret_cast< CWebSock * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_ValidateWebRequestCSRFCheck" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_ValidateWebRequestCSRFCheck" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->ValidateWebRequestCSRFCheck(*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_GetSubPages(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  VWebSubPages *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_GetSubPages" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  result = (VWebSubPages *) &(arg1)->GetSubPages();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPreRehash(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPreRehash" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  (arg1)->OnPreRehash();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPostRehash(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPostRehash" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  (arg1)->OnPostRehash();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnIRCDisconnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnIRCDisconnected" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  (arg1)->OnIRCDisconnected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnIRCConnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnIRCConnected" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  (arg1)->OnIRCConnected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnIRCConnecting(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CIRCSock *arg2 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnIRCConnecting", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnIRCConnecting" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnIRCConnecting" "', argument " "2"" of type '" "CIRCSock *""'"); 
  }
  arg2 = reinterpret_cast< CIRCSock * >(argp2);
  result = (CModule::EModRet)(arg1)->OnIRCConnecting(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnIRCConnectionError(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CIRCSock *arg2 = (CIRCSock *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnIRCConnectionError", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnIRCConnectionError" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCSock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnIRCConnectionError" "', argument " "2"" of type '" "CIRCSock *""'"); 
  }
  arg2 = reinterpret_cast< CIRCSock * >(argp2);
  (arg1)->OnIRCConnectionError(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnIRCRegistration(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[5] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnIRCRegistration", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnIRCRegistration" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sPass");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sNick");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sIdent");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[4], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg5 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 5 sRealName");
    }
  }
  result = (CModule::EModRet)(arg1)->OnIRCRegistration(*arg2,*arg3,*arg4,*arg5);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnBroadcast(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnBroadcast", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnBroadcast" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnBroadcast(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanPermission2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  unsigned char arg5 ;
  bool arg6 ;
  bool arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  unsigned char val5 ;
  int ecode5 = 0 ;
  bool val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[7] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanPermission2", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanPermission2" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanPermission2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnChanPermission2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanPermission2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CPyModule_OnChanPermission2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanPermission2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_unsigned_SS_char(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CPyModule_OnChanPermission2" "', argument " "5"" of type '" "unsigned char""'");
  } 
  arg5 = static_cast< unsigned char >(val5);
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CPyModule_OnChanPermission2" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  ecode7 = SWIG_AsVal_bool(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CPyModule_OnChanPermission2" "', argument " "7"" of type '" "bool""'");
  } 
  arg7 = static_cast< bool >(val7);
  (arg1)->OnChanPermission2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5,arg6,arg7);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnOp2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnOp2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnOp2" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnOp2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnOp2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnOp2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CPyModule_OnOp2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnOp2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CPyModule_OnOp2" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  (arg1)->OnOp2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnDeop2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnDeop2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnDeop2" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnDeop2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnDeop2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnDeop2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CPyModule_OnDeop2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnDeop2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CPyModule_OnDeop2" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  (arg1)->OnDeop2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnVoice2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnVoice2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnVoice2" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnVoice2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnVoice2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnVoice2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CPyModule_OnVoice2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnVoice2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CPyModule_OnVoice2" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  (arg1)->OnVoice2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnDevoice2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CNick *arg3 = 0 ;
  CChan *arg4 = 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnDevoice2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnDevoice2" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnDevoice2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnDevoice2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnDevoice2" "', argument " "3"" of type '" "CNick const &""'"); 
  }
  arg3 = reinterpret_cast< CNick * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CPyModule_OnDevoice2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnDevoice2" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CPyModule_OnDevoice2" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  (arg1)->OnDevoice2((CNick const *)arg2,(CNick const &)*arg3,*arg4,arg5);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnMode2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CChan *arg3 = 0 ;
  char arg4 ;
  CString *arg5 = 0 ;
  bool arg6 ;
  bool arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  char val4 ;
  int ecode4 = 0 ;
  int res5 = SWIG_OLDOBJ ;
  bool val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  PyObject *swig_obj[7] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnMode2", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnMode2" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnMode2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnMode2" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnMode2" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  ecode4 = SWIG_AsVal_char(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CPyModule_OnMode2" "', argument " "4"" of type '" "char""'");
  } 
  arg4 = static_cast< char >(val4);
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CPyModule_OnMode2" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnMode2" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CPyModule_OnMode2" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  ecode7 = SWIG_AsVal_bool(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "CPyModule_OnMode2" "', argument " "7"" of type '" "bool""'");
  } 
  arg7 = static_cast< bool >(val7);
  (arg1)->OnMode2((CNick const *)arg2,*arg3,arg4,(CString const &)*arg5,arg6,arg7);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnRawMode2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = (CNick *) 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnRawMode2", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnRawMode2" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CNick, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnRawMode2" "', argument " "2"" of type '" "CNick const *""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnRawMode2" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnRawMode2" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CPyModule_OnRawMode2" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnRawMode2" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CPyModule_OnRawMode2" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnRawMode2" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  (arg1)->OnRawMode2((CNick const *)arg2,*arg3,(CString const &)*arg4,(CString const &)*arg5);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnRaw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnRaw", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnRaw" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  result = (CModule::EModRet)(arg1)->OnRaw(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnStatusCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnStatusCommand", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnStatusCommand" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sCommand");
    }
  }
  result = (CModule::EModRet)(arg1)->OnStatusCommand(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnModCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnModCommand", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnModCommand" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnModCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnModCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->OnModCommand((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnModNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnModNotice", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnModNotice" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnModNotice" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->OnModNotice((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnModCTCP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnModCTCP", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnModCTCP" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnModCTCP" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnModCTCP" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->OnModCTCP((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnQuit(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnQuit", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnQuit" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnQuit" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnQuit" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnQuit" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnQuit" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CPyModule_OnQuit" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnQuit" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->OnQuit((CNick const &)*arg2,(CString const &)*arg3,(std::vector< CChan *,std::allocator< CChan * > > const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnNick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnNick", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnNick" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnNick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnNick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnNick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnNick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CPyModule_OnNick" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnNick" "', argument " "4"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->OnNick((CNick const &)*arg2,(CString const &)*arg3,(std::vector< CChan *,std::allocator< CChan * > > const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnKick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  CChan *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 = SWIG_OLDOBJ ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnKick", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnKick" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnKick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnKick" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnKick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnKick" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CPyModule_OnKick" "', argument " "4"" of type '" "CChan &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnKick" "', argument " "4"" of type '" "CChan &""'"); 
  }
  arg4 = reinterpret_cast< CChan * >(argp4);
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CPyModule_OnKick" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnKick" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  (arg1)->OnKick((CNick const &)*arg2,(CString const &)*arg3,*arg4,(CString const &)*arg5);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnJoining(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CChan *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnJoining", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnJoining" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnJoining" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnJoining" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  result = (CModule::EModRet)(arg1)->OnJoining(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnJoin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnJoin", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnJoin" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnJoin" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnJoin" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnJoin" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnJoin" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  (arg1)->OnJoin((CNick const &)*arg2,*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPart(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnPart", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPart" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnPart" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPart" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnPart" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPart" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CPyModule_OnPart" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPart" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->OnPart((CNick const &)*arg2,*arg3,(CString const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnInvite(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnInvite", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnInvite" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnInvite" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnInvite" "', argument " "2"" of type '" "CNick const &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnInvite" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnInvite" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (CModule::EModRet)(arg1)->OnInvite((CNick const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanBufferStarting(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CChan *arg2 = 0 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanBufferStarting", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanBufferStarting" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanBufferStarting" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanBufferStarting" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnChanBufferStarting" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanBufferStarting" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (CModule::EModRet)(arg1)->OnChanBufferStarting(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanBufferEnding(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CChan *arg2 = 0 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanBufferEnding", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanBufferEnding" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanBufferEnding" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanBufferEnding" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnChanBufferEnding" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanBufferEnding" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (CModule::EModRet)(arg1)->OnChanBufferEnding(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanBufferPlayLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CChan *arg2 = 0 ;
  CClient *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanBufferPlayLine", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanBufferPlayLine" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanBufferPlayLine" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanBufferPlayLine" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnChanBufferPlayLine" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanBufferPlayLine" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sLine");
    }
  }
  result = (CModule::EModRet)(arg1)->OnChanBufferPlayLine(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPrivBufferPlayLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CClient *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnPrivBufferPlayLine", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPrivBufferPlayLine" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnPrivBufferPlayLine" "', argument " "2"" of type '" "CClient &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPrivBufferPlayLine" "', argument " "2"" of type '" "CClient &""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sLine");
    }
  }
  result = (CModule::EModRet)(arg1)->OnPrivBufferPlayLine(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnClientLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnClientLogin" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  (arg1)->OnClientLogin();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnClientDisconnect(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnClientDisconnect" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  (arg1)->OnClientDisconnect();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserRaw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserRaw", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserRaw" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserRaw(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserCTCPReply(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserCTCPReply", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserCTCPReply" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserCTCPReply(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserCTCP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserCTCP", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserCTCP" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserCTCP(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserAction(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserAction", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserAction" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserAction(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserMsg", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserMsg" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserMsg(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserNotice", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserNotice" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sTarget");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserNotice(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserJoin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserJoin", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserJoin" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sChannel");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sKey");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserJoin(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserPart(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserPart", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserPart" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sChannel");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserPart(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserTopic(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserTopic", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserTopic" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sChannel");
    }
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sTopic");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserTopic(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserTopicRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserTopicRequest", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserTopicRequest" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sChannel");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserTopicRequest(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserQuit(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserQuit", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserQuit" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUserQuit(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnCTCPReply(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnCTCPReply", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnCTCPReply" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnCTCPReply" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnCTCPReply" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnCTCPReply(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPrivCTCP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnPrivCTCP", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPrivCTCP" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnPrivCTCP" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPrivCTCP" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnPrivCTCP(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanCTCP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanCTCP", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanCTCP" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanCTCP" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanCTCP" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnChanCTCP" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanCTCP" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnChanCTCP(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPrivAction(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnPrivAction", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPrivAction" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnPrivAction" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPrivAction" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnPrivAction(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanAction(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanAction", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanAction" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanAction" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanAction" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnChanAction" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanAction" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnChanAction(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPrivMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnPrivMsg", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPrivMsg" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnPrivMsg" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPrivMsg" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnPrivMsg(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanMsg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanMsg", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanMsg" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanMsg" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanMsg" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnChanMsg" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanMsg" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnChanMsg(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPrivNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnPrivNotice", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPrivNotice" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnPrivNotice" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPrivNotice" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnPrivNotice(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanNotice(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanNotice", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanNotice" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanNotice" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanNotice" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnChanNotice" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanNotice" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sMessage");
    }
  }
  result = (CModule::EModRet)(arg1)->OnChanNotice(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnTopic(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNick *arg2 = 0 ;
  CChan *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnTopic", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnTopic" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNick,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnTopic" "', argument " "2"" of type '" "CNick &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnTopic" "', argument " "2"" of type '" "CNick &""'"); 
  }
  arg2 = reinterpret_cast< CNick * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnTopic" "', argument " "3"" of type '" "CChan &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnTopic" "', argument " "3"" of type '" "CChan &""'"); 
  }
  arg3 = reinterpret_cast< CChan * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sTopic");
    }
  }
  result = (CModule::EModRet)(arg1)->OnTopic(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnServerCapAvailable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnServerCapAvailable", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnServerCapAvailable" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnServerCapAvailable" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnServerCapAvailable" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  result = (bool)(arg1)->OnServerCapAvailable((CString const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnServerCap302Available(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnServerCap302Available", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnServerCap302Available" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnServerCap302Available" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnServerCap302Available" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnServerCap302Available" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnServerCap302Available" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (bool)(arg1)->OnServerCap302Available((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnServerCapResult(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnServerCapResult", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnServerCapResult" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnServerCapResult" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnServerCapResult" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CPyModule_OnServerCapResult" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->OnServerCapResult((CString const &)*arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnClientAttached(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnClientAttached" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  (arg1)->OnClientAttached();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnClientDetached(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnClientDetached" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  (arg1)->OnClientDetached();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnTimerAutoJoin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CChan *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnTimerAutoJoin", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnTimerAutoJoin" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CChan,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnTimerAutoJoin" "', argument " "2"" of type '" "CChan &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnTimerAutoJoin" "', argument " "2"" of type '" "CChan &""'"); 
  }
  arg2 = reinterpret_cast< CChan * >(argp2);
  result = (CModule::EModRet)(arg1)->OnTimerAutoJoin(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnEmbeddedWebRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CWebSock *arg2 = 0 ;
  CString *arg3 = 0 ;
  CTemplate *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnEmbeddedWebRequest", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnEmbeddedWebRequest" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CWebSock,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnEmbeddedWebRequest" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnEmbeddedWebRequest" "', argument " "2"" of type '" "CWebSock &""'"); 
  }
  arg2 = reinterpret_cast< CWebSock * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnEmbeddedWebRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnEmbeddedWebRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_CTemplate,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CPyModule_OnEmbeddedWebRequest" "', argument " "4"" of type '" "CTemplate &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnEmbeddedWebRequest" "', argument " "4"" of type '" "CTemplate &""'"); 
  }
  arg4 = reinterpret_cast< CTemplate * >(argp4);
  result = (bool)(arg1)->OnEmbeddedWebRequest(*arg2,(CString const &)*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnAddNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CIRCNetwork *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnAddNetwork", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnAddNetwork" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CIRCNetwork,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnAddNetwork" "', argument " "2"" of type '" "CIRCNetwork &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnAddNetwork" "', argument " "2"" of type '" "CIRCNetwork &""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sErrorRet");
    }
  }
  result = (CModule::EModRet)(arg1)->OnAddNetwork(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnDeleteNetwork(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CIRCNetwork *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnDeleteNetwork", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnDeleteNetwork" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CIRCNetwork,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnDeleteNetwork" "', argument " "2"" of type '" "CIRCNetwork &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnDeleteNetwork" "', argument " "2"" of type '" "CIRCNetwork &""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  result = (CModule::EModRet)(arg1)->OnDeleteNetwork(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnSendToClient(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CClient *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnSendToClient", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnSendToClient" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_CClient,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnSendToClient" "', argument " "3"" of type '" "CClient &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnSendToClient" "', argument " "3"" of type '" "CClient &""'"); 
  }
  arg3 = reinterpret_cast< CClient * >(argp3);
  result = (CModule::EModRet)(arg1)->OnSendToClient(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnSendToIRC(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnSendToIRC", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnSendToIRC" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[1], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg2 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 2 sLine");
    }
  }
  result = (CModule::EModRet)(arg1)->OnSendToIRC(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnRawMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnRawMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnRawMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnRawMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnNumericMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNumericMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnNumericMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnNumericMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNumericMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnNumericMessage" "', argument " "2"" of type '" "CNumericMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnNumericMessage" "', argument " "2"" of type '" "CNumericMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNumericMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnNumericMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnQuitMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CQuitMessage *arg2 = 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnQuitMessage", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnQuitMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CQuitMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnQuitMessage" "', argument " "2"" of type '" "CQuitMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnQuitMessage" "', argument " "2"" of type '" "CQuitMessage &""'"); 
  }
  arg2 = reinterpret_cast< CQuitMessage * >(argp2);
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnQuitMessage" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnQuitMessage" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->OnQuitMessage(*arg2,(std::vector< CChan *,std::allocator< CChan * > > const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnNickMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNickMessage *arg2 = 0 ;
  std::vector< CChan *,std::allocator< CChan * > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnNickMessage", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnNickMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNickMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnNickMessage" "', argument " "2"" of type '" "CNickMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnNickMessage" "', argument " "2"" of type '" "CNickMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNickMessage * >(argp2);
  {
    std::vector< CChan*,std::allocator< CChan * > > *ptr = (std::vector< CChan*,std::allocator< CChan * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnNickMessage" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnNickMessage" "', argument " "3"" of type '" "std::vector< CChan *,std::allocator< CChan * > > const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->OnNickMessage(*arg2,(std::vector< CChan *,std::allocator< CChan * > > const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnKickMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CKickMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnKickMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnKickMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CKickMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnKickMessage" "', argument " "2"" of type '" "CKickMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnKickMessage" "', argument " "2"" of type '" "CKickMessage &""'"); 
  }
  arg2 = reinterpret_cast< CKickMessage * >(argp2);
  (arg1)->OnKickMessage(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnJoinMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CJoinMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnJoinMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnJoinMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CJoinMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnJoinMessage" "', argument " "2"" of type '" "CJoinMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnJoinMessage" "', argument " "2"" of type '" "CJoinMessage &""'"); 
  }
  arg2 = reinterpret_cast< CJoinMessage * >(argp2);
  (arg1)->OnJoinMessage(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPartMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CPartMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnPartMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPartMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CPartMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnPartMessage" "', argument " "2"" of type '" "CPartMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPartMessage" "', argument " "2"" of type '" "CPartMessage &""'"); 
  }
  arg2 = reinterpret_cast< CPartMessage * >(argp2);
  (arg1)->OnPartMessage(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanBufferPlayMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanBufferPlayMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanBufferPlayMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanBufferPlayMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanBufferPlayMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnChanBufferPlayMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPrivBufferPlayMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnPrivBufferPlayMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPrivBufferPlayMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnPrivBufferPlayMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPrivBufferPlayMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnPrivBufferPlayMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserRawMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserRawMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserRawMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnUserRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnUserRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserRawMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserCTCPReplyMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserCTCPReplyMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserCTCPReplyMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnUserCTCPReplyMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnUserCTCPReplyMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserCTCPReplyMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserCTCPMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserCTCPMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserCTCPMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnUserCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnUserCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserCTCPMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserActionMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CActionMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserActionMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserActionMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CActionMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnUserActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnUserActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  arg2 = reinterpret_cast< CActionMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserActionMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserTextMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CTextMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserTextMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserTextMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTextMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnUserTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnUserTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTextMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserTextMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserNoticeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNoticeMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserNoticeMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserNoticeMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNoticeMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnUserNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnUserNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNoticeMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserNoticeMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserJoinMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CJoinMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserJoinMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserJoinMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CJoinMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnUserJoinMessage" "', argument " "2"" of type '" "CJoinMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnUserJoinMessage" "', argument " "2"" of type '" "CJoinMessage &""'"); 
  }
  arg2 = reinterpret_cast< CJoinMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserJoinMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserPartMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CPartMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserPartMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserPartMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CPartMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnUserPartMessage" "', argument " "2"" of type '" "CPartMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnUserPartMessage" "', argument " "2"" of type '" "CPartMessage &""'"); 
  }
  arg2 = reinterpret_cast< CPartMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserPartMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserTopicMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CTopicMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserTopicMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserTopicMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTopicMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnUserTopicMessage" "', argument " "2"" of type '" "CTopicMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnUserTopicMessage" "', argument " "2"" of type '" "CTopicMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTopicMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserTopicMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserQuitMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CQuitMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserQuitMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserQuitMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CQuitMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnUserQuitMessage" "', argument " "2"" of type '" "CQuitMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnUserQuitMessage" "', argument " "2"" of type '" "CQuitMessage &""'"); 
  }
  arg2 = reinterpret_cast< CQuitMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserQuitMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnCTCPReplyMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnCTCPReplyMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnCTCPReplyMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnCTCPReplyMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnCTCPReplyMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnCTCPReplyMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPrivCTCPMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnPrivCTCPMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPrivCTCPMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnPrivCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPrivCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnPrivCTCPMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanCTCPMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CCTCPMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanCTCPMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanCTCPMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CCTCPMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanCTCPMessage" "', argument " "2"" of type '" "CCTCPMessage &""'"); 
  }
  arg2 = reinterpret_cast< CCTCPMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnChanCTCPMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPrivActionMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CActionMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnPrivActionMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPrivActionMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CActionMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnPrivActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPrivActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  arg2 = reinterpret_cast< CActionMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnPrivActionMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanActionMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CActionMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanActionMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanActionMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CActionMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanActionMessage" "', argument " "2"" of type '" "CActionMessage &""'"); 
  }
  arg2 = reinterpret_cast< CActionMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnChanActionMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPrivTextMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CTextMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnPrivTextMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPrivTextMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTextMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnPrivTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPrivTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTextMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnPrivTextMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanTextMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CTextMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanTextMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanTextMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTextMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanTextMessage" "', argument " "2"" of type '" "CTextMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTextMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnChanTextMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPrivNoticeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNoticeMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnPrivNoticeMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPrivNoticeMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNoticeMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnPrivNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPrivNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNoticeMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnPrivNoticeMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanNoticeMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CNoticeMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanNoticeMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanNoticeMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CNoticeMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanNoticeMessage" "', argument " "2"" of type '" "CNoticeMessage &""'"); 
  }
  arg2 = reinterpret_cast< CNoticeMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnChanNoticeMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnTopicMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CTopicMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnTopicMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnTopicMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTopicMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnTopicMessage" "', argument " "2"" of type '" "CTopicMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnTopicMessage" "', argument " "2"" of type '" "CTopicMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTopicMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnTopicMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnSendToClientMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnSendToClientMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnSendToClientMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnSendToClientMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnSendToClientMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnSendToClientMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnSendToIRCMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnSendToIRCMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnSendToIRCMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnSendToIRCMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnSendToIRCMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnSendToIRCMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUserTagMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CTargetMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUserTagMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUserTagMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTargetMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnUserTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnUserTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTargetMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUserTagMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnChanTagMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CTargetMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnChanTagMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnChanTagMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTargetMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnChanTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnChanTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTargetMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnChanTagMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnPrivTagMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CTargetMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnPrivTagMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnPrivTagMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CTargetMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnPrivTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnPrivTagMessage" "', argument " "2"" of type '" "CTargetMessage &""'"); 
  }
  arg2 = reinterpret_cast< CTargetMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnPrivTagMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnInviteMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CInviteMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnInviteMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnInviteMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CInviteMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnInviteMessage" "', argument " "2"" of type '" "CInviteMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnInviteMessage" "', argument " "2"" of type '" "CInviteMessage &""'"); 
  }
  arg2 = reinterpret_cast< CInviteMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnInviteMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnAddUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CUser *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnAddUser", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnAddUser" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnAddUser" "', argument " "2"" of type '" "CUser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnAddUser" "', argument " "2"" of type '" "CUser &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sErrorRet");
    }
  }
  result = (CModule::EModRet)(arg1)->OnAddUser(*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnDeleteUser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CUser *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnDeleteUser", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnDeleteUser" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CUser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnDeleteUser" "', argument " "2"" of type '" "CUser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnDeleteUser" "', argument " "2"" of type '" "CUser &""'"); 
  }
  arg2 = reinterpret_cast< CUser * >(argp2);
  result = (CModule::EModRet)(arg1)->OnDeleteUser(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnClientConnect(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CZNCSock *arg2 = (CZNCSock *) 0 ;
  CString *arg3 = 0 ;
  unsigned short arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  unsigned short val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnClientConnect", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnClientConnect" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CZNCSock, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnClientConnect" "', argument " "2"" of type '" "CZNCSock *""'"); 
  }
  arg2 = reinterpret_cast< CZNCSock * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnClientConnect" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnClientConnect" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_unsigned_SS_short(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CPyModule_OnClientConnect" "', argument " "4"" of type '" "unsigned short""'");
  } 
  arg4 = static_cast< unsigned short >(val4);
  (arg1)->OnClientConnect(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnFailedLogin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnFailedLogin", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnFailedLogin" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnFailedLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnFailedLogin" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnFailedLogin" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnFailedLogin" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->OnFailedLogin((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUnknownUserRaw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUnknownUserRaw", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUnknownUserRaw" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnUnknownUserRaw" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sLine");
    }
  }
  result = (CModule::EModRet)(arg1)->OnUnknownUserRaw(arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnUnknownUserRawMessage(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CMessage *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnUnknownUserRawMessage", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnUnknownUserRawMessage" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CMessage,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnUnknownUserRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnUnknownUserRawMessage" "', argument " "2"" of type '" "CMessage &""'"); 
  }
  arg2 = reinterpret_cast< CMessage * >(argp2);
  result = (CModule::EModRet)(arg1)->OnUnknownUserRawMessage(*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_IsClientCapSupported(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_IsClientCapSupported", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_IsClientCapSupported" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_IsClientCapSupported" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_IsClientCapSupported" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_IsClientCapSupported" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CPyModule_IsClientCapSupported" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  result = (bool)(arg1)->IsClientCapSupported(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnClientCapRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  CString *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnClientCapRequest", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnClientCapRequest" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnClientCapRequest" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnClientCapRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnClientCapRequest" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CPyModule_OnClientCapRequest" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  (arg1)->OnClientCapRequest(arg2,(CString const &)*arg3,arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnClientGetSASLMechanisms(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  SCString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnClientGetSASLMechanisms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnClientGetSASLMechanisms" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__setT_CString_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnClientGetSASLMechanisms" "', argument " "2"" of type '" "SCString &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnClientGetSASLMechanisms" "', argument " "2"" of type '" "SCString &""'"); 
  }
  arg2 = reinterpret_cast< SCString * >(argp2);
  (arg1)->OnClientGetSASLMechanisms(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnClientSASLServerInitialChallenge(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnClientSASLServerInitialChallenge", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnClientSASLServerInitialChallenge" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnClientSASLServerInitialChallenge" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnClientSASLServerInitialChallenge" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[2], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg3 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 3 sResponse");
    }
  }
  result = (CModule::EModRet)(arg1)->OnClientSASLServerInitialChallenge((CString const &)*arg2,*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnClientSASLAuthenticate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnClientSASLAuthenticate", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnClientSASLAuthenticate" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnClientSASLAuthenticate" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnClientSASLAuthenticate" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnClientSASLAuthenticate" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnClientSASLAuthenticate" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  result = (CModule::EModRet)(arg1)->OnClientSASLAuthenticate((CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnClientSASLAborted(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnClientSASLAborted" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  (arg1)->OnClientSASLAborted();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnModuleLoading(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  CModInfo::EModuleType arg4 ;
  bool *arg5 = 0 ;
  CString *arg6 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject *swig_obj[6] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnModuleLoading", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnModuleLoading" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnModuleLoading" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnModuleLoading" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnModuleLoading" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnModuleLoading" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CPyModule_OnModuleLoading" "', argument " "4"" of type '" "CModInfo::EModuleType""'");
  } 
  arg4 = static_cast< CModInfo::EModuleType >(val4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_bool,  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CPyModule_OnModuleLoading" "', argument " "5"" of type '" "bool &""'"); 
  }
  if (!argp5) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnModuleLoading" "', argument " "5"" of type '" "bool &""'"); 
  }
  arg5 = reinterpret_cast< bool * >(argp5);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[5], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg6 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 6 sRetMsg");
    }
  }
  result = (CModule::EModRet)(arg1)->OnModuleLoading((CString const &)*arg2,(CString const &)*arg3,arg4,*arg5,*arg6);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnModuleUnloading(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CModule *arg2 = (CModule *) 0 ;
  bool *arg3 = 0 ;
  CString *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[4] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnModuleUnloading", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnModuleUnloading" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnModuleUnloading" "', argument " "2"" of type '" "CModule *""'"); 
  }
  arg2 = reinterpret_cast< CModule * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_bool,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnModuleUnloading" "', argument " "3"" of type '" "bool &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnModuleUnloading" "', argument " "3"" of type '" "bool &""'"); 
  }
  arg3 = reinterpret_cast< bool * >(argp3);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[3], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg4 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 4 sRetMsg");
    }
  }
  result = (CModule::EModRet)(arg1)->OnModuleUnloading(arg2,*arg3,*arg4);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnGetModInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CModInfo *arg2 = 0 ;
  CString *arg3 = 0 ;
  bool *arg4 = 0 ;
  CString *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[5] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnGetModInfo", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnGetModInfo" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_CModInfo,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnGetModInfo" "', argument " "2"" of type '" "CModInfo &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnGetModInfo" "', argument " "2"" of type '" "CModInfo &""'"); 
  }
  arg2 = reinterpret_cast< CModInfo * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnGetModInfo" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnGetModInfo" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_bool,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CPyModule_OnGetModInfo" "', argument " "4"" of type '" "bool &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnGetModInfo" "', argument " "4"" of type '" "bool &""'"); 
  }
  arg4 = reinterpret_cast< bool * >(argp4);
  {
    String* p;
    int res = SWIG_IsOK(SWIG_ConvertPtr(swig_obj[4], (void**)&p, SWIG_TypeQuery("String*"), 0));
    if (SWIG_IsOK(res)) {
      arg5 = &p->s;
    } else {
      SWIG_exception_fail(SWIG_ArgError(res), "need znc.String object as argument 5 sRetMsg");
    }
  }
  result = (CModule::EModRet)(arg1)->OnGetModInfo(*arg2,(CString const &)*arg3,*arg4,*arg5);
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnGetAvailableMods(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > *arg2 = 0 ;
  CModInfo::EModuleType arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnGetAvailableMods", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnGetAvailableMods" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__setT_CModInfo_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnGetAvailableMods" "', argument " "2"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnGetAvailableMods" "', argument " "2"" of type '" "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > &""'"); 
  }
  arg2 = reinterpret_cast< std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > * >(argp2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CPyModule_OnGetAvailableMods" "', argument " "3"" of type '" "CModInfo::EModuleType""'");
  } 
  arg3 = static_cast< CModInfo::EModuleType >(val3);
  (arg1)->OnGetAvailableMods(*arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnClientCapLs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  SCString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnClientCapLs", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnClientCapLs" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnClientCapLs" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__setT_CString_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CPyModule_OnClientCapLs" "', argument " "3"" of type '" "SCString &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyModule_OnClientCapLs" "', argument " "3"" of type '" "SCString &""'"); 
  }
  arg3 = reinterpret_cast< SCString * >(argp3);
  (arg1)->OnClientCapLs(arg2,*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModule_OnLoginAttempt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  std::shared_ptr< CAuthBase > arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  CModule::EModRet result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyModule_OnLoginAttempt", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModule_OnLoginAttempt" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    int newmem = 0;
    res2 = SWIG_ConvertPtrAndOwn(swig_obj[1], &argp2, SWIGTYPE_p_std__shared_ptrT_CAuthBase_t,  0 , &newmem);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyModule_OnLoginAttempt" "', argument " "2"" of type '" "std::shared_ptr< CAuthBase >""'");
    }
    if (argp2) arg2 = *(reinterpret_cast< std::shared_ptr< CAuthBase > * >(argp2));
    if (newmem & SWIG_CAST_NEW_MEMORY) delete reinterpret_cast< std::shared_ptr< CAuthBase > * >(argp2);
  }
  result = (CModule::EModRet)(arg1)->OnLoginAttempt(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CPyModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CPyModule" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CPyModule_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CPyModule, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CPyModule_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_AsPyModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModule *arg1 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CPyModule *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AsPyModule" "', argument " "1"" of type '" "CModule *""'"); 
  }
  arg1 = reinterpret_cast< CModule * >(argp1);
  result = (CPyModule *)AsPyModule(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CPyModule, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CreatePyModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CUser *arg1 = (CUser *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  CString *arg3 = 0 ;
  CString *arg4 = 0 ;
  CModInfo::EModuleType arg5 ;
  PyObject *arg6 = (PyObject *) 0 ;
  CModPython *arg7 = (CModPython *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  int val5 ;
  int ecode5 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  PyObject *swig_obj[7] ;
  CPyModule *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CreatePyModule", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CUser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CreatePyModule" "', argument " "1"" of type '" "CUser *""'"); 
  }
  arg1 = reinterpret_cast< CUser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CreatePyModule" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CreatePyModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CreatePyModule" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CreatePyModule" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CreatePyModule" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CreatePyModule" "', argument " "5"" of type '" "CModInfo::EModuleType""'");
  } 
  arg5 = static_cast< CModInfo::EModuleType >(val5);
  arg6 = swig_obj[5];
  res7 = SWIG_ConvertPtr(swig_obj[6], &argp7,SWIGTYPE_p_CModPython, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "CreatePyModule" "', argument " "7"" of type '" "CModPython *""'"); 
  }
  arg7 = reinterpret_cast< CModPython * >(argp7);
  result = (CPyModule *)CreatePyModule(arg1,arg2,(CString const &)*arg3,(CString const &)*arg4,arg5,arg6,arg7);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CPyModule, 0 |  0 );
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CPyTimer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  PyObject *arg6 = (PyObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  PyObject *swig_obj[6] ;
  CPyTimer *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CPyTimer", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CPyTimer" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CPyTimer" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CPyTimer" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CPyTimer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CPyTimer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_CPyTimer" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CPyTimer" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  arg6 = swig_obj[5];
  result = (CPyTimer *)new CPyTimer(arg1,arg2,arg3,(CString const &)*arg4,(CString const &)*arg5,arg6);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CPyTimer, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyTimer_RunJob(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyTimer *arg1 = (CPyTimer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyTimer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyTimer_RunJob" "', argument " "1"" of type '" "CPyTimer *""'"); 
  }
  arg1 = reinterpret_cast< CPyTimer * >(argp1);
  (arg1)->RunJob();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyTimer_GetPyObj(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyTimer *arg1 = (CPyTimer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyTimer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyTimer_GetPyObj" "', argument " "1"" of type '" "CPyTimer *""'"); 
  }
  arg1 = reinterpret_cast< CPyTimer * >(argp1);
  result = (PyObject *)(arg1)->GetPyObj();
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyTimer_GetNewPyObj(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyTimer *arg1 = (CPyTimer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyTimer, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyTimer_GetNewPyObj" "', argument " "1"" of type '" "CPyTimer *""'"); 
  }
  arg1 = reinterpret_cast< CPyTimer * >(argp1);
  result = (PyObject *)(arg1)->GetNewPyObj();
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CPyTimer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyTimer *arg1 = (CPyTimer *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyTimer, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CPyTimer" "', argument " "1"" of type '" "CPyTimer *""'"); 
  }
  arg1 = reinterpret_cast< CPyTimer * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CPyTimer_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CPyTimer, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CPyTimer_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CreatePyTimer(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  CString *arg4 = 0 ;
  CString *arg5 = 0 ;
  PyObject *arg6 = (PyObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  PyObject *swig_obj[6] ;
  CPyTimer *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CreatePyTimer", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CreatePyTimer" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CreatePyTimer" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CreatePyTimer" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    CString *ptr = (CString *)0;
    res4 = SWIG_AsPtr_CString(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CreatePyTimer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CreatePyTimer" "', argument " "4"" of type '" "CString const &""'"); 
    }
    arg4 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res5 = SWIG_AsPtr_CString(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "CreatePyTimer" "', argument " "5"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CreatePyTimer" "', argument " "5"" of type '" "CString const &""'"); 
    }
    arg5 = ptr;
  }
  arg6 = swig_obj[5];
  result = (CPyTimer *)CreatePyTimer(arg1,arg2,arg3,(CString const &)*arg4,(CString const &)*arg5,arg6);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CPyTimer, 0 |  0 );
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CPySocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  PyObject *arg2 = (PyObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CPySocket *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CPySocket", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CPySocket" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  arg2 = swig_obj[1];
  result = (CPySocket *)new CPySocket(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CPySocket, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPySocket_GetPyObj(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPySocket *arg1 = (CPySocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPySocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPySocket_GetPyObj" "', argument " "1"" of type '" "CPySocket *""'"); 
  }
  arg1 = reinterpret_cast< CPySocket * >(argp1);
  result = (PyObject *)(arg1)->GetPyObj();
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPySocket_GetNewPyObj(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPySocket *arg1 = (CPySocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPySocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPySocket_GetNewPyObj" "', argument " "1"" of type '" "CPySocket *""'"); 
  }
  arg1 = reinterpret_cast< CPySocket * >(argp1);
  result = (PyObject *)(arg1)->GetNewPyObj();
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CPySocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPySocket *arg1 = (CPySocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPySocket, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CPySocket" "', argument " "1"" of type '" "CPySocket *""'"); 
  }
  arg1 = reinterpret_cast< CPySocket * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPySocket_Connected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPySocket *arg1 = (CPySocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPySocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPySocket_Connected" "', argument " "1"" of type '" "CPySocket *""'"); 
  }
  arg1 = reinterpret_cast< CPySocket * >(argp1);
  (arg1)->Connected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPySocket_Disconnected(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPySocket *arg1 = (CPySocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPySocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPySocket_Disconnected" "', argument " "1"" of type '" "CPySocket *""'"); 
  }
  arg1 = reinterpret_cast< CPySocket * >(argp1);
  (arg1)->Disconnected();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPySocket_Timeout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPySocket *arg1 = (CPySocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPySocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPySocket_Timeout" "', argument " "1"" of type '" "CPySocket *""'"); 
  }
  arg1 = reinterpret_cast< CPySocket * >(argp1);
  (arg1)->Timeout();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPySocket_ConnectionRefused(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPySocket *arg1 = (CPySocket *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPySocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPySocket_ConnectionRefused" "', argument " "1"" of type '" "CPySocket *""'"); 
  }
  arg1 = reinterpret_cast< CPySocket * >(argp1);
  (arg1)->ConnectionRefused();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPySocket_ReadData(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPySocket *arg1 = (CPySocket *) 0 ;
  char *arg2 = (char *) 0 ;
  size_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  size_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPySocket_ReadData", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPySocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPySocket_ReadData" "', argument " "1"" of type '" "CPySocket *""'"); 
  }
  arg1 = reinterpret_cast< CPySocket * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPySocket_ReadData" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CPySocket_ReadData" "', argument " "3"" of type '" "size_t""'");
  } 
  arg3 = static_cast< size_t >(val3);
  (arg1)->ReadData((char const *)arg2,arg3);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPySocket_ReadLine(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPySocket *arg1 = (CPySocket *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPySocket_ReadLine", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPySocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPySocket_ReadLine" "', argument " "1"" of type '" "CPySocket *""'"); 
  }
  arg1 = reinterpret_cast< CPySocket * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPySocket_ReadLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPySocket_ReadLine" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->ReadLine((CString const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPySocket_GetSockObj(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPySocket *arg1 = (CPySocket *) 0 ;
  CString *arg2 = 0 ;
  unsigned short arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  unsigned short val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Csock *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPySocket_GetSockObj", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPySocket, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPySocket_GetSockObj" "', argument " "1"" of type '" "CPySocket *""'"); 
  }
  arg1 = reinterpret_cast< CPySocket * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPySocket_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPySocket_GetSockObj" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_unsigned_SS_short(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CPySocket_GetSockObj" "', argument " "3"" of type '" "unsigned short""'");
  } 
  arg3 = static_cast< unsigned short >(val3);
  result = (Csock *)(arg1)->GetSockObj((CString const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Csock, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *CPySocket_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CPySocket, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CPySocket_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CreatePySocket(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  PyObject *arg2 = (PyObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  CPySocket *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CreatePySocket", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CreatePySocket" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  arg2 = swig_obj[1];
  result = (CPySocket *)CreatePySocket(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CPySocket, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_HaveIPv6_(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "HaveIPv6_", 0, 0, 0)) SWIG_fail;
  result = (bool)HaveIPv6_();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_HaveSSL_(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "HaveSSL_", 0, 0, 0)) SWIG_fail;
  result = (bool)HaveSSL_();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_HaveCharset_(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "HaveCharset_", 0, 0, 0)) SWIG_fail;
  result = (bool)HaveCharset_();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GetSOMAXCONN(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "GetSOMAXCONN", 0, 0, 0)) SWIG_fail;
  result = (int)GetSOMAXCONN();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GetVersionMajor(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "GetVersionMajor", 0, 0, 0)) SWIG_fail;
  result = (int)GetVersionMajor();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GetVersionMinor(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "GetVersionMinor", 0, 0, 0)) SWIG_fail;
  result = (int)GetVersionMinor();
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GetVersion(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  double result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "GetVersion", 0, 0, 0)) SWIG_fail;
  result = (double)GetVersion();
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GetVersionExtra(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "GetVersionExtra", 0, 0, 0)) SWIG_fail;
  result = GetVersionExtra();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_MCString_iter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  MCString_iter *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (MCString_iter *)new MCString_iter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_MCString_iter, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MCString_iter_x_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  MCString_iter *arg1 = (MCString_iter *) 0 ;
  MCString::iterator *arg2 = (MCString::iterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "MCString_iter_x_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_MCString_iter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MCString_iter_x_set" "', argument " "1"" of type '" "MCString_iter *""'"); 
  }
  arg1 = reinterpret_cast< MCString_iter * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__mapT_CString_CString_t__iterator, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MCString_iter_x_set" "', argument " "2"" of type '" "MCString::iterator *""'"); 
  }
  arg2 = reinterpret_cast< MCString::iterator * >(argp2);
  if (arg1) (arg1)->x = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MCString_iter_x_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  MCString_iter *arg1 = (MCString_iter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  MCString::iterator *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_MCString_iter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MCString_iter_x_get" "', argument " "1"" of type '" "MCString_iter *""'"); 
  }
  arg1 = reinterpret_cast< MCString_iter * >(argp1);
  result = (MCString::iterator *)& ((arg1)->x);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_CString_CString_t__iterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_MCString_iter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  MCString::iterator arg1 ;
  swig::SwigPyIterator *iter1 = 0 ;
  int res1 ;
  MCString_iter *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], SWIG_as_voidptrptr(&iter1), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res1) || !iter1) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "new_MCString_iter" "', argument " "1"" of type '" "MCString::iterator""'");
  } else {
    swig::SwigPyIterator_T<MCString::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<MCString::iterator > *>(iter1);
    if (iter_t) {
      arg1 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "new_MCString_iter" "', argument " "1"" of type '" "MCString::iterator""'");
    }
  }
  result = (MCString_iter *)new MCString_iter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_MCString_iter, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_MCString_iter(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_MCString_iter", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_MCString_iter__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    swig::SwigPyIterator *iter = 0;
    int res = SWIG_ConvertPtr(argv[0], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
    _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<MCString::iterator > *>(iter) != 0));
    if (_v) {
      return _wrap_new_MCString_iter__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_MCString_iter'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    MCString_iter::MCString_iter()\n"
    "    MCString_iter::MCString_iter(MCString::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_MCString_iter_plusplus(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  MCString_iter *arg1 = (MCString_iter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_MCString_iter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MCString_iter_plusplus" "', argument " "1"" of type '" "MCString_iter *""'"); 
  }
  arg1 = reinterpret_cast< MCString_iter * >(argp1);
  (arg1)->plusplus();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MCString_iter_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  MCString_iter *arg1 = (MCString_iter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_MCString_iter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MCString_iter_get" "', argument " "1"" of type '" "MCString_iter *""'"); 
  }
  arg1 = reinterpret_cast< MCString_iter * >(argp1);
  result = (arg1)->get();
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MCString_iter_is_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  MCString_iter *arg1 = (MCString_iter *) 0 ;
  CModule *arg2 = (CModule *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "MCString_iter_is_end", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_MCString_iter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MCString_iter_is_end" "', argument " "1"" of type '" "MCString_iter *""'"); 
  }
  arg1 = reinterpret_cast< MCString_iter * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CModule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MCString_iter_is_end" "', argument " "2"" of type '" "CModule *""'"); 
  }
  arg2 = reinterpret_cast< CModule * >(argp2);
  result = (bool)(arg1)->is_end(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_MCString_iter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  MCString_iter *arg1 = (MCString_iter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_MCString_iter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_MCString_iter" "', argument " "1"" of type '" "MCString_iter *""'"); 
  }
  arg1 = reinterpret_cast< MCString_iter * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *MCString_iter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_MCString_iter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *MCString_iter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CModulesIter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModules *arg1 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CModulesIter *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CModulesIter" "', argument " "1"" of type '" "CModules *""'"); 
  }
  arg1 = reinterpret_cast< CModules * >(argp1);
  result = (CModulesIter *)new CModulesIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModulesIter, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModulesIter_plusplus(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModulesIter *arg1 = (CModulesIter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModulesIter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModulesIter_plusplus" "', argument " "1"" of type '" "CModulesIter *""'"); 
  }
  arg1 = reinterpret_cast< CModulesIter * >(argp1);
  (arg1)->plusplus();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModulesIter_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModulesIter *arg1 = (CModulesIter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CModule *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModulesIter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModulesIter_get" "', argument " "1"" of type '" "CModulesIter const *""'"); 
  }
  arg1 = reinterpret_cast< CModulesIter * >(argp1);
  result = (CModule *)((CModulesIter const *)arg1)->get();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModule, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModulesIter_is_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModulesIter *arg1 = (CModulesIter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModulesIter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModulesIter_is_end" "', argument " "1"" of type '" "CModulesIter const *""'"); 
  }
  arg1 = reinterpret_cast< CModulesIter * >(argp1);
  result = (bool)((CModulesIter const *)arg1)->is_end();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModulesIter_m_pModules_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModulesIter *arg1 = (CModulesIter *) 0 ;
  CModules *arg2 = (CModules *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModulesIter_m_pModules_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModulesIter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModulesIter_m_pModules_set" "', argument " "1"" of type '" "CModulesIter *""'"); 
  }
  arg1 = reinterpret_cast< CModulesIter * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CModules, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModulesIter_m_pModules_set" "', argument " "2"" of type '" "CModules *""'"); 
  }
  arg2 = reinterpret_cast< CModules * >(argp2);
  if (arg1) (arg1)->m_pModules = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModulesIter_m_pModules_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModulesIter *arg1 = (CModulesIter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CModules *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModulesIter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModulesIter_m_pModules_get" "', argument " "1"" of type '" "CModulesIter *""'"); 
  }
  arg1 = reinterpret_cast< CModulesIter * >(argp1);
  result = (CModules *) ((arg1)->m_pModules);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CModules, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModulesIter_m_it_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModulesIter *arg1 = (CModulesIter *) 0 ;
  CModules::const_iterator *arg2 = (CModules::const_iterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CModulesIter_m_it_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModulesIter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModulesIter_m_it_set" "', argument " "1"" of type '" "CModulesIter *""'"); 
  }
  arg1 = reinterpret_cast< CModulesIter * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__vectorT_CModule_p_t__const_iterator, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CModulesIter_m_it_set" "', argument " "2"" of type '" "CModules::const_iterator *""'"); 
  }
  arg2 = reinterpret_cast< CModules::const_iterator * >(argp2);
  if (arg1) (arg1)->m_it = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CModulesIter_m_it_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModulesIter *arg1 = (CModulesIter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CModules::const_iterator *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModulesIter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CModulesIter_m_it_get" "', argument " "1"" of type '" "CModulesIter *""'"); 
  }
  arg1 = reinterpret_cast< CModulesIter * >(argp1);
  result = (CModules::const_iterator *)& ((arg1)->m_it);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_CModule_p_t__const_iterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CModulesIter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CModulesIter *arg1 = (CModulesIter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CModulesIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CModulesIter" "', argument " "1"" of type '" "CModulesIter *""'"); 
  }
  arg1 = reinterpret_cast< CModulesIter * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CModulesIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CModulesIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CModulesIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CPyModCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  COptionalTranslation *arg3 = 0 ;
  COptionalTranslation *arg4 = 0 ;
  PyObject *arg5 = (PyObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[5] ;
  CPyModCommand *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CPyModCommand", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_CPyModCommand" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_CPyModCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CPyModCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_COptionalTranslation,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_CPyModCommand" "', argument " "3"" of type '" "COptionalTranslation const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CPyModCommand" "', argument " "3"" of type '" "COptionalTranslation const &""'"); 
  }
  arg3 = reinterpret_cast< COptionalTranslation * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_COptionalTranslation,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_CPyModCommand" "', argument " "4"" of type '" "COptionalTranslation const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_CPyModCommand" "', argument " "4"" of type '" "COptionalTranslation const &""'"); 
  }
  arg4 = reinterpret_cast< COptionalTranslation * >(argp4);
  arg5 = swig_obj[4];
  result = (CPyModCommand *)new CPyModCommand(arg1,(CString const &)*arg2,(COptionalTranslation const &)*arg3,(COptionalTranslation const &)*arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CPyModCommand, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CPyModCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModCommand *arg1 = (CPyModCommand *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModCommand, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CPyModCommand" "', argument " "1"" of type '" "CPyModCommand *""'"); 
  }
  arg1 = reinterpret_cast< CPyModCommand * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyModCommand_GetModule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModCommand *arg1 = (CPyModCommand *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CPyModule *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModCommand, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyModCommand_GetModule" "', argument " "1"" of type '" "CPyModCommand *""'"); 
  }
  arg1 = reinterpret_cast< CPyModCommand * >(argp1);
  result = (CPyModule *)(arg1)->GetModule();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CPyModule, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CPyModCommand_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CPyModCommand, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CPyModCommand_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CreatePyModCommand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyModule *arg1 = (CPyModule *) 0 ;
  CString *arg2 = 0 ;
  COptionalTranslation *arg3 = 0 ;
  COptionalTranslation *arg4 = 0 ;
  PyObject *arg5 = (PyObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[5] ;
  CPyModCommand *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CreatePyModCommand", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyModule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CreatePyModCommand" "', argument " "1"" of type '" "CPyModule *""'"); 
  }
  arg1 = reinterpret_cast< CPyModule * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CreatePyModCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CreatePyModCommand" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_COptionalTranslation,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CreatePyModCommand" "', argument " "3"" of type '" "COptionalTranslation const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CreatePyModCommand" "', argument " "3"" of type '" "COptionalTranslation const &""'"); 
  }
  arg3 = reinterpret_cast< COptionalTranslation * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_COptionalTranslation,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CreatePyModCommand" "', argument " "4"" of type '" "COptionalTranslation const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CreatePyModCommand" "', argument " "4"" of type '" "COptionalTranslation const &""'"); 
  }
  arg4 = reinterpret_cast< COptionalTranslation * >(argp4);
  arg5 = swig_obj[4];
  result = (CPyModCommand *)CreatePyModCommand(arg1,(CString const &)*arg2,(COptionalTranslation const &)*arg3,(COptionalTranslation const &)*arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CPyModCommand, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_CPyCapability(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  PyObject *arg1 = (PyObject *) 0 ;
  PyObject *arg2 = (PyObject *) 0 ;
  PyObject *swig_obj[2] ;
  CPyCapability *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_CPyCapability", 2, 2, swig_obj)) SWIG_fail;
  arg1 = swig_obj[0];
  arg2 = swig_obj[1];
  result = (CPyCapability *)new CPyCapability(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_CPyCapability, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CPyCapability(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyCapability *arg1 = (CPyCapability *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyCapability, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CPyCapability" "', argument " "1"" of type '" "CPyCapability *""'"); 
  }
  arg1 = reinterpret_cast< CPyCapability * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyCapability_OnServerChangedSupport(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyCapability *arg1 = (CPyCapability *) 0 ;
  CIRCNetwork *arg2 = (CIRCNetwork *) 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyCapability_OnServerChangedSupport", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyCapability, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyCapability_OnServerChangedSupport" "', argument " "1"" of type '" "CPyCapability *""'"); 
  }
  arg1 = reinterpret_cast< CPyCapability * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CIRCNetwork, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyCapability_OnServerChangedSupport" "', argument " "2"" of type '" "CIRCNetwork *""'"); 
  }
  arg2 = reinterpret_cast< CIRCNetwork * >(argp2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CPyCapability_OnServerChangedSupport" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->OnServerChangedSupport(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyCapability_OnClientChangedSupport(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyCapability *arg1 = (CPyCapability *) 0 ;
  CClient *arg2 = (CClient *) 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyCapability_OnClientChangedSupport", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyCapability, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyCapability_OnClientChangedSupport" "', argument " "1"" of type '" "CPyCapability *""'"); 
  }
  arg1 = reinterpret_cast< CPyCapability * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_CClient, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyCapability_OnClientChangedSupport" "', argument " "2"" of type '" "CClient *""'"); 
  }
  arg2 = reinterpret_cast< CClient * >(argp2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CPyCapability_OnClientChangedSupport" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  (arg1)->OnClientChangedSupport(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CPyCapability_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CPyCapability, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CPyCapability_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CPyRetString_s_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyRetString *arg1 = (CPyRetString *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyRetString_s_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyRetString, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyRetString_s_set" "', argument " "1"" of type '" "CPyRetString *""'"); 
  }
  arg1 = reinterpret_cast< CPyRetString * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CPyRetString_s_set" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CPyRetString_s_set" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  if (arg1) (arg1)->s = *arg2;
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyRetString_s_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyRetString *arg1 = (CPyRetString *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyRetString, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyRetString_s_get" "', argument " "1"" of type '" "CPyRetString *""'"); 
  }
  arg1 = reinterpret_cast< CPyRetString * >(argp1);
  result = (CString *) & ((arg1)->s);
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyRetString___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyRetString *arg1 = (CPyRetString *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyRetString, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyRetString___str__" "', argument " "1"" of type '" "CPyRetString *""'"); 
  }
  arg1 = reinterpret_cast< CPyRetString * >(argp1);
  result = CPyRetString___str__(arg1);
  resultobj = SWIG_From_CString(static_cast< CString >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CPyRetString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyRetString *arg1 = (CPyRetString *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyRetString, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CPyRetString" "', argument " "1"" of type '" "CPyRetString *""'"); 
  }
  arg1 = reinterpret_cast< CPyRetString * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CPyRetString_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CPyRetString, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_CPyRetBool_b_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyRetBool *arg1 = (CPyRetBool *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CPyRetBool_b_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyRetBool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyRetBool_b_set" "', argument " "1"" of type '" "CPyRetBool *""'"); 
  }
  arg1 = reinterpret_cast< CPyRetBool * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CPyRetBool_b_set" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  if (arg1) (arg1)->b = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyRetBool_b_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyRetBool *arg1 = (CPyRetBool *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyRetBool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyRetBool_b_get" "', argument " "1"" of type '" "CPyRetBool *""'"); 
  }
  arg1 = reinterpret_cast< CPyRetBool * >(argp1);
  result = (bool) ((arg1)->b);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CPyRetBool___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyRetBool *arg1 = (CPyRetBool *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyRetBool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CPyRetBool___bool__" "', argument " "1"" of type '" "CPyRetBool *""'"); 
  }
  arg1 = reinterpret_cast< CPyRetBool * >(argp1);
  result = (bool)CPyRetBool___bool__(arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CPyRetBool(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CPyRetBool *arg1 = (CPyRetBool *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_CPyRetBool, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CPyRetBool" "', argument " "1"" of type '" "CPyRetBool *""'"); 
  }
  arg1 = reinterpret_cast< CPyRetBool * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CPyRetBool_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CPyRetBool, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CModPython_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_CModPython, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_StrPair__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::pair< CString,CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::pair< CString,CString > *)new std::pair< CString,CString >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__pairT_CString_CString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_StrPair__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  CString arg1 ;
  CString arg2 ;
  std::pair< CString,CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    int res = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_StrPair" "', argument " "1"" of type '" "CString""'"); 
    }
    arg1 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  {
    CString *ptr = (CString *)0;
    int res = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_StrPair" "', argument " "2"" of type '" "CString""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  result = (std::pair< CString,CString > *)new std::pair< CString,CString >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__pairT_CString_CString_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_StrPair__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::pair< CString,CString > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::pair< CString,CString > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::pair< CString,CString > *ptr = (std::pair< CString,CString > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_StrPair" "', argument " "1"" of type '" "std::pair< CString,CString > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_StrPair" "', argument " "1"" of type '" "std::pair< CString,CString > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::pair< CString,CString > *)new std::pair< CString,CString >((std::pair< CString,CString > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__pairT_CString_CString_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_StrPair(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_StrPair", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_StrPair__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::pair< CString,CString >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_StrPair__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = SWIG_AsPtr_CString(argv[0], (CString**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_CString(argv[1], (CString**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_StrPair__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_StrPair'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::pair< CString,CString >::pair()\n"
    "    std::pair< CString,CString >::pair(CString,CString)\n"
    "    std::pair< CString,CString >::pair(std::pair< CString,CString > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_StrPair_first_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< CString,CString > *arg1 = (std::pair< CString,CString > *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "StrPair_first_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StrPair_first_set" "', argument " "1"" of type '" "std::pair< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< CString,CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StrPair_first_set" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "StrPair_first_set" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  if (arg1) (arg1)->first = *arg2;
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StrPair_first_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< CString,CString > *arg1 = (std::pair< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StrPair_first_get" "', argument " "1"" of type '" "std::pair< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< CString,CString > * >(argp1);
  result = (CString *) & ((arg1)->first);
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StrPair_second_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< CString,CString > *arg1 = (std::pair< CString,CString > *) 0 ;
  CString *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "StrPair_second_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StrPair_second_set" "', argument " "1"" of type '" "std::pair< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< CString,CString > * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StrPair_second_set" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "StrPair_second_set" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  if (arg1) (arg1)->second = *arg2;
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StrPair_second_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< CString,CString > *arg1 = (std::pair< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  CString *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_CString_CString_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StrPair_second_get" "', argument " "1"" of type '" "std::pair< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< CString,CString > * >(argp1);
  result = (CString *) & ((arg1)->second);
  resultobj = SWIG_From_CString(static_cast< CString >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_StrPair(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< CString,CString > *arg1 = (std::pair< CString,CString > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_CString_CString_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_StrPair" "', argument " "1"" of type '" "std::pair< CString,CString > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< CString,CString > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *StrPair_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__pairT_CString_CString_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *StrPair_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_VPair_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_iterator" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___nonzero__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = (bool)std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____nonzero__((std::vector< std::pair< CString,CString > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___bool__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = (bool)std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____bool__((std::vector< std::pair< CString,CString > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::pair< CString,CString > >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___len__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____len__((std::vector< std::pair< CString,CString > > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___getslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::difference_type arg2 ;
  std::vector< std::pair< CString,CString > >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VPair___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___getslice__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VPair___getslice__" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::pair< CString,CString > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VPair___getslice__" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::pair< CString,CString > >::difference_type >(val3);
  try {
    result = (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *)std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::difference_type arg2 ;
  std::vector< std::pair< CString,CString > >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___setslice__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VPair___setslice__" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::pair< CString,CString > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VPair___setslice__" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::pair< CString,CString > >::difference_type >(val3);
  try {
    std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::difference_type arg2 ;
  std::vector< std::pair< CString,CString > >::difference_type arg3 ;
  std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___setslice__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VPair___setslice__" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::pair< CString,CString > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VPair___setslice__" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::pair< CString,CString > >::difference_type >(val3);
  {
    std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *ptr = (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VPair___setslice__" "', argument " "4"" of type '" "std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VPair___setslice__" "', argument " "4"" of type '" "std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VPair___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_VPair___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VPair___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VPair___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::pair< CString,CString > >::__setslice__(std::vector< std::pair< CString,CString > >::difference_type,std::vector< std::pair< CString,CString > >::difference_type)\n"
    "    std::vector< std::pair< CString,CString > >::__setslice__(std::vector< std::pair< CString,CString > >::difference_type,std::vector< std::pair< CString,CString > >::difference_type,std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VPair___delslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::difference_type arg2 ;
  std::vector< std::pair< CString,CString > >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VPair___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___delslice__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VPair___delslice__" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::pair< CString,CString > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VPair___delslice__" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::pair< CString,CString > >::difference_type >(val3);
  try {
    std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___delitem__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VPair___delitem__" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::pair< CString,CString > >::difference_type >(val2);
  try {
    std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___getitem__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    result = (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *)std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___setitem__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  {
    std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *ptr = (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VPair___setitem__" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VPair___setitem__" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___setitem__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___delitem__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VPair___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VPair___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VPair___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VPair___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::pair< CString,CString > >::__delitem__(std::vector< std::pair< CString,CString > >::difference_type)\n"
    "    std::vector< std::pair< CString,CString > >::__delitem__(SWIGPY_SLICEOBJECT *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VPair___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< std::pair< CString,CString > >::value_type *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___getitem__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VPair___getitem__" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::pair< CString,CString > >::difference_type >(val2);
  try {
    result = (std::vector< std::pair< CString,CString > >::value_type *) &std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____getitem____SWIG_1((std::vector< std::pair< CString,CString > > const *)arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = swig::from(static_cast< std::pair< CString,CString > >(*result));
  (void)swig::container_owner<swig::traits<std::vector< std::pair< CString,CString > >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VPair___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VPair___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VPair___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VPair___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::pair< CString,CString > >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< std::pair< CString,CString > >::__getitem__(std::vector< std::pair< CString,CString > >::difference_type) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VPair___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::difference_type arg2 ;
  std::vector< std::pair< CString,CString > >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair___setitem__" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VPair___setitem__" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::pair< CString,CString > >::difference_type >(val2);
  {
    std::pair< CString,CString > *ptr = (std::pair< CString,CString > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VPair___setitem__" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VPair___setitem__" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(std::pair< CString,CString > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VPair___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VPair___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VPair___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::pair< CString,CString >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VPair___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VPair___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::pair< CString,CString > >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > const &)\n"
    "    std::vector< std::pair< CString,CString > >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< std::pair< CString,CString > >::__setitem__(std::vector< std::pair< CString,CString > >::difference_type,std::vector< std::pair< CString,CString > >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VPair_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::pair< CString,CString > >::value_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_pop" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  try {
    result = std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = swig::from(static_cast< std::pair< CString,CString > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VPair_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_append" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  {
    std::pair< CString,CString > *ptr = (std::pair< CString,CString > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VPair_append" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VPair_append" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__append(arg1,(std::pair< CString,CString > const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VPair__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< std::pair< CString,CString > > *)new std::vector< std::pair< CString,CString > >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VPair__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< std::pair< CString,CString > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *ptr = (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_VPair" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VPair" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< std::pair< CString,CString > > *)new std::vector< std::pair< CString,CString > >((std::vector< std::pair< CString,CString > > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_empty" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = (bool)((std::vector< std::pair< CString,CString > > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::pair< CString,CString > >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_size" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = ((std::vector< std::pair< CString,CString > > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VPair_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_swap" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VPair_swap" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VPair_swap" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::pair< CString,CString > >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_begin" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::pair< CString,CString > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::pair< CString,CString > >::iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_end" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::pair< CString,CString > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::pair< CString,CString > >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_rbegin" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::pair< CString,CString > >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::pair< CString,CString > >::reverse_iterator result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_rend" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::pair< CString,CString > >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_clear" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< std::pair< CString,CString > > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_get_allocator" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = ((std::vector< std::pair< CString,CString > > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< std::pair< CString,CString > >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_std__pairT_CString_CString_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VPair__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< std::pair< CString,CString > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VPair" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > >::size_type""'");
  } 
  arg1 = static_cast< std::vector< std::pair< CString,CString > >::size_type >(val1);
  result = (std::vector< std::pair< CString,CString > > *)new std::vector< std::pair< CString,CString > >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_pop_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_pop_back" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_resize" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VPair_resize" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::pair< CString,CString > >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< std::pair< CString,CString > >::iterator result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_erase" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair_erase" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair_erase" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::iterator""'");
    }
  }
  result = std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::pair< CString,CString > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::iterator arg2 ;
  std::vector< std::pair< CString,CString > >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::vector< std::pair< CString,CString > >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_erase" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair_erase" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair_erase" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair_erase" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair_erase" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::iterator""'");
    }
  }
  result = std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::pair< CString,CString > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VPair_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_VPair_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_VPair_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VPair_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::pair< CString,CString > >::erase(std::vector< std::pair< CString,CString > >::iterator)\n"
    "    std::vector< std::pair< CString,CString > >::erase(std::vector< std::pair< CString,CString > >::iterator,std::vector< std::pair< CString,CString > >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_VPair__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > >::size_type arg1 ;
  std::vector< std::pair< CString,CString > >::value_type *arg2 = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  std::vector< std::pair< CString,CString > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VPair" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > >::size_type""'");
  } 
  arg1 = static_cast< std::vector< std::pair< CString,CString > >::size_type >(val1);
  {
    std::pair< CString,CString > *ptr = (std::pair< CString,CString > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_VPair" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VPair" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (std::vector< std::pair< CString,CString > > *)new std::vector< std::pair< CString,CString > >(arg1,(std::vector< std::pair< CString,CString > >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VPair(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_VPair", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_VPair__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_VPair__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_VPair__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      int res = swig::asptr(argv[1], (std::pair< CString,CString >**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_VPair__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_VPair'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::pair< CString,CString > >::vector()\n"
    "    std::vector< std::pair< CString,CString > >::vector(std::vector< std::pair< CString,CString > > const &)\n"
    "    std::vector< std::pair< CString,CString > >::vector(std::vector< std::pair< CString,CString > >::size_type)\n"
    "    std::vector< std::pair< CString,CString > >::vector(std::vector< std::pair< CString,CString > >::size_type,std::vector< std::pair< CString,CString > >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VPair_push_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VPair_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_push_back" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  {
    std::pair< CString,CString > *ptr = (std::pair< CString,CString > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VPair_push_back" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VPair_push_back" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->push_back((std::vector< std::pair< CString,CString > >::value_type const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::pair< CString,CString > >::value_type *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_front" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = (std::vector< std::pair< CString,CString > >::value_type *) &((std::vector< std::pair< CString,CString > > const *)arg1)->front();
  resultobj = swig::from(static_cast< std::pair< CString,CString > >(*result));
  (void)swig::container_owner<swig::traits<std::vector< std::pair< CString,CString > >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::pair< CString,CString > >::value_type *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_back" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = (std::vector< std::pair< CString,CString > >::value_type *) &((std::vector< std::pair< CString,CString > > const *)arg1)->back();
  resultobj = swig::from(static_cast< std::pair< CString,CString > >(*result));
  (void)swig::container_owner<swig::traits<std::vector< std::pair< CString,CString > >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_assign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::size_type arg2 ;
  std::vector< std::pair< CString,CString > >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VPair_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_assign" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VPair_assign" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::pair< CString,CString > >::size_type >(val2);
  {
    std::pair< CString,CString > *ptr = (std::pair< CString,CString > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VPair_assign" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VPair_assign" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->assign(arg2,(std::vector< std::pair< CString,CString > >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::size_type arg2 ;
  std::vector< std::pair< CString,CString > >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_resize" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VPair_resize" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::pair< CString,CString > >::size_type >(val2);
  {
    std::pair< CString,CString > *ptr = (std::pair< CString,CString > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VPair_resize" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VPair_resize" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->resize(arg2,(std::vector< std::pair< CString,CString > >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VPair_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VPair_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::pair< CString,CString >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VPair_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VPair_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::pair< CString,CString > >::resize(std::vector< std::pair< CString,CString > >::size_type)\n"
    "    std::vector< std::pair< CString,CString > >::resize(std::vector< std::pair< CString,CString > >::size_type,std::vector< std::pair< CString,CString > >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VPair_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::iterator arg2 ;
  std::vector< std::pair< CString,CString > >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  int res3 = SWIG_OLDOBJ ;
  std::vector< std::pair< CString,CString > >::iterator result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_insert" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair_insert" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair_insert" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::iterator""'");
    }
  }
  {
    std::pair< CString,CString > *ptr = (std::pair< CString,CString > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VPair_insert" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VPair_insert" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  result = std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),(std::pair< CString,CString > const &)*arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::pair< CString,CString > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::iterator arg2 ;
  std::vector< std::pair< CString,CString > >::size_type arg3 ;
  std::vector< std::pair< CString,CString > >::value_type *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_insert" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair_insert" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VPair_insert" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VPair_insert" "', argument " "3"" of type '" "std::vector< std::pair< CString,CString > >::size_type""'");
  } 
  arg3 = static_cast< std::vector< std::pair< CString,CString > >::size_type >(val3);
  {
    std::pair< CString,CString > *ptr = (std::pair< CString,CString > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VPair_insert" "', argument " "4"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VPair_insert" "', argument " "4"" of type '" "std::vector< std::pair< CString,CString > >::value_type const &""'"); 
    }
    arg4 = ptr;
  }
  std_vector_Sl_std_pair_Sl_CString_Sc_CString_Sg__Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::pair< CString,CString > const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VPair_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *>(iter) != 0));
      if (_v) {
        int res = swig::asptr(argv[2], (std::pair< CString,CString >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VPair_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::pair< CString,CString > >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::pair< CString,CString >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VPair_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VPair_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::pair< CString,CString > >::insert(std::vector< std::pair< CString,CString > >::iterator,std::vector< std::pair< CString,CString > >::value_type const &)\n"
    "    std::vector< std::pair< CString,CString > >::insert(std::vector< std::pair< CString,CString > >::iterator,std::vector< std::pair< CString,CString > >::size_type,std::vector< std::pair< CString,CString > >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VPair_reserve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  std::vector< std::pair< CString,CString > >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VPair_reserve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_reserve" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VPair_reserve" "', argument " "2"" of type '" "std::vector< std::pair< CString,CString > >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::pair< CString,CString > >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VPair_capacity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::pair< CString,CString > >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_capacity" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  result = ((std::vector< std::pair< CString,CString > > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_VPair(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::pair< CString,CString > > *arg1 = (std::vector< std::pair< CString,CString > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VPair" "', argument " "1"" of type '" "std::vector< std::pair< CString,CString > > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::pair< CString,CString > > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *VPair_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *VPair_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_VWebSubPages_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_iterator" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_TWebSubPage_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___nonzero__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___nonzero__" "', argument " "1"" of type '" "std::vector< TWebSubPage > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = (bool)std_vector_Sl_TWebSubPage_Sg____nonzero__((std::vector< std::shared_ptr< CWebSubPage > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___bool__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___bool__" "', argument " "1"" of type '" "std::vector< TWebSubPage > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = (bool)std_vector_Sl_TWebSubPage_Sg____bool__((std::vector< std::shared_ptr< CWebSubPage > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::shared_ptr< CWebSubPage > >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___len__" "', argument " "1"" of type '" "std::vector< TWebSubPage > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = std_vector_Sl_TWebSubPage_Sg____len__((std::vector< std::shared_ptr< CWebSubPage > > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___getslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::difference_type arg2 ;
  std::vector< std::shared_ptr< CWebSubPage > >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VWebSubPages___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___getslice__" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VWebSubPages___getslice__" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VWebSubPages___getslice__" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::difference_type >(val3);
  try {
    result = (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *)std_vector_Sl_TWebSubPage_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_TWebSubPage_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::difference_type arg2 ;
  std::vector< std::shared_ptr< CWebSubPage > >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___setslice__" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VWebSubPages___setslice__" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VWebSubPages___setslice__" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::difference_type >(val3);
  try {
    std_vector_Sl_TWebSubPage_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::difference_type arg2 ;
  std::vector< std::shared_ptr< CWebSubPage > >::difference_type arg3 ;
  std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___setslice__" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VWebSubPages___setslice__" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VWebSubPages___setslice__" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::difference_type >(val3);
  {
    std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *ptr = (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VWebSubPages___setslice__" "', argument " "4"" of type '" "std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VWebSubPages___setslice__" "', argument " "4"" of type '" "std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_TWebSubPage_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VWebSubPages___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_VWebSubPages___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VWebSubPages___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VWebSubPages___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< TWebSubPage >::__setslice__(std::vector< std::shared_ptr< CWebSubPage > >::difference_type,std::vector< std::shared_ptr< CWebSubPage > >::difference_type)\n"
    "    std::vector< TWebSubPage >::__setslice__(std::vector< std::shared_ptr< CWebSubPage > >::difference_type,std::vector< std::shared_ptr< CWebSubPage > >::difference_type,std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___delslice__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::difference_type arg2 ;
  std::vector< std::shared_ptr< CWebSubPage > >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VWebSubPages___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___delslice__" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VWebSubPages___delslice__" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VWebSubPages___delslice__" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::difference_type >(val3);
  try {
    std_vector_Sl_TWebSubPage_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___delitem__" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VWebSubPages___delitem__" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::difference_type >(val2);
  try {
    std_vector_Sl_TWebSubPage_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___getitem__" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    result = (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *)std_vector_Sl_TWebSubPage_Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_TWebSubPage_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___setitem__" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  {
    std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *ptr = (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VWebSubPages___setitem__" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VWebSubPages___setitem__" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_TWebSubPage_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___setitem__" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_TWebSubPage_Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  SWIGPY_SLICEOBJECT *arg2 = (SWIGPY_SLICEOBJECT *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___delitem__" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'");
    }
    arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1];
  }
  try {
    std_vector_Sl_TWebSubPage_Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VWebSubPages___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VWebSubPages___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VWebSubPages___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VWebSubPages___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< TWebSubPage >::__delitem__(std::vector< std::shared_ptr< CWebSubPage > >::difference_type)\n"
    "    std::vector< TWebSubPage >::__delitem__(SWIGPY_SLICEOBJECT *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::value_type *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___getitem__" "', argument " "1"" of type '" "std::vector< TWebSubPage > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VWebSubPages___getitem__" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::difference_type >(val2);
  try {
    result = (std::vector< std::shared_ptr< CWebSubPage > >::value_type *) &std_vector_Sl_TWebSubPage_Sg____getitem____SWIG_1((std::vector< std::shared_ptr< CWebSubPage > > const *)arg1,SWIG_STD_MOVE(arg2));
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t, 0 |  0 );
  (void)swig::container_owner<swig::traits<std::vector< std::shared_ptr< CWebSubPage > >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VWebSubPages___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VWebSubPages___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VWebSubPages___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VWebSubPages___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< TWebSubPage >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< TWebSubPage >::__getitem__(std::vector< std::shared_ptr< CWebSubPage > >::difference_type) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::difference_type arg2 ;
  std::vector< std::shared_ptr< CWebSubPage > >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages___setitem__" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VWebSubPages___setitem__" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VWebSubPages___setitem__" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VWebSubPages___setitem__" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  arg3 = reinterpret_cast< std::vector< std::shared_ptr< CWebSubPage > >::value_type * >(argp3);
  try {
    std_vector_Sl_TWebSubPage_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(std::shared_ptr< CWebSubPage > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VWebSubPages___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_VWebSubPages___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VWebSubPages___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VWebSubPages___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VWebSubPages___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< TWebSubPage >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > const &)\n"
    "    std::vector< TWebSubPage >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
    "    std::vector< TWebSubPage >::__setitem__(std::vector< std::shared_ptr< CWebSubPage > >::difference_type,std::vector< std::shared_ptr< CWebSubPage > >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::shared_ptr< CWebSubPage > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_pop" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  try {
    result = std_vector_Sl_TWebSubPage_Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj((new std::vector< std::shared_ptr< CWebSubPage > >::value_type(result)), SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VWebSubPages_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_append" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VWebSubPages_append" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VWebSubPages_append" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< std::shared_ptr< CWebSubPage > >::value_type * >(argp2);
  std_vector_Sl_TWebSubPage_Sg__append(arg1,(std::shared_ptr< CWebSubPage > const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VWebSubPages__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *result = 0 ;
  
  (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< TWebSubPage > *)new std::vector< TWebSubPage >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_TWebSubPage_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VWebSubPages__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< TWebSubPage > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *ptr = (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_VWebSubPages" "', argument " "1"" of type '" "std::vector< TWebSubPage > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VWebSubPages" "', argument " "1"" of type '" "std::vector< TWebSubPage > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< TWebSubPage > *)new std::vector< TWebSubPage >((std::vector< TWebSubPage > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_TWebSubPage_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_empty" "', argument " "1"" of type '" "std::vector< TWebSubPage > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = (bool)((std::vector< TWebSubPage > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_size(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::shared_ptr< CWebSubPage > >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_size" "', argument " "1"" of type '" "std::vector< TWebSubPage > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = ((std::vector< TWebSubPage > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_swap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< TWebSubPage > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VWebSubPages_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_swap" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_TWebSubPage_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VWebSubPages_swap" "', argument " "2"" of type '" "std::vector< TWebSubPage > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VWebSubPages_swap" "', argument " "2"" of type '" "std::vector< TWebSubPage > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< TWebSubPage > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_begin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< std::shared_ptr< CWebSubPage > >::iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_begin" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::shared_ptr< CWebSubPage > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< std::shared_ptr< CWebSubPage > >::iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_end" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::shared_ptr< CWebSubPage > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_rbegin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< std::shared_ptr< CWebSubPage > >::reverse_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_rbegin" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::shared_ptr< CWebSubPage > >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_rend(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< std::shared_ptr< CWebSubPage > >::reverse_iterator > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_rend" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::shared_ptr< CWebSubPage > >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_clear" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_get_allocator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< std::shared_ptr< CWebSubPage > > > result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_get_allocator" "', argument " "1"" of type '" "std::vector< TWebSubPage > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = ((std::vector< TWebSubPage > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< std::shared_ptr< CWebSubPage > >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_std__shared_ptrT_CWebSubPage_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VWebSubPages__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::shared_ptr< CWebSubPage > >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< TWebSubPage > *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VWebSubPages" "', argument " "1"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::size_type""'");
  } 
  arg1 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::size_type >(val1);
  result = (std::vector< TWebSubPage > *)new std::vector< TWebSubPage >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_TWebSubPage_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_pop_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_pop_back" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_resize" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VWebSubPages_resize" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  SwigValueWrapper< std::vector< std::shared_ptr< CWebSubPage > >::iterator > arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  SwigValueWrapper< std::vector< std::shared_ptr< CWebSubPage > >::iterator > result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_erase" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages_erase" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages_erase" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::iterator""'");
    }
  }
  result = std_vector_Sl_TWebSubPage_Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::shared_ptr< CWebSubPage > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  SwigValueWrapper< std::vector< std::shared_ptr< CWebSubPage > >::iterator > arg2 ;
  SwigValueWrapper< std::vector< std::shared_ptr< CWebSubPage > >::iterator > arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  SwigValueWrapper< std::vector< std::shared_ptr< CWebSubPage > >::iterator > result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_erase" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages_erase" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages_erase" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages_erase" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages_erase" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::iterator""'");
    }
  }
  result = std_vector_Sl_TWebSubPage_Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::shared_ptr< CWebSubPage > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VWebSubPages_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_VWebSubPages_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_VWebSubPages_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VWebSubPages_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< TWebSubPage >::erase(std::vector< std::shared_ptr< CWebSubPage > >::iterator)\n"
    "    std::vector< TWebSubPage >::erase(std::vector< std::shared_ptr< CWebSubPage > >::iterator,std::vector< std::shared_ptr< CWebSubPage > >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_VWebSubPages__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::shared_ptr< CWebSubPage > >::size_type arg1 ;
  std::vector< std::shared_ptr< CWebSubPage > >::value_type *arg2 = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::vector< TWebSubPage > *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_VWebSubPages" "', argument " "1"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::size_type""'");
  } 
  arg1 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_VWebSubPages" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VWebSubPages" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< std::shared_ptr< CWebSubPage > >::value_type * >(argp2);
  result = (std::vector< TWebSubPage > *)new std::vector< TWebSubPage >(arg1,(std::vector< std::shared_ptr< CWebSubPage > >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_TWebSubPage_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_VWebSubPages(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_VWebSubPages", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_VWebSubPages__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_VWebSubPages__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_VWebSubPages__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_VWebSubPages__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_VWebSubPages'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< TWebSubPage >::vector()\n"
    "    std::vector< TWebSubPage >::vector(std::vector< TWebSubPage > const &)\n"
    "    std::vector< TWebSubPage >::vector(std::vector< std::shared_ptr< CWebSubPage > >::size_type)\n"
    "    std::vector< TWebSubPage >::vector(std::vector< std::shared_ptr< CWebSubPage > >::size_type,std::vector< std::shared_ptr< CWebSubPage > >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_push_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VWebSubPages_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_push_back" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VWebSubPages_push_back" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VWebSubPages_push_back" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< std::shared_ptr< CWebSubPage > >::value_type * >(argp2);
  (arg1)->push_back((std::vector< std::shared_ptr< CWebSubPage > >::value_type const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_front(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::shared_ptr< CWebSubPage > >::value_type *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_front" "', argument " "1"" of type '" "std::vector< TWebSubPage > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = (std::vector< std::shared_ptr< CWebSubPage > >::value_type *) &((std::vector< TWebSubPage > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t, 0 |  0 );
  (void)swig::container_owner<swig::traits<std::vector< std::shared_ptr< CWebSubPage > >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_back(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::shared_ptr< CWebSubPage > >::value_type *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_back" "', argument " "1"" of type '" "std::vector< TWebSubPage > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = (std::vector< std::shared_ptr< CWebSubPage > >::value_type *) &((std::vector< TWebSubPage > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t, 0 |  0 );
  (void)swig::container_owner<swig::traits<std::vector< std::shared_ptr< CWebSubPage > >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_assign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::size_type arg2 ;
  std::vector< std::shared_ptr< CWebSubPage > >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VWebSubPages_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_assign" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VWebSubPages_assign" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VWebSubPages_assign" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VWebSubPages_assign" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  arg3 = reinterpret_cast< std::vector< std::shared_ptr< CWebSubPage > >::value_type * >(argp3);
  (arg1)->assign(arg2,(std::vector< std::shared_ptr< CWebSubPage > >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::size_type arg2 ;
  std::vector< std::shared_ptr< CWebSubPage > >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_resize" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VWebSubPages_resize" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::size_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VWebSubPages_resize" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VWebSubPages_resize" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  arg3 = reinterpret_cast< std::vector< std::shared_ptr< CWebSubPage > >::value_type * >(argp3);
  (arg1)->resize(arg2,(std::vector< std::shared_ptr< CWebSubPage > >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VWebSubPages_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VWebSubPages_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VWebSubPages_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VWebSubPages_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< TWebSubPage >::resize(std::vector< std::shared_ptr< CWebSubPage > >::size_type)\n"
    "    std::vector< TWebSubPage >::resize(std::vector< std::shared_ptr< CWebSubPage > >::size_type,std::vector< std::shared_ptr< CWebSubPage > >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  SwigValueWrapper< std::vector< std::shared_ptr< CWebSubPage > >::iterator > arg2 ;
  std::vector< std::shared_ptr< CWebSubPage > >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  SwigValueWrapper< std::vector< std::shared_ptr< CWebSubPage > >::iterator > result;
  
  (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_insert" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages_insert" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages_insert" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VWebSubPages_insert" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VWebSubPages_insert" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  arg3 = reinterpret_cast< std::vector< std::shared_ptr< CWebSubPage > >::value_type * >(argp3);
  result = std_vector_Sl_TWebSubPage_Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),(std::shared_ptr< CWebSubPage > const &)*arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::shared_ptr< CWebSubPage > >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  SwigValueWrapper< std::vector< std::shared_ptr< CWebSubPage > >::iterator > arg2 ;
  std::vector< std::shared_ptr< CWebSubPage > >::size_type arg3 ;
  std::vector< std::shared_ptr< CWebSubPage > >::value_type *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  
  (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_insert" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages_insert" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "VWebSubPages_insert" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VWebSubPages_insert" "', argument " "3"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::size_type""'");
  } 
  arg3 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::size_type >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "VWebSubPages_insert" "', argument " "4"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VWebSubPages_insert" "', argument " "4"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::value_type const &""'"); 
  }
  arg4 = reinterpret_cast< std::vector< std::shared_ptr< CWebSubPage > >::value_type * >(argp4);
  std_vector_Sl_TWebSubPage_Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::shared_ptr< CWebSubPage > const &)*arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VWebSubPages_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *>(iter) != 0));
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_VWebSubPages_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v = 0;
    int res = swig::asptr(argv[0], (std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::shared_ptr< CWebSubPage > >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_ConvertPtr(argv[3], 0, SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t, SWIG_POINTER_NO_NULL | 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_VWebSubPages_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VWebSubPages_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< TWebSubPage >::insert(std::vector< std::shared_ptr< CWebSubPage > >::iterator,std::vector< std::shared_ptr< CWebSubPage > >::value_type const &)\n"
    "    std::vector< TWebSubPage >::insert(std::vector< std::shared_ptr< CWebSubPage > >::iterator,std::vector< std::shared_ptr< CWebSubPage > >::size_type,std::vector< std::shared_ptr< CWebSubPage > >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_reserve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  std::vector< std::shared_ptr< CWebSubPage > >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VWebSubPages_reserve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_reserve" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VWebSubPages_reserve" "', argument " "2"" of type '" "std::vector< std::shared_ptr< CWebSubPage > >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::shared_ptr< CWebSubPage > >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VWebSubPages_capacity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::shared_ptr< CWebSubPage > >::size_type result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VWebSubPages_capacity" "', argument " "1"" of type '" "std::vector< TWebSubPage > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  result = ((std::vector< TWebSubPage > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_VWebSubPages(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< TWebSubPage > *arg1 = (std::vector< TWebSubPage > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_TWebSubPage_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VWebSubPages" "', argument " "1"" of type '" "std::vector< TWebSubPage > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< TWebSubPage > * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *VWebSubPages_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_TWebSubPage_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *VWebSubPages_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_VPair_Add2Str_(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  VPair *arg1 = (VPair *) 0 ;
  CString *arg2 = 0 ;
  CString *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "VPair_Add2Str_", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_CString_CString_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VPair_Add2Str_" "', argument " "1"" of type '" "VPair *""'"); 
  }
  arg1 = reinterpret_cast< VPair * >(argp1);
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VPair_Add2Str_" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VPair_Add2Str_" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res3 = SWIG_AsPtr_CString(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VPair_Add2Str_" "', argument " "3"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VPair_Add2Str_" "', argument " "3"" of type '" "CString const &""'"); 
    }
    arg3 = ptr;
  }
  VPair_Add2Str_(arg1,(CString const &)*arg2,(CString const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CreateWebSubPage_(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  CString *arg1 = 0 ;
  CString *arg2 = 0 ;
  VPair *arg3 = 0 ;
  unsigned int arg4 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  SwigValueWrapper< std::shared_ptr< CWebSubPage > > result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CreateWebSubPage_", 4, 4, swig_obj)) SWIG_fail;
  {
    CString *ptr = (CString *)0;
    res1 = SWIG_AsPtr_CString(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CreateWebSubPage_" "', argument " "1"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CreateWebSubPage_" "', argument " "1"" of type '" "CString const &""'"); 
    }
    arg1 = ptr;
  }
  {
    CString *ptr = (CString *)0;
    res2 = SWIG_AsPtr_CString(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CreateWebSubPage_" "', argument " "2"" of type '" "CString const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CreateWebSubPage_" "', argument " "2"" of type '" "CString const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *ptr = (std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CreateWebSubPage_" "', argument " "3"" of type '" "VPair const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "CreateWebSubPage_" "', argument " "3"" of type '" "VPair const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CreateWebSubPage_" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  result = CreateWebSubPage_((CString const &)*arg1,(CString const &)*arg2,(std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > const &)*arg3,arg4);
  resultobj = SWIG_NewPointerObj((new TWebSubPage(result)), SWIGTYPE_p_std__shared_ptrT_CWebSubPage_t, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


static PyMethodDef SwigMethods[] = {
	 { "delete_SwigPyIterator", _wrap_delete_SwigPyIterator, METH_O, NULL},
	 { "SwigPyIterator_value", _wrap_SwigPyIterator_value, METH_O, NULL},
	 { "SwigPyIterator_incr", _wrap_SwigPyIterator_incr, METH_VARARGS, NULL},
	 { "SwigPyIterator_decr", _wrap_SwigPyIterator_decr, METH_VARARGS, NULL},
	 { "SwigPyIterator_distance", _wrap_SwigPyIterator_distance, METH_VARARGS, NULL},
	 { "SwigPyIterator_equal", _wrap_SwigPyIterator_equal, METH_VARARGS, NULL},
	 { "SwigPyIterator_copy", _wrap_SwigPyIterator_copy, METH_O, NULL},
	 { "SwigPyIterator_next", _wrap_SwigPyIterator_next, METH_O, NULL},
	 { "SwigPyIterator___next__", _wrap_SwigPyIterator___next__, METH_O, NULL},
	 { "SwigPyIterator_previous", _wrap_SwigPyIterator_previous, METH_O, NULL},
	 { "SwigPyIterator_advance", _wrap_SwigPyIterator_advance, METH_VARARGS, NULL},
	 { "SwigPyIterator___eq__", _wrap_SwigPyIterator___eq__, METH_VARARGS, NULL},
	 { "SwigPyIterator___ne__", _wrap_SwigPyIterator___ne__, METH_VARARGS, NULL},
	 { "SwigPyIterator___iadd__", _wrap_SwigPyIterator___iadd__, METH_VARARGS, NULL},
	 { "SwigPyIterator___isub__", _wrap_SwigPyIterator___isub__, METH_VARARGS, NULL},
	 { "SwigPyIterator___add__", _wrap_SwigPyIterator___add__, METH_VARARGS, NULL},
	 { "SwigPyIterator___sub__", _wrap_SwigPyIterator___sub__, METH_VARARGS, NULL},
	 { "SwigPyIterator_swigregister", SwigPyIterator_swigregister, METH_O, NULL},
	 { "_stringlist_iterator", _wrap__stringlist_iterator, METH_O, NULL},
	 { "_stringlist___nonzero__", _wrap__stringlist___nonzero__, METH_O, NULL},
	 { "_stringlist___bool__", _wrap__stringlist___bool__, METH_O, NULL},
	 { "_stringlist___len__", _wrap__stringlist___len__, METH_O, NULL},
	 { "_stringlist___getslice__", _wrap__stringlist___getslice__, METH_VARARGS, NULL},
	 { "_stringlist___setslice__", _wrap__stringlist___setslice__, METH_VARARGS, NULL},
	 { "_stringlist___delslice__", _wrap__stringlist___delslice__, METH_VARARGS, NULL},
	 { "_stringlist___delitem__", _wrap__stringlist___delitem__, METH_VARARGS, NULL},
	 { "_stringlist___getitem__", _wrap__stringlist___getitem__, METH_VARARGS, NULL},
	 { "_stringlist___setitem__", _wrap__stringlist___setitem__, METH_VARARGS, NULL},
	 { "_stringlist_pop", _wrap__stringlist_pop, METH_O, NULL},
	 { "_stringlist_append", _wrap__stringlist_append, METH_VARARGS, NULL},
	 { "_stringlist_empty", _wrap__stringlist_empty, METH_O, NULL},
	 { "_stringlist_size", _wrap__stringlist_size, METH_O, NULL},
	 { "_stringlist_swap", _wrap__stringlist_swap, METH_VARARGS, NULL},
	 { "_stringlist_begin", _wrap__stringlist_begin, METH_O, NULL},
	 { "_stringlist_end", _wrap__stringlist_end, METH_O, NULL},
	 { "_stringlist_rbegin", _wrap__stringlist_rbegin, METH_O, NULL},
	 { "_stringlist_rend", _wrap__stringlist_rend, METH_O, NULL},
	 { "_stringlist_clear", _wrap__stringlist_clear, METH_O, NULL},
	 { "_stringlist_get_allocator", _wrap__stringlist_get_allocator, METH_O, NULL},
	 { "_stringlist_pop_back", _wrap__stringlist_pop_back, METH_O, NULL},
	 { "_stringlist_erase", _wrap__stringlist_erase, METH_VARARGS, NULL},
	 { "new__stringlist", _wrap_new__stringlist, METH_VARARGS, NULL},
	 { "_stringlist_push_back", _wrap__stringlist_push_back, METH_VARARGS, NULL},
	 { "_stringlist_front", _wrap__stringlist_front, METH_O, NULL},
	 { "_stringlist_back", _wrap__stringlist_back, METH_O, NULL},
	 { "_stringlist_assign", _wrap__stringlist_assign, METH_VARARGS, NULL},
	 { "_stringlist_resize", _wrap__stringlist_resize, METH_VARARGS, NULL},
	 { "_stringlist_insert", _wrap__stringlist_insert, METH_VARARGS, NULL},
	 { "_stringlist_pop_front", _wrap__stringlist_pop_front, METH_O, NULL},
	 { "_stringlist_push_front", _wrap__stringlist_push_front, METH_VARARGS, NULL},
	 { "_stringlist_reverse", _wrap__stringlist_reverse, METH_O, NULL},
	 { "delete__stringlist", _wrap_delete__stringlist, METH_O, NULL},
	 { "_stringlist_swigregister", _stringlist_swigregister, METH_O, NULL},
	 { "_stringlist_swiginit", _stringlist_swiginit, METH_VARARGS, NULL},
	 { "VIRCNetworks_iterator", _wrap_VIRCNetworks_iterator, METH_O, NULL},
	 { "VIRCNetworks___nonzero__", _wrap_VIRCNetworks___nonzero__, METH_O, NULL},
	 { "VIRCNetworks___bool__", _wrap_VIRCNetworks___bool__, METH_O, NULL},
	 { "VIRCNetworks___len__", _wrap_VIRCNetworks___len__, METH_O, NULL},
	 { "VIRCNetworks___getslice__", _wrap_VIRCNetworks___getslice__, METH_VARARGS, NULL},
	 { "VIRCNetworks___setslice__", _wrap_VIRCNetworks___setslice__, METH_VARARGS, NULL},
	 { "VIRCNetworks___delslice__", _wrap_VIRCNetworks___delslice__, METH_VARARGS, NULL},
	 { "VIRCNetworks___delitem__", _wrap_VIRCNetworks___delitem__, METH_VARARGS, NULL},
	 { "VIRCNetworks___getitem__", _wrap_VIRCNetworks___getitem__, METH_VARARGS, NULL},
	 { "VIRCNetworks___setitem__", _wrap_VIRCNetworks___setitem__, METH_VARARGS, NULL},
	 { "VIRCNetworks_pop", _wrap_VIRCNetworks_pop, METH_O, NULL},
	 { "VIRCNetworks_append", _wrap_VIRCNetworks_append, METH_VARARGS, NULL},
	 { "VIRCNetworks_empty", _wrap_VIRCNetworks_empty, METH_O, NULL},
	 { "VIRCNetworks_size", _wrap_VIRCNetworks_size, METH_O, NULL},
	 { "VIRCNetworks_swap", _wrap_VIRCNetworks_swap, METH_VARARGS, NULL},
	 { "VIRCNetworks_begin", _wrap_VIRCNetworks_begin, METH_O, NULL},
	 { "VIRCNetworks_end", _wrap_VIRCNetworks_end, METH_O, NULL},
	 { "VIRCNetworks_rbegin", _wrap_VIRCNetworks_rbegin, METH_O, NULL},
	 { "VIRCNetworks_rend", _wrap_VIRCNetworks_rend, METH_O, NULL},
	 { "VIRCNetworks_clear", _wrap_VIRCNetworks_clear, METH_O, NULL},
	 { "VIRCNetworks_get_allocator", _wrap_VIRCNetworks_get_allocator, METH_O, NULL},
	 { "VIRCNetworks_pop_back", _wrap_VIRCNetworks_pop_back, METH_O, NULL},
	 { "VIRCNetworks_erase", _wrap_VIRCNetworks_erase, METH_VARARGS, NULL},
	 { "new_VIRCNetworks", _wrap_new_VIRCNetworks, METH_VARARGS, NULL},
	 { "VIRCNetworks_push_back", _wrap_VIRCNetworks_push_back, METH_VARARGS, NULL},
	 { "VIRCNetworks_front", _wrap_VIRCNetworks_front, METH_O, NULL},
	 { "VIRCNetworks_back", _wrap_VIRCNetworks_back, METH_O, NULL},
	 { "VIRCNetworks_assign", _wrap_VIRCNetworks_assign, METH_VARARGS, NULL},
	 { "VIRCNetworks_resize", _wrap_VIRCNetworks_resize, METH_VARARGS, NULL},
	 { "VIRCNetworks_insert", _wrap_VIRCNetworks_insert, METH_VARARGS, NULL},
	 { "VIRCNetworks_reserve", _wrap_VIRCNetworks_reserve, METH_VARARGS, NULL},
	 { "VIRCNetworks_capacity", _wrap_VIRCNetworks_capacity, METH_O, NULL},
	 { "delete_VIRCNetworks", _wrap_delete_VIRCNetworks, METH_O, NULL},
	 { "VIRCNetworks_swigregister", VIRCNetworks_swigregister, METH_O, NULL},
	 { "VIRCNetworks_swiginit", VIRCNetworks_swiginit, METH_VARARGS, NULL},
	 { "VChannels_iterator", _wrap_VChannels_iterator, METH_O, NULL},
	 { "VChannels___nonzero__", _wrap_VChannels___nonzero__, METH_O, NULL},
	 { "VChannels___bool__", _wrap_VChannels___bool__, METH_O, NULL},
	 { "VChannels___len__", _wrap_VChannels___len__, METH_O, NULL},
	 { "VChannels___getslice__", _wrap_VChannels___getslice__, METH_VARARGS, NULL},
	 { "VChannels___setslice__", _wrap_VChannels___setslice__, METH_VARARGS, NULL},
	 { "VChannels___delslice__", _wrap_VChannels___delslice__, METH_VARARGS, NULL},
	 { "VChannels___delitem__", _wrap_VChannels___delitem__, METH_VARARGS, NULL},
	 { "VChannels___getitem__", _wrap_VChannels___getitem__, METH_VARARGS, NULL},
	 { "VChannels___setitem__", _wrap_VChannels___setitem__, METH_VARARGS, NULL},
	 { "VChannels_pop", _wrap_VChannels_pop, METH_O, NULL},
	 { "VChannels_append", _wrap_VChannels_append, METH_VARARGS, NULL},
	 { "VChannels_empty", _wrap_VChannels_empty, METH_O, NULL},
	 { "VChannels_size", _wrap_VChannels_size, METH_O, NULL},
	 { "VChannels_swap", _wrap_VChannels_swap, METH_VARARGS, NULL},
	 { "VChannels_begin", _wrap_VChannels_begin, METH_O, NULL},
	 { "VChannels_end", _wrap_VChannels_end, METH_O, NULL},
	 { "VChannels_rbegin", _wrap_VChannels_rbegin, METH_O, NULL},
	 { "VChannels_rend", _wrap_VChannels_rend, METH_O, NULL},
	 { "VChannels_clear", _wrap_VChannels_clear, METH_O, NULL},
	 { "VChannels_get_allocator", _wrap_VChannels_get_allocator, METH_O, NULL},
	 { "VChannels_pop_back", _wrap_VChannels_pop_back, METH_O, NULL},
	 { "VChannels_erase", _wrap_VChannels_erase, METH_VARARGS, NULL},
	 { "new_VChannels", _wrap_new_VChannels, METH_VARARGS, NULL},
	 { "VChannels_push_back", _wrap_VChannels_push_back, METH_VARARGS, NULL},
	 { "VChannels_front", _wrap_VChannels_front, METH_O, NULL},
	 { "VChannels_back", _wrap_VChannels_back, METH_O, NULL},
	 { "VChannels_assign", _wrap_VChannels_assign, METH_VARARGS, NULL},
	 { "VChannels_resize", _wrap_VChannels_resize, METH_VARARGS, NULL},
	 { "VChannels_insert", _wrap_VChannels_insert, METH_VARARGS, NULL},
	 { "VChannels_reserve", _wrap_VChannels_reserve, METH_VARARGS, NULL},
	 { "VChannels_capacity", _wrap_VChannels_capacity, METH_O, NULL},
	 { "delete_VChannels", _wrap_delete_VChannels, METH_O, NULL},
	 { "VChannels_swigregister", VChannels_swigregister, METH_O, NULL},
	 { "VChannels_swiginit", VChannels_swiginit, METH_VARARGS, NULL},
	 { "MNicks_iterator", _wrap_MNicks_iterator, METH_O, NULL},
	 { "MNicks___nonzero__", _wrap_MNicks___nonzero__, METH_O, NULL},
	 { "MNicks___bool__", _wrap_MNicks___bool__, METH_O, NULL},
	 { "MNicks___len__", _wrap_MNicks___len__, METH_O, NULL},
	 { "MNicks___getitem__", _wrap_MNicks___getitem__, METH_VARARGS, NULL},
	 { "MNicks___delitem__", _wrap_MNicks___delitem__, METH_VARARGS, NULL},
	 { "MNicks_has_key", _wrap_MNicks_has_key, METH_VARARGS, NULL},
	 { "MNicks_keys", _wrap_MNicks_keys, METH_O, NULL},
	 { "MNicks_values", _wrap_MNicks_values, METH_O, NULL},
	 { "MNicks_items", _wrap_MNicks_items, METH_O, NULL},
	 { "MNicks___contains__", _wrap_MNicks___contains__, METH_VARARGS, NULL},
	 { "MNicks_key_iterator", _wrap_MNicks_key_iterator, METH_O, NULL},
	 { "MNicks_value_iterator", _wrap_MNicks_value_iterator, METH_O, NULL},
	 { "MNicks___setitem__", _wrap_MNicks___setitem__, METH_VARARGS, NULL},
	 { "MNicks_asdict", _wrap_MNicks_asdict, METH_O, NULL},
	 { "new_MNicks", _wrap_new_MNicks, METH_VARARGS, NULL},
	 { "MNicks_empty", _wrap_MNicks_empty, METH_O, NULL},
	 { "MNicks_size", _wrap_MNicks_size, METH_O, NULL},
	 { "MNicks_swap", _wrap_MNicks_swap, METH_VARARGS, NULL},
	 { "MNicks_begin", _wrap_MNicks_begin, METH_O, NULL},
	 { "MNicks_end", _wrap_MNicks_end, METH_O, NULL},
	 { "MNicks_rbegin", _wrap_MNicks_rbegin, METH_O, NULL},
	 { "MNicks_rend", _wrap_MNicks_rend, METH_O, NULL},
	 { "MNicks_clear", _wrap_MNicks_clear, METH_O, NULL},
	 { "MNicks_get_allocator", _wrap_MNicks_get_allocator, METH_O, NULL},
	 { "MNicks_count", _wrap_MNicks_count, METH_VARARGS, NULL},
	 { "MNicks_erase", _wrap_MNicks_erase, METH_VARARGS, NULL},
	 { "MNicks_find", _wrap_MNicks_find, METH_VARARGS, NULL},
	 { "MNicks_lower_bound", _wrap_MNicks_lower_bound, METH_VARARGS, NULL},
	 { "MNicks_upper_bound", _wrap_MNicks_upper_bound, METH_VARARGS, NULL},
	 { "delete_MNicks", _wrap_delete_MNicks, METH_O, NULL},
	 { "MNicks_swigregister", MNicks_swigregister, METH_O, NULL},
	 { "MNicks_swiginit", MNicks_swiginit, METH_VARARGS, NULL},
	 { "SModInfo_iterator", _wrap_SModInfo_iterator, METH_O, NULL},
	 { "SModInfo___nonzero__", _wrap_SModInfo___nonzero__, METH_O, NULL},
	 { "SModInfo___bool__", _wrap_SModInfo___bool__, METH_O, NULL},
	 { "SModInfo___len__", _wrap_SModInfo___len__, METH_O, NULL},
	 { "SModInfo_append", _wrap_SModInfo_append, METH_VARARGS, NULL},
	 { "SModInfo___contains__", _wrap_SModInfo___contains__, METH_VARARGS, NULL},
	 { "SModInfo___getitem__", _wrap_SModInfo___getitem__, METH_VARARGS, NULL},
	 { "SModInfo_add", _wrap_SModInfo_add, METH_VARARGS, NULL},
	 { "SModInfo_discard", _wrap_SModInfo_discard, METH_VARARGS, NULL},
	 { "new_SModInfo", _wrap_new_SModInfo, METH_VARARGS, NULL},
	 { "SModInfo_empty", _wrap_SModInfo_empty, METH_O, NULL},
	 { "SModInfo_size", _wrap_SModInfo_size, METH_O, NULL},
	 { "SModInfo_clear", _wrap_SModInfo_clear, METH_O, NULL},
	 { "SModInfo_swap", _wrap_SModInfo_swap, METH_VARARGS, NULL},
	 { "SModInfo_count", _wrap_SModInfo_count, METH_VARARGS, NULL},
	 { "SModInfo_begin", _wrap_SModInfo_begin, METH_O, NULL},
	 { "SModInfo_end", _wrap_SModInfo_end, METH_O, NULL},
	 { "SModInfo_rbegin", _wrap_SModInfo_rbegin, METH_O, NULL},
	 { "SModInfo_rend", _wrap_SModInfo_rend, METH_O, NULL},
	 { "SModInfo_erase", _wrap_SModInfo_erase, METH_VARARGS, NULL},
	 { "SModInfo_find", _wrap_SModInfo_find, METH_VARARGS, NULL},
	 { "SModInfo_lower_bound", _wrap_SModInfo_lower_bound, METH_VARARGS, NULL},
	 { "SModInfo_upper_bound", _wrap_SModInfo_upper_bound, METH_VARARGS, NULL},
	 { "SModInfo_equal_range", _wrap_SModInfo_equal_range, METH_VARARGS, NULL},
	 { "SModInfo_insert", _wrap_SModInfo_insert, METH_VARARGS, NULL},
	 { "delete_SModInfo", _wrap_delete_SModInfo, METH_O, NULL},
	 { "SModInfo_swigregister", SModInfo_swigregister, METH_O, NULL},
	 { "SModInfo_swiginit", SModInfo_swiginit, METH_VARARGS, NULL},
	 { "SCString_iterator", _wrap_SCString_iterator, METH_O, NULL},
	 { "SCString___nonzero__", _wrap_SCString___nonzero__, METH_O, NULL},
	 { "SCString___bool__", _wrap_SCString___bool__, METH_O, NULL},
	 { "SCString___len__", _wrap_SCString___len__, METH_O, NULL},
	 { "SCString_append", _wrap_SCString_append, METH_VARARGS, NULL},
	 { "SCString___contains__", _wrap_SCString___contains__, METH_VARARGS, NULL},
	 { "SCString___getitem__", _wrap_SCString___getitem__, METH_VARARGS, NULL},
	 { "SCString_add", _wrap_SCString_add, METH_VARARGS, NULL},
	 { "SCString_discard", _wrap_SCString_discard, METH_VARARGS, NULL},
	 { "new_SCString", _wrap_new_SCString, METH_VARARGS, NULL},
	 { "SCString_empty", _wrap_SCString_empty, METH_O, NULL},
	 { "SCString_size", _wrap_SCString_size, METH_O, NULL},
	 { "SCString_clear", _wrap_SCString_clear, METH_O, NULL},
	 { "SCString_swap", _wrap_SCString_swap, METH_VARARGS, NULL},
	 { "SCString_count", _wrap_SCString_count, METH_VARARGS, NULL},
	 { "SCString_begin", _wrap_SCString_begin, METH_O, NULL},
	 { "SCString_end", _wrap_SCString_end, METH_O, NULL},
	 { "SCString_rbegin", _wrap_SCString_rbegin, METH_O, NULL},
	 { "SCString_rend", _wrap_SCString_rend, METH_O, NULL},
	 { "SCString_erase", _wrap_SCString_erase, METH_VARARGS, NULL},
	 { "SCString_find", _wrap_SCString_find, METH_VARARGS, NULL},
	 { "SCString_lower_bound", _wrap_SCString_lower_bound, METH_VARARGS, NULL},
	 { "SCString_upper_bound", _wrap_SCString_upper_bound, METH_VARARGS, NULL},
	 { "SCString_equal_range", _wrap_SCString_equal_range, METH_VARARGS, NULL},
	 { "SCString_insert", _wrap_SCString_insert, METH_VARARGS, NULL},
	 { "delete_SCString", _wrap_delete_SCString, METH_O, NULL},
	 { "SCString_swigregister", SCString_swigregister, METH_O, NULL},
	 { "SCString_swiginit", SCString_swiginit, METH_VARARGS, NULL},
	 { "VCString_iterator", _wrap_VCString_iterator, METH_O, NULL},
	 { "VCString___nonzero__", _wrap_VCString___nonzero__, METH_O, NULL},
	 { "VCString___bool__", _wrap_VCString___bool__, METH_O, NULL},
	 { "VCString___len__", _wrap_VCString___len__, METH_O, NULL},
	 { "VCString___getslice__", _wrap_VCString___getslice__, METH_VARARGS, NULL},
	 { "VCString___setslice__", _wrap_VCString___setslice__, METH_VARARGS, NULL},
	 { "VCString___delslice__", _wrap_VCString___delslice__, METH_VARARGS, NULL},
	 { "VCString___delitem__", _wrap_VCString___delitem__, METH_VARARGS, NULL},
	 { "VCString___getitem__", _wrap_VCString___getitem__, METH_VARARGS, NULL},
	 { "VCString___setitem__", _wrap_VCString___setitem__, METH_VARARGS, NULL},
	 { "VCString_pop", _wrap_VCString_pop, METH_O, NULL},
	 { "VCString_append", _wrap_VCString_append, METH_VARARGS, NULL},
	 { "VCString_empty", _wrap_VCString_empty, METH_O, NULL},
	 { "VCString_size", _wrap_VCString_size, METH_O, NULL},
	 { "VCString_swap", _wrap_VCString_swap, METH_VARARGS, NULL},
	 { "VCString_begin", _wrap_VCString_begin, METH_O, NULL},
	 { "VCString_end", _wrap_VCString_end, METH_O, NULL},
	 { "VCString_rbegin", _wrap_VCString_rbegin, METH_O, NULL},
	 { "VCString_rend", _wrap_VCString_rend, METH_O, NULL},
	 { "VCString_clear", _wrap_VCString_clear, METH_O, NULL},
	 { "VCString_get_allocator", _wrap_VCString_get_allocator, METH_O, NULL},
	 { "VCString_pop_back", _wrap_VCString_pop_back, METH_O, NULL},
	 { "VCString_erase", _wrap_VCString_erase, METH_VARARGS, NULL},
	 { "new_VCString", _wrap_new_VCString, METH_VARARGS, NULL},
	 { "VCString_push_back", _wrap_VCString_push_back, METH_VARARGS, NULL},
	 { "VCString_front", _wrap_VCString_front, METH_O, NULL},
	 { "VCString_back", _wrap_VCString_back, METH_O, NULL},
	 { "VCString_assign", _wrap_VCString_assign, METH_VARARGS, NULL},
	 { "VCString_resize", _wrap_VCString_resize, METH_VARARGS, NULL},
	 { "VCString_insert", _wrap_VCString_insert, METH_VARARGS, NULL},
	 { "VCString_reserve", _wrap_VCString_reserve, METH_VARARGS, NULL},
	 { "VCString_capacity", _wrap_VCString_capacity, METH_O, NULL},
	 { "delete_VCString", _wrap_delete_VCString, METH_O, NULL},
	 { "VCString_swigregister", VCString_swigregister, METH_O, NULL},
	 { "VCString_swiginit", VCString_swiginit, METH_VARARGS, NULL},
	 { "PyMCString_iterator", _wrap_PyMCString_iterator, METH_O, NULL},
	 { "PyMCString___nonzero__", _wrap_PyMCString___nonzero__, METH_O, NULL},
	 { "PyMCString___bool__", _wrap_PyMCString___bool__, METH_O, NULL},
	 { "PyMCString___len__", _wrap_PyMCString___len__, METH_O, NULL},
	 { "PyMCString___getitem__", _wrap_PyMCString___getitem__, METH_VARARGS, NULL},
	 { "PyMCString___delitem__", _wrap_PyMCString___delitem__, METH_VARARGS, NULL},
	 { "PyMCString_has_key", _wrap_PyMCString_has_key, METH_VARARGS, NULL},
	 { "PyMCString_keys", _wrap_PyMCString_keys, METH_O, NULL},
	 { "PyMCString_values", _wrap_PyMCString_values, METH_O, NULL},
	 { "PyMCString_items", _wrap_PyMCString_items, METH_O, NULL},
	 { "PyMCString___contains__", _wrap_PyMCString___contains__, METH_VARARGS, NULL},
	 { "PyMCString_key_iterator", _wrap_PyMCString_key_iterator, METH_O, NULL},
	 { "PyMCString_value_iterator", _wrap_PyMCString_value_iterator, METH_O, NULL},
	 { "PyMCString___setitem__", _wrap_PyMCString___setitem__, METH_VARARGS, NULL},
	 { "PyMCString_asdict", _wrap_PyMCString_asdict, METH_O, NULL},
	 { "new_PyMCString", _wrap_new_PyMCString, METH_VARARGS, NULL},
	 { "PyMCString_empty", _wrap_PyMCString_empty, METH_O, NULL},
	 { "PyMCString_size", _wrap_PyMCString_size, METH_O, NULL},
	 { "PyMCString_swap", _wrap_PyMCString_swap, METH_VARARGS, NULL},
	 { "PyMCString_begin", _wrap_PyMCString_begin, METH_O, NULL},
	 { "PyMCString_end", _wrap_PyMCString_end, METH_O, NULL},
	 { "PyMCString_rbegin", _wrap_PyMCString_rbegin, METH_O, NULL},
	 { "PyMCString_rend", _wrap_PyMCString_rend, METH_O, NULL},
	 { "PyMCString_clear", _wrap_PyMCString_clear, METH_O, NULL},
	 { "PyMCString_get_allocator", _wrap_PyMCString_get_allocator, METH_O, NULL},
	 { "PyMCString_count", _wrap_PyMCString_count, METH_VARARGS, NULL},
	 { "PyMCString_erase", _wrap_PyMCString_erase, METH_VARARGS, NULL},
	 { "PyMCString_find", _wrap_PyMCString_find, METH_VARARGS, NULL},
	 { "PyMCString_lower_bound", _wrap_PyMCString_lower_bound, METH_VARARGS, NULL},
	 { "PyMCString_upper_bound", _wrap_PyMCString_upper_bound, METH_VARARGS, NULL},
	 { "delete_PyMCString", _wrap_delete_PyMCString, METH_O, NULL},
	 { "PyMCString_swigregister", PyMCString_swigregister, METH_O, NULL},
	 { "PyMCString_swiginit", PyMCString_swiginit, METH_VARARGS, NULL},
	 { "PyMStringVString_iterator", _wrap_PyMStringVString_iterator, METH_O, NULL},
	 { "PyMStringVString___nonzero__", _wrap_PyMStringVString___nonzero__, METH_O, NULL},
	 { "PyMStringVString___bool__", _wrap_PyMStringVString___bool__, METH_O, NULL},
	 { "PyMStringVString___len__", _wrap_PyMStringVString___len__, METH_O, NULL},
	 { "PyMStringVString___getitem__", _wrap_PyMStringVString___getitem__, METH_VARARGS, NULL},
	 { "PyMStringVString___delitem__", _wrap_PyMStringVString___delitem__, METH_VARARGS, NULL},
	 { "PyMStringVString_has_key", _wrap_PyMStringVString_has_key, METH_VARARGS, NULL},
	 { "PyMStringVString_keys", _wrap_PyMStringVString_keys, METH_O, NULL},
	 { "PyMStringVString_values", _wrap_PyMStringVString_values, METH_O, NULL},
	 { "PyMStringVString_items", _wrap_PyMStringVString_items, METH_O, NULL},
	 { "PyMStringVString___contains__", _wrap_PyMStringVString___contains__, METH_VARARGS, NULL},
	 { "PyMStringVString_key_iterator", _wrap_PyMStringVString_key_iterator, METH_O, NULL},
	 { "PyMStringVString_value_iterator", _wrap_PyMStringVString_value_iterator, METH_O, NULL},
	 { "PyMStringVString___setitem__", _wrap_PyMStringVString___setitem__, METH_VARARGS, NULL},
	 { "PyMStringVString_asdict", _wrap_PyMStringVString_asdict, METH_O, NULL},
	 { "new_PyMStringVString", _wrap_new_PyMStringVString, METH_VARARGS, NULL},
	 { "PyMStringVString_empty", _wrap_PyMStringVString_empty, METH_O, NULL},
	 { "PyMStringVString_size", _wrap_PyMStringVString_size, METH_O, NULL},
	 { "PyMStringVString_swap", _wrap_PyMStringVString_swap, METH_VARARGS, NULL},
	 { "PyMStringVString_begin", _wrap_PyMStringVString_begin, METH_O, NULL},
	 { "PyMStringVString_end", _wrap_PyMStringVString_end, METH_O, NULL},
	 { "PyMStringVString_rbegin", _wrap_PyMStringVString_rbegin, METH_O, NULL},
	 { "PyMStringVString_rend", _wrap_PyMStringVString_rend, METH_O, NULL},
	 { "PyMStringVString_clear", _wrap_PyMStringVString_clear, METH_O, NULL},
	 { "PyMStringVString_get_allocator", _wrap_PyMStringVString_get_allocator, METH_O, NULL},
	 { "PyMStringVString_count", _wrap_PyMStringVString_count, METH_VARARGS, NULL},
	 { "PyMStringVString_erase", _wrap_PyMStringVString_erase, METH_VARARGS, NULL},
	 { "PyMStringVString_find", _wrap_PyMStringVString_find, METH_VARARGS, NULL},
	 { "PyMStringVString_lower_bound", _wrap_PyMStringVString_lower_bound, METH_VARARGS, NULL},
	 { "PyMStringVString_upper_bound", _wrap_PyMStringVString_upper_bound, METH_VARARGS, NULL},
	 { "delete_PyMStringVString", _wrap_delete_PyMStringVString, METH_O, NULL},
	 { "PyMStringVString_swigregister", PyMStringVString_swigregister, METH_O, NULL},
	 { "PyMStringVString_swiginit", PyMStringVString_swiginit, METH_VARARGS, NULL},
	 { "new_MCString", _wrap_new_MCString, METH_NOARGS, NULL},
	 { "delete_MCString", _wrap_delete_MCString, METH_O, NULL},
	 { "MCString_swigregister", MCString_swigregister, METH_O, NULL},
	 { "MCString_swiginit", MCString_swiginit, METH_VARARGS, NULL},
	 { "PyModulesVector_iterator", _wrap_PyModulesVector_iterator, METH_O, NULL},
	 { "PyModulesVector___nonzero__", _wrap_PyModulesVector___nonzero__, METH_O, NULL},
	 { "PyModulesVector___bool__", _wrap_PyModulesVector___bool__, METH_O, NULL},
	 { "PyModulesVector___len__", _wrap_PyModulesVector___len__, METH_O, NULL},
	 { "PyModulesVector___getslice__", _wrap_PyModulesVector___getslice__, METH_VARARGS, NULL},
	 { "PyModulesVector___setslice__", _wrap_PyModulesVector___setslice__, METH_VARARGS, NULL},
	 { "PyModulesVector___delslice__", _wrap_PyModulesVector___delslice__, METH_VARARGS, NULL},
	 { "PyModulesVector___delitem__", _wrap_PyModulesVector___delitem__, METH_VARARGS, NULL},
	 { "PyModulesVector___getitem__", _wrap_PyModulesVector___getitem__, METH_VARARGS, NULL},
	 { "PyModulesVector___setitem__", _wrap_PyModulesVector___setitem__, METH_VARARGS, NULL},
	 { "PyModulesVector_pop", _wrap_PyModulesVector_pop, METH_O, NULL},
	 { "PyModulesVector_append", _wrap_PyModulesVector_append, METH_VARARGS, NULL},
	 { "PyModulesVector_empty", _wrap_PyModulesVector_empty, METH_O, NULL},
	 { "PyModulesVector_size", _wrap_PyModulesVector_size, METH_O, NULL},
	 { "PyModulesVector_swap", _wrap_PyModulesVector_swap, METH_VARARGS, NULL},
	 { "PyModulesVector_begin", _wrap_PyModulesVector_begin, METH_O, NULL},
	 { "PyModulesVector_end", _wrap_PyModulesVector_end, METH_O, NULL},
	 { "PyModulesVector_rbegin", _wrap_PyModulesVector_rbegin, METH_O, NULL},
	 { "PyModulesVector_rend", _wrap_PyModulesVector_rend, METH_O, NULL},
	 { "PyModulesVector_clear", _wrap_PyModulesVector_clear, METH_O, NULL},
	 { "PyModulesVector_get_allocator", _wrap_PyModulesVector_get_allocator, METH_O, NULL},
	 { "PyModulesVector_pop_back", _wrap_PyModulesVector_pop_back, METH_O, NULL},
	 { "PyModulesVector_erase", _wrap_PyModulesVector_erase, METH_VARARGS, NULL},
	 { "new_PyModulesVector", _wrap_new_PyModulesVector, METH_VARARGS, NULL},
	 { "PyModulesVector_push_back", _wrap_PyModulesVector_push_back, METH_VARARGS, NULL},
	 { "PyModulesVector_front", _wrap_PyModulesVector_front, METH_O, NULL},
	 { "PyModulesVector_back", _wrap_PyModulesVector_back, METH_O, NULL},
	 { "PyModulesVector_assign", _wrap_PyModulesVector_assign, METH_VARARGS, NULL},
	 { "PyModulesVector_resize", _wrap_PyModulesVector_resize, METH_VARARGS, NULL},
	 { "PyModulesVector_insert", _wrap_PyModulesVector_insert, METH_VARARGS, NULL},
	 { "PyModulesVector_reserve", _wrap_PyModulesVector_reserve, METH_VARARGS, NULL},
	 { "PyModulesVector_capacity", _wrap_PyModulesVector_capacity, METH_O, NULL},
	 { "delete_PyModulesVector", _wrap_delete_PyModulesVector, METH_O, NULL},
	 { "PyModulesVector_swigregister", PyModulesVector_swigregister, METH_O, NULL},
	 { "PyModulesVector_swiginit", PyModulesVector_swiginit, METH_VARARGS, NULL},
	 { "VListeners_iterator", _wrap_VListeners_iterator, METH_O, NULL},
	 { "VListeners___nonzero__", _wrap_VListeners___nonzero__, METH_O, NULL},
	 { "VListeners___bool__", _wrap_VListeners___bool__, METH_O, NULL},
	 { "VListeners___len__", _wrap_VListeners___len__, METH_O, NULL},
	 { "VListeners___getslice__", _wrap_VListeners___getslice__, METH_VARARGS, NULL},
	 { "VListeners___setslice__", _wrap_VListeners___setslice__, METH_VARARGS, NULL},
	 { "VListeners___delslice__", _wrap_VListeners___delslice__, METH_VARARGS, NULL},
	 { "VListeners___delitem__", _wrap_VListeners___delitem__, METH_VARARGS, NULL},
	 { "VListeners___getitem__", _wrap_VListeners___getitem__, METH_VARARGS, NULL},
	 { "VListeners___setitem__", _wrap_VListeners___setitem__, METH_VARARGS, NULL},
	 { "VListeners_pop", _wrap_VListeners_pop, METH_O, NULL},
	 { "VListeners_append", _wrap_VListeners_append, METH_VARARGS, NULL},
	 { "VListeners_empty", _wrap_VListeners_empty, METH_O, NULL},
	 { "VListeners_size", _wrap_VListeners_size, METH_O, NULL},
	 { "VListeners_swap", _wrap_VListeners_swap, METH_VARARGS, NULL},
	 { "VListeners_begin", _wrap_VListeners_begin, METH_O, NULL},
	 { "VListeners_end", _wrap_VListeners_end, METH_O, NULL},
	 { "VListeners_rbegin", _wrap_VListeners_rbegin, METH_O, NULL},
	 { "VListeners_rend", _wrap_VListeners_rend, METH_O, NULL},
	 { "VListeners_clear", _wrap_VListeners_clear, METH_O, NULL},
	 { "VListeners_get_allocator", _wrap_VListeners_get_allocator, METH_O, NULL},
	 { "VListeners_pop_back", _wrap_VListeners_pop_back, METH_O, NULL},
	 { "VListeners_erase", _wrap_VListeners_erase, METH_VARARGS, NULL},
	 { "new_VListeners", _wrap_new_VListeners, METH_VARARGS, NULL},
	 { "VListeners_push_back", _wrap_VListeners_push_back, METH_VARARGS, NULL},
	 { "VListeners_front", _wrap_VListeners_front, METH_O, NULL},
	 { "VListeners_back", _wrap_VListeners_back, METH_O, NULL},
	 { "VListeners_assign", _wrap_VListeners_assign, METH_VARARGS, NULL},
	 { "VListeners_resize", _wrap_VListeners_resize, METH_VARARGS, NULL},
	 { "VListeners_insert", _wrap_VListeners_insert, METH_VARARGS, NULL},
	 { "VListeners_reserve", _wrap_VListeners_reserve, METH_VARARGS, NULL},
	 { "VListeners_capacity", _wrap_VListeners_capacity, METH_O, NULL},
	 { "delete_VListeners", _wrap_delete_VListeners, METH_O, NULL},
	 { "VListeners_swigregister", VListeners_swigregister, METH_O, NULL},
	 { "VListeners_swiginit", VListeners_swiginit, METH_VARARGS, NULL},
	 { "BufLines_iterator", _wrap_BufLines_iterator, METH_O, NULL},
	 { "BufLines___nonzero__", _wrap_BufLines___nonzero__, METH_O, NULL},
	 { "BufLines___bool__", _wrap_BufLines___bool__, METH_O, NULL},
	 { "BufLines___len__", _wrap_BufLines___len__, METH_O, NULL},
	 { "BufLines___getslice__", _wrap_BufLines___getslice__, METH_VARARGS, NULL},
	 { "BufLines___setslice__", _wrap_BufLines___setslice__, METH_VARARGS, NULL},
	 { "BufLines___delslice__", _wrap_BufLines___delslice__, METH_VARARGS, NULL},
	 { "BufLines___delitem__", _wrap_BufLines___delitem__, METH_VARARGS, NULL},
	 { "BufLines___getitem__", _wrap_BufLines___getitem__, METH_VARARGS, NULL},
	 { "BufLines___setitem__", _wrap_BufLines___setitem__, METH_VARARGS, NULL},
	 { "BufLines_pop", _wrap_BufLines_pop, METH_O, NULL},
	 { "BufLines_append", _wrap_BufLines_append, METH_VARARGS, NULL},
	 { "BufLines_empty", _wrap_BufLines_empty, METH_O, NULL},
	 { "BufLines_size", _wrap_BufLines_size, METH_O, NULL},
	 { "BufLines_swap", _wrap_BufLines_swap, METH_VARARGS, NULL},
	 { "BufLines_begin", _wrap_BufLines_begin, METH_O, NULL},
	 { "BufLines_end", _wrap_BufLines_end, METH_O, NULL},
	 { "BufLines_rbegin", _wrap_BufLines_rbegin, METH_O, NULL},
	 { "BufLines_rend", _wrap_BufLines_rend, METH_O, NULL},
	 { "BufLines_clear", _wrap_BufLines_clear, METH_O, NULL},
	 { "BufLines_get_allocator", _wrap_BufLines_get_allocator, METH_O, NULL},
	 { "BufLines_pop_back", _wrap_BufLines_pop_back, METH_O, NULL},
	 { "BufLines_erase", _wrap_BufLines_erase, METH_VARARGS, NULL},
	 { "new_BufLines", _wrap_new_BufLines, METH_VARARGS, NULL},
	 { "BufLines_push_back", _wrap_BufLines_push_back, METH_VARARGS, NULL},
	 { "BufLines_front", _wrap_BufLines_front, METH_O, NULL},
	 { "BufLines_back", _wrap_BufLines_back, METH_O, NULL},
	 { "BufLines_assign", _wrap_BufLines_assign, METH_VARARGS, NULL},
	 { "BufLines_resize", _wrap_BufLines_resize, METH_VARARGS, NULL},
	 { "BufLines_insert", _wrap_BufLines_insert, METH_VARARGS, NULL},
	 { "BufLines_pop_front", _wrap_BufLines_pop_front, METH_O, NULL},
	 { "BufLines_push_front", _wrap_BufLines_push_front, METH_VARARGS, NULL},
	 { "delete_BufLines", _wrap_delete_BufLines, METH_O, NULL},
	 { "BufLines_swigregister", BufLines_swigregister, METH_O, NULL},
	 { "BufLines_swiginit", BufLines_swiginit, METH_VARARGS, NULL},
	 { "VVString_iterator", _wrap_VVString_iterator, METH_O, NULL},
	 { "VVString___nonzero__", _wrap_VVString___nonzero__, METH_O, NULL},
	 { "VVString___bool__", _wrap_VVString___bool__, METH_O, NULL},
	 { "VVString___len__", _wrap_VVString___len__, METH_O, NULL},
	 { "VVString___getslice__", _wrap_VVString___getslice__, METH_VARARGS, NULL},
	 { "VVString___setslice__", _wrap_VVString___setslice__, METH_VARARGS, NULL},
	 { "VVString___delslice__", _wrap_VVString___delslice__, METH_VARARGS, NULL},
	 { "VVString___delitem__", _wrap_VVString___delitem__, METH_VARARGS, NULL},
	 { "VVString___getitem__", _wrap_VVString___getitem__, METH_VARARGS, NULL},
	 { "VVString___setitem__", _wrap_VVString___setitem__, METH_VARARGS, NULL},
	 { "VVString_pop", _wrap_VVString_pop, METH_O, NULL},
	 { "VVString_append", _wrap_VVString_append, METH_VARARGS, NULL},
	 { "VVString_empty", _wrap_VVString_empty, METH_O, NULL},
	 { "VVString_size", _wrap_VVString_size, METH_O, NULL},
	 { "VVString_swap", _wrap_VVString_swap, METH_VARARGS, NULL},
	 { "VVString_begin", _wrap_VVString_begin, METH_O, NULL},
	 { "VVString_end", _wrap_VVString_end, METH_O, NULL},
	 { "VVString_rbegin", _wrap_VVString_rbegin, METH_O, NULL},
	 { "VVString_rend", _wrap_VVString_rend, METH_O, NULL},
	 { "VVString_clear", _wrap_VVString_clear, METH_O, NULL},
	 { "VVString_get_allocator", _wrap_VVString_get_allocator, METH_O, NULL},
	 { "VVString_pop_back", _wrap_VVString_pop_back, METH_O, NULL},
	 { "VVString_erase", _wrap_VVString_erase, METH_VARARGS, NULL},
	 { "new_VVString", _wrap_new_VVString, METH_VARARGS, NULL},
	 { "VVString_push_back", _wrap_VVString_push_back, METH_VARARGS, NULL},
	 { "VVString_front", _wrap_VVString_front, METH_O, NULL},
	 { "VVString_back", _wrap_VVString_back, METH_O, NULL},
	 { "VVString_assign", _wrap_VVString_assign, METH_VARARGS, NULL},
	 { "VVString_resize", _wrap_VVString_resize, METH_VARARGS, NULL},
	 { "VVString_insert", _wrap_VVString_insert, METH_VARARGS, NULL},
	 { "VVString_reserve", _wrap_VVString_reserve, METH_VARARGS, NULL},
	 { "VVString_capacity", _wrap_VVString_capacity, METH_O, NULL},
	 { "delete_VVString", _wrap_delete_VVString, METH_O, NULL},
	 { "VVString_swigregister", VVString_swigregister, METH_O, NULL},
	 { "VVString_swiginit", VVString_swiginit, METH_VARARGS, NULL},
	 { "VClients_iterator", _wrap_VClients_iterator, METH_O, NULL},
	 { "VClients___nonzero__", _wrap_VClients___nonzero__, METH_O, NULL},
	 { "VClients___bool__", _wrap_VClients___bool__, METH_O, NULL},
	 { "VClients___len__", _wrap_VClients___len__, METH_O, NULL},
	 { "VClients___getslice__", _wrap_VClients___getslice__, METH_VARARGS, NULL},
	 { "VClients___setslice__", _wrap_VClients___setslice__, METH_VARARGS, NULL},
	 { "VClients___delslice__", _wrap_VClients___delslice__, METH_VARARGS, NULL},
	 { "VClients___delitem__", _wrap_VClients___delitem__, METH_VARARGS, NULL},
	 { "VClients___getitem__", _wrap_VClients___getitem__, METH_VARARGS, NULL},
	 { "VClients___setitem__", _wrap_VClients___setitem__, METH_VARARGS, NULL},
	 { "VClients_pop", _wrap_VClients_pop, METH_O, NULL},
	 { "VClients_append", _wrap_VClients_append, METH_VARARGS, NULL},
	 { "VClients_empty", _wrap_VClients_empty, METH_O, NULL},
	 { "VClients_size", _wrap_VClients_size, METH_O, NULL},
	 { "VClients_swap", _wrap_VClients_swap, METH_VARARGS, NULL},
	 { "VClients_begin", _wrap_VClients_begin, METH_O, NULL},
	 { "VClients_end", _wrap_VClients_end, METH_O, NULL},
	 { "VClients_rbegin", _wrap_VClients_rbegin, METH_O, NULL},
	 { "VClients_rend", _wrap_VClients_rend, METH_O, NULL},
	 { "VClients_clear", _wrap_VClients_clear, METH_O, NULL},
	 { "VClients_get_allocator", _wrap_VClients_get_allocator, METH_O, NULL},
	 { "VClients_pop_back", _wrap_VClients_pop_back, METH_O, NULL},
	 { "VClients_erase", _wrap_VClients_erase, METH_VARARGS, NULL},
	 { "new_VClients", _wrap_new_VClients, METH_VARARGS, NULL},
	 { "VClients_push_back", _wrap_VClients_push_back, METH_VARARGS, NULL},
	 { "VClients_front", _wrap_VClients_front, METH_O, NULL},
	 { "VClients_back", _wrap_VClients_back, METH_O, NULL},
	 { "VClients_assign", _wrap_VClients_assign, METH_VARARGS, NULL},
	 { "VClients_resize", _wrap_VClients_resize, METH_VARARGS, NULL},
	 { "VClients_insert", _wrap_VClients_insert, METH_VARARGS, NULL},
	 { "VClients_reserve", _wrap_VClients_reserve, METH_VARARGS, NULL},
	 { "VClients_capacity", _wrap_VClients_capacity, METH_O, NULL},
	 { "delete_VClients", _wrap_delete_VClients, METH_O, NULL},
	 { "VClients_swigregister", VClients_swigregister, METH_O, NULL},
	 { "VClients_swiginit", VClients_swiginit, METH_VARARGS, NULL},
	 { "VServers_iterator", _wrap_VServers_iterator, METH_O, NULL},
	 { "VServers___nonzero__", _wrap_VServers___nonzero__, METH_O, NULL},
	 { "VServers___bool__", _wrap_VServers___bool__, METH_O, NULL},
	 { "VServers___len__", _wrap_VServers___len__, METH_O, NULL},
	 { "VServers___getslice__", _wrap_VServers___getslice__, METH_VARARGS, NULL},
	 { "VServers___setslice__", _wrap_VServers___setslice__, METH_VARARGS, NULL},
	 { "VServers___delslice__", _wrap_VServers___delslice__, METH_VARARGS, NULL},
	 { "VServers___delitem__", _wrap_VServers___delitem__, METH_VARARGS, NULL},
	 { "VServers___getitem__", _wrap_VServers___getitem__, METH_VARARGS, NULL},
	 { "VServers___setitem__", _wrap_VServers___setitem__, METH_VARARGS, NULL},
	 { "VServers_pop", _wrap_VServers_pop, METH_O, NULL},
	 { "VServers_append", _wrap_VServers_append, METH_VARARGS, NULL},
	 { "VServers_empty", _wrap_VServers_empty, METH_O, NULL},
	 { "VServers_size", _wrap_VServers_size, METH_O, NULL},
	 { "VServers_swap", _wrap_VServers_swap, METH_VARARGS, NULL},
	 { "VServers_begin", _wrap_VServers_begin, METH_O, NULL},
	 { "VServers_end", _wrap_VServers_end, METH_O, NULL},
	 { "VServers_rbegin", _wrap_VServers_rbegin, METH_O, NULL},
	 { "VServers_rend", _wrap_VServers_rend, METH_O, NULL},
	 { "VServers_clear", _wrap_VServers_clear, METH_O, NULL},
	 { "VServers_get_allocator", _wrap_VServers_get_allocator, METH_O, NULL},
	 { "VServers_pop_back", _wrap_VServers_pop_back, METH_O, NULL},
	 { "VServers_erase", _wrap_VServers_erase, METH_VARARGS, NULL},
	 { "new_VServers", _wrap_new_VServers, METH_VARARGS, NULL},
	 { "VServers_push_back", _wrap_VServers_push_back, METH_VARARGS, NULL},
	 { "VServers_front", _wrap_VServers_front, METH_O, NULL},
	 { "VServers_back", _wrap_VServers_back, METH_O, NULL},
	 { "VServers_assign", _wrap_VServers_assign, METH_VARARGS, NULL},
	 { "VServers_resize", _wrap_VServers_resize, METH_VARARGS, NULL},
	 { "VServers_insert", _wrap_VServers_insert, METH_VARARGS, NULL},
	 { "VServers_reserve", _wrap_VServers_reserve, METH_VARARGS, NULL},
	 { "VServers_capacity", _wrap_VServers_capacity, METH_O, NULL},
	 { "delete_VServers", _wrap_delete_VServers, METH_O, NULL},
	 { "VServers_swigregister", VServers_swigregister, METH_O, NULL},
	 { "VServers_swiginit", VServers_swiginit, METH_VARARGS, NULL},
	 { "VQueries_iterator", _wrap_VQueries_iterator, METH_O, NULL},
	 { "VQueries___nonzero__", _wrap_VQueries___nonzero__, METH_O, NULL},
	 { "VQueries___bool__", _wrap_VQueries___bool__, METH_O, NULL},
	 { "VQueries___len__", _wrap_VQueries___len__, METH_O, NULL},
	 { "VQueries___getslice__", _wrap_VQueries___getslice__, METH_VARARGS, NULL},
	 { "VQueries___setslice__", _wrap_VQueries___setslice__, METH_VARARGS, NULL},
	 { "VQueries___delslice__", _wrap_VQueries___delslice__, METH_VARARGS, NULL},
	 { "VQueries___delitem__", _wrap_VQueries___delitem__, METH_VARARGS, NULL},
	 { "VQueries___getitem__", _wrap_VQueries___getitem__, METH_VARARGS, NULL},
	 { "VQueries___setitem__", _wrap_VQueries___setitem__, METH_VARARGS, NULL},
	 { "VQueries_pop", _wrap_VQueries_pop, METH_O, NULL},
	 { "VQueries_append", _wrap_VQueries_append, METH_VARARGS, NULL},
	 { "VQueries_empty", _wrap_VQueries_empty, METH_O, NULL},
	 { "VQueries_size", _wrap_VQueries_size, METH_O, NULL},
	 { "VQueries_swap", _wrap_VQueries_swap, METH_VARARGS, NULL},
	 { "VQueries_begin", _wrap_VQueries_begin, METH_O, NULL},
	 { "VQueries_end", _wrap_VQueries_end, METH_O, NULL},
	 { "VQueries_rbegin", _wrap_VQueries_rbegin, METH_O, NULL},
	 { "VQueries_rend", _wrap_VQueries_rend, METH_O, NULL},
	 { "VQueries_clear", _wrap_VQueries_clear, METH_O, NULL},
	 { "VQueries_get_allocator", _wrap_VQueries_get_allocator, METH_O, NULL},
	 { "VQueries_pop_back", _wrap_VQueries_pop_back, METH_O, NULL},
	 { "VQueries_erase", _wrap_VQueries_erase, METH_VARARGS, NULL},
	 { "new_VQueries", _wrap_new_VQueries, METH_VARARGS, NULL},
	 { "VQueries_push_back", _wrap_VQueries_push_back, METH_VARARGS, NULL},
	 { "VQueries_front", _wrap_VQueries_front, METH_O, NULL},
	 { "VQueries_back", _wrap_VQueries_back, METH_O, NULL},
	 { "VQueries_assign", _wrap_VQueries_assign, METH_VARARGS, NULL},
	 { "VQueries_resize", _wrap_VQueries_resize, METH_VARARGS, NULL},
	 { "VQueries_insert", _wrap_VQueries_insert, METH_VARARGS, NULL},
	 { "VQueries_reserve", _wrap_VQueries_reserve, METH_VARARGS, NULL},
	 { "VQueries_capacity", _wrap_VQueries_capacity, METH_O, NULL},
	 { "delete_VQueries", _wrap_delete_VQueries, METH_O, NULL},
	 { "VQueries_swigregister", VQueries_swigregister, METH_O, NULL},
	 { "VQueries_swiginit", VQueries_swiginit, METH_VARARGS, NULL},
	 { "CTranslationInfo_GetTranslations", _wrap_CTranslationInfo_GetTranslations, METH_NOARGS, NULL},
	 { "CTranslationInfo_sSelfName_set", _wrap_CTranslationInfo_sSelfName_set, METH_VARARGS, NULL},
	 { "CTranslationInfo_sSelfName_get", _wrap_CTranslationInfo_sSelfName_get, METH_O, NULL},
	 { "new_CTranslationInfo", _wrap_new_CTranslationInfo, METH_NOARGS, NULL},
	 { "delete_CTranslationInfo", _wrap_delete_CTranslationInfo, METH_O, NULL},
	 { "CTranslationInfo_swigregister", CTranslationInfo_swigregister, METH_O, NULL},
	 { "CTranslationInfo_swiginit", CTranslationInfo_swiginit, METH_VARARGS, NULL},
	 { "CTranslation_Get", _wrap_CTranslation_Get, METH_NOARGS, NULL},
	 { "CTranslation_Singular", _wrap_CTranslation_Singular, METH_VARARGS, NULL},
	 { "CTranslation_Plural", _wrap_CTranslation_Plural, METH_VARARGS, NULL},
	 { "CTranslation_PushLanguage", _wrap_CTranslation_PushLanguage, METH_VARARGS, NULL},
	 { "CTranslation_PopLanguage", _wrap_CTranslation_PopLanguage, METH_O, NULL},
	 { "CTranslation_NewReference", _wrap_CTranslation_NewReference, METH_VARARGS, NULL},
	 { "CTranslation_DelReference", _wrap_CTranslation_DelReference, METH_VARARGS, NULL},
	 { "new_CTranslation", _wrap_new_CTranslation, METH_NOARGS, NULL},
	 { "delete_CTranslation", _wrap_delete_CTranslation, METH_O, NULL},
	 { "CTranslation_swigregister", CTranslation_swigregister, METH_O, NULL},
	 { "CTranslation_swiginit", CTranslation_swiginit, METH_VARARGS, NULL},
	 { "new_CLanguageScope", _wrap_new_CLanguageScope, METH_O, NULL},
	 { "delete_CLanguageScope", _wrap_delete_CLanguageScope, METH_O, NULL},
	 { "CLanguageScope_swigregister", CLanguageScope_swigregister, METH_O, NULL},
	 { "CLanguageScope_swiginit", CLanguageScope_swiginit, METH_VARARGS, NULL},
	 { "new_CTranslationDomainRefHolder", _wrap_new_CTranslationDomainRefHolder, METH_O, NULL},
	 { "delete_CTranslationDomainRefHolder", _wrap_delete_CTranslationDomainRefHolder, METH_O, NULL},
	 { "CTranslationDomainRefHolder_swigregister", CTranslationDomainRefHolder_swigregister, METH_O, NULL},
	 { "CTranslationDomainRefHolder_swiginit", CTranslationDomainRefHolder_swiginit, METH_VARARGS, NULL},
	 { "new_CDelayedTranslation", _wrap_new_CDelayedTranslation, METH_VARARGS, NULL},
	 { "CDelayedTranslation_Resolve", _wrap_CDelayedTranslation_Resolve, METH_O, NULL},
	 { "delete_CDelayedTranslation", _wrap_delete_CDelayedTranslation, METH_O, NULL},
	 { "CDelayedTranslation_swigregister", CDelayedTranslation_swigregister, METH_O, NULL},
	 { "CDelayedTranslation_swiginit", CDelayedTranslation_swiginit, METH_VARARGS, NULL},
	 { "new_COptionalTranslation", _wrap_new_COptionalTranslation, METH_VARARGS, NULL},
	 { "COptionalTranslation_Resolve", _wrap_COptionalTranslation_Resolve, METH_O, NULL},
	 { "delete_COptionalTranslation", _wrap_delete_COptionalTranslation, METH_O, NULL},
	 { "COptionalTranslation_swigregister", COptionalTranslation_swigregister, METH_O, NULL},
	 { "COptionalTranslation_swiginit", COptionalTranslation_swiginit, METH_VARARGS, NULL},
	 { "new_CCoreTranslationMixin", _wrap_new_CCoreTranslationMixin, METH_NOARGS, NULL},
	 { "delete_CCoreTranslationMixin", _wrap_delete_CCoreTranslationMixin, METH_O, NULL},
	 { "CCoreTranslationMixin_swigregister", CCoreTranslationMixin_swigregister, METH_O, NULL},
	 { "CCoreTranslationMixin_swiginit", CCoreTranslationMixin_swiginit, METH_VARARGS, NULL},
	 { "SetFdCloseOnExec", _wrap_SetFdCloseOnExec, METH_O, NULL},
	 { "new_CUtils", _wrap_new_CUtils, METH_NOARGS, NULL},
	 { "delete_CUtils", _wrap_delete_CUtils, METH_O, NULL},
	 { "CUtils_GetIP", _wrap_CUtils_GetIP, METH_O, NULL},
	 { "CUtils_GetLongIP", _wrap_CUtils_GetLongIP, METH_O, NULL},
	 { "CUtils_GetHostName", _wrap_CUtils_GetHostName, METH_NOARGS, NULL},
	 { "CUtils_PrintError", _wrap_CUtils_PrintError, METH_O, NULL},
	 { "CUtils_PrintMessage", _wrap_CUtils_PrintMessage, METH_VARARGS, NULL},
	 { "CUtils_PrintPrompt", _wrap_CUtils_PrintPrompt, METH_O, NULL},
	 { "CUtils_PrintAction", _wrap_CUtils_PrintAction, METH_O, NULL},
	 { "CUtils_PrintStatus", _wrap_CUtils_PrintStatus, METH_VARARGS, NULL},
	 { "CUtils_AskSaltedHashPassForConfig", _wrap_CUtils_AskSaltedHashPassForConfig, METH_NOARGS, NULL},
	 { "CUtils_GetSalt", _wrap_CUtils_GetSalt, METH_NOARGS, NULL},
	 { "CUtils_SaltedMD5Hash", _wrap_CUtils_SaltedMD5Hash, METH_VARARGS, NULL},
	 { "CUtils_SaltedSHA256Hash", _wrap_CUtils_SaltedSHA256Hash, METH_VARARGS, NULL},
	 { "CUtils_SaltedHash", _wrap_CUtils_SaltedHash, METH_VARARGS, NULL},
	 { "CUtils_GetPass", _wrap_CUtils_GetPass, METH_O, NULL},
	 { "CUtils_GetInput", _wrap_CUtils_GetInput, METH_VARARGS, NULL},
	 { "CUtils_GetBoolInput", _wrap_CUtils_GetBoolInput, METH_VARARGS, NULL},
	 { "CUtils_GetNumInput", _wrap_CUtils_GetNumInput, METH_VARARGS, NULL},
	 { "CUtils_GetTime", _wrap_CUtils_GetTime, METH_NOARGS, NULL},
	 { "CUtils_GetMillTime", _wrap_CUtils_GetMillTime, METH_NOARGS, NULL},
	 { "CUtils_CTime", _wrap_CUtils_CTime, METH_VARARGS, NULL},
	 { "CUtils_FormatTime", _wrap_CUtils_FormatTime, METH_VARARGS, NULL},
	 { "CUtils_FormatServerTime", _wrap_CUtils_FormatServerTime, METH_O, NULL},
	 { "CUtils_ParseServerTime", _wrap_CUtils_ParseServerTime, METH_O, NULL},
	 { "CUtils_GetTimezones", _wrap_CUtils_GetTimezones, METH_NOARGS, NULL},
	 { "CUtils_GetEncodings", _wrap_CUtils_GetEncodings, METH_NOARGS, NULL},
	 { "CUtils_CheckCIDR", _wrap_CUtils_CheckCIDR, METH_VARARGS, NULL},
	 { "CUtils_GetMessageTags", _wrap_CUtils_GetMessageTags, METH_O, NULL},
	 { "CUtils_SetMessageTags", _wrap_CUtils_SetMessageTags, METH_VARARGS, NULL},
	 { "CUtils_swigregister", CUtils_swigregister, METH_O, NULL},
	 { "CUtils_swiginit", CUtils_swiginit, METH_VARARGS, NULL},
	 { "new_CException", _wrap_new_CException, METH_O, NULL},
	 { "delete_CException", _wrap_delete_CException, METH_O, NULL},
	 { "CException_GetType", _wrap_CException_GetType, METH_O, NULL},
	 { "CException_swigregister", CException_swigregister, METH_O, NULL},
	 { "CException_swiginit", CException_swiginit, METH_VARARGS, NULL},
	 { "new_CTable", _wrap_new_CTable, METH_NOARGS, NULL},
	 { "delete_CTable", _wrap_delete_CTable, METH_O, NULL},
	 { "CTable_AddColumn", _wrap_CTable_AddColumn, METH_VARARGS, NULL},
	 { "CTable_SetStyle", _wrap_CTable_SetStyle, METH_VARARGS, NULL},
	 { "CTable_AddRow", _wrap_CTable_AddRow, METH_O, NULL},
	 { "CTable_SetCell", _wrap_CTable_SetCell, METH_VARARGS, NULL},
	 { "CTable_GetLine", _wrap_CTable_GetLine, METH_VARARGS, NULL},
	 { "CTable_GetColumnWidth", _wrap_CTable_GetColumnWidth, METH_VARARGS, NULL},
	 { "CTable_Clear", _wrap_CTable_Clear, METH_O, NULL},
	 { "CTable_size", _wrap_CTable_size, METH_O, NULL},
	 { "CTable_empty", _wrap_CTable_empty, METH_O, NULL},
	 { "CTable_swigregister", CTable_swigregister, METH_O, NULL},
	 { "CTable_swiginit", CTable_swiginit, METH_VARARGS, NULL},
	 { "new_CConfigEntry", _wrap_new_CConfigEntry, METH_VARARGS, NULL},
	 { "delete_CConfigEntry", _wrap_delete_CConfigEntry, METH_O, NULL},
	 { "CConfigEntry_m_pSubConfig_set", _wrap_CConfigEntry_m_pSubConfig_set, METH_VARARGS, NULL},
	 { "CConfigEntry_m_pSubConfig_get", _wrap_CConfigEntry_m_pSubConfig_get, METH_O, NULL},
	 { "CConfigEntry_swigregister", CConfigEntry_swigregister, METH_O, NULL},
	 { "CConfigEntry_swiginit", CConfigEntry_swiginit, METH_VARARGS, NULL},
	 { "new_CConfig", _wrap_new_CConfig, METH_NOARGS, NULL},
	 { "CConfig_BeginEntries", _wrap_CConfig_BeginEntries, METH_O, NULL},
	 { "CConfig_EndEntries", _wrap_CConfig_EndEntries, METH_O, NULL},
	 { "CConfig_BeginSubConfigs", _wrap_CConfig_BeginSubConfigs, METH_O, NULL},
	 { "CConfig_EndSubConfigs", _wrap_CConfig_EndSubConfigs, METH_O, NULL},
	 { "CConfig_AddKeyValuePair", _wrap_CConfig_AddKeyValuePair, METH_VARARGS, NULL},
	 { "CConfig_AddSubConfig", _wrap_CConfig_AddSubConfig, METH_VARARGS, NULL},
	 { "CConfig_FindStringVector", _wrap_CConfig_FindStringVector, METH_VARARGS, NULL},
	 { "CConfig_FindStringEntry", _wrap_CConfig_FindStringEntry, METH_VARARGS, NULL},
	 { "CConfig_FindBoolEntry", _wrap_CConfig_FindBoolEntry, METH_VARARGS, NULL},
	 { "CConfig_FindUIntEntry", _wrap_CConfig_FindUIntEntry, METH_VARARGS, NULL},
	 { "CConfig_FindUShortEntry", _wrap_CConfig_FindUShortEntry, METH_VARARGS, NULL},
	 { "CConfig_FindDoubleEntry", _wrap_CConfig_FindDoubleEntry, METH_VARARGS, NULL},
	 { "CConfig_FindSubConfig", _wrap_CConfig_FindSubConfig, METH_VARARGS, NULL},
	 { "CConfig_empty", _wrap_CConfig_empty, METH_O, NULL},
	 { "CConfig_Parse", _wrap_CConfig_Parse, METH_VARARGS, NULL},
	 { "CConfig_Write", _wrap_CConfig_Write, METH_VARARGS, NULL},
	 { "delete_CConfig", _wrap_delete_CConfig, METH_O, NULL},
	 { "CConfig_swigregister", CConfig_swigregister, METH_O, NULL},
	 { "CConfig_swiginit", CConfig_swiginit, METH_VARARGS, NULL},
	 { "new_CSCharBuffer", _wrap_new_CSCharBuffer, METH_O, NULL},
	 { "delete_CSCharBuffer", _wrap_delete_CSCharBuffer, METH_O, NULL},
	 { "CSCharBuffer___call__", _wrap_CSCharBuffer___call__, METH_O, NULL},
	 { "CSCharBuffer_swigregister", CSCharBuffer_swigregister, METH_O, NULL},
	 { "CSCharBuffer_swiginit", CSCharBuffer_swiginit, METH_VARARGS, NULL},
	 { "new_CSSockAddr", _wrap_new_CSSockAddr, METH_NOARGS, NULL},
	 { "delete_CSSockAddr", _wrap_delete_CSSockAddr, METH_O, NULL},
	 { "CSSockAddr_SinFamily", _wrap_CSSockAddr_SinFamily, METH_O, NULL},
	 { "CSSockAddr_SinPort", _wrap_CSSockAddr_SinPort, METH_VARARGS, NULL},
	 { "CSSockAddr_SetIPv6", _wrap_CSSockAddr_SetIPv6, METH_VARARGS, NULL},
	 { "CSSockAddr_GetIPv6", _wrap_CSSockAddr_GetIPv6, METH_O, NULL},
	 { "CSSockAddr_GetSockAddrLen", _wrap_CSSockAddr_GetSockAddrLen, METH_O, NULL},
	 { "CSSockAddr_GetSockAddr", _wrap_CSSockAddr_GetSockAddr, METH_O, NULL},
	 { "CSSockAddr_GetAddr", _wrap_CSSockAddr_GetAddr, METH_O, NULL},
	 { "CSSockAddr_SetAFRequire", _wrap_CSSockAddr_SetAFRequire, METH_VARARGS, NULL},
	 { "CSSockAddr_GetAFRequire", _wrap_CSSockAddr_GetAFRequire, METH_O, NULL},
	 { "CSSockAddr_swigregister", CSSockAddr_swigregister, METH_O, NULL},
	 { "CSSockAddr_swiginit", CSSockAddr_swiginit, METH_VARARGS, NULL},
	 { "new_CGetAddrInfo", _wrap_new_CGetAddrInfo, METH_VARARGS, NULL},
	 { "delete_CGetAddrInfo", _wrap_delete_CGetAddrInfo, METH_O, NULL},
	 { "CGetAddrInfo_Init", _wrap_CGetAddrInfo_Init, METH_O, NULL},
	 { "CGetAddrInfo_Process", _wrap_CGetAddrInfo_Process, METH_O, NULL},
	 { "CGetAddrInfo_Finish", _wrap_CGetAddrInfo_Finish, METH_O, NULL},
	 { "CGetAddrInfo_swigregister", CGetAddrInfo_swigregister, METH_O, NULL},
	 { "CGetAddrInfo_swiginit", CGetAddrInfo_swiginit, METH_VARARGS, NULL},
	 { "CS_GetAddrInfo", _wrap_CS_GetAddrInfo, METH_VARARGS, NULL},
	 { "GetCsockSSLIdx", _wrap_GetCsockSSLIdx, METH_NOARGS, NULL},
	 { "InitCsocket", _wrap_InitCsocket, METH_NOARGS, NULL},
	 { "ShutdownCsocket", _wrap_ShutdownCsocket, METH_NOARGS, NULL},
	 { "GetSockError", _wrap_GetSockError, METH_NOARGS, NULL},
	 { "TFD_ZERO", _wrap_TFD_ZERO, METH_O, NULL},
	 { "TFD_SET", _wrap_TFD_SET, METH_VARARGS, NULL},
	 { "TFD_ISSET", _wrap_TFD_ISSET, METH_VARARGS, NULL},
	 { "TFD_CLR", _wrap_TFD_CLR, METH_VARARGS, NULL},
	 { "__Perror", _wrap___Perror, METH_VARARGS, NULL},
	 { "new_CCron", _wrap_new_CCron, METH_NOARGS, NULL},
	 { "delete_CCron", _wrap_delete_CCron, METH_O, NULL},
	 { "CCron_run", _wrap_CCron_run, METH_VARARGS, NULL},
	 { "CCron_StartMaxCycles", _wrap_CCron_StartMaxCycles, METH_VARARGS, NULL},
	 { "CCron_Start", _wrap_CCron_Start, METH_VARARGS, NULL},
	 { "CCron_Stop", _wrap_CCron_Stop, METH_O, NULL},
	 { "CCron_Pause", _wrap_CCron_Pause, METH_O, NULL},
	 { "CCron_UnPause", _wrap_CCron_UnPause, METH_O, NULL},
	 { "CCron_Reset", _wrap_CCron_Reset, METH_O, NULL},
	 { "CCron_GetInterval", _wrap_CCron_GetInterval, METH_O, NULL},
	 { "CCron_GetMaxCycles", _wrap_CCron_GetMaxCycles, METH_O, NULL},
	 { "CCron_GetCyclesLeft", _wrap_CCron_GetCyclesLeft, METH_O, NULL},
	 { "CCron_isValid", _wrap_CCron_isValid, METH_O, NULL},
	 { "CCron_GetName", _wrap_CCron_GetName, METH_O, NULL},
	 { "CCron_SetName", _wrap_CCron_SetName, METH_VARARGS, NULL},
	 { "CCron_GetNextRun", _wrap_CCron_GetNextRun, METH_O, NULL},
	 { "CCron_RunJob", _wrap_CCron_RunJob, METH_O, NULL},
	 { "CCron_swigregister", CCron_swigregister, METH_O, NULL},
	 { "CCron_swiginit", CCron_swiginit, METH_VARARGS, NULL},
	 { "new_CSMonitorFD", _wrap_new_CSMonitorFD, METH_NOARGS, NULL},
	 { "delete_CSMonitorFD", _wrap_delete_CSMonitorFD, METH_O, NULL},
	 { "CSMonitorFD_GatherFDsForSelect", _wrap_CSMonitorFD_GatherFDsForSelect, METH_VARARGS, NULL},
	 { "CSMonitorFD_FDsThatTriggered", _wrap_CSMonitorFD_FDsThatTriggered, METH_VARARGS, NULL},
	 { "CSMonitorFD_CheckFDs", _wrap_CSMonitorFD_CheckFDs, METH_VARARGS, NULL},
	 { "CSMonitorFD_Add", _wrap_CSMonitorFD_Add, METH_VARARGS, NULL},
	 { "CSMonitorFD_Remove", _wrap_CSMonitorFD_Remove, METH_VARARGS, NULL},
	 { "CSMonitorFD_DisableMonitor", _wrap_CSMonitorFD_DisableMonitor, METH_O, NULL},
	 { "CSMonitorFD_IsEnabled", _wrap_CSMonitorFD_IsEnabled, METH_O, NULL},
	 { "CSMonitorFD_swigregister", CSMonitorFD_swigregister, METH_O, NULL},
	 { "CSMonitorFD_swiginit", CSMonitorFD_swiginit, METH_VARARGS, NULL},
	 { "new_CSockCommon", _wrap_new_CSockCommon, METH_NOARGS, NULL},
	 { "delete_CSockCommon", _wrap_delete_CSockCommon, METH_O, NULL},
	 { "CSockCommon_CleanupCrons", _wrap_CSockCommon_CleanupCrons, METH_O, NULL},
	 { "CSockCommon_CleanupFDMonitors", _wrap_CSockCommon_CleanupFDMonitors, METH_O, NULL},
	 { "CSockCommon_GetCrons", _wrap_CSockCommon_GetCrons, METH_O, NULL},
	 { "CSockCommon_Cron", _wrap_CSockCommon_Cron, METH_O, NULL},
	 { "CSockCommon_AddCron", _wrap_CSockCommon_AddCron, METH_VARARGS, NULL},
	 { "CSockCommon_DelCron", _wrap_CSockCommon_DelCron, METH_VARARGS, NULL},
	 { "CSockCommon_DelCronByAddr", _wrap_CSockCommon_DelCronByAddr, METH_VARARGS, NULL},
	 { "CSockCommon_CheckFDs", _wrap_CSockCommon_CheckFDs, METH_VARARGS, NULL},
	 { "CSockCommon_AssignFDs", _wrap_CSockCommon_AssignFDs, METH_VARARGS, NULL},
	 { "CSockCommon_MonitorFD", _wrap_CSockCommon_MonitorFD, METH_VARARGS, NULL},
	 { "CSockCommon_swigregister", CSockCommon_swigregister, METH_O, NULL},
	 { "CSockCommon_swiginit", CSockCommon_swiginit, METH_VARARGS, NULL},
	 { "new_Csock", _wrap_new_Csock, METH_VARARGS, NULL},
	 { "Csock_GetSockObj", _wrap_Csock_GetSockObj, METH_VARARGS, NULL},
	 { "delete_Csock", _wrap_delete_Csock, METH_O, NULL},
	 { "Csock_Dereference", _wrap_Csock_Dereference, METH_O, NULL},
	 { "Csock_Copy", _wrap_Csock_Copy, METH_VARARGS, NULL},
	 { "Csock___lshift__", _wrap_Csock___lshift__, METH_VARARGS, NULL},
	 { "Csock_Connect", _wrap_Csock_Connect, METH_O, NULL},
	 { "Csock_ConnectUnixInternal", _wrap_Csock_ConnectUnixInternal, METH_VARARGS, NULL},
	 { "Csock_ListenUnixInternal", _wrap_Csock_ListenUnixInternal, METH_VARARGS, NULL},
	 { "Csock_Listen", _wrap_Csock_Listen, METH_VARARGS, NULL},
	 { "Csock_Accept", _wrap_Csock_Accept, METH_VARARGS, NULL},
	 { "Csock_AcceptSSL", _wrap_Csock_AcceptSSL, METH_O, NULL},
	 { "Csock_SSLClientSetup", _wrap_Csock_SSLClientSetup, METH_O, NULL},
	 { "Csock_SSLServerSetup", _wrap_Csock_SSLServerSetup, METH_O, NULL},
	 { "Csock_ConnectSSL", _wrap_Csock_ConnectSSL, METH_O, NULL},
	 { "Csock_StartTLS", _wrap_Csock_StartTLS, METH_O, NULL},
	 { "Csock_Write", _wrap_Csock_Write, METH_VARARGS, NULL},
	 { "Csock_Read", _wrap_Csock_Read, METH_VARARGS, NULL},
	 { "Csock_GetLocalIP", _wrap_Csock_GetLocalIP, METH_O, NULL},
	 { "Csock_GetRemoteIP", _wrap_Csock_GetRemoteIP, METH_O, NULL},
	 { "Csock_IsConnected", _wrap_Csock_IsConnected, METH_O, NULL},
	 { "Csock_SetIsConnected", _wrap_Csock_SetIsConnected, METH_VARARGS, NULL},
	 { "Csock_GetRSock", _wrap_Csock_GetRSock, METH_VARARGS, NULL},
	 { "Csock_SetRSock", _wrap_Csock_SetRSock, METH_VARARGS, NULL},
	 { "Csock_GetWSock", _wrap_Csock_GetWSock, METH_VARARGS, NULL},
	 { "Csock_SetWSock", _wrap_Csock_SetWSock, METH_VARARGS, NULL},
	 { "Csock_SetSock", _wrap_Csock_SetSock, METH_VARARGS, NULL},
	 { "Csock_GetSock", _wrap_Csock_GetSock, METH_VARARGS, NULL},
	 { "Csock_CallSockError", _wrap_Csock_CallSockError, METH_VARARGS, NULL},
	 { "Csock_ResetTimer", _wrap_Csock_ResetTimer, METH_O, NULL},
	 { "Csock_PauseRead", _wrap_Csock_PauseRead, METH_O, NULL},
	 { "Csock_UnPauseRead", _wrap_Csock_UnPauseRead, METH_O, NULL},
	 { "Csock_IsReadPaused", _wrap_Csock_IsReadPaused, METH_O, NULL},
	 { "Csock_SetTimeout", _wrap_Csock_SetTimeout, METH_VARARGS, NULL},
	 { "Csock_SetTimeoutType", _wrap_Csock_SetTimeoutType, METH_VARARGS, NULL},
	 { "Csock_GetTimeout", _wrap_Csock_GetTimeout, METH_O, NULL},
	 { "Csock_GetTimeoutType", _wrap_Csock_GetTimeoutType, METH_O, NULL},
	 { "Csock_CheckTimeout", _wrap_Csock_CheckTimeout, METH_VARARGS, NULL},
	 { "Csock_PushBuff", _wrap_Csock_PushBuff, METH_VARARGS, NULL},
	 { "Csock_GetInternalReadBuffer", _wrap_Csock_GetInternalReadBuffer, METH_O, NULL},
	 { "Csock_GetInternalWriteBuffer", _wrap_Csock_GetInternalWriteBuffer, METH_O, NULL},
	 { "Csock_SetMaxBufferThreshold", _wrap_Csock_SetMaxBufferThreshold, METH_VARARGS, NULL},
	 { "Csock_GetMaxBufferThreshold", _wrap_Csock_GetMaxBufferThreshold, METH_O, NULL},
	 { "Csock_GetType", _wrap_Csock_GetType, METH_O, NULL},
	 { "Csock_SetType", _wrap_Csock_SetType, METH_VARARGS, NULL},
	 { "Csock_GetSockName", _wrap_Csock_GetSockName, METH_O, NULL},
	 { "Csock_SetSockName", _wrap_Csock_SetSockName, METH_VARARGS, NULL},
	 { "Csock_GetHostName", _wrap_Csock_GetHostName, METH_O, NULL},
	 { "Csock_SetHostName", _wrap_Csock_SetHostName, METH_VARARGS, NULL},
	 { "Csock_GetStartTime", _wrap_Csock_GetStartTime, METH_O, NULL},
	 { "Csock_ResetStartTime", _wrap_Csock_ResetStartTime, METH_O, NULL},
	 { "Csock_GetBytesRead", _wrap_Csock_GetBytesRead, METH_O, NULL},
	 { "Csock_ResetBytesRead", _wrap_Csock_ResetBytesRead, METH_O, NULL},
	 { "Csock_GetBytesWritten", _wrap_Csock_GetBytesWritten, METH_O, NULL},
	 { "Csock_ResetBytesWritten", _wrap_Csock_ResetBytesWritten, METH_O, NULL},
	 { "Csock_GetAvgRead", _wrap_Csock_GetAvgRead, METH_VARARGS, NULL},
	 { "Csock_GetAvgWrite", _wrap_Csock_GetAvgWrite, METH_VARARGS, NULL},
	 { "Csock_GetRemotePort", _wrap_Csock_GetRemotePort, METH_O, NULL},
	 { "Csock_GetLocalPort", _wrap_Csock_GetLocalPort, METH_O, NULL},
	 { "Csock_GetPort", _wrap_Csock_GetPort, METH_O, NULL},
	 { "Csock_SetPort", _wrap_Csock_SetPort, METH_VARARGS, NULL},
	 { "Csock_Close", _wrap_Csock_Close, METH_VARARGS, NULL},
	 { "Csock_GetCloseType", _wrap_Csock_GetCloseType, METH_O, NULL},
	 { "Csock_IsClosed", _wrap_Csock_IsClosed, METH_O, NULL},
	 { "Csock_NonBlockingIO", _wrap_Csock_NonBlockingIO, METH_O, NULL},
	 { "Csock_GetSSL", _wrap_Csock_GetSSL, METH_O, NULL},
	 { "Csock_SetSSL", _wrap_Csock_SetSSL, METH_VARARGS, NULL},
	 { "Csock_HasWriteBuffer", _wrap_Csock_HasWriteBuffer, METH_O, NULL},
	 { "Csock_ClearWriteBuffer", _wrap_Csock_ClearWriteBuffer, METH_O, NULL},
	 { "Csock_SslIsEstablished", _wrap_Csock_SslIsEstablished, METH_O, NULL},
	 { "Csock_ConnectInetd", _wrap_Csock_ConnectInetd, METH_VARARGS, NULL},
	 { "Csock_ConnectFD", _wrap_Csock_ConnectFD, METH_VARARGS, NULL},
	 { "Csock_SetParentSockName", _wrap_Csock_SetParentSockName, METH_VARARGS, NULL},
	 { "Csock_GetParentSockName", _wrap_Csock_GetParentSockName, METH_O, NULL},
	 { "Csock_SetRate", _wrap_Csock_SetRate, METH_VARARGS, NULL},
	 { "Csock_GetRateBytes", _wrap_Csock_GetRateBytes, METH_O, NULL},
	 { "Csock_GetRateTime", _wrap_Csock_GetRateTime, METH_O, NULL},
	 { "Csock_Connected", _wrap_Csock_Connected, METH_O, NULL},
	 { "Csock_Disconnected", _wrap_Csock_Disconnected, METH_O, NULL},
	 { "Csock_Timeout", _wrap_Csock_Timeout, METH_O, NULL},
	 { "Csock_ReadData", _wrap_Csock_ReadData, METH_VARARGS, NULL},
	 { "Csock_ReadLine", _wrap_Csock_ReadLine, METH_VARARGS, NULL},
	 { "Csock_EnableReadLine", _wrap_Csock_EnableReadLine, METH_O, NULL},
	 { "Csock_DisableReadLine", _wrap_Csock_DisableReadLine, METH_O, NULL},
	 { "Csock_HasReadLine", _wrap_Csock_HasReadLine, METH_O, NULL},
	 { "Csock_ReachedMaxBuffer", _wrap_Csock_ReachedMaxBuffer, METH_O, NULL},
	 { "Csock_SockError", _wrap_Csock_SockError, METH_VARARGS, NULL},
	 { "Csock_ConnectionFrom", _wrap_Csock_ConnectionFrom, METH_VARARGS, NULL},
	 { "Csock_Listening", _wrap_Csock_Listening, METH_VARARGS, NULL},
	 { "Csock_ConnectionRefused", _wrap_Csock_ConnectionRefused, METH_O, NULL},
	 { "Csock_ReadPaused", _wrap_Csock_ReadPaused, METH_O, NULL},
	 { "Csock_GetTimeSinceLastDataTransaction", _wrap_Csock_GetTimeSinceLastDataTransaction, METH_VARARGS, NULL},
	 { "Csock_GetLastCheckTimeout", _wrap_Csock_GetLastCheckTimeout, METH_O, NULL},
	 { "Csock_GetNextCheckTimeout", _wrap_Csock_GetNextCheckTimeout, METH_VARARGS, NULL},
	 { "Csock_GetPending", _wrap_Csock_GetPending, METH_O, NULL},
	 { "Csock_GetConState", _wrap_Csock_GetConState, METH_O, NULL},
	 { "Csock_SetConState", _wrap_Csock_SetConState, METH_VARARGS, NULL},
	 { "Csock_CreateSocksFD", _wrap_Csock_CreateSocksFD, METH_O, NULL},
	 { "Csock_CloseSocksFD", _wrap_Csock_CloseSocksFD, METH_O, NULL},
	 { "Csock_GetBindHost", _wrap_Csock_GetBindHost, METH_O, NULL},
	 { "Csock_SetBindHost", _wrap_Csock_SetBindHost, METH_VARARGS, NULL},
	 { "Csock_DNSLookup", _wrap_Csock_DNSLookup, METH_VARARGS, NULL},
	 { "Csock_SetupVHost", _wrap_Csock_SetupVHost, METH_O, NULL},
	 { "Csock_GetIPv6", _wrap_Csock_GetIPv6, METH_O, NULL},
	 { "Csock_SetIPv6", _wrap_Csock_SetIPv6, METH_VARARGS, NULL},
	 { "Csock_SetAFRequire", _wrap_Csock_SetAFRequire, METH_VARARGS, NULL},
	 { "Csock_AllowWrite", _wrap_Csock_AllowWrite, METH_VARARGS, NULL},
	 { "Csock_SetSkipConnect", _wrap_Csock_SetSkipConnect, METH_VARARGS, NULL},
	 { "Csock_GetAddrInfo", _wrap_Csock_GetAddrInfo, METH_VARARGS, NULL},
	 { "Csock_ConvertAddress", _wrap_Csock_ConvertAddress, METH_VARARGS, NULL},
	 { "Csock_GetMaxConns", _wrap_Csock_GetMaxConns, METH_O, NULL},
	 { "Csock_WriteBytes", _wrap_Csock_WriteBytes, METH_VARARGS, NULL},
	 { "Csock_swigregister", Csock_swigregister, METH_O, NULL},
	 { "Csock_swiginit", Csock_swiginit, METH_VARARGS, NULL},
	 { "new_CSConnection", _wrap_new_CSConnection, METH_VARARGS, NULL},
	 { "delete_CSConnection", _wrap_delete_CSConnection, METH_O, NULL},
	 { "CSConnection_GetHostname", _wrap_CSConnection_GetHostname, METH_O, NULL},
	 { "CSConnection_GetSockName", _wrap_CSConnection_GetSockName, METH_O, NULL},
	 { "CSConnection_GetBindHost", _wrap_CSConnection_GetBindHost, METH_O, NULL},
	 { "CSConnection_GetPort", _wrap_CSConnection_GetPort, METH_O, NULL},
	 { "CSConnection_GetTimeout", _wrap_CSConnection_GetTimeout, METH_O, NULL},
	 { "CSConnection_GetIsSSL", _wrap_CSConnection_GetIsSSL, METH_O, NULL},
	 { "CSConnection_GetAFRequire", _wrap_CSConnection_GetAFRequire, METH_O, NULL},
	 { "CSConnection_SetHostname", _wrap_CSConnection_SetHostname, METH_VARARGS, NULL},
	 { "CSConnection_SetSockName", _wrap_CSConnection_SetSockName, METH_VARARGS, NULL},
	 { "CSConnection_SetBindHost", _wrap_CSConnection_SetBindHost, METH_VARARGS, NULL},
	 { "CSConnection_SetPort", _wrap_CSConnection_SetPort, METH_VARARGS, NULL},
	 { "CSConnection_SetTimeout", _wrap_CSConnection_SetTimeout, METH_VARARGS, NULL},
	 { "CSConnection_SetIsSSL", _wrap_CSConnection_SetIsSSL, METH_VARARGS, NULL},
	 { "CSConnection_SetAFRequire", _wrap_CSConnection_SetAFRequire, METH_VARARGS, NULL},
	 { "CSConnection_swigregister", CSConnection_swigregister, METH_O, NULL},
	 { "CSConnection_swiginit", CSConnection_swiginit, METH_VARARGS, NULL},
	 { "new_CSSSLConnection", _wrap_new_CSSSLConnection, METH_VARARGS, NULL},
	 { "delete_CSSSLConnection", _wrap_delete_CSSSLConnection, METH_O, NULL},
	 { "CSSSLConnection_swigregister", CSSSLConnection_swigregister, METH_O, NULL},
	 { "CSSSLConnection_swiginit", CSSSLConnection_swiginit, METH_VARARGS, NULL},
	 { "new_CSListener", _wrap_new_CSListener, METH_VARARGS, NULL},
	 { "delete_CSListener", _wrap_delete_CSListener, METH_O, NULL},
	 { "CSListener_SetDetach", _wrap_CSListener_SetDetach, METH_VARARGS, NULL},
	 { "CSListener_GetDetach", _wrap_CSListener_GetDetach, METH_O, NULL},
	 { "CSListener_GetPort", _wrap_CSListener_GetPort, METH_O, NULL},
	 { "CSListener_GetSockName", _wrap_CSListener_GetSockName, METH_O, NULL},
	 { "CSListener_GetBindHost", _wrap_CSListener_GetBindHost, METH_O, NULL},
	 { "CSListener_GetIsSSL", _wrap_CSListener_GetIsSSL, METH_O, NULL},
	 { "CSListener_GetMaxConns", _wrap_CSListener_GetMaxConns, METH_O, NULL},
	 { "CSListener_GetTimeout", _wrap_CSListener_GetTimeout, METH_O, NULL},
	 { "CSListener_GetAFRequire", _wrap_CSListener_GetAFRequire, METH_O, NULL},
	 { "CSListener_SetPort", _wrap_CSListener_SetPort, METH_VARARGS, NULL},
	 { "CSListener_SetSockName", _wrap_CSListener_SetSockName, METH_VARARGS, NULL},
	 { "CSListener_SetBindHost", _wrap_CSListener_SetBindHost, METH_VARARGS, NULL},
	 { "CSListener_SetIsSSL", _wrap_CSListener_SetIsSSL, METH_VARARGS, NULL},
	 { "CSListener_SetMaxConns", _wrap_CSListener_SetMaxConns, METH_VARARGS, NULL},
	 { "CSListener_SetTimeout", _wrap_CSListener_SetTimeout, METH_VARARGS, NULL},
	 { "CSListener_SetAFRequire", _wrap_CSListener_SetAFRequire, METH_VARARGS, NULL},
	 { "CSListener_swigregister", CSListener_swigregister, METH_O, NULL},
	 { "CSListener_swiginit", CSListener_swiginit, METH_VARARGS, NULL},
	 { "new_CSocketManager", _wrap_new_CSocketManager, METH_NOARGS, NULL},
	 { "delete_CSocketManager", _wrap_delete_CSocketManager, METH_O, NULL},
	 { "CSocketManager_clear", _wrap_CSocketManager_clear, METH_O, NULL},
	 { "CSocketManager_Cleanup", _wrap_CSocketManager_Cleanup, METH_O, NULL},
	 { "CSocketManager_GetSockObj", _wrap_CSocketManager_GetSockObj, METH_VARARGS, NULL},
	 { "CSocketManager_Connect", _wrap_CSocketManager_Connect, METH_VARARGS, NULL},
	 { "CSocketManager_Listen", _wrap_CSocketManager_Listen, METH_VARARGS, NULL},
	 { "CSocketManager_HasFDs", _wrap_CSocketManager_HasFDs, METH_O, NULL},
	 { "CSocketManager_Loop", _wrap_CSocketManager_Loop, METH_O, NULL},
	 { "CSocketManager_DynamicSelectLoop", _wrap_CSocketManager_DynamicSelectLoop, METH_VARARGS, NULL},
	 { "CSocketManager_AddSock", _wrap_CSocketManager_AddSock, METH_VARARGS, NULL},
	 { "CSocketManager_FindSockByRemotePort", _wrap_CSocketManager_FindSockByRemotePort, METH_VARARGS, NULL},
	 { "CSocketManager_FindSockByLocalPort", _wrap_CSocketManager_FindSockByLocalPort, METH_VARARGS, NULL},
	 { "CSocketManager_FindSockByName", _wrap_CSocketManager_FindSockByName, METH_VARARGS, NULL},
	 { "CSocketManager_FindSockByFD", _wrap_CSocketManager_FindSockByFD, METH_VARARGS, NULL},
	 { "CSocketManager_FindSocksByName", _wrap_CSocketManager_FindSocksByName, METH_VARARGS, NULL},
	 { "CSocketManager_FindSocksByRemoteHost", _wrap_CSocketManager_FindSocksByRemoteHost, METH_VARARGS, NULL},
	 { "CSocketManager_GetErrno", _wrap_CSocketManager_GetErrno, METH_O, NULL},
	 { "CSocketManager_GetSelectTimeout", _wrap_CSocketManager_GetSelectTimeout, METH_O, NULL},
	 { "CSocketManager_SetSelectTimeout", _wrap_CSocketManager_SetSelectTimeout, METH_VARARGS, NULL},
	 { "CSocketManager_DelSockByAddr", _wrap_CSocketManager_DelSockByAddr, METH_VARARGS, NULL},
	 { "CSocketManager_DelSock", _wrap_CSocketManager_DelSock, METH_VARARGS, NULL},
	 { "CSocketManager_SwapSockByIdx", _wrap_CSocketManager_SwapSockByIdx, METH_VARARGS, NULL},
	 { "CSocketManager_SwapSockByAddr", _wrap_CSocketManager_SwapSockByAddr, METH_VARARGS, NULL},
	 { "CSocketManager_GetBytesRead", _wrap_CSocketManager_GetBytesRead, METH_O, NULL},
	 { "CSocketManager_GetBytesWritten", _wrap_CSocketManager_GetBytesWritten, METH_O, NULL},
	 { "CSocketManager_FDSetCheck", _wrap_CSocketManager_FDSetCheck, METH_VARARGS, NULL},
	 { "CSocketManager_FDHasCheck", _wrap_CSocketManager_FDHasCheck, METH_VARARGS, NULL},
	 { "CSocketManager_swigregister", CSocketManager_swigregister, METH_O, NULL},
	 { "CSocketManager_swiginit", CSocketManager_swiginit, METH_VARARGS, NULL},
	 { "new_ZNCSocketManager", _wrap_new_ZNCSocketManager, METH_NOARGS, NULL},
	 { "delete_ZNCSocketManager", _wrap_delete_ZNCSocketManager, METH_O, NULL},
	 { "ZNCSocketManager_GetSockObj", _wrap_ZNCSocketManager_GetSockObj, METH_VARARGS, NULL},
	 { "ZNCSocketManager_swigregister", ZNCSocketManager_swigregister, METH_O, NULL},
	 { "ZNCSocketManager_swiginit", ZNCSocketManager_swiginit, METH_VARARGS, NULL},
	 { "new_CZNCSock", _wrap_new_CZNCSock, METH_VARARGS, NULL},
	 { "delete_CZNCSock", _wrap_delete_CZNCSock, METH_O, NULL},
	 { "CZNCSock_ConvertAddress", _wrap_CZNCSock_ConvertAddress, METH_VARARGS, NULL},
	 { "CZNCSock_GetSSLPeerFingerprint", _wrap_CZNCSock_GetSSLPeerFingerprint, METH_O, NULL},
	 { "CZNCSock_SetHostToVerifySSL", _wrap_CZNCSock_SetHostToVerifySSL, METH_VARARGS, NULL},
	 { "CZNCSock_SetSSLTrustedPeerFingerprints", _wrap_CZNCSock_SetSSLTrustedPeerFingerprints, METH_VARARGS, NULL},
	 { "CZNCSock_SetTrustAllCerts", _wrap_CZNCSock_SetTrustAllCerts, METH_VARARGS, NULL},
	 { "CZNCSock_GetTrustAllCerts", _wrap_CZNCSock_GetTrustAllCerts, METH_O, NULL},
	 { "CZNCSock_SetTrustPKI", _wrap_CZNCSock_SetTrustPKI, METH_VARARGS, NULL},
	 { "CZNCSock_GetTrustPKI", _wrap_CZNCSock_GetTrustPKI, METH_O, NULL},
	 { "CZNCSock_SetEncoding", _wrap_CZNCSock_SetEncoding, METH_VARARGS, NULL},
	 { "CZNCSock_GetRemoteIP", _wrap_CZNCSock_GetRemoteIP, METH_O, NULL},
	 { "CZNCSock_swigregister", CZNCSock_swigregister, METH_O, NULL},
	 { "CZNCSock_swiginit", CZNCSock_swiginit, METH_VARARGS, NULL},
	 { "new_CSockManager", _wrap_new_CSockManager, METH_NOARGS, NULL},
	 { "delete_CSockManager", _wrap_delete_CSockManager, METH_O, NULL},
	 { "CSockManager_ListenHost", _wrap_CSockManager_ListenHost, METH_VARARGS, NULL},
	 { "CSockManager_ListenAll", _wrap_CSockManager_ListenAll, METH_VARARGS, NULL},
	 { "CSockManager_ListenRand", _wrap_CSockManager_ListenRand, METH_VARARGS, NULL},
	 { "CSockManager_ListenAllRand", _wrap_CSockManager_ListenAllRand, METH_VARARGS, NULL},
	 { "CSockManager_Connect", _wrap_CSockManager_Connect, METH_VARARGS, NULL},
	 { "CSockManager_ListenUnix", _wrap_CSockManager_ListenUnix, METH_VARARGS, NULL},
	 { "CSockManager_ConnectUnix", _wrap_CSockManager_ConnectUnix, METH_VARARGS, NULL},
	 { "CSockManager_GetAnonConnectionCount", _wrap_CSockManager_GetAnonConnectionCount, METH_VARARGS, NULL},
	 { "CSockManager_DelSockByAddr", _wrap_CSockManager_DelSockByAddr, METH_VARARGS, NULL},
	 { "CSockManager_swigregister", CSockManager_swigregister, METH_O, NULL},
	 { "CSockManager_swiginit", CSockManager_swiginit, METH_VARARGS, NULL},
	 { "new_CSocket", _wrap_new_CSocket, METH_VARARGS, NULL},
	 { "delete_CSocket", _wrap_delete_CSocket, METH_O, NULL},
	 { "CSocket_ReachedMaxBuffer", _wrap_CSocket_ReachedMaxBuffer, METH_O, NULL},
	 { "CSocket_SockError", _wrap_CSocket_SockError, METH_VARARGS, NULL},
	 { "CSocket_ConnectionFrom", _wrap_CSocket_ConnectionFrom, METH_VARARGS, NULL},
	 { "CSocket_Connect", _wrap_CSocket_Connect, METH_VARARGS, NULL},
	 { "CSocket_Listen", _wrap_CSocket_Listen, METH_VARARGS, NULL},
	 { "CSocket_ConnectUnix", _wrap_CSocket_ConnectUnix, METH_VARARGS, NULL},
	 { "CSocket_ListenUnix", _wrap_CSocket_ListenUnix, METH_VARARGS, NULL},
	 { "CSocket_ConstructSockName", _wrap_CSocket_ConstructSockName, METH_VARARGS, NULL},
	 { "CSocket_GetModule", _wrap_CSocket_GetModule, METH_O, NULL},
	 { "CSocket_swigregister", CSocket_swigregister, METH_O, NULL},
	 { "CSocket_swiginit", CSocket_swiginit, METH_VARARGS, NULL},
	 { "new_CIRCSocket", _wrap_new_CIRCSocket, METH_NOARGS, NULL},
	 { "delete_CIRCSocket", _wrap_delete_CIRCSocket, METH_O, NULL},
	 { "CIRCSocket_swigregister", CIRCSocket_swigregister, METH_O, NULL},
	 { "CIRCSocket_swiginit", CIRCSocket_swiginit, METH_VARARGS, NULL},
	 { "new_CFile", _wrap_new_CFile, METH_VARARGS, NULL},
	 { "delete_CFile", _wrap_delete_CFile, METH_O, NULL},
	 { "CFile_SetFileName", _wrap_CFile_SetFileName, METH_VARARGS, NULL},
	 { "CFile_IsReg", _wrap_CFile_IsReg, METH_VARARGS, NULL},
	 { "CFile_IsDir", _wrap_CFile_IsDir, METH_VARARGS, NULL},
	 { "CFile_IsChr", _wrap_CFile_IsChr, METH_VARARGS, NULL},
	 { "CFile_IsBlk", _wrap_CFile_IsBlk, METH_VARARGS, NULL},
	 { "CFile_IsFifo", _wrap_CFile_IsFifo, METH_VARARGS, NULL},
	 { "CFile_IsLnk", _wrap_CFile_IsLnk, METH_VARARGS, NULL},
	 { "CFile_IsSock", _wrap_CFile_IsSock, METH_VARARGS, NULL},
	 { "CFile_FType", _wrap_CFile_FType, METH_VARARGS, NULL},
	 { "CFile_Exists", _wrap_CFile_Exists, METH_VARARGS, NULL},
	 { "CFile_GetSize", _wrap_CFile_GetSize, METH_VARARGS, NULL},
	 { "CFile_GetATime", _wrap_CFile_GetATime, METH_VARARGS, NULL},
	 { "CFile_GetMTime", _wrap_CFile_GetMTime, METH_VARARGS, NULL},
	 { "CFile_GetCTime", _wrap_CFile_GetCTime, METH_VARARGS, NULL},
	 { "CFile_GetUID", _wrap_CFile_GetUID, METH_VARARGS, NULL},
	 { "CFile_GetGID", _wrap_CFile_GetGID, METH_VARARGS, NULL},
	 { "CFile_GetInfo", _wrap_CFile_GetInfo, METH_VARARGS, NULL},
	 { "CFile_Delete", _wrap_CFile_Delete, METH_VARARGS, NULL},
	 { "CFile_Move", _wrap_CFile_Move, METH_VARARGS, NULL},
	 { "CFile_Copy", _wrap_CFile_Copy, METH_VARARGS, NULL},
	 { "CFile_Chmod", _wrap_CFile_Chmod, METH_VARARGS, NULL},
	 { "CFile_Seek", _wrap_CFile_Seek, METH_VARARGS, NULL},
	 { "CFile_Truncate", _wrap_CFile_Truncate, METH_O, NULL},
	 { "CFile_Sync", _wrap_CFile_Sync, METH_O, NULL},
	 { "CFile_Open", _wrap_CFile_Open, METH_VARARGS, NULL},
	 { "CFile_Read", _wrap_CFile_Read, METH_VARARGS, NULL},
	 { "CFile_ReadLine", _wrap_CFile_ReadLine, METH_VARARGS, NULL},
	 { "CFile_ReadFile", _wrap_CFile_ReadFile, METH_VARARGS, NULL},
	 { "CFile_Write", _wrap_CFile_Write, METH_VARARGS, NULL},
	 { "CFile_Close", _wrap_CFile_Close, METH_O, NULL},
	 { "CFile_ClearBuffer", _wrap_CFile_ClearBuffer, METH_O, NULL},
	 { "CFile_TryExLock", _wrap_CFile_TryExLock, METH_VARARGS, NULL},
	 { "CFile_ExLock", _wrap_CFile_ExLock, METH_O, NULL},
	 { "CFile_UnLock", _wrap_CFile_UnLock, METH_O, NULL},
	 { "CFile_IsOpen", _wrap_CFile_IsOpen, METH_O, NULL},
	 { "CFile_GetLongName", _wrap_CFile_GetLongName, METH_O, NULL},
	 { "CFile_GetShortName", _wrap_CFile_GetShortName, METH_O, NULL},
	 { "CFile_GetDir", _wrap_CFile_GetDir, METH_O, NULL},
	 { "CFile_HadError", _wrap_CFile_HadError, METH_O, NULL},
	 { "CFile_ResetError", _wrap_CFile_ResetError, METH_O, NULL},
	 { "CFile_InitHomePath", _wrap_CFile_InitHomePath, METH_O, NULL},
	 { "CFile_GetHomePath", _wrap_CFile_GetHomePath, METH_NOARGS, NULL},
	 { "CFile_swigregister", CFile_swigregister, METH_O, NULL},
	 { "CFile_swiginit", CFile_swiginit, METH_VARARGS, NULL},
	 { "new_CDir", _wrap_new_CDir, METH_VARARGS, NULL},
	 { "delete_CDir", _wrap_delete_CDir, METH_O, NULL},
	 { "CDir_CleanUp", _wrap_CDir_CleanUp, METH_O, NULL},
	 { "CDir_Fill", _wrap_CDir_Fill, METH_VARARGS, NULL},
	 { "CDir_FillByWildcard", _wrap_CDir_FillByWildcard, METH_VARARGS, NULL},
	 { "CDir_Chmod", _wrap_CDir_Chmod, METH_VARARGS, NULL},
	 { "CDir_Delete", _wrap_CDir_Delete, METH_VARARGS, NULL},
	 { "CDir_GetSortAttr", _wrap_CDir_GetSortAttr, METH_O, NULL},
	 { "CDir_IsDescending", _wrap_CDir_IsDescending, METH_O, NULL},
	 { "CDir_CheckPathPrefix", _wrap_CDir_CheckPathPrefix, METH_VARARGS, NULL},
	 { "CDir_ChangeDir", _wrap_CDir_ChangeDir, METH_VARARGS, NULL},
	 { "CDir_MakeDir", _wrap_CDir_MakeDir, METH_VARARGS, NULL},
	 { "CDir_GetCWD", _wrap_CDir_GetCWD, METH_NOARGS, NULL},
	 { "CDir_swigregister", CDir_swigregister, METH_O, NULL},
	 { "CDir_swiginit", CDir_swiginit, METH_VARARGS, NULL},
	 { "new_CMessage", _wrap_new_CMessage, METH_VARARGS, NULL},
	 { "CMessage_GetType", _wrap_CMessage_GetType, METH_O, NULL},
	 { "CMessage_Equals", _wrap_CMessage_Equals, METH_VARARGS, NULL},
	 { "CMessage_Clone", _wrap_CMessage_Clone, METH_VARARGS, NULL},
	 { "CMessage_GetNetwork", _wrap_CMessage_GetNetwork, METH_O, NULL},
	 { "CMessage_SetNetwork", _wrap_CMessage_SetNetwork, METH_VARARGS, NULL},
	 { "CMessage_GetClient", _wrap_CMessage_GetClient, METH_O, NULL},
	 { "CMessage_SetClient", _wrap_CMessage_SetClient, METH_VARARGS, NULL},
	 { "CMessage_GetChan", _wrap_CMessage_GetChan, METH_O, NULL},
	 { "CMessage_SetChan", _wrap_CMessage_SetChan, METH_VARARGS, NULL},
	 { "CMessage_GetNick", _wrap_CMessage_GetNick, METH_VARARGS, NULL},
	 { "CMessage_SetNick", _wrap_CMessage_SetNick, METH_VARARGS, NULL},
	 { "CMessage_GetCommand", _wrap_CMessage_GetCommand, METH_O, NULL},
	 { "CMessage_SetCommand", _wrap_CMessage_SetCommand, METH_VARARGS, NULL},
	 { "CMessage_GetParamsSplit", _wrap_CMessage_GetParamsSplit, METH_VARARGS, NULL},
	 { "CMessage_SetParams", _wrap_CMessage_SetParams, METH_VARARGS, NULL},
	 { "CMessage_GetParams", _wrap_CMessage_GetParams, METH_VARARGS, NULL},
	 { "CMessage_GetParamsColon", _wrap_CMessage_GetParamsColon, METH_VARARGS, NULL},
	 { "CMessage_GetParam", _wrap_CMessage_GetParam, METH_VARARGS, NULL},
	 { "CMessage_SetParam", _wrap_CMessage_SetParam, METH_VARARGS, NULL},
	 { "CMessage_GetTime", _wrap_CMessage_GetTime, METH_O, NULL},
	 { "CMessage_SetTime", _wrap_CMessage_SetTime, METH_VARARGS, NULL},
	 { "CMessage_GetTags", _wrap_CMessage_GetTags, METH_VARARGS, NULL},
	 { "CMessage_SetTags", _wrap_CMessage_SetTags, METH_VARARGS, NULL},
	 { "CMessage_GetTag", _wrap_CMessage_GetTag, METH_VARARGS, NULL},
	 { "CMessage_SetTag", _wrap_CMessage_SetTag, METH_VARARGS, NULL},
	 { "CMessage_ToString", _wrap_CMessage_ToString, METH_VARARGS, NULL},
	 { "CMessage_Parse", _wrap_CMessage_Parse, METH_VARARGS, NULL},
	 { "CMessage_As_CTargetMessage", _wrap_CMessage_As_CTargetMessage, METH_O, NULL},
	 { "CMessage_As_CActionMessage", _wrap_CMessage_As_CActionMessage, METH_O, NULL},
	 { "CMessage_As_CAuthenticateMessage", _wrap_CMessage_As_CAuthenticateMessage, METH_O, NULL},
	 { "CMessage_As_CCTCPMessage", _wrap_CMessage_As_CCTCPMessage, METH_O, NULL},
	 { "CMessage_As_CJoinMessage", _wrap_CMessage_As_CJoinMessage, METH_O, NULL},
	 { "CMessage_As_CModeMessage", _wrap_CMessage_As_CModeMessage, METH_O, NULL},
	 { "CMessage_As_CNickMessage", _wrap_CMessage_As_CNickMessage, METH_O, NULL},
	 { "CMessage_As_CNoticeMessage", _wrap_CMessage_As_CNoticeMessage, METH_O, NULL},
	 { "CMessage_As_CNumericMessage", _wrap_CMessage_As_CNumericMessage, METH_O, NULL},
	 { "CMessage_As_CKickMessage", _wrap_CMessage_As_CKickMessage, METH_O, NULL},
	 { "CMessage_As_CInviteMessage", _wrap_CMessage_As_CInviteMessage, METH_O, NULL},
	 { "CMessage_As_CPartMessage", _wrap_CMessage_As_CPartMessage, METH_O, NULL},
	 { "CMessage_As_CQuitMessage", _wrap_CMessage_As_CQuitMessage, METH_O, NULL},
	 { "CMessage_As_CTextMessage", _wrap_CMessage_As_CTextMessage, METH_O, NULL},
	 { "CMessage_As_CTopicMessage", _wrap_CMessage_As_CTopicMessage, METH_O, NULL},
	 { "CMessage_As_CChgHostMessage", _wrap_CMessage_As_CChgHostMessage, METH_O, NULL},
	 { "CMessage___str__", _wrap_CMessage___str__, METH_O, NULL},
	 { "CMessage___repr__", _wrap_CMessage___repr__, METH_O, NULL},
	 { "delete_CMessage", _wrap_delete_CMessage, METH_O, NULL},
	 { "CMessage_swigregister", CMessage_swigregister, METH_O, NULL},
	 { "CMessage_swiginit", CMessage_swiginit, METH_VARARGS, NULL},
	 { "CTargetMessage_GetTarget", _wrap_CTargetMessage_GetTarget, METH_O, NULL},
	 { "CTargetMessage_SetTarget", _wrap_CTargetMessage_SetTarget, METH_VARARGS, NULL},
	 { "new_CTargetMessage", _wrap_new_CTargetMessage, METH_NOARGS, NULL},
	 { "delete_CTargetMessage", _wrap_delete_CTargetMessage, METH_O, NULL},
	 { "CTargetMessage_swigregister", CTargetMessage_swigregister, METH_O, NULL},
	 { "CTargetMessage_swiginit", CTargetMessage_swiginit, METH_VARARGS, NULL},
	 { "CActionMessage_GetText", _wrap_CActionMessage_GetText, METH_O, NULL},
	 { "CActionMessage_SetText", _wrap_CActionMessage_SetText, METH_VARARGS, NULL},
	 { "new_CActionMessage", _wrap_new_CActionMessage, METH_NOARGS, NULL},
	 { "delete_CActionMessage", _wrap_delete_CActionMessage, METH_O, NULL},
	 { "CActionMessage_swigregister", CActionMessage_swigregister, METH_O, NULL},
	 { "CActionMessage_swiginit", CActionMessage_swiginit, METH_VARARGS, NULL},
	 { "CAuthenticateMessage_GetText", _wrap_CAuthenticateMessage_GetText, METH_O, NULL},
	 { "CAuthenticateMessage_SetText", _wrap_CAuthenticateMessage_SetText, METH_VARARGS, NULL},
	 { "new_CAuthenticateMessage", _wrap_new_CAuthenticateMessage, METH_NOARGS, NULL},
	 { "delete_CAuthenticateMessage", _wrap_delete_CAuthenticateMessage, METH_O, NULL},
	 { "CAuthenticateMessage_swigregister", CAuthenticateMessage_swigregister, METH_O, NULL},
	 { "CAuthenticateMessage_swiginit", CAuthenticateMessage_swiginit, METH_VARARGS, NULL},
	 { "CCTCPMessage_IsReply", _wrap_CCTCPMessage_IsReply, METH_O, NULL},
	 { "CCTCPMessage_GetText", _wrap_CCTCPMessage_GetText, METH_O, NULL},
	 { "CCTCPMessage_SetText", _wrap_CCTCPMessage_SetText, METH_VARARGS, NULL},
	 { "new_CCTCPMessage", _wrap_new_CCTCPMessage, METH_NOARGS, NULL},
	 { "delete_CCTCPMessage", _wrap_delete_CCTCPMessage, METH_O, NULL},
	 { "CCTCPMessage_swigregister", CCTCPMessage_swigregister, METH_O, NULL},
	 { "CCTCPMessage_swiginit", CCTCPMessage_swiginit, METH_VARARGS, NULL},
	 { "CJoinMessage_GetKey", _wrap_CJoinMessage_GetKey, METH_O, NULL},
	 { "CJoinMessage_SetKey", _wrap_CJoinMessage_SetKey, METH_VARARGS, NULL},
	 { "new_CJoinMessage", _wrap_new_CJoinMessage, METH_NOARGS, NULL},
	 { "delete_CJoinMessage", _wrap_delete_CJoinMessage, METH_O, NULL},
	 { "CJoinMessage_swigregister", CJoinMessage_swigregister, METH_O, NULL},
	 { "CJoinMessage_swiginit", CJoinMessage_swiginit, METH_VARARGS, NULL},
	 { "CModeMessage_GetModes", _wrap_CModeMessage_GetModes, METH_O, NULL},
	 { "CModeMessage_GetModeList", _wrap_CModeMessage_GetModeList, METH_O, NULL},
	 { "CModeMessage_GetModeParams", _wrap_CModeMessage_GetModeParams, METH_O, NULL},
	 { "CModeMessage_HasModes", _wrap_CModeMessage_HasModes, METH_O, NULL},
	 { "new_CModeMessage", _wrap_new_CModeMessage, METH_NOARGS, NULL},
	 { "delete_CModeMessage", _wrap_delete_CModeMessage, METH_O, NULL},
	 { "CModeMessage_swigregister", CModeMessage_swigregister, METH_O, NULL},
	 { "CModeMessage_swiginit", CModeMessage_swiginit, METH_VARARGS, NULL},
	 { "CNickMessage_GetOldNick", _wrap_CNickMessage_GetOldNick, METH_O, NULL},
	 { "CNickMessage_GetNewNick", _wrap_CNickMessage_GetNewNick, METH_O, NULL},
	 { "CNickMessage_SetNewNick", _wrap_CNickMessage_SetNewNick, METH_VARARGS, NULL},
	 { "new_CNickMessage", _wrap_new_CNickMessage, METH_NOARGS, NULL},
	 { "delete_CNickMessage", _wrap_delete_CNickMessage, METH_O, NULL},
	 { "CNickMessage_swigregister", CNickMessage_swigregister, METH_O, NULL},
	 { "CNickMessage_swiginit", CNickMessage_swiginit, METH_VARARGS, NULL},
	 { "CNoticeMessage_GetText", _wrap_CNoticeMessage_GetText, METH_O, NULL},
	 { "CNoticeMessage_SetText", _wrap_CNoticeMessage_SetText, METH_VARARGS, NULL},
	 { "new_CNoticeMessage", _wrap_new_CNoticeMessage, METH_NOARGS, NULL},
	 { "delete_CNoticeMessage", _wrap_delete_CNoticeMessage, METH_O, NULL},
	 { "CNoticeMessage_swigregister", CNoticeMessage_swigregister, METH_O, NULL},
	 { "CNoticeMessage_swiginit", CNoticeMessage_swiginit, METH_VARARGS, NULL},
	 { "CNumericMessage_GetCode", _wrap_CNumericMessage_GetCode, METH_O, NULL},
	 { "new_CNumericMessage", _wrap_new_CNumericMessage, METH_NOARGS, NULL},
	 { "delete_CNumericMessage", _wrap_delete_CNumericMessage, METH_O, NULL},
	 { "CNumericMessage_swigregister", CNumericMessage_swigregister, METH_O, NULL},
	 { "CNumericMessage_swiginit", CNumericMessage_swiginit, METH_VARARGS, NULL},
	 { "CKickMessage_GetKickedNick", _wrap_CKickMessage_GetKickedNick, METH_O, NULL},
	 { "CKickMessage_SetKickedNick", _wrap_CKickMessage_SetKickedNick, METH_VARARGS, NULL},
	 { "CKickMessage_GetReason", _wrap_CKickMessage_GetReason, METH_O, NULL},
	 { "CKickMessage_SetReason", _wrap_CKickMessage_SetReason, METH_VARARGS, NULL},
	 { "CKickMessage_GetText", _wrap_CKickMessage_GetText, METH_O, NULL},
	 { "CKickMessage_SetText", _wrap_CKickMessage_SetText, METH_VARARGS, NULL},
	 { "new_CKickMessage", _wrap_new_CKickMessage, METH_NOARGS, NULL},
	 { "delete_CKickMessage", _wrap_delete_CKickMessage, METH_O, NULL},
	 { "CKickMessage_swigregister", CKickMessage_swigregister, METH_O, NULL},
	 { "CKickMessage_swiginit", CKickMessage_swiginit, METH_VARARGS, NULL},
	 { "CInviteMessage_GetInvitedNick", _wrap_CInviteMessage_GetInvitedNick, METH_O, NULL},
	 { "CInviteMessage_SetInvitedNick", _wrap_CInviteMessage_SetInvitedNick, METH_VARARGS, NULL},
	 { "CInviteMessage_GetChannel", _wrap_CInviteMessage_GetChannel, METH_O, NULL},
	 { "CInviteMessage_SetChannel", _wrap_CInviteMessage_SetChannel, METH_VARARGS, NULL},
	 { "new_CInviteMessage", _wrap_new_CInviteMessage, METH_NOARGS, NULL},
	 { "delete_CInviteMessage", _wrap_delete_CInviteMessage, METH_O, NULL},
	 { "CInviteMessage_swigregister", CInviteMessage_swigregister, METH_O, NULL},
	 { "CInviteMessage_swiginit", CInviteMessage_swiginit, METH_VARARGS, NULL},
	 { "CPartMessage_GetReason", _wrap_CPartMessage_GetReason, METH_O, NULL},
	 { "CPartMessage_SetReason", _wrap_CPartMessage_SetReason, METH_VARARGS, NULL},
	 { "CPartMessage_GetText", _wrap_CPartMessage_GetText, METH_O, NULL},
	 { "CPartMessage_SetText", _wrap_CPartMessage_SetText, METH_VARARGS, NULL},
	 { "new_CPartMessage", _wrap_new_CPartMessage, METH_NOARGS, NULL},
	 { "delete_CPartMessage", _wrap_delete_CPartMessage, METH_O, NULL},
	 { "CPartMessage_swigregister", CPartMessage_swigregister, METH_O, NULL},
	 { "CPartMessage_swiginit", CPartMessage_swiginit, METH_VARARGS, NULL},
	 { "CQuitMessage_GetReason", _wrap_CQuitMessage_GetReason, METH_O, NULL},
	 { "CQuitMessage_SetReason", _wrap_CQuitMessage_SetReason, METH_VARARGS, NULL},
	 { "CQuitMessage_GetText", _wrap_CQuitMessage_GetText, METH_O, NULL},
	 { "CQuitMessage_SetText", _wrap_CQuitMessage_SetText, METH_VARARGS, NULL},
	 { "new_CQuitMessage", _wrap_new_CQuitMessage, METH_NOARGS, NULL},
	 { "delete_CQuitMessage", _wrap_delete_CQuitMessage, METH_O, NULL},
	 { "CQuitMessage_swigregister", CQuitMessage_swigregister, METH_O, NULL},
	 { "CQuitMessage_swiginit", CQuitMessage_swiginit, METH_VARARGS, NULL},
	 { "CTextMessage_GetText", _wrap_CTextMessage_GetText, METH_O, NULL},
	 { "CTextMessage_SetText", _wrap_CTextMessage_SetText, METH_VARARGS, NULL},
	 { "new_CTextMessage", _wrap_new_CTextMessage, METH_NOARGS, NULL},
	 { "delete_CTextMessage", _wrap_delete_CTextMessage, METH_O, NULL},
	 { "CTextMessage_swigregister", CTextMessage_swigregister, METH_O, NULL},
	 { "CTextMessage_swiginit", CTextMessage_swiginit, METH_VARARGS, NULL},
	 { "CTopicMessage_GetTopic", _wrap_CTopicMessage_GetTopic, METH_O, NULL},
	 { "CTopicMessage_SetTopic", _wrap_CTopicMessage_SetTopic, METH_VARARGS, NULL},
	 { "CTopicMessage_GetText", _wrap_CTopicMessage_GetText, METH_O, NULL},
	 { "CTopicMessage_SetText", _wrap_CTopicMessage_SetText, METH_VARARGS, NULL},
	 { "new_CTopicMessage", _wrap_new_CTopicMessage, METH_NOARGS, NULL},
	 { "delete_CTopicMessage", _wrap_delete_CTopicMessage, METH_O, NULL},
	 { "CTopicMessage_swigregister", CTopicMessage_swigregister, METH_O, NULL},
	 { "CTopicMessage_swiginit", CTopicMessage_swiginit, METH_VARARGS, NULL},
	 { "CChgHostMessage_GetNewIdent", _wrap_CChgHostMessage_GetNewIdent, METH_O, NULL},
	 { "CChgHostMessage_SetNewIdent", _wrap_CChgHostMessage_SetNewIdent, METH_VARARGS, NULL},
	 { "CChgHostMessage_GetNewHost", _wrap_CChgHostMessage_GetNewHost, METH_O, NULL},
	 { "CChgHostMessage_SetNewHost", _wrap_CChgHostMessage_SetNewHost, METH_VARARGS, NULL},
	 { "new_CChgHostMessage", _wrap_new_CChgHostMessage, METH_NOARGS, NULL},
	 { "delete_CChgHostMessage", _wrap_delete_CChgHostMessage, METH_O, NULL},
	 { "CChgHostMessage_swigregister", CChgHostMessage_swigregister, METH_O, NULL},
	 { "CChgHostMessage_swiginit", CChgHostMessage_swiginit, METH_VARARGS, NULL},
	 { "CModuleEntry_pcVersion_set", _wrap_CModuleEntry_pcVersion_set, METH_VARARGS, NULL},
	 { "CModuleEntry_pcVersion_get", _wrap_CModuleEntry_pcVersion_get, METH_O, NULL},
	 { "CModuleEntry_pcVersionExtra_set", _wrap_CModuleEntry_pcVersionExtra_set, METH_VARARGS, NULL},
	 { "CModuleEntry_pcVersionExtra_get", _wrap_CModuleEntry_pcVersionExtra_get, METH_O, NULL},
	 { "CModuleEntry_pcCompileOptions_set", _wrap_CModuleEntry_pcCompileOptions_set, METH_VARARGS, NULL},
	 { "CModuleEntry_pcCompileOptions_get", _wrap_CModuleEntry_pcCompileOptions_get, METH_O, NULL},
	 { "CModuleEntry_fpFillModInfo_set", _wrap_CModuleEntry_fpFillModInfo_set, METH_VARARGS, NULL},
	 { "CModuleEntry_fpFillModInfo_get", _wrap_CModuleEntry_fpFillModInfo_get, METH_O, NULL},
	 { "new_CModuleEntry", _wrap_new_CModuleEntry, METH_NOARGS, NULL},
	 { "delete_CModuleEntry", _wrap_delete_CModuleEntry, METH_O, NULL},
	 { "CModuleEntry_swigregister", CModuleEntry_swigregister, METH_O, NULL},
	 { "CModuleEntry_swiginit", CModuleEntry_swiginit, METH_VARARGS, NULL},
	 { "delete_CCapability", _wrap_delete_CCapability, METH_O, NULL},
	 { "CCapability_OnServerChangedSupport", _wrap_CCapability_OnServerChangedSupport, METH_VARARGS, NULL},
	 { "CCapability_OnClientChangedSupport", _wrap_CCapability_OnClientChangedSupport, METH_VARARGS, NULL},
	 { "CCapability_GetModule", _wrap_CCapability_GetModule, METH_O, NULL},
	 { "CCapability_SetModule", _wrap_CCapability_SetModule, METH_VARARGS, NULL},
	 { "new_CCapability", _wrap_new_CCapability, METH_NOARGS, NULL},
	 { "CCapability_swigregister", CCapability_swigregister, METH_O, NULL},
	 { "CCapability_swiginit", CCapability_swiginit, METH_VARARGS, NULL},
	 { "new_CTimer", _wrap_new_CTimer, METH_VARARGS, NULL},
	 { "delete_CTimer", _wrap_delete_CTimer, METH_O, NULL},
	 { "CTimer_SetModule", _wrap_CTimer_SetModule, METH_VARARGS, NULL},
	 { "CTimer_SetDescription", _wrap_CTimer_SetDescription, METH_VARARGS, NULL},
	 { "CTimer_GetModule", _wrap_CTimer_GetModule, METH_O, NULL},
	 { "CTimer_GetDescription", _wrap_CTimer_GetDescription, METH_O, NULL},
	 { "CTimer_swigregister", CTimer_swigregister, METH_O, NULL},
	 { "CTimer_swiginit", CTimer_swiginit, METH_VARARGS, NULL},
	 { "new_CFPTimer", _wrap_new_CFPTimer, METH_VARARGS, NULL},
	 { "delete_CFPTimer", _wrap_delete_CFPTimer, METH_O, NULL},
	 { "CFPTimer_SetFPCallback", _wrap_CFPTimer_SetFPCallback, METH_VARARGS, NULL},
	 { "CFPTimer_swigregister", CFPTimer_swigregister, METH_O, NULL},
	 { "CFPTimer_swiginit", CFPTimer_swiginit, METH_VARARGS, NULL},
	 { "new_CModInfo", _wrap_new_CModInfo, METH_VARARGS, NULL},
	 { "delete_CModInfo", _wrap_delete_CModInfo, METH_O, NULL},
	 { "CModInfo___lt__", _wrap_CModInfo___lt__, METH_VARARGS, NULL},
	 { "CModInfo_SupportsType", _wrap_CModInfo_SupportsType, METH_VARARGS, NULL},
	 { "CModInfo_AddType", _wrap_CModInfo_AddType, METH_VARARGS, NULL},
	 { "CModInfo_ModuleTypeToString", _wrap_CModInfo_ModuleTypeToString, METH_O, NULL},
	 { "CModInfo_GetName", _wrap_CModInfo_GetName, METH_O, NULL},
	 { "CModInfo_GetPath", _wrap_CModInfo_GetPath, METH_O, NULL},
	 { "CModInfo_GetDescription", _wrap_CModInfo_GetDescription, METH_O, NULL},
	 { "CModInfo_GetWikiPage", _wrap_CModInfo_GetWikiPage, METH_O, NULL},
	 { "CModInfo_GetArgsHelpText", _wrap_CModInfo_GetArgsHelpText, METH_O, NULL},
	 { "CModInfo_GetHasArgs", _wrap_CModInfo_GetHasArgs, METH_O, NULL},
	 { "CModInfo_GetLoader", _wrap_CModInfo_GetLoader, METH_O, NULL},
	 { "CModInfo_GetDefaultType", _wrap_CModInfo_GetDefaultType, METH_O, NULL},
	 { "CModInfo_SetName", _wrap_CModInfo_SetName, METH_VARARGS, NULL},
	 { "CModInfo_SetPath", _wrap_CModInfo_SetPath, METH_VARARGS, NULL},
	 { "CModInfo_SetDescription", _wrap_CModInfo_SetDescription, METH_VARARGS, NULL},
	 { "CModInfo_SetWikiPage", _wrap_CModInfo_SetWikiPage, METH_VARARGS, NULL},
	 { "CModInfo_SetArgsHelpText", _wrap_CModInfo_SetArgsHelpText, METH_VARARGS, NULL},
	 { "CModInfo_SetHasArgs", _wrap_CModInfo_SetHasArgs, METH_VARARGS, NULL},
	 { "CModInfo_SetLoader", _wrap_CModInfo_SetLoader, METH_VARARGS, NULL},
	 { "CModInfo_SetDefaultType", _wrap_CModInfo_SetDefaultType, METH_VARARGS, NULL},
	 { "CModInfo_t_s", _wrap_CModInfo_t_s, METH_VARARGS, NULL},
	 { "CModInfo_swigregister", CModInfo_swigregister, METH_O, NULL},
	 { "CModInfo_swiginit", CModInfo_swiginit, METH_VARARGS, NULL},
	 { "new_CModCommand", _wrap_new_CModCommand, METH_VARARGS, NULL},
	 { "CModCommand_InitHelp", _wrap_CModCommand_InitHelp, METH_O, NULL},
	 { "CModCommand_AddHelp", _wrap_CModCommand_AddHelp, METH_VARARGS, NULL},
	 { "CModCommand_GetCommand", _wrap_CModCommand_GetCommand, METH_O, NULL},
	 { "CModCommand_GetFunction", _wrap_CModCommand_GetFunction, METH_O, NULL},
	 { "CModCommand_GetArgs", _wrap_CModCommand_GetArgs, METH_O, NULL},
	 { "CModCommand_GetDescription", _wrap_CModCommand_GetDescription, METH_O, NULL},
	 { "CModCommand_Call", _wrap_CModCommand_Call, METH_VARARGS, NULL},
	 { "delete_CModCommand", _wrap_delete_CModCommand, METH_O, NULL},
	 { "CModCommand_swigregister", CModCommand_swigregister, METH_O, NULL},
	 { "CModCommand_swiginit", CModCommand_swiginit, METH_VARARGS, NULL},
	 { "new_CModule", _wrap_new_CModule, METH_VARARGS, NULL},
	 { "delete_CModule", _wrap_delete_CModule, METH_O, NULL},
	 { "CModule_SetUser", _wrap_CModule_SetUser, METH_VARARGS, NULL},
	 { "CModule_SetNetwork", _wrap_CModule_SetNetwork, METH_VARARGS, NULL},
	 { "CModule_SetClient", _wrap_CModule_SetClient, METH_VARARGS, NULL},
	 { "CModule_Unload", _wrap_CModule_Unload, METH_O, NULL},
	 { "CModule_OnLoad", _wrap_CModule_OnLoad, METH_VARARGS, NULL},
	 { "CModule_OnBoot", _wrap_CModule_OnBoot, METH_O, NULL},
	 { "CModule_WebRequiresLogin", _wrap_CModule_WebRequiresLogin, METH_O, NULL},
	 { "CModule_WebRequiresAdmin", _wrap_CModule_WebRequiresAdmin, METH_O, NULL},
	 { "CModule_GetWebMenuTitle", _wrap_CModule_GetWebMenuTitle, METH_O, NULL},
	 { "CModule_GetWebPath", _wrap_CModule_GetWebPath, METH_O, NULL},
	 { "CModule_GetWebFilesPath", _wrap_CModule_GetWebFilesPath, METH_O, NULL},
	 { "CModule_OnWebPreRequest", _wrap_CModule_OnWebPreRequest, METH_VARARGS, NULL},
	 { "CModule_OnWebRequest", _wrap_CModule_OnWebRequest, METH_VARARGS, NULL},
	 { "CModule_ValidateWebRequestCSRFCheck", _wrap_CModule_ValidateWebRequestCSRFCheck, METH_VARARGS, NULL},
	 { "CModule_AddSubPage", _wrap_CModule_AddSubPage, METH_VARARGS, NULL},
	 { "CModule_ClearSubPages", _wrap_CModule_ClearSubPages, METH_O, NULL},
	 { "CModule_GetSubPages", _wrap_CModule_GetSubPages, METH_O, NULL},
	 { "CModule_OnEmbeddedWebRequest", _wrap_CModule_OnEmbeddedWebRequest, METH_VARARGS, NULL},
	 { "CModule_OnPreRehash", _wrap_CModule_OnPreRehash, METH_O, NULL},
	 { "CModule_OnPostRehash", _wrap_CModule_OnPostRehash, METH_O, NULL},
	 { "CModule_OnIRCDisconnected", _wrap_CModule_OnIRCDisconnected, METH_O, NULL},
	 { "CModule_OnIRCConnected", _wrap_CModule_OnIRCConnected, METH_O, NULL},
	 { "CModule_OnIRCConnecting", _wrap_CModule_OnIRCConnecting, METH_VARARGS, NULL},
	 { "CModule_OnIRCConnectionError", _wrap_CModule_OnIRCConnectionError, METH_VARARGS, NULL},
	 { "CModule_OnIRCRegistration", _wrap_CModule_OnIRCRegistration, METH_VARARGS, NULL},
	 { "CModule_OnBroadcast", _wrap_CModule_OnBroadcast, METH_VARARGS, NULL},
	 { "CModule_OnChanPermission3", _wrap_CModule_OnChanPermission3, METH_VARARGS, NULL},
	 { "CModule_OnChanPermission2", _wrap_CModule_OnChanPermission2, METH_VARARGS, NULL},
	 { "CModule_OnChanPermission", _wrap_CModule_OnChanPermission, METH_VARARGS, NULL},
	 { "CModule_OnOp2", _wrap_CModule_OnOp2, METH_VARARGS, NULL},
	 { "CModule_OnOp", _wrap_CModule_OnOp, METH_VARARGS, NULL},
	 { "CModule_OnDeop2", _wrap_CModule_OnDeop2, METH_VARARGS, NULL},
	 { "CModule_OnDeop", _wrap_CModule_OnDeop, METH_VARARGS, NULL},
	 { "CModule_OnVoice2", _wrap_CModule_OnVoice2, METH_VARARGS, NULL},
	 { "CModule_OnVoice", _wrap_CModule_OnVoice, METH_VARARGS, NULL},
	 { "CModule_OnDevoice2", _wrap_CModule_OnDevoice2, METH_VARARGS, NULL},
	 { "CModule_OnDevoice", _wrap_CModule_OnDevoice, METH_VARARGS, NULL},
	 { "CModule_OnMode2", _wrap_CModule_OnMode2, METH_VARARGS, NULL},
	 { "CModule_OnMode", _wrap_CModule_OnMode, METH_VARARGS, NULL},
	 { "CModule_OnRawMode2", _wrap_CModule_OnRawMode2, METH_VARARGS, NULL},
	 { "CModule_OnRawMode", _wrap_CModule_OnRawMode, METH_VARARGS, NULL},
	 { "CModule_OnRaw", _wrap_CModule_OnRaw, METH_VARARGS, NULL},
	 { "CModule_OnRawMessage", _wrap_CModule_OnRawMessage, METH_VARARGS, NULL},
	 { "CModule_OnNumericMessage", _wrap_CModule_OnNumericMessage, METH_VARARGS, NULL},
	 { "CModule_OnStatusCommand", _wrap_CModule_OnStatusCommand, METH_VARARGS, NULL},
	 { "CModule_OnModCommand", _wrap_CModule_OnModCommand, METH_VARARGS, NULL},
	 { "CModule_OnUnknownModCommand", _wrap_CModule_OnUnknownModCommand, METH_VARARGS, NULL},
	 { "CModule_OnModNotice", _wrap_CModule_OnModNotice, METH_VARARGS, NULL},
	 { "CModule_OnModCTCP", _wrap_CModule_OnModCTCP, METH_VARARGS, NULL},
	 { "CModule_OnQuitMessage", _wrap_CModule_OnQuitMessage, METH_VARARGS, NULL},
	 { "CModule_OnQuit", _wrap_CModule_OnQuit, METH_VARARGS, NULL},
	 { "CModule_OnNickMessage", _wrap_CModule_OnNickMessage, METH_VARARGS, NULL},
	 { "CModule_OnNick", _wrap_CModule_OnNick, METH_VARARGS, NULL},
	 { "CModule_OnKickMessage", _wrap_CModule_OnKickMessage, METH_VARARGS, NULL},
	 { "CModule_OnKick", _wrap_CModule_OnKick, METH_VARARGS, NULL},
	 { "CModule_OnJoining", _wrap_CModule_OnJoining, METH_VARARGS, NULL},
	 { "CModule_OnJoinMessage", _wrap_CModule_OnJoinMessage, METH_VARARGS, NULL},
	 { "CModule_OnJoin", _wrap_CModule_OnJoin, METH_VARARGS, NULL},
	 { "CModule_OnPartMessage", _wrap_CModule_OnPartMessage, METH_VARARGS, NULL},
	 { "CModule_OnPart", _wrap_CModule_OnPart, METH_VARARGS, NULL},
	 { "CModule_OnInviteMessage", _wrap_CModule_OnInviteMessage, METH_VARARGS, NULL},
	 { "CModule_OnInvite", _wrap_CModule_OnInvite, METH_VARARGS, NULL},
	 { "CModule_OnChanBufferStarting", _wrap_CModule_OnChanBufferStarting, METH_VARARGS, NULL},
	 { "CModule_OnChanBufferEnding", _wrap_CModule_OnChanBufferEnding, METH_VARARGS, NULL},
	 { "CModule_OnChanBufferPlayMessage", _wrap_CModule_OnChanBufferPlayMessage, METH_VARARGS, NULL},
	 { "CModule_OnChanBufferPlayLine2", _wrap_CModule_OnChanBufferPlayLine2, METH_VARARGS, NULL},
	 { "CModule_OnChanBufferPlayLine", _wrap_CModule_OnChanBufferPlayLine, METH_VARARGS, NULL},
	 { "CModule_OnPrivBufferStarting", _wrap_CModule_OnPrivBufferStarting, METH_VARARGS, NULL},
	 { "CModule_OnPrivBufferEnding", _wrap_CModule_OnPrivBufferEnding, METH_VARARGS, NULL},
	 { "CModule_OnPrivBufferPlayMessage", _wrap_CModule_OnPrivBufferPlayMessage, METH_VARARGS, NULL},
	 { "CModule_OnPrivBufferPlayLine2", _wrap_CModule_OnPrivBufferPlayLine2, METH_VARARGS, NULL},
	 { "CModule_OnPrivBufferPlayLine", _wrap_CModule_OnPrivBufferPlayLine, METH_VARARGS, NULL},
	 { "CModule_OnClientLogin", _wrap_CModule_OnClientLogin, METH_O, NULL},
	 { "CModule_OnClientDisconnect", _wrap_CModule_OnClientDisconnect, METH_O, NULL},
	 { "CModule_OnUserRaw", _wrap_CModule_OnUserRaw, METH_VARARGS, NULL},
	 { "CModule_OnUserRawMessage", _wrap_CModule_OnUserRawMessage, METH_VARARGS, NULL},
	 { "CModule_OnUserCTCPReplyMessage", _wrap_CModule_OnUserCTCPReplyMessage, METH_VARARGS, NULL},
	 { "CModule_OnUserCTCPReply", _wrap_CModule_OnUserCTCPReply, METH_VARARGS, NULL},
	 { "CModule_OnUserCTCPMessage", _wrap_CModule_OnUserCTCPMessage, METH_VARARGS, NULL},
	 { "CModule_OnUserCTCP", _wrap_CModule_OnUserCTCP, METH_VARARGS, NULL},
	 { "CModule_OnUserActionMessage", _wrap_CModule_OnUserActionMessage, METH_VARARGS, NULL},
	 { "CModule_OnUserAction", _wrap_CModule_OnUserAction, METH_VARARGS, NULL},
	 { "CModule_OnUserTextMessage", _wrap_CModule_OnUserTextMessage, METH_VARARGS, NULL},
	 { "CModule_OnUserMsg", _wrap_CModule_OnUserMsg, METH_VARARGS, NULL},
	 { "CModule_OnUserNoticeMessage", _wrap_CModule_OnUserNoticeMessage, METH_VARARGS, NULL},
	 { "CModule_OnUserNotice", _wrap_CModule_OnUserNotice, METH_VARARGS, NULL},
	 { "CModule_OnUserJoinMessage", _wrap_CModule_OnUserJoinMessage, METH_VARARGS, NULL},
	 { "CModule_OnUserJoin", _wrap_CModule_OnUserJoin, METH_VARARGS, NULL},
	 { "CModule_OnUserPartMessage", _wrap_CModule_OnUserPartMessage, METH_VARARGS, NULL},
	 { "CModule_OnUserPart", _wrap_CModule_OnUserPart, METH_VARARGS, NULL},
	 { "CModule_OnUserTopicMessage", _wrap_CModule_OnUserTopicMessage, METH_VARARGS, NULL},
	 { "CModule_OnUserTopic", _wrap_CModule_OnUserTopic, METH_VARARGS, NULL},
	 { "CModule_OnUserTopicRequest", _wrap_CModule_OnUserTopicRequest, METH_VARARGS, NULL},
	 { "CModule_OnUserQuitMessage", _wrap_CModule_OnUserQuitMessage, METH_VARARGS, NULL},
	 { "CModule_OnUserQuit", _wrap_CModule_OnUserQuit, METH_VARARGS, NULL},
	 { "CModule_OnCTCPReplyMessage", _wrap_CModule_OnCTCPReplyMessage, METH_VARARGS, NULL},
	 { "CModule_OnCTCPReply", _wrap_CModule_OnCTCPReply, METH_VARARGS, NULL},
	 { "CModule_OnPrivCTCPMessage", _wrap_CModule_OnPrivCTCPMessage, METH_VARARGS, NULL},
	 { "CModule_OnPrivCTCP", _wrap_CModule_OnPrivCTCP, METH_VARARGS, NULL},
	 { "CModule_OnChanCTCPMessage", _wrap_CModule_OnChanCTCPMessage, METH_VARARGS, NULL},
	 { "CModule_OnChanCTCP", _wrap_CModule_OnChanCTCP, METH_VARARGS, NULL},
	 { "CModule_OnPrivActionMessage", _wrap_CModule_OnPrivActionMessage, METH_VARARGS, NULL},
	 { "CModule_OnPrivAction", _wrap_CModule_OnPrivAction, METH_VARARGS, NULL},
	 { "CModule_OnChanActionMessage", _wrap_CModule_OnChanActionMessage, METH_VARARGS, NULL},
	 { "CModule_OnChanAction", _wrap_CModule_OnChanAction, METH_VARARGS, NULL},
	 { "CModule_OnPrivTextMessage", _wrap_CModule_OnPrivTextMessage, METH_VARARGS, NULL},
	 { "CModule_OnPrivMsg", _wrap_CModule_OnPrivMsg, METH_VARARGS, NULL},
	 { "CModule_OnChanTextMessage", _wrap_CModule_OnChanTextMessage, METH_VARARGS, NULL},
	 { "CModule_OnChanMsg", _wrap_CModule_OnChanMsg, METH_VARARGS, NULL},
	 { "CModule_OnPrivNoticeMessage", _wrap_CModule_OnPrivNoticeMessage, METH_VARARGS, NULL},
	 { "CModule_OnPrivNotice", _wrap_CModule_OnPrivNotice, METH_VARARGS, NULL},
	 { "CModule_OnChanNoticeMessage", _wrap_CModule_OnChanNoticeMessage, METH_VARARGS, NULL},
	 { "CModule_OnChanNotice", _wrap_CModule_OnChanNotice, METH_VARARGS, NULL},
	 { "CModule_OnTopicMessage", _wrap_CModule_OnTopicMessage, METH_VARARGS, NULL},
	 { "CModule_OnTopic", _wrap_CModule_OnTopic, METH_VARARGS, NULL},
	 { "CModule_OnServerCapAvailable", _wrap_CModule_OnServerCapAvailable, METH_VARARGS, NULL},
	 { "CModule_OnServerCap302Available", _wrap_CModule_OnServerCap302Available, METH_VARARGS, NULL},
	 { "CModule_OnServerCapResult", _wrap_CModule_OnServerCapResult, METH_VARARGS, NULL},
	 { "CModule_OnTimerAutoJoin", _wrap_CModule_OnTimerAutoJoin, METH_VARARGS, NULL},
	 { "CModule_OnAddNetwork", _wrap_CModule_OnAddNetwork, METH_VARARGS, NULL},
	 { "CModule_OnDeleteNetwork", _wrap_CModule_OnDeleteNetwork, METH_VARARGS, NULL},
	 { "CModule_OnSendToClientMessage", _wrap_CModule_OnSendToClientMessage, METH_VARARGS, NULL},
	 { "CModule_OnSendToClient", _wrap_CModule_OnSendToClient, METH_VARARGS, NULL},
	 { "CModule_OnSendToIRCMessage", _wrap_CModule_OnSendToIRCMessage, METH_VARARGS, NULL},
	 { "CModule_OnSendToIRC", _wrap_CModule_OnSendToIRC, METH_VARARGS, NULL},
	 { "CModule_OnUserTagMessage", _wrap_CModule_OnUserTagMessage, METH_VARARGS, NULL},
	 { "CModule_OnChanTagMessage", _wrap_CModule_OnChanTagMessage, METH_VARARGS, NULL},
	 { "CModule_OnPrivTagMessage", _wrap_CModule_OnPrivTagMessage, METH_VARARGS, NULL},
	 { "CModule_GetDLL", _wrap_CModule_GetDLL, METH_O, NULL},
	 { "CModule_PutIRC", _wrap_CModule_PutIRC, METH_VARARGS, NULL},
	 { "CModule_PutUser", _wrap_CModule_PutUser, METH_VARARGS, NULL},
	 { "CModule_PutStatus", _wrap_CModule_PutStatus, METH_VARARGS, NULL},
	 { "CModule_PutModule", _wrap_CModule_PutModule, METH_VARARGS, NULL},
	 { "CModule_PutModNotice", _wrap_CModule_PutModNotice, METH_VARARGS, NULL},
	 { "CModule_GetModName", _wrap_CModule_GetModName, METH_O, NULL},
	 { "CModule_GetModNick", _wrap_CModule_GetModNick, METH_O, NULL},
	 { "CModule_GetModDataDir", _wrap_CModule_GetModDataDir, METH_O, NULL},
	 { "CModule_AddTimer", _wrap_CModule_AddTimer, METH_VARARGS, NULL},
	 { "CModule_RemTimer", _wrap_CModule_RemTimer, METH_VARARGS, NULL},
	 { "CModule_UnlinkTimer", _wrap_CModule_UnlinkTimer, METH_VARARGS, NULL},
	 { "CModule_FindTimer", _wrap_CModule_FindTimer, METH_VARARGS, NULL},
	 { "CModule_BeginTimers", _wrap_CModule_BeginTimers, METH_O, NULL},
	 { "CModule_EndTimers", _wrap_CModule_EndTimers, METH_O, NULL},
	 { "CModule_ListTimers", _wrap_CModule_ListTimers, METH_O, NULL},
	 { "CModule_AddSocket", _wrap_CModule_AddSocket, METH_VARARGS, NULL},
	 { "CModule_RemSocket", _wrap_CModule_RemSocket, METH_VARARGS, NULL},
	 { "CModule_UnlinkSocket", _wrap_CModule_UnlinkSocket, METH_VARARGS, NULL},
	 { "CModule_FindSocket", _wrap_CModule_FindSocket, METH_VARARGS, NULL},
	 { "CModule_BeginSockets", _wrap_CModule_BeginSockets, METH_O, NULL},
	 { "CModule_EndSockets", _wrap_CModule_EndSockets, METH_O, NULL},
	 { "CModule_ListSockets", _wrap_CModule_ListSockets, METH_O, NULL},
	 { "CModule_AddHelpCommand", _wrap_CModule_AddHelpCommand, METH_O, NULL},
	 { "CModule_AddCommand", _wrap_CModule_AddCommand, METH_VARARGS, NULL},
	 { "CModule_RemCommand", _wrap_CModule_RemCommand, METH_VARARGS, NULL},
	 { "CModule_FindCommand", _wrap_CModule_FindCommand, METH_VARARGS, NULL},
	 { "CModule_HandleCommand", _wrap_CModule_HandleCommand, METH_VARARGS, NULL},
	 { "CModule_HandleHelpCommand", _wrap_CModule_HandleHelpCommand, METH_VARARGS, NULL},
	 { "CModule_LoadRegistry", _wrap_CModule_LoadRegistry, METH_O, NULL},
	 { "CModule_SaveRegistry", _wrap_CModule_SaveRegistry, METH_O, NULL},
	 { "CModule_MoveRegistry", _wrap_CModule_MoveRegistry, METH_VARARGS, NULL},
	 { "CModule_SetNV", _wrap_CModule_SetNV, METH_VARARGS, NULL},
	 { "CModule_GetNV", _wrap_CModule_GetNV, METH_VARARGS, NULL},
	 { "CModule_HasNV", _wrap_CModule_HasNV, METH_VARARGS, NULL},
	 { "CModule_FindNV", _wrap_CModule_FindNV, METH_VARARGS, NULL},
	 { "CModule_EndNV", _wrap_CModule_EndNV, METH_O, NULL},
	 { "CModule_BeginNV", _wrap_CModule_BeginNV, METH_O, NULL},
	 { "CModule_DelNV", _wrap_CModule_DelNV, METH_VARARGS, NULL},
	 { "CModule_ClearNV", _wrap_CModule_ClearNV, METH_VARARGS, NULL},
	 { "CModule_GetSavePath", _wrap_CModule_GetSavePath, METH_O, NULL},
	 { "CModule_ExpandString", _wrap_CModule_ExpandString, METH_VARARGS, NULL},
	 { "CModule_SetType", _wrap_CModule_SetType, METH_VARARGS, NULL},
	 { "CModule_SetDescription", _wrap_CModule_SetDescription, METH_VARARGS, NULL},
	 { "CModule_SetModPath", _wrap_CModule_SetModPath, METH_VARARGS, NULL},
	 { "CModule_SetArgs", _wrap_CModule_SetArgs, METH_VARARGS, NULL},
	 { "CModule_GetType", _wrap_CModule_GetType, METH_O, NULL},
	 { "CModule_GetDescription", _wrap_CModule_GetDescription, METH_O, NULL},
	 { "CModule_GetArgs", _wrap_CModule_GetArgs, METH_O, NULL},
	 { "CModule_GetModPath", _wrap_CModule_GetModPath, METH_O, NULL},
	 { "CModule_GetUser", _wrap_CModule_GetUser, METH_O, NULL},
	 { "CModule_GetNetwork", _wrap_CModule_GetNetwork, METH_O, NULL},
	 { "CModule_GetClient", _wrap_CModule_GetClient, METH_O, NULL},
	 { "CModule_GetManager", _wrap_CModule_GetManager, METH_O, NULL},
	 { "CModule_OnAddUser", _wrap_CModule_OnAddUser, METH_VARARGS, NULL},
	 { "CModule_OnDeleteUser", _wrap_CModule_OnDeleteUser, METH_VARARGS, NULL},
	 { "CModule_OnClientConnect", _wrap_CModule_OnClientConnect, METH_VARARGS, NULL},
	 { "CModule_OnLoginAttempt", _wrap_CModule_OnLoginAttempt, METH_VARARGS, NULL},
	 { "CModule_OnFailedLogin", _wrap_CModule_OnFailedLogin, METH_VARARGS, NULL},
	 { "CModule_OnUnknownUserRaw", _wrap_CModule_OnUnknownUserRaw, METH_VARARGS, NULL},
	 { "CModule_OnUnknownUserRawMessage", _wrap_CModule_OnUnknownUserRawMessage, METH_VARARGS, NULL},
	 { "CModule_OnClientAttached", _wrap_CModule_OnClientAttached, METH_O, NULL},
	 { "CModule_OnClientDetached", _wrap_CModule_OnClientDetached, METH_O, NULL},
	 { "CModule_OnClientCapLs", _wrap_CModule_OnClientCapLs, METH_VARARGS, NULL},
	 { "CModule_IsClientCapSupported", _wrap_CModule_IsClientCapSupported, METH_VARARGS, NULL},
	 { "CModule_OnClientCapRequest", _wrap_CModule_OnClientCapRequest, METH_VARARGS, NULL},
	 { "CModule_OnClientGetSASLMechanisms", _wrap_CModule_OnClientGetSASLMechanisms, METH_VARARGS, NULL},
	 { "CModule_OnClientSASLServerInitialChallenge", _wrap_CModule_OnClientSASLServerInitialChallenge, METH_VARARGS, NULL},
	 { "CModule_OnClientSASLAuthenticate", _wrap_CModule_OnClientSASLAuthenticate, METH_VARARGS, NULL},
	 { "CModule_OnClientSASLAborted", _wrap_CModule_OnClientSASLAborted, METH_O, NULL},
	 { "CModule_OnModuleLoading", _wrap_CModule_OnModuleLoading, METH_VARARGS, NULL},
	 { "CModule_OnModuleUnloading", _wrap_CModule_OnModuleUnloading, METH_VARARGS, NULL},
	 { "CModule_OnGetModInfo", _wrap_CModule_OnGetModInfo, METH_VARARGS, NULL},
	 { "CModule_OnGetAvailableMods", _wrap_CModule_OnGetAvailableMods, METH_VARARGS, NULL},
	 { "CModule_InternalServerDependentCapsOnServerCap302Available", _wrap_CModule_InternalServerDependentCapsOnServerCap302Available, METH_VARARGS, NULL},
	 { "CModule_InternalServerDependentCapsOnServerCapResult", _wrap_CModule_InternalServerDependentCapsOnServerCapResult, METH_VARARGS, NULL},
	 { "CModule_InternalServerDependentCapsOnClientCapLs", _wrap_CModule_InternalServerDependentCapsOnClientCapLs, METH_VARARGS, NULL},
	 { "CModule_InternalServerDependentCapsIsClientCapSupported", _wrap_CModule_InternalServerDependentCapsIsClientCapSupported, METH_VARARGS, NULL},
	 { "CModule_InternalServerDependentCapsOnClientCapRequest", _wrap_CModule_InternalServerDependentCapsOnClientCapRequest, METH_VARARGS, NULL},
	 { "CModule_InternalServerDependentCapsOnClientAttached", _wrap_CModule_InternalServerDependentCapsOnClientAttached, METH_O, NULL},
	 { "CModule_InternalServerDependentCapsOnClientDetached", _wrap_CModule_InternalServerDependentCapsOnClientDetached, METH_O, NULL},
	 { "CModule_InternalServerDependentCapsOnIRCConnected", _wrap_CModule_InternalServerDependentCapsOnIRCConnected, METH_O, NULL},
	 { "CModule_InternalServerDependentCapsOnIRCDisconnected", _wrap_CModule_InternalServerDependentCapsOnIRCDisconnected, METH_O, NULL},
	 { "CModule___str__", _wrap_CModule___str__, METH_O, NULL},
	 { "CModule_BeginNV_", _wrap_CModule_BeginNV_, METH_O, NULL},
	 { "CModule_ExistsNV", _wrap_CModule_ExistsNV, METH_VARARGS, NULL},
	 { "CModule_AddServerDependentCapability", _wrap_CModule_AddServerDependentCapability, METH_VARARGS, NULL},
	 { "CModule_swigregister", CModule_swigregister, METH_O, NULL},
	 { "CModule_swiginit", CModule_swiginit, METH_VARARGS, NULL},
	 { "delete_CModules", _wrap_delete_CModules, METH_O, NULL},
	 { "new_CModules", _wrap_new_CModules, METH_VARARGS, NULL},
	 { "CModules_SetUser", _wrap_CModules_SetUser, METH_VARARGS, NULL},
	 { "CModules_SetNetwork", _wrap_CModules_SetNetwork, METH_VARARGS, NULL},
	 { "CModules_SetClient", _wrap_CModules_SetClient, METH_VARARGS, NULL},
	 { "CModules_GetUser", _wrap_CModules_GetUser, METH_O, NULL},
	 { "CModules_GetNetwork", _wrap_CModules_GetNetwork, METH_O, NULL},
	 { "CModules_GetClient", _wrap_CModules_GetClient, METH_O, NULL},
	 { "CModules_UnloadAll", _wrap_CModules_UnloadAll, METH_O, NULL},
	 { "CModules_OnBoot", _wrap_CModules_OnBoot, METH_O, NULL},
	 { "CModules_OnPreRehash", _wrap_CModules_OnPreRehash, METH_O, NULL},
	 { "CModules_OnPostRehash", _wrap_CModules_OnPostRehash, METH_O, NULL},
	 { "CModules_OnIRCDisconnected", _wrap_CModules_OnIRCDisconnected, METH_O, NULL},
	 { "CModules_OnIRCConnected", _wrap_CModules_OnIRCConnected, METH_O, NULL},
	 { "CModules_OnIRCConnecting", _wrap_CModules_OnIRCConnecting, METH_VARARGS, NULL},
	 { "CModules_OnIRCConnectionError", _wrap_CModules_OnIRCConnectionError, METH_VARARGS, NULL},
	 { "CModules_OnIRCRegistration", _wrap_CModules_OnIRCRegistration, METH_VARARGS, NULL},
	 { "CModules_OnBroadcast", _wrap_CModules_OnBroadcast, METH_VARARGS, NULL},
	 { "CModules_OnChanPermission3", _wrap_CModules_OnChanPermission3, METH_VARARGS, NULL},
	 { "CModules_OnChanPermission2", _wrap_CModules_OnChanPermission2, METH_VARARGS, NULL},
	 { "CModules_OnChanPermission", _wrap_CModules_OnChanPermission, METH_VARARGS, NULL},
	 { "CModules_OnOp2", _wrap_CModules_OnOp2, METH_VARARGS, NULL},
	 { "CModules_OnOp", _wrap_CModules_OnOp, METH_VARARGS, NULL},
	 { "CModules_OnDeop2", _wrap_CModules_OnDeop2, METH_VARARGS, NULL},
	 { "CModules_OnDeop", _wrap_CModules_OnDeop, METH_VARARGS, NULL},
	 { "CModules_OnVoice2", _wrap_CModules_OnVoice2, METH_VARARGS, NULL},
	 { "CModules_OnVoice", _wrap_CModules_OnVoice, METH_VARARGS, NULL},
	 { "CModules_OnDevoice2", _wrap_CModules_OnDevoice2, METH_VARARGS, NULL},
	 { "CModules_OnDevoice", _wrap_CModules_OnDevoice, METH_VARARGS, NULL},
	 { "CModules_OnRawMode2", _wrap_CModules_OnRawMode2, METH_VARARGS, NULL},
	 { "CModules_OnRawMode", _wrap_CModules_OnRawMode, METH_VARARGS, NULL},
	 { "CModules_OnMode2", _wrap_CModules_OnMode2, METH_VARARGS, NULL},
	 { "CModules_OnMode", _wrap_CModules_OnMode, METH_VARARGS, NULL},
	 { "CModules_OnRaw", _wrap_CModules_OnRaw, METH_VARARGS, NULL},
	 { "CModules_OnRawMessage", _wrap_CModules_OnRawMessage, METH_VARARGS, NULL},
	 { "CModules_OnNumericMessage", _wrap_CModules_OnNumericMessage, METH_VARARGS, NULL},
	 { "CModules_OnStatusCommand", _wrap_CModules_OnStatusCommand, METH_VARARGS, NULL},
	 { "CModules_OnModCommand", _wrap_CModules_OnModCommand, METH_VARARGS, NULL},
	 { "CModules_OnModNotice", _wrap_CModules_OnModNotice, METH_VARARGS, NULL},
	 { "CModules_OnModCTCP", _wrap_CModules_OnModCTCP, METH_VARARGS, NULL},
	 { "CModules_OnQuit", _wrap_CModules_OnQuit, METH_VARARGS, NULL},
	 { "CModules_OnQuitMessage", _wrap_CModules_OnQuitMessage, METH_VARARGS, NULL},
	 { "CModules_OnNick", _wrap_CModules_OnNick, METH_VARARGS, NULL},
	 { "CModules_OnNickMessage", _wrap_CModules_OnNickMessage, METH_VARARGS, NULL},
	 { "CModules_OnKick", _wrap_CModules_OnKick, METH_VARARGS, NULL},
	 { "CModules_OnKickMessage", _wrap_CModules_OnKickMessage, METH_VARARGS, NULL},
	 { "CModules_OnJoining", _wrap_CModules_OnJoining, METH_VARARGS, NULL},
	 { "CModules_OnJoin", _wrap_CModules_OnJoin, METH_VARARGS, NULL},
	 { "CModules_OnJoinMessage", _wrap_CModules_OnJoinMessage, METH_VARARGS, NULL},
	 { "CModules_OnPart", _wrap_CModules_OnPart, METH_VARARGS, NULL},
	 { "CModules_OnPartMessage", _wrap_CModules_OnPartMessage, METH_VARARGS, NULL},
	 { "CModules_OnInvite", _wrap_CModules_OnInvite, METH_VARARGS, NULL},
	 { "CModules_OnInviteMessage", _wrap_CModules_OnInviteMessage, METH_VARARGS, NULL},
	 { "CModules_OnChanBufferStarting", _wrap_CModules_OnChanBufferStarting, METH_VARARGS, NULL},
	 { "CModules_OnChanBufferEnding", _wrap_CModules_OnChanBufferEnding, METH_VARARGS, NULL},
	 { "CModules_OnChanBufferPlayLine2", _wrap_CModules_OnChanBufferPlayLine2, METH_VARARGS, NULL},
	 { "CModules_OnChanBufferPlayLine", _wrap_CModules_OnChanBufferPlayLine, METH_VARARGS, NULL},
	 { "CModules_OnPrivBufferStarting", _wrap_CModules_OnPrivBufferStarting, METH_VARARGS, NULL},
	 { "CModules_OnPrivBufferEnding", _wrap_CModules_OnPrivBufferEnding, METH_VARARGS, NULL},
	 { "CModules_OnPrivBufferPlayLine2", _wrap_CModules_OnPrivBufferPlayLine2, METH_VARARGS, NULL},
	 { "CModules_OnPrivBufferPlayLine", _wrap_CModules_OnPrivBufferPlayLine, METH_VARARGS, NULL},
	 { "CModules_OnChanBufferPlayMessage", _wrap_CModules_OnChanBufferPlayMessage, METH_VARARGS, NULL},
	 { "CModules_OnPrivBufferPlayMessage", _wrap_CModules_OnPrivBufferPlayMessage, METH_VARARGS, NULL},
	 { "CModules_OnClientLogin", _wrap_CModules_OnClientLogin, METH_O, NULL},
	 { "CModules_OnClientDisconnect", _wrap_CModules_OnClientDisconnect, METH_O, NULL},
	 { "CModules_OnUserRaw", _wrap_CModules_OnUserRaw, METH_VARARGS, NULL},
	 { "CModules_OnUserRawMessage", _wrap_CModules_OnUserRawMessage, METH_VARARGS, NULL},
	 { "CModules_OnUserCTCPReply", _wrap_CModules_OnUserCTCPReply, METH_VARARGS, NULL},
	 { "CModules_OnUserCTCPReplyMessage", _wrap_CModules_OnUserCTCPReplyMessage, METH_VARARGS, NULL},
	 { "CModules_OnUserCTCP", _wrap_CModules_OnUserCTCP, METH_VARARGS, NULL},
	 { "CModules_OnUserCTCPMessage", _wrap_CModules_OnUserCTCPMessage, METH_VARARGS, NULL},
	 { "CModules_OnUserAction", _wrap_CModules_OnUserAction, METH_VARARGS, NULL},
	 { "CModules_OnUserActionMessage", _wrap_CModules_OnUserActionMessage, METH_VARARGS, NULL},
	 { "CModules_OnUserMsg", _wrap_CModules_OnUserMsg, METH_VARARGS, NULL},
	 { "CModules_OnUserTextMessage", _wrap_CModules_OnUserTextMessage, METH_VARARGS, NULL},
	 { "CModules_OnUserNotice", _wrap_CModules_OnUserNotice, METH_VARARGS, NULL},
	 { "CModules_OnUserNoticeMessage", _wrap_CModules_OnUserNoticeMessage, METH_VARARGS, NULL},
	 { "CModules_OnUserJoin", _wrap_CModules_OnUserJoin, METH_VARARGS, NULL},
	 { "CModules_OnUserJoinMessage", _wrap_CModules_OnUserJoinMessage, METH_VARARGS, NULL},
	 { "CModules_OnUserPart", _wrap_CModules_OnUserPart, METH_VARARGS, NULL},
	 { "CModules_OnUserPartMessage", _wrap_CModules_OnUserPartMessage, METH_VARARGS, NULL},
	 { "CModules_OnUserTopic", _wrap_CModules_OnUserTopic, METH_VARARGS, NULL},
	 { "CModules_OnUserTopicMessage", _wrap_CModules_OnUserTopicMessage, METH_VARARGS, NULL},
	 { "CModules_OnUserTopicRequest", _wrap_CModules_OnUserTopicRequest, METH_VARARGS, NULL},
	 { "CModules_OnUserQuit", _wrap_CModules_OnUserQuit, METH_VARARGS, NULL},
	 { "CModules_OnUserQuitMessage", _wrap_CModules_OnUserQuitMessage, METH_VARARGS, NULL},
	 { "CModules_OnUserTagMessage", _wrap_CModules_OnUserTagMessage, METH_VARARGS, NULL},
	 { "CModules_OnChanTagMessage", _wrap_CModules_OnChanTagMessage, METH_VARARGS, NULL},
	 { "CModules_OnPrivTagMessage", _wrap_CModules_OnPrivTagMessage, METH_VARARGS, NULL},
	 { "CModules_OnCTCPReply", _wrap_CModules_OnCTCPReply, METH_VARARGS, NULL},
	 { "CModules_OnCTCPReplyMessage", _wrap_CModules_OnCTCPReplyMessage, METH_VARARGS, NULL},
	 { "CModules_OnPrivCTCP", _wrap_CModules_OnPrivCTCP, METH_VARARGS, NULL},
	 { "CModules_OnPrivCTCPMessage", _wrap_CModules_OnPrivCTCPMessage, METH_VARARGS, NULL},
	 { "CModules_OnChanCTCP", _wrap_CModules_OnChanCTCP, METH_VARARGS, NULL},
	 { "CModules_OnChanCTCPMessage", _wrap_CModules_OnChanCTCPMessage, METH_VARARGS, NULL},
	 { "CModules_OnPrivAction", _wrap_CModules_OnPrivAction, METH_VARARGS, NULL},
	 { "CModules_OnPrivActionMessage", _wrap_CModules_OnPrivActionMessage, METH_VARARGS, NULL},
	 { "CModules_OnChanAction", _wrap_CModules_OnChanAction, METH_VARARGS, NULL},
	 { "CModules_OnChanActionMessage", _wrap_CModules_OnChanActionMessage, METH_VARARGS, NULL},
	 { "CModules_OnPrivMsg", _wrap_CModules_OnPrivMsg, METH_VARARGS, NULL},
	 { "CModules_OnPrivTextMessage", _wrap_CModules_OnPrivTextMessage, METH_VARARGS, NULL},
	 { "CModules_OnChanMsg", _wrap_CModules_OnChanMsg, METH_VARARGS, NULL},
	 { "CModules_OnChanTextMessage", _wrap_CModules_OnChanTextMessage, METH_VARARGS, NULL},
	 { "CModules_OnPrivNotice", _wrap_CModules_OnPrivNotice, METH_VARARGS, NULL},
	 { "CModules_OnPrivNoticeMessage", _wrap_CModules_OnPrivNoticeMessage, METH_VARARGS, NULL},
	 { "CModules_OnChanNotice", _wrap_CModules_OnChanNotice, METH_VARARGS, NULL},
	 { "CModules_OnChanNoticeMessage", _wrap_CModules_OnChanNoticeMessage, METH_VARARGS, NULL},
	 { "CModules_OnTopic", _wrap_CModules_OnTopic, METH_VARARGS, NULL},
	 { "CModules_OnTopicMessage", _wrap_CModules_OnTopicMessage, METH_VARARGS, NULL},
	 { "CModules_OnTimerAutoJoin", _wrap_CModules_OnTimerAutoJoin, METH_VARARGS, NULL},
	 { "CModules_OnAddNetwork", _wrap_CModules_OnAddNetwork, METH_VARARGS, NULL},
	 { "CModules_OnDeleteNetwork", _wrap_CModules_OnDeleteNetwork, METH_VARARGS, NULL},
	 { "CModules_OnSendToClient", _wrap_CModules_OnSendToClient, METH_VARARGS, NULL},
	 { "CModules_OnSendToClientMessage", _wrap_CModules_OnSendToClientMessage, METH_VARARGS, NULL},
	 { "CModules_OnSendToIRC", _wrap_CModules_OnSendToIRC, METH_VARARGS, NULL},
	 { "CModules_OnSendToIRCMessage", _wrap_CModules_OnSendToIRCMessage, METH_VARARGS, NULL},
	 { "CModules_OnClientAttached", _wrap_CModules_OnClientAttached, METH_O, NULL},
	 { "CModules_OnClientDetached", _wrap_CModules_OnClientDetached, METH_O, NULL},
	 { "CModules_OnServerCapAvailable", _wrap_CModules_OnServerCapAvailable, METH_VARARGS, NULL},
	 { "CModules_OnServerCapResult", _wrap_CModules_OnServerCapResult, METH_VARARGS, NULL},
	 { "CModules_FindModule", _wrap_CModules_FindModule, METH_VARARGS, NULL},
	 { "CModules_LoadModule", _wrap_CModules_LoadModule, METH_VARARGS, NULL},
	 { "CModules_UnloadModule", _wrap_CModules_UnloadModule, METH_VARARGS, NULL},
	 { "CModules_ReloadModule", _wrap_CModules_ReloadModule, METH_VARARGS, NULL},
	 { "CModules_GetModInfo", _wrap_CModules_GetModInfo, METH_VARARGS, NULL},
	 { "CModules_GetModPathInfo", _wrap_CModules_GetModPathInfo, METH_VARARGS, NULL},
	 { "CModules_GetAvailableMods", _wrap_CModules_GetAvailableMods, METH_VARARGS, NULL},
	 { "CModules_GetDefaultMods", _wrap_CModules_GetDefaultMods, METH_VARARGS, NULL},
	 { "CModules_FindModPath", _wrap_CModules_FindModPath, METH_VARARGS, NULL},
	 { "CModules_GetModDirs", _wrap_CModules_GetModDirs, METH_NOARGS, NULL},
	 { "CModules_OnAddUser", _wrap_CModules_OnAddUser, METH_VARARGS, NULL},
	 { "CModules_OnDeleteUser", _wrap_CModules_OnDeleteUser, METH_VARARGS, NULL},
	 { "CModules_OnClientConnect", _wrap_CModules_OnClientConnect, METH_VARARGS, NULL},
	 { "CModules_OnLoginAttempt", _wrap_CModules_OnLoginAttempt, METH_VARARGS, NULL},
	 { "CModules_OnFailedLogin", _wrap_CModules_OnFailedLogin, METH_VARARGS, NULL},
	 { "CModules_OnUnknownUserRaw", _wrap_CModules_OnUnknownUserRaw, METH_VARARGS, NULL},
	 { "CModules_OnUnknownUserRawMessage", _wrap_CModules_OnUnknownUserRawMessage, METH_VARARGS, NULL},
	 { "CModules_OnClientCapLs", _wrap_CModules_OnClientCapLs, METH_VARARGS, NULL},
	 { "CModules_IsClientCapSupported", _wrap_CModules_IsClientCapSupported, METH_VARARGS, NULL},
	 { "CModules_OnClientCapRequest", _wrap_CModules_OnClientCapRequest, METH_VARARGS, NULL},
	 { "CModules_OnClientGetSASLMechanisms", _wrap_CModules_OnClientGetSASLMechanisms, METH_VARARGS, NULL},
	 { "CModules_OnClientSASLAborted", _wrap_CModules_OnClientSASLAborted, METH_O, NULL},
	 { "CModules_OnClientSASLServerInitialChallenge", _wrap_CModules_OnClientSASLServerInitialChallenge, METH_VARARGS, NULL},
	 { "CModules_OnClientSASLAuthenticate", _wrap_CModules_OnClientSASLAuthenticate, METH_VARARGS, NULL},
	 { "CModules_OnModuleLoading", _wrap_CModules_OnModuleLoading, METH_VARARGS, NULL},
	 { "CModules_OnModuleUnloading", _wrap_CModules_OnModuleUnloading, METH_VARARGS, NULL},
	 { "CModules_OnGetModInfo", _wrap_CModules_OnGetModInfo, METH_VARARGS, NULL},
	 { "CModules_OnGetAvailableMods", _wrap_CModules_OnGetAvailableMods, METH_VARARGS, NULL},
	 { "CModules_removeModule", _wrap_CModules_removeModule, METH_VARARGS, NULL},
	 { "CModules_swigregister", CModules_swigregister, METH_O, NULL},
	 { "CModules_swiginit", CModules_swiginit, METH_VARARGS, NULL},
	 { "delete_CNick", _wrap_delete_CNick, METH_O, NULL},
	 { "new_CNick", _wrap_new_CNick, METH_VARARGS, NULL},
	 { "CNick_Reset", _wrap_CNick_Reset, METH_O, NULL},
	 { "CNick_Parse", _wrap_CNick_Parse, METH_VARARGS, NULL},
	 { "CNick_GetHostMask", _wrap_CNick_GetHostMask, METH_O, NULL},
	 { "CNick_GetCommonChans", _wrap_CNick_GetCommonChans, METH_VARARGS, NULL},
	 { "CNick_NickEquals", _wrap_CNick_NickEquals, METH_VARARGS, NULL},
	 { "CNick_SetNetwork", _wrap_CNick_SetNetwork, METH_VARARGS, NULL},
	 { "CNick_SetNick", _wrap_CNick_SetNick, METH_VARARGS, NULL},
	 { "CNick_SetIdent", _wrap_CNick_SetIdent, METH_VARARGS, NULL},
	 { "CNick_SetHost", _wrap_CNick_SetHost, METH_VARARGS, NULL},
	 { "CNick_AddPerm", _wrap_CNick_AddPerm, METH_VARARGS, NULL},
	 { "CNick_RemPerm", _wrap_CNick_RemPerm, METH_VARARGS, NULL},
	 { "CNick_GetPermStr", _wrap_CNick_GetPermStr, METH_O, NULL},
	 { "CNick_GetPermChar", _wrap_CNick_GetPermChar, METH_O, NULL},
	 { "CNick_HasPerm", _wrap_CNick_HasPerm, METH_VARARGS, NULL},
	 { "CNick_GetNick", _wrap_CNick_GetNick, METH_O, NULL},
	 { "CNick_GetIdent", _wrap_CNick_GetIdent, METH_O, NULL},
	 { "CNick_GetHost", _wrap_CNick_GetHost, METH_O, NULL},
	 { "CNick_GetNickMask", _wrap_CNick_GetNickMask, METH_O, NULL},
	 { "CNick_Clone", _wrap_CNick_Clone, METH_VARARGS, NULL},
	 { "CNick___str__", _wrap_CNick___str__, METH_O, NULL},
	 { "CNick___repr__", _wrap_CNick___repr__, METH_O, NULL},
	 { "CNick_swigregister", CNick_swigregister, METH_O, NULL},
	 { "CNick_swiginit", CNick_swiginit, METH_VARARGS, NULL},
	 { "new_CChan", _wrap_new_CChan, METH_VARARGS, NULL},
	 { "delete_CChan", _wrap_delete_CChan, METH_O, NULL},
	 { "CChan_Reset", _wrap_CChan_Reset, METH_O, NULL},
	 { "CChan_ToConfig", _wrap_CChan_ToConfig, METH_O, NULL},
	 { "CChan_Clone", _wrap_CChan_Clone, METH_VARARGS, NULL},
	 { "CChan_Cycle", _wrap_CChan_Cycle, METH_O, NULL},
	 { "CChan_JoinUser", _wrap_CChan_JoinUser, METH_VARARGS, NULL},
	 { "CChan_AttachUser", _wrap_CChan_AttachUser, METH_VARARGS, NULL},
	 { "CChan_DetachUser", _wrap_CChan_DetachUser, METH_O, NULL},
	 { "CChan_OnWho", _wrap_CChan_OnWho, METH_VARARGS, NULL},
	 { "CChan_SetModes", _wrap_CChan_SetModes, METH_VARARGS, NULL},
	 { "CChan_ModeChange", _wrap_CChan_ModeChange, METH_VARARGS, NULL},
	 { "CChan_AddMode", _wrap_CChan_AddMode, METH_VARARGS, NULL},
	 { "CChan_RemMode", _wrap_CChan_RemMode, METH_VARARGS, NULL},
	 { "CChan_GetModeString", _wrap_CChan_GetModeString, METH_O, NULL},
	 { "CChan_GetModeForNames", _wrap_CChan_GetModeForNames, METH_O, NULL},
	 { "CChan_ClearNicks", _wrap_CChan_ClearNicks, METH_O, NULL},
	 { "CChan_FindNick", _wrap_CChan_FindNick, METH_VARARGS, NULL},
	 { "CChan_AddNicks", _wrap_CChan_AddNicks, METH_VARARGS, NULL},
	 { "CChan_AddNick", _wrap_CChan_AddNick, METH_VARARGS, NULL},
	 { "CChan_RemNick", _wrap_CChan_RemNick, METH_VARARGS, NULL},
	 { "CChan_ChangeNick", _wrap_CChan_ChangeNick, METH_VARARGS, NULL},
	 { "CChan_GetBuffer", _wrap_CChan_GetBuffer, METH_O, NULL},
	 { "CChan_GetBufferCount", _wrap_CChan_GetBufferCount, METH_O, NULL},
	 { "CChan_SetBufferCount", _wrap_CChan_SetBufferCount, METH_VARARGS, NULL},
	 { "CChan_InheritBufferCount", _wrap_CChan_InheritBufferCount, METH_VARARGS, NULL},
	 { "CChan_ResetBufferCount", _wrap_CChan_ResetBufferCount, METH_O, NULL},
	 { "CChan_AddBuffer", _wrap_CChan_AddBuffer, METH_VARARGS, NULL},
	 { "CChan_ClearBuffer", _wrap_CChan_ClearBuffer, METH_O, NULL},
	 { "CChan_SendBuffer", _wrap_CChan_SendBuffer, METH_VARARGS, NULL},
	 { "CChan_GetPermStr", _wrap_CChan_GetPermStr, METH_O, NULL},
	 { "CChan_HasPerm", _wrap_CChan_HasPerm, METH_VARARGS, NULL},
	 { "CChan_AddPerm", _wrap_CChan_AddPerm, METH_VARARGS, NULL},
	 { "CChan_RemPerm", _wrap_CChan_RemPerm, METH_VARARGS, NULL},
	 { "CChan_SetModeKnown", _wrap_CChan_SetModeKnown, METH_VARARGS, NULL},
	 { "CChan_SetIsOn", _wrap_CChan_SetIsOn, METH_VARARGS, NULL},
	 { "CChan_SetKey", _wrap_CChan_SetKey, METH_VARARGS, NULL},
	 { "CChan_SetTopic", _wrap_CChan_SetTopic, METH_VARARGS, NULL},
	 { "CChan_SetTopicOwner", _wrap_CChan_SetTopicOwner, METH_VARARGS, NULL},
	 { "CChan_SetTopicDate", _wrap_CChan_SetTopicDate, METH_VARARGS, NULL},
	 { "CChan_SetDefaultModes", _wrap_CChan_SetDefaultModes, METH_VARARGS, NULL},
	 { "CChan_SetAutoClearChanBuffer", _wrap_CChan_SetAutoClearChanBuffer, METH_VARARGS, NULL},
	 { "CChan_InheritAutoClearChanBuffer", _wrap_CChan_InheritAutoClearChanBuffer, METH_VARARGS, NULL},
	 { "CChan_ResetAutoClearChanBuffer", _wrap_CChan_ResetAutoClearChanBuffer, METH_O, NULL},
	 { "CChan_SetDetached", _wrap_CChan_SetDetached, METH_VARARGS, NULL},
	 { "CChan_SetInConfig", _wrap_CChan_SetInConfig, METH_VARARGS, NULL},
	 { "CChan_SetCreationDate", _wrap_CChan_SetCreationDate, METH_VARARGS, NULL},
	 { "CChan_Disable", _wrap_CChan_Disable, METH_O, NULL},
	 { "CChan_Enable", _wrap_CChan_Enable, METH_O, NULL},
	 { "CChan_IncJoinTries", _wrap_CChan_IncJoinTries, METH_O, NULL},
	 { "CChan_ResetJoinTries", _wrap_CChan_ResetJoinTries, METH_O, NULL},
	 { "CChan_GetNetwork", _wrap_CChan_GetNetwork, METH_O, NULL},
	 { "CChan_IsModeKnown", _wrap_CChan_IsModeKnown, METH_O, NULL},
	 { "CChan_HasMode", _wrap_CChan_HasMode, METH_VARARGS, NULL},
	 { "CChan_GetOptions", _wrap_CChan_GetOptions, METH_O, NULL},
	 { "CChan_GetModeArg", _wrap_CChan_GetModeArg, METH_VARARGS, NULL},
	 { "CChan_GetPermCounts", _wrap_CChan_GetPermCounts, METH_O, NULL},
	 { "CChan_IsOn", _wrap_CChan_IsOn, METH_O, NULL},
	 { "CChan_GetName", _wrap_CChan_GetName, METH_O, NULL},
	 { "CChan_GetModes", _wrap_CChan_GetModes, METH_O, NULL},
	 { "CChan_GetKey", _wrap_CChan_GetKey, METH_O, NULL},
	 { "CChan_GetTopic", _wrap_CChan_GetTopic, METH_O, NULL},
	 { "CChan_GetTopicOwner", _wrap_CChan_GetTopicOwner, METH_O, NULL},
	 { "CChan_GetTopicDate", _wrap_CChan_GetTopicDate, METH_O, NULL},
	 { "CChan_GetDefaultModes", _wrap_CChan_GetDefaultModes, METH_O, NULL},
	 { "CChan_GetNicks", _wrap_CChan_GetNicks, METH_O, NULL},
	 { "CChan_GetNickCount", _wrap_CChan_GetNickCount, METH_O, NULL},
	 { "CChan_AutoClearChanBuffer", _wrap_CChan_AutoClearChanBuffer, METH_O, NULL},
	 { "CChan_IsDetached", _wrap_CChan_IsDetached, METH_O, NULL},
	 { "CChan_InConfig", _wrap_CChan_InConfig, METH_O, NULL},
	 { "CChan_GetCreationDate", _wrap_CChan_GetCreationDate, METH_O, NULL},
	 { "CChan_IsDisabled", _wrap_CChan_IsDisabled, METH_O, NULL},
	 { "CChan_GetJoinTries", _wrap_CChan_GetJoinTries, METH_O, NULL},
	 { "CChan_HasBufferCountSet", _wrap_CChan_HasBufferCountSet, METH_O, NULL},
	 { "CChan_HasAutoClearChanBufferSet", _wrap_CChan_HasAutoClearChanBufferSet, METH_O, NULL},
	 { "CChan___str__", _wrap_CChan___str__, METH_O, NULL},
	 { "CChan___repr__", _wrap_CChan___repr__, METH_O, NULL},
	 { "CChan_GetNicks_", _wrap_CChan_GetNicks_, METH_O, NULL},
	 { "CChan_swigregister", CChan_swigregister, METH_O, NULL},
	 { "CChan_swiginit", CChan_swiginit, METH_VARARGS, NULL},
	 { "new_CUser", _wrap_new_CUser, METH_O, NULL},
	 { "delete_CUser", _wrap_delete_CUser, METH_O, NULL},
	 { "CUser_ParseConfig", _wrap_CUser_ParseConfig, METH_VARARGS, NULL},
	 { "CUser_SaltedHash", _wrap_CUser_SaltedHash, METH_VARARGS, NULL},
	 { "CUser_ToConfig", _wrap_CUser_ToConfig, METH_O, NULL},
	 { "CUser_CheckPass", _wrap_CUser_CheckPass, METH_VARARGS, NULL},
	 { "CUser_AddAllowedHost", _wrap_CUser_AddAllowedHost, METH_VARARGS, NULL},
	 { "CUser_RemAllowedHost", _wrap_CUser_RemAllowedHost, METH_VARARGS, NULL},
	 { "CUser_ClearAllowedHosts", _wrap_CUser_ClearAllowedHosts, METH_O, NULL},
	 { "CUser_IsHostAllowed", _wrap_CUser_IsHostAllowed, METH_VARARGS, NULL},
	 { "CUser_IsValid", _wrap_CUser_IsValid, METH_VARARGS, NULL},
	 { "CUser_IsValidUsername", _wrap_CUser_IsValidUsername, METH_O, NULL},
	 { "CUser_IsValidUserName", _wrap_CUser_IsValidUserName, METH_O, NULL},
	 { "CUser_MakeCleanUsername", _wrap_CUser_MakeCleanUsername, METH_O, NULL},
	 { "CUser_MakeCleanUserName", _wrap_CUser_MakeCleanUserName, METH_O, NULL},
	 { "CUser_GetModules", _wrap_CUser_GetModules, METH_VARARGS, NULL},
	 { "CUser_DeleteNetwork", _wrap_CUser_DeleteNetwork, METH_VARARGS, NULL},
	 { "CUser_AddNetwork", _wrap_CUser_AddNetwork, METH_VARARGS, NULL},
	 { "CUser_RemoveNetwork", _wrap_CUser_RemoveNetwork, METH_VARARGS, NULL},
	 { "CUser_FindNetwork", _wrap_CUser_FindNetwork, METH_VARARGS, NULL},
	 { "CUser_GetNetworks", _wrap_CUser_GetNetworks, METH_O, NULL},
	 { "CUser_HasSpaceForNewNetwork", _wrap_CUser_HasSpaceForNewNetwork, METH_O, NULL},
	 { "CUser_PutUser", _wrap_CUser_PutUser, METH_VARARGS, NULL},
	 { "CUser_PutAllUser", _wrap_CUser_PutAllUser, METH_VARARGS, NULL},
	 { "CUser_PutStatus", _wrap_CUser_PutStatus, METH_VARARGS, NULL},
	 { "CUser_PutStatusNotice", _wrap_CUser_PutStatusNotice, METH_VARARGS, NULL},
	 { "CUser_PutModule", _wrap_CUser_PutModule, METH_VARARGS, NULL},
	 { "CUser_PutModNotice", _wrap_CUser_PutModNotice, METH_VARARGS, NULL},
	 { "CUser_IsUserAttached", _wrap_CUser_IsUserAttached, METH_O, NULL},
	 { "CUser_UserConnected", _wrap_CUser_UserConnected, METH_VARARGS, NULL},
	 { "CUser_UserDisconnected", _wrap_CUser_UserDisconnected, METH_VARARGS, NULL},
	 { "CUser_GetLocalDCCIP", _wrap_CUser_GetLocalDCCIP, METH_O, NULL},
	 { "CUser_ExpandString", _wrap_CUser_ExpandString, METH_VARARGS, NULL},
	 { "CUser_AddTimestamp", _wrap_CUser_AddTimestamp, METH_VARARGS, NULL},
	 { "CUser_CloneNetworks", _wrap_CUser_CloneNetworks, METH_VARARGS, NULL},
	 { "CUser_Clone", _wrap_CUser_Clone, METH_VARARGS, NULL},
	 { "CUser_BounceAllClients", _wrap_CUser_BounceAllClients, METH_O, NULL},
	 { "CUser_AddBytesRead", _wrap_CUser_AddBytesRead, METH_VARARGS, NULL},
	 { "CUser_AddBytesWritten", _wrap_CUser_AddBytesWritten, METH_VARARGS, NULL},
	 { "CUser_SetNick", _wrap_CUser_SetNick, METH_VARARGS, NULL},
	 { "CUser_SetAltNick", _wrap_CUser_SetAltNick, METH_VARARGS, NULL},
	 { "CUser_SetIdent", _wrap_CUser_SetIdent, METH_VARARGS, NULL},
	 { "CUser_SetRealName", _wrap_CUser_SetRealName, METH_VARARGS, NULL},
	 { "CUser_SetBindHost", _wrap_CUser_SetBindHost, METH_VARARGS, NULL},
	 { "CUser_SetDCCBindHost", _wrap_CUser_SetDCCBindHost, METH_VARARGS, NULL},
	 { "CUser_SetPass", _wrap_CUser_SetPass, METH_VARARGS, NULL},
	 { "CUser_SetMultiClients", _wrap_CUser_SetMultiClients, METH_VARARGS, NULL},
	 { "CUser_SetDenyLoadMod", _wrap_CUser_SetDenyLoadMod, METH_VARARGS, NULL},
	 { "CUser_SetAdmin", _wrap_CUser_SetAdmin, METH_VARARGS, NULL},
	 { "CUser_SetDenySetBindHost", _wrap_CUser_SetDenySetBindHost, METH_VARARGS, NULL},
	 { "CUser_SetDenySetIdent", _wrap_CUser_SetDenySetIdent, METH_VARARGS, NULL},
	 { "CUser_SetDenySetNetwork", _wrap_CUser_SetDenySetNetwork, METH_VARARGS, NULL},
	 { "CUser_SetDenySetRealName", _wrap_CUser_SetDenySetRealName, METH_VARARGS, NULL},
	 { "CUser_SetDenySetQuitMsg", _wrap_CUser_SetDenySetQuitMsg, METH_VARARGS, NULL},
	 { "CUser_SetDenySetCTCPReplies", _wrap_CUser_SetDenySetCTCPReplies, METH_VARARGS, NULL},
	 { "CUser_SetStatusPrefix", _wrap_CUser_SetStatusPrefix, METH_VARARGS, NULL},
	 { "CUser_SetDefaultChanModes", _wrap_CUser_SetDefaultChanModes, METH_VARARGS, NULL},
	 { "CUser_SetClientEncoding", _wrap_CUser_SetClientEncoding, METH_VARARGS, NULL},
	 { "CUser_SetQuitMsg", _wrap_CUser_SetQuitMsg, METH_VARARGS, NULL},
	 { "CUser_AddCTCPReply", _wrap_CUser_AddCTCPReply, METH_VARARGS, NULL},
	 { "CUser_DelCTCPReply", _wrap_CUser_DelCTCPReply, METH_VARARGS, NULL},
	 { "CUser_SetBufferCount", _wrap_CUser_SetBufferCount, METH_VARARGS, NULL},
	 { "CUser_SetChanBufferSize", _wrap_CUser_SetChanBufferSize, METH_VARARGS, NULL},
	 { "CUser_SetQueryBufferSize", _wrap_CUser_SetQueryBufferSize, METH_VARARGS, NULL},
	 { "CUser_SetAutoClearChanBuffer", _wrap_CUser_SetAutoClearChanBuffer, METH_VARARGS, NULL},
	 { "CUser_SetAutoClearQueryBuffer", _wrap_CUser_SetAutoClearQueryBuffer, METH_VARARGS, NULL},
	 { "CUser_SetLanguage", _wrap_CUser_SetLanguage, METH_VARARGS, NULL},
	 { "CUser_SetBeingDeleted", _wrap_CUser_SetBeingDeleted, METH_VARARGS, NULL},
	 { "CUser_SetTimestampFormat", _wrap_CUser_SetTimestampFormat, METH_VARARGS, NULL},
	 { "CUser_SetTimestampAppend", _wrap_CUser_SetTimestampAppend, METH_VARARGS, NULL},
	 { "CUser_SetTimestampPrepend", _wrap_CUser_SetTimestampPrepend, METH_VARARGS, NULL},
	 { "CUser_SetAuthOnlyViaModule", _wrap_CUser_SetAuthOnlyViaModule, METH_VARARGS, NULL},
	 { "CUser_SetTimezone", _wrap_CUser_SetTimezone, METH_VARARGS, NULL},
	 { "CUser_SetJoinTries", _wrap_CUser_SetJoinTries, METH_VARARGS, NULL},
	 { "CUser_SetMaxJoins", _wrap_CUser_SetMaxJoins, METH_VARARGS, NULL},
	 { "CUser_SetSkinName", _wrap_CUser_SetSkinName, METH_VARARGS, NULL},
	 { "CUser_SetMaxNetworks", _wrap_CUser_SetMaxNetworks, METH_VARARGS, NULL},
	 { "CUser_SetMaxQueryBuffers", _wrap_CUser_SetMaxQueryBuffers, METH_VARARGS, NULL},
	 { "CUser_SetNoTrafficTimeout", _wrap_CUser_SetNoTrafficTimeout, METH_VARARGS, NULL},
	 { "CUser_GetUserClients", _wrap_CUser_GetUserClients, METH_O, NULL},
	 { "CUser_GetAllClients", _wrap_CUser_GetAllClients, METH_O, NULL},
	 { "CUser_GetUserName", _wrap_CUser_GetUserName, METH_O, NULL},
	 { "CUser_GetUsername", _wrap_CUser_GetUsername, METH_O, NULL},
	 { "CUser_GetCleanUserName", _wrap_CUser_GetCleanUserName, METH_O, NULL},
	 { "CUser_GetNick", _wrap_CUser_GetNick, METH_VARARGS, NULL},
	 { "CUser_GetAltNick", _wrap_CUser_GetAltNick, METH_VARARGS, NULL},
	 { "CUser_GetIdent", _wrap_CUser_GetIdent, METH_VARARGS, NULL},
	 { "CUser_GetRealName", _wrap_CUser_GetRealName, METH_O, NULL},
	 { "CUser_GetBindHost", _wrap_CUser_GetBindHost, METH_O, NULL},
	 { "CUser_GetDCCBindHost", _wrap_CUser_GetDCCBindHost, METH_O, NULL},
	 { "CUser_GetPass", _wrap_CUser_GetPass, METH_O, NULL},
	 { "CUser_GetPassHashType", _wrap_CUser_GetPassHashType, METH_O, NULL},
	 { "CUser_GetPassSalt", _wrap_CUser_GetPassSalt, METH_O, NULL},
	 { "CUser_GetAllowedHosts", _wrap_CUser_GetAllowedHosts, METH_O, NULL},
	 { "CUser_GetTimestampFormat", _wrap_CUser_GetTimestampFormat, METH_O, NULL},
	 { "CUser_GetClientEncoding", _wrap_CUser_GetClientEncoding, METH_O, NULL},
	 { "CUser_GetTimestampAppend", _wrap_CUser_GetTimestampAppend, METH_O, NULL},
	 { "CUser_GetTimestampPrepend", _wrap_CUser_GetTimestampPrepend, METH_O, NULL},
	 { "CUser_GetUserPath", _wrap_CUser_GetUserPath, METH_O, NULL},
	 { "CUser_DenyLoadMod", _wrap_CUser_DenyLoadMod, METH_O, NULL},
	 { "CUser_IsAdmin", _wrap_CUser_IsAdmin, METH_O, NULL},
	 { "CUser_DenySetBindHost", _wrap_CUser_DenySetBindHost, METH_O, NULL},
	 { "CUser_DenySetIdent", _wrap_CUser_DenySetIdent, METH_O, NULL},
	 { "CUser_DenySetNetwork", _wrap_CUser_DenySetNetwork, METH_O, NULL},
	 { "CUser_DenySetRealName", _wrap_CUser_DenySetRealName, METH_O, NULL},
	 { "CUser_DenySetQuitMsg", _wrap_CUser_DenySetQuitMsg, METH_O, NULL},
	 { "CUser_DenySetCTCPReplies", _wrap_CUser_DenySetCTCPReplies, METH_O, NULL},
	 { "CUser_MultiClients", _wrap_CUser_MultiClients, METH_O, NULL},
	 { "CUser_AuthOnlyViaModule", _wrap_CUser_AuthOnlyViaModule, METH_O, NULL},
	 { "CUser_GetStatusPrefix", _wrap_CUser_GetStatusPrefix, METH_O, NULL},
	 { "CUser_GetDefaultChanModes", _wrap_CUser_GetDefaultChanModes, METH_O, NULL},
	 { "CUser_GetPingFrequency", _wrap_CUser_GetPingFrequency, METH_O, NULL},
	 { "CUser_GetPingSlack", _wrap_CUser_GetPingSlack, METH_O, NULL},
	 { "CUser_GetNoTrafficTimeout", _wrap_CUser_GetNoTrafficTimeout, METH_O, NULL},
	 { "CUser_GetQuitMsg", _wrap_CUser_GetQuitMsg, METH_O, NULL},
	 { "CUser_GetCTCPReplies", _wrap_CUser_GetCTCPReplies, METH_O, NULL},
	 { "CUser_GetBufferCount", _wrap_CUser_GetBufferCount, METH_O, NULL},
	 { "CUser_GetChanBufferSize", _wrap_CUser_GetChanBufferSize, METH_O, NULL},
	 { "CUser_GetQueryBufferSize", _wrap_CUser_GetQueryBufferSize, METH_O, NULL},
	 { "CUser_AutoClearChanBuffer", _wrap_CUser_AutoClearChanBuffer, METH_O, NULL},
	 { "CUser_AutoClearQueryBuffer", _wrap_CUser_AutoClearQueryBuffer, METH_O, NULL},
	 { "CUser_IsBeingDeleted", _wrap_CUser_IsBeingDeleted, METH_O, NULL},
	 { "CUser_GetTimezone", _wrap_CUser_GetTimezone, METH_O, NULL},
	 { "CUser_BytesRead", _wrap_CUser_BytesRead, METH_O, NULL},
	 { "CUser_BytesWritten", _wrap_CUser_BytesWritten, METH_O, NULL},
	 { "CUser_JoinTries", _wrap_CUser_JoinTries, METH_O, NULL},
	 { "CUser_MaxJoins", _wrap_CUser_MaxJoins, METH_O, NULL},
	 { "CUser_GetSkinName", _wrap_CUser_GetSkinName, METH_O, NULL},
	 { "CUser_GetLanguage", _wrap_CUser_GetLanguage, METH_O, NULL},
	 { "CUser_MaxNetworks", _wrap_CUser_MaxNetworks, METH_O, NULL},
	 { "CUser_MaxQueryBuffers", _wrap_CUser_MaxQueryBuffers, METH_O, NULL},
	 { "CUser___str__", _wrap_CUser___str__, METH_O, NULL},
	 { "CUser___repr__", _wrap_CUser___repr__, METH_O, NULL},
	 { "CUser_GetNetworks_", _wrap_CUser_GetNetworks_, METH_O, NULL},
	 { "CUser_swigregister", CUser_swigregister, METH_O, NULL},
	 { "CUser_swiginit", CUser_swiginit, METH_VARARGS, NULL},
	 { "CIRCNetwork_IsValidNetwork", _wrap_CIRCNetwork_IsValidNetwork, METH_O, NULL},
	 { "new_CIRCNetwork", _wrap_new_CIRCNetwork, METH_VARARGS, NULL},
	 { "delete_CIRCNetwork", _wrap_delete_CIRCNetwork, METH_O, NULL},
	 { "CIRCNetwork_Clone", _wrap_CIRCNetwork_Clone, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetNetworkPath", _wrap_CIRCNetwork_GetNetworkPath, METH_O, NULL},
	 { "CIRCNetwork_DelServers", _wrap_CIRCNetwork_DelServers, METH_O, NULL},
	 { "CIRCNetwork_ParseConfig", _wrap_CIRCNetwork_ParseConfig, METH_VARARGS, NULL},
	 { "CIRCNetwork_ToConfig", _wrap_CIRCNetwork_ToConfig, METH_O, NULL},
	 { "CIRCNetwork_BounceAllClients", _wrap_CIRCNetwork_BounceAllClients, METH_O, NULL},
	 { "CIRCNetwork_IsUserAttached", _wrap_CIRCNetwork_IsUserAttached, METH_O, NULL},
	 { "CIRCNetwork_IsUserOnline", _wrap_CIRCNetwork_IsUserOnline, METH_O, NULL},
	 { "CIRCNetwork_ClientConnected", _wrap_CIRCNetwork_ClientConnected, METH_VARARGS, NULL},
	 { "CIRCNetwork_ClientDisconnected", _wrap_CIRCNetwork_ClientDisconnected, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetUser", _wrap_CIRCNetwork_GetUser, METH_O, NULL},
	 { "CIRCNetwork_GetName", _wrap_CIRCNetwork_GetName, METH_O, NULL},
	 { "CIRCNetwork_IsNetworkAttached", _wrap_CIRCNetwork_IsNetworkAttached, METH_O, NULL},
	 { "CIRCNetwork_GetClients", _wrap_CIRCNetwork_GetClients, METH_O, NULL},
	 { "CIRCNetwork_FindClients", _wrap_CIRCNetwork_FindClients, METH_VARARGS, NULL},
	 { "CIRCNetwork_SetUser", _wrap_CIRCNetwork_SetUser, METH_VARARGS, NULL},
	 { "CIRCNetwork_SetName", _wrap_CIRCNetwork_SetName, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetModules", _wrap_CIRCNetwork_GetModules, METH_VARARGS, NULL},
	 { "CIRCNetwork_PutUser", _wrap_CIRCNetwork_PutUser, METH_VARARGS, NULL},
	 { "CIRCNetwork_PutStatus", _wrap_CIRCNetwork_PutStatus, METH_VARARGS, NULL},
	 { "CIRCNetwork_PutModule", _wrap_CIRCNetwork_PutModule, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetChans", _wrap_CIRCNetwork_GetChans, METH_O, NULL},
	 { "CIRCNetwork_FindChan", _wrap_CIRCNetwork_FindChan, METH_VARARGS, NULL},
	 { "CIRCNetwork_FindChans", _wrap_CIRCNetwork_FindChans, METH_VARARGS, NULL},
	 { "CIRCNetwork_AddChan", _wrap_CIRCNetwork_AddChan, METH_VARARGS, NULL},
	 { "CIRCNetwork_DelChan", _wrap_CIRCNetwork_DelChan, METH_VARARGS, NULL},
	 { "CIRCNetwork_MoveChan", _wrap_CIRCNetwork_MoveChan, METH_VARARGS, NULL},
	 { "CIRCNetwork_SwapChans", _wrap_CIRCNetwork_SwapChans, METH_VARARGS, NULL},
	 { "CIRCNetwork_JoinChans", _wrap_CIRCNetwork_JoinChans, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetQueries", _wrap_CIRCNetwork_GetQueries, METH_O, NULL},
	 { "CIRCNetwork_FindQuery", _wrap_CIRCNetwork_FindQuery, METH_VARARGS, NULL},
	 { "CIRCNetwork_FindQueries", _wrap_CIRCNetwork_FindQueries, METH_VARARGS, NULL},
	 { "CIRCNetwork_AddQuery", _wrap_CIRCNetwork_AddQuery, METH_VARARGS, NULL},
	 { "CIRCNetwork_DelQuery", _wrap_CIRCNetwork_DelQuery, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetChanPrefixes", _wrap_CIRCNetwork_GetChanPrefixes, METH_O, NULL},
	 { "CIRCNetwork_SetChanPrefixes", _wrap_CIRCNetwork_SetChanPrefixes, METH_VARARGS, NULL},
	 { "CIRCNetwork_IsChan", _wrap_CIRCNetwork_IsChan, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetServers", _wrap_CIRCNetwork_GetServers, METH_O, NULL},
	 { "CIRCNetwork_HasServers", _wrap_CIRCNetwork_HasServers, METH_O, NULL},
	 { "CIRCNetwork_FindServer", _wrap_CIRCNetwork_FindServer, METH_VARARGS, NULL},
	 { "CIRCNetwork_DelServer", _wrap_CIRCNetwork_DelServer, METH_VARARGS, NULL},
	 { "CIRCNetwork_AddServer", _wrap_CIRCNetwork_AddServer, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetNextServer", _wrap_CIRCNetwork_GetNextServer, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetCurrentServer", _wrap_CIRCNetwork_GetCurrentServer, METH_O, NULL},
	 { "CIRCNetwork_SetIRCServer", _wrap_CIRCNetwork_SetIRCServer, METH_VARARGS, NULL},
	 { "CIRCNetwork_SetNextServer", _wrap_CIRCNetwork_SetNextServer, METH_VARARGS, NULL},
	 { "CIRCNetwork_IsLastServer", _wrap_CIRCNetwork_IsLastServer, METH_O, NULL},
	 { "CIRCNetwork_GetTrustedFingerprints", _wrap_CIRCNetwork_GetTrustedFingerprints, METH_O, NULL},
	 { "CIRCNetwork_AddTrustedFingerprint", _wrap_CIRCNetwork_AddTrustedFingerprint, METH_VARARGS, NULL},
	 { "CIRCNetwork_DelTrustedFingerprint", _wrap_CIRCNetwork_DelTrustedFingerprint, METH_VARARGS, NULL},
	 { "CIRCNetwork_ClearTrustedFingerprints", _wrap_CIRCNetwork_ClearTrustedFingerprints, METH_O, NULL},
	 { "CIRCNetwork_SetIRCConnectEnabled", _wrap_CIRCNetwork_SetIRCConnectEnabled, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetIRCConnectEnabled", _wrap_CIRCNetwork_GetIRCConnectEnabled, METH_O, NULL},
	 { "CIRCNetwork_GetIRCSock", _wrap_CIRCNetwork_GetIRCSock, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetIRCServer", _wrap_CIRCNetwork_GetIRCServer, METH_O, NULL},
	 { "CIRCNetwork_GetIRCNick", _wrap_CIRCNetwork_GetIRCNick, METH_O, NULL},
	 { "CIRCNetwork_SetIRCNick", _wrap_CIRCNetwork_SetIRCNick, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetCurNick", _wrap_CIRCNetwork_GetCurNick, METH_O, NULL},
	 { "CIRCNetwork_IsIRCAway", _wrap_CIRCNetwork_IsIRCAway, METH_O, NULL},
	 { "CIRCNetwork_SetIRCAway", _wrap_CIRCNetwork_SetIRCAway, METH_VARARGS, NULL},
	 { "CIRCNetwork_Connect", _wrap_CIRCNetwork_Connect, METH_O, NULL},
	 { "CIRCNetwork_IsIRCConnected", _wrap_CIRCNetwork_IsIRCConnected, METH_O, NULL},
	 { "CIRCNetwork_SetIRCSocket", _wrap_CIRCNetwork_SetIRCSocket, METH_VARARGS, NULL},
	 { "CIRCNetwork_IRCConnected", _wrap_CIRCNetwork_IRCConnected, METH_O, NULL},
	 { "CIRCNetwork_IRCDisconnected", _wrap_CIRCNetwork_IRCDisconnected, METH_O, NULL},
	 { "CIRCNetwork_CheckIRCConnect", _wrap_CIRCNetwork_CheckIRCConnect, METH_O, NULL},
	 { "CIRCNetwork_NotifyClientsAboutServerDependentCap", _wrap_CIRCNetwork_NotifyClientsAboutServerDependentCap, METH_VARARGS, NULL},
	 { "CIRCNetwork_IsServerCapAccepted", _wrap_CIRCNetwork_IsServerCapAccepted, METH_VARARGS, NULL},
	 { "CIRCNetwork_PutIRC", _wrap_CIRCNetwork_PutIRC, METH_VARARGS, NULL},
	 { "CIRCNetwork_ClearRawBuffer", _wrap_CIRCNetwork_ClearRawBuffer, METH_O, NULL},
	 { "CIRCNetwork_AddRawBuffer", _wrap_CIRCNetwork_AddRawBuffer, METH_VARARGS, NULL},
	 { "CIRCNetwork_UpdateRawBuffer", _wrap_CIRCNetwork_UpdateRawBuffer, METH_VARARGS, NULL},
	 { "CIRCNetwork_UpdateExactRawBuffer", _wrap_CIRCNetwork_UpdateExactRawBuffer, METH_VARARGS, NULL},
	 { "CIRCNetwork_ClearMotdBuffer", _wrap_CIRCNetwork_ClearMotdBuffer, METH_O, NULL},
	 { "CIRCNetwork_AddMotdBuffer", _wrap_CIRCNetwork_AddMotdBuffer, METH_VARARGS, NULL},
	 { "CIRCNetwork_UpdateMotdBuffer", _wrap_CIRCNetwork_UpdateMotdBuffer, METH_VARARGS, NULL},
	 { "CIRCNetwork_ClearNoticeBuffer", _wrap_CIRCNetwork_ClearNoticeBuffer, METH_O, NULL},
	 { "CIRCNetwork_AddNoticeBuffer", _wrap_CIRCNetwork_AddNoticeBuffer, METH_VARARGS, NULL},
	 { "CIRCNetwork_UpdateNoticeBuffer", _wrap_CIRCNetwork_UpdateNoticeBuffer, METH_VARARGS, NULL},
	 { "CIRCNetwork_ClearQueryBuffer", _wrap_CIRCNetwork_ClearQueryBuffer, METH_O, NULL},
	 { "CIRCNetwork_GetNick", _wrap_CIRCNetwork_GetNick, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetAltNick", _wrap_CIRCNetwork_GetAltNick, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetIdent", _wrap_CIRCNetwork_GetIdent, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetRealName", _wrap_CIRCNetwork_GetRealName, METH_O, NULL},
	 { "CIRCNetwork_GetBindHost", _wrap_CIRCNetwork_GetBindHost, METH_O, NULL},
	 { "CIRCNetwork_GetEncoding", _wrap_CIRCNetwork_GetEncoding, METH_O, NULL},
	 { "CIRCNetwork_GetQuitMsg", _wrap_CIRCNetwork_GetQuitMsg, METH_O, NULL},
	 { "CIRCNetwork_SetNick", _wrap_CIRCNetwork_SetNick, METH_VARARGS, NULL},
	 { "CIRCNetwork_SetAltNick", _wrap_CIRCNetwork_SetAltNick, METH_VARARGS, NULL},
	 { "CIRCNetwork_SetIdent", _wrap_CIRCNetwork_SetIdent, METH_VARARGS, NULL},
	 { "CIRCNetwork_SetRealName", _wrap_CIRCNetwork_SetRealName, METH_VARARGS, NULL},
	 { "CIRCNetwork_SetBindHost", _wrap_CIRCNetwork_SetBindHost, METH_VARARGS, NULL},
	 { "CIRCNetwork_SetEncoding", _wrap_CIRCNetwork_SetEncoding, METH_VARARGS, NULL},
	 { "CIRCNetwork_SetQuitMsg", _wrap_CIRCNetwork_SetQuitMsg, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetFloodRate", _wrap_CIRCNetwork_GetFloodRate, METH_O, NULL},
	 { "CIRCNetwork_GetFloodBurst", _wrap_CIRCNetwork_GetFloodBurst, METH_O, NULL},
	 { "CIRCNetwork_SetFloodRate", _wrap_CIRCNetwork_SetFloodRate, METH_VARARGS, NULL},
	 { "CIRCNetwork_SetFloodBurst", _wrap_CIRCNetwork_SetFloodBurst, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetJoinDelay", _wrap_CIRCNetwork_GetJoinDelay, METH_O, NULL},
	 { "CIRCNetwork_SetJoinDelay", _wrap_CIRCNetwork_SetJoinDelay, METH_VARARGS, NULL},
	 { "CIRCNetwork_SetTrustAllCerts", _wrap_CIRCNetwork_SetTrustAllCerts, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetTrustAllCerts", _wrap_CIRCNetwork_GetTrustAllCerts, METH_O, NULL},
	 { "CIRCNetwork_SetTrustPKI", _wrap_CIRCNetwork_SetTrustPKI, METH_VARARGS, NULL},
	 { "CIRCNetwork_GetTrustPKI", _wrap_CIRCNetwork_GetTrustPKI, METH_O, NULL},
	 { "CIRCNetwork_BytesRead", _wrap_CIRCNetwork_BytesRead, METH_O, NULL},
	 { "CIRCNetwork_BytesWritten", _wrap_CIRCNetwork_BytesWritten, METH_O, NULL},
	 { "CIRCNetwork_AddBytesRead", _wrap_CIRCNetwork_AddBytesRead, METH_VARARGS, NULL},
	 { "CIRCNetwork_AddBytesWritten", _wrap_CIRCNetwork_AddBytesWritten, METH_VARARGS, NULL},
	 { "CIRCNetwork_ExpandString", _wrap_CIRCNetwork_ExpandString, METH_VARARGS, NULL},
	 { "CIRCNetwork___str__", _wrap_CIRCNetwork___str__, METH_O, NULL},
	 { "CIRCNetwork___repr__", _wrap_CIRCNetwork___repr__, METH_O, NULL},
	 { "CIRCNetwork_GetChans_", _wrap_CIRCNetwork_GetChans_, METH_O, NULL},
	 { "CIRCNetwork_GetServers_", _wrap_CIRCNetwork_GetServers_, METH_O, NULL},
	 { "CIRCNetwork_GetQueries_", _wrap_CIRCNetwork_GetQueries_, METH_O, NULL},
	 { "CIRCNetwork_swigregister", CIRCNetwork_swigregister, METH_O, NULL},
	 { "CIRCNetwork_swiginit", CIRCNetwork_swiginit, METH_VARARGS, NULL},
	 { "new_CQuery", _wrap_new_CQuery, METH_VARARGS, NULL},
	 { "delete_CQuery", _wrap_delete_CQuery, METH_O, NULL},
	 { "CQuery_GetBuffer", _wrap_CQuery_GetBuffer, METH_O, NULL},
	 { "CQuery_GetBufferCount", _wrap_CQuery_GetBufferCount, METH_O, NULL},
	 { "CQuery_SetBufferCount", _wrap_CQuery_SetBufferCount, METH_VARARGS, NULL},
	 { "CQuery_AddBuffer", _wrap_CQuery_AddBuffer, METH_VARARGS, NULL},
	 { "CQuery_ClearBuffer", _wrap_CQuery_ClearBuffer, METH_O, NULL},
	 { "CQuery_SendBuffer", _wrap_CQuery_SendBuffer, METH_VARARGS, NULL},
	 { "CQuery_GetName", _wrap_CQuery_GetName, METH_O, NULL},
	 { "CQuery_swigregister", CQuery_swigregister, METH_O, NULL},
	 { "CQuery_swiginit", CQuery_swiginit, METH_VARARGS, NULL},
	 { "delete_CAuthBase", _wrap_delete_CAuthBase, METH_O, NULL},
	 { "CAuthBase_SetLoginInfo", _wrap_CAuthBase_SetLoginInfo, METH_VARARGS, NULL},
	 { "CAuthBase_AcceptLogin", _wrap_CAuthBase_AcceptLogin, METH_VARARGS, NULL},
	 { "CAuthBase_RefuseLogin", _wrap_CAuthBase_RefuseLogin, METH_VARARGS, NULL},
	 { "CAuthBase_GetUsername", _wrap_CAuthBase_GetUsername, METH_O, NULL},
	 { "CAuthBase_GetPassword", _wrap_CAuthBase_GetPassword, METH_O, NULL},
	 { "CAuthBase_GetSocket", _wrap_CAuthBase_GetSocket, METH_O, NULL},
	 { "CAuthBase_GetRemoteIP", _wrap_CAuthBase_GetRemoteIP, METH_O, NULL},
	 { "CAuthBase_Invalidate", _wrap_CAuthBase_Invalidate, METH_O, NULL},
	 { "CAuthBase_swigregister", CAuthBase_swigregister, METH_O, NULL},
	 { "new_CClientAuth", _wrap_new_CClientAuth, METH_VARARGS, NULL},
	 { "delete_CClientAuth", _wrap_delete_CClientAuth, METH_O, NULL},
	 { "CClientAuth_Invalidate", _wrap_CClientAuth_Invalidate, METH_O, NULL},
	 { "CClientAuth_AcceptedLogin", _wrap_CClientAuth_AcceptedLogin, METH_VARARGS, NULL},
	 { "CClientAuth_RefusedLogin", _wrap_CClientAuth_RefusedLogin, METH_VARARGS, NULL},
	 { "CClientAuth_swigregister", CClientAuth_swigregister, METH_O, NULL},
	 { "CClientAuth_swiginit", CClientAuth_swiginit, METH_VARARGS, NULL},
	 { "new_CClient", _wrap_new_CClient, METH_NOARGS, NULL},
	 { "delete_CClient", _wrap_delete_CClient, METH_O, NULL},
	 { "CClient_SendRequiredPasswordNotice", _wrap_CClient_SendRequiredPasswordNotice, METH_O, NULL},
	 { "CClient_AcceptLogin", _wrap_CClient_AcceptLogin, METH_VARARGS, NULL},
	 { "CClient_RefuseLogin", _wrap_CClient_RefuseLogin, METH_VARARGS, NULL},
	 { "CClient_GetNick", _wrap_CClient_GetNick, METH_VARARGS, NULL},
	 { "CClient_GetNickMask", _wrap_CClient_GetNickMask, METH_O, NULL},
	 { "CClient_GetIdentifier", _wrap_CClient_GetIdentifier, METH_O, NULL},
	 { "CClient_CapVersion", _wrap_CClient_CapVersion, METH_O, NULL},
	 { "CClient_HasCap302", _wrap_CClient_HasCap302, METH_O, NULL},
	 { "CClient_HasCapNotify", _wrap_CClient_HasCapNotify, METH_O, NULL},
	 { "CClient_HasAwayNotify", _wrap_CClient_HasAwayNotify, METH_O, NULL},
	 { "CClient_HasAccountNotify", _wrap_CClient_HasAccountNotify, METH_O, NULL},
	 { "CClient_HasInviteNotify", _wrap_CClient_HasInviteNotify, METH_O, NULL},
	 { "CClient_HasExtendedJoin", _wrap_CClient_HasExtendedJoin, METH_O, NULL},
	 { "CClient_HasNamesx", _wrap_CClient_HasNamesx, METH_O, NULL},
	 { "CClient_HasUHNames", _wrap_CClient_HasUHNames, METH_O, NULL},
	 { "CClient_HasChgHost", _wrap_CClient_HasChgHost, METH_O, NULL},
	 { "CClient_IsAway", _wrap_CClient_IsAway, METH_O, NULL},
	 { "CClient_HasServerTime", _wrap_CClient_HasServerTime, METH_O, NULL},
	 { "CClient_HasBatch", _wrap_CClient_HasBatch, METH_O, NULL},
	 { "CClient_HasEchoMessage", _wrap_CClient_HasEchoMessage, METH_O, NULL},
	 { "CClient_HasSelfMessage", _wrap_CClient_HasSelfMessage, METH_O, NULL},
	 { "CClient_IsValidIdentifier", _wrap_CClient_IsValidIdentifier, METH_O, NULL},
	 { "CClient_UserCommand", _wrap_CClient_UserCommand, METH_VARARGS, NULL},
	 { "CClient_UserPortCommand", _wrap_CClient_UserPortCommand, METH_VARARGS, NULL},
	 { "CClient_StatusCTCP", _wrap_CClient_StatusCTCP, METH_VARARGS, NULL},
	 { "CClient_BouncedOff", _wrap_CClient_BouncedOff, METH_O, NULL},
	 { "CClient_IsAttached", _wrap_CClient_IsAttached, METH_O, NULL},
	 { "CClient_IsPlaybackActive", _wrap_CClient_IsPlaybackActive, METH_O, NULL},
	 { "CClient_SetPlaybackActive", _wrap_CClient_SetPlaybackActive, METH_VARARGS, NULL},
	 { "CClient_PutIRC", _wrap_CClient_PutIRC, METH_VARARGS, NULL},
	 { "CClient_PutIRCStripping", _wrap_CClient_PutIRCStripping, METH_VARARGS, NULL},
	 { "CClient_PutClientRaw", _wrap_CClient_PutClientRaw, METH_VARARGS, NULL},
	 { "CClient_PutClient", _wrap_CClient_PutClient, METH_VARARGS, NULL},
	 { "CClient_PutStatus", _wrap_CClient_PutStatus, METH_VARARGS, NULL},
	 { "CClient_PutStatusNotice", _wrap_CClient_PutStatusNotice, METH_VARARGS, NULL},
	 { "CClient_PutModule", _wrap_CClient_PutModule, METH_VARARGS, NULL},
	 { "CClient_PutModNotice", _wrap_CClient_PutModNotice, METH_VARARGS, NULL},
	 { "CClient_IsCapEnabled", _wrap_CClient_IsCapEnabled, METH_VARARGS, NULL},
	 { "CClient_IsTagEnabled", _wrap_CClient_IsTagEnabled, METH_VARARGS, NULL},
	 { "CClient_SetTagSupport", _wrap_CClient_SetTagSupport, METH_VARARGS, NULL},
	 { "CClient_NotifyServerDependentCap", _wrap_CClient_NotifyServerDependentCap, METH_VARARGS, NULL},
	 { "CClient_ReadLine", _wrap_CClient_ReadLine, METH_VARARGS, NULL},
	 { "CClient_SendMotd", _wrap_CClient_SendMotd, METH_O, NULL},
	 { "CClient_HelpUser", _wrap_CClient_HelpUser, METH_VARARGS, NULL},
	 { "CClient_AuthUser", _wrap_CClient_AuthUser, METH_O, NULL},
	 { "CClient_Connected", _wrap_CClient_Connected, METH_O, NULL},
	 { "CClient_Timeout", _wrap_CClient_Timeout, METH_O, NULL},
	 { "CClient_Disconnected", _wrap_CClient_Disconnected, METH_O, NULL},
	 { "CClient_ConnectionRefused", _wrap_CClient_ConnectionRefused, METH_O, NULL},
	 { "CClient_ReachedMaxBuffer", _wrap_CClient_ReachedMaxBuffer, METH_O, NULL},
	 { "CClient_SetNick", _wrap_CClient_SetNick, METH_VARARGS, NULL},
	 { "CClient_SetAway", _wrap_CClient_SetAway, METH_VARARGS, NULL},
	 { "CClient_GetUser", _wrap_CClient_GetUser, METH_O, NULL},
	 { "CClient_SetNetwork", _wrap_CClient_SetNetwork, METH_VARARGS, NULL},
	 { "CClient_GetNetwork", _wrap_CClient_GetNetwork, METH_O, NULL},
	 { "CClient_GetClients", _wrap_CClient_GetClients, METH_O, NULL},
	 { "CClient_GetIRCSock", _wrap_CClient_GetIRCSock, METH_VARARGS, NULL},
	 { "CClient_GetFullName", _wrap_CClient_GetFullName, METH_O, NULL},
	 { "CClient_SendSASLChallenge", _wrap_CClient_SendSASLChallenge, METH_VARARGS, NULL},
	 { "CClient_RefuseSASLLogin", _wrap_CClient_RefuseSASLLogin, METH_VARARGS, NULL},
	 { "CClient_AcceptSASLLogin", _wrap_CClient_AcceptSASLLogin, METH_VARARGS, NULL},
	 { "CClient_StartSASLPasswordCheck", _wrap_CClient_StartSASLPasswordCheck, METH_VARARGS, NULL},
	 { "CClient_ParseUser", _wrap_CClient_ParseUser, METH_VARARGS, NULL},
	 { "CClient_swigregister", CClient_swigregister, METH_O, NULL},
	 { "CClient_swiginit", CClient_swiginit, METH_VARARGS, NULL},
	 { "new_CIRCSock", _wrap_new_CIRCSock, METH_O, NULL},
	 { "delete_CIRCSock", _wrap_delete_CIRCSock, METH_O, NULL},
	 { "CIRCSock_ReadLine", _wrap_CIRCSock_ReadLine, METH_VARARGS, NULL},
	 { "CIRCSock_Connected", _wrap_CIRCSock_Connected, METH_O, NULL},
	 { "CIRCSock_Disconnected", _wrap_CIRCSock_Disconnected, METH_O, NULL},
	 { "CIRCSock_ConnectionRefused", _wrap_CIRCSock_ConnectionRefused, METH_O, NULL},
	 { "CIRCSock_SockError", _wrap_CIRCSock_SockError, METH_VARARGS, NULL},
	 { "CIRCSock_Timeout", _wrap_CIRCSock_Timeout, METH_O, NULL},
	 { "CIRCSock_ReachedMaxBuffer", _wrap_CIRCSock_ReachedMaxBuffer, METH_O, NULL},
	 { "CIRCSock_PutIRCRaw", _wrap_CIRCSock_PutIRCRaw, METH_VARARGS, NULL},
	 { "CIRCSock_PutIRC", _wrap_CIRCSock_PutIRC, METH_VARARGS, NULL},
	 { "CIRCSock_PutIRCQuick", _wrap_CIRCSock_PutIRCQuick, METH_VARARGS, NULL},
	 { "CIRCSock_ResetChans", _wrap_CIRCSock_ResetChans, METH_O, NULL},
	 { "CIRCSock_Quit", _wrap_CIRCSock_Quit, METH_VARARGS, NULL},
	 { "CIRCSock_PauseCap", _wrap_CIRCSock_PauseCap, METH_O, NULL},
	 { "CIRCSock_ResumeCap", _wrap_CIRCSock_ResumeCap, METH_O, NULL},
	 { "CIRCSock_IsTagEnabled", _wrap_CIRCSock_IsTagEnabled, METH_VARARGS, NULL},
	 { "CIRCSock_SetTagSupport", _wrap_CIRCSock_SetTagSupport, METH_VARARGS, NULL},
	 { "CIRCSock_SetPass", _wrap_CIRCSock_SetPass, METH_VARARGS, NULL},
	 { "CIRCSock_GetMaxNickLen", _wrap_CIRCSock_GetMaxNickLen, METH_O, NULL},
	 { "CIRCSock_GetModeType", _wrap_CIRCSock_GetModeType, METH_VARARGS, NULL},
	 { "CIRCSock_GetPermFromMode", _wrap_CIRCSock_GetPermFromMode, METH_VARARGS, NULL},
	 { "CIRCSock_GetModeFromPerm", _wrap_CIRCSock_GetModeFromPerm, METH_VARARGS, NULL},
	 { "CIRCSock_GetChanModes", _wrap_CIRCSock_GetChanModes, METH_O, NULL},
	 { "CIRCSock_IsPermChar", _wrap_CIRCSock_IsPermChar, METH_VARARGS, NULL},
	 { "CIRCSock_IsPermMode", _wrap_CIRCSock_IsPermMode, METH_VARARGS, NULL},
	 { "CIRCSock_GetPerms", _wrap_CIRCSock_GetPerms, METH_O, NULL},
	 { "CIRCSock_GetPermModes", _wrap_CIRCSock_GetPermModes, METH_O, NULL},
	 { "CIRCSock_GetNickMask", _wrap_CIRCSock_GetNickMask, METH_O, NULL},
	 { "CIRCSock_GetNick", _wrap_CIRCSock_GetNick, METH_O, NULL},
	 { "CIRCSock_GetPass", _wrap_CIRCSock_GetPass, METH_O, NULL},
	 { "CIRCSock_GetNetwork", _wrap_CIRCSock_GetNetwork, METH_O, NULL},
	 { "CIRCSock_HasNamesx", _wrap_CIRCSock_HasNamesx, METH_O, NULL},
	 { "CIRCSock_HasUHNames", _wrap_CIRCSock_HasUHNames, METH_O, NULL},
	 { "CIRCSock_HasAwayNotify", _wrap_CIRCSock_HasAwayNotify, METH_O, NULL},
	 { "CIRCSock_HasAccountNotify", _wrap_CIRCSock_HasAccountNotify, METH_O, NULL},
	 { "CIRCSock_HasExtendedJoin", _wrap_CIRCSock_HasExtendedJoin, METH_O, NULL},
	 { "CIRCSock_HasServerTime", _wrap_CIRCSock_HasServerTime, METH_O, NULL},
	 { "CIRCSock_HasMessageTagCap", _wrap_CIRCSock_HasMessageTagCap, METH_O, NULL},
	 { "CIRCSock_GetUserModes", _wrap_CIRCSock_GetUserModes, METH_O, NULL},
	 { "CIRCSock_IsAuthed", _wrap_CIRCSock_IsAuthed, METH_O, NULL},
	 { "CIRCSock_GetAcceptedCaps", _wrap_CIRCSock_GetAcceptedCaps, METH_O, NULL},
	 { "CIRCSock_IsCapAccepted", _wrap_CIRCSock_IsCapAccepted, METH_VARARGS, NULL},
	 { "CIRCSock_GetCapLsValue", _wrap_CIRCSock_GetCapLsValue, METH_VARARGS, NULL},
	 { "CIRCSock_GetISupport", _wrap_CIRCSock_GetISupport, METH_VARARGS, NULL},
	 { "CIRCSock_IsFloodProtected", _wrap_CIRCSock_IsFloodProtected, METH_O, NULL},
	 { "CIRCSock_swigregister", CIRCSock_swigregister, METH_O, NULL},
	 { "CIRCSock_swiginit", CIRCSock_swiginit, METH_VARARGS, NULL},
	 { "delete_CListener", _wrap_delete_CListener, METH_O, NULL},
	 { "CListener_IsSSL", _wrap_CListener_IsSSL, METH_O, NULL},
	 { "CListener_GetRealListener", _wrap_CListener_GetRealListener, METH_O, NULL},
	 { "CListener_GetURIPrefix", _wrap_CListener_GetURIPrefix, METH_O, NULL},
	 { "CListener_GetAcceptType", _wrap_CListener_GetAcceptType, METH_O, NULL},
	 { "CListener_SetAcceptType", _wrap_CListener_SetAcceptType, METH_VARARGS, NULL},
	 { "CListener_Listen", _wrap_CListener_Listen, METH_O, NULL},
	 { "CListener_ResetRealListener", _wrap_CListener_ResetRealListener, METH_O, NULL},
	 { "CListener_ToConfig", _wrap_CListener_ToConfig, METH_O, NULL},
	 { "CListener_swigregister", CListener_swigregister, METH_O, NULL},
	 { "new_CTCPListener", _wrap_new_CTCPListener, METH_VARARGS, NULL},
	 { "delete_CTCPListener", _wrap_delete_CTCPListener, METH_O, NULL},
	 { "CTCPListener_GetAddrType", _wrap_CTCPListener_GetAddrType, METH_O, NULL},
	 { "CTCPListener_GetPort", _wrap_CTCPListener_GetPort, METH_O, NULL},
	 { "CTCPListener_GetBindHost", _wrap_CTCPListener_GetBindHost, METH_O, NULL},
	 { "CTCPListener_Listen", _wrap_CTCPListener_Listen, METH_O, NULL},
	 { "CTCPListener_ToConfig", _wrap_CTCPListener_ToConfig, METH_O, NULL},
	 { "CTCPListener_swigregister", CTCPListener_swigregister, METH_O, NULL},
	 { "CTCPListener_swiginit", CTCPListener_swiginit, METH_VARARGS, NULL},
	 { "new_CUnixListener", _wrap_new_CUnixListener, METH_VARARGS, NULL},
	 { "delete_CUnixListener", _wrap_delete_CUnixListener, METH_O, NULL},
	 { "CUnixListener_GetPath", _wrap_CUnixListener_GetPath, METH_O, NULL},
	 { "CUnixListener_GetGroup", _wrap_CUnixListener_GetGroup, METH_O, NULL},
	 { "CUnixListener_GetMode", _wrap_CUnixListener_GetMode, METH_O, NULL},
	 { "CUnixListener_Listen", _wrap_CUnixListener_Listen, METH_O, NULL},
	 { "CUnixListener_ToConfig", _wrap_CUnixListener_ToConfig, METH_O, NULL},
	 { "CUnixListener_swigregister", CUnixListener_swigregister, METH_O, NULL},
	 { "CUnixListener_swiginit", CUnixListener_swiginit, METH_VARARGS, NULL},
	 { "new_CRealListener", _wrap_new_CRealListener, METH_O, NULL},
	 { "delete_CRealListener", _wrap_delete_CRealListener, METH_O, NULL},
	 { "CRealListener_ConnectionFrom", _wrap_CRealListener_ConnectionFrom, METH_VARARGS, NULL},
	 { "CRealListener_GetSockObj", _wrap_CRealListener_GetSockObj, METH_VARARGS, NULL},
	 { "CRealListener_SockError", _wrap_CRealListener_SockError, METH_VARARGS, NULL},
	 { "CRealListener_swigregister", CRealListener_swigregister, METH_O, NULL},
	 { "CRealListener_swiginit", CRealListener_swiginit, METH_VARARGS, NULL},
	 { "new_CIncomingConnection", _wrap_new_CIncomingConnection, METH_VARARGS, NULL},
	 { "delete_CIncomingConnection", _wrap_delete_CIncomingConnection, METH_O, NULL},
	 { "CIncomingConnection_ReadLine", _wrap_CIncomingConnection_ReadLine, METH_VARARGS, NULL},
	 { "CIncomingConnection_ReachedMaxBuffer", _wrap_CIncomingConnection_ReachedMaxBuffer, METH_O, NULL},
	 { "CIncomingConnection_swigregister", CIncomingConnection_swigregister, METH_O, NULL},
	 { "CIncomingConnection_swiginit", CIncomingConnection_swiginit, METH_VARARGS, NULL},
	 { "delete_CHTTPSock", _wrap_delete_CHTTPSock, METH_O, NULL},
	 { "CHTTPSock_ReadData", _wrap_CHTTPSock_ReadData, METH_VARARGS, NULL},
	 { "CHTTPSock_ReadLine", _wrap_CHTTPSock_ReadLine, METH_VARARGS, NULL},
	 { "CHTTPSock_Connected", _wrap_CHTTPSock_Connected, METH_O, NULL},
	 { "CHTTPSock_GetSockObj", _wrap_CHTTPSock_GetSockObj, METH_VARARGS, NULL},
	 { "CHTTPSock_ForceLogin", _wrap_CHTTPSock_ForceLogin, METH_O, NULL},
	 { "CHTTPSock_OnLogin", _wrap_CHTTPSock_OnLogin, METH_VARARGS, NULL},
	 { "CHTTPSock_OnPageRequest", _wrap_CHTTPSock_OnPageRequest, METH_VARARGS, NULL},
	 { "CHTTPSock_PrintFile", _wrap_CHTTPSock_PrintFile, METH_VARARGS, NULL},
	 { "CHTTPSock_CheckPost", _wrap_CHTTPSock_CheckPost, METH_O, NULL},
	 { "CHTTPSock_SentHeader", _wrap_CHTTPSock_SentHeader, METH_O, NULL},
	 { "CHTTPSock_PrintHeader", _wrap_CHTTPSock_PrintHeader, METH_VARARGS, NULL},
	 { "CHTTPSock_AddHeader", _wrap_CHTTPSock_AddHeader, METH_VARARGS, NULL},
	 { "CHTTPSock_SetContentType", _wrap_CHTTPSock_SetContentType, METH_VARARGS, NULL},
	 { "CHTTPSock_PrintNotFound", _wrap_CHTTPSock_PrintNotFound, METH_O, NULL},
	 { "CHTTPSock_Redirect", _wrap_CHTTPSock_Redirect, METH_VARARGS, NULL},
	 { "CHTTPSock_PrintErrorPage", _wrap_CHTTPSock_PrintErrorPage, METH_VARARGS, NULL},
	 { "CHTTPSock_ParseParams", _wrap_CHTTPSock_ParseParams, METH_VARARGS, NULL},
	 { "CHTTPSock_ParseURI", _wrap_CHTTPSock_ParseURI, METH_O, NULL},
	 { "CHTTPSock_GetPage", _wrap_CHTTPSock_GetPage, METH_O, NULL},
	 { "CHTTPSock_GetDate", _wrap_CHTTPSock_GetDate, METH_VARARGS, NULL},
	 { "CHTTPSock_GetRemoteIP", _wrap_CHTTPSock_GetRemoteIP, METH_O, NULL},
	 { "CHTTPSock_GetRequestCookie", _wrap_CHTTPSock_GetRequestCookie, METH_VARARGS, NULL},
	 { "CHTTPSock_SendCookie", _wrap_CHTTPSock_SendCookie, METH_VARARGS, NULL},
	 { "CHTTPSock_SetDocRoot", _wrap_CHTTPSock_SetDocRoot, METH_VARARGS, NULL},
	 { "CHTTPSock_SetLoggedIn", _wrap_CHTTPSock_SetLoggedIn, METH_VARARGS, NULL},
	 { "CHTTPSock_GetPath", _wrap_CHTTPSock_GetPath, METH_O, NULL},
	 { "CHTTPSock_IsLoggedIn", _wrap_CHTTPSock_IsLoggedIn, METH_O, NULL},
	 { "CHTTPSock_GetDocRoot", _wrap_CHTTPSock_GetDocRoot, METH_O, NULL},
	 { "CHTTPSock_GetUser", _wrap_CHTTPSock_GetUser, METH_O, NULL},
	 { "CHTTPSock_GetPass", _wrap_CHTTPSock_GetPass, METH_O, NULL},
	 { "CHTTPSock_GetParamString", _wrap_CHTTPSock_GetParamString, METH_O, NULL},
	 { "CHTTPSock_GetContentType", _wrap_CHTTPSock_GetContentType, METH_O, NULL},
	 { "CHTTPSock_GetURI", _wrap_CHTTPSock_GetURI, METH_O, NULL},
	 { "CHTTPSock_GetURIPrefix", _wrap_CHTTPSock_GetURIPrefix, METH_O, NULL},
	 { "CHTTPSock_IsPost", _wrap_CHTTPSock_IsPost, METH_O, NULL},
	 { "CHTTPSock_GetParam", _wrap_CHTTPSock_GetParam, METH_VARARGS, NULL},
	 { "CHTTPSock_GetRawParam", _wrap_CHTTPSock_GetRawParam, METH_VARARGS, NULL},
	 { "CHTTPSock_HasParam", _wrap_CHTTPSock_HasParam, METH_VARARGS, NULL},
	 { "CHTTPSock_GetParams", _wrap_CHTTPSock_GetParams, METH_VARARGS, NULL},
	 { "CHTTPSock_GetParamValues", _wrap_CHTTPSock_GetParamValues, METH_VARARGS, NULL},
	 { "CHTTPSock_swigregister", CHTTPSock_swigregister, METH_O, NULL},
	 { "new_CTemplateTagHandler", _wrap_new_CTemplateTagHandler, METH_NOARGS, NULL},
	 { "delete_CTemplateTagHandler", _wrap_delete_CTemplateTagHandler, METH_O, NULL},
	 { "CTemplateTagHandler_HandleVar", _wrap_CTemplateTagHandler_HandleVar, METH_VARARGS, NULL},
	 { "CTemplateTagHandler_HandleTag", _wrap_CTemplateTagHandler_HandleTag, METH_VARARGS, NULL},
	 { "CTemplateTagHandler_HandleIf", _wrap_CTemplateTagHandler_HandleIf, METH_VARARGS, NULL},
	 { "CTemplateTagHandler_HandleValue", _wrap_CTemplateTagHandler_HandleValue, METH_VARARGS, NULL},
	 { "CTemplateTagHandler_swigregister", CTemplateTagHandler_swigregister, METH_O, NULL},
	 { "CTemplateTagHandler_swiginit", CTemplateTagHandler_swiginit, METH_VARARGS, NULL},
	 { "new_CTemplateOptions", _wrap_new_CTemplateOptions, METH_NOARGS, NULL},
	 { "delete_CTemplateOptions", _wrap_delete_CTemplateOptions, METH_O, NULL},
	 { "CTemplateOptions_Parse", _wrap_CTemplateOptions_Parse, METH_VARARGS, NULL},
	 { "CTemplateOptions_GetEscapeFrom", _wrap_CTemplateOptions_GetEscapeFrom, METH_O, NULL},
	 { "CTemplateOptions_GetEscapeTo", _wrap_CTemplateOptions_GetEscapeTo, METH_O, NULL},
	 { "CTemplateOptions_swigregister", CTemplateOptions_swigregister, METH_O, NULL},
	 { "CTemplateOptions_swiginit", CTemplateOptions_swiginit, METH_VARARGS, NULL},
	 { "delete_CTemplateLoopContext", _wrap_delete_CTemplateLoopContext, METH_O, NULL},
	 { "new_CTemplateLoopContext", _wrap_new_CTemplateLoopContext, METH_VARARGS, NULL},
	 { "CTemplateLoopContext_SetHasData", _wrap_CTemplateLoopContext_SetHasData, METH_VARARGS, NULL},
	 { "CTemplateLoopContext_SetName", _wrap_CTemplateLoopContext_SetName, METH_VARARGS, NULL},
	 { "CTemplateLoopContext_SetRowIndex", _wrap_CTemplateLoopContext_SetRowIndex, METH_VARARGS, NULL},
	 { "CTemplateLoopContext_IncRowIndex", _wrap_CTemplateLoopContext_IncRowIndex, METH_O, NULL},
	 { "CTemplateLoopContext_DecRowIndex", _wrap_CTemplateLoopContext_DecRowIndex, METH_O, NULL},
	 { "CTemplateLoopContext_SetFilePosition", _wrap_CTemplateLoopContext_SetFilePosition, METH_VARARGS, NULL},
	 { "CTemplateLoopContext_HasData", _wrap_CTemplateLoopContext_HasData, METH_O, NULL},
	 { "CTemplateLoopContext_GetName", _wrap_CTemplateLoopContext_GetName, METH_O, NULL},
	 { "CTemplateLoopContext_GetFilePosition", _wrap_CTemplateLoopContext_GetFilePosition, METH_O, NULL},
	 { "CTemplateLoopContext_GetRowIndex", _wrap_CTemplateLoopContext_GetRowIndex, METH_O, NULL},
	 { "CTemplateLoopContext_GetRowCount", _wrap_CTemplateLoopContext_GetRowCount, METH_O, NULL},
	 { "CTemplateLoopContext_GetRows", _wrap_CTemplateLoopContext_GetRows, METH_O, NULL},
	 { "CTemplateLoopContext_GetNextRow", _wrap_CTemplateLoopContext_GetNextRow, METH_O, NULL},
	 { "CTemplateLoopContext_GetCurRow", _wrap_CTemplateLoopContext_GetCurRow, METH_O, NULL},
	 { "CTemplateLoopContext_GetRow", _wrap_CTemplateLoopContext_GetRow, METH_VARARGS, NULL},
	 { "CTemplateLoopContext_GetValue", _wrap_CTemplateLoopContext_GetValue, METH_VARARGS, NULL},
	 { "CTemplateLoopContext_swigregister", CTemplateLoopContext_swigregister, METH_O, NULL},
	 { "CTemplateLoopContext_swiginit", CTemplateLoopContext_swiginit, METH_VARARGS, NULL},
	 { "delete_CTemplate", _wrap_delete_CTemplate, METH_O, NULL},
	 { "new_CTemplate", _wrap_new_CTemplate, METH_VARARGS, NULL},
	 { "CTemplate_AddTagHandler", _wrap_CTemplate_AddTagHandler, METH_VARARGS, NULL},
	 { "CTemplate_GetTagHandlers", _wrap_CTemplate_GetTagHandlers, METH_O, NULL},
	 { "CTemplate_ResolveLiteral", _wrap_CTemplate_ResolveLiteral, METH_VARARGS, NULL},
	 { "CTemplate_Init", _wrap_CTemplate_Init, METH_O, NULL},
	 { "CTemplate_GetParent", _wrap_CTemplate_GetParent, METH_VARARGS, NULL},
	 { "CTemplate_ExpandFile", _wrap_CTemplate_ExpandFile, METH_VARARGS, NULL},
	 { "CTemplate_SetFile", _wrap_CTemplate_SetFile, METH_VARARGS, NULL},
	 { "CTemplate_SetPath", _wrap_CTemplate_SetPath, METH_VARARGS, NULL},
	 { "CTemplate_MakePath", _wrap_CTemplate_MakePath, METH_VARARGS, NULL},
	 { "CTemplate_PrependPath", _wrap_CTemplate_PrependPath, METH_VARARGS, NULL},
	 { "CTemplate_AppendPath", _wrap_CTemplate_AppendPath, METH_VARARGS, NULL},
	 { "CTemplate_RemovePath", _wrap_CTemplate_RemovePath, METH_VARARGS, NULL},
	 { "CTemplate_ClearPaths", _wrap_CTemplate_ClearPaths, METH_O, NULL},
	 { "CTemplate_PrintString", _wrap_CTemplate_PrintString, METH_VARARGS, NULL},
	 { "CTemplate_Print", _wrap_CTemplate_Print, METH_VARARGS, NULL},
	 { "CTemplate_ValidIf", _wrap_CTemplate_ValidIf, METH_VARARGS, NULL},
	 { "CTemplate_ValidExpr", _wrap_CTemplate_ValidExpr, METH_VARARGS, NULL},
	 { "CTemplate_IsTrue", _wrap_CTemplate_IsTrue, METH_VARARGS, NULL},
	 { "CTemplate_HasLoop", _wrap_CTemplate_HasLoop, METH_VARARGS, NULL},
	 { "CTemplate_GetValue", _wrap_CTemplate_GetValue, METH_VARARGS, NULL},
	 { "CTemplate_AddRow", _wrap_CTemplate_AddRow, METH_VARARGS, NULL},
	 { "CTemplate_GetRow", _wrap_CTemplate_GetRow, METH_VARARGS, NULL},
	 { "CTemplate_GetLoop", _wrap_CTemplate_GetLoop, METH_VARARGS, NULL},
	 { "CTemplate_DelCurLoopContext", _wrap_CTemplate_DelCurLoopContext, METH_O, NULL},
	 { "CTemplate_GetCurLoopContext", _wrap_CTemplate_GetCurLoopContext, METH_O, NULL},
	 { "CTemplate_GetCurTemplate", _wrap_CTemplate_GetCurTemplate, METH_O, NULL},
	 { "CTemplate_GetFileName", _wrap_CTemplate_GetFileName, METH_O, NULL},
	 { "CTemplate_set", _wrap_CTemplate_set, METH_VARARGS, NULL},
	 { "CTemplate_swigregister", CTemplate_swigregister, METH_O, NULL},
	 { "CTemplate_swiginit", CTemplate_swiginit, METH_VARARGS, NULL},
	 { "new_CZNCTagHandler", _wrap_new_CZNCTagHandler, METH_O, NULL},
	 { "delete_CZNCTagHandler", _wrap_delete_CZNCTagHandler, METH_O, NULL},
	 { "CZNCTagHandler_HandleTag", _wrap_CZNCTagHandler_HandleTag, METH_VARARGS, NULL},
	 { "CZNCTagHandler_swigregister", CZNCTagHandler_swigregister, METH_O, NULL},
	 { "CZNCTagHandler_swiginit", CZNCTagHandler_swiginit, METH_VARARGS, NULL},
	 { "new_CWebSession", _wrap_new_CWebSession, METH_VARARGS, NULL},
	 { "delete_CWebSession", _wrap_delete_CWebSession, METH_O, NULL},
	 { "CWebSession_GetId", _wrap_CWebSession_GetId, METH_O, NULL},
	 { "CWebSession_GetIP", _wrap_CWebSession_GetIP, METH_O, NULL},
	 { "CWebSession_GetUser", _wrap_CWebSession_GetUser, METH_O, NULL},
	 { "CWebSession_GetLastActive", _wrap_CWebSession_GetLastActive, METH_O, NULL},
	 { "CWebSession_IsLoggedIn", _wrap_CWebSession_IsLoggedIn, METH_O, NULL},
	 { "CWebSession_IsAdmin", _wrap_CWebSession_IsAdmin, METH_O, NULL},
	 { "CWebSession_UpdateLastActive", _wrap_CWebSession_UpdateLastActive, METH_O, NULL},
	 { "CWebSession_SetUser", _wrap_CWebSession_SetUser, METH_VARARGS, NULL},
	 { "CWebSession_ClearMessageLoops", _wrap_CWebSession_ClearMessageLoops, METH_O, NULL},
	 { "CWebSession_FillMessageLoops", _wrap_CWebSession_FillMessageLoops, METH_VARARGS, NULL},
	 { "CWebSession_AddError", _wrap_CWebSession_AddError, METH_VARARGS, NULL},
	 { "CWebSession_AddSuccess", _wrap_CWebSession_AddSuccess, METH_VARARGS, NULL},
	 { "CWebSession_swigregister", CWebSession_swigregister, METH_O, NULL},
	 { "CWebSession_swiginit", CWebSession_swiginit, METH_VARARGS, NULL},
	 { "new_CWebSubPage", _wrap_new_CWebSubPage, METH_VARARGS, NULL},
	 { "delete_CWebSubPage", _wrap_delete_CWebSubPage, METH_O, NULL},
	 { "CWebSubPage_SetName", _wrap_CWebSubPage_SetName, METH_VARARGS, NULL},
	 { "CWebSubPage_SetTitle", _wrap_CWebSubPage_SetTitle, METH_VARARGS, NULL},
	 { "CWebSubPage_AddParam", _wrap_CWebSubPage_AddParam, METH_VARARGS, NULL},
	 { "CWebSubPage_RequiresAdmin", _wrap_CWebSubPage_RequiresAdmin, METH_O, NULL},
	 { "CWebSubPage_GetName", _wrap_CWebSubPage_GetName, METH_O, NULL},
	 { "CWebSubPage_GetTitle", _wrap_CWebSubPage_GetTitle, METH_O, NULL},
	 { "CWebSubPage_GetParams", _wrap_CWebSubPage_GetParams, METH_O, NULL},
	 { "CWebSubPage_swigregister", CWebSubPage_swigregister, METH_O, NULL},
	 { "CWebSubPage_swiginit", CWebSubPage_swiginit, METH_VARARGS, NULL},
	 { "new_CWebSessionMap", _wrap_new_CWebSessionMap, METH_VARARGS, NULL},
	 { "CWebSessionMap_FinishUserSessions", _wrap_CWebSessionMap_FinishUserSessions, METH_VARARGS, NULL},
	 { "delete_CWebSessionMap", _wrap_delete_CWebSessionMap, METH_O, NULL},
	 { "CWebSessionMap_swigregister", CWebSessionMap_swigregister, METH_O, NULL},
	 { "CWebSessionMap_swiginit", CWebSessionMap_swiginit, METH_VARARGS, NULL},
	 { "new_CWebSock", _wrap_new_CWebSock, METH_O, NULL},
	 { "delete_CWebSock", _wrap_delete_CWebSock, METH_O, NULL},
	 { "CWebSock_ForceLogin", _wrap_CWebSock_ForceLogin, METH_O, NULL},
	 { "CWebSock_OnLogin", _wrap_CWebSock_OnLogin, METH_VARARGS, NULL},
	 { "CWebSock_OnPageRequest", _wrap_CWebSock_OnPageRequest, METH_VARARGS, NULL},
	 { "CWebSock_PrintTemplate", _wrap_CWebSock_PrintTemplate, METH_VARARGS, NULL},
	 { "CWebSock_PrintStaticFile", _wrap_CWebSock_PrintStaticFile, METH_VARARGS, NULL},
	 { "CWebSock_FindTmpl", _wrap_CWebSock_FindTmpl, METH_VARARGS, NULL},
	 { "CWebSock_PrintErrorPage", _wrap_CWebSock_PrintErrorPage, METH_VARARGS, NULL},
	 { "CWebSock_GetSession", _wrap_CWebSock_GetSession, METH_O, NULL},
	 { "CWebSock_GetSockObj", _wrap_CWebSock_GetSockObj, METH_VARARGS, NULL},
	 { "CWebSock_GetSkinPath", _wrap_CWebSock_GetSkinPath, METH_O, NULL},
	 { "CWebSock_GetAvailSkins", _wrap_CWebSock_GetAvailSkins, METH_VARARGS, NULL},
	 { "CWebSock_GetSkinName", _wrap_CWebSock_GetSkinName, METH_O, NULL},
	 { "CWebSock_GetRequestCookie", _wrap_CWebSock_GetRequestCookie, METH_VARARGS, NULL},
	 { "CWebSock_SendCookie", _wrap_CWebSock_SendCookie, METH_VARARGS, NULL},
	 { "CWebSock_FinishUserSessions", _wrap_CWebSock_FinishUserSessions, METH_O, NULL},
	 { "CWebSock_GetCSRFCheck", _wrap_CWebSock_GetCSRFCheck, METH_O, NULL},
	 { "CWebSock_ValidateCSRFCheck", _wrap_CWebSock_ValidateCSRFCheck, METH_VARARGS, NULL},
	 { "CWebSock_swigregister", CWebSock_swigregister, METH_O, NULL},
	 { "CWebSock_swiginit", CWebSock_swiginit, METH_VARARGS, NULL},
	 { "new_CZNC", _wrap_new_CZNC, METH_NOARGS, NULL},
	 { "delete_CZNC", _wrap_delete_CZNC, METH_O, NULL},
	 { "CZNC_DeleteUsers", _wrap_CZNC_DeleteUsers, METH_O, NULL},
	 { "CZNC_Loop", _wrap_CZNC_Loop, METH_O, NULL},
	 { "CZNC_WritePidFile", _wrap_CZNC_WritePidFile, METH_VARARGS, NULL},
	 { "CZNC_DeletePidFile", _wrap_CZNC_DeletePidFile, METH_O, NULL},
	 { "CZNC_WaitForChildLock", _wrap_CZNC_WaitForChildLock, METH_O, NULL},
	 { "CZNC_IsHostAllowed", _wrap_CZNC_IsHostAllowed, METH_VARARGS, NULL},
	 { "CZNC_AllowConnectionFrom", _wrap_CZNC_AllowConnectionFrom, METH_VARARGS, NULL},
	 { "CZNC_InitDirs", _wrap_CZNC_InitDirs, METH_VARARGS, NULL},
	 { "CZNC_OnBoot", _wrap_CZNC_OnBoot, METH_O, NULL},
	 { "CZNC_ExpandConfigPath", _wrap_CZNC_ExpandConfigPath, METH_VARARGS, NULL},
	 { "CZNC_WriteNewConfig", _wrap_CZNC_WriteNewConfig, METH_VARARGS, NULL},
	 { "CZNC_WriteConfig", _wrap_CZNC_WriteConfig, METH_O, NULL},
	 { "CZNC_ParseConfig", _wrap_CZNC_ParseConfig, METH_VARARGS, NULL},
	 { "CZNC_RehashConfig", _wrap_CZNC_RehashConfig, METH_VARARGS, NULL},
	 { "CZNC_BackupConfigOnce", _wrap_CZNC_BackupConfigOnce, METH_VARARGS, NULL},
	 { "CZNC_GetVersion", _wrap_CZNC_GetVersion, METH_NOARGS, NULL},
	 { "CZNC_GetTag", _wrap_CZNC_GetTag, METH_VARARGS, NULL},
	 { "CZNC_GetCompileOptionsString", _wrap_CZNC_GetCompileOptionsString, METH_NOARGS, NULL},
	 { "CZNC_GetUptime", _wrap_CZNC_GetUptime, METH_O, NULL},
	 { "CZNC_ClearBindHosts", _wrap_CZNC_ClearBindHosts, METH_O, NULL},
	 { "CZNC_AddBindHost", _wrap_CZNC_AddBindHost, METH_VARARGS, NULL},
	 { "CZNC_RemBindHost", _wrap_CZNC_RemBindHost, METH_VARARGS, NULL},
	 { "CZNC_ClearTrustedProxies", _wrap_CZNC_ClearTrustedProxies, METH_O, NULL},
	 { "CZNC_AddTrustedProxy", _wrap_CZNC_AddTrustedProxy, METH_VARARGS, NULL},
	 { "CZNC_RemTrustedProxy", _wrap_CZNC_RemTrustedProxy, METH_VARARGS, NULL},
	 { "CZNC_GetClientCapBlacklist", _wrap_CZNC_GetClientCapBlacklist, METH_O, NULL},
	 { "CZNC_GetServerCapBlacklist", _wrap_CZNC_GetServerCapBlacklist, METH_O, NULL},
	 { "CZNC_Broadcast", _wrap_CZNC_Broadcast, METH_VARARGS, NULL},
	 { "CZNC_AddBytesRead", _wrap_CZNC_AddBytesRead, METH_VARARGS, NULL},
	 { "CZNC_AddBytesWritten", _wrap_CZNC_AddBytesWritten, METH_VARARGS, NULL},
	 { "CZNC_BytesRead", _wrap_CZNC_BytesRead, METH_O, NULL},
	 { "CZNC_BytesWritten", _wrap_CZNC_BytesWritten, METH_O, NULL},
	 { "CZNC_GetTrafficStats", _wrap_CZNC_GetTrafficStats, METH_VARARGS, NULL},
	 { "CZNC_GetNetworkTrafficStats", _wrap_CZNC_GetNetworkTrafficStats, METH_VARARGS, NULL},
	 { "CZNC_AuthUser", _wrap_CZNC_AuthUser, METH_VARARGS, NULL},
	 { "CZNC_SetConfigState", _wrap_CZNC_SetConfigState, METH_VARARGS, NULL},
	 { "CZNC_SetSkinName", _wrap_CZNC_SetSkinName, METH_VARARGS, NULL},
	 { "CZNC_SetStatusPrefix", _wrap_CZNC_SetStatusPrefix, METH_VARARGS, NULL},
	 { "CZNC_SetMaxBufferSize", _wrap_CZNC_SetMaxBufferSize, METH_VARARGS, NULL},
	 { "CZNC_SetAnonIPLimit", _wrap_CZNC_SetAnonIPLimit, METH_VARARGS, NULL},
	 { "CZNC_SetServerThrottle", _wrap_CZNC_SetServerThrottle, METH_VARARGS, NULL},
	 { "CZNC_SetProtectWebSessions", _wrap_CZNC_SetProtectWebSessions, METH_VARARGS, NULL},
	 { "CZNC_SetHideVersion", _wrap_CZNC_SetHideVersion, METH_VARARGS, NULL},
	 { "CZNC_SetAuthOnlyViaModule", _wrap_CZNC_SetAuthOnlyViaModule, METH_VARARGS, NULL},
	 { "CZNC_SetConnectDelay", _wrap_CZNC_SetConnectDelay, METH_VARARGS, NULL},
	 { "CZNC_SetSSLCiphers", _wrap_CZNC_SetSSLCiphers, METH_VARARGS, NULL},
	 { "CZNC_SetSSLProtocols", _wrap_CZNC_SetSSLProtocols, METH_VARARGS, NULL},
	 { "CZNC_SetSSLCertFile", _wrap_CZNC_SetSSLCertFile, METH_VARARGS, NULL},
	 { "CZNC_SetConfigWriteDelay", _wrap_CZNC_SetConfigWriteDelay, METH_VARARGS, NULL},
	 { "CZNC_GetConfigState", _wrap_CZNC_GetConfigState, METH_O, NULL},
	 { "CZNC_GetManager", _wrap_CZNC_GetManager, METH_VARARGS, NULL},
	 { "CZNC_GetModules", _wrap_CZNC_GetModules, METH_O, NULL},
	 { "CZNC_GetSkinName", _wrap_CZNC_GetSkinName, METH_O, NULL},
	 { "CZNC_GetStatusPrefix", _wrap_CZNC_GetStatusPrefix, METH_O, NULL},
	 { "CZNC_GetCurPath", _wrap_CZNC_GetCurPath, METH_O, NULL},
	 { "CZNC_GetHomePath", _wrap_CZNC_GetHomePath, METH_O, NULL},
	 { "CZNC_GetZNCPath", _wrap_CZNC_GetZNCPath, METH_O, NULL},
	 { "CZNC_GetConfPath", _wrap_CZNC_GetConfPath, METH_VARARGS, NULL},
	 { "CZNC_GetUserPath", _wrap_CZNC_GetUserPath, METH_O, NULL},
	 { "CZNC_GetModPath", _wrap_CZNC_GetModPath, METH_O, NULL},
	 { "CZNC_GetPemLocation", _wrap_CZNC_GetPemLocation, METH_O, NULL},
	 { "CZNC_GetKeyLocation", _wrap_CZNC_GetKeyLocation, METH_O, NULL},
	 { "CZNC_GetDHParamLocation", _wrap_CZNC_GetDHParamLocation, METH_O, NULL},
	 { "CZNC_GetConfigFile", _wrap_CZNC_GetConfigFile, METH_O, NULL},
	 { "CZNC_WritePemFile", _wrap_CZNC_WritePemFile, METH_O, NULL},
	 { "CZNC_GetBindHosts", _wrap_CZNC_GetBindHosts, METH_O, NULL},
	 { "CZNC_GetTrustedProxies", _wrap_CZNC_GetTrustedProxies, METH_O, NULL},
	 { "CZNC_GetListeners", _wrap_CZNC_GetListeners, METH_O, NULL},
	 { "CZNC_TimeStarted", _wrap_CZNC_TimeStarted, METH_O, NULL},
	 { "CZNC_GetMaxBufferSize", _wrap_CZNC_GetMaxBufferSize, METH_O, NULL},
	 { "CZNC_GetAnonIPLimit", _wrap_CZNC_GetAnonIPLimit, METH_O, NULL},
	 { "CZNC_GetConnectDelay", _wrap_CZNC_GetConnectDelay, METH_O, NULL},
	 { "CZNC_GetProtectWebSessions", _wrap_CZNC_GetProtectWebSessions, METH_O, NULL},
	 { "CZNC_GetHideVersion", _wrap_CZNC_GetHideVersion, METH_O, NULL},
	 { "CZNC_GetAuthOnlyViaModule", _wrap_CZNC_GetAuthOnlyViaModule, METH_O, NULL},
	 { "CZNC_GetSSLCiphers", _wrap_CZNC_GetSSLCiphers, METH_O, NULL},
	 { "CZNC_GetSSLProtocols", _wrap_CZNC_GetSSLProtocols, METH_O, NULL},
	 { "CZNC_GetDisabledSSLProtocols", _wrap_CZNC_GetDisabledSSLProtocols, METH_O, NULL},
	 { "CZNC_GetSSLCertFile", _wrap_CZNC_GetSSLCertFile, METH_O, NULL},
	 { "CZNC_GetAvailableSSLProtocols", _wrap_CZNC_GetAvailableSSLProtocols, METH_NOARGS, NULL},
	 { "CZNC_GetConfigWriteDelay", _wrap_CZNC_GetConfigWriteDelay, METH_O, NULL},
	 { "CZNC_CreateInstance", _wrap_CZNC_CreateInstance, METH_NOARGS, NULL},
	 { "CZNC_Get", _wrap_CZNC_Get, METH_NOARGS, NULL},
	 { "CZNC_DestroyInstance", _wrap_CZNC_DestroyInstance, METH_NOARGS, NULL},
	 { "CZNC_FindUser", _wrap_CZNC_FindUser, METH_VARARGS, NULL},
	 { "CZNC_FindModule", _wrap_CZNC_FindModule, METH_VARARGS, NULL},
	 { "CZNC_UpdateModule", _wrap_CZNC_UpdateModule, METH_VARARGS, NULL},
	 { "CZNC_DeleteUser", _wrap_CZNC_DeleteUser, METH_VARARGS, NULL},
	 { "CZNC_AddUser", _wrap_CZNC_AddUser, METH_VARARGS, NULL},
	 { "CZNC_GetUserMap", _wrap_CZNC_GetUserMap, METH_O, NULL},
	 { "CZNC_FindListener", _wrap_CZNC_FindListener, METH_VARARGS, NULL},
	 { "CZNC_FindUnixListener", _wrap_CZNC_FindUnixListener, METH_VARARGS, NULL},
	 { "CZNC_AddTCPListener", _wrap_CZNC_AddTCPListener, METH_VARARGS, NULL},
	 { "CZNC_AddUnixListener", _wrap_CZNC_AddUnixListener, METH_VARARGS, NULL},
	 { "CZNC_DelListener", _wrap_CZNC_DelListener, METH_VARARGS, NULL},
	 { "CZNC_AddListener", _wrap_CZNC_AddListener, METH_VARARGS, NULL},
	 { "CZNC_SetMotd", _wrap_CZNC_SetMotd, METH_VARARGS, NULL},
	 { "CZNC_AddMotd", _wrap_CZNC_AddMotd, METH_VARARGS, NULL},
	 { "CZNC_ClearMotd", _wrap_CZNC_ClearMotd, METH_O, NULL},
	 { "CZNC_GetMotd", _wrap_CZNC_GetMotd, METH_O, NULL},
	 { "CZNC_AddServerThrottle", _wrap_CZNC_AddServerThrottle, METH_VARARGS, NULL},
	 { "CZNC_GetServerThrottle", _wrap_CZNC_GetServerThrottle, METH_VARARGS, NULL},
	 { "CZNC_AddNetworkToQueue", _wrap_CZNC_AddNetworkToQueue, METH_VARARGS, NULL},
	 { "CZNC_GetConnectionQueue", _wrap_CZNC_GetConnectionQueue, METH_O, NULL},
	 { "CZNC_EnableConnectQueue", _wrap_CZNC_EnableConnectQueue, METH_O, NULL},
	 { "CZNC_DisableConnectQueue", _wrap_CZNC_DisableConnectQueue, METH_O, NULL},
	 { "CZNC_PauseConnectQueue", _wrap_CZNC_PauseConnectQueue, METH_O, NULL},
	 { "CZNC_ResumeConnectQueue", _wrap_CZNC_ResumeConnectQueue, METH_O, NULL},
	 { "CZNC_ForceEncoding", _wrap_CZNC_ForceEncoding, METH_O, NULL},
	 { "CZNC_UnforceEncoding", _wrap_CZNC_UnforceEncoding, METH_O, NULL},
	 { "CZNC_IsForcingEncoding", _wrap_CZNC_IsForcingEncoding, METH_O, NULL},
	 { "CZNC_FixupEncoding", _wrap_CZNC_FixupEncoding, METH_VARARGS, NULL},
	 { "CZNC_LeakConnectQueueTimer", _wrap_CZNC_LeakConnectQueueTimer, METH_VARARGS, NULL},
	 { "CZNC_DisableConfigTimer", _wrap_CZNC_DisableConfigTimer, METH_O, NULL},
	 { "CZNC_DumpConfig", _wrap_CZNC_DumpConfig, METH_O, NULL},
	 { "CZNC_GetUserMap_", _wrap_CZNC_GetUserMap_, METH_O, NULL},
	 { "CZNC_swigregister", CZNC_swigregister, METH_O, NULL},
	 { "CZNC_swiginit", CZNC_swiginit, METH_VARARGS, NULL},
	 { "new_CServer", _wrap_new_CServer, METH_VARARGS, NULL},
	 { "delete_CServer", _wrap_delete_CServer, METH_O, NULL},
	 { "CServer___eq__", _wrap_CServer___eq__, METH_VARARGS, NULL},
	 { "CServer___lt__", _wrap_CServer___lt__, METH_VARARGS, NULL},
	 { "CServer_Parse", _wrap_CServer_Parse, METH_O, NULL},
	 { "CServer_GetName", _wrap_CServer_GetName, METH_O, NULL},
	 { "CServer_GetPort", _wrap_CServer_GetPort, METH_O, NULL},
	 { "CServer_GetPass", _wrap_CServer_GetPass, METH_O, NULL},
	 { "CServer_IsSSL", _wrap_CServer_IsSSL, METH_O, NULL},
	 { "CServer_IsUnixSocket", _wrap_CServer_IsUnixSocket, METH_O, NULL},
	 { "CServer_GetString", _wrap_CServer_GetString, METH_VARARGS, NULL},
	 { "CServer_IsValidHostName", _wrap_CServer_IsValidHostName, METH_O, NULL},
	 { "CServer_swigregister", CServer_swigregister, METH_O, NULL},
	 { "CServer_swiginit", CServer_swiginit, METH_VARARGS, NULL},
	 { "CDebug_SetStdoutIsTTY", _wrap_CDebug_SetStdoutIsTTY, METH_O, NULL},
	 { "CDebug_StdoutIsTTY", _wrap_CDebug_StdoutIsTTY, METH_NOARGS, NULL},
	 { "CDebug_SetDebug", _wrap_CDebug_SetDebug, METH_O, NULL},
	 { "CDebug_Debug", _wrap_CDebug_Debug, METH_NOARGS, NULL},
	 { "CDebug_Filter", _wrap_CDebug_Filter, METH_O, NULL},
	 { "new_CDebug", _wrap_new_CDebug, METH_NOARGS, NULL},
	 { "delete_CDebug", _wrap_delete_CDebug, METH_O, NULL},
	 { "CDebug_swigregister", CDebug_swigregister, METH_O, NULL},
	 { "CDebug_swiginit", CDebug_swiginit, METH_VARARGS, NULL},
	 { "delete_CDebugStream", _wrap_delete_CDebugStream, METH_O, NULL},
	 { "new_CDebugStream", _wrap_new_CDebugStream, METH_NOARGS, NULL},
	 { "CDebugStream_swigregister", CDebugStream_swigregister, METH_O, NULL},
	 { "CDebugStream_swiginit", CDebugStream_swiginit, METH_VARARGS, NULL},
	 { "new_CExecSock", _wrap_new_CExecSock, METH_NOARGS, NULL},
	 { "CExecSock_Execute", _wrap_CExecSock_Execute, METH_VARARGS, NULL},
	 { "CExecSock_Kill", _wrap_CExecSock_Kill, METH_VARARGS, NULL},
	 { "delete_CExecSock", _wrap_delete_CExecSock, METH_O, NULL},
	 { "CExecSock_popen2", _wrap_CExecSock_popen2, METH_VARARGS, NULL},
	 { "CExecSock_close2", _wrap_CExecSock_close2, METH_VARARGS, NULL},
	 { "CExecSock_swigregister", CExecSock_swigregister, METH_O, NULL},
	 { "CExecSock_swiginit", CExecSock_swiginit, METH_VARARGS, NULL},
	 { "new_CBufLine", _wrap_new_CBufLine, METH_VARARGS, NULL},
	 { "delete_CBufLine", _wrap_delete_CBufLine, METH_O, NULL},
	 { "CBufLine_ToMessage", _wrap_CBufLine_ToMessage, METH_VARARGS, NULL},
	 { "CBufLine_GetLine", _wrap_CBufLine_GetLine, METH_VARARGS, NULL},
	 { "CBufLine_UpdateTime", _wrap_CBufLine_UpdateTime, METH_O, NULL},
	 { "CBufLine_Equals", _wrap_CBufLine_Equals, METH_VARARGS, NULL},
	 { "CBufLine_SetFormat", _wrap_CBufLine_SetFormat, METH_VARARGS, NULL},
	 { "CBufLine_SetText", _wrap_CBufLine_SetText, METH_VARARGS, NULL},
	 { "CBufLine_SetTime", _wrap_CBufLine_SetTime, METH_VARARGS, NULL},
	 { "CBufLine_SetTags", _wrap_CBufLine_SetTags, METH_VARARGS, NULL},
	 { "CBufLine_GetCommand", _wrap_CBufLine_GetCommand, METH_O, NULL},
	 { "CBufLine_GetFormat", _wrap_CBufLine_GetFormat, METH_O, NULL},
	 { "CBufLine_GetText", _wrap_CBufLine_GetText, METH_O, NULL},
	 { "CBufLine_GetTime", _wrap_CBufLine_GetTime, METH_O, NULL},
	 { "CBufLine_GetTags", _wrap_CBufLine_GetTags, METH_O, NULL},
	 { "CBufLine_swigregister", CBufLine_swigregister, METH_O, NULL},
	 { "CBufLine_swiginit", CBufLine_swiginit, METH_VARARGS, NULL},
	 { "new_CBuffer", _wrap_new_CBuffer, METH_VARARGS, NULL},
	 { "delete_CBuffer", _wrap_delete_CBuffer, METH_O, NULL},
	 { "CBuffer_AddLine", _wrap_CBuffer_AddLine, METH_VARARGS, NULL},
	 { "CBuffer_UpdateLine", _wrap_CBuffer_UpdateLine, METH_VARARGS, NULL},
	 { "CBuffer_UpdateExactLine", _wrap_CBuffer_UpdateExactLine, METH_VARARGS, NULL},
	 { "CBuffer_GetBufLine", _wrap_CBuffer_GetBufLine, METH_VARARGS, NULL},
	 { "CBuffer_GetLine", _wrap_CBuffer_GetLine, METH_VARARGS, NULL},
	 { "CBuffer_Size", _wrap_CBuffer_Size, METH_O, NULL},
	 { "CBuffer_IsEmpty", _wrap_CBuffer_IsEmpty, METH_O, NULL},
	 { "CBuffer_Clear", _wrap_CBuffer_Clear, METH_O, NULL},
	 { "CBuffer_SetLineCount", _wrap_CBuffer_SetLineCount, METH_VARARGS, NULL},
	 { "CBuffer_GetLineCount", _wrap_CBuffer_GetLineCount, METH_O, NULL},
	 { "CBuffer_swigregister", CBuffer_swigregister, METH_O, NULL},
	 { "CBuffer_swiginit", CBuffer_swiginit, METH_VARARGS, NULL},
	 { "String_s_set", _wrap_String_s_set, METH_VARARGS, NULL},
	 { "String_s_get", _wrap_String_s_get, METH_O, NULL},
	 { "new_String", _wrap_new_String, METH_VARARGS, NULL},
	 { "String___str__", _wrap_String___str__, METH_O, NULL},
	 { "delete_String", _wrap_delete_String, METH_O, NULL},
	 { "String_swigregister", String_swigregister, METH_O, NULL},
	 { "String_swiginit", String_swiginit, METH_VARARGS, NULL},
	 { "new_CPyModule", _wrap_new_CPyModule, METH_VARARGS, NULL},
	 { "CPyModule_GetPyObj", _wrap_CPyModule_GetPyObj, METH_O, NULL},
	 { "CPyModule_GetNewPyObj", _wrap_CPyModule_GetNewPyObj, METH_O, NULL},
	 { "CPyModule_DeletePyModule", _wrap_CPyModule_DeletePyModule, METH_O, NULL},
	 { "CPyModule_GetPyExceptionStr", _wrap_CPyModule_GetPyExceptionStr, METH_O, NULL},
	 { "CPyModule_GetModPython", _wrap_CPyModule_GetModPython, METH_O, NULL},
	 { "CPyModule_OnBoot", _wrap_CPyModule_OnBoot, METH_O, NULL},
	 { "CPyModule_WebRequiresLogin", _wrap_CPyModule_WebRequiresLogin, METH_O, NULL},
	 { "CPyModule_WebRequiresAdmin", _wrap_CPyModule_WebRequiresAdmin, METH_O, NULL},
	 { "CPyModule_GetWebMenuTitle", _wrap_CPyModule_GetWebMenuTitle, METH_O, NULL},
	 { "CPyModule_OnWebPreRequest", _wrap_CPyModule_OnWebPreRequest, METH_VARARGS, NULL},
	 { "CPyModule_OnWebRequest", _wrap_CPyModule_OnWebRequest, METH_VARARGS, NULL},
	 { "CPyModule_ValidateWebRequestCSRFCheck", _wrap_CPyModule_ValidateWebRequestCSRFCheck, METH_VARARGS, NULL},
	 { "CPyModule_GetSubPages", _wrap_CPyModule_GetSubPages, METH_O, NULL},
	 { "CPyModule_OnPreRehash", _wrap_CPyModule_OnPreRehash, METH_O, NULL},
	 { "CPyModule_OnPostRehash", _wrap_CPyModule_OnPostRehash, METH_O, NULL},
	 { "CPyModule_OnIRCDisconnected", _wrap_CPyModule_OnIRCDisconnected, METH_O, NULL},
	 { "CPyModule_OnIRCConnected", _wrap_CPyModule_OnIRCConnected, METH_O, NULL},
	 { "CPyModule_OnIRCConnecting", _wrap_CPyModule_OnIRCConnecting, METH_VARARGS, NULL},
	 { "CPyModule_OnIRCConnectionError", _wrap_CPyModule_OnIRCConnectionError, METH_VARARGS, NULL},
	 { "CPyModule_OnIRCRegistration", _wrap_CPyModule_OnIRCRegistration, METH_VARARGS, NULL},
	 { "CPyModule_OnBroadcast", _wrap_CPyModule_OnBroadcast, METH_VARARGS, NULL},
	 { "CPyModule_OnChanPermission2", _wrap_CPyModule_OnChanPermission2, METH_VARARGS, NULL},
	 { "CPyModule_OnOp2", _wrap_CPyModule_OnOp2, METH_VARARGS, NULL},
	 { "CPyModule_OnDeop2", _wrap_CPyModule_OnDeop2, METH_VARARGS, NULL},
	 { "CPyModule_OnVoice2", _wrap_CPyModule_OnVoice2, METH_VARARGS, NULL},
	 { "CPyModule_OnDevoice2", _wrap_CPyModule_OnDevoice2, METH_VARARGS, NULL},
	 { "CPyModule_OnMode2", _wrap_CPyModule_OnMode2, METH_VARARGS, NULL},
	 { "CPyModule_OnRawMode2", _wrap_CPyModule_OnRawMode2, METH_VARARGS, NULL},
	 { "CPyModule_OnRaw", _wrap_CPyModule_OnRaw, METH_VARARGS, NULL},
	 { "CPyModule_OnStatusCommand", _wrap_CPyModule_OnStatusCommand, METH_VARARGS, NULL},
	 { "CPyModule_OnModCommand", _wrap_CPyModule_OnModCommand, METH_VARARGS, NULL},
	 { "CPyModule_OnModNotice", _wrap_CPyModule_OnModNotice, METH_VARARGS, NULL},
	 { "CPyModule_OnModCTCP", _wrap_CPyModule_OnModCTCP, METH_VARARGS, NULL},
	 { "CPyModule_OnQuit", _wrap_CPyModule_OnQuit, METH_VARARGS, NULL},
	 { "CPyModule_OnNick", _wrap_CPyModule_OnNick, METH_VARARGS, NULL},
	 { "CPyModule_OnKick", _wrap_CPyModule_OnKick, METH_VARARGS, NULL},
	 { "CPyModule_OnJoining", _wrap_CPyModule_OnJoining, METH_VARARGS, NULL},
	 { "CPyModule_OnJoin", _wrap_CPyModule_OnJoin, METH_VARARGS, NULL},
	 { "CPyModule_OnPart", _wrap_CPyModule_OnPart, METH_VARARGS, NULL},
	 { "CPyModule_OnInvite", _wrap_CPyModule_OnInvite, METH_VARARGS, NULL},
	 { "CPyModule_OnChanBufferStarting", _wrap_CPyModule_OnChanBufferStarting, METH_VARARGS, NULL},
	 { "CPyModule_OnChanBufferEnding", _wrap_CPyModule_OnChanBufferEnding, METH_VARARGS, NULL},
	 { "CPyModule_OnChanBufferPlayLine", _wrap_CPyModule_OnChanBufferPlayLine, METH_VARARGS, NULL},
	 { "CPyModule_OnPrivBufferPlayLine", _wrap_CPyModule_OnPrivBufferPlayLine, METH_VARARGS, NULL},
	 { "CPyModule_OnClientLogin", _wrap_CPyModule_OnClientLogin, METH_O, NULL},
	 { "CPyModule_OnClientDisconnect", _wrap_CPyModule_OnClientDisconnect, METH_O, NULL},
	 { "CPyModule_OnUserRaw", _wrap_CPyModule_OnUserRaw, METH_VARARGS, NULL},
	 { "CPyModule_OnUserCTCPReply", _wrap_CPyModule_OnUserCTCPReply, METH_VARARGS, NULL},
	 { "CPyModule_OnUserCTCP", _wrap_CPyModule_OnUserCTCP, METH_VARARGS, NULL},
	 { "CPyModule_OnUserAction", _wrap_CPyModule_OnUserAction, METH_VARARGS, NULL},
	 { "CPyModule_OnUserMsg", _wrap_CPyModule_OnUserMsg, METH_VARARGS, NULL},
	 { "CPyModule_OnUserNotice", _wrap_CPyModule_OnUserNotice, METH_VARARGS, NULL},
	 { "CPyModule_OnUserJoin", _wrap_CPyModule_OnUserJoin, METH_VARARGS, NULL},
	 { "CPyModule_OnUserPart", _wrap_CPyModule_OnUserPart, METH_VARARGS, NULL},
	 { "CPyModule_OnUserTopic", _wrap_CPyModule_OnUserTopic, METH_VARARGS, NULL},
	 { "CPyModule_OnUserTopicRequest", _wrap_CPyModule_OnUserTopicRequest, METH_VARARGS, NULL},
	 { "CPyModule_OnUserQuit", _wrap_CPyModule_OnUserQuit, METH_VARARGS, NULL},
	 { "CPyModule_OnCTCPReply", _wrap_CPyModule_OnCTCPReply, METH_VARARGS, NULL},
	 { "CPyModule_OnPrivCTCP", _wrap_CPyModule_OnPrivCTCP, METH_VARARGS, NULL},
	 { "CPyModule_OnChanCTCP", _wrap_CPyModule_OnChanCTCP, METH_VARARGS, NULL},
	 { "CPyModule_OnPrivAction", _wrap_CPyModule_OnPrivAction, METH_VARARGS, NULL},
	 { "CPyModule_OnChanAction", _wrap_CPyModule_OnChanAction, METH_VARARGS, NULL},
	 { "CPyModule_OnPrivMsg", _wrap_CPyModule_OnPrivMsg, METH_VARARGS, NULL},
	 { "CPyModule_OnChanMsg", _wrap_CPyModule_OnChanMsg, METH_VARARGS, NULL},
	 { "CPyModule_OnPrivNotice", _wrap_CPyModule_OnPrivNotice, METH_VARARGS, NULL},
	 { "CPyModule_OnChanNotice", _wrap_CPyModule_OnChanNotice, METH_VARARGS, NULL},
	 { "CPyModule_OnTopic", _wrap_CPyModule_OnTopic, METH_VARARGS, NULL},
	 { "CPyModule_OnServerCapAvailable", _wrap_CPyModule_OnServerCapAvailable, METH_VARARGS, NULL},
	 { "CPyModule_OnServerCap302Available", _wrap_CPyModule_OnServerCap302Available, METH_VARARGS, NULL},
	 { "CPyModule_OnServerCapResult", _wrap_CPyModule_OnServerCapResult, METH_VARARGS, NULL},
	 { "CPyModule_OnClientAttached", _wrap_CPyModule_OnClientAttached, METH_O, NULL},
	 { "CPyModule_OnClientDetached", _wrap_CPyModule_OnClientDetached, METH_O, NULL},
	 { "CPyModule_OnTimerAutoJoin", _wrap_CPyModule_OnTimerAutoJoin, METH_VARARGS, NULL},
	 { "CPyModule_OnEmbeddedWebRequest", _wrap_CPyModule_OnEmbeddedWebRequest, METH_VARARGS, NULL},
	 { "CPyModule_OnAddNetwork", _wrap_CPyModule_OnAddNetwork, METH_VARARGS, NULL},
	 { "CPyModule_OnDeleteNetwork", _wrap_CPyModule_OnDeleteNetwork, METH_VARARGS, NULL},
	 { "CPyModule_OnSendToClient", _wrap_CPyModule_OnSendToClient, METH_VARARGS, NULL},
	 { "CPyModule_OnSendToIRC", _wrap_CPyModule_OnSendToIRC, METH_VARARGS, NULL},
	 { "CPyModule_OnRawMessage", _wrap_CPyModule_OnRawMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnNumericMessage", _wrap_CPyModule_OnNumericMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnQuitMessage", _wrap_CPyModule_OnQuitMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnNickMessage", _wrap_CPyModule_OnNickMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnKickMessage", _wrap_CPyModule_OnKickMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnJoinMessage", _wrap_CPyModule_OnJoinMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnPartMessage", _wrap_CPyModule_OnPartMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnChanBufferPlayMessage", _wrap_CPyModule_OnChanBufferPlayMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnPrivBufferPlayMessage", _wrap_CPyModule_OnPrivBufferPlayMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnUserRawMessage", _wrap_CPyModule_OnUserRawMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnUserCTCPReplyMessage", _wrap_CPyModule_OnUserCTCPReplyMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnUserCTCPMessage", _wrap_CPyModule_OnUserCTCPMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnUserActionMessage", _wrap_CPyModule_OnUserActionMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnUserTextMessage", _wrap_CPyModule_OnUserTextMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnUserNoticeMessage", _wrap_CPyModule_OnUserNoticeMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnUserJoinMessage", _wrap_CPyModule_OnUserJoinMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnUserPartMessage", _wrap_CPyModule_OnUserPartMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnUserTopicMessage", _wrap_CPyModule_OnUserTopicMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnUserQuitMessage", _wrap_CPyModule_OnUserQuitMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnCTCPReplyMessage", _wrap_CPyModule_OnCTCPReplyMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnPrivCTCPMessage", _wrap_CPyModule_OnPrivCTCPMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnChanCTCPMessage", _wrap_CPyModule_OnChanCTCPMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnPrivActionMessage", _wrap_CPyModule_OnPrivActionMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnChanActionMessage", _wrap_CPyModule_OnChanActionMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnPrivTextMessage", _wrap_CPyModule_OnPrivTextMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnChanTextMessage", _wrap_CPyModule_OnChanTextMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnPrivNoticeMessage", _wrap_CPyModule_OnPrivNoticeMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnChanNoticeMessage", _wrap_CPyModule_OnChanNoticeMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnTopicMessage", _wrap_CPyModule_OnTopicMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnSendToClientMessage", _wrap_CPyModule_OnSendToClientMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnSendToIRCMessage", _wrap_CPyModule_OnSendToIRCMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnUserTagMessage", _wrap_CPyModule_OnUserTagMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnChanTagMessage", _wrap_CPyModule_OnChanTagMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnPrivTagMessage", _wrap_CPyModule_OnPrivTagMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnInviteMessage", _wrap_CPyModule_OnInviteMessage, METH_VARARGS, NULL},
	 { "CPyModule_OnAddUser", _wrap_CPyModule_OnAddUser, METH_VARARGS, NULL},
	 { "CPyModule_OnDeleteUser", _wrap_CPyModule_OnDeleteUser, METH_VARARGS, NULL},
	 { "CPyModule_OnClientConnect", _wrap_CPyModule_OnClientConnect, METH_VARARGS, NULL},
	 { "CPyModule_OnFailedLogin", _wrap_CPyModule_OnFailedLogin, METH_VARARGS, NULL},
	 { "CPyModule_OnUnknownUserRaw", _wrap_CPyModule_OnUnknownUserRaw, METH_VARARGS, NULL},
	 { "CPyModule_OnUnknownUserRawMessage", _wrap_CPyModule_OnUnknownUserRawMessage, METH_VARARGS, NULL},
	 { "CPyModule_IsClientCapSupported", _wrap_CPyModule_IsClientCapSupported, METH_VARARGS, NULL},
	 { "CPyModule_OnClientCapRequest", _wrap_CPyModule_OnClientCapRequest, METH_VARARGS, NULL},
	 { "CPyModule_OnClientGetSASLMechanisms", _wrap_CPyModule_OnClientGetSASLMechanisms, METH_VARARGS, NULL},
	 { "CPyModule_OnClientSASLServerInitialChallenge", _wrap_CPyModule_OnClientSASLServerInitialChallenge, METH_VARARGS, NULL},
	 { "CPyModule_OnClientSASLAuthenticate", _wrap_CPyModule_OnClientSASLAuthenticate, METH_VARARGS, NULL},
	 { "CPyModule_OnClientSASLAborted", _wrap_CPyModule_OnClientSASLAborted, METH_O, NULL},
	 { "CPyModule_OnModuleLoading", _wrap_CPyModule_OnModuleLoading, METH_VARARGS, NULL},
	 { "CPyModule_OnModuleUnloading", _wrap_CPyModule_OnModuleUnloading, METH_VARARGS, NULL},
	 { "CPyModule_OnGetModInfo", _wrap_CPyModule_OnGetModInfo, METH_VARARGS, NULL},
	 { "CPyModule_OnGetAvailableMods", _wrap_CPyModule_OnGetAvailableMods, METH_VARARGS, NULL},
	 { "CPyModule_OnClientCapLs", _wrap_CPyModule_OnClientCapLs, METH_VARARGS, NULL},
	 { "CPyModule_OnLoginAttempt", _wrap_CPyModule_OnLoginAttempt, METH_VARARGS, NULL},
	 { "delete_CPyModule", _wrap_delete_CPyModule, METH_O, NULL},
	 { "CPyModule_swigregister", CPyModule_swigregister, METH_O, NULL},
	 { "CPyModule_swiginit", CPyModule_swiginit, METH_VARARGS, NULL},
	 { "AsPyModule", _wrap_AsPyModule, METH_O, NULL},
	 { "CreatePyModule", _wrap_CreatePyModule, METH_VARARGS, NULL},
	 { "new_CPyTimer", _wrap_new_CPyTimer, METH_VARARGS, NULL},
	 { "CPyTimer_RunJob", _wrap_CPyTimer_RunJob, METH_O, NULL},
	 { "CPyTimer_GetPyObj", _wrap_CPyTimer_GetPyObj, METH_O, NULL},
	 { "CPyTimer_GetNewPyObj", _wrap_CPyTimer_GetNewPyObj, METH_O, NULL},
	 { "delete_CPyTimer", _wrap_delete_CPyTimer, METH_O, NULL},
	 { "CPyTimer_swigregister", CPyTimer_swigregister, METH_O, NULL},
	 { "CPyTimer_swiginit", CPyTimer_swiginit, METH_VARARGS, NULL},
	 { "CreatePyTimer", _wrap_CreatePyTimer, METH_VARARGS, NULL},
	 { "new_CPySocket", _wrap_new_CPySocket, METH_VARARGS, NULL},
	 { "CPySocket_GetPyObj", _wrap_CPySocket_GetPyObj, METH_O, NULL},
	 { "CPySocket_GetNewPyObj", _wrap_CPySocket_GetNewPyObj, METH_O, NULL},
	 { "delete_CPySocket", _wrap_delete_CPySocket, METH_O, NULL},
	 { "CPySocket_Connected", _wrap_CPySocket_Connected, METH_O, NULL},
	 { "CPySocket_Disconnected", _wrap_CPySocket_Disconnected, METH_O, NULL},
	 { "CPySocket_Timeout", _wrap_CPySocket_Timeout, METH_O, NULL},
	 { "CPySocket_ConnectionRefused", _wrap_CPySocket_ConnectionRefused, METH_O, NULL},
	 { "CPySocket_ReadData", _wrap_CPySocket_ReadData, METH_VARARGS, NULL},
	 { "CPySocket_ReadLine", _wrap_CPySocket_ReadLine, METH_VARARGS, NULL},
	 { "CPySocket_GetSockObj", _wrap_CPySocket_GetSockObj, METH_VARARGS, NULL},
	 { "CPySocket_swigregister", CPySocket_swigregister, METH_O, NULL},
	 { "CPySocket_swiginit", CPySocket_swiginit, METH_VARARGS, NULL},
	 { "CreatePySocket", _wrap_CreatePySocket, METH_VARARGS, NULL},
	 { "HaveIPv6_", _wrap_HaveIPv6_, METH_NOARGS, NULL},
	 { "HaveSSL_", _wrap_HaveSSL_, METH_NOARGS, NULL},
	 { "HaveCharset_", _wrap_HaveCharset_, METH_NOARGS, NULL},
	 { "GetSOMAXCONN", _wrap_GetSOMAXCONN, METH_NOARGS, NULL},
	 { "GetVersionMajor", _wrap_GetVersionMajor, METH_NOARGS, NULL},
	 { "GetVersionMinor", _wrap_GetVersionMinor, METH_NOARGS, NULL},
	 { "GetVersion", _wrap_GetVersion, METH_NOARGS, NULL},
	 { "GetVersionExtra", _wrap_GetVersionExtra, METH_NOARGS, NULL},
	 { "MCString_iter_x_set", _wrap_MCString_iter_x_set, METH_VARARGS, NULL},
	 { "MCString_iter_x_get", _wrap_MCString_iter_x_get, METH_O, NULL},
	 { "new_MCString_iter", _wrap_new_MCString_iter, METH_VARARGS, NULL},
	 { "MCString_iter_plusplus", _wrap_MCString_iter_plusplus, METH_O, NULL},
	 { "MCString_iter_get", _wrap_MCString_iter_get, METH_O, NULL},
	 { "MCString_iter_is_end", _wrap_MCString_iter_is_end, METH_VARARGS, NULL},
	 { "delete_MCString_iter", _wrap_delete_MCString_iter, METH_O, NULL},
	 { "MCString_iter_swigregister", MCString_iter_swigregister, METH_O, NULL},
	 { "MCString_iter_swiginit", MCString_iter_swiginit, METH_VARARGS, NULL},
	 { "new_CModulesIter", _wrap_new_CModulesIter, METH_O, NULL},
	 { "CModulesIter_plusplus", _wrap_CModulesIter_plusplus, METH_O, NULL},
	 { "CModulesIter_get", _wrap_CModulesIter_get, METH_O, NULL},
	 { "CModulesIter_is_end", _wrap_CModulesIter_is_end, METH_O, NULL},
	 { "CModulesIter_m_pModules_set", _wrap_CModulesIter_m_pModules_set, METH_VARARGS, NULL},
	 { "CModulesIter_m_pModules_get", _wrap_CModulesIter_m_pModules_get, METH_O, NULL},
	 { "CModulesIter_m_it_set", _wrap_CModulesIter_m_it_set, METH_VARARGS, NULL},
	 { "CModulesIter_m_it_get", _wrap_CModulesIter_m_it_get, METH_O, NULL},
	 { "delete_CModulesIter", _wrap_delete_CModulesIter, METH_O, NULL},
	 { "CModulesIter_swigregister", CModulesIter_swigregister, METH_O, NULL},
	 { "CModulesIter_swiginit", CModulesIter_swiginit, METH_VARARGS, NULL},
	 { "new_CPyModCommand", _wrap_new_CPyModCommand, METH_VARARGS, NULL},
	 { "delete_CPyModCommand", _wrap_delete_CPyModCommand, METH_O, NULL},
	 { "CPyModCommand_GetModule", _wrap_CPyModCommand_GetModule, METH_O, NULL},
	 { "CPyModCommand_swigregister", CPyModCommand_swigregister, METH_O, NULL},
	 { "CPyModCommand_swiginit", CPyModCommand_swiginit, METH_VARARGS, NULL},
	 { "CreatePyModCommand", _wrap_CreatePyModCommand, METH_VARARGS, NULL},
	 { "new_CPyCapability", _wrap_new_CPyCapability, METH_VARARGS, NULL},
	 { "delete_CPyCapability", _wrap_delete_CPyCapability, METH_O, NULL},
	 { "CPyCapability_OnServerChangedSupport", _wrap_CPyCapability_OnServerChangedSupport, METH_VARARGS, NULL},
	 { "CPyCapability_OnClientChangedSupport", _wrap_CPyCapability_OnClientChangedSupport, METH_VARARGS, NULL},
	 { "CPyCapability_swigregister", CPyCapability_swigregister, METH_O, NULL},
	 { "CPyCapability_swiginit", CPyCapability_swiginit, METH_VARARGS, NULL},
	 { "CPyRetString_s_set", _wrap_CPyRetString_s_set, METH_VARARGS, NULL},
	 { "CPyRetString_s_get", _wrap_CPyRetString_s_get, METH_O, NULL},
	 { "CPyRetString___str__", _wrap_CPyRetString___str__, METH_O, NULL},
	 { "delete_CPyRetString", _wrap_delete_CPyRetString, METH_O, NULL},
	 { "CPyRetString_swigregister", CPyRetString_swigregister, METH_O, NULL},
	 { "CPyRetBool_b_set", _wrap_CPyRetBool_b_set, METH_VARARGS, NULL},
	 { "CPyRetBool_b_get", _wrap_CPyRetBool_b_get, METH_O, NULL},
	 { "CPyRetBool___bool__", _wrap_CPyRetBool___bool__, METH_O, NULL},
	 { "delete_CPyRetBool", _wrap_delete_CPyRetBool, METH_O, NULL},
	 { "CPyRetBool_swigregister", CPyRetBool_swigregister, METH_O, NULL},
	 { "CModPython_swigregister", CModPython_swigregister, METH_O, NULL},
	 { "new_StrPair", _wrap_new_StrPair, METH_VARARGS, NULL},
	 { "StrPair_first_set", _wrap_StrPair_first_set, METH_VARARGS, NULL},
	 { "StrPair_first_get", _wrap_StrPair_first_get, METH_O, NULL},
	 { "StrPair_second_set", _wrap_StrPair_second_set, METH_VARARGS, NULL},
	 { "StrPair_second_get", _wrap_StrPair_second_get, METH_O, NULL},
	 { "delete_StrPair", _wrap_delete_StrPair, METH_O, NULL},
	 { "StrPair_swigregister", StrPair_swigregister, METH_O, NULL},
	 { "StrPair_swiginit", StrPair_swiginit, METH_VARARGS, NULL},
	 { "VPair_iterator", _wrap_VPair_iterator, METH_O, NULL},
	 { "VPair___nonzero__", _wrap_VPair___nonzero__, METH_O, NULL},
	 { "VPair___bool__", _wrap_VPair___bool__, METH_O, NULL},
	 { "VPair___len__", _wrap_VPair___len__, METH_O, NULL},
	 { "VPair___getslice__", _wrap_VPair___getslice__, METH_VARARGS, NULL},
	 { "VPair___setslice__", _wrap_VPair___setslice__, METH_VARARGS, NULL},
	 { "VPair___delslice__", _wrap_VPair___delslice__, METH_VARARGS, NULL},
	 { "VPair___delitem__", _wrap_VPair___delitem__, METH_VARARGS, NULL},
	 { "VPair___getitem__", _wrap_VPair___getitem__, METH_VARARGS, NULL},
	 { "VPair___setitem__", _wrap_VPair___setitem__, METH_VARARGS, NULL},
	 { "VPair_pop", _wrap_VPair_pop, METH_O, NULL},
	 { "VPair_append", _wrap_VPair_append, METH_VARARGS, NULL},
	 { "VPair_empty", _wrap_VPair_empty, METH_O, NULL},
	 { "VPair_size", _wrap_VPair_size, METH_O, NULL},
	 { "VPair_swap", _wrap_VPair_swap, METH_VARARGS, NULL},
	 { "VPair_begin", _wrap_VPair_begin, METH_O, NULL},
	 { "VPair_end", _wrap_VPair_end, METH_O, NULL},
	 { "VPair_rbegin", _wrap_VPair_rbegin, METH_O, NULL},
	 { "VPair_rend", _wrap_VPair_rend, METH_O, NULL},
	 { "VPair_clear", _wrap_VPair_clear, METH_O, NULL},
	 { "VPair_get_allocator", _wrap_VPair_get_allocator, METH_O, NULL},
	 { "VPair_pop_back", _wrap_VPair_pop_back, METH_O, NULL},
	 { "VPair_erase", _wrap_VPair_erase, METH_VARARGS, NULL},
	 { "new_VPair", _wrap_new_VPair, METH_VARARGS, NULL},
	 { "VPair_push_back", _wrap_VPair_push_back, METH_VARARGS, NULL},
	 { "VPair_front", _wrap_VPair_front, METH_O, NULL},
	 { "VPair_back", _wrap_VPair_back, METH_O, NULL},
	 { "VPair_assign", _wrap_VPair_assign, METH_VARARGS, NULL},
	 { "VPair_resize", _wrap_VPair_resize, METH_VARARGS, NULL},
	 { "VPair_insert", _wrap_VPair_insert, METH_VARARGS, NULL},
	 { "VPair_reserve", _wrap_VPair_reserve, METH_VARARGS, NULL},
	 { "VPair_capacity", _wrap_VPair_capacity, METH_O, NULL},
	 { "delete_VPair", _wrap_delete_VPair, METH_O, NULL},
	 { "VPair_swigregister", VPair_swigregister, METH_O, NULL},
	 { "VPair_swiginit", VPair_swiginit, METH_VARARGS, NULL},
	 { "VWebSubPages_iterator", _wrap_VWebSubPages_iterator, METH_O, NULL},
	 { "VWebSubPages___nonzero__", _wrap_VWebSubPages___nonzero__, METH_O, NULL},
	 { "VWebSubPages___bool__", _wrap_VWebSubPages___bool__, METH_O, NULL},
	 { "VWebSubPages___len__", _wrap_VWebSubPages___len__, METH_O, NULL},
	 { "VWebSubPages___getslice__", _wrap_VWebSubPages___getslice__, METH_VARARGS, NULL},
	 { "VWebSubPages___setslice__", _wrap_VWebSubPages___setslice__, METH_VARARGS, NULL},
	 { "VWebSubPages___delslice__", _wrap_VWebSubPages___delslice__, METH_VARARGS, NULL},
	 { "VWebSubPages___delitem__", _wrap_VWebSubPages___delitem__, METH_VARARGS, NULL},
	 { "VWebSubPages___getitem__", _wrap_VWebSubPages___getitem__, METH_VARARGS, NULL},
	 { "VWebSubPages___setitem__", _wrap_VWebSubPages___setitem__, METH_VARARGS, NULL},
	 { "VWebSubPages_pop", _wrap_VWebSubPages_pop, METH_O, NULL},
	 { "VWebSubPages_append", _wrap_VWebSubPages_append, METH_VARARGS, NULL},
	 { "VWebSubPages_empty", _wrap_VWebSubPages_empty, METH_O, NULL},
	 { "VWebSubPages_size", _wrap_VWebSubPages_size, METH_O, NULL},
	 { "VWebSubPages_swap", _wrap_VWebSubPages_swap, METH_VARARGS, NULL},
	 { "VWebSubPages_begin", _wrap_VWebSubPages_begin, METH_O, NULL},
	 { "VWebSubPages_end", _wrap_VWebSubPages_end, METH_O, NULL},
	 { "VWebSubPages_rbegin", _wrap_VWebSubPages_rbegin, METH_O, NULL},
	 { "VWebSubPages_rend", _wrap_VWebSubPages_rend, METH_O, NULL},
	 { "VWebSubPages_clear", _wrap_VWebSubPages_clear, METH_O, NULL},
	 { "VWebSubPages_get_allocator", _wrap_VWebSubPages_get_allocator, METH_O, NULL},
	 { "VWebSubPages_pop_back", _wrap_VWebSubPages_pop_back, METH_O, NULL},
	 { "VWebSubPages_erase", _wrap_VWebSubPages_erase, METH_VARARGS, NULL},
	 { "new_VWebSubPages", _wrap_new_VWebSubPages, METH_VARARGS, NULL},
	 { "VWebSubPages_push_back", _wrap_VWebSubPages_push_back, METH_VARARGS, NULL},
	 { "VWebSubPages_front", _wrap_VWebSubPages_front, METH_O, NULL},
	 { "VWebSubPages_back", _wrap_VWebSubPages_back, METH_O, NULL},
	 { "VWebSubPages_assign", _wrap_VWebSubPages_assign, METH_VARARGS, NULL},
	 { "VWebSubPages_resize", _wrap_VWebSubPages_resize, METH_VARARGS, NULL},
	 { "VWebSubPages_insert", _wrap_VWebSubPages_insert, METH_VARARGS, NULL},
	 { "VWebSubPages_reserve", _wrap_VWebSubPages_reserve, METH_VARARGS, NULL},
	 { "VWebSubPages_capacity", _wrap_VWebSubPages_capacity, METH_O, NULL},
	 { "delete_VWebSubPages", _wrap_delete_VWebSubPages, METH_O, NULL},
	 { "VWebSubPages_swigregister", VWebSubPages_swigregister, METH_O, NULL},
	 { "VWebSubPages_swiginit", VWebSubPages_swiginit, METH_VARARGS, NULL},
	 { "VPair_Add2Str_", _wrap_VPair_Add2Str_, METH_VARARGS, NULL},
	 { "CreateWebSubPage_", _wrap_CreateWebSubPage_, METH_VARARGS, NULL},
	 { NULL, NULL, 0, NULL }
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static void *_p_CClientAuthTo_p_CAuthBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CAuthBase *)  ((CClientAuth *) x));
}
static void *_p_CPyCapabilityTo_p_CCapability(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CCapability *)  ((CPyCapability *) x));
}
static void *_p_CAuthBaseTo_p_CCoreTranslationMixin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CCoreTranslationMixin *)  ((CAuthBase *) x));
}
static void *_p_CChanTo_p_CCoreTranslationMixin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CCoreTranslationMixin *)  ((CChan *) x));
}
static void *_p_CIRCNetworkTo_p_CCoreTranslationMixin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CCoreTranslationMixin *)  ((CIRCNetwork *) x));
}
static void *_p_CModCommandTo_p_CCoreTranslationMixin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CCoreTranslationMixin *)  ((CModCommand *) x));
}
static void *_p_CModulesTo_p_CCoreTranslationMixin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CCoreTranslationMixin *)  ((CModules *) x));
}
static void *_p_CSockManagerTo_p_CCoreTranslationMixin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CCoreTranslationMixin *)  ((CSockManager *) x));
}
static void *_p_CUserTo_p_CCoreTranslationMixin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CCoreTranslationMixin *)  ((CUser *) x));
}
static void *_p_CZNCTo_p_CCoreTranslationMixin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CCoreTranslationMixin *)  ((CZNC *) x));
}
static void *_p_CZNCSockTo_p_CCoreTranslationMixin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CCoreTranslationMixin *)  ((CZNCSock *) x));
}
static void *_p_CFPTimerTo_p_CCron(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CCron *) (CTimer *) ((CFPTimer *) x));
}
static void *_p_CPyTimerTo_p_CCron(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CCron *) (CTimer *) ((CPyTimer *) x));
}
static void *_p_CTimerTo_p_CCron(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CCron *)  ((CTimer *) x));
}
static void *_p_CWebSockTo_p_CHTTPSock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CHTTPSock *)  ((CWebSock *) x));
}
static void *_p_CClientTo_p_CIRCSocket(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CIRCSocket *)  ((CClient *) x));
}
static void *_p_CIRCSockTo_p_CIRCSocket(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CIRCSocket *)  ((CIRCSock *) x));
}
static void *_p_CTCPListenerTo_p_CListener(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CListener *)  ((CTCPListener *) x));
}
static void *_p_CUnixListenerTo_p_CListener(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CListener *)  ((CUnixListener *) x));
}
static void *_p_CActionMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *) (CTargetMessage *) ((CActionMessage *) x));
}
static void *_p_CAuthenticateMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *)  ((CAuthenticateMessage *) x));
}
static void *_p_CCTCPMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *) (CTargetMessage *) ((CCTCPMessage *) x));
}
static void *_p_CChgHostMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *)  ((CChgHostMessage *) x));
}
static void *_p_CInviteMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *)  ((CInviteMessage *) x));
}
static void *_p_CJoinMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *) (CTargetMessage *) ((CJoinMessage *) x));
}
static void *_p_CKickMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *) (CTargetMessage *) ((CKickMessage *) x));
}
static void *_p_CModeMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *) (CTargetMessage *) ((CModeMessage *) x));
}
static void *_p_CNickMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *)  ((CNickMessage *) x));
}
static void *_p_CNoticeMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *) (CTargetMessage *) ((CNoticeMessage *) x));
}
static void *_p_CNumericMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *)  ((CNumericMessage *) x));
}
static void *_p_CPartMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *) (CTargetMessage *) ((CPartMessage *) x));
}
static void *_p_CQuitMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *)  ((CQuitMessage *) x));
}
static void *_p_CTargetMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *)  ((CTargetMessage *) x));
}
static void *_p_CTextMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *) (CTargetMessage *) ((CTextMessage *) x));
}
static void *_p_CTopicMessageTo_p_CMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CMessage *) (CTargetMessage *) ((CTopicMessage *) x));
}
static void *_p_CPyModCommandTo_p_CModCommand(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CModCommand *)  ((CPyModCommand *) x));
}
static void *_p_CModPythonTo_p_CModule(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CModule *)  ((CModPython *) x));
}
static void *_p_CPyModuleTo_p_CModule(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CModule *)  ((CPyModule *) x));
}
static void *_p_CSSSLConnectionTo_p_CSConnection(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSConnection *)  ((CSSSLConnection *) x));
}
static void *_p_CClientTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *) (Csock *)(CZNCSock *)(CIRCSocket *) ((CClient *) x));
}
static void *_p_CExecSockTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *) (Csock *)(CZNCSock *) ((CExecSock *) x));
}
static void *_p_CHTTPSockTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *) (Csock *)(CZNCSock *)(CSocket *) ((CHTTPSock *) x));
}
static void *_p_CIRCSockTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *) (Csock *)(CZNCSock *)(CIRCSocket *) ((CIRCSock *) x));
}
static void *_p_CIRCSocketTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *) (Csock *)(CZNCSock *) ((CIRCSocket *) x));
}
static void *_p_CIncomingConnectionTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *) (Csock *)(CZNCSock *) ((CIncomingConnection *) x));
}
static void *_p_CPySocketTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *) (Csock *)(CZNCSock *)(CSocket *) ((CPySocket *) x));
}
static void *_p_CRealListenerTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *) (Csock *)(CZNCSock *) ((CRealListener *) x));
}
static void *_p_CSockManagerTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *) (CSocketManager *)(TSocketManager< CZNCSock > *) ((CSockManager *) x));
}
static void *_p_CSocketTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *) (Csock *)(CZNCSock *) ((CSocket *) x));
}
static void *_p_CSocketManagerTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *)  ((CSocketManager *) x));
}
static void *_p_CWebSockTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *) (Csock *)(CZNCSock *)(CSocket *)(CHTTPSock *) ((CWebSock *) x));
}
static void *_p_CZNCSockTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *) (Csock *) ((CZNCSock *) x));
}
static void *_p_CsockTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *)  ((Csock *) x));
}
static void *_p_TSocketManagerT_CZNCSock_tTo_p_CSockCommon(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSockCommon *) (CSocketManager *) ((TSocketManager< CZNCSock > *) x));
}
static void *_p_CHTTPSockTo_p_CSocket(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSocket *)  ((CHTTPSock *) x));
}
static void *_p_CPySocketTo_p_CSocket(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSocket *)  ((CPySocket *) x));
}
static void *_p_CWebSockTo_p_CSocket(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSocket *) (CHTTPSock *) ((CWebSock *) x));
}
static void *_p_CSockManagerTo_p_CSocketManager(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSocketManager *) (TSocketManager< CZNCSock > *) ((CSockManager *) x));
}
static void *_p_TSocketManagerT_CZNCSock_tTo_p_CSocketManager(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CSocketManager *)  ((TSocketManager< CZNCSock > *) x));
}
static void *_p_CActionMessageTo_p_CTargetMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CTargetMessage *)  ((CActionMessage *) x));
}
static void *_p_CCTCPMessageTo_p_CTargetMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CTargetMessage *)  ((CCTCPMessage *) x));
}
static void *_p_CJoinMessageTo_p_CTargetMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CTargetMessage *)  ((CJoinMessage *) x));
}
static void *_p_CKickMessageTo_p_CTargetMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CTargetMessage *)  ((CKickMessage *) x));
}
static void *_p_CModeMessageTo_p_CTargetMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CTargetMessage *)  ((CModeMessage *) x));
}
static void *_p_CNoticeMessageTo_p_CTargetMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CTargetMessage *)  ((CNoticeMessage *) x));
}
static void *_p_CPartMessageTo_p_CTargetMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CTargetMessage *)  ((CPartMessage *) x));
}
static void *_p_CTextMessageTo_p_CTargetMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CTargetMessage *)  ((CTextMessage *) x));
}
static void *_p_CTopicMessageTo_p_CTargetMessage(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CTargetMessage *)  ((CTopicMessage *) x));
}
static void *_p_CZNCTagHandlerTo_p_CTemplateTagHandler(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CTemplateTagHandler *)  ((CZNCTagHandler *) x));
}
static void *_p_CFPTimerTo_p_CTimer(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CTimer *)  ((CFPTimer *) x));
}
static void *_p_CPyTimerTo_p_CTimer(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CTimer *)  ((CPyTimer *) x));
}
static void *_p_CClientTo_p_CZNCSock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CZNCSock *) (CIRCSocket *) ((CClient *) x));
}
static void *_p_CExecSockTo_p_CZNCSock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CZNCSock *)  ((CExecSock *) x));
}
static void *_p_CHTTPSockTo_p_CZNCSock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CZNCSock *) (CSocket *) ((CHTTPSock *) x));
}
static void *_p_CIRCSockTo_p_CZNCSock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CZNCSock *) (CIRCSocket *) ((CIRCSock *) x));
}
static void *_p_CIRCSocketTo_p_CZNCSock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CZNCSock *)  ((CIRCSocket *) x));
}
static void *_p_CIncomingConnectionTo_p_CZNCSock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CZNCSock *)  ((CIncomingConnection *) x));
}
static void *_p_CPySocketTo_p_CZNCSock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CZNCSock *) (CSocket *) ((CPySocket *) x));
}
static void *_p_CRealListenerTo_p_CZNCSock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CZNCSock *)  ((CRealListener *) x));
}
static void *_p_CSocketTo_p_CZNCSock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CZNCSock *)  ((CSocket *) x));
}
static void *_p_CWebSockTo_p_CZNCSock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CZNCSock *) (CSocket *)(CHTTPSock *) ((CWebSock *) x));
}
static void *_p_CClientTo_p_Csock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((Csock *) (CZNCSock *)(CIRCSocket *) ((CClient *) x));
}
static void *_p_CExecSockTo_p_Csock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((Csock *) (CZNCSock *) ((CExecSock *) x));
}
static void *_p_CHTTPSockTo_p_Csock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((Csock *) (CZNCSock *)(CSocket *) ((CHTTPSock *) x));
}
static void *_p_CIRCSockTo_p_Csock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((Csock *) (CZNCSock *)(CIRCSocket *) ((CIRCSock *) x));
}
static void *_p_CIRCSocketTo_p_Csock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((Csock *) (CZNCSock *) ((CIRCSocket *) x));
}
static void *_p_CIncomingConnectionTo_p_Csock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((Csock *) (CZNCSock *) ((CIncomingConnection *) x));
}
static void *_p_CPySocketTo_p_Csock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((Csock *) (CZNCSock *)(CSocket *) ((CPySocket *) x));
}
static void *_p_CRealListenerTo_p_Csock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((Csock *) (CZNCSock *) ((CRealListener *) x));
}
static void *_p_CSocketTo_p_Csock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((Csock *) (CZNCSock *) ((CSocket *) x));
}
static void *_p_CWebSockTo_p_Csock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((Csock *) (CZNCSock *)(CSocket *)(CHTTPSock *) ((CWebSock *) x));
}
static void *_p_CZNCSockTo_p_Csock(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((Csock *)  ((CZNCSock *) x));
}
static void *_p_CTemplateTo_p_MCString(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((MCString *)  ((CTemplate *) x));
}
static void *_p_CSockManagerTo_p_TSocketManagerT_CZNCSock_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((TSocketManager< CZNCSock > *)  ((CSockManager *) x));
}
static void *_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModPythonTo_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModule(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CModule *(*)(ModHandle,CUser *,CIRCNetwork *,CString const &,CString const &,CModInfo::EModuleType))  ((CModPython *(*)(ModHandle,CUser *,CIRCNetwork *,CString const &,CString const &,CModInfo::EModuleType)) x));
}
static void *_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CPyModuleTo_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModule(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((CModule *(*)(ModHandle,CUser *,CIRCNetwork *,CString const &,CString const &,CModInfo::EModuleType))  ((CPyModule *(*)(ModHandle,CUser *,CIRCNetwork *,CString const &,CString const &,CModInfo::EModuleType)) x));
}
static void *_p_CBufferTo_p_std__dequeT_CBufLine_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((std::deque< CBufLine,std::allocator< CBufLine > > *)  ((CBuffer *) x));
}
static void *_p_CTemplateTo_p_std__mapT_CString_CString_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > > *) (MCString *) ((CTemplate *) x));
}
static void *_p_MCStringTo_p_std__mapT_CString_CString_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > > *)  ((MCString *) x));
}
static void *_p_std__shared_ptrT_CClientAuth_tTo_p_std__shared_ptrT_CAuthBase_t(void *x, int *newmemory) {
    *newmemory = SWIG_CAST_NEW_MEMORY;
    return (void *) new std::shared_ptr< CAuthBase >(*(std::shared_ptr< CClientAuth > *)x);
}
static void *_p_CModulesTo_p_std__vectorT_CModule_p_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((std::vector< CModule *,std::allocator< CModule * > > *)  ((CModules *) x));
}
static void *_p_CSocketManagerTo_p_std__vectorT_Csock_p_std__allocatorT_Csock_p_t_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((std::vector< Csock *,std::allocator< Csock * > > *)  ((CSocketManager *) x));
}
static void *_p_CTableTo_p_std__vectorT_VCString_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *)  ((CTable *) x));
}
static swig_type_info _swigt__m_CModule__f_r_q_const__CString__void = {"_m_CModule__f_r_q_const__CString__void", "CModCommand::ModCmdFunc|void (CModule::*)(CString const &)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CActionMessage = {"_p_CActionMessage", "CActionMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CAuthBase = {"_p_CAuthBase", "CAuthBase *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CAuthenticateMessage = {"_p_CAuthenticateMessage", "CAuthenticateMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CBufLine = {"_p_CBufLine", "std::deque< CBufLine >::value_type *|CBufLine *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CBuffer = {"_p_CBuffer", "CBuffer *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CCTCPMessage = {"_p_CCTCPMessage", "CCTCPMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CCapability = {"_p_CCapability", "CCapability *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CChan = {"_p_CChan", "std::vector< CChan * >::value_type|CChan *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CChgHostMessage = {"_p_CChgHostMessage", "CChgHostMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CClient = {"_p_CClient", "std::vector< CClient * >::value_type|CClient *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CClientAuth = {"_p_CClientAuth", "CClientAuth *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CConfig = {"_p_CConfig", "CConfig *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CConfigEntry = {"_p_CConfigEntry", "CConfigEntry *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CConnectQueueTimer = {"_p_CConnectQueueTimer", "CConnectQueueTimer *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CCoreTranslationMixin = {"_p_CCoreTranslationMixin", "CCoreTranslationMixin *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CCron = {"_p_CCron", "CCron *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CDebug = {"_p_CDebug", "CDebug *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CDebugStream = {"_p_CDebugStream", "CDebugStream *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CDelayedTranslation = {"_p_CDelayedTranslation", "CDelayedTranslation *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CDir = {"_p_CDir", "CDir *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CException = {"_p_CException", "CException *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CExecSock = {"_p_CExecSock", "CExecSock *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CFPTimer = {"_p_CFPTimer", "CFPTimer *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CFile = {"_p_CFile", "CFile *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CGetAddrInfo = {"_p_CGetAddrInfo", "CGetAddrInfo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CHTTPSock = {"_p_CHTTPSock", "CHTTPSock *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CIRCNetwork = {"_p_CIRCNetwork", "std::vector< CIRCNetwork * >::value_type|CIRCNetwork *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CIRCSock = {"_p_CIRCSock", "CIRCSock *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CIRCSocket = {"_p_CIRCSocket", "CIRCSocket *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CIncomingConnection = {"_p_CIncomingConnection", "CIncomingConnection *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CInviteMessage = {"_p_CInviteMessage", "CInviteMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CJoinMessage = {"_p_CJoinMessage", "CJoinMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CKickMessage = {"_p_CKickMessage", "CKickMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CLanguageScope = {"_p_CLanguageScope", "CLanguageScope *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CListener = {"_p_CListener", "std::vector< CListener * >::value_type|CListener *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CMessage = {"_p_CMessage", "CMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CModCommand = {"_p_CModCommand", "CModCommand *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CModInfo = {"_p_CModInfo", "std::set< CModInfo >::key_type *|std::set< CModInfo >::value_type *|CModInfo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CModPython = {"_p_CModPython", "CModPython *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CModeMessage = {"_p_CModeMessage", "CModeMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CModule = {"_p_CModule", "std::vector< CModule * >::value_type|CModule *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CModuleEntry = {"_p_CModuleEntry", "CModuleEntry *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CModules = {"_p_CModules", "CModules *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CModulesIter = {"_p_CModulesIter", "CModulesIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CNick = {"_p_CNick", "std::map< CString,CNick >::mapped_type *|CNick *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CNickMessage = {"_p_CNickMessage", "CNickMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CNoticeMessage = {"_p_CNoticeMessage", "CNoticeMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CNumericMessage = {"_p_CNumericMessage", "CNumericMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_COptionalTranslation = {"_p_COptionalTranslation", "COptionalTranslation *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CPartMessage = {"_p_CPartMessage", "CPartMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CPyCapability = {"_p_CPyCapability", "CPyCapability *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CPyModCommand = {"_p_CPyModCommand", "CPyModCommand *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CPyModule = {"_p_CPyModule", "CPyModule *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CPyRetBool = {"_p_CPyRetBool", "CPyRetBool *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CPyRetString = {"_p_CPyRetString", "CPyRetString *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CPySocket = {"_p_CPySocket", "CPySocket *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CPyTimer = {"_p_CPyTimer", "CPyTimer *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CQuery = {"_p_CQuery", "std::vector< CQuery * >::value_type|CQuery *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CQuitMessage = {"_p_CQuitMessage", "CQuitMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CRealListener = {"_p_CRealListener", "CRealListener *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CSCharBuffer = {"_p_CSCharBuffer", "CSCharBuffer *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CSConnection = {"_p_CSConnection", "CSConnection *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CSListener = {"_p_CSListener", "CSListener *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CSMonitorFD = {"_p_CSMonitorFD", "CSMonitorFD *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CSSSLConnection = {"_p_CSSSLConnection", "CSSSLConnection *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CSSockAddr = {"_p_CSSockAddr", "CSSockAddr *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CServer = {"_p_CServer", "std::vector< CServer * >::value_type|CServer *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CSockCommon = {"_p_CSockCommon", "CSockCommon *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CSockManager = {"_p_CSockManager", "CSockManager *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CSocket = {"_p_CSocket", "CSocket *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CSocketManager = {"_p_CSocketManager", "CSocketManager *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CString__EEscape = {"_p_CString__EEscape", "CString::EEscape *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CString__size_type = {"_p_CString__size_type", "CString::size_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CTCPListener = {"_p_CTCPListener", "CTCPListener *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CTable = {"_p_CTable", "CTable *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CTargetMessage = {"_p_CTargetMessage", "CTargetMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CTemplate = {"_p_CTemplate", "CTemplate *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CTemplateLoopContext = {"_p_CTemplateLoopContext", "CTemplateLoopContext *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CTemplateOptions = {"_p_CTemplateOptions", "CTemplateOptions *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CTemplateTagHandler = {"_p_CTemplateTagHandler", "CTemplateTagHandler *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CTextMessage = {"_p_CTextMessage", "CTextMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CTimer = {"_p_CTimer", "CTimer *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CTopicMessage = {"_p_CTopicMessage", "CTopicMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CTranslation = {"_p_CTranslation", "CTranslation *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CTranslationDomainRefHolder = {"_p_CTranslationDomainRefHolder", "CTranslationDomainRefHolder *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CTranslationInfo = {"_p_CTranslationInfo", "CTranslationInfo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CUnixListener = {"_p_CUnixListener", "CUnixListener *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CUser = {"_p_CUser", "CUser *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CUtils = {"_p_CUtils", "CUtils *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CWebSession = {"_p_CWebSession", "CWebSession *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CWebSessionMap = {"_p_CWebSessionMap", "CWebSessionMap *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CWebSock = {"_p_CWebSock", "CWebSock *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CWebSubPage = {"_p_CWebSubPage", "CWebSubPage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CZNC = {"_p_CZNC", "CZNC *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CZNCSock = {"_p_CZNCSock", "CZNCSock *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CZNCTagHandler = {"_p_CZNCTagHandler", "CZNCTagHandler *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_CmdFunc = {"_p_CmdFunc", "CmdFunc *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Csock = {"_p_Csock", "Csock *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_EAcceptType = {"_p_EAcceptType", "EAcceptType *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_EChanModeArgs = {"_p_EChanModeArgs", "EChanModeArgs *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_EModException = {"_p_EModException", "EModException *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_EModRet = {"_p_EModRet", "EModRet *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_EModes = {"_p_EModes", "EModes *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_EModuleType = {"_p_EModuleType", "EModuleType *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_EType = {"_p_EType", "EType *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_EUserPerms = {"_p_EUserPerms", "EUserPerms *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_EntryMap = {"_p_EntryMap", "EntryMap *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_EntryMapIterator = {"_p_EntryMapIterator", "EntryMapIterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_MCString = {"_p_MCString", "MCString *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_MCString_iter = {"_p_MCString_iter", "MCString_iter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ModCmdFunc = {"_p_ModCmdFunc", "ModCmdFunc *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ModDirList = {"_p_ModDirList", "ModDirList *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_String = {"_p_String", "String *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_SubConfig = {"_p_SubConfig", "SubConfig *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_SubConfigEntry = {"_p_SubConfigEntry", "SubConfigEntry *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_SubConfigMap = {"_p_SubConfigMap", "SubConfigMap *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_SubConfigMapIterator = {"_p_SubConfigMapIterator", "SubConfigMapIterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TSocketManagerT_CZNCSock_t = {"_p_TSocketManagerT_CZNCSock_t", "TSocketManager< CZNCSock > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TrafficStatsMap = {"_p_TrafficStatsMap", "TrafficStatsMap *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TrafficStatsPair = {"_p_TrafficStatsPair", "TrafficStatsPair *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_allocator_type = {"_p_allocator_type", "allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_bool = {"_p_bool", "bool *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_difference_type = {"_p_difference_type", "difference_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_double = {"_p_double", "double *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModule = {"_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModule", "CModInfo::ModLoader|CModule *(*)(ModHandle,CUser *,CIRCNetwork *,CString const &,CString const &,CModInfo::EModuleType)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModPython = {"_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModPython", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CPyModule = {"_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CPyModule", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_f_p_CModule_p_CFPTimer__void = {"_p_f_p_CModule_p_CFPTimer__void", "FPTimer_t|void (*)(CModule *,CFPTimer *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_r_CModInfo__void = {"_p_f_r_CModInfo__void", "void (*)(CModInfo &)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_r_std__ostream__r_std__ostream = {"_p_f_r_std__ostream__r_std__ostream", "std::ostream &(*)(std::ostream &)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_fd_set = {"_p_fd_set", "fd_set *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_first_type = {"_p_first_type", "first_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_gid_t = {"_p_gid_t", "gid_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_in_addr = {"_p_in_addr", "in_addr *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "cs_sock_t *|int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int32_t = {"_p_int32_t", "int32_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int64_t = {"_p_int64_t", "int64_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_key_type = {"_p_key_type", "key_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_long = {"_p_long", "long *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_mapped_type = {"_p_mapped_type", "mapped_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_mode_t = {"_p_mode_t", "mode_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_PyObject = {"_p_p_PyObject", "PyObject **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_second_type = {"_p_second_type", "second_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_size_type = {"_p_size_type", "size_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_sockaddr_in = {"_p_sockaddr_in", "sockaddr_in *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_sockaddr_storage = {"_p_sockaddr_storage", "sockaddr_storage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_socklen_t = {"_p_socklen_t", "socklen_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ssize_t = {"_p_ssize_t", "cs_ssize_t *|ssize_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_stat = {"_p_stat", "stat *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_CBufLine_t = {"_p_std__allocatorT_CBufLine_t", "std::deque< CBufLine >::allocator_type *|std::allocator< CBufLine > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_CChan_p_t = {"_p_std__allocatorT_CChan_p_t", "std::vector< CChan * >::allocator_type *|std::allocator< CChan * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_CClient_p_t = {"_p_std__allocatorT_CClient_p_t", "std::vector< CClient * >::allocator_type *|std::allocator< CClient * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_CIRCNetwork_p_t = {"_p_std__allocatorT_CIRCNetwork_p_t", "std::vector< CIRCNetwork * >::allocator_type *|std::allocator< CIRCNetwork * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_CListener_p_t = {"_p_std__allocatorT_CListener_p_t", "std::vector< CListener * >::allocator_type *|std::allocator< CListener * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_CModule_p_t = {"_p_std__allocatorT_CModule_p_t", "std::vector< CModule * >::allocator_type *|std::allocator< CModule * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_CQuery_p_t = {"_p_std__allocatorT_CQuery_p_t", "std::vector< CQuery * >::allocator_type *|std::allocator< CQuery * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_CServer_p_t = {"_p_std__allocatorT_CServer_p_t", "std::vector< CServer * >::allocator_type *|std::allocator< CServer * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_CString_t = {"_p_std__allocatorT_CString_t", "std::list< CString >::allocator_type *|std::vector< CString >::allocator_type *|std::allocator< CString > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_std__pairT_CString_CString_t_t = {"_p_std__allocatorT_std__pairT_CString_CString_t_t", "std::vector< std::pair< CString,CString > >::allocator_type *|std::allocator< std::pair< CString,CString > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_std__pairT_CString_const_CNick_t_t = {"_p_std__allocatorT_std__pairT_CString_const_CNick_t_t", "std::map< CString,CNick >::allocator_type *|std::allocator< std::pair< CString const,CNick > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_std__pairT_CString_const_CString_t_t = {"_p_std__allocatorT_std__pairT_CString_const_CString_t_t", "std::map< CString,CString >::allocator_type *|std::allocator< std::pair< CString const,CString > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_std__pairT_CString_const_std__vectorT_CString_t_t_t = {"_p_std__allocatorT_std__pairT_CString_const_std__vectorT_CString_t_t_t", "std::map< CString,std::vector< CString,std::allocator< CString > > >::allocator_type *|std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_std__shared_ptrT_CWebSubPage_t_t = {"_p_std__allocatorT_std__shared_ptrT_CWebSubPage_t_t", "std::vector< std::shared_ptr< CWebSubPage > >::allocator_type *|std::allocator< std::shared_ptr< CWebSubPage > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_std__vectorT_CString_t_t = {"_p_std__allocatorT_std__vectorT_CString_t_t", "std::vector< std::vector< CString,std::allocator< CString > > >::allocator_type *|std::allocator< std::vector< CString,std::allocator< CString > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__dequeT_CBufLine_t = {"_p_std__dequeT_CBufLine_t", "std::deque< CBufLine,std::allocator< CBufLine > > *|std::deque< CBufLine > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__functionT_void_fCString_const_RF_t = {"_p_std__functionT_void_fCString_const_RF_t", "CModCommand::CmdFunc *|std::function< void (CString const &) > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__invalid_argument = {"_p_std__invalid_argument", "std::invalid_argument *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__lessT_CModInfo_t = {"_p_std__lessT_CModInfo_t", "std::less< CModInfo > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__lessT_CString_t = {"_p_std__lessT_CString_t", "std::less< CString > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__listT_CIRCNetwork_p_std__allocatorT_CIRCNetwork_p_t_t = {"_p_std__listT_CIRCNetwork_p_std__allocatorT_CIRCNetwork_p_t_t", "std::list< CIRCNetwork *,std::allocator< CIRCNetwork * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__listT_CString_t = {"_p_std__listT_CString_t", "std::list< CString,std::allocator< CString > > *|std::list< CString > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_CString_CNick_t = {"_p_std__mapT_CString_CNick_t", "std::map< CString,CNick,std::less< CString >,std::allocator< std::pair< CString const,CNick > > > *|std::map< CString,CNick > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_CString_CString_t = {"_p_std__mapT_CString_CString_t", "std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > > *|std::map< CString,CString > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_CString_CString_t__iterator = {"_p_std__mapT_CString_CString_t__iterator", "MCString::iterator *|std::map< CString,CString,std::less< CString >,std::allocator< std::pair< CString const,CString > > >::iterator *|std::map< CString,CString >::iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_CString_CTranslationInfo_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CTranslationInfo_t_t_t = {"_p_std__mapT_CString_CTranslationInfo_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CTranslationInfo_t_t_t", "std::map< CString,CTranslationInfo,std::less< CString >,std::allocator< std::pair< CString const,CTranslationInfo > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_CString_CUser_p_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CUser_p_t_t_t = {"_p_std__mapT_CString_CUser_p_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CUser_p_t_t_t", "std::map< CString,CUser *,std::less< CString >,std::allocator< std::pair< CString const,CUser * > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_CString_VCString_t = {"_p_std__mapT_CString_VCString_t", "std::map< CString,VCString > *|std::map< CString,VCString,std::less< CString >,std::allocator< std::pair< CString const,VCString > > > *|std::map< CString,std::vector< CString,std::allocator< CString > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< CString,std::allocator< CString > > > > > *|std::map< CString,std::vector< CString,std::allocator< CString > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_CString_std__pairT_unsigned_long_long_unsigned_long_long_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__pairT_unsigned_long_long_unsigned_long_long_t_t_t_t = {"_p_std__mapT_CString_std__pairT_unsigned_long_long_unsigned_long_long_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__pairT_unsigned_long_long_unsigned_long_long_t_t_t_t", "CZNC::TrafficStatsMap *|std::map< CString,std::pair< unsigned long long,unsigned long long >,std::less< CString >,std::allocator< std::pair< CString const,std::pair< unsigned long long,unsigned long long > > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_CString_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_t_t_t__const_iterator = {"_p_std__mapT_CString_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_t_t_t__const_iterator", "CConfig::SubConfigMapIterator *|std::map< CString,std::vector< std::pair< CString,CConfigEntry >,std::allocator< std::pair< CString,CConfigEntry > > >,std::less< CString >,std::allocator< std::pair< CString const,std::vector< std::pair< CString,CConfigEntry >,std::allocator< std::pair< CString,CConfigEntry > > > > > >::const_iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_char_CIRCSock__EChanModeArgs_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CIRCSock__EChanModeArgs_t_t_t = {"_p_std__mapT_char_CIRCSock__EChanModeArgs_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CIRCSock__EChanModeArgs_t_t_t", "std::map< char,CIRCSock::EChanModeArgs,std::less< char >,std::allocator< std::pair< char const,CIRCSock::EChanModeArgs > > > *|std::map< char,enum CIRCSock::EChanModeArgs,std::less< char >,std::allocator< std::pair< char const,enum CIRCSock::EChanModeArgs > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_char_CString_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CString_t_t_t = {"_p_std__mapT_char_CString_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CString_t_t_t", "std::map< char,CString,std::less< char >,std::allocator< std::pair< char const,CString > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_char_unsigned_int_std__lessT_char_t_std__allocatorT_std__pairT_char_const_unsigned_int_t_t_t = {"_p_std__mapT_char_unsigned_int_std__lessT_char_t_std__allocatorT_std__pairT_char_const_unsigned_int_t_t_t", "std::map< char,unsigned int,std::less< char >,std::allocator< std::pair< char const,unsigned int > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t = {"_p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t", "std::map< cs_sock_t,short,std::less< cs_sock_t >,std::allocator< std::pair< cs_sock_t const,short > > > *|std::map< int,short,std::less< int >,std::allocator< std::pair< int const,short > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__ostream = {"_p_std__ostream", "std::ostream *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__pairT_CString_CString_t = {"_p_std__pairT_CString_CString_t", "std::pair< CString,CString > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__pairT_unsigned_long_long_unsigned_long_long_t = {"_p_std__pairT_unsigned_long_long_unsigned_long_long_t", "CZNC::TrafficStatsPair *|std::pair< unsigned long long,unsigned long long > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__setT_CChan_p_std__lessT_CChan_p_t_std__allocatorT_CChan_p_t_t = {"_p_std__setT_CChan_p_std__lessT_CChan_p_t_std__allocatorT_CChan_p_t_t", "std::set< CChan *,std::less< CChan * >,std::allocator< CChan * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__setT_CModInfo_t = {"_p_std__setT_CModInfo_t", "std::set< CModInfo,std::less< CModInfo >,std::allocator< CModInfo > > *|std::set< CModInfo > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__setT_CSocket_p_std__lessT_CSocket_p_t_std__allocatorT_CSocket_p_t_t__const_iterator = {"_p_std__setT_CSocket_p_std__lessT_CSocket_p_t_std__allocatorT_CSocket_p_t_t__const_iterator", "std::set< CSocket *,std::less< CSocket * >,std::allocator< CSocket * > >::const_iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__setT_CString_t = {"_p_std__setT_CString_t", "SCString *|std::set< CString,std::less< CString >,std::allocator< CString > > *|std::set< CString > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__setT_CTimer_p_std__lessT_CTimer_p_t_std__allocatorT_CTimer_p_t_t__const_iterator = {"_p_std__setT_CTimer_p_std__lessT_CTimer_p_t_std__allocatorT_CTimer_p_t_t__const_iterator", "std::set< CTimer *,std::less< CTimer * >,std::allocator< CTimer * > >::const_iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__setT_char_std__lessT_char_t_std__allocatorT_char_t_t = {"_p_std__setT_char_std__lessT_char_t_std__allocatorT_char_t_t", "std::set< char,std::less< char >,std::allocator< char > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__shared_ptrT_CAuthBase_t = {"_p_std__shared_ptrT_CAuthBase_t", "std::shared_ptr< CAuthBase > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__shared_ptrT_CClientAuth_t = {"_p_std__shared_ptrT_CClientAuth_t", "std::shared_ptr< CClientAuth > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__shared_ptrT_CTemplateOptions_t = {"_p_std__shared_ptrT_CTemplateOptions_t", "std::shared_ptr< CTemplateOptions > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__shared_ptrT_CTemplateTagHandler_t = {"_p_std__shared_ptrT_CTemplateTagHandler_t", "std::shared_ptr< CTemplateTagHandler > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__shared_ptrT_CWebSession_t = {"_p_std__shared_ptrT_CWebSession_t", "std::shared_ptr< CWebSession > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__shared_ptrT_CWebSubPage_t = {"_p_std__shared_ptrT_CWebSubPage_t", "TWebSubPage *|std::vector< std::shared_ptr< CWebSubPage > >::value_type *|std::shared_ptr< CWebSubPage > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_CChan_p_t = {"_p_std__vectorT_CChan_p_t", "std::vector< CChan *,std::allocator< CChan * > > *|std::vector< CChan * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_CClient_p_t = {"_p_std__vectorT_CClient_p_t", "std::vector< CClient *,std::allocator< CClient * > > *|std::vector< CClient * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_CCron_p_std__allocatorT_CCron_p_t_t = {"_p_std__vectorT_CCron_p_std__allocatorT_CCron_p_t_t", "std::vector< CCron *,std::allocator< CCron * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_CIRCNetwork_p_t = {"_p_std__vectorT_CIRCNetwork_p_t", "std::vector< CIRCNetwork *,std::allocator< CIRCNetwork * > > *|std::vector< CIRCNetwork * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_CListener_p_t = {"_p_std__vectorT_CListener_p_t", "std::vector< CListener *,std::allocator< CListener * > > *|std::vector< CListener * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_CModule_p_t = {"_p_std__vectorT_CModule_p_t", "std::vector< CModule *,std::allocator< CModule * > > *|std::vector< CModule * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_CModule_p_t__const_iterator = {"_p_std__vectorT_CModule_p_t__const_iterator", "CModules::const_iterator *|std::vector< CModule *,std::allocator< CModule * > >::const_iterator *|std::vector< CModule * >::const_iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_CQuery_p_t = {"_p_std__vectorT_CQuery_p_t", "std::vector< CQuery *,std::allocator< CQuery * > > *|std::vector< CQuery * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_CServer_p_t = {"_p_std__vectorT_CServer_p_t", "std::vector< CServer *,std::allocator< CServer * > > *|std::vector< CServer * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_CString_t = {"_p_std__vectorT_CString_t", "VCString *|std::vector< CString,std::allocator< CString > > *|std::vector< CString > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_CTemplate_p_std__allocatorT_CTemplate_p_t_t = {"_p_std__vectorT_CTemplate_p_std__allocatorT_CTemplate_p_t_t", "std::vector< CTemplate *,std::allocator< CTemplate * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_Csock_p_std__allocatorT_Csock_p_t_t = {"_p_std__vectorT_Csock_p_std__allocatorT_Csock_p_t_t", "std::vector< Csock *,std::allocator< Csock * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_TWebSubPage_t = {"_p_std__vectorT_TWebSubPage_t", "VWebSubPages *|std::vector< TWebSubPage > *|std::vector< std::shared_ptr< CWebSubPage >,std::allocator< std::shared_ptr< CWebSubPage > > > *|std::vector< std::shared_ptr< CWebSubPage > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_VCString_t = {"_p_std__vectorT_VCString_t", "std::vector< VCString > *|std::vector< std::vector< CString,std::allocator< CString > >,std::allocator< std::vector< CString,std::allocator< CString > > > > *|std::vector< std::vector< CString,std::allocator< CString > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t = {"_p_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t", "CConfig::SubConfig *|std::vector< std::pair< CString,CConfigEntry >,std::allocator< std::pair< CString,CConfigEntry > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_std__pairT_CString_CString_t_t = {"_p_std__vectorT_std__pairT_CString_CString_t_t", "VPair *|std::vector< std::pair< CString,CString >,std::allocator< std::pair< CString,CString > > > *|std::vector< std::pair< CString,CString > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_std__shared_ptrT_CTemplateTagHandler_t_std__allocatorT_std__shared_ptrT_CTemplateTagHandler_t_t_t = {"_p_std__vectorT_std__shared_ptrT_CTemplateTagHandler_t_std__allocatorT_std__shared_ptrT_CTemplateTagHandler_t_t_t", "std::vector< std::shared_ptr< CTemplateTagHandler >,std::allocator< std::shared_ptr< CTemplateTagHandler > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_swig__SwigPyIterator = {"_p_swig__SwigPyIterator", "swig::SwigPyIterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_time_t = {"_p_time_t", "time_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_timeval = {"_p_timeval", "timeval *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_uid_t = {"_p_uid_t", "uid_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_uint16_t = {"_p_uint16_t", "uint16_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_uint64_t = {"_p_uint64_t", "uint64_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_int = {"_p_unsigned_int", "unsigned int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_short = {"_p_unsigned_short", "unsigned short *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_value_type = {"_p_value_type", "value_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_void = {"_p_void", "ModHandle|void *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__m_CModule__f_r_q_const__CString__void,
  &_swigt__p_CActionMessage,
  &_swigt__p_CAuthBase,
  &_swigt__p_CAuthenticateMessage,
  &_swigt__p_CBufLine,
  &_swigt__p_CBuffer,
  &_swigt__p_CCTCPMessage,
  &_swigt__p_CCapability,
  &_swigt__p_CChan,
  &_swigt__p_CChgHostMessage,
  &_swigt__p_CClient,
  &_swigt__p_CClientAuth,
  &_swigt__p_CConfig,
  &_swigt__p_CConfigEntry,
  &_swigt__p_CConnectQueueTimer,
  &_swigt__p_CCoreTranslationMixin,
  &_swigt__p_CCron,
  &_swigt__p_CDebug,
  &_swigt__p_CDebugStream,
  &_swigt__p_CDelayedTranslation,
  &_swigt__p_CDir,
  &_swigt__p_CException,
  &_swigt__p_CExecSock,
  &_swigt__p_CFPTimer,
  &_swigt__p_CFile,
  &_swigt__p_CGetAddrInfo,
  &_swigt__p_CHTTPSock,
  &_swigt__p_CIRCNetwork,
  &_swigt__p_CIRCSock,
  &_swigt__p_CIRCSocket,
  &_swigt__p_CIncomingConnection,
  &_swigt__p_CInviteMessage,
  &_swigt__p_CJoinMessage,
  &_swigt__p_CKickMessage,
  &_swigt__p_CLanguageScope,
  &_swigt__p_CListener,
  &_swigt__p_CMessage,
  &_swigt__p_CModCommand,
  &_swigt__p_CModInfo,
  &_swigt__p_CModPython,
  &_swigt__p_CModeMessage,
  &_swigt__p_CModule,
  &_swigt__p_CModuleEntry,
  &_swigt__p_CModules,
  &_swigt__p_CModulesIter,
  &_swigt__p_CNick,
  &_swigt__p_CNickMessage,
  &_swigt__p_CNoticeMessage,
  &_swigt__p_CNumericMessage,
  &_swigt__p_COptionalTranslation,
  &_swigt__p_CPartMessage,
  &_swigt__p_CPyCapability,
  &_swigt__p_CPyModCommand,
  &_swigt__p_CPyModule,
  &_swigt__p_CPyRetBool,
  &_swigt__p_CPyRetString,
  &_swigt__p_CPySocket,
  &_swigt__p_CPyTimer,
  &_swigt__p_CQuery,
  &_swigt__p_CQuitMessage,
  &_swigt__p_CRealListener,
  &_swigt__p_CSCharBuffer,
  &_swigt__p_CSConnection,
  &_swigt__p_CSListener,
  &_swigt__p_CSMonitorFD,
  &_swigt__p_CSSSLConnection,
  &_swigt__p_CSSockAddr,
  &_swigt__p_CServer,
  &_swigt__p_CSockCommon,
  &_swigt__p_CSockManager,
  &_swigt__p_CSocket,
  &_swigt__p_CSocketManager,
  &_swigt__p_CString__EEscape,
  &_swigt__p_CString__size_type,
  &_swigt__p_CTCPListener,
  &_swigt__p_CTable,
  &_swigt__p_CTargetMessage,
  &_swigt__p_CTemplate,
  &_swigt__p_CTemplateLoopContext,
  &_swigt__p_CTemplateOptions,
  &_swigt__p_CTemplateTagHandler,
  &_swigt__p_CTextMessage,
  &_swigt__p_CTimer,
  &_swigt__p_CTopicMessage,
  &_swigt__p_CTranslation,
  &_swigt__p_CTranslationDomainRefHolder,
  &_swigt__p_CTranslationInfo,
  &_swigt__p_CUnixListener,
  &_swigt__p_CUser,
  &_swigt__p_CUtils,
  &_swigt__p_CWebSession,
  &_swigt__p_CWebSessionMap,
  &_swigt__p_CWebSock,
  &_swigt__p_CWebSubPage,
  &_swigt__p_CZNC,
  &_swigt__p_CZNCSock,
  &_swigt__p_CZNCTagHandler,
  &_swigt__p_CmdFunc,
  &_swigt__p_Csock,
  &_swigt__p_EAcceptType,
  &_swigt__p_EChanModeArgs,
  &_swigt__p_EModException,
  &_swigt__p_EModRet,
  &_swigt__p_EModes,
  &_swigt__p_EModuleType,
  &_swigt__p_EType,
  &_swigt__p_EUserPerms,
  &_swigt__p_EntryMap,
  &_swigt__p_EntryMapIterator,
  &_swigt__p_MCString,
  &_swigt__p_MCString_iter,
  &_swigt__p_ModCmdFunc,
  &_swigt__p_ModDirList,
  &_swigt__p_String,
  &_swigt__p_SubConfig,
  &_swigt__p_SubConfigEntry,
  &_swigt__p_SubConfigMap,
  &_swigt__p_SubConfigMapIterator,
  &_swigt__p_TSocketManagerT_CZNCSock_t,
  &_swigt__p_TrafficStatsMap,
  &_swigt__p_TrafficStatsPair,
  &_swigt__p_allocator_type,
  &_swigt__p_bool,
  &_swigt__p_char,
  &_swigt__p_difference_type,
  &_swigt__p_double,
  &_swigt__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModPython,
  &_swigt__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModule,
  &_swigt__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CPyModule,
  &_swigt__p_f_p_CModule_p_CFPTimer__void,
  &_swigt__p_f_r_CModInfo__void,
  &_swigt__p_f_r_std__ostream__r_std__ostream,
  &_swigt__p_fd_set,
  &_swigt__p_first_type,
  &_swigt__p_gid_t,
  &_swigt__p_in_addr,
  &_swigt__p_int,
  &_swigt__p_int32_t,
  &_swigt__p_int64_t,
  &_swigt__p_key_type,
  &_swigt__p_long,
  &_swigt__p_mapped_type,
  &_swigt__p_mode_t,
  &_swigt__p_p_PyObject,
  &_swigt__p_second_type,
  &_swigt__p_size_type,
  &_swigt__p_sockaddr_in,
  &_swigt__p_sockaddr_storage,
  &_swigt__p_socklen_t,
  &_swigt__p_ssize_t,
  &_swigt__p_stat,
  &_swigt__p_std__allocatorT_CBufLine_t,
  &_swigt__p_std__allocatorT_CChan_p_t,
  &_swigt__p_std__allocatorT_CClient_p_t,
  &_swigt__p_std__allocatorT_CIRCNetwork_p_t,
  &_swigt__p_std__allocatorT_CListener_p_t,
  &_swigt__p_std__allocatorT_CModule_p_t,
  &_swigt__p_std__allocatorT_CQuery_p_t,
  &_swigt__p_std__allocatorT_CServer_p_t,
  &_swigt__p_std__allocatorT_CString_t,
  &_swigt__p_std__allocatorT_std__pairT_CString_CString_t_t,
  &_swigt__p_std__allocatorT_std__pairT_CString_const_CNick_t_t,
  &_swigt__p_std__allocatorT_std__pairT_CString_const_CString_t_t,
  &_swigt__p_std__allocatorT_std__pairT_CString_const_std__vectorT_CString_t_t_t,
  &_swigt__p_std__allocatorT_std__shared_ptrT_CWebSubPage_t_t,
  &_swigt__p_std__allocatorT_std__vectorT_CString_t_t,
  &_swigt__p_std__dequeT_CBufLine_t,
  &_swigt__p_std__functionT_void_fCString_const_RF_t,
  &_swigt__p_std__invalid_argument,
  &_swigt__p_std__lessT_CModInfo_t,
  &_swigt__p_std__lessT_CString_t,
  &_swigt__p_std__listT_CIRCNetwork_p_std__allocatorT_CIRCNetwork_p_t_t,
  &_swigt__p_std__listT_CString_t,
  &_swigt__p_std__mapT_CString_CNick_t,
  &_swigt__p_std__mapT_CString_CString_t,
  &_swigt__p_std__mapT_CString_CString_t__iterator,
  &_swigt__p_std__mapT_CString_CTranslationInfo_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CTranslationInfo_t_t_t,
  &_swigt__p_std__mapT_CString_CUser_p_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CUser_p_t_t_t,
  &_swigt__p_std__mapT_CString_VCString_t,
  &_swigt__p_std__mapT_CString_std__pairT_unsigned_long_long_unsigned_long_long_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__pairT_unsigned_long_long_unsigned_long_long_t_t_t_t,
  &_swigt__p_std__mapT_CString_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_t_t_t__const_iterator,
  &_swigt__p_std__mapT_char_CIRCSock__EChanModeArgs_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CIRCSock__EChanModeArgs_t_t_t,
  &_swigt__p_std__mapT_char_CString_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CString_t_t_t,
  &_swigt__p_std__mapT_char_unsigned_int_std__lessT_char_t_std__allocatorT_std__pairT_char_const_unsigned_int_t_t_t,
  &_swigt__p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t,
  &_swigt__p_std__ostream,
  &_swigt__p_std__pairT_CString_CString_t,
  &_swigt__p_std__pairT_unsigned_long_long_unsigned_long_long_t,
  &_swigt__p_std__setT_CChan_p_std__lessT_CChan_p_t_std__allocatorT_CChan_p_t_t,
  &_swigt__p_std__setT_CModInfo_t,
  &_swigt__p_std__setT_CSocket_p_std__lessT_CSocket_p_t_std__allocatorT_CSocket_p_t_t__const_iterator,
  &_swigt__p_std__setT_CString_t,
  &_swigt__p_std__setT_CTimer_p_std__lessT_CTimer_p_t_std__allocatorT_CTimer_p_t_t__const_iterator,
  &_swigt__p_std__setT_char_std__lessT_char_t_std__allocatorT_char_t_t,
  &_swigt__p_std__shared_ptrT_CAuthBase_t,
  &_swigt__p_std__shared_ptrT_CClientAuth_t,
  &_swigt__p_std__shared_ptrT_CTemplateOptions_t,
  &_swigt__p_std__shared_ptrT_CTemplateTagHandler_t,
  &_swigt__p_std__shared_ptrT_CWebSession_t,
  &_swigt__p_std__shared_ptrT_CWebSubPage_t,
  &_swigt__p_std__vectorT_CChan_p_t,
  &_swigt__p_std__vectorT_CClient_p_t,
  &_swigt__p_std__vectorT_CCron_p_std__allocatorT_CCron_p_t_t,
  &_swigt__p_std__vectorT_CIRCNetwork_p_t,
  &_swigt__p_std__vectorT_CListener_p_t,
  &_swigt__p_std__vectorT_CModule_p_t,
  &_swigt__p_std__vectorT_CModule_p_t__const_iterator,
  &_swigt__p_std__vectorT_CQuery_p_t,
  &_swigt__p_std__vectorT_CServer_p_t,
  &_swigt__p_std__vectorT_CString_t,
  &_swigt__p_std__vectorT_CTemplate_p_std__allocatorT_CTemplate_p_t_t,
  &_swigt__p_std__vectorT_Csock_p_std__allocatorT_Csock_p_t_t,
  &_swigt__p_std__vectorT_TWebSubPage_t,
  &_swigt__p_std__vectorT_VCString_t,
  &_swigt__p_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t,
  &_swigt__p_std__vectorT_std__pairT_CString_CString_t_t,
  &_swigt__p_std__vectorT_std__shared_ptrT_CTemplateTagHandler_t_std__allocatorT_std__shared_ptrT_CTemplateTagHandler_t_t_t,
  &_swigt__p_swig__SwigPyIterator,
  &_swigt__p_time_t,
  &_swigt__p_timeval,
  &_swigt__p_uid_t,
  &_swigt__p_uint16_t,
  &_swigt__p_uint64_t,
  &_swigt__p_unsigned_int,
  &_swigt__p_unsigned_short,
  &_swigt__p_value_type,
  &_swigt__p_void,
};

static swig_cast_info _swigc__m_CModule__f_r_q_const__CString__void[] = {  {&_swigt__m_CModule__f_r_q_const__CString__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CActionMessage[] = {  {&_swigt__p_CActionMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CAuthBase[] = {  {&_swigt__p_CAuthBase, 0, 0, 0},  {&_swigt__p_CClientAuth, _p_CClientAuthTo_p_CAuthBase, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CAuthenticateMessage[] = {  {&_swigt__p_CAuthenticateMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CBufLine[] = {  {&_swigt__p_CBufLine, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CBuffer[] = {  {&_swigt__p_CBuffer, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CCTCPMessage[] = {  {&_swigt__p_CCTCPMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CCapability[] = {  {&_swigt__p_CCapability, 0, 0, 0},  {&_swigt__p_CPyCapability, _p_CPyCapabilityTo_p_CCapability, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CChan[] = {  {&_swigt__p_CChan, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CChgHostMessage[] = {  {&_swigt__p_CChgHostMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CClient[] = {  {&_swigt__p_CClient, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CClientAuth[] = {  {&_swigt__p_CClientAuth, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CConfig[] = {  {&_swigt__p_CConfig, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CConfigEntry[] = {  {&_swigt__p_CConfigEntry, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CConnectQueueTimer[] = {  {&_swigt__p_CConnectQueueTimer, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CCoreTranslationMixin[] = {  {&_swigt__p_CClient, 0, 0, 0},  {&_swigt__p_CClientAuth, 0, 0, 0},  {&_swigt__p_CCoreTranslationMixin, 0, 0, 0},  {&_swigt__p_CExecSock, 0, 0, 0},  {&_swigt__p_CHTTPSock, 0, 0, 0},  {&_swigt__p_CIRCSock, 0, 0, 0},  {&_swigt__p_CIRCSocket, 0, 0, 0},  {&_swigt__p_CIncomingConnection, 0, 0, 0},  {&_swigt__p_CPyModCommand, 0, 0, 0},  {&_swigt__p_CPySocket, 0, 0, 0},  {&_swigt__p_CRealListener, 0, 0, 0},  {&_swigt__p_CSocket, 0, 0, 0},  {&_swigt__p_CWebSock, 0, 0, 0},  {&_swigt__p_CAuthBase, _p_CAuthBaseTo_p_CCoreTranslationMixin, 0, 0},  {&_swigt__p_CChan, _p_CChanTo_p_CCoreTranslationMixin, 0, 0},  {&_swigt__p_CIRCNetwork, _p_CIRCNetworkTo_p_CCoreTranslationMixin, 0, 0},  {&_swigt__p_CModCommand, _p_CModCommandTo_p_CCoreTranslationMixin, 0, 0},  {&_swigt__p_CModules, _p_CModulesTo_p_CCoreTranslationMixin, 0, 0},  {&_swigt__p_CSockManager, _p_CSockManagerTo_p_CCoreTranslationMixin, 0, 0},  {&_swigt__p_CUser, _p_CUserTo_p_CCoreTranslationMixin, 0, 0},  {&_swigt__p_CZNC, _p_CZNCTo_p_CCoreTranslationMixin, 0, 0},  {&_swigt__p_CZNCSock, _p_CZNCSockTo_p_CCoreTranslationMixin, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CCron[] = {  {&_swigt__p_CCron, 0, 0, 0},  {&_swigt__p_CFPTimer, _p_CFPTimerTo_p_CCron, 0, 0},  {&_swigt__p_CPyTimer, _p_CPyTimerTo_p_CCron, 0, 0},  {&_swigt__p_CTimer, _p_CTimerTo_p_CCron, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CDebug[] = {  {&_swigt__p_CDebug, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CDebugStream[] = {  {&_swigt__p_CDebugStream, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CDelayedTranslation[] = {  {&_swigt__p_CDelayedTranslation, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CDir[] = {  {&_swigt__p_CDir, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CException[] = {  {&_swigt__p_CException, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CExecSock[] = {  {&_swigt__p_CExecSock, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CFPTimer[] = {  {&_swigt__p_CFPTimer, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CFile[] = {  {&_swigt__p_CFile, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CGetAddrInfo[] = {  {&_swigt__p_CGetAddrInfo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CHTTPSock[] = {  {&_swigt__p_CHTTPSock, 0, 0, 0},  {&_swigt__p_CWebSock, _p_CWebSockTo_p_CHTTPSock, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CIRCNetwork[] = {  {&_swigt__p_CIRCNetwork, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CIRCSock[] = {  {&_swigt__p_CIRCSock, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CIRCSocket[] = {  {&_swigt__p_CIRCSocket, 0, 0, 0},  {&_swigt__p_CClient, _p_CClientTo_p_CIRCSocket, 0, 0},  {&_swigt__p_CIRCSock, _p_CIRCSockTo_p_CIRCSocket, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CIncomingConnection[] = {  {&_swigt__p_CIncomingConnection, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CInviteMessage[] = {  {&_swigt__p_CInviteMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CJoinMessage[] = {  {&_swigt__p_CJoinMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CKickMessage[] = {  {&_swigt__p_CKickMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CLanguageScope[] = {  {&_swigt__p_CLanguageScope, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CListener[] = {  {&_swigt__p_CListener, 0, 0, 0},  {&_swigt__p_CTCPListener, _p_CTCPListenerTo_p_CListener, 0, 0},  {&_swigt__p_CUnixListener, _p_CUnixListenerTo_p_CListener, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CMessage[] = {  {&_swigt__p_CMessage, 0, 0, 0},  {&_swigt__p_CActionMessage, _p_CActionMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CAuthenticateMessage, _p_CAuthenticateMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CCTCPMessage, _p_CCTCPMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CChgHostMessage, _p_CChgHostMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CInviteMessage, _p_CInviteMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CJoinMessage, _p_CJoinMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CKickMessage, _p_CKickMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CModeMessage, _p_CModeMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CNickMessage, _p_CNickMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CNoticeMessage, _p_CNoticeMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CNumericMessage, _p_CNumericMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CPartMessage, _p_CPartMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CQuitMessage, _p_CQuitMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CTargetMessage, _p_CTargetMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CTextMessage, _p_CTextMessageTo_p_CMessage, 0, 0},  {&_swigt__p_CTopicMessage, _p_CTopicMessageTo_p_CMessage, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CModCommand[] = {  {&_swigt__p_CModCommand, 0, 0, 0},  {&_swigt__p_CPyModCommand, _p_CPyModCommandTo_p_CModCommand, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CModInfo[] = {  {&_swigt__p_CModInfo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CModPython[] = {  {&_swigt__p_CModPython, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CModeMessage[] = {  {&_swigt__p_CModeMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CModule[] = {  {&_swigt__p_CModule, 0, 0, 0},  {&_swigt__p_CModPython, _p_CModPythonTo_p_CModule, 0, 0},  {&_swigt__p_CPyModule, _p_CPyModuleTo_p_CModule, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CModuleEntry[] = {  {&_swigt__p_CModuleEntry, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CModules[] = {  {&_swigt__p_CModules, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CModulesIter[] = {  {&_swigt__p_CModulesIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CNick[] = {  {&_swigt__p_CNick, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CNickMessage[] = {  {&_swigt__p_CNickMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CNoticeMessage[] = {  {&_swigt__p_CNoticeMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CNumericMessage[] = {  {&_swigt__p_CNumericMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_COptionalTranslation[] = {  {&_swigt__p_COptionalTranslation, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CPartMessage[] = {  {&_swigt__p_CPartMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CPyCapability[] = {  {&_swigt__p_CPyCapability, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CPyModCommand[] = {  {&_swigt__p_CPyModCommand, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CPyModule[] = {  {&_swigt__p_CPyModule, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CPyRetBool[] = {  {&_swigt__p_CPyRetBool, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CPyRetString[] = {  {&_swigt__p_CPyRetString, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CPySocket[] = {  {&_swigt__p_CPySocket, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CPyTimer[] = {  {&_swigt__p_CPyTimer, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CQuery[] = {  {&_swigt__p_CQuery, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CQuitMessage[] = {  {&_swigt__p_CQuitMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CRealListener[] = {  {&_swigt__p_CRealListener, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CSCharBuffer[] = {  {&_swigt__p_CSCharBuffer, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CSConnection[] = {  {&_swigt__p_CSConnection, 0, 0, 0},  {&_swigt__p_CSSSLConnection, _p_CSSSLConnectionTo_p_CSConnection, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CSListener[] = {  {&_swigt__p_CSListener, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CSMonitorFD[] = {  {&_swigt__p_CSMonitorFD, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CSSSLConnection[] = {  {&_swigt__p_CSSSLConnection, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CSSockAddr[] = {  {&_swigt__p_CSSockAddr, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CServer[] = {  {&_swigt__p_CServer, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CSockCommon[] = {  {&_swigt__p_CSockCommon, 0, 0, 0},  {&_swigt__p_CClient, _p_CClientTo_p_CSockCommon, 0, 0},  {&_swigt__p_CExecSock, _p_CExecSockTo_p_CSockCommon, 0, 0},  {&_swigt__p_CHTTPSock, _p_CHTTPSockTo_p_CSockCommon, 0, 0},  {&_swigt__p_CIRCSock, _p_CIRCSockTo_p_CSockCommon, 0, 0},  {&_swigt__p_CIRCSocket, _p_CIRCSocketTo_p_CSockCommon, 0, 0},  {&_swigt__p_CIncomingConnection, _p_CIncomingConnectionTo_p_CSockCommon, 0, 0},  {&_swigt__p_CPySocket, _p_CPySocketTo_p_CSockCommon, 0, 0},  {&_swigt__p_CRealListener, _p_CRealListenerTo_p_CSockCommon, 0, 0},  {&_swigt__p_CSockManager, _p_CSockManagerTo_p_CSockCommon, 0, 0},  {&_swigt__p_CSocket, _p_CSocketTo_p_CSockCommon, 0, 0},  {&_swigt__p_CSocketManager, _p_CSocketManagerTo_p_CSockCommon, 0, 0},  {&_swigt__p_CWebSock, _p_CWebSockTo_p_CSockCommon, 0, 0},  {&_swigt__p_CZNCSock, _p_CZNCSockTo_p_CSockCommon, 0, 0},  {&_swigt__p_Csock, _p_CsockTo_p_CSockCommon, 0, 0},  {&_swigt__p_TSocketManagerT_CZNCSock_t, _p_TSocketManagerT_CZNCSock_tTo_p_CSockCommon, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CSockManager[] = {  {&_swigt__p_CSockManager, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CSocket[] = {  {&_swigt__p_CSocket, 0, 0, 0},  {&_swigt__p_CHTTPSock, _p_CHTTPSockTo_p_CSocket, 0, 0},  {&_swigt__p_CPySocket, _p_CPySocketTo_p_CSocket, 0, 0},  {&_swigt__p_CWebSock, _p_CWebSockTo_p_CSocket, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CSocketManager[] = {  {&_swigt__p_CSocketManager, 0, 0, 0},  {&_swigt__p_CSockManager, _p_CSockManagerTo_p_CSocketManager, 0, 0},  {&_swigt__p_TSocketManagerT_CZNCSock_t, _p_TSocketManagerT_CZNCSock_tTo_p_CSocketManager, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CString__EEscape[] = {  {&_swigt__p_CString__EEscape, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CString__size_type[] = {  {&_swigt__p_CString__size_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CTCPListener[] = {  {&_swigt__p_CTCPListener, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CTable[] = {  {&_swigt__p_CTable, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CTargetMessage[] = {  {&_swigt__p_CTargetMessage, 0, 0, 0},  {&_swigt__p_CActionMessage, _p_CActionMessageTo_p_CTargetMessage, 0, 0},  {&_swigt__p_CCTCPMessage, _p_CCTCPMessageTo_p_CTargetMessage, 0, 0},  {&_swigt__p_CJoinMessage, _p_CJoinMessageTo_p_CTargetMessage, 0, 0},  {&_swigt__p_CKickMessage, _p_CKickMessageTo_p_CTargetMessage, 0, 0},  {&_swigt__p_CModeMessage, _p_CModeMessageTo_p_CTargetMessage, 0, 0},  {&_swigt__p_CNoticeMessage, _p_CNoticeMessageTo_p_CTargetMessage, 0, 0},  {&_swigt__p_CPartMessage, _p_CPartMessageTo_p_CTargetMessage, 0, 0},  {&_swigt__p_CTextMessage, _p_CTextMessageTo_p_CTargetMessage, 0, 0},  {&_swigt__p_CTopicMessage, _p_CTopicMessageTo_p_CTargetMessage, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CTemplate[] = {  {&_swigt__p_CTemplate, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CTemplateLoopContext[] = {  {&_swigt__p_CTemplateLoopContext, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CTemplateOptions[] = {  {&_swigt__p_CTemplateOptions, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CTemplateTagHandler[] = {  {&_swigt__p_CTemplateTagHandler, 0, 0, 0},  {&_swigt__p_CZNCTagHandler, _p_CZNCTagHandlerTo_p_CTemplateTagHandler, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CTextMessage[] = {  {&_swigt__p_CTextMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CTimer[] = {  {&_swigt__p_CTimer, 0, 0, 0},  {&_swigt__p_CFPTimer, _p_CFPTimerTo_p_CTimer, 0, 0},  {&_swigt__p_CPyTimer, _p_CPyTimerTo_p_CTimer, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CTopicMessage[] = {  {&_swigt__p_CTopicMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CTranslation[] = {  {&_swigt__p_CTranslation, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CTranslationDomainRefHolder[] = {  {&_swigt__p_CTranslationDomainRefHolder, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CTranslationInfo[] = {  {&_swigt__p_CTranslationInfo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CUnixListener[] = {  {&_swigt__p_CUnixListener, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CUser[] = {  {&_swigt__p_CUser, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CUtils[] = {  {&_swigt__p_CUtils, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CWebSession[] = {  {&_swigt__p_CWebSession, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CWebSessionMap[] = {  {&_swigt__p_CWebSessionMap, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CWebSock[] = {  {&_swigt__p_CWebSock, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CWebSubPage[] = {  {&_swigt__p_CWebSubPage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CZNC[] = {  {&_swigt__p_CZNC, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CZNCSock[] = {  {&_swigt__p_CZNCSock, 0, 0, 0},  {&_swigt__p_CClient, _p_CClientTo_p_CZNCSock, 0, 0},  {&_swigt__p_CExecSock, _p_CExecSockTo_p_CZNCSock, 0, 0},  {&_swigt__p_CHTTPSock, _p_CHTTPSockTo_p_CZNCSock, 0, 0},  {&_swigt__p_CIRCSock, _p_CIRCSockTo_p_CZNCSock, 0, 0},  {&_swigt__p_CIRCSocket, _p_CIRCSocketTo_p_CZNCSock, 0, 0},  {&_swigt__p_CIncomingConnection, _p_CIncomingConnectionTo_p_CZNCSock, 0, 0},  {&_swigt__p_CPySocket, _p_CPySocketTo_p_CZNCSock, 0, 0},  {&_swigt__p_CRealListener, _p_CRealListenerTo_p_CZNCSock, 0, 0},  {&_swigt__p_CSocket, _p_CSocketTo_p_CZNCSock, 0, 0},  {&_swigt__p_CWebSock, _p_CWebSockTo_p_CZNCSock, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CZNCTagHandler[] = {  {&_swigt__p_CZNCTagHandler, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_CmdFunc[] = {  {&_swigt__p_CmdFunc, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Csock[] = {  {&_swigt__p_Csock, 0, 0, 0},  {&_swigt__p_CClient, _p_CClientTo_p_Csock, 0, 0},  {&_swigt__p_CExecSock, _p_CExecSockTo_p_Csock, 0, 0},  {&_swigt__p_CHTTPSock, _p_CHTTPSockTo_p_Csock, 0, 0},  {&_swigt__p_CIRCSock, _p_CIRCSockTo_p_Csock, 0, 0},  {&_swigt__p_CIRCSocket, _p_CIRCSocketTo_p_Csock, 0, 0},  {&_swigt__p_CIncomingConnection, _p_CIncomingConnectionTo_p_Csock, 0, 0},  {&_swigt__p_CPySocket, _p_CPySocketTo_p_Csock, 0, 0},  {&_swigt__p_CRealListener, _p_CRealListenerTo_p_Csock, 0, 0},  {&_swigt__p_CSocket, _p_CSocketTo_p_Csock, 0, 0},  {&_swigt__p_CWebSock, _p_CWebSockTo_p_Csock, 0, 0},  {&_swigt__p_CZNCSock, _p_CZNCSockTo_p_Csock, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_EAcceptType[] = {  {&_swigt__p_EAcceptType, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_EChanModeArgs[] = {  {&_swigt__p_EChanModeArgs, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_EModException[] = {  {&_swigt__p_EModException, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_EModRet[] = {  {&_swigt__p_EModRet, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_EModes[] = {  {&_swigt__p_EModes, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_EModuleType[] = {  {&_swigt__p_EModuleType, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_EType[] = {  {&_swigt__p_EType, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_EUserPerms[] = {  {&_swigt__p_EUserPerms, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_EntryMap[] = {  {&_swigt__p_EntryMap, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_EntryMapIterator[] = {  {&_swigt__p_EntryMapIterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_MCString[] = {  {&_swigt__p_MCString, 0, 0, 0},  {&_swigt__p_CTemplate, _p_CTemplateTo_p_MCString, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_MCString_iter[] = {  {&_swigt__p_MCString_iter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ModCmdFunc[] = {  {&_swigt__p_ModCmdFunc, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ModDirList[] = {  {&_swigt__p_ModDirList, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_String[] = {  {&_swigt__p_String, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_SubConfig[] = {  {&_swigt__p_SubConfig, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_SubConfigEntry[] = {  {&_swigt__p_SubConfigEntry, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_SubConfigMap[] = {  {&_swigt__p_SubConfigMap, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_SubConfigMapIterator[] = {  {&_swigt__p_SubConfigMapIterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TSocketManagerT_CZNCSock_t[] = {  {&_swigt__p_TSocketManagerT_CZNCSock_t, 0, 0, 0},  {&_swigt__p_CSockManager, _p_CSockManagerTo_p_TSocketManagerT_CZNCSock_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TrafficStatsMap[] = {  {&_swigt__p_TrafficStatsMap, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TrafficStatsPair[] = {  {&_swigt__p_TrafficStatsPair, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_allocator_type[] = {  {&_swigt__p_allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_bool[] = {  {&_swigt__p_bool, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_difference_type[] = {  {&_swigt__p_difference_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_double[] = {  {&_swigt__p_double, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModPython[] = {{&_swigt__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModPython, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CPyModule[] = {{&_swigt__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CPyModule, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModule[] = {  {&_swigt__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModule, 0, 0, 0},  {&_swigt__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModPython, _p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModPythonTo_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModule, 0, 0},  {&_swigt__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CPyModule, _p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CPyModuleTo_p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModule, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_CModule_p_CFPTimer__void[] = {  {&_swigt__p_f_p_CModule_p_CFPTimer__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_r_CModInfo__void[] = {  {&_swigt__p_f_r_CModInfo__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_r_std__ostream__r_std__ostream[] = {  {&_swigt__p_f_r_std__ostream__r_std__ostream, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_fd_set[] = {  {&_swigt__p_fd_set, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_first_type[] = {  {&_swigt__p_first_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_gid_t[] = {  {&_swigt__p_gid_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_in_addr[] = {  {&_swigt__p_in_addr, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int32_t[] = {  {&_swigt__p_int32_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int64_t[] = {  {&_swigt__p_int64_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_key_type[] = {  {&_swigt__p_key_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_long[] = {  {&_swigt__p_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_mapped_type[] = {  {&_swigt__p_mapped_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_mode_t[] = {  {&_swigt__p_mode_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_PyObject[] = {  {&_swigt__p_p_PyObject, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_second_type[] = {  {&_swigt__p_second_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_size_type[] = {  {&_swigt__p_size_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_sockaddr_in[] = {  {&_swigt__p_sockaddr_in, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_sockaddr_storage[] = {  {&_swigt__p_sockaddr_storage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_socklen_t[] = {  {&_swigt__p_socklen_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ssize_t[] = {  {&_swigt__p_ssize_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_stat[] = {  {&_swigt__p_stat, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_CBufLine_t[] = {  {&_swigt__p_std__allocatorT_CBufLine_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_CChan_p_t[] = {  {&_swigt__p_std__allocatorT_CChan_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_CClient_p_t[] = {  {&_swigt__p_std__allocatorT_CClient_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_CIRCNetwork_p_t[] = {  {&_swigt__p_std__allocatorT_CIRCNetwork_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_CListener_p_t[] = {  {&_swigt__p_std__allocatorT_CListener_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_CModule_p_t[] = {  {&_swigt__p_std__allocatorT_CModule_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_CQuery_p_t[] = {  {&_swigt__p_std__allocatorT_CQuery_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_CServer_p_t[] = {  {&_swigt__p_std__allocatorT_CServer_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_CString_t[] = {  {&_swigt__p_std__allocatorT_CString_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_std__pairT_CString_CString_t_t[] = {  {&_swigt__p_std__allocatorT_std__pairT_CString_CString_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_std__pairT_CString_const_CNick_t_t[] = {  {&_swigt__p_std__allocatorT_std__pairT_CString_const_CNick_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_std__pairT_CString_const_CString_t_t[] = {  {&_swigt__p_std__allocatorT_std__pairT_CString_const_CString_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_std__pairT_CString_const_std__vectorT_CString_t_t_t[] = {  {&_swigt__p_std__allocatorT_std__pairT_CString_const_std__vectorT_CString_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_std__shared_ptrT_CWebSubPage_t_t[] = {  {&_swigt__p_std__allocatorT_std__shared_ptrT_CWebSubPage_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_std__vectorT_CString_t_t[] = {  {&_swigt__p_std__allocatorT_std__vectorT_CString_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__dequeT_CBufLine_t[] = {  {&_swigt__p_std__dequeT_CBufLine_t, 0, 0, 0},  {&_swigt__p_CBuffer, _p_CBufferTo_p_std__dequeT_CBufLine_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__functionT_void_fCString_const_RF_t[] = {  {&_swigt__p_std__functionT_void_fCString_const_RF_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__invalid_argument[] = {  {&_swigt__p_std__invalid_argument, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__lessT_CModInfo_t[] = {  {&_swigt__p_std__lessT_CModInfo_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__lessT_CString_t[] = {  {&_swigt__p_std__lessT_CString_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__listT_CIRCNetwork_p_std__allocatorT_CIRCNetwork_p_t_t[] = {  {&_swigt__p_std__listT_CIRCNetwork_p_std__allocatorT_CIRCNetwork_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__listT_CString_t[] = {  {&_swigt__p_std__listT_CString_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_CString_CNick_t[] = {  {&_swigt__p_std__mapT_CString_CNick_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_CString_CString_t[] = {  {&_swigt__p_std__mapT_CString_CString_t, 0, 0, 0},  {&_swigt__p_CTemplate, _p_CTemplateTo_p_std__mapT_CString_CString_t, 0, 0},  {&_swigt__p_MCString, _p_MCStringTo_p_std__mapT_CString_CString_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_CString_CString_t__iterator[] = {  {&_swigt__p_std__mapT_CString_CString_t__iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_CString_CTranslationInfo_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CTranslationInfo_t_t_t[] = {  {&_swigt__p_std__mapT_CString_CTranslationInfo_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CTranslationInfo_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_CString_CUser_p_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CUser_p_t_t_t[] = {  {&_swigt__p_std__mapT_CString_CUser_p_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CUser_p_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_CString_VCString_t[] = {  {&_swigt__p_std__mapT_CString_VCString_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_CString_std__pairT_unsigned_long_long_unsigned_long_long_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__pairT_unsigned_long_long_unsigned_long_long_t_t_t_t[] = {  {&_swigt__p_std__mapT_CString_std__pairT_unsigned_long_long_unsigned_long_long_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__pairT_unsigned_long_long_unsigned_long_long_t_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_CString_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_t_t_t__const_iterator[] = {  {&_swigt__p_std__mapT_CString_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_t_t_t__const_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_char_CIRCSock__EChanModeArgs_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CIRCSock__EChanModeArgs_t_t_t[] = {  {&_swigt__p_std__mapT_char_CIRCSock__EChanModeArgs_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CIRCSock__EChanModeArgs_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_char_CString_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CString_t_t_t[] = {  {&_swigt__p_std__mapT_char_CString_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CString_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_char_unsigned_int_std__lessT_char_t_std__allocatorT_std__pairT_char_const_unsigned_int_t_t_t[] = {  {&_swigt__p_std__mapT_char_unsigned_int_std__lessT_char_t_std__allocatorT_std__pairT_char_const_unsigned_int_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t[] = {  {&_swigt__p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__ostream[] = {  {&_swigt__p_std__ostream, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__pairT_CString_CString_t[] = {  {&_swigt__p_std__pairT_CString_CString_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__pairT_unsigned_long_long_unsigned_long_long_t[] = {  {&_swigt__p_std__pairT_unsigned_long_long_unsigned_long_long_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__setT_CChan_p_std__lessT_CChan_p_t_std__allocatorT_CChan_p_t_t[] = {  {&_swigt__p_std__setT_CChan_p_std__lessT_CChan_p_t_std__allocatorT_CChan_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__setT_CModInfo_t[] = {  {&_swigt__p_std__setT_CModInfo_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__setT_CSocket_p_std__lessT_CSocket_p_t_std__allocatorT_CSocket_p_t_t__const_iterator[] = {  {&_swigt__p_std__setT_CSocket_p_std__lessT_CSocket_p_t_std__allocatorT_CSocket_p_t_t__const_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__setT_CString_t[] = {  {&_swigt__p_std__setT_CString_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__setT_CTimer_p_std__lessT_CTimer_p_t_std__allocatorT_CTimer_p_t_t__const_iterator[] = {  {&_swigt__p_std__setT_CTimer_p_std__lessT_CTimer_p_t_std__allocatorT_CTimer_p_t_t__const_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__setT_char_std__lessT_char_t_std__allocatorT_char_t_t[] = {  {&_swigt__p_std__setT_char_std__lessT_char_t_std__allocatorT_char_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__shared_ptrT_CAuthBase_t[] = {  {&_swigt__p_std__shared_ptrT_CAuthBase_t, 0, 0, 0},  {&_swigt__p_std__shared_ptrT_CClientAuth_t, _p_std__shared_ptrT_CClientAuth_tTo_p_std__shared_ptrT_CAuthBase_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__shared_ptrT_CClientAuth_t[] = {  {&_swigt__p_std__shared_ptrT_CClientAuth_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__shared_ptrT_CTemplateOptions_t[] = {  {&_swigt__p_std__shared_ptrT_CTemplateOptions_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__shared_ptrT_CTemplateTagHandler_t[] = {  {&_swigt__p_std__shared_ptrT_CTemplateTagHandler_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__shared_ptrT_CWebSession_t[] = {  {&_swigt__p_std__shared_ptrT_CWebSession_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__shared_ptrT_CWebSubPage_t[] = {  {&_swigt__p_std__shared_ptrT_CWebSubPage_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_CChan_p_t[] = {  {&_swigt__p_std__vectorT_CChan_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_CClient_p_t[] = {  {&_swigt__p_std__vectorT_CClient_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_CCron_p_std__allocatorT_CCron_p_t_t[] = {  {&_swigt__p_std__vectorT_CCron_p_std__allocatorT_CCron_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_CIRCNetwork_p_t[] = {  {&_swigt__p_std__vectorT_CIRCNetwork_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_CListener_p_t[] = {  {&_swigt__p_std__vectorT_CListener_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_CModule_p_t[] = {  {&_swigt__p_std__vectorT_CModule_p_t, 0, 0, 0},  {&_swigt__p_CModules, _p_CModulesTo_p_std__vectorT_CModule_p_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_CModule_p_t__const_iterator[] = {  {&_swigt__p_std__vectorT_CModule_p_t__const_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_CQuery_p_t[] = {  {&_swigt__p_std__vectorT_CQuery_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_CServer_p_t[] = {  {&_swigt__p_std__vectorT_CServer_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_CString_t[] = {  {&_swigt__p_std__vectorT_CString_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_CTemplate_p_std__allocatorT_CTemplate_p_t_t[] = {  {&_swigt__p_std__vectorT_CTemplate_p_std__allocatorT_CTemplate_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_Csock_p_std__allocatorT_Csock_p_t_t[] = {  {&_swigt__p_CSockManager, 0, 0, 0},  {&_swigt__p_TSocketManagerT_CZNCSock_t, 0, 0, 0},  {&_swigt__p_std__vectorT_Csock_p_std__allocatorT_Csock_p_t_t, 0, 0, 0},  {&_swigt__p_CSocketManager, _p_CSocketManagerTo_p_std__vectorT_Csock_p_std__allocatorT_Csock_p_t_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_TWebSubPage_t[] = {  {&_swigt__p_std__vectorT_TWebSubPage_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_VCString_t[] = {  {&_swigt__p_std__vectorT_VCString_t, 0, 0, 0},  {&_swigt__p_CTable, _p_CTableTo_p_std__vectorT_VCString_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t[] = {  {&_swigt__p_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_std__pairT_CString_CString_t_t[] = {  {&_swigt__p_std__vectorT_std__pairT_CString_CString_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_std__shared_ptrT_CTemplateTagHandler_t_std__allocatorT_std__shared_ptrT_CTemplateTagHandler_t_t_t[] = {  {&_swigt__p_std__vectorT_std__shared_ptrT_CTemplateTagHandler_t_std__allocatorT_std__shared_ptrT_CTemplateTagHandler_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_swig__SwigPyIterator[] = {  {&_swigt__p_swig__SwigPyIterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_time_t[] = {  {&_swigt__p_time_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_timeval[] = {  {&_swigt__p_timeval, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_uid_t[] = {  {&_swigt__p_uid_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_uint16_t[] = {  {&_swigt__p_uint16_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_uint64_t[] = {  {&_swigt__p_uint64_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_int[] = {  {&_swigt__p_unsigned_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_short[] = {  {&_swigt__p_unsigned_short, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_value_type[] = {  {&_swigt__p_value_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_void[] = {  {&_swigt__p_void, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__m_CModule__f_r_q_const__CString__void,
  _swigc__p_CActionMessage,
  _swigc__p_CAuthBase,
  _swigc__p_CAuthenticateMessage,
  _swigc__p_CBufLine,
  _swigc__p_CBuffer,
  _swigc__p_CCTCPMessage,
  _swigc__p_CCapability,
  _swigc__p_CChan,
  _swigc__p_CChgHostMessage,
  _swigc__p_CClient,
  _swigc__p_CClientAuth,
  _swigc__p_CConfig,
  _swigc__p_CConfigEntry,
  _swigc__p_CConnectQueueTimer,
  _swigc__p_CCoreTranslationMixin,
  _swigc__p_CCron,
  _swigc__p_CDebug,
  _swigc__p_CDebugStream,
  _swigc__p_CDelayedTranslation,
  _swigc__p_CDir,
  _swigc__p_CException,
  _swigc__p_CExecSock,
  _swigc__p_CFPTimer,
  _swigc__p_CFile,
  _swigc__p_CGetAddrInfo,
  _swigc__p_CHTTPSock,
  _swigc__p_CIRCNetwork,
  _swigc__p_CIRCSock,
  _swigc__p_CIRCSocket,
  _swigc__p_CIncomingConnection,
  _swigc__p_CInviteMessage,
  _swigc__p_CJoinMessage,
  _swigc__p_CKickMessage,
  _swigc__p_CLanguageScope,
  _swigc__p_CListener,
  _swigc__p_CMessage,
  _swigc__p_CModCommand,
  _swigc__p_CModInfo,
  _swigc__p_CModPython,
  _swigc__p_CModeMessage,
  _swigc__p_CModule,
  _swigc__p_CModuleEntry,
  _swigc__p_CModules,
  _swigc__p_CModulesIter,
  _swigc__p_CNick,
  _swigc__p_CNickMessage,
  _swigc__p_CNoticeMessage,
  _swigc__p_CNumericMessage,
  _swigc__p_COptionalTranslation,
  _swigc__p_CPartMessage,
  _swigc__p_CPyCapability,
  _swigc__p_CPyModCommand,
  _swigc__p_CPyModule,
  _swigc__p_CPyRetBool,
  _swigc__p_CPyRetString,
  _swigc__p_CPySocket,
  _swigc__p_CPyTimer,
  _swigc__p_CQuery,
  _swigc__p_CQuitMessage,
  _swigc__p_CRealListener,
  _swigc__p_CSCharBuffer,
  _swigc__p_CSConnection,
  _swigc__p_CSListener,
  _swigc__p_CSMonitorFD,
  _swigc__p_CSSSLConnection,
  _swigc__p_CSSockAddr,
  _swigc__p_CServer,
  _swigc__p_CSockCommon,
  _swigc__p_CSockManager,
  _swigc__p_CSocket,
  _swigc__p_CSocketManager,
  _swigc__p_CString__EEscape,
  _swigc__p_CString__size_type,
  _swigc__p_CTCPListener,
  _swigc__p_CTable,
  _swigc__p_CTargetMessage,
  _swigc__p_CTemplate,
  _swigc__p_CTemplateLoopContext,
  _swigc__p_CTemplateOptions,
  _swigc__p_CTemplateTagHandler,
  _swigc__p_CTextMessage,
  _swigc__p_CTimer,
  _swigc__p_CTopicMessage,
  _swigc__p_CTranslation,
  _swigc__p_CTranslationDomainRefHolder,
  _swigc__p_CTranslationInfo,
  _swigc__p_CUnixListener,
  _swigc__p_CUser,
  _swigc__p_CUtils,
  _swigc__p_CWebSession,
  _swigc__p_CWebSessionMap,
  _swigc__p_CWebSock,
  _swigc__p_CWebSubPage,
  _swigc__p_CZNC,
  _swigc__p_CZNCSock,
  _swigc__p_CZNCTagHandler,
  _swigc__p_CmdFunc,
  _swigc__p_Csock,
  _swigc__p_EAcceptType,
  _swigc__p_EChanModeArgs,
  _swigc__p_EModException,
  _swigc__p_EModRet,
  _swigc__p_EModes,
  _swigc__p_EModuleType,
  _swigc__p_EType,
  _swigc__p_EUserPerms,
  _swigc__p_EntryMap,
  _swigc__p_EntryMapIterator,
  _swigc__p_MCString,
  _swigc__p_MCString_iter,
  _swigc__p_ModCmdFunc,
  _swigc__p_ModDirList,
  _swigc__p_String,
  _swigc__p_SubConfig,
  _swigc__p_SubConfigEntry,
  _swigc__p_SubConfigMap,
  _swigc__p_SubConfigMapIterator,
  _swigc__p_TSocketManagerT_CZNCSock_t,
  _swigc__p_TrafficStatsMap,
  _swigc__p_TrafficStatsPair,
  _swigc__p_allocator_type,
  _swigc__p_bool,
  _swigc__p_char,
  _swigc__p_difference_type,
  _swigc__p_double,
  _swigc__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModPython,
  _swigc__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CModule,
  _swigc__p_f_ModHandle_p_CUser_p_CIRCNetwork_r_q_const__CString_r_q_const__CString_CModInfo__EModuleType__p_CPyModule,
  _swigc__p_f_p_CModule_p_CFPTimer__void,
  _swigc__p_f_r_CModInfo__void,
  _swigc__p_f_r_std__ostream__r_std__ostream,
  _swigc__p_fd_set,
  _swigc__p_first_type,
  _swigc__p_gid_t,
  _swigc__p_in_addr,
  _swigc__p_int,
  _swigc__p_int32_t,
  _swigc__p_int64_t,
  _swigc__p_key_type,
  _swigc__p_long,
  _swigc__p_mapped_type,
  _swigc__p_mode_t,
  _swigc__p_p_PyObject,
  _swigc__p_second_type,
  _swigc__p_size_type,
  _swigc__p_sockaddr_in,
  _swigc__p_sockaddr_storage,
  _swigc__p_socklen_t,
  _swigc__p_ssize_t,
  _swigc__p_stat,
  _swigc__p_std__allocatorT_CBufLine_t,
  _swigc__p_std__allocatorT_CChan_p_t,
  _swigc__p_std__allocatorT_CClient_p_t,
  _swigc__p_std__allocatorT_CIRCNetwork_p_t,
  _swigc__p_std__allocatorT_CListener_p_t,
  _swigc__p_std__allocatorT_CModule_p_t,
  _swigc__p_std__allocatorT_CQuery_p_t,
  _swigc__p_std__allocatorT_CServer_p_t,
  _swigc__p_std__allocatorT_CString_t,
  _swigc__p_std__allocatorT_std__pairT_CString_CString_t_t,
  _swigc__p_std__allocatorT_std__pairT_CString_const_CNick_t_t,
  _swigc__p_std__allocatorT_std__pairT_CString_const_CString_t_t,
  _swigc__p_std__allocatorT_std__pairT_CString_const_std__vectorT_CString_t_t_t,
  _swigc__p_std__allocatorT_std__shared_ptrT_CWebSubPage_t_t,
  _swigc__p_std__allocatorT_std__vectorT_CString_t_t,
  _swigc__p_std__dequeT_CBufLine_t,
  _swigc__p_std__functionT_void_fCString_const_RF_t,
  _swigc__p_std__invalid_argument,
  _swigc__p_std__lessT_CModInfo_t,
  _swigc__p_std__lessT_CString_t,
  _swigc__p_std__listT_CIRCNetwork_p_std__allocatorT_CIRCNetwork_p_t_t,
  _swigc__p_std__listT_CString_t,
  _swigc__p_std__mapT_CString_CNick_t,
  _swigc__p_std__mapT_CString_CString_t,
  _swigc__p_std__mapT_CString_CString_t__iterator,
  _swigc__p_std__mapT_CString_CTranslationInfo_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CTranslationInfo_t_t_t,
  _swigc__p_std__mapT_CString_CUser_p_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_CUser_p_t_t_t,
  _swigc__p_std__mapT_CString_VCString_t,
  _swigc__p_std__mapT_CString_std__pairT_unsigned_long_long_unsigned_long_long_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__pairT_unsigned_long_long_unsigned_long_long_t_t_t_t,
  _swigc__p_std__mapT_CString_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_std__lessT_CString_t_std__allocatorT_std__pairT_CString_const_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t_t_t_t__const_iterator,
  _swigc__p_std__mapT_char_CIRCSock__EChanModeArgs_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CIRCSock__EChanModeArgs_t_t_t,
  _swigc__p_std__mapT_char_CString_std__lessT_char_t_std__allocatorT_std__pairT_char_const_CString_t_t_t,
  _swigc__p_std__mapT_char_unsigned_int_std__lessT_char_t_std__allocatorT_std__pairT_char_const_unsigned_int_t_t_t,
  _swigc__p_std__mapT_int_short_std__lessT_int_t_std__allocatorT_std__pairT_int_const_short_t_t_t,
  _swigc__p_std__ostream,
  _swigc__p_std__pairT_CString_CString_t,
  _swigc__p_std__pairT_unsigned_long_long_unsigned_long_long_t,
  _swigc__p_std__setT_CChan_p_std__lessT_CChan_p_t_std__allocatorT_CChan_p_t_t,
  _swigc__p_std__setT_CModInfo_t,
  _swigc__p_std__setT_CSocket_p_std__lessT_CSocket_p_t_std__allocatorT_CSocket_p_t_t__const_iterator,
  _swigc__p_std__setT_CString_t,
  _swigc__p_std__setT_CTimer_p_std__lessT_CTimer_p_t_std__allocatorT_CTimer_p_t_t__const_iterator,
  _swigc__p_std__setT_char_std__lessT_char_t_std__allocatorT_char_t_t,
  _swigc__p_std__shared_ptrT_CAuthBase_t,
  _swigc__p_std__shared_ptrT_CClientAuth_t,
  _swigc__p_std__shared_ptrT_CTemplateOptions_t,
  _swigc__p_std__shared_ptrT_CTemplateTagHandler_t,
  _swigc__p_std__shared_ptrT_CWebSession_t,
  _swigc__p_std__shared_ptrT_CWebSubPage_t,
  _swigc__p_std__vectorT_CChan_p_t,
  _swigc__p_std__vectorT_CClient_p_t,
  _swigc__p_std__vectorT_CCron_p_std__allocatorT_CCron_p_t_t,
  _swigc__p_std__vectorT_CIRCNetwork_p_t,
  _swigc__p_std__vectorT_CListener_p_t,
  _swigc__p_std__vectorT_CModule_p_t,
  _swigc__p_std__vectorT_CModule_p_t__const_iterator,
  _swigc__p_std__vectorT_CQuery_p_t,
  _swigc__p_std__vectorT_CServer_p_t,
  _swigc__p_std__vectorT_CString_t,
  _swigc__p_std__vectorT_CTemplate_p_std__allocatorT_CTemplate_p_t_t,
  _swigc__p_std__vectorT_Csock_p_std__allocatorT_Csock_p_t_t,
  _swigc__p_std__vectorT_TWebSubPage_t,
  _swigc__p_std__vectorT_VCString_t,
  _swigc__p_std__vectorT_std__pairT_CString_CConfigEntry_t_std__allocatorT_std__pairT_CString_CConfigEntry_t_t_t,
  _swigc__p_std__vectorT_std__pairT_CString_CString_t_t,
  _swigc__p_std__vectorT_std__shared_ptrT_CTemplateTagHandler_t_std__allocatorT_std__shared_ptrT_CTemplateTagHandler_t_t_t,
  _swigc__p_swig__SwigPyIterator,
  _swigc__p_time_t,
  _swigc__p_timeval,
  _swigc__p_uid_t,
  _swigc__p_uint16_t,
  _swigc__p_uint64_t,
  _swigc__p_unsigned_int,
  _swigc__p_unsigned_short,
  _swigc__p_value_type,
  _swigc__p_void,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_const_info swig_const_table[] = {
{0, 0, 0, 0.0, 0, 0}};

#ifdef __cplusplus
}
#endif
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast linked list. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns. So to actually build the cast linked list,
 * we find the array of casts associated with the type, and loop through it
 * adding the casts to the list. The one last trick we need to do is making
 * sure the type pointer in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the linked list (because the cast->type) pointer will
 *     be correct.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif

#ifndef SWIG_INIT_CLIENT_DATA_TYPE
#define SWIG_INIT_CLIENT_DATA_TYPE void *
#endif

SWIGRUNTIME void
SWIG_InitializeModule(SWIG_INIT_CLIENT_DATA_TYPE clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %lu\n", (unsigned long)swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *ret;
    swig_cast_info *cast;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      ret = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        ret = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
#ifdef SWIGRUNTIME_DEBUG
        if (ret) printf("SWIG_InitializeModule: found cast %s\n", ret->name);
#endif
      }
      if (ret) {
        if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: skip old type %s\n", ret->name);
#endif
          cast->type = ret;
          ret = 0;
        } else {
          /* Check for casting already in the list */
          swig_cast_info *ocast = SWIG_TypeCheck(ret->name, type);
#ifdef SWIGRUNTIME_DEBUG
          if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", ret->name);
#endif
          if (!ocast) ret = 0;
        }
      }
      
      if (!ret) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        if (type->cast) {
          type->cast->prev = cast;
          cast->next = type->cast;
        }
        type->cast = cast;
      }
      cast++;
    }
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  swig_cast_info *equiv;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      equiv = swig_module.types[i]->cast;
      while (equiv) {
        if (!equiv->converter) {
          if (equiv->type && !equiv->type->clientdata)
          SWIG_TypeClientData(equiv->type, swig_module.types[i]->clientdata);
        }
        equiv = equiv->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#ifdef __cplusplus
extern "C" {
#endif
  
  /* -----------------------------------------------------------------------------
   * constants/methods manipulation
   * ----------------------------------------------------------------------------- */
  
  /* Install Constants */
  SWIGINTERN void
  SWIG_Python_InstallConstants(PyObject *d, swig_const_info constants[]) {
    PyObject *obj = 0;
    size_t i;
    for (i = 0; constants[i].type; ++i) {
      switch(constants[i].type) {
      case SWIG_PY_POINTER:
        obj = SWIG_InternalNewPointerObj(constants[i].pvalue, *(constants[i]).ptype,0);
        break;
      case SWIG_PY_BINARY:
        obj = SWIG_NewPackedObj(constants[i].pvalue, constants[i].lvalue, *(constants[i].ptype));
        break;
      default:
        obj = 0;
        break;
      }
      if (obj) {
        PyDict_SetItemString(d, constants[i].name, obj);
        Py_DECREF(obj);
      }
    }
  }
  
  /* -----------------------------------------------------------------------------
   * Patch %callback methods' docstrings to hold the callback ptrs
   * -----------------------------------------------------------------------------*/
  
  SWIGINTERN void
  SWIG_Python_FixMethods(PyMethodDef *methods, const swig_const_info *const_table, swig_type_info **types, swig_type_info **types_initial) {
    size_t i;
    for (i = 0; methods[i].ml_name; ++i) {
      const char *c = methods[i].ml_doc;
      if (!c) continue;
      c = strstr(c, "swig_ptr: ");
      if (c) {
        int j;
        const swig_const_info *ci = 0;
        const char *name = c + 10;
        for (j = 0; const_table[j].type; ++j) {
          if (strncmp(const_table[j].name, name, 
              strlen(const_table[j].name)) == 0) {
            ci = &(const_table[j]);
            break;
          }
        }
        if (ci) {
          void *ptr = (ci->type == SWIG_PY_POINTER) ? ci->pvalue : 0;
          if (ptr) {
            size_t shift = (ci->ptype) - types;
            swig_type_info *ty = types_initial[shift];
            size_t ldoc = (c - methods[i].ml_doc);
            size_t lptr = strlen(ty->name)+2*sizeof(void*)+2;
            char *ndoc = (char*)malloc(ldoc + lptr + 10);
            if (ndoc) {
              char *buff = ndoc;
              memcpy(buff, methods[i].ml_doc, ldoc);
              buff += ldoc;
              memcpy(buff, "swig_ptr: ", 10);
              buff += 10;
              SWIG_PackVoidPtr(buff, ptr, ty->name, lptr);
              methods[i].ml_doc = ndoc;
            }
          }
        }
      }
    }
  } 
  
#ifdef __cplusplus
}
#endif




/* -----------------------------------------------------------------------------*
 *  Partial Init method
 * -----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C"
#endif

SWIGEXPORT 
#if PY_VERSION_HEX >= 0x03000000
PyObject*
#else
void
#endif
SWIG_init(void) {
  PyObject *m, *d, *md, *globals;
  
#if PY_VERSION_HEX >= 0x03000000
  static struct PyModuleDef SWIG_module = {
    PyModuleDef_HEAD_INIT,
    SWIG_name,
    NULL,
    -1,
    SwigMethods,
    NULL,
    NULL,
    NULL,
    NULL
  };
#endif
  
#if defined(SWIGPYTHON_BUILTIN)
  static SwigPyClientData SwigPyObject_clientdata = {
    0, 0, 0, 0, 0, 0, 0
  };
  static PyGetSetDef this_getset_def = {
    (char *)"this", &SwigPyBuiltin_ThisClosure, NULL, NULL, NULL
  };
  static SwigPyGetSet thisown_getset_closure = {
    SwigPyObject_own,
    SwigPyObject_own
  };
  static PyGetSetDef thisown_getset_def = {
    (char *)"thisown", SwigPyBuiltin_GetterClosure, SwigPyBuiltin_SetterClosure, NULL, &thisown_getset_closure
  };
  PyTypeObject *builtin_pytype;
  int builtin_base_count;
  swig_type_info *builtin_basetype;
  PyObject *tuple;
  PyGetSetDescrObject *static_getset;
  PyTypeObject *metatype;
  PyTypeObject *swigpyobject;
  SwigPyClientData *cd;
  PyObject *public_interface, *public_symbol;
  PyObject *this_descr;
  PyObject *thisown_descr;
  PyObject *self = 0;
  int i;
  
  (void)builtin_pytype;
  (void)builtin_base_count;
  (void)builtin_basetype;
  (void)tuple;
  (void)static_getset;
  (void)self;
  
  /* Metaclass is used to implement static member variables */
  metatype = SwigPyObjectType();
  assert(metatype);
#endif
  
  (void)globals;
  
  /* Create singletons now to avoid potential deadlocks with multi-threaded usage after module initialization */
  SWIG_This();
  SWIG_Python_TypeCache();
  SwigPyPacked_type();
#ifndef SWIGPYTHON_BUILTIN
  SwigPyObject_type();
#endif
  
  /* Fix SwigMethods to carry the callback ptrs when needed */
  SWIG_Python_FixMethods(SwigMethods, swig_const_table, swig_types, swig_type_initial);
  
#if PY_VERSION_HEX >= 0x03000000
  m = PyModule_Create(&SWIG_module);
#else
  m = Py_InitModule(SWIG_name, SwigMethods);
#endif
  
  md = d = PyModule_GetDict(m);
  (void)md;
  
  SWIG_InitializeModule(0);
  
#ifdef SWIGPYTHON_BUILTIN
  swigpyobject = SwigPyObject_TypeOnce();
  
  SwigPyObject_stype = SWIG_MangledTypeQuery("_p_SwigPyObject");
  assert(SwigPyObject_stype);
  cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
  if (!cd) {
    SwigPyObject_stype->clientdata = &SwigPyObject_clientdata;
    SwigPyObject_clientdata.pytype = swigpyobject;
  } else if (swigpyobject->tp_basicsize != cd->pytype->tp_basicsize) {
    PyErr_SetString(PyExc_RuntimeError, "Import error: attempted to load two incompatible swig-generated modules.");
# if PY_VERSION_HEX >= 0x03000000
    return NULL;
# else
    return;
# endif
  }
  
  /* All objects have a 'this' attribute */
  this_descr = PyDescr_NewGetSet(SwigPyObject_type(), &this_getset_def);
  (void)this_descr;
  
  /* All objects have a 'thisown' attribute */
  thisown_descr = PyDescr_NewGetSet(SwigPyObject_type(), &thisown_getset_def);
  (void)thisown_descr;
  
  public_interface = PyList_New(0);
  public_symbol = 0;
  (void)public_symbol;
  
  PyDict_SetItemString(md, "__all__", public_interface);
  Py_DECREF(public_interface);
  for (i = 0; SwigMethods[i].ml_name != NULL; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, SwigMethods[i].ml_name);
  for (i = 0; swig_const_table[i].name != 0; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, swig_const_table[i].name);
#endif
  
  SWIG_InstallConstants(d,swig_const_table);
  
  SWIG_Python_SetConstant(d, "SHARED_PTR_DISOWN",SWIG_From_int(static_cast< int >(0)));
  
  // thread safe initialization
  swig::container_owner_attribute();
  
  globals = SWIG_globals();
  if (!globals) {
    PyErr_SetString(PyExc_TypeError, "Failure to create SWIG globals.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  PyDict_SetItemString(md, "cvar", globals);
  SWIG_addvarlink(globals, "g_HexDigits", Swig_var_g_HexDigits_get, Swig_var_g_HexDigits_set);
  SWIG_Python_SetConstant(d, "CException_EX_Shutdown",SWIG_From_int(static_cast< int >(CException::EX_Shutdown)));
  SWIG_Python_SetConstant(d, "CException_EX_Restart",SWIG_From_int(static_cast< int >(CException::EX_Restart)));
  SWIG_Python_SetConstant(d, "CTable_GridStyle",SWIG_From_int(static_cast< int >(CTable::GridStyle)));
  SWIG_Python_SetConstant(d, "CTable_ListStyle",SWIG_From_int(static_cast< int >(CTable::ListStyle)));
  SWIG_Python_SetConstant(d, "CS_INVALID_SOCK",SWIG_From_int(static_cast< int >(-1)));
  SWIG_Python_SetConstant(d, "CSSockAddr_RAF_ANY",SWIG_From_int(static_cast< int >(CSSockAddr::RAF_ANY)));
  SWIG_Python_SetConstant(d, "CSSockAddr_RAF_INET",SWIG_From_int(static_cast< int >(CSSockAddr::RAF_INET)));
  SWIG_addvarlink(globals, "CS_BLOCKSIZE", Swig_var_CS_BLOCKSIZE_get, Swig_var_CS_BLOCKSIZE_set);
  SWIG_Python_SetConstant(d, "Csock_OUTBOUND",SWIG_From_int(static_cast< int >(Csock::OUTBOUND)));
  SWIG_Python_SetConstant(d, "Csock_LISTENER",SWIG_From_int(static_cast< int >(Csock::LISTENER)));
  SWIG_Python_SetConstant(d, "Csock_INBOUND",SWIG_From_int(static_cast< int >(Csock::INBOUND)));
  SWIG_Python_SetConstant(d, "Csock_READ_EOF",SWIG_From_int(static_cast< int >(Csock::READ_EOF)));
  SWIG_Python_SetConstant(d, "Csock_READ_ERR",SWIG_From_int(static_cast< int >(Csock::READ_ERR)));
  SWIG_Python_SetConstant(d, "Csock_READ_EAGAIN",SWIG_From_int(static_cast< int >(Csock::READ_EAGAIN)));
  SWIG_Python_SetConstant(d, "Csock_READ_CONNREFUSED",SWIG_From_int(static_cast< int >(Csock::READ_CONNREFUSED)));
  SWIG_Python_SetConstant(d, "Csock_READ_TIMEDOUT",SWIG_From_int(static_cast< int >(Csock::READ_TIMEDOUT)));
  SWIG_Python_SetConstant(d, "Csock_SEL_OK",SWIG_From_int(static_cast< int >(Csock::SEL_OK)));
  SWIG_Python_SetConstant(d, "Csock_SEL_TIMEOUT",SWIG_From_int(static_cast< int >(Csock::SEL_TIMEOUT)));
  SWIG_Python_SetConstant(d, "Csock_SEL_EAGAIN",SWIG_From_int(static_cast< int >(Csock::SEL_EAGAIN)));
  SWIG_Python_SetConstant(d, "Csock_SEL_ERR",SWIG_From_int(static_cast< int >(Csock::SEL_ERR)));
  SWIG_Python_SetConstant(d, "Csock_TLS",SWIG_From_int(static_cast< int >(Csock::TLS)));
  SWIG_Python_SetConstant(d, "Csock_SSL23",SWIG_From_int(static_cast< int >(Csock::SSL23)));
  SWIG_Python_SetConstant(d, "Csock_SSL2",SWIG_From_int(static_cast< int >(Csock::SSL2)));
  SWIG_Python_SetConstant(d, "Csock_SSL3",SWIG_From_int(static_cast< int >(Csock::SSL3)));
  SWIG_Python_SetConstant(d, "Csock_TLS1",SWIG_From_int(static_cast< int >(Csock::TLS1)));
  SWIG_Python_SetConstant(d, "Csock_TLS11",SWIG_From_int(static_cast< int >(Csock::TLS11)));
  SWIG_Python_SetConstant(d, "Csock_TLS12",SWIG_From_int(static_cast< int >(Csock::TLS12)));
  SWIG_Python_SetConstant(d, "Csock_EDP_None",SWIG_From_int(static_cast< int >(Csock::EDP_None)));
  SWIG_Python_SetConstant(d, "Csock_EDP_SSLv2",SWIG_From_int(static_cast< int >(Csock::EDP_SSLv2)));
  SWIG_Python_SetConstant(d, "Csock_EDP_SSLv3",SWIG_From_int(static_cast< int >(Csock::EDP_SSLv3)));
  SWIG_Python_SetConstant(d, "Csock_EDP_TLSv1",SWIG_From_int(static_cast< int >(Csock::EDP_TLSv1)));
  SWIG_Python_SetConstant(d, "Csock_EDP_TLSv1_1",SWIG_From_int(static_cast< int >(Csock::EDP_TLSv1_1)));
  SWIG_Python_SetConstant(d, "Csock_EDP_TLSv1_2",SWIG_From_int(static_cast< int >(Csock::EDP_TLSv1_2)));
  SWIG_Python_SetConstant(d, "Csock_EDP_SSL",SWIG_From_int(static_cast< int >(Csock::EDP_SSL)));
  SWIG_Python_SetConstant(d, "Csock_CST_START",SWIG_From_int(static_cast< int >(Csock::CST_START)));
  SWIG_Python_SetConstant(d, "Csock_CST_DNS",SWIG_From_int(static_cast< int >(Csock::CST_DNS)));
  SWIG_Python_SetConstant(d, "Csock_CST_BINDVHOST",SWIG_From_int(static_cast< int >(Csock::CST_BINDVHOST)));
  SWIG_Python_SetConstant(d, "Csock_CST_DESTDNS",SWIG_From_int(static_cast< int >(Csock::CST_DESTDNS)));
  SWIG_Python_SetConstant(d, "Csock_CST_CONNECT",SWIG_From_int(static_cast< int >(Csock::CST_CONNECT)));
  SWIG_Python_SetConstant(d, "Csock_CST_CONNECTWAIT",SWIG_From_int(static_cast< int >(Csock::CST_CONNECTWAIT)));
  SWIG_Python_SetConstant(d, "Csock_CST_CONNECTSSL",SWIG_From_int(static_cast< int >(Csock::CST_CONNECTSSL)));
  SWIG_Python_SetConstant(d, "Csock_CST_OK",SWIG_From_int(static_cast< int >(Csock::CST_OK)));
  SWIG_Python_SetConstant(d, "Csock_CLT_DONT",SWIG_From_int(static_cast< int >(Csock::CLT_DONT)));
  SWIG_Python_SetConstant(d, "Csock_CLT_NOW",SWIG_From_int(static_cast< int >(Csock::CLT_NOW)));
  SWIG_Python_SetConstant(d, "Csock_CLT_AFTERWRITE",SWIG_From_int(static_cast< int >(Csock::CLT_AFTERWRITE)));
  SWIG_Python_SetConstant(d, "Csock_CLT_DEREFERENCE",SWIG_From_int(static_cast< int >(Csock::CLT_DEREFERENCE)));
  SWIG_Python_SetConstant(d, "Csock_TMO_READ",SWIG_From_int(static_cast< int >(Csock::TMO_READ)));
  SWIG_Python_SetConstant(d, "Csock_TMO_WRITE",SWIG_From_int(static_cast< int >(Csock::TMO_WRITE)));
  SWIG_Python_SetConstant(d, "Csock_TMO_ACCEPT",SWIG_From_int(static_cast< int >(Csock::TMO_ACCEPT)));
  SWIG_Python_SetConstant(d, "Csock_TMO_ALL",SWIG_From_int(static_cast< int >(Csock::TMO_ALL)));
  SWIG_Python_SetConstant(d, "Csock_DNS_VHOST",SWIG_From_int(static_cast< int >(Csock::DNS_VHOST)));
  SWIG_Python_SetConstant(d, "Csock_DNS_DEST",SWIG_From_int(static_cast< int >(Csock::DNS_DEST)));
  SWIG_Python_SetConstant(d, "CSocketManager_SUCCESS",SWIG_From_int(static_cast< int >(CSocketManager::SUCCESS)));
  SWIG_Python_SetConstant(d, "CSocketManager_SELECT_ERROR",SWIG_From_int(static_cast< int >(CSocketManager::SELECT_ERROR)));
  SWIG_Python_SetConstant(d, "CSocketManager_SELECT_TIMEOUT",SWIG_From_int(static_cast< int >(CSocketManager::SELECT_TIMEOUT)));
  SWIG_Python_SetConstant(d, "CSocketManager_SELECT_TRYAGAIN",SWIG_From_int(static_cast< int >(CSocketManager::SELECT_TRYAGAIN)));
  SWIG_Python_SetConstant(d, "CSocketManager_ECT_Read",SWIG_From_int(static_cast< int >(CSocketManager::ECT_Read)));
  SWIG_Python_SetConstant(d, "CSocketManager_ECT_Write",SWIG_From_int(static_cast< int >(CSocketManager::ECT_Write)));
  SWIG_Python_SetConstant(d, "ADDR_IPV4ONLY",SWIG_From_int(static_cast< int >(ADDR_IPV4ONLY)));
  SWIG_Python_SetConstant(d, "ADDR_IPV6ONLY",SWIG_From_int(static_cast< int >(ADDR_IPV6ONLY)));
  SWIG_Python_SetConstant(d, "ADDR_ALL",SWIG_From_int(static_cast< int >(ADDR_ALL)));
  SWIG_Python_SetConstant(d, "CFile_FT_REGULAR",SWIG_From_int(static_cast< int >(CFile::FT_REGULAR)));
  SWIG_Python_SetConstant(d, "CFile_FT_DIRECTORY",SWIG_From_int(static_cast< int >(CFile::FT_DIRECTORY)));
  SWIG_Python_SetConstant(d, "CFile_FT_CHARACTER",SWIG_From_int(static_cast< int >(CFile::FT_CHARACTER)));
  SWIG_Python_SetConstant(d, "CFile_FT_BLOCK",SWIG_From_int(static_cast< int >(CFile::FT_BLOCK)));
  SWIG_Python_SetConstant(d, "CFile_FT_FIFO",SWIG_From_int(static_cast< int >(CFile::FT_FIFO)));
  SWIG_Python_SetConstant(d, "CFile_FT_LINK",SWIG_From_int(static_cast< int >(CFile::FT_LINK)));
  SWIG_Python_SetConstant(d, "CFile_FT_SOCK",SWIG_From_int(static_cast< int >(CFile::FT_SOCK)));
  SWIG_Python_SetConstant(d, "CFile_FA_Name",SWIG_From_int(static_cast< int >(CFile::FA_Name)));
  SWIG_Python_SetConstant(d, "CFile_FA_Size",SWIG_From_int(static_cast< int >(CFile::FA_Size)));
  SWIG_Python_SetConstant(d, "CFile_FA_ATime",SWIG_From_int(static_cast< int >(CFile::FA_ATime)));
  SWIG_Python_SetConstant(d, "CFile_FA_MTime",SWIG_From_int(static_cast< int >(CFile::FA_MTime)));
  SWIG_Python_SetConstant(d, "CFile_FA_CTime",SWIG_From_int(static_cast< int >(CFile::FA_CTime)));
  SWIG_Python_SetConstant(d, "CFile_FA_UID",SWIG_From_int(static_cast< int >(CFile::FA_UID)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Unknown",SWIG_From_int(static_cast< int >(CMessage::Type::Unknown)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Account",SWIG_From_int(static_cast< int >(CMessage::Type::Account)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Action",SWIG_From_int(static_cast< int >(CMessage::Type::Action)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Authenticate",SWIG_From_int(static_cast< int >(CMessage::Type::Authenticate)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Away",SWIG_From_int(static_cast< int >(CMessage::Type::Away)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Capability",SWIG_From_int(static_cast< int >(CMessage::Type::Capability)));
  SWIG_Python_SetConstant(d, "CMessage_Type_ChgHost",SWIG_From_int(static_cast< int >(CMessage::Type::ChgHost)));
  SWIG_Python_SetConstant(d, "CMessage_Type_CTCP",SWIG_From_int(static_cast< int >(CMessage::Type::CTCP)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Error",SWIG_From_int(static_cast< int >(CMessage::Type::Error)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Invite",SWIG_From_int(static_cast< int >(CMessage::Type::Invite)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Join",SWIG_From_int(static_cast< int >(CMessage::Type::Join)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Kick",SWIG_From_int(static_cast< int >(CMessage::Type::Kick)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Mode",SWIG_From_int(static_cast< int >(CMessage::Type::Mode)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Nick",SWIG_From_int(static_cast< int >(CMessage::Type::Nick)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Notice",SWIG_From_int(static_cast< int >(CMessage::Type::Notice)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Numeric",SWIG_From_int(static_cast< int >(CMessage::Type::Numeric)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Part",SWIG_From_int(static_cast< int >(CMessage::Type::Part)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Ping",SWIG_From_int(static_cast< int >(CMessage::Type::Ping)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Pong",SWIG_From_int(static_cast< int >(CMessage::Type::Pong)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Quit",SWIG_From_int(static_cast< int >(CMessage::Type::Quit)));
  SWIG_Python_SetConstant(d, "CMessage_Type_TagMsg",SWIG_From_int(static_cast< int >(CMessage::Type::TagMsg)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Text",SWIG_From_int(static_cast< int >(CMessage::Type::Text)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Topic",SWIG_From_int(static_cast< int >(CMessage::Type::Topic)));
  SWIG_Python_SetConstant(d, "CMessage_Type_Wallops",SWIG_From_int(static_cast< int >(CMessage::Type::Wallops)));
  SWIG_Python_SetConstant(d, "CMessage_IncludeAll",SWIG_From_int(static_cast< int >(CMessage::IncludeAll)));
  SWIG_Python_SetConstant(d, "CMessage_ExcludePrefix",SWIG_From_int(static_cast< int >(CMessage::ExcludePrefix)));
  SWIG_Python_SetConstant(d, "CMessage_ExcludeTags",SWIG_From_int(static_cast< int >(CMessage::ExcludeTags)));
  SWIG_Python_SetConstant(d, "CModInfo_GlobalModule",SWIG_From_int(static_cast< int >(CModInfo::GlobalModule)));
  SWIG_Python_SetConstant(d, "CModInfo_UserModule",SWIG_From_int(static_cast< int >(CModInfo::UserModule)));
  SWIG_Python_SetConstant(d, "CModInfo_NetworkModule",SWIG_From_int(static_cast< int >(CModInfo::NetworkModule)));
  SWIG_Python_SetConstant(d, "CModule_CONTINUE",SWIG_From_int(static_cast< int >(CModule::CONTINUE)));
  SWIG_Python_SetConstant(d, "CModule_HALT",SWIG_From_int(static_cast< int >(CModule::HALT)));
  SWIG_Python_SetConstant(d, "CModule_HALTMODS",SWIG_From_int(static_cast< int >(CModule::HALTMODS)));
  SWIG_Python_SetConstant(d, "CModule_HALTCORE",SWIG_From_int(static_cast< int >(CModule::HALTCORE)));
  SWIG_Python_SetConstant(d, "CModule_UNLOAD",SWIG_From_int(static_cast< int >(CModule::UNLOAD)));
  SWIG_Python_SetConstant(d, "CChan_Voice",SWIG_From_char(static_cast< char >(CChan::Voice)));
  SWIG_Python_SetConstant(d, "CChan_HalfOp",SWIG_From_char(static_cast< char >(CChan::HalfOp)));
  SWIG_Python_SetConstant(d, "CChan_Op",SWIG_From_char(static_cast< char >(CChan::Op)));
  SWIG_Python_SetConstant(d, "CChan_Admin",SWIG_From_char(static_cast< char >(CChan::Admin)));
  SWIG_Python_SetConstant(d, "CChan_Owner",SWIG_From_char(static_cast< char >(CChan::Owner)));
  SWIG_Python_SetConstant(d, "CChan_M_Private",SWIG_From_char(static_cast< char >(CChan::M_Private)));
  SWIG_Python_SetConstant(d, "CChan_M_Secret",SWIG_From_char(static_cast< char >(CChan::M_Secret)));
  SWIG_Python_SetConstant(d, "CChan_M_Moderated",SWIG_From_char(static_cast< char >(CChan::M_Moderated)));
  SWIG_Python_SetConstant(d, "CChan_M_InviteOnly",SWIG_From_char(static_cast< char >(CChan::M_InviteOnly)));
  SWIG_Python_SetConstant(d, "CChan_M_NoMessages",SWIG_From_char(static_cast< char >(CChan::M_NoMessages)));
  SWIG_Python_SetConstant(d, "CChan_M_OpTopic",SWIG_From_char(static_cast< char >(CChan::M_OpTopic)));
  SWIG_Python_SetConstant(d, "CChan_M_Limit",SWIG_From_char(static_cast< char >(CChan::M_Limit)));
  SWIG_Python_SetConstant(d, "CChan_M_Key",SWIG_From_char(static_cast< char >(CChan::M_Key)));
  SWIG_Python_SetConstant(d, "CChan_M_Op",SWIG_From_char(static_cast< char >(CChan::M_Op)));
  SWIG_Python_SetConstant(d, "CChan_M_Voice",SWIG_From_char(static_cast< char >(CChan::M_Voice)));
  SWIG_Python_SetConstant(d, "CChan_M_Ban",SWIG_From_char(static_cast< char >(CChan::M_Ban)));
  SWIG_Python_SetConstant(d, "CChan_M_Except",SWIG_From_char(static_cast< char >(CChan::M_Except)));
  SWIG_Python_SetConstant(d, "CUser_HASH_NONE",SWIG_From_int(static_cast< int >(CUser::HASH_NONE)));
  SWIG_Python_SetConstant(d, "CUser_HASH_MD5",SWIG_From_int(static_cast< int >(CUser::HASH_MD5)));
  SWIG_Python_SetConstant(d, "CUser_HASH_SHA256",SWIG_From_int(static_cast< int >(CUser::HASH_SHA256)));
  SWIG_Python_SetConstant(d, "CUser_HASH_ARGON2ID",SWIG_From_int(static_cast< int >(CUser::HASH_ARGON2ID)));
  SWIG_Python_SetConstant(d, "CUser_HASH_DEFAULT",SWIG_From_int(static_cast< int >(CUser::HASH_DEFAULT)));
  SWIG_Python_SetConstant(d, "CIRCSock_ListArg",SWIG_From_int(static_cast< int >(CIRCSock::ListArg)));
  SWIG_Python_SetConstant(d, "CIRCSock_HasArg",SWIG_From_int(static_cast< int >(CIRCSock::HasArg)));
  SWIG_Python_SetConstant(d, "CIRCSock_ArgWhenSet",SWIG_From_int(static_cast< int >(CIRCSock::ArgWhenSet)));
  SWIG_Python_SetConstant(d, "CIRCSock_NoArg",SWIG_From_int(static_cast< int >(CIRCSock::NoArg)));
  SWIG_Python_SetConstant(d, "CListener_ACCEPT_IRC",SWIG_From_int(static_cast< int >(CListener::ACCEPT_IRC)));
  SWIG_Python_SetConstant(d, "CListener_ACCEPT_HTTP",SWIG_From_int(static_cast< int >(CListener::ACCEPT_HTTP)));
  SWIG_Python_SetConstant(d, "CListener_ACCEPT_ALL",SWIG_From_int(static_cast< int >(CListener::ACCEPT_ALL)));
  SWIG_Python_SetConstant(d, "CWebSubPage_F_ADMIN",SWIG_From_int(static_cast< int >(CWebSubPage::F_ADMIN)));
  SWIG_Python_SetConstant(d, "CWebSock_PAGE_NOTFOUND",SWIG_From_int(static_cast< int >(CWebSock::PAGE_NOTFOUND)));
  SWIG_Python_SetConstant(d, "CWebSock_PAGE_PRINT",SWIG_From_int(static_cast< int >(CWebSock::PAGE_PRINT)));
  SWIG_Python_SetConstant(d, "CWebSock_PAGE_DEFERRED",SWIG_From_int(static_cast< int >(CWebSock::PAGE_DEFERRED)));
  SWIG_Python_SetConstant(d, "CWebSock_PAGE_DONE",SWIG_From_int(static_cast< int >(CWebSock::PAGE_DONE)));
  SWIG_Python_SetConstant(d, "CZNC_ECONFIG_NOTHING",SWIG_From_int(static_cast< int >(CZNC::ECONFIG_NOTHING)));
  SWIG_Python_SetConstant(d, "CZNC_ECONFIG_NEED_REHASH",SWIG_From_int(static_cast< int >(CZNC::ECONFIG_NEED_REHASH)));
  SWIG_Python_SetConstant(d, "CZNC_ECONFIG_NEED_WRITE",SWIG_From_int(static_cast< int >(CZNC::ECONFIG_NEED_WRITE)));
  SWIG_Python_SetConstant(d, "CZNC_ECONFIG_NEED_VERBOSE_WRITE",SWIG_From_int(static_cast< int >(CZNC::ECONFIG_NEED_VERBOSE_WRITE)));
  SWIG_Python_SetConstant(d, "CZNC_ECONFIG_DELAYED_WRITE",SWIG_From_int(static_cast< int >(CZNC::ECONFIG_DELAYED_WRITE)));
  SWIG_Python_SetConstant(d, "CZNC_ECONFIG_NEED_QUIT",SWIG_From_int(static_cast< int >(CZNC::ECONFIG_NEED_QUIT)));
#if PY_VERSION_HEX >= 0x03000000
  return m;
#else
  return;
#endif
}

